(* Theorema 
    Copyright (C) 2010 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

(*
	This package is largely inspired by
	[1]	http://mathematica.stackexchange.com/questions/838/functional-style-using-lazy-lists
*)

BeginPackage["LazyList`"]

With[ {lng = "English"},
	MessageName[ LL, "usage", lng] = "LL is a data structure which generalizes lists in Mathematica to \"lazy lists\". LL combines lazy lists and other expressions, e.g. standard lists.";
	
	MessageName[ lAppend, "usage", lng] = "";
	MessageName[ lCases, "usage", lng] = "";
	MessageName[ lDeleteDuplicates, "usage", lng] = "";
	MessageName[ lDrop, "usage", lng] = "";
	MessageName[ lEmptyQ, "usage", lng] = "";
	MessageName[ lFirst, "usage", lng] = "";
	MessageName[ lFlatten, "usage", lng] = "";
	MessageName[ lFlattenTranspose, "usage", lng] = "";
	MessageName[ lFromFunction, "usage", lng] = "";
	MessageName[ lFromList, "usage", lng] = "";
	MessageName[ lGetFirst, "usage", lng] = "";
	MessageName[ lInterleave, "usage", lng] = "";
	MessageName[ lJoin, "usage", lng] = "";
	MessageName[ lMap, "usage", lng] = "";
	MessageName[ lMapIndexed, "usage", lng] = "";
	MessageName[ lMemberQ, "usage", lng] = "";
	MessageName[ lMerge, "usage", lng] = "";
	MessageName[ lPrepend, "usage", lng] = "";
	MessageName[ lRest, "usage", lng] = "";
	MessageName[ lScan, "usage", lng] = "";
	MessageName[ lSelect, "usage", lng] = "";
	MessageName[ lTake, "usage", lng] = "";
	MessageName[ lToList, "usage", lng] = "";
	MessageName[ lTrim, "usage", lng] = "";
	
	MessageName[ clearIterator, "usage", lng] = "";
	MessageName[ makeIterator, "usage", lng] = "";
	MessageName[ makeMemoization, "usage", lng] = "";
]

(*  The names of all the functions below have prefix "l";
	although it would of course be possible to overload existing functions ('First', 'Rest', etc.),
	using fresh names is better for debugging, where it is absoluetly clear now where a lazy list may potentially appear. *)

Begin["`Private`"]


(* ::Section:: *)
(* Public Functions*)

SetAttributes[ LL, HoldAll];

lEmptyQ[ e_] := MatchQ[ e, _[]]

SetAttributes[ lFirst, HoldRest];
lFirst[ l_] := lFirst[ l, $Failed, Identity]
lFirst[ l_, def_] := lFirst[ l, def, Identity]
lFirst[ _[], def_, h_] := h[ def]
lFirst[ _[ f_, ___], _, h_] := h[ f]

SetAttributes[ lRest, HoldRest];
lRest[ expr_] := lRest[ expr, $Failed]
lRest[ _[], def_] := def
lRest[ LL[ _, t_], _] := t
lRest[ expr:Except[ _LL], _] := Rest[ expr]

(* The third argument of 'lMap' and 'lMapIndexed' specifies whether the given function shall immediately be evaluated on the first argument of the lazy list.
	The fourth argument specifies how to treat standard lists: just use built-in 'Map' resp. 'MapIndexed' ('True'),
	or create a lazy list where the given function is only applied to the first element ('False'). *)
lMap[ fn_, expr_] := lMap[ fn, expr, True, False]
lMap[ fn_, expr_, eval_] := lMap[ fn, expr, eval, False]
lMap[ _, _[], _, False] := LL[]
lMap[ _, h_[], _, True] := h[]
lMap[ fn_, LL[ f_, lMap[ fn1_, t_, False, eager2_]], False, eager1_] :=
	With[ {newFn = Composition[ fn, fn1], newEager = TrueQ[ eager1 && eager2]},
		LL[ fn[ f], lMap[ newFn, t, False, newEager]]
	]
lMap[ fn_, LL[ f_, t_], False, eager_] := LL[ fn[ f], lMap[ fn, t, False, eager]]
lMap[ fn_, LL[ f_, lMap[ fn1_, t_, _, eager2_]], True, eager1_] :=
	With[ {new = fn[ f], newFn = Composition[ fn, fn1], newEager = TrueQ[ eager1 && eager2]},
		LL[ new, lMap[ newFn, t, True, newEager]]
	]
lMap[ fn_, LL[ f_, t_], True, eager_] := With[ {new = fn[ f]}, LL[ new, lMap[ fn, t, True, eager]]]
lMap[ fn_, (h:Except[ LL])[ f_, r___], False, False] := LL[ fn[ f], lMap[ fn, h[ r], False, False]]
lMap[ fn_, (h:Except[ LL])[ f_, r___], True, False] := With[ {new = fn[ f]}, LL[ new, lMap[ fn, h[ r], True, False]]]
lMap[ fn_, expr:Except[ _LL], _, True] := Map[ fn, expr]

(* In contrast to buil-in 'MapIndexed', the second argument that is passed to the given function is just a plain integer, not a singleton list. *)
lMapIndexed[ fn_, expr_] := lMapIndexed[ fn, expr, True, False, 1]
lMapIndexed[ fn_, expr_, eval_] := lMapIndexed[ fn, expr, eval, False, 1]
lMapIndexed[ _, _[], _, False, _] := LL[]
lMapIndexed[ _, h_[], _, True, _] := h[]
lMapIndexed[ fn_, LL[ f_, t_], False, eager_, i_] := With[ {newI = i + 1}, LL[ fn[ f, i], lMapIndexed[ fn, t, False, eager, newI]]]
lMapIndexed[ fn_, LL[ f_, t_], True, eager_, i_] := With[ {new = fn[ f, i], newI = i + 1}, LL[ new, lMapIndexed[ fn, t, True, eager, newI]]]
lMapIndexed[ fn_, (h:Except[ LL])[ f_, r___], False, False, i_] := With[ {newI = i + 1}, LL[ fn[ f, i], lMapIndexed[ fn, h[ r], False, False, newI]]]
lMapIndexed[ fn_, (h:Except[ LL])[ f_, r___], True, False, i_] := With[ {new = fn[ f, i], newI = i + 1}, LL[ new, lMapIndexed[ fn, h[ r], True, False, newI]]]
lMapIndexed[ fn_, expr:Except[ _LL], _, True, i_] := With[ {offset = i - 1}, MapIndexed[ fn[ #1, offset + First[ #2]]&, expr]]

(* The fourth argument specifies whether the function shall be "safe", i.e. only scan elements that are
	already known and therefore not run into an infinite loop, or not. *)
lScan[ fn_, expr_] := lScan[ fn, expr, Identity, True]
lScan[ fn_, expr_, h_] := lScan[ fn, expr, h, True]
lScan[ _, _[], _, _] := Null
lScan[ fn_, LL[ f_, t_], h_, False] := (fn[ h[ f]]; lScan[ fn, t, h])
lScan[ fn_, LL[ f_, t:(_List|_Hold|_HoldComplete|_LL)], h_, True] := (fn[ h[ f]]; lScan[ fn, t, h])
lScan[ fn_, LL[ f_, _], h_, True] := (fn[ h[ f]]; Null)
lScan[ fn_, expr:Except[ _LL], Identity, _] := Scan[ fn, expr]
lScan[ fn_, Except[ LL][ f_, r___], h_, _] := (fn[ h[ f]]; lScan[ fn, Hold[ r], h, True])

(* The third argument of 'lSelect' specifies whether 'Hold' shall be wrapped around the elements of the list before the predicate is called on them.
	The fourth argument specifies how to treat standard lists: just use built-in 'Select' ('True'),
	or create a lazy list where the given predicate is only applied to the initial elements until the first gives 'True' ('False'). *)
lSelect[ expr_, p_] := lSelect[ expr, p, False, False]
lSelect[ expr_, p_, hold_] := lSelect[ expr, p, hold, False]
lSelect[ _[], _, _, False] := LL[]
lSelect[ h_[], _, _, True] := h[]
lSelect[ LL[ f_, lSelect[ t_, p2_, False, eager2_]], p1_, False, eager1_] :=
	With[ {newP = conjPredicate[ p2, p1], newEager = TrueQ[ eager1 && eager2]},
		If[ TrueQ[ p1[ f]],
			LL[ f, lSelect[ t, newP, False, newEager]],
		(*else*)
			lSelect[ t, newP, False, newEager]
		]
	]
lSelect[ LL[ f_, t_], p_, False, eager_] :=
	If[ TrueQ[ p[ f]],
		LL[ f, lSelect[ t, p, False, eager]],
	(*else*)
		lSelect[ t, p, False, eager]
	]
lSelect[ LL[ f_, lSelect[ t_, p2_, True, eager2_]], p1_, True, eager1_] :=
	With[ {newP = conjPredicate[ p2, p1], newEager = TrueQ[ eager1 && eager2]},
		If[ TrueQ[ p1[ Hold[ f]]],
			LL[ f, lSelect[ t, newP, True, newEager]],
		(*else*)
			lSelect[ t, newP, True, newEager]
		]
	]
lSelect[ LL[ f_, t_], p_, True, eager_] :=
	If[ TrueQ[ p[ Hold[ f]]],
		LL[ f, lSelect[ t, p, True, eager]],
	(*else*)
		lSelect[ t, p, True, eager]
	]
lSelect[ (h:Except[ LL])[ f_, r___], p_, False, False] :=
	If[ TrueQ[ p[ f]],
		LL[ f, lSelect[ h[ r], p, False, False]],
	(*else*)
		lSelect[ h[ r], p, False, False]
	]
lSelect[ (h:Except[ LL])[ f_, r___], p_, True, False] :=
	If[ TrueQ[ p[ Hold[ f]]],
		LL[ f, lSelect[ h[ r], p, True, False]],
	(*else*)
		lSelect[ h[ r], p, True, False]
	]
lSelect[ expr:Except[ _LL], p_, False, True] := Select[ expr, p]
lSelect[ expr:Except[ _LL], p_, True, True] := Select[ expr, Function[ {x}, p[ Hold[ x]], {HoldFirst}]]

(* The following implementation of filtering elements from lazy lists, taken from [1], is a bit slower than 'lSelect'
	and does not work for combinations of lazy- and standard lists; however, it does not necessitate adjusting '$IterationLimit'. *)
(*
lFilter[ l_LL, pred_] := NestWhile[ lRest, l, (!lEmptyQ[ #] && !pred[ lFirst[ #]])&] /. LL[ h_, t_] :> LL[ h, lFilter[ t, pred]]
*)

lCases[ h_[], _] := h[]
lCases[ LL[ f_, t_], rl:(_Rule|_RuleDelayed)] :=
	With[ {fEval = f},
	With[ {new = Cases[ {fEval}, rl]},
		If[ new === {},
			lCases[ t, rl],
		(*else*)
			With[ {a = First[ new]},
				LL[ a, lCases[ t, rl]]
			]
		]
	]]
lCases[ LL[ f_, t_], p_] :=
	With[ {feval = f},
		If[ MatchQ[ feval, p],
			LL[ feval, lCases[ t, p]],
		(*else*)
			lCases[ t, p]
		]
	]
lCases[ expr:Except[ _LL], p_] := Cases[ expr, p]

(* 'lTrim[ expr, p]' removes the initial elements of 'expr' which satisfy 'p'. *)
lTrim[ expr_, p_] := lTrim[ expr, p, False, Infinity]
lTrim[ expr_, p_, hold_] := lTrim[ expr, p, hold, Infinity]
lTrim[ h_[], _, _, _] := h[]
lTrim[ expr_, _, _, _?NonPositive] := expr
lTrim[ LL[ f_, t_], p_, False, bound_] :=
	With[ {fEval = f},
		If[ p[ f] === False,
			LL[ fEval, t],
		(*else*)
			lTrim[ t, p, False, bound - 1]
		]
	]
lTrim[ LL[ f_, t_], p_, True, bound_] :=
	If[ p[ Hold[ f]] === False,
		LL[ f, t],
	(*else*)
		lTrim[ t, p, True, bound - 1]
	]
lTrim[ expr:Except[ _LL], p_, False, bound_] :=
	With[ {res = Position[ expr, _?(!p[ #] &), {1}, 1]},
		If[ res === {}, Head[ expr][], Drop[ expr, res[[1, 1]] - 1]]
	]
lTrim[ expr:Except[ _LL], p_, True, bound_] :=
	With[ {res = Position[ expr, _?(Function[ {x}, !p[ Hold[ x]], {HoldFirst}]), {1}, 1]},
		If[ res === {}, Head[ expr][], Drop[ expr, res[[1, 1]] - 1]]
	]

(* The third argument is a symbol that is wrapped around every element before it is tested and returned.
	The fourth argument specifies whether the function shall be "safe", i.e. only look at elements that are
	already known and therefore not run into an infinite loop, or not. *)
lGetFirst[ expr_, p_] := lGetFirst[ expr, p, Identity, True, Indeterminate, $Failed]
lGetFirst[ expr_, p_, h_] := lGetFirst[ expr, p, h, True, Indeterminate, $Failed]
lGetFirst[ expr_, p_, h_, safe_] := lGetFirst[ expr, p, h, safe, Indeterminate, $Failed]
lGetFirst[ expr_, p_, h_, safe_, indet_] := lGetFirst[ expr, p, h, safe, indet, $Failed]
lGetFirst[ _[], _, h_, _, _, def_] := h[ def]
lGetFirst[ LL[ f_, t_], p_, h_, False, indet_, def_] :=
	If[ TrueQ[ p[ h[ f]]],
		h[ f],
	(*else*)
		lGetFirst[ t, p, h, False, indet, def]
	]
lGetFirst[ LL[ f_, t:(_List|_Hold|_HoldComplete|_LL)], p_, h_, True, indet_, def_] :=
	If[ TrueQ[ p[ h[ f]]],
		h[ f],
	(*else*)
		lGetFirst[ t, p, h, True, indet, def]
	]
lGetFirst[ LL[ f_, t_], p_, h_, True, indet_, def_] :=
	If[ TrueQ[ p[ h[ f]]],
		h[ f],
	(*else*)
		h[ indet]
	]
lGetFirst[ expr:Except[ _LL], p_, h_, _, _, def_] :=
	With[ {res = Select[ expr, Function[ {x}, p[ h[ x]], {HoldFirst}], 1]},
		If[ res === {}, h[ def], First[ res]]
	]

lMemberQ[ expr_, p_] := lMemberQ[ expr, p, True, Indeterminate]
lMemberQ[ expr_, p_, safe_] := lMemberQ[ expr, p, safe, Indeterminate]
lMemberQ[ _[], _, _, _] := False
lMemberQ[ LL[ f_, t_], p_, False, indet_] :=
	If[ MatchQ[ f, p],
		True,
	(*else*)
		lMemberQ[ t, p, False, indet]
	]
lMemberQ[ LL[ f_, t:(_List|_Hold|_HoldComplete)], p_, True, indet_] :=
	If[ MatchQ[ f, p],
		True,
	(*else*)
		lMemberQ[ t, p, True, indet]
	]
lMemberQ[ LL[ f_, t_], p_, True, indet_] :=
	If[ MatchQ[ f, p],
		True,
	(*else*)
		indet
	]
lMemberQ[ expr:Except[ _LL], p_, _, _] := MemberQ[ expr, p]

(* 'lTake[ list, n]' takes *up to* 'n' elements from 'expr', i.e. it does not throw an error if there are not enough elements. *)
lTake[ h_[ ___], 0] := h[]
lTake[ h_[], _] := h[]
lTake[ LL[ f_, t_], n_Integer?Positive] := With[ {nn = n - 1}, LL[ f, lTake[ t, nn]]]
lTake[ expr:Except[ _LL], n_Integer] := If[ Length[ expr] >= n, Take[ expr, n], expr]

(* 'lDrop' drops *at most* 'n' elements, but does not throw an error if there are not enough elements. *)
lDrop[ expr_, 0] := expr
lDrop[ h_[], _] := h[]
lDrop[ LL[ _, t_], n_Integer?Positive] := lDrop[ t, n - 1]
lDrop[ expr:Except[ _LL], n_Integer?Positive] := If[ Length[ expr] >= n, Drop[ expr, n], Head[ expr][]]

(* The last argument of 'lJoin', 'lAppend', 'lInterleave' and 'lPrepend' specifies how to treat its second
	(resp. first) argument if it is inside 'Hold' (or 'HoldComplete'):
	if 'True', the enclosing 'Hold' of the second argument is considered only a wrapper to prevent evaluation,
	and *not* the "true" head of it. Default is 'False'. *)
lJoin[ a_, b_] := lJoin[ a, b, False]
lJoin[ a_, b:Except[ _Hold|_HoldComplete], True] := lJoin[ a, b, False]
lJoin[ expr_, _[], False] := expr
lJoin[ _[], expr_, False] := expr
lJoin[ expr_, _[ _[]], True] := expr
lJoin[ _[], _[ expr_], True] := expr
lJoin[ LL[ f_, lAppend[ t_, expr1_, False]], expr_, hold_] := With[ {new = lPrepend[ expr, expr1, hold]}, LL[ f, lJoin[ t, new, False]]]
lJoin[ LL[ f_, lJoin[ t_, expr1_, False]], expr_, hold_] := With[ {new = lJoin[ expr1, expr, hold]}, LL[ f, lJoin[ t, new, False]]]
lJoin[ LL[ f_, t_], expr_, hold_] := LL[ f, lJoin[ t, expr, hold]]
lJoin[ a:(Except[ _LL]), b:LL[ _, _], False] := Fold[ Function[ {x, y}, LL[ y, x], {HoldAll}], b, Reverse[ a]]
lJoin[ a:(Except[ _LL]), b:(Except[ _LL]), False] := Join[ a, b]
lJoin[ (h:Except[ LL])[ f_, r___], b_, True] := LL[ f, lJoin[ h[ r], b, True]]

lAppend[ a_, b_] := lAppend[ a, b, False]
lAppend[ a_, b:Except[ _Hold|_HoldComplete], True] := lAppend[ a, b, False]
lAppend[ LL[], expr_, False] := LL[ expr, LL[]]
lAppend[ LL[], _[ expr_], True] := LL[ expr, LL[]]
lAppend[ LL[ f_, lAppend[ t_, expr1_, False]], expr_, False] := LL[ f, lJoin[ t, {expr1, expr}, False]]
lAppend[ LL[ f_, lJoin[ t_, expr1_, False]], expr_, hold_] := With[ {new = lAppend[ expr1, expr, hold]}, LL[ f, lJoin[ t, new, False]]]
lAppend[ LL[ f_, t_], expr_, hold_] := LL[ f, lAppend[ t, expr, hold]]
lAppend[ a:(Except[ _LL]), expr_, False] := Append[ a, expr]
lAppend[ a:(Except[ _LL]), _[ expr_], True] := Append[ a, expr]

lPrepend[ a_, b_] := lPrepend[ a, b, False]
lPrepend[ a:Except[ _Hold|_HoldComplete], expr_, True] := lPrepend[ a, expr, False]
lPrepend[ a:(LL[]|LL[ _, _]), expr_, False] := LL[ expr, a]
lPrepend[ _[ a:(LL[]|LL[ _, _])], expr_, True] := LL[ expr, a]
lPrepend[ a:(Except[ _LL]), expr_, False] := Prepend[ a, expr]
lPrepend[ _[ a:(Except[ _LL])], expr_, True] := lPrepend[ a, expr, False]

lInterleave[ a_, b_] := lInterleave[ a, b, False]
lInterleave[ a_, b:Except[ _Hold|_HoldComplete], True] := lInterleave[ a, b, False]
lInterleave[ expr_, _[], False] := expr
lInterleave[ _[], expr_, False] := expr
lInterleave[ expr_, _[ _[]], True] := expr
lInterleave[ _[], _[ expr_], True] := expr
lInterleave[ LL[ f_, t_], expr_, False] := LL[ f, lInterleave[ expr, t, False]]
lInterleave[ LL[ f_, t_], _[ expr_], True] := LL[ f, lInterleave[ expr, t, False]]
lInterleave[ (h:Except[LL])[ f_, rest___], b:LL[ _, _], False] := LL[ f, lInterleave[ b, h[ rest], False]]
lInterleave[ (h:Except[LL])[ f_, rest___], _[ b:LL[ _, _]], True] := LL[ f, lInterleave[ b, h[ rest], False]]
lInterleave[ a:(Except[ _LL]), b:(Except[ _LL]), False] :=
	With[ {la = Length[ a], lb = Length[ b]},
		Which[
			lb + 1 < la,
			Join[ Riffle[ Take[ a, lb], b], Drop[ a, lb]],
			la < lb,
			Join[ Riffle[ a, Take[ b, la]], Drop[ b, la]],
			True,
			Riffle[ a, b]
		]
	]
lInterleave[ a:(Except[ _LL]), _[ b:(Except[ _LL])], True] := lInterleave[ a, b, False]

(* The third argument of 'lMerge' specifies whether 'Hold' shall be wrapped around the list-elements
	before they are compared using the given relation.
	'lMerge' is parameterized over the relation to be used for comparing elements, such that it may
	easily be used as an argument in 'lFlatten' (just like 'lJoin' etc.).
	Note that the result is *always* a lazy list with head 'LL'! *)
lMerge[ rel_][ a_, b_] := lMerge[ rel][ a, b, False]
lMerge[ _][ expr_, _[], _] := expr
lMerge[ _][ _[], expr_, _] := expr
lMerge[ rel_][ LL[ f1_, t1_], LL[ f2_, t2_], False] :=
	With[ {f1Eval = f1, f2Eval = f2},
		If[ TrueQ[ rel[ f1Eval, f2Eval]],
			LL[ f1Eval, lMerge[ rel][ t1, LL[ f2Eval, t2], False]],
		(*else*)
			LL[ f2Eval, lMerge[ rel][ LL[ f1Eval, t1], t2, False]]
		]
	]
lMerge[ rel_][ LL[ f1_, t_], (h:(Except[ LL]))[ f2_, r___], False] :=
	With[ {f1Eval = f1, f2Eval = f2},
		If[ TrueQ[ rel[ f1Eval, f2Eval]],
			LL[ f1Eval, lMerge[ rel][ t, h[ f2Eval, r], False]],
		(*else*)
			LL[ f2Eval, lMerge[ rel][ LL[ f1Eval, t], h[ r], False]]
		]
	]
lMerge[ rel_][ (h:Except[LL])[ f1_, r___], LL[ f2_, t_], False] :=
	With[ {f1Eval = f1, f2Eval = f2},
		If[ TrueQ[ rel[ f1Eval, f2Eval]],
			LL[ f1Eval, lMerge[ rel][ h[ r], LL[ f2Eval, t], False]],
		(*else*)
			LL[ f2Eval, lMerge[ rel][ h[ f1Eval, r], t, False]]
		]
	]
lMerge[ rel_][ (h1:Except[LL])[ f1_, r1___], (h2:Except[LL])[ f2_, r2___], False] :=
	With[ {f1Eval = f1, f2Eval = f2},
		If[ TrueQ[ rel[ f1Eval, f2Eval]],
			LL[ f1Eval, lMerge[ rel][ h1[ r1], h2[ f2Eval, r2], False]],
		(*else*)
			LL[ f2Eval, lMerge[ rel][ h1[ f1Eval, r1], h2[ r2], False]]
		]
	]
lMerge[ rel_][ LL[ f1_, t1_], LL[ f2_, t2_], True] :=
	If[ TrueQ[ rel[ Hold[ f1], Hold[ f2]]],
		LL[ f1, lMerge[ rel][ t1, LL[ f2, t2], True]],
	(*else*)
		LL[ f2, lMerge[ rel][ LL[ f1, t1], t2, True]]
	]
lMerge[ rel_][ LL[ f1_, t_], (h:(Except[ LL]))[ f2_, r___], True] :=
	If[ TrueQ[ rel[ Hold[ f1], Hold[ f2]]],
		LL[ f1, lMerge[ rel][ t, h[ f2, r], True]],
	(*else*)
		LL[ f2, lMerge[ rel][ LL[ f1, t], h[ r], True]]
	]
lMerge[ rel_][ (h:Except[LL])[ f1_, r___], LL[ f2_, t_], True] :=
	If[ TrueQ[ rel[ Hold[ f1], Hold[ f2]]],
		LL[ f1, lMerge[ rel][ h[ r], LL[ f2, t], True]],
	(*else*)
		LL[ f2, lMerge[ rel][ h[ f1, r], t, True]]
	]
lMerge[ rel_][ (h1:Except[LL])[ f1_, r1___], (h2:Except[LL])[ f2_, r2___], True] :=
	If[ TrueQ[ rel[ Hold[ f1], Hold[ f2]]],
		LL[ f1, lMerge[ rel][ h1[ r1], h2[ f2, r2], True]],
	(*else*)
		LL[ f2, lMerge[ rel][ h1[ f1, r1], h2[ r2], True]]
	]
	
lDeleteDuplicates[ expr_] := lDeleteDuplicates[ expr, SameQ, {}]
lDeleteDuplicates[ expr_, test_] := lDeleteDuplicates[ expr, test, {}]
lDeleteDuplicates[ h_[], _, _] := h[]
lDeleteDuplicates[ LL[ f_, lDeleteDuplicates[ t_, test_, found1_List]], test_, found_List] :=
	If[ isMember[ found, f, test],
		lDeleteDuplicates[ t, test, Join[ found1, found]],
	(*else*)
		With[ {new = Join[ found1, found, {f}]},
			LL[ f, lDeleteDuplicates[ t, test, new]]
		]
	]
lDeleteDuplicates[ LL[ f_, t_], test_, found_List] :=
	If[ isMember[ found, f, test],
		lDeleteDuplicates[ t, test, found],
	(*else*)
		With[ {new = Append[ found, f]},
			LL[ f, lDeleteDuplicates[ t, test, new]]
		]
	]
lDeleteDuplicates[ expr:Except[ _LL], test_, found_List] := DeleteDuplicates[ Select[ expr, !isMember[ found, #, test] &], test]

(* The second argument of 'lFlatten' is the "combinator" that is used to combine the sublists.
	'lInterleave' (default) is a fair combinator, whereas 'lJoin' is an unfair combinator.
	The third argument specifies whether the sublists shall combined in reversed order (default) or not.
		Example:
		'lFlatten[ {{a, b}, {c, d}}, lJoin, True]' --> '{a, c, b, d}'
		'lFlatten[ {{a, b}, {c, d}}, lJoin, False]' --> '{a, b, c, d}'
	Arguments 'lInterleave, True' yield the "fairest" way of flattening, arguments 'lJoin, False' resemble built-in 'Flatten'.
	Note that the result is *always* a lazy list with head 'LL'. *)
lFlatten[ expr_] := lFlatten[ expr, lInterleave, True]
lFlatten[ expr_, op_] := lFlatten[ expr, op, True]
lFlatten[ _[], _, _] := LL[]
lFlatten[ LL[ f_, t_], op_, True] :=
	With[ {fEval = f},
		If[ lEmptyQ[ fEval],
			lFlatten[ t, op, True],
		(*else*)
			With[ {new = Join[ Hold[ lFlatten[ t, op, True]], Hold @@ Rest[ fEval]]},
				Join[ lFirst[ fEval, $Failed, LL], op @@@ LL[ new]]
			]
		]
	]
lFlatten[ LL[ f_, t_], op_, False] :=
	With[ {fEval = f},
		If[ lEmptyQ[ fEval],
			lFlatten[ t, op, False],
		(*else*)
			With[ {new = Join[ Hold @@ Rest[ fEval], Hold[ lFlatten[ t, op, False]]]},
				Join[ lFirst[ fEval, $Failed, LL], op @@@ LL[ new]]
			]
		]
	]
lFlatten[ (h:Except[ LL])[ _[], r___], op_, rev_] := lFlatten[ h[ r], op, rev]
lFlatten[ (h:Except[ LL])[ LL[ f_, t_], r___], op_, True] := LL[ f, op[ lFlatten[ h[ r], op, True], t]]
lFlatten[ (h:Except[ LL])[ LL[ f_, t_], r___], op_, False] := LL[ f, op[ t, lFlatten[ h[ r], op, False]]]
lFlatten[ (h:Except[ LL])[ Except[ LL][ f_, r2___], r1___], op_, True] := LL[ f, op[ lFlatten[ h[ r1], op, True], h[ r2]]]
lFlatten[ (h:Except[ LL])[ Except[ LL][ f_, r2___], r1___], op_, False] := LL[ f, op[ h[ r2], lFlatten[ h[ r1], op, False]]]

(* The composition of transposition and flattening can be achieved much easier than transposition alone.
	As with 'lFlatten', the result *always* is a lazy list with head 'LL'. *)
lFlattenTranspose[ _[]] := LL[]
lFlattenTranspose[ LL[ f_, t_]] :=
	With[ {fEval = f},
		If[ lEmptyQ[ fEval],
			lFlattenTranspose[ t],
		(*else*)
			With[ {new = Join[ Hold[ t], Hold @@ Rest[ fEval]]},
				Join[ lFirst[ fEval, $Failed, LL], lFlattenTranspose /@ lAppend @@@ LL[ new]]
			]
		]
	]
lFlattenTranspose[ (h:Except[ LL])[ _[], r___]] := lFlattenTranspose[ h[ r]]
lFlattenTranspose[ (h:Except[ LL])[ LL[ f_, t_], r___]] := LL[ f, lFlattenTranspose[ h[ r, t]]]
lFlattenTranspose[ (h:Except[ LL])[ Except[ LL][ f_, r2___], r1___]] := LL[ f, lFlattenTranspose[ h[ r1, h[ r2]]]]

lToList[ expr_] := lToList[ expr, {}]
lToList[ expr_, h_] := lToList[ expr, h[]]
lToList[ LL[], out_] := out
lToList[ LL[ f_, t_], h_[ acc___]] := lToList[ t, h[ acc, f]]
lToList[ expr:Except[ _LL], acc:(h_[ ___])] := Join[ acc, h @@ expr]

lFromList[ _[]] := LL[]
lFromList[ LL[ f_, t:lFromList[ _]]] := LL[ f, t]
lFromList[ LL[ f_, t_]] := LL[ f, lFromList[ t]]
lFromList[ (h:Except[ LL])[ f_, r___]] := LL[ f, lFromList[ h[ r]]]

lFromFunction[ fn_] := lFromFunction[ fn, 1, False]
lFromFunction[ fn_, init_] := lFromFunction[ fn, init, False]
lFromFunction[ fn_, init_, False] := With[ {next = init + 1}, LL[ fn[ init], lFromFunction[ fn, next]]]
lFromFunction[ fn_, init_, True] := With[ {next = init + 1, f = fn[ init]}, LL[ f, lFromFunction[ fn, next]]]


(* ::Subsection:: *)
(*Auxiliary Functions*)

SetAttributes[ conjPredicate, {Flat, OneIdentity}];
conjPredicate[ preds___][ x_] := And @@ Map[ #[ x] &, {preds}]

isMember[ l_List, elem_, test_] :=
	Catch[
		Scan[
			If[ TrueQ[ test[ #, elem]], Throw[ True]]&,
			l
		];
		Throw[ False]
	]


(* ::Section:: *)
(*Iterators*)

(* Iterators heavily depend on lazy lists, and lazy lists act like a stack. Hence, in order to decrease the size of the lazy list a little,
	we assign the iterators below to individual symbols. This has the effect that instead of a compound expression like
	'generateFair[ fun, mem]' only a single symbol like 'fair$5715' appears (hundreds of times) in the lazy list. *)

SetAttributes[ clearIterator, HoldAll];
clearIterator[] := Null
clearIterator[ f_, r__] := (clearIterator[ f]; clearIterator[ r])
clearIterator[ f_] := Clear[ f]

makeIterator[ fun_, opts___?OptionQ] :=
	With[ {control = Replace[ "Control", {opts, _ -> "fair"}],
			mem = Replace[ "Memoization", {opts, _ -> None}],
			name = Replace[ "Name", {opts, _ :> Unique[ "iterator"]}],
			aux = Replace[ "NameAux", {opts, _ :> Unique[ "iterator$aux"]}]},
		Switch[ control,
			"fair",
			generateFair[ fun, mem, name],
			"depth-first",
			generateDF[ fun, mem, name],
			"breadth-first",
			generateBF[ fun, mem, name, aux],
			_,
			generateBest[ fun, control, mem, name]
		]
	]

makeMemoization[] :=
	With[ {mem = Unique[ "mem"]},
		makeMemoization[ mem]
	]
makeMemoization[ mem_] := makeMemoization[ mem, Identity]
makeMemoization[ mem_, proj_] := makeMemoization[ mem, proj, False&]
makeMemoization[ mem_, proj_, unique_] :=
	(Clear[ mem]; mem[ _] := $Failed; {mem, proj, unique})

generateFair[ fun_, {mem_, proj_, unique_}, name_] :=
	(
		clearIterator[ name];
		name[ o_] :=
			With[ {p = proj[ o]},
				If[ TrueQ[ mem[ p]],
					LL[],
				(*else*)
					If[ unique[ p] === False, mem[ p] = True];
					LL[
						o,
						lFlatten[
							lMap[ name, fun[ o], True, False],
							lInterleave,
							True
						]
					]
				]
			];
		name
	)
generateFair[ fun_, _, name_] :=
	(
		clearIterator[ name];
		name[ o_] :=
			LL[
				o,
				lFlatten[
					lMap[ name, fun[ o], True, False],
					lInterleave,
					True
				]
			];
		name
	)

generateBest[ fun_, rel_, {mem_, proj_, unique_}, name_] :=
	(
		clearIterator[ name];
		name[ o_] :=
			With[ {p = proj[ o]},
				If[ TrueQ[ mem[ p]],
					LL[],
				(*else*)
					If[ unique[ p] === False, mem[ p] = True];
					LL[
						o,
						lFlatten[
							lMap[ name, fun[ o], True, False],
							lMerge[ rel],
							False
						]
					]
				]
			];
		name
	)
generateBest[ fun_, rel_, _, name_] :=
	(
		clearIterator[ name];
		name[ o_] :=
			LL[
				o,
				lFlatten[
					lMap[ name, fun[ o], True, False],
					lMerge[ rel],
					False
				]
			];
		name
	)

generateDF[ fun_, {mem_, proj_, unique_}, name_] :=
	(
		clearIterator[ name];
		name[ o_] :=
			With[ {p = proj[ o]},
				If[ TrueQ[ mem[ p]],
					LL[],
				(*else*)
					If[ unique[ p] === False, mem[ p] = True];
					LL[
						o,
						lFlatten[
							lMap[ name, fun[ o], True, False],
							lJoin,
							False
						]
					]
				]
			];
		name
	)
generateDF[ fun_, _, name_] :=
	(
		clearIterator[ name];
		name[ o_] :=
			LL[
				o,
				lFlatten[
					lMap[ name, fun[ o], True, False],
					lJoin,
					False
				]
			];
		name
	)
	
(* Experiments suggest that breadth-first search is the method of choice for finitely branching graphs. *)
generateBF[ fun_, mem_, name_, aux_] :=
	(
		clearIterator[ name, aux];
		generateBF[ fun, mem, aux];
		name /: clearIterator[ name] := Clear[ name, aux];
		name[ o_] := LL[ o, aux[ {o}]];
		name
	)
generateBF[ fun_, {mem_, proj_, unique_}, name_] :=
	(
		name[ queue_] :=
			If[ lEmptyQ[ queue],
				LL[],
			(*else*)
				With[ {f = lFirst[ queue, $Failed, Identity], r = lRest[ queue]},
					With[ {p = proj[ f]},
						If[ unique[ p] === False, mem[ p] = True]
					];
					With[ {new = lSelect[ fun[ f], pruneNewBF[ mem, proj, unique, r]]},
						lJoin[ new, Hold[ name[ lJoin[ r, new]]], True]
					]
				]
			]
	)
generateBF[ fun_, _, name_] :=
	(
		name[ queue_] :=
			If[ lEmptyQ[ queue],
				LL[],
			(*else*)
				With[ {f = lFirst[ queue, $Failed, Identity]},
				With[ {new = fun[ f]},
					lJoin[ new, Hold[ name[ lJoin[ lRest[ queue], new]]], True]
				]]
			]
	)
pruneNewBF[ mem_, proj_, unique_, queue_][ o_] :=
	mem[ proj[ o]] =!= True &&
	Catch[
		lScan[ With[ {p = proj[ #]}, If[ unique[ p] === False, mem[ p] = True]; If[ # === o, Throw[ False]]]&, queue, Identity, False];
		True
	]


End[]

EndPackage[]