(* ::Section:: *)
(*Comments*)


(* ::Section:: *)
(*Implementation*)


(* ::Subsection::Closed:: *)
(*Begin*)


BeginPackage[ "SequentialEvaluation`"];

MessageName[ closeSE, "usage", "English"] = "";
MessageName[ initiateSE, "usage", "English"] = "";
MessageName[ runSE, "usage", "English"] = "";

MessageName[ $cacheTime, "usage", "English"] = "";
MessageName[ $timeConstrained, "usage", "English"] = "";
MessageName[ $timingFunction, "usage", "English"] = "";

(* The following global variables are used:
	- '$active[$id]': Specifies whether the SE is active or has already been closed.

	- '$input[$id]': Notebook containing the input to the SE.

	- '$data[$id]'

	- '$init[$id]': Initialization commands for the SE.

	- '$startExecution[$id]': Specifies whether the execution of the SE shall be started.
		If negative, the execution is started and 'startExecution' is set to its absolute value.

	- '$cache[$id]': List of pairs of the form '{index, state}', where 'index' is an index of the 'cells'-list and
		'state' is a state of the SE (represented as a sequence of variable assignments wrapped in Hold).

	- '$currentState[$id]': Either a pair '{index, state}' with the same meaning as the elements of 'cache',
		respresenting the current state of the SE, or only 'index' if the current state is contained in 'cache' anyway.

	- '$cells[$id]': List of elements of the form '{p, id, s, c, l:{{s1, c1}...}, m, t}', where 'id' is a cell-ID from
		$input, 'p' its position among all cells in $input, 's' is its style, 'c' is its content, and 't' is the time
		needed for evaluating the cell (including all of its children) the last time (or $Failed if unknown).
		'm' is the list of all messages attached to the cell, and 'l' is the list of children of the main cell,
		specified by their style 's1' and content 'c1'.

	- '$executionTarget[$id]': Index of the cell in 'cells' corresponding to the requested state of the SE; 0 if the
		initial state of the SE is requested.

	- '$executionIndex[$id]': Index of the cell in 'cells' currently being executed; 0 if the initial state is
		executed; -1 if no evaluation is going on right now because the SE is up to date.

	- '$errorIndex[$id]': Index of the cell in 'cells' where the first execution error occurs, or -1 if no error occurs
		at all.

	- '$abortable[$id]': [OBSOLETE]

	- '$refresh': The value of this variable is changed whenever the status window shall be refreshed.
		Note that $refresh does not depend on $id!
*)

Begin[ "`Private`"];


(* ::Subsection::Closed:: *)
(*Public*)

$cacheTime = 1.;
$timeConstrained = Infinity;
$timingFunction = Timing;	(* MUST return a pair '{t, r}'! *)

(* 'initiateSE[ vars, $id, opts...]' opens a new sequential evaluation interface, consisting of an input notebook and
	a status notebook. 'vars' is supposed to be list of assignments of values to symbols, defining the initial state of
	the sequential evaluation, '$id' is supposed to be an integer, and 'opts' can be optional parameters (see below). *)
Options[ initiateSE] =
	{
		"DisplayFunction" -> printStatus,
		"ParsingFunction" -> parseInput,
		"ExecutionFunction" -> executeCommand,
		"ClosingFunction" -> (Null&),
		"MainCellStyle" -> "Input",
		"MinorCellStyle" -> {},
		"Input" -> {},
		"InputWindowOptions" -> {},
		"StatusWindowOptions" -> {}
	};
initiateSE[ {vars___}, $id_, opts___?OptionQ] :=
	initiateSE[ Hold[ vars], $id, opts]
initiateSE[ vars:Hold[ ((Rule|RuleDelayed|Set)[ _, _])...], $id_, opts___?OptionQ] :=
	(
		$data[$id] = Most /@ (List @@ (Hold @@@ vars));
		$init[$id] = Set @@@ vars;
		$cache[$id] = {};
		$currentState[$id] = 0;
		$errorIndex[$id] = -1;
		$executionIndex[$id] = 0;	(*is set to -1 in 'openSE', but must be 0 here because '$executionTarget' is set to it*)
		initState[ $id];
		openSE[ $id, None, {}, opts]
	)

(* 'runSE[ vars, cmd, $id, opts...]' evaluates the sequence of commands 'cmd' sequentially and, if something goes
	wrong, opens a sequential evaluation interface where the insertion point is at the cell that caused the error.
	'cmd' is supposed to be a list of pairs '{c, index}', where 'c' is the actual command (or list thereof), and
	'index' is the index of the cell 'c' originates from. The list of cells may be passed in the optional parameter
	"input", but it does not need to; if it is not, an empty input notebook will pop up if an execution error occurs.
	'opts' can be any optional parameters 'initiateSE' accepts, plus the ones given below. *)
Options[ runSE] =
	{
		"ExecutionFunction1" -> executeCommand,
		"Postprocessing" -> (Null&),
		"ClosingFunction" -> (Null&),
		"StepFunction" -> (Null&),
		"EnterInteractive" -> Hold[ False],
		"EnterInteractiveFunction" -> (Null&)
		(* "EnterInteractiveFunction" is called on 'reason', 'step', 'len'.
			'reason' is
			- '0' if interactive mode is entered because "EnterInteractive" was set to 'True',
			- '1' if interactive mode is entered because "Postprocessing" returned '$Failed',
			- '2' if an error occurred. *)
	};
runSE[ {vars___}, cmd_List, $id_, opts___?OptionQ] :=
	runSE[ Hold[ vars], cmd, $id, opts]
runSE[ vars:Hold[ ((Rule|RuleDelayed|Set)[ _, _])...], cmd_List, $id_, opts___?OptionQ] :=
	With[ {startTime = SessionTime[], len = If[ cmd === {}, 0, cmd[[-1, -1]]]},
	Module[ {exec1, stepF, enterInteractive, enterInterF, postprocess, mt = Table[ {{}, $Failed}, {Length[ cmd]}], tt = 0., t, r},
		{exec1, stepF, enterInteractive, enterInterF, postprocess, $closingF[$id]} =
			Replace[
				{"ExecutionFunction1", "StepFunction", "EnterInteractive", "EnterInteractiveFunction", "Postprocessing", "ClosingFunction"},
				Join[ {opts}, Options[ runSE]],
				{1}
			];
		
		$data[$id] = Most /@ (List @@ (Hold @@@ vars));
		$init[$id] = Set @@@ vars;
		$cache[$id] = {};
		$currentState[$id] = 0;
		$errorIndex[$id] = -1;
		stepF[ Evaluate[ $executionIndex[$id] = 0], len];
		initState[ $id];
		
		Quiet[
			Check[	(* because when aborting evaluation (manually), '$executionIndex' etc. might not have values any more *)
				Catch[
					Scan[
						If[ TrueQ[ ReleaseHold[ enterInteractive]],
							enterInterF[ 0, $executionIndex[$id], SessionTime[] - startTime];
							openSE[ $id, mt, {}, opts];
							FrontEndExecute[ FrontEndToken[ "EvaluatorAbort"]];
							Throw[ Null],
						(*else*)
							With[ {c = First[ #]},
								stepF[ Evaluate[ $executionIndex[$id] = Last[ #]], len];
								{t, r} =
									Replace[ c,
										{
											Hold[ "Stop"] :> {0., {$Failed, "Execution stopped."}},
											Hold[ "Ignore"] :> {0., Null},
											Hold[ "Cache"] :> {0., "AddToCache"},
											_Hold :> $timingFunction[ TimeConstrained[ exec1[ $id, c], $timeConstrained, {$Failed, "Time limit exceeded."}]]
										}
									];
								Switch[ r,
									{$Failed, __},
									$errorIndex[$id] = $executionIndex[$id];
									enterInterF[ 2, $executionIndex[$id], SessionTime[] - startTime];
									openSE[ $id, mt, Rest[ r], opts];
									FrontEndExecute[ FrontEndToken[ "EvaluatorAbort"]];
									Throw[ Null],
				
									$Failed,
									$errorIndex[$id] = $executionIndex[$id];
									enterInterF[ 2, $executionIndex[$id], SessionTime[] - startTime];
									openSE[ $id, mt, {"Unknown error."}, opts];
									FrontEndExecute[ FrontEndToken[ "EvaluatorAbort"]];
									Throw[ Null],
				
									_,
									$currentState[$id] = {$executionIndex[$id], cacheState[ $id]};
									mt =
										If[ MatchQ[ r, {__}],
											ReplacePart[ mt, {{$executionIndex[$id], 1} -> r, {$executionIndex[$id], 2} -> t}],
										(*else*)
											ReplacePart[ mt, {$executionIndex[$id], 2} -> t]
										];
									tt += t;
									If[ tt >= $cacheTime || r === "AddToCache",
										tt = 0.;
										AppendTo[ $cache[$id], $currentState[$id]];
										$currentState[$id] = $executionIndex[$id]
									]
								]
							]
						]&,
						cmd
					];
					Replace[ postprocess[ $id, SessionTime[] - startTime],
						{
							$Failed :>
								(
									enterInterF[ 1, $executionIndex[$id], SessionTime[] - startTime];
									openSE[ $id, mt, {}, opts];
									FrontEndExecute[ FrontEndToken[ "EvaluatorAbort"]];
								),
							out_ :> (closeSE[ $id, False]; out)
						}
					]
				],
				Null
			]
		]
	]]


(* ::Subsection:: *)
(*Private*)

openSE[ $id_, mt_, err_List, opts___?OptionQ] :=
	Module[ {input, inOpts, statOpts, newCellStyle, cid},
		{$displayF[$id], $parsingF[$id], $executionF[$id], $closingF[$id], $mainStyle[$id], $minorStyle[$id], input, inOpts, statOpts} =
			Replace[
				{"DisplayFunction", "ParsingFunction", "ExecutionFunction", "ClosingFunction", "MainCellStyle", "MinorCellStyle", "Input", "InputWindowOptions", "StatusWindowOptions"},
				Join[ {opts}, Options[ initiateSE]],
				{1}
			];
		If[ ListQ[ $mainStyle[$id]],
			newCellStyle = First[ $mainStyle[$id]];
			$mainStyle[$id] = Alternatives @@ $mainStyle[$id],
		(*else*)
			newCellStyle = $mainStyle[$id]
		];
		If[ ListQ[ $minorStyle[$id]], $minorStyle[$id] = Alternatives @@ $minorStyle[$id]];
		inOpts = Sequence @@ Join[
				{
					CreateCellID -> True
				},
				inOpts,
				{
					DefaultNewCellStyle -> newCellStyle,
					WindowTitle -> "Input",
					WindowMargins -> {{0, Automatic}, {Automatic, 0}}
				}
			];
		statOpts = Sequence @@ Join[
				{
					NotebookEventActions -> {"WindowClose" :> closeSE[ $id, True]},
					NotebookDynamicExpression :>
						Dynamic[
							Refresh[
								If[ TrueQ[ $active[$id]] && updateSE[ $id] === False,
									closeSE[ $id, True]
								],
								UpdateInterval -> 0.5
							]
						]
				},
				statOpts,
				{
					Saveable -> False,
					Editable -> False,
					Selectable -> False,
					ShowCellBracket -> False,
					ShowCellLabel -> False,
					WindowElements -> {},
					WindowTitle -> "Status",
					WindowMargins -> {{Automatic, 0}, {Automatic, 0}},
					WindowSize -> {350, All}
				}
			];
		
		$active[$id] = True;
		$input[$id] = generateSEI[ input, $id, inOpts];
		$startExecution[$id] = 0;
		$cells[$id] = If[ ListQ[ mt], getCellStructure[ $id, mt], {}];
		If[ $errorIndex[$id] >= 1,
			cid = $cells[$id][[$errorIndex[$id], 2]];
			$executionIndex[$id] = $errorIndex[$id];
			executionFailure[ $id, err],
		(*else*)
			cid =
				Which[
					$executionIndex[$id] < 1,
					Begin,
				(*else*)
					$executionIndex[$id] >= Length[ $cells[$id]],
					End,
				(*else*)
					True,
					$cells[$id][[$executionIndex[$id] + 1, 2]]
				];
			$executionTarget[$id] = $executionIndex[$id]
		];
		$executionIndex[$id] = -1;
		$abortable[$id] = False;
		$refresh = 0;

		$stateWindow[$id] = CreateDocument[
			{
				Cell[ BoxData[ RowBox[ {"UpdateState", "[", "]"}]], "Input",
					CellProlog :> (
						$PreRead =
							(Unset[ $PreRead];
							If[ TrueQ[ $startExecution[$id] < 0],
								$startExecution[$id] = -$startExecution[$id];
								"True",
								"False"
							])&;
						$Pre = (Unset[ $Pre]; If[ TrueQ[ #], executeSE[ $id]]; Null)&;
						$Post = Null&;
						Null
					),
					CellEpilog :> (Unset[ $Post]; Null),
					ShowCellBracket -> True
				],
				Dynamic[
					Refresh[
						($refresh; $displayF[$id][ $id]),
						TrackedSymbols :> {$refresh}
					]
				]
			},
			statOpts
		];
		
		$executionCell[$id] = First[ Cells[ $stateWindow[$id]]];
		
		Switch[ cid,
			Begin,
			Null,
			
			End,
			SelectionMove[ $input[$id], After, Notebook],
			
			_,
			Replace[ Cells[ $input[$id], CellID -> cid],
				{co_CellObject, ___} :> SelectionMove[ co, Before, Cell]
			]
		];
		
		SetSelectedNotebook[ $input[$id]];
	]

generateSEI[ cmd_List, $id_, opts___?OptionQ] :=
	Module[ {maxID = Max @@ Append[ Cases[ cmd, Cell[ ___, CellID -> i_Integer, ___] :> i], 0]},
		With[ {main = $mainStyle[$id]},
		With[ {cells =
					Replace[
						cmd,
						{
							s_String :> Cell[ BoxData[ s], main, CellID -> (++maxID)],
							c:Cell[ _, main, ___] :> DeleteCases[ c, CellDingbat -> Except[ _TooltipBox]]
						},
						{1}
					]
				},
			CreateDocument[
				Cases[ cells, _Cell],
				opts
			]
		]]
	]

printInternalStatus[ $id_] :=
	Grid[
		Append[ Map[ {TextCell[ SymbolName[ #]], TextCell[ " = "], #[$id]}&, {$executionIndex, $executionTarget, $errorIndex}],
			{TextCell[ "$cache"], TextCell[ " = "], First /@ $cache[$id]}
		],
		Alignment -> Left
	]

initState[ $id_] := ReleaseHold[ $init[$id]]

(* 'getCellStructure[ $id, mt]' obtains the cell structure of '$input[$id]' w.r.t. the message-time-list 'mt'.
	It proceeds exactly as 'updateSE', but without setting 'selection', 'target' and 'devIndex'. *)
getCellStructure[ $id_, mt_List] :=
	With[ {$MainStyle = $mainStyle[$id], $MinorStyle = $minorStyle[$id]},
	With[ {cells =
			Split[
				Replace[ MapIndexed[ {First[ #2], #1}&, NotebookRead /@ Cells[ $input[$id]]],
					{
						{p_Integer, Cell[ content_, $MainStyle, ___, CellID -> id_, ___]} :> {p, id, $MainStyle, content},
						{_Integer, Cell[ content_, s:$MinorStyle, ___]} :> {s, content},
						_ -> Sequence[]
					},
					{1}
				],
				(!IntegerQ[ First[ #2]])&
			]},
		Quiet[
			Check[
				(* yields an error if the lengths of 'cells' and 'mt' are different, in which case the second alternative is returned *)
				MapThread[
					Replace[ #1,
						{{p_, id_, f___}, r___} :>
							Join[
								{p, id, f, {r}},
								Replace[ #2,
									{
										{{_, {}, t_}} :> {{}, t},
										{{_, m_List, t_}} :> (CurrentValue[ Cells[ $input[$id], CellID -> id], CellDingbat] = makeMessageDingbat[ m]; {m, t})
									}
								]
							]
					]&,
					{cells, mt}
				],
			(*else*)
				Replace[
					cells,
					{{f___}, r___} :> {f, {r}, {}, $Failed},
					{1}
				]
			]
		]
	]]

updateSE[ $id_] :=
	With[ {$MainStyle = $mainStyle[$id], $MinorStyle = $minorStyle[$id]},
	Module[ {cells, selection, target, devIndex = 1},
		selection = Quiet[ Check[ SelectedCells[ $input[$id]], $Failed]];
		If[ selection === $Failed || !IntegerQ[ $startExecution[$id]],
			False,
		(*else*)
			(*collect all cells in SEI*)
			cells = Cases[ MapIndexed[ {First[ #2], #1}&, NotebookRead /@ Cells[ $input[$id]]],
						{p_Integer, Cell[ content_, style_, ___, CellID -> id_, ___]} :> {p, id, style, content}
					];

			If[ selection === {},
				SelectionMove[ $input[$id], Next, Cell, AutoScroll -> False];
				selection = SelectedCells[ $input[$id]];
				SelectionMove[ $input[$id], Before, Cell, AutoScroll -> False];
				target = selection,
			(*else*)
				target = selection;
				selection = $Failed
			];
			target = Map[ With[ {id = CurrentValue[ #, CellID]},
						If[ IntegerQ[ id],
							Position[ cells, {_, id, ___}, {1}, 1],
						(*else*)
							$executionTarget[$id]
						]
					]&,
					target
				];
			target = Max@@Cases[ target, {{p_Integer}} :> p];
			If[ IntegerQ[ target],
				While[ target <= Length[ cells] && !MatchQ[ cells[[target, 3]], $MainStyle|$MinorStyle],
					++target
				];
				If[ target > Length[ cells], target = -Infinity]
			];

			cells = Replace[
					cells,
					{
						q:{_, _, $MainStyle, _} :> q,
						{_, _, s:$MinorStyle, c_} :> {s, c},
						_ -> $Failed
					},
					{1}
				];
			cells = Replace[ Split[ DeleteCases[ cells, $Failed], (!IntegerQ[ First[ #2]])&], {{f___}, r___} :> {f, {r}, {}, $Failed}, {1}];
			(* 'cells' is now a list consisting of elements of the form '{p, id, s, c, l:{{s1, c1}...}, {}, $Failed}',
				where all elements have main-style, 'p' is the position of the respective cell in SEI,
				's' is its style, 'id' is its ID, 'c' its content, and 'l' the list of all "child"-cells
				coming afterward (before the next main cell). Each child-cell is given by its style 's1'
				and its content 'c1'. *)

			If[ IntegerQ[ target],
				target = Max[ 0, Count[ cells, {_?(# <= target &), ___}] - 1],
			(*else*)
				target = Length[ cells]
			];

			Catch[ Scan[
				If[ devIndex > Length[ cells] || cells[[devIndex, -4]] =!= #[[-4]] || cells[[devIndex, -3]] =!= #[[-3]],
					Throw[ Null],
				(*else*)
					cells[[devIndex, -1]] = Last[ #];
					cells[[devIndex, -2]] = #[[-2]];
					++devIndex
				]&,
				$cells[$id]
			]];

			$cells[$id] = cells;

			(*remove all entries coming after 'devIndex' from the cache*)
			$cache[$id] = DeleteCases[ $cache[$id], {_Integer?(# >= devIndex &), ___}];
			If[ MatchQ[ $currentState[$id], {_Integer?(# >= devIndex &), ___}], $currentState[$id] = First[ $currentState[$id]]];
			If[ $errorIndex[$id] >= devIndex,
				(*remove error-indicators ("x", "?") from all cells coming after $errorIndex*)
				CurrentValue[ Cells[ $input[$id], CellID -> $cells[$id][[$errorIndex[$id];;, 2]]], CellDingbat] = None;
				$errorIndex[$id] = -1
			];

			If[ devIndex <= $executionIndex[$id] ||
				target < $executionIndex[$id] ||
				(0 <= $executionIndex[$id] < target && MemberQ[ $cache[$id], {_Integer?($executionIndex[$id] < # <= target &), ___}]),
				If[ $executionIndex[ $id] >= 0,
					(*abort the current execution*)
					abortExecution[ $id];
					$executionIndex[$id] = -1
				];
				$executionTarget[$id] = -1
			];

			If[ 1 <= $errorIndex[$id] <= target, target = $errorIndex[$id] - 1];

			If[ target =!= $executionTarget[$id],
				(*update the execution target*)
				Which[
					$executionIndex[$id] < 0,
					If[ $errorIndex[$id] <= 0,
						CurrentValue[
							Cells[
								$input[$id],
								CellID -> Cases[ $cells[$id], {_, id_, ___, {}, _} :> id]
							],
							CellDingbat
						] = None,
					(*else*)
						CurrentValue[
							Cells[
								$input[$id],
								CellID -> Cases[ $cells[$id][[1;;($errorIndex[$id] - 1)]], {_, id_, ___, {}, _} :> id]
							],
							CellDingbat
						] = None
					],
					
					target < $executionTarget[$id],
					(*remove evaluation-indicators (orange triangles)*)
					CurrentValue[ Cells[ $input[$id], CellID -> $cells[$id][[(target + 1);;$executionTarget[$id], 2]]], CellDingbat] =
						None,

					True,
					(*add evaluation-indicators*)
					CurrentValue[ Cells[ $input[$id], CellID -> $cells[$id][[($executionTarget[$id] + 1);;target, 2]]], CellDingbat] =
						makeDingbat[ "\[RightTriangle]", FontColor -> Orange]
				];
				$executionTarget[$id] = target;
				If[ TrueQ[ $executionTarget[$id] >= 0 && $executionIndex[$id] < 0],
					(*start new execution*)
					$startExecution[$id] = -(Abs[ $startExecution[$id]] + 1)
				]
			];

			If[ $startExecution[$id] < 0 && MatchQ[ $executionCell[$id], _CellObject],
				SelectionMove[ $executionCell[$id], All, Cell, AutoScroll -> False];
				SelectionEvaluateCreateCell[ $stateWindow[$id], AutoScroll -> False];
				SelectionMove[ $executionCell[$id], Before, Cell, AutoScroll -> False]
			];

			True
		]
	]]

abortExecution[ $id_] :=
	If[ True, (*TrueQ[ $abortable[$id]],*)
		$abortable[$id] = False;
		FrontEndExecute[ FrontEndToken[ "EvaluatorAbort"]] (*problematic if several SEs are active*)
	]

executeSE[ $id_] :=
	Module[ {cacheIndex = Count[ $cache[$id], {_Integer?(# <= $executionTarget[$id] &), _}], cmd, args, t, r, tt = 0.},
		$executionIndex[$id] = If[ cacheIndex === 0, 0, $cache[$id][[cacheIndex, 1]]];
		Switch[ $currentState[$id],
			{_Integer?($executionIndex[$id] < # <= $executionTarget[$id] &), _},
			tt = Total[ Cases[ $cells[$id][[($executionIndex[$id] + 1);;First[ $currentState[$id]], -1]], _?NumberQ]];
			$executionIndex[$id] = First[ $currentState[$id]];
			ReleaseHold[ Last[ $currentState[$id]]],

			_,
			If[ cacheIndex === 0,
				initState[ $id],
			(*else*)
				ReleaseHold[ $cache[$id][[cacheIndex, -1]]]
			]
		];
		(*$currentState[$id] = $executionIndex[$id];*)
		CurrentValue[ Cells[ $input[$id], CellID -> $cells[$id][[($executionIndex[$id] + 1);;$executionTarget[$id], 2]]], CellDingbat] =
			makeDingbat[ "\[RightTriangle]", FontColor -> Orange];

		While[ $executionIndex[$id] < $executionTarget[$id],
			++$executionIndex[$id];
			If[ $executionIndex[$id] > 0,	(* '$executionIndex[$id]' could be '0', for some strange reason ... *)
				With[ {cell = $cells[$id][[$executionIndex[$id]]]},	(*parse expressions*)
					cmd = parseInput[ $id][ {$mainStyle[$id], cell[[-4]]}];
					If[ cmd === $Failed,
						(*main command could not be parsed*)
						executionFailure[ $id, "Syntax error: command cannot be parsed."],
					(*else*)
						args = parseInput[ $id] /@ cell[[-3]];
						If[ MemberQ[ args, $Failed],
							(*at least one argument could not be parsed*)
							executionFailure[ $id, "Syntax error: arguments at positions " <>
													ToString[ First /@ Position[ args, $Failed, {1}]] <>
													" cannot be parsed."
							],
						(*else*)
							CurrentValue[ Cells[ $input[$id], CellID -> cell[[2]]], CellDingbat] = makeDingbat[ "\[FilledRightTriangle]", FontColor -> Orange];
							$abortable[$id] = True;
							{t, r} =
								Replace[ cmd,
									{
										Hold[ "Stop"] :> {0., {$Failed, "Execution stopped."}},
										Hold[ "Ignore"] :> {0., Null},
										Hold[ "Cache"] :> {0., "AddToCache"},
										_Hold :> $timingFunction[ TimeConstrained[ $executionF[$id][ $id, cmd, args], $timeConstrained, {$Failed, "Time limit exceeded."}]]
									}
								];
							$abortable[$id] = False;
							Switch[ r,
								{$Failed, __},
								(*execution of command failed*)
								executionFailure[ $id, Rest[ r]],
	
								$Failed,
								executionFailure[ $id, "Unknown error."],
	
								_,
								$currentState[$id] = {$executionIndex[$id], cacheState[ $id]};
								$cells[$id] = ReplacePart[ $cells[$id], {$executionIndex[$id], -1} -> t];
								If[ MatchQ[ r, {__}],
									$cells[$id] = ReplacePart[ $cells[$id], {$executionIndex[$id], -2} -> r];
									CurrentValue[ Cells[ $input[$id], CellID -> cell[[2]]], CellDingbat] = makeMessageDingbat[ r],
								(*else*)
									CurrentValue[ Cells[ $input[$id], CellID -> cell[[2]]], CellDingbat] = None
								];
								tt += t;
								If[ tt >= $cacheTime || r === "AddToCache",
									tt = 0.;
									++cacheIndex;
									If[ cacheIndex <= Length[ $cache[$id]],
										$cache[$id] = ReplacePart[ $cache[$id], cacheIndex -> $currentState[$id]],
									(*else*)
										AppendTo[ $cache[$id], $currentState[$id]]
									];
									$currentState[$id] = $executionIndex[$id]
								]
							]
						]
					]
				]
			]
		];
		If[ 1 <= $executionTarget[$id] <= Length[ $cells[$id]],
			With[ {cell = $cells[$id][[$executionTarget[$id]]]},
				If[ cell[[-2]] === {},
					CurrentValue[ Cells[ $input[$id], CellID -> cell[[2]]], CellDingbat] =
						makeDingbat[ "\[FilledRightTriangle]", FontColor -> Darker[ Green]]
				]
			]
		];
		$executionIndex[$id] = -1;
		++$refresh;
	]

cacheState[ $id_] :=
	With[ {val = ReleaseHold[ $data[$id]]},
		Set @@@ (Hold @@ MapThread[ Append, {$data[$id], val}])
	]

executionFailure[ $id_, msg_List] :=
	With[ {id1 = $cells[$id][[$executionIndex[$id], 2]], id2 = (#[[2]] &) /@ Drop[ $cells[$id], $executionIndex[$id]]},
		CurrentValue[ Cells[ $input[$id], CellID -> id1], CellDingbat] = makeDingbat[ msg, "\[WarningSign]", FontColor -> Red];
		CurrentValue[ Cells[ $input[$id], CellID -> id2], CellDingbat] = makeDingbat[ "?", FontSize -> Medium, FontColor -> RGBColor[ 0.5, 0, 1]];
		$errorIndex[$id] = $executionIndex[$id];
		$executionTarget[$id] = $executionIndex[$id] - 1
	]
executionFailure[ $id_, msg_String] :=
	executionFailure[ $id, {msg}]

makeDingbat[ text_String, opts___] :=
	makeDingbat[ {}, text, opts]
makeDingbat[ tt_List, text_String, opts___] :=
	If[ tt === {},
		StyleBox[ text, "Text", opts],
	(*else*)
		ToBoxes[ Tooltip[ Style[ text, ShowStringCharacters -> False, opts], Replace[ tt, {{tt0_} :> tt0, _ :> Column[ tt]}]]]
	]

makeMessageDingbat[ msg_List, opts___] :=
	makeMessageDingbat[ "\[LightBulb]", msg, opts]
makeMessageDingbat[ text_String, msg_List, opts___] :=
	makeDingbat[ msg, text, opts, FontColor -> Blue]

closeSE[ $id_, True] :=
	If[ TrueQ[ $active[$id]],
		abortExecution[ $id];
		$active[$id] = False;
		Quiet[ $abortable[$id] =., Unset::norep];
		Quiet[ $executionCell[$id] =., Unset::norep];
		Quiet[ $startExecution[$id] =., Unset::norep];
		Quiet[ $cells[$id] =., Unset::norep];
		Quiet[ $executionTarget[$id] =., Unset::norep];
		Quiet[ NotebookClose[ $input[$id]]];
		Quiet[ NotebookClose[ $stateWindow[$id]]];
		Quiet[ $input[$id] =., Unset::norep];
		Quiet[ $stateWindow[$id] =., Unset::norep];
		Unset[ $PreRead];
		Unset[ $Pre];
		Unset[ $Post];
		closeSE[ $id, False]
	]
closeSE[ $id_, False] :=
(
	Quiet[ $cache[$id] =., Unset::norep];
	Quiet[ $currentState[$id] =., Unset::norep];
	Quiet[ $data[$id] =., Unset::norep];
	Quiet[ $errorIndex[$id] =., Unset::norep];
	Quiet[ $executionIndex[$id] =., Unset::norep];
	Quiet[ $init[$id] =., Unset::norep];
	$closingF[$id][ $id];
)

printStatus[ $id_] := Grid[ Map[ {HoldForm @@ #, TextCell[ " = "], ReleaseHold[ #]}&, $data[$id]]]

parseInput[ $id_][ {s_, c_}] := Quiet[ Check[ $parsingF[$id][ $id, s, c], $Failed]]
parseInput[ _, _, box_] := ToExpression[ box, StandardForm, Hold]

executeCommand[ _, Hold[ expr_], _List] := expr
executeCommand[ _, Hold[ expr_]] := expr


(* ::Subsection::Closed:: *)
(*End*)


End[];

EndPackage[];
