(* ::Package:: *)

(* Theorema 
    Copyright (C) 2010 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Core`HOUnification`"]

Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]
Needs[ "Theorema`Language`"]
Needs[ "Theorema`Common`"]
Needs[ "LazyList`", FileNameJoin[ {Theorema`$TheoremaDirectory, "Theorema", "Components", "LazyList.m"}]]

With[ {lang = "English"},
	MessageName[ unify, "usage", lang] = "'unify[ system, meta, fix, tfix, tvar, tag]' attempts to pre-unify the given list 'system' of higher-order
		unification problems. It returns a (lazy) list of the unifiers found; these unifiers do not need to be most general.";
	MessageName[ unifyFirst, "usage", lang] = "'unifyFirst[ system, meta, fix, tfix, tvar, mode]' attempts to pre-unify the given list 'system' of
		higher-order unification problems and, in case of success, returns the first unifier found (discarding all others).";
];

Begin["`Private`"]

(* ::Section:: *)
(*Public Functions*)

(* 'unify[ system, meta, fix, tfix, tvar, tag]' attempts to pre-unify the given list 'system' of higher-order
	unification problems. It returns a (lazy) list of the unifiers found; these unifiers do not need to be
	most general.
	- 'system' is the list of unification problems that shall be solved. Each unification problems is a
		pair of beta-eta reduced, typed lambda terms.
	- 'meta' is a list of rules of the form 'var_ -> info_', where 'var' is a free variable
		of the form 'META$[ _Symbol, _Integer]' and 'info' is a pair of the form '{t_, dep_}',
		where 't' is the type of 'var' and 'dep' specifies which a.b.f. constants the instance of 'var' may depend upon:
		an integer (or 'Infinity') has the same meaning as in Core proof situations, whereas a pair consisting of an
		integer/'Infinity' and a list of indices specifies additional a.b.f. constants 'var' may *not* depend on.
		'dep' may also be 'False', which means that the corresponding 'var' must not be instantiated and is hence treated
		like a constant in the unification procedure.
	- 'fix' is a list of rules of the form 'const_ -> type_', where 'const' is an a.b.f.
		constant and 'type' is its type.
	- 'tfix' and 'tvar' are associations of sort constraints on a.b.f. type constants and type variables, respectively.
	- 'tag' is an arbitrary symbol that serves as the "tag" associated to the given unification problem, or, more
		precisely, to the output of the function. Executing 'clearIterator[ tag]' clears all auxiliary objects
		generated for solving the unification problem. 'tag' shall not be a fixed symbol, but a symbol generated by
		'Unique' or by 'Module'.
	The following options may be given:
	- "FlexFlexToTerms" can be either 'True' (default) or 'False'. With 'True', remaining flex-flex pairs are returned as
		pairs of proper terms, where all variables are bound by abstractions. With 'False', flex-flex pairs are returned
		as triples of the form '{{ah, aargs}, {bh, bargs}, loose}', where 'ah' and 'bh' are the heads of the terms (both
		are meta variables), 'aargs' and 'bargs' are their respective argument lists, and 'loose' is the list of types of
		the loose bound variables (in reverse order, as usual).
	- "AllowFlexFlex" can be either 'True' (default) or 'False'. It specifies whether solutions may contain flex-flex
		pairs, or whether such solutions shall be dropped immediately.
	- "TypeConstThreshold" can be an arbitrary integer or 'Infinity' (default is '0'). When unifying types, all type
		variables with an index strictly below "TypeConstThreshold" are treated as constants, i.e. not instantiated. '0'
		means that all type variables are treated as variables, whereas 'Infinity' means that no type variables may be
		instantiated; but of course global parameter '$instantiateTypes' is also taken into account.
	The output is a (lazy) list of 5-tuples '{flexflex, rem, subst, tsubst, newTVar}', where
	- 'flexflex' is a list of flex-flex pairs,
	- 'rem' is the list of free variables remaining after applying 'subst', with type- and dependency information
		as usual, and
	- 'subst' and 'tsubst' are the term- and type substitutions, respectively, whose application to the original system
		yields the system of flex-flex pairs. Note that 'subst' is in triangular form, i.e. must be applied sequentially!
		Also note that 'tsubst' has already been applied to 'flexflex', 'rem' and 'subst'.
	- 'newTVar' is the new association of sort constraints on type variables (might be the same as input 'tvar').
	The following preconditions must be satisfied to ensure the correct behavior of the function:
	- no loose bound variables or meta variables may appear under 'TYPED$',
	- the system is beta-reduced (but it neither needs to be in eta-long- nor eta-short form),
	- every a.b.f. constant and meta variable occurring in 'system' is listed in 'fix' and 'meta', respectively.
	These preconditions are not checked during execution, but simply assumed!
	If the last assumption is violated the function may crash during execution; otherwise, it does not crash but might
	miss unifiers or compute incorrect solutions.
	At each time a unifier, i.e. an element of the lazy list returned by 'unify', shall be computed, the following global
	variables are set:
	- '$unifFix' is set to the list of a.b.f. constants,
	- '$unifMeta' is set to the current list of meta variables,
	- '$unifTFix' is set to the sort constraints on a.b.f. type constants,
	- '$unifTVar' is set to the current sort constraints on type variables,
	- '$unifSubst' is set to the so-far computed substitution, and
	- '$unifTSubst' is set to the so-far computed type substitution.
	All these values could also be passed as explicit parameters of the various functions that need them, but global
	variables make the implementation easier (and the execution faster?). Of course, it is possible to run several
	unification problems "in parallel", i.e. to work with several lazy lists of unifiers at once, without having to worry
	about setting the global variables to the right values (which clearly may be different in different unification problems);
	'unify' and 'transformUP' take care of all this. Furthermore, the global variables are in fact parametrized over 'tag',
	which allows for "real" parallel executions using 'Parallelize' etc. that do not interfere with each other (as long as
	'tag' is different in each invocation).
	
	Theoretical background: 'unify' essentially implements the pre-unification procedure presented in
	
	[1]	Christian Prehofer. Solving Higher-Order Equations: From Logic to Programming.
		Progress in Theoretical Computer Science, Birkhaeuser, 1997. ISBN 0-8176-4032-0.
	
	but also incorporates a unification algorithm for higher-order patterns, which is applied automatically if an equation
	is detected to belong to this category. *)
unify[ system_List, meta_List, fix_List, tfix_Association, tvar_Association, tag_Symbol, opts___?OptionQ] :=
	Module[ {aux},
	With[ {ffPatt = If[ TrueQ[ Replace[ "AllowFlexFlex", {opts, _ -> True}]], Blank[], {}],
			id = tag},
		$unifFlexFlexToTerms[ id] = Replace[ "FlexFlexToTerms", {opts, _ -> True}];
		$unifTypeConstThreshold[ id] = If[ TrueQ[ $instantiateTypes], Replace[ "TypeConstThreshold", {opts, _ -> 0}], Infinity];
		$unifTFix[ id] = tfix;	(* does not change *)
		With[ {out =
				makeIterator[
					transformUP[ id],
					"Control" -> "breadth-first",	(* do not use any parallel controls (if there are some in the future)! --> global variables '$unifFix' etc. *)
					"Memoization" -> None,	(* the derivation graph should actually be a tree, i.e. it should not be possible to reach one unifier in two different ways *)
					"Name" -> tag,
					"NameAux" -> aux
				]
			},
			tag /: clearIterator[ tag] :=
				(
					Clear[ tag, aux];
					Unset[ $unifFlexFlexToTerms[ id]];
					Unset[ $unifTypeConstThreshold[ id]];
					Unset[ $unifFix[ id]];
					Unset[ $unifMeta[ id]];
					Unset[ $unifTFix[ id]];
					Unset[ $unifTVar[ id]];
					Unset[ $unifSubst[ id]];
					Unset[ $unifTSubst[ id]]
				);
			lCases[
				out[ {Map[ {Append[ splitHeadArgs[ First[ #]], Indeterminate], Append[ splitHeadArgs[ Last[ #]], Indeterminate], {}}&, system], {}, {}, fix, meta, tvar, {}, {}}],
				{ffPatt, _, _, _, _}
			]
		]
	]]
(* The below variant of 'unify' assumes that the given system consists only of higher-order patterns, and tries to
	compute the most general unifier. If the system does not consist of higher-order patterns only, the result might
	be incorrect or incomplete.
	Although the general procedure incorporates pattern unification as well, it is more efficient to use the below
	variant if the system is known in advance to consist of patterns only.
	Note that the output is the same as with the above variant, i.e. a *list* of 5-tuples; this in particular also
	includes lists of flex-flex pairs, which will be empty, though. *)
unify[ system_List, meta_List, fix_List, tfix_Association, tvar_Association, "pattern", opts___?OptionQ] :=
	Module[ {tag},
		$unifTypeConstThreshold[ tag] = If[ TrueQ[ $instantiateTypes], Replace[ "TypeConstThreshold", {opts, _ -> 0}], Infinity];
		$unifTFix[ tag] = tfix;		(* does not change *)
		With[ {res =
				Replace[
					transformUP[ tag][ {Map[ {Append[ splitHeadArgs[ First[ #]], True], Append[ splitHeadArgs[ Last[ #]], True], {}}&, system], {}, {}, fix, meta, tvar, {}, {}}],
					Except[ {{{}, _, _, _, _}}] -> {}
				]},
			Unset[ $unifTypeConstThreshold[ tag]];
			Unset[ $unifFix[ tag]];
			Unset[ $unifMeta[ tag]];
			Unset[ $unifTFix[ tag]];
			Unset[ $unifTVar[ tag]];
			Unset[ $unifSubst[ tag]];
			Unset[ $unifTSubst[ tag]];
			res
		]
	]
(* The below variant of 'unify' performs purely syntactic unification of the given system.
	As with "pattern", the output is still a list of 5-tuples. *)
unify[ system0_List, meta_List, fix_List, tfix_Association, tvar_Association, "syntactic", opts___?OptionQ] :=
	Module[ {tag, system = system0},
		$unifTypeConstThreshold[ tag] = If[ TrueQ[ $instantiateTypes], Replace[ "TypeConstThreshold", {opts, _ -> 0}], Infinity];
		$unifFix[ tag] = fix;
		$unifMeta[ tag] = meta;
		$unifTFix[ tag] = tfix;
		$unifTVar[ tag] = tvar;
		$unifSubst[ tag] = {};
		$unifTSubst[ tag] = {};
		While[ ListQ[ system] && system =!= {},
			system = updateUPSyn[ Rest[ system], unifySynAux[ First[ system], tag], tag]
		];
		system =
			If[ system === {},
				{{{}, $unifMeta[ tag], $unifSubst[ tag], $unifTSubst[ tag], $unifTVar[ tag]}},
			(*else*)
				{}
			];
		Unset[ $unifTypeConstThreshold[ tag]];
		Unset[ $unifFix[ tag]];
		Unset[ $unifMeta[ tag]];
		Unset[ $unifTFix[ tag]];
		Unset[ $unifTVar[ tag]];
		Unset[ $unifSubst[ tag]];
		Unset[ $unifTSubst[ tag]];
		system
	]

(* 'unifyFirst[ system, meta, fix, tfix, tvar, mode, opts]' tries to unify the given system and, in case of success,
	returns the first unifier found. 'mode' can be either "pattern", "syntactic" or anything else; in the latter case,
	full-fledged higher-order unification is employed.
	The advantage of this function over 'unify' merely lies in the fact that the calling function does not need to care
	about tags and iterators and the like. *)
unifyFirst[ system_List, meta_List, fix_List, tfix_Association, tvar_Association, mode_String, opts___] :=
	lFirst[ unify[ system, meta, fix, tfix, tvar, mode, opts]]
unifyFirst[ system_List, meta_List, fix_List, tfix_Association, tvar_Association, _, opts___] :=
	Module[ {tag},
	With[ {res = lFirst[ unify[ system, meta, fix, tfix, tvar, tag, opts]]},
		clearIterator[ tag];
		res
	]]


(* ::Section:: *)
(*Private Functions*)

(* ::Subsection:: *)
(*Higher-Order Unification*)

(* 'transformUP' operates on 8-tuples '{system, flexrigid, flexflex, fix, meta, tvar, subst, tsubst}', where
	- 'system', 'flexrigid' and 'flexflex' are lists of tuples '{{ah, aargs, astruct}, {bh, bargs, bstruct}, loose, opts...}'
		representing the equation 'ah[ aargs] = bh[ bargs]' (with 'ah' and 'bh' atomic), the structures of 'a'
		and 'b' ('True' = higher-order pattern, 'False' = no higher-order pattern,
		list = list of meta variables whose instantiation might render the argument list a higher-order pattern,
		'Indeterminate' = unknown), the types of the loose bound variables that have been stripped off already,
		and optional parameters guiding the unification procedure,
	- 'fix' is the list of a.b.f. constants, together with type information,
	- 'meta' is the list of meta variables, together with type- and dependency information (see above),
	- 'tvar' is the association of sort constraints on type variables,
	- 'subst' is the so-far computed term substitution, and
	- 'tsubst' is the so-far computed type substitution. *)
(* initialization *)
transformUP[ tag_][ {system_, flexrigid_, flexflex_, fix_, meta_, tvar_, subst_, tsubst_}] :=
	(
		$unifFix[ tag] = fix;
		$unifMeta[ tag] = meta;
		$unifTVar[ tag] = tvar;
		$unifSubst[ tag] = subst;
		$unifTSubst[ tag] = tsubst;
		transformUP[ system, flexrigid, flexflex, tag]
	)
transformUP[ _][ __] := {}
(* postprocessing *)
transformUP[ {}, {}, flexflex_, tag_] :=
	transformUP[ {}, {}, flexflex, {}, tag]
transformUP[ {}, {}, {}, out_, tag_] :=
	{Join[ {out, $unifMeta[ tag], $unifSubst[ tag]} /. META$[ n_, i_, True] :> META$[ n, i], {$unifTSubst[ tag], $unifTVar[ tag]}]}
transformUP[ {}, {}, {{{ah_, aargs_, _}, {bh_, bargs_, _}, loose_, ___}, r___}, out_, tag_] :=
	With[ {ta = stripArgumentTypes[ First[ Replace[ ah, $unifMeta[ tag]]], Length[ aargs]], tb = stripArgumentTypes[ First[ Replace[ bh, $unifMeta[ tag]]], Length[ bargs]]},
		If[ ta =!= $Failed && tb =!= $Failed,
			Replace[ compareTypes[ ta, tb, hasSort[ $unifTFix[ tag]], {$unifTVar[ tag], {}}, $unifTypeConstThreshold[ tag]],
					(* 'ta' and 'tb' are types of meta variables, hence they must be full types *)
				{
					{newTVar_Association, tunif_List} :>
						(
							$unifFix[ tag] = $unifFix[ tag] /. tunif;
							$unifMeta[ tag] = $unifMeta[ tag] /. tunif;
							$unifSubst[ tag] = replaceTypes[ $unifSubst[ tag], tunif];
							$unifTSubst[ tag] = Join[ tunif, $unifTSubst[ tag] /. tunif];
							$unifTVar[ tag] = newTVar;
							transformUP[
								{},
								{},
								replaceTypes[ {r}, tunif],
								replaceTypes[
									Append[ out,
										If[ TrueQ[ $unifFlexFlexToTerms[ tag]], {applyCurried[ ah, aargs, loose], applyCurried[ bh, bargs, loose]}, {{ah, aargs}, {bh, bargs}, loose}]
									],
									tunif
								],
								tag
							]
						),
					_ -> {}
				}
			],
		(*else*)
			{}
		]
	]
(* imitation, projection *)
transformUP[ {}, {f_, r___}, flexflex_, tag_] :=
	With[ {res = DeleteCases[ imitProjUP[ f, tag], $Failed]},
		If[ MatchQ[ res, {_}],
			updateUP[ {}, {r}, flexflex, tag] @@ res,
		(*else*)
			updateUP[ {r}, flexflex, tag] /@ res
		]
	]
(* elimination, decomposition, trivial *)
transformUP[ {f_, r___}, acc_, flexflex_, tag_] :=
	updateUP[ {r}, acc, flexflex, tag][ elimDecompUP[ f, tag]]

updateUP[ sys1_, sys2_, flexflex_, tag_][ {new1_, newMeta_, inst_, tinst_, tvar_}] :=
	updateUP[ sys1, sys2, flexflex, tag][ {new1, {}, {}, newMeta, inst, tinst, tvar}]
updateUP[ sys1_, sys2_, flexflex_, tag_][ {new1_, new2_, newFF_, newMeta_, {}, tinst_, tvar_}] :=
	(
		$unifFix[ tag] = $unifFix[ tag] /. tinst;
		$unifMeta[ tag] = Join[ $unifMeta[ tag], newMeta] /. tinst;	(* 'replaceTypes' does not work here *)
		If[ tvar =!= Null, $unifTVar[ tag] = tvar];
		$unifSubst[ tag] = replaceTypes[ $unifSubst[ tag], tinst];
		$unifTSubst[ tag] = Join[ tinst, $unifTSubst[ tag] /. tinst];
		transformUP[
			replaceTypes[ Join[ new1, sys1], tinst],
			replaceTypes[ Join[ sys2, new2], tinst],
			replaceTypes[ Join[ flexflex, newFF], tinst],
			tag
		]
	)
updateUP[ sys1_, sys2_, flexflex_, tag_][ {new1_, new2_, newFF_, newMeta_, inst_, tinst_, tvar_}] :=
	With[ {instVars = Alternatives @@ inst[[All, 1]]},
	With[ {newSys2 = instantiateEquation[ inst, instVars] /@ sys2,
			newFF0 = instantiateEquation[ inst, instVars] /@ flexflex},
	With[ {posSys2 = Position[ newSys2, {{_META$, _, as_}, {_, _, bs_}, ___} /; (as === False || ListQ[ as] || bs === False || ListQ[ bs]), {1}, Heads -> False],
			posFF = Position[ newFF0, {{_META$, _, as_}, {_META$, _, bs_}, ___} /; (as === False || ListQ[ as] || bs === False || ListQ[ bs]), {1}, Heads -> False]},
		(* some flex-rigid and flex-flex equations might become higher-order patterns after instantiation
			--> are detected and put into the list of equations to be looked at *)
		$unifFix[ tag] = $unifFix[ tag] /. tinst;
		$unifMeta[ tag] = Join[ DeleteCases[ $unifMeta[ tag], instVars -> _], newMeta] /. tinst;
		If[ tvar =!= Null, $unifTVar[ tag] = tvar];
		$unifSubst[ tag] = replaceTypes[ Join[ normalize[ $unifSubst[ tag] /. Cases[ inst, _[ META$[ _, _, True], _]], True], DeleteCases[ inst, META$[ _, _, True] -> _]], tinst];
		$unifTSubst[ tag] = Join[ tinst, $unifTSubst[ tag] /. tinst];
		transformUP[
			replaceTypes[ Join[ Delete[ newSys2, posSys2], Delete[ newFF0, posFF], new1, instantiateEquation[ inst, instVars] /@ sys1], tinst],
			replaceTypes[ Join[ Extract[ newSys2, posSys2], new2], tinst],
			replaceTypes[ Join[ Extract[ newFF0, posFF], newFF], tinst],
			tag
		]
	]]]
updateUP[ sys_, flexflex_, tag_][ {newSys_, newMeta_, inst_, tinst_, tvar_}] :=
	With[ {instVars = Alternatives @@ inst[[All, 1]]},
	With[ {newSys2 = instantiateEquation[ inst, instVars] /@ sys,
			newFF0 = instantiateEquation[ inst, instVars] /@ flexflex},
	With[ {posSys2 = Position[ newSys2, {{_META$, _, as_}, {_, _, bs_}, ___} /; (as === False || ListQ[ as] || bs === False || ListQ[ bs]), {1}, Heads -> False],
			posFF = Position[ newFF0, {{_META$, _, as_}, {_META$, _, bs_}, ___} /; (as === False || ListQ[ as] || bs === False || ListQ[ bs]), {1}, Heads -> False]},
		{
			replaceTypes[ Join[ Delete[ newSys2, posSys2], Delete[ newFF0, posFF], newSys], tinst],
			replaceTypes[ Extract[ newSys2, posSys2], tinst],
			replaceTypes[ Extract[ newFF0, posFF], tinst],
			$unifFix[ tag] /. tinst,
			Join[ DeleteCases[ $unifMeta[ tag], instVars -> _], newMeta] /. tinst,
			If[ tvar === Null, $unifTVar[ tag], tvar],
			replaceTypes[ Join[ normalize[ $unifSubst[ tag] /. Cases[ inst, _[ META$[ _, _, True], _]], True], DeleteCases[ inst, META$[ _, _, True] -> _]], tinst],
			Join[ tinst, $unifTSubst[ tag] /. tinst]
		}
	]]]
updateUP[ ___][ ___] := {}

instantiateEquation[ inst_, instVars_][ {a_, b_, rest___}] :=
	{instantiateEquation[ inst, instVars] @@ a, instantiateEquation[ inst, instVars] @@ b, rest}
instantiateEquation[ inst_, instVars_][ h_, args_, struct_List] :=
	Append[ splitHeadArgs[ normalize[ h /. inst, True], normalize[ args /. inst, True]], Replace[ DeleteCases[ struct, instVars], {} -> Indeterminate]]
instantiateEquation[ inst_, _][ h_, args_, struct_] :=
	Append[ splitHeadArgs[ normalize[ h /. inst, True], normalize[ args /. inst, True]], struct]

elimDecompUP[ {{Lambda$TM[ _, e_], {a_, r___}, struct_}, b_, rest__}, tag_] :=
	elimDecompUP[ {{betaReduce[ e, a], {r}, struct}, b, rest}, tag]
elimDecompUP[ {a_, b:{_Lambda$TM, {__}, _}, rest__}, tag_] :=
	elimDecompUP[ {b, a, rest}, tag]
elimDecompUP[ {{Lambda$TM[ RNG$[ TYPRNG$[ _, tx_]], a_], {}, astruct_}, {Lambda$TM[ RNG$[ TYPRNG$[ _, ty_]], b_], {}, bstruct_}, loose_, opts___}, tag_] :=
	Replace[ unifyTypes[ tx, ty, hasSort[ $unifTFix[ tag]], {$unifTVar[ tag], {}}, $unifTypeConstThreshold[ tag]],
		{
			{newTVar_Association, tunif_List} :>
				(* instantiations could kill loose bound variables, so pruning lose bound variables here (or in decomposition) does not really make sense:
					has to be done anyway when trying to make variable bindings *)
				{{{Append[ splitHeadArgs[ a], astruct], Append[ splitHeadArgs[ b], bstruct], Prepend[ loose, tx /. tunif], opts}}, {}, {}, {}, {}, tunif, newTVar},
			_ -> $Failed
		}
	]
elimDecompUP[ {{expr_Lambda$TM, {}, astruct_}, {bh_, bargs_, bstruct_}, loose_, opts___}, tag_] :=
	With[ {split = stripAbstractions[ expr, {}]},
	With[ {a = First[ split], types = Last[ split]},
	Replace[
		compareTypes[
			makeFunctionType[ types, getType[ a, Join[ types, loose], $unifFix[ tag], $unifMeta[ tag]]],
			getType[ bh @@ bargs, loose, $unifFix[ tag], $unifMeta[ tag]],
			hasSort[ $unifTFix[ tag]],
			{$unifTVar[ tag], {}},
			$unifTypeConstThreshold[ tag]
		],
		{
			{newTVar_Association, tunif_List} :>
				With[ {l = Length[ types]},
				With[ {new = {
								Append[ splitHeadArgs[ a], astruct],
								{mapFreeVars[ bh, VAR$[ #3 + l]&], Join[ mapFreeVars[ bargs, VAR$[ #3 + l]&], Table[ VAR$[ l - i], {i, l}]], bstruct},
								Join[ types, loose],
								opts
							}},
					If[ tunif === {},
						(* in this case 'newTVar === $unifTVar[ tag]' *)
						elimDecompUP[ new, tag],
					(*else*)
						{{new}, {}, {}, {}, {}, tunif, newTVar}
					]
				]],
			_ -> $Failed
		}
	]]]
elimDecompUP[ {a_, b:{_Lambda$TM, {}, _}, rest__}, tag_] :=
	elimDecompUP[ {b, a, rest}, tag]
elimDecompUP[ eq:{a:{v_META$, aargs_, astruct_}, b:{v_, bargs_, bstruct_}, loose_, opts___}, tag_] :=
	If[ TrueQ[ astruct] || (!MatchQ[ astruct, {__}] && MatchQ[ aargs, {(_VAR$|_FIX$)...}]),
		If[ aargs === bargs,
			{{}, {}, {}, {}, {}, {}, Null},	(* trivial *)
		(*else*)
			Replace[ Replace[ v, $unifMeta[ tag]],
				{
					{_, False} :>
						elimDecompUP[ {{Null, aargs, astruct}, {Null, bargs, bstruct}, loose, opts}, tag],
					{typ_, dep_} :>
						With[ {m = Length[ aargs]},
							If[ m === Length[ bargs],
								With[ {as = isHOPattern[ aargs, astruct, dep, tag], bs = isHOPattern[ bargs, bstruct, dep, tag]},
									If[ TrueQ[ as] && TrueQ[ bs],	(* flex-flex same *)
										With[ {types = flattenTypes[ typ, {}, m]},
											If[ ListQ[ types],
												With[ {newArgs = Position[ Transpose[ {aargs, bargs}], {x_, x_}, {1}, Heads -> False],
														newVar = META$[ First[ v], getFreshVarIndex[ v, tag], True],
														at = Most[ types], rt = Last[ types]},
													{
														{},
														{},
														{},
														{newVar -> {makeFunctionType[ Reverse[ Extract[ at, newArgs]], rt], dep}},
														{v -> applyCurried[ newVar, VAR$[ m - First[ #]]& /@ newArgs, Reverse[ at]]},
														{},
														Null
													}
												],
											(*else*)
												$Failed
											]
										],
									(*else*)
										{{}, {}, {{{v, aargs, as}, {v, bargs, bs}, loose, opts}}, {}, {}, {}, Null}	(* no higher-order pattern -> add to list of flex-flex pairs *)
									]
								],
							(*else*)
								$Failed
							]
						],
					_ -> $Failed
				}
			]
		],
	(*else*)
		{{}, {}, {eq}, {}, {}, {}, Null}
	]
elimDecompUP[ {a:{u_META$, uargs_, ustruct_}, b:{v_META$, vargs_, vstruct_}, loose_, opts___}, tag_] :=
	With[ {ui = Replace[ u, $unifMeta[ tag]], vi = Replace[ v, $unifMeta[ tag]]},
		Which[
			MatchQ[ ui, {_, False}],
			elimDecompUP[ {b, a}, vi, First[ ui], tag],
		(*else*)
			MatchQ[ vi, {_, False}],
			elimDecompUP[ {a, b}, ui, First[ vi], tag],
		(*else*)
			True,
			With[ {us = isHOPattern[ uargs, ustruct, Last[ ui], tag]},
				If[ TrueQ[ us],
					(* higher-order pattern *)
					If[ MatchQ[ uargs, {___VAR$}] && vargs === uargs,
						makeFFBinding[ {u, {}, ui}, {v, {}, vi}, tag],
					(*else*)
						With[ {vs = isHOPattern[ vargs, vstruct, Last[ vi], tag]},
							If[ TrueQ[ vs],
								makeFFBinding[ {v, vargs, vi}, {u, uargs, ui}, tag],
							(*else*)
								{{}, {}, {{{u, uargs, us}, {v, vargs, vs}, loose, opts}}, {}, {}, {}, Null}
							]
						]
					],
				(*else*)
					{{}, {}, {{{u, uargs, us}, {v, vargs, vstruct}, loose, opts}}, {}, {}, {}, Null}
				]
			]
		]
	]
elimDecompUP[ eq:{{v_META$, vargs_, vstruct_}, {bh_, bargs_, bstruct_}, loose_, opts___}, tag_] :=
	elimDecompUP[ eq, Replace[ v, $unifMeta[ tag]], getType[ bh, loose, $unifFix[ tag], $unifMeta[ tag]], tag]
elimDecompUP[ {b:{bh_, _, _}, a:{v_META$, _, _}, loose_, opts___}, tag_] :=
	elimDecompUP[ {a, b, loose, opts}, Replace[ v, $unifMeta[ tag]], getType[ bh, loose, $unifFix[ tag], $unifMeta[ tag]], tag]
elimDecompUP[ eq:{{v_META$, vargs_, vstruct_}, {bh_, bargs_, bstruct_}, loose_, opts___}, {typ_, dep:Except[ False]}, tbh:Except[ $Failed], tag_] :=
	(* this is an auxiliary definition of 'elimDecompUP', which takes two additional arguments: the info of 'v' and the type of 'bh' *)
	With[ {ft = flattenTypes[ typ, {}, Length[ vargs]]},
		If[ ft === $Failed,
			$Failed,
		(*else*)
			With[ {tv = Last[ ft], tb = stripArgumentTypes[ tbh, Length[ bargs]]},
				If[ tb === $Failed,
					$Failed,
				(*else*)
					Replace[ compareTypes[ tv, tb, hasSort[ $unifTFix[ tag]], {$unifTVar[ tag], {}}, $unifTypeConstThreshold[ tag]],
						{
							{newTVar_Association, tunif_List} :>
								If[ TrueQ[ bstruct] || bstruct === Indeterminate,
									With[ {vs = isHOPattern[ vargs, vstruct, dep, tag]},
										If[ TrueQ[ vs],
											Catch[	(* catch anything thrown below, e.g. '$Failed' during execution of 'scanFR[ b, tag]' *)
												With[ {b = applyCurried[ bh, bargs, {}]},
												With[ {scan1 =
														Replace[
															Last[ Reap[ $unifTempData[ tag] = {v, dep, vargs, Length[ vargs]}; scanFR[ b, tag], {"meta", "repl"}]],
															{lst_List} :> lst,
															{1}
														]},
													(* Get the positions and number of enclosing abstractions of all rigid occurrences of
														meta-variables in 'bargs', represented by rules '{var, info2} -> {{args1, abs1, pos1}, ...}'.
														Any of the positions 'pos1' etc., and all other positions below, are represented as pairs
														'{base, zeros}', where 'base' is a list of integers and 'zeros' is an integer such that
														the "standard" representation of positions can be obtained as
														'Join[ base, Table[ 0, {zeros}]]', and the last element of 'base' is different from '0'. *)
												With[ {metaOccs =
															Replace[
																GatherBy[ First[ scan1], First],
																all:{{var_, __}, ___} :>
																	With[ {info2 = Replace[ var, $unifMeta[ tag]]},
																		(* 'var' can certainly be instantiated, since otherwise 'scanFR' would not sow it *)
																		{var, info2} ->
																			Map[
																				With[ {args = argsFromPos[ Last[ #], b]},
																				With[ {struct0 = isHOPattern[ args, bstruct, Last[ info2], tag]},
																					If[ TrueQ[ struct0],
																						ReplacePart[ #, 1 -> args],
																					(*else*)
																						(* RHS is no higher-order pattern -> postpone equation to imitation/projection *)
																						Throw[ {{}, {{{v, vargs, vs}, {bh, bargs, struct0}, loose, opts}}, {}, {}, {}, tunif, newTVar}]
																					]
																				]]&,
																				all
																			]
																	],
																{1}
															],
														repl = Last[ scan1]},
													(* Construct suitable substitutions for 'v' and the meta variables appearing in 'b'.
														The basic idea is taken from
														
														[2] Gopalan Nadathur, Natalie Linnell. Practical Higher-Order Pattern Unification with On-the-Fly Raising.
															Logic Programming (ICLP 2005). LNCS volume 3668, pp 371-386, Springer, 2005. DOI 10.1007/11562931_28.
															
														However, for reasons of efficiency, the substitutions for the meta-variables appearing in 'b' are not
														constructed incrementally by looking at their occurrences one after the other, but at once. This happens
														in the one-argument, parameterized version of 'makeFFBinding'. *)
												With[ {scan2 =
														($unifFreshVarIndex[ tag] = {Indeterminate, v};
														$unifTempData[ tag] = {dep, vargs, Length[ vargs]};
														Reap[ Map[ makeFFBinding[ tag], metaOccs], {"inst", "new"}])},
													{
														{},
														{},
														{},
														Join @@ scan2[[2, 2]],
														Append[ Join @@ scan2[[2, 1]], v -> applyCurried[ ReplacePart[ b, Join[ repl, Join @@ First[ scan2]]], {}, Reverse[ Most[ ft]]]],
														tunif,
														newTVar
													}
												]]]]
											],
										(*else*)
											(* LHS is no higher-order pattern -> postpone equation to imitation/projection *)
											{{}, {{{v, vargs, vs}, {bh, bargs, bstruct}, loose, opts}}, {}, {}, {}, tunif, newTVar}
										]
									],
								(*else*)
									(* RHS is no higher-order pattern -> postpone equation to imitation/projection *)
									{{}, {eq}, {}, {}, {}, tunif, newTVar}
								],
							_ -> $Failed
						}
					]
				]
			]
		]
	]
elimDecompUP[ {{a:(_VAR$|_FIX$), {}, _}, {b_, {}, _}, __}, _] :=
	If[ a === b,
		{{}, {}, {}, {}, {}, {}, Null},
	(*else*)
		$Failed
	]
elimDecompUP[ {{_, {}, _}, {_VAR$|_FIX$, {}, _}, __}, _] :=
	$Failed
elimDecompUP[ {{A:TYPED$[ a_, _], {}, astruct_}, {B:TYPED$[ b_, _], {}, bstruct_}, rest__}, tag_] :=
	Replace[ compareTypes[ HoldComplete[ A], HoldComplete[ B], hasSort[ $unifTFix[ tag]], {$unifTVar[ tag], {}}, $unifTypeConstThreshold[ tag]],
		{
			{newTVar_Association, tunif_List} :>
				With[ {new = {Append[ splitHeadArgs[ a], astruct], Append[ splitHeadArgs[ b], bstruct], rest}},
					If[ tunif === {},
						(* in this case '$unifTVar[ tag] === newTVar' *)
						elimDecompUP[ new, tag],
					(*else*)
						{{new}, {}, {}, {}, {}, tunif, newTVar}
					]
				],
			_ -> $Failed
		}
	]
elimDecompUP[ eq:{{ah_, aargs:Except[ {}], astruct_}, {bh_, bargs_, bstruct_}, loose_, opts___}, tag_] :=
	If[ Length[ aargs] === Length[ bargs],
		With[ {as = If[ TrueQ[ astruct], True, Indeterminate], bs = If[ TrueQ[ bstruct], True, Indeterminate]},
			{
				Prepend[ MapThread[ {Append[ splitHeadArgs[ #1], as], Append[ splitHeadArgs[ #2], bs], loose, opts}&, {aargs, bargs}], {{ah, {}, as}, {bh, {}, bs}, loose, opts}],
				{},
				{},
				{},
				{},
				{},
				Null
			}
		],
	(*else*)
		$Failed
	]
elimDecompUP[ {_, b:{_, Except[ {}], _}, __}, _] :=
	$Failed
elimDecompUP[ {{A:TYPED$[ a_, _], {}, astruct_}, b:{bh_, bargs_, _}, loose_, opts___}, tag_] :=
	With[ {tb = stripArgumentTypes[ getType[ bh, loose, $unifFix[ tag], $unifMeta[ tag]], Length[ bargs]]},
		Replace[ compareTypes[ HoldComplete[ A], HoldComplete[ TYPED$[ Null, tb]], hasSort[ $unifTFix[ tag]], {$unifTVar[ tag], {}}, $unifTypeConstThreshold[ tag]],
			{
				{newTVar_Association, tunif_List} :>
					With[ {new = {Append[ splitHeadArgs[ a], astruct], b, loose, opts}},
						If[ tunif === {},
							(* in this case '$unifTVar[ tag] === newTVar' *)
							elimDecompUP[ new, tag],
						(*else*)
							{{new}, {}, {}, {}, {}, tunif, newTVar}
						]
					],
				_ -> $Failed
			}
		]
	]
elimDecompUP[ {a_, b:{_TYPED$, {}, _}, rest__}, tag_] :=
	elimDecompUP[ {b, a, rest}, tag]
elimDecompUP[ {{a_, {}, _}, {b_, {}, _}, __}, _] :=
	If[ a === b,
		{{}, {}, {}, {}, {}, {}, Null},
	(*else*)
		$Failed
	]
elimDecompUP[ ___] := $Failed

imitProjUP[ {_, {_Lambda$TM, _, _}, ___}, _] :=
	{}		(* system is not beta-reduced *)
imitProjUP[ {{v:(META$[ s_, __]), vargs_, vstruct_}, {bh_, bargs_, bstruct_}, loose_, opts___}, tag_] :=
	With[ {vi = getFreshVarIndex[ v, tag],
			vargt = Catch[ Map[ With[ {ty = getType[ #, loose, $unifFix[ tag], $unifMeta[ tag]]}, If[ ty === $Failed, Throw[ $Failed], ty]]&, vargs]],
			bargt = Catch[ Map[ With[ {ty = getType[ #, loose, $unifFix[ tag], $unifMeta[ tag]]}, If[ ty === $Failed, Throw[ $Failed], ty]]&, bargs]],
			n = Length[ vargs], m = Length[ bargs]},
		If[ vargt === $Failed || bargt === $Failed,
			{},
		(*else*)
			With[ {rev = Reverse[ vargt]},
			With[ {info = Replace[ v, $unifMeta[ tag]],	(* 'v' can certainly be instantiated, since otherwise 'imitProjUP' would not be called on it *)
				template = applyCurried[ v, {}, rev],	(* 'n' abstractions around 'v' ('v' only serves as a placeholder here; is replaced below) *)
					templateT = makeFunctionType[ rev, 0],	(* 'vargt[[1]] -> ... -> vargt[[n]] -> 0' ('0' only serves as a placeholder here) *)
					pos = {Table[ 2, {n}]},
					templateV = applyCurried[ META$[ s, 0, True], Table[ VAR$[ n - k], {k, n}], {}],	(* 'META$[ s, 0, True][ VAR$[ n-1]]...[ VAR$[ 0]]' *)
					posV = Append[ Table[ 0, {n}], 2]},
			With[ {vrt = stripArgumentTypes[ First[ info], n], dep = Last[ info],	(* 'vrt' is the type of 'v' after stripping off the first 'n' argument types *)
					bs = If[ TrueQ[ bstruct] || bstruct === False, bstruct, Indeterminate]},
				If[ vrt === $Failed,
					$Failed,
				(*else*)
					Join[
						
						(* imitation *)
						If[ mayDependOn[ dep, bh, tag],
							(* imitation is possible *)
							{
								With[ {inst = {v -> ReplacePart[
														template,
														pos -> applyCurried[ bh, Table[ ReplacePart[ templateV, posV -> vi + k], {k, 0, m - 1}], {}]
												]},
										vs = If[ TrueQ[ vstruct] || vstruct === False, vstruct, Indeterminate]},
									{
										MapIndexed[
											{
												{META$[ s, vi + First[ #2] - 1, True], normalize[ vargs /. inst, True], vs},
												Append[ splitHeadArgs[ normalize[ #1 /. inst, True]], bs],
												loose,
												opts
											}&,
											bargs
										],		(* new equations *)
										MapIndexed[
											(META$[ s, vi + First[ #2] - 1, True] -> {ReplacePart[ templateT, pos -> #1], dep}) &,
											bargt
										],		(* new meta variables *)
										inst,	(* term substitution *)
										{},	(* type substitution *)
										Null	(* sort constraints *)
									}
								]
							},
						(*else*)
							(* 'b' is a loose bound variable or an illegal a.b.f. constant --> no imitation possible *)
							{}
						],
						
						(* projections *)
						With[ {templateVargs = applyCurried[ META$[ s, 0, True], vargs, {}],	(* '0' only serves as a placeholder here; is replaced below *)
								vs = If[ TrueQ[ vstruct], True, Indeterminate]},
							MapThread[
								Function[ {a, t, i},	(* 'a' is an argument of 'v' in 'vargs', 't' is its type, 'i' is its position in the list *)
									With[ {at = flattenTypes[ t, {}]},		(* 'at' is the list of argument- and result-types of 'a' *)
										Replace[ compareTypes[ Last[ at], vrt, hasSort[ $unifTFix[ tag]], {$unifTVar[ tag], {}}, $unifTypeConstThreshold[ tag]],
											{
												{newTVar_Association, tunif_List} :>
													With[ {p = Length[ at] - 1},
													With[ {inst = {v -> ReplacePart[
																			template,
																			pos -> applyCurried[ VAR$[ n - i], Table[ ReplacePart[ templateV, posV -> vi + k], {k, 0, p - 1}], {}]
																	]}
															},
														{
															{
																{
																	Append[
																		splitHeadArgs[
																			normalize[ a /. inst, True],
																			normalize[ Table[ ReplacePart[ templateVargs, posV -> vi + k], {k, 0, p - 1}] /. inst, True]
																		],
																		vs
																	],
																	{bh, normalize[ bargs /. inst, True], bs},	(* 'bh' does not need to be instantiated, since it certainly differs from 'v' *)
																	loose,
																	opts
																}
															},		(* new equations *)
															MapIndexed[
																(META$[ s, vi + First[ #2] - 1, True] -> {ReplacePart[ templateT, pos -> #1], dep}) &,
																Most[ at]
															],		(* new meta variables *)
															inst,	(* term substitution *)
															tunif,	(* type substitution *)
															newTVar	(* sort constraints *)
														}
													]],
												_ -> $Failed
											}
										]
									]
								],
								{vargs, vargt, Range[ n]}
							]
						]
					]
				]
			]]]
		]
	]
imitProjUP[ ___] := {}

(* 'isHOPattern' returns either 'True' or 'False' or a list of meta variables whose instantiation could
	possibly render the argument list a higher-order pattern. *)
isHOPattern[ args_, Indeterminate, dep_, tag_] :=
	If[ MatchQ[ args, {___, x:(_VAR$|_FIX$), ___, x_, ___}] || MemberQ[ args, _FIX$?(mayDependOn[ dep, tag])],
		False,
	(*else*)
		With[ {thinned = Cases[ args, e:Except[ _VAR$|_FIX$] :> First[ getHeadAtom[ e, 0]]]},
			Switch[ thinned,
				{},
				True,
				{__META$},
				DeleteDuplicates[ thinned],
				_,
				False
			]
		]
	]
isHOPattern[ _, struct_, _, _] :=
	struct

makeFFBinding[ {u_META$, {}, {tu_, du_}}, {v_META$, {}, {tv_, dv_}}, tag_] :=
	Replace[ compareTypes[ tu, tv, hasSort[ $unifTFix[ tag]], {$unifTVar[ tag], {}}, $unifTypeConstThreshold[ tag]],
		{
			{newTVar_Association, tunif_List} :>
				With[ {interDep = intersectDependencies[ du, dv]},
					Switch[ interDep,
						du,
						{{}, {}, {}, {}, {v -> u}, tunif, newTVar},
					(*else*)
						dv,
						{{}, {}, {}, {}, {u -> v}, tunif, newTVar},
					(*else*)
						_,
						With[ {newVar = META$[ First[ u], getFreshVarIndex[ u, tag], True]},
							{{}, {}, {}, {newVar -> {tu, interDep}}, {u -> newVar, v -> newVar}, tunif, newTVar}
						]
					]
				],
			_ -> $Failed
		}
	]
makeFFBinding[ a:{u_META$, uargs_, {tu_, du_}}, b:{v_META$, vargs_, {tv_, dv_}}, tag_] :=
	With[ {m = Length[ uargs], n = Length[ vargs]},
	With[ {tu1 = flattenTypes[ tu, {}, m], tv1 = flattenTypes[ tv, {}, n]},
		If[ tu1 === $Failed || tv1 === $Failed,
			$Failed,
		(*else*)
			With[ {atu = Most[ tu1], atv = Most[ tv1], rtu = Last[ tu1], rtv = Last[ tv1]},
				Replace[ compareTypes[ rtu, rtv, hasSort[ $unifTFix[ tag]], {$unifTVar[ tag], {}}, $unifTypeConstThreshold[ tag]],
					{
						{newTVar_Association, tunif_List} :>
							With[ {interMap =
								DeleteCases[
									MapIndexed[
										With[ {i = Position[ vargs, #1, {1}, 1, Heads -> False]},
											If[ i === {},
												Null,
											(*else*)
												First[ #2] -> i[[1, 1]]
											]
										]&,
										uargs
									],
									Null
								]},
							With[ {inter1 = First /@ interMap, inter2 = Last /@ interMap},
							With[ {consts1 = Position[ uargs, _FIX$?(mayDependOn[ dv, tag]), {1}, Heads -> False][[All, 1]],
									consts2 = Position[ vargs, _FIX$?(mayDependOn[ du, tag]), {1}, Heads -> False][[All, 1]],
									newVar = META$[ First[ u], getFreshVarIndex[ u, tag], True]},
								{
									{},
									{},
									{},
									{
										newVar ->
											{
												makeFunctionType[ Join[ Reverse[ Part[ atu, inter1]], Reverse[ Part[ atv, consts2]], Reverse[ Part[ atu, consts1]]], rtu],
												intersectDependencies[ du, dv]
											}
									},
									{
										u -> applyCurried[ newVar, Join[ VAR$[ m - #]& /@ consts1, Part[ vargs, consts2], VAR$[ m - #]& /@ inter1], Reverse[ atu]],
										v -> applyCurried[ newVar, Join[ Part[ uargs, consts1], VAR$[ n - #]& /@ consts2, VAR$[ n - #]& /@ inter2], Reverse[ atv]]
									},
									tunif,
									newTVar
								}
							]]],
						_ -> $Failed
					}
				]
			]
		]
	]]
(* Before calling 'makeFFBinding' with a single argument, '$unifTempData[ tag]' shall be set to the 3-tuple '{du, uargs, m}', where
	- 'du' are the dependencies of the main variable to be eliminated,
	- 'uargs' is its argument list, and
	- 'm' is the length of 'uargs'.
	Moreover, '$unifFreshVarIndex[ tag]' shall be the pair '{Indeterminate, u'}, where 'u' is the main variable to be eliminated.
	Below, 'l' is a non-empty list of elements of the form '{args, abs, pos}'.
	Actually, the two definitions of 'makeFFBinding' above and below are quite similar, the one above being kind of a special case
	of the one below. The one above is kept mainly for efficiency reasons. *)
makeFFBinding[ tag_][ {v_, {tv_, dv_}} -> l_List] :=
	With[ {m = $unifTempData[ tag][[3]],
			constIndices = Position[ $unifTempData[ tag][[2]], _FIX$?(mayDependOn[ dv, tag]), {1}, Heads -> False][[All, 1]]},
	With[ {consts = Part[ $unifTempData[ tag][[2]], constIndices], constNum = Length[ constIndices],
			interDep = intersectDependencies[ First[ $unifTempData[ tag]], dv]},
	(*With[ {addDep = addDependencies[ intersectDependencies[ First[ $unifTempData[ tag]], dv], consts0]},
	With[ {constIndices = Extract[ constIndices0, Last[ addDep]], consts = Extract[ consts0, Last[ addDep]]},*)
		(* Rather than explicitly listing all a.b.f. constants appearing among the arguments of 'u' in the arguments
			of the instance of 'v', it might be better to instead allow 'newVar' to depend on these constants. *)
	With[ {newArgs =
				Reap[
					Map[
						With[ {abs = #[[2]]},
							{
								#[[3, 1]],	(* recall that the position is given as '{base, zeros}', but we are only interested in 'base' *)
								Join[
									VAR$[ abs + m - #]& /@ constIndices,
									MapIndexed[
										Function[ {a, i},
											Switch[ a,
												_FIX$,
												If[ mayDependOn[ First[ $unifTempData[ tag]], a, tag],
													a,
												(*else*)
													Replace[
														Position[ $unifTempData[ tag][[2]], a, {1}, 1, Heads -> False],
														{{{j_}} :> VAR$[ abs + m - j], _ :> Sow[ constNum + First[ i]]}
													]
												],
												_VAR$,
												If[ First[ a] < abs,
													a,
												(*else*)
													Replace[
														Position[ $unifTempData[ tag][[2]], VAR$[ First[ a] - abs], {1}, 1, Heads -> False],
														{{{j_}} :> VAR$[ abs + m - j], _ :> Sow[ constNum + First[ i]]}
													]
												]
											]
										],
										First[ #]
									]
								]
							}
						]&,
						l
					]
				]},
	With[ {badIndices = Union @@ Last[ newArgs], constTypes = Reverse[ Replace[ consts, $unifFix[ tag], {1}]]},
		If[ badIndices === {},
			If[ consts === {} && interDep === dv,
				Replace[ First[ newArgs], {pos_, args_} :> (pos -> applyCurried[ v, args, {}]), {1}],
			(*else*)
				With[ {newVar = getFreshVar[ tag]},
					Sow[ newVar -> {makeFunctionType[ constTypes, tv], interDep}, "new"];
					Sow[ v -> applyCurried[ newVar, consts, {}], "inst"];
					Replace[ First[ newArgs], {pos_, args_} :> (pos -> applyCurried[ newVar, args, {}]), {1}]
				]
			],
		(*else*)
			With[ {badMax = Last[ badIndices], newVar = getFreshVar[ tag]},
			With[ {ftv = flattenTypes[ tv, {}, badMax - constNum]},
				If[ ftv === $Failed,
					Throw[ $Failed],
				(*else*)
					With[ {atv = Reverse[ Most[ ftv]], badIndices2 = List /@ badIndices},
						Sow[ newVar -> {makeFunctionType[ Delete[ Join[ atv, constTypes], List /@ -badIndices], Last[ ftv]], interDep}, "new"];
						Sow[ v -> applyCurried[ newVar, Delete[ Join[ consts, Table[ VAR$[ badMax - k], {k, constNum + 1, badMax}]], badIndices2], atv], "inst"];
						Replace[
							First[ newArgs],
							{pos_, args_} :>
								With[ {diff = badMax - Length[ args]},
									pos ->
										If[ diff > 0,
											applyCurried[
												newVar,
												Delete[
													Join[ Replace[ args, VAR$[ k_] :> VAR$[ k + diff], {1}], Table[ VAR$[ diff - k], {k, 1, diff}]],
													badIndices2
												],
												Take[ atv, diff]	(* 'atv' is already in reversed order *)
											],
										(*else*)
											applyCurried[ newVar, Delete[ args, badIndices2], {}]
										]
								],
							{1}
						]
					]
				]
			]]
		]
	]]]]


(* ::Subsection:: *)
(*Syntactic Unification*)

unifySynAux[ {u_META$, v_META$}, tag_] :=
	If[ u === v,
		{{}, {}, {}, {}, Null},
	(*else*)
		With[ {ui = Replace[ u, $unifMeta[ tag]], vi = Replace[ v, $unifMeta[ tag]]},
			Which[
				MatchQ[ ui, {_, False}],
				If[ MatchQ[ vi, {_, False}],
					$Failed,
				(*else*)
					makeSynInst[ v, u, vi, First[ ui], tag]
				],
			(*else*)
				MatchQ[ vi, {_, False}],
				makeSynInst[ u, v, ui, First[ vi], tag],
			(*else*)
				True,
				With[ {du = Last[ ui], dv = Last[ vi]},
					Replace[ compareTypes[ First[ ui], First[ vi], hasSort[ $unifTFix[ tag]], {$unifTVar[ tag], {}}, $unifTypeConstThreshold[ tag]],
						{
							{newTVar_Association, tunif_List} :>
								With[ {interDep = intersectDependencies[ du, dv]},
									Switch[ interDep,
										du,
										{{}, {}, {v -> u}, tunif, newTVar},
										
										dv,
										{{}, {}, {u -> v}, tunif, newTVar},
										
										_,
										With[ {newVar = META$[ First[ u], getFreshVarIndex[ u, tag], True]},
											{{}, {}, {}, {newVar -> {First[ ui], interDep}}, {u -> newVar, v -> newVar}, tunif, newTVar}
										]
									]
								],
							_ -> $Failed
						}
					]
				]
			]
		]
	]
unifySynAux[ {v_META$, b_}, tag_] :=
	makeSynInst[ v, b, Replace[ v, $unifMeta[ tag]], getType[ b, {}, $unifFix[ tag], $unifMeta[ tag]], tag]	(* there must not be any loose bound variables in 'b' *)
unifySynAux[ {a_, v_META$}, tag_] :=
	unifySynAux[ {v, a}, tag]
unifySynAux[ {a:(_FIX$|_VAR$), b_}, tag_] :=
	If[ a === b,
		{{}, {}, {}, {}, Null},
	(*else*)
		$Failed
	]
unifySynAux[ {_, _FIX$|_VAR$}, _] :=
	$Failed
unifySynAux[ {Lambda$TM[ RNG$[ TYPRNG$[ _, typ1_]], a_], b0_}, tag_] :=
	Replace[ b0,
		{
			Lambda$TM[ RNG$[ TYPRNG$[ _, typ2_]], b_] :>
				Replace[ unifyTypes[ typ1, typ2, hasSort[ $unifTFix[ tag]], {$unifTVar[ tag], {}}, $unifTypeConstThreshold[ tag]],
					{
						{_, {}} :> unifySynAux[ {a, b}, tag],
						{newTVar_Association, tunif_List} :> {{{a, b}}, {}, {}, tunif, newTVar},
						_ -> $Failed
					}
				],
			_ :> $Failed
		}
	]
unifySynAux[ {a0:TYPED$[ a_, _], b0_}, tag_] :=
	Replace[ b0,
		{
			TYPED$[ b_, _] :>
				Replace[ compareTypes[ HoldComplete[ a0], HoldComplete[ b0], hasSort[ $unifTFix[ tag]], {$unifTVar[ tag], {}}, $unifTypeConstThreshold[ tag]],
					{
						{_Association, {}} :> unifySynAux[ {a, b}, tag],
						{newTVar_Association, tunif_List} :> {{{a, b}}, {}, {}, tunif, newTVar},
						_ -> $Failed
					}
				],
			_ :> $Failed
		}
	]
unifySynAux[ {f_[ a_], b0_}, tag_] :=
	Replace[ b0,
		{
			g_[ b_] :> {{{f, g}, {a, b}}, {}, {}, {}, Null},
			_ :> $Failed
		}
	]
unifySynAux[ {a_, b_}, _] :=
	If[ a === b,
		{{}, {}, {}, {}, Null},
	(*else*)
		$Failed
	]
unifySynAux[ _, _] :=
	$Failed

makeSynInst[ _, _ {_, False}, _, _] :=
	$Failed
makeSynInst[ _, _ _, $Failed, _] :=
	$Failed
makeSynInst[ v_META$, t_, {vt_, dep_}, tt_, tag_] :=
	If[ FreeQ[ t, v] && !containsVar[ t, #3 >= #2 &] && (dep === Infinity || FreeQ[ t, _FIX$?mayNotDependOn[ dep, tag]]),
		Replace[ compareTypes[ vt, tt, hasSort[ $unifTFix[ tag]], {$unifTVar[ tag], {}}, $unifTypeConstThreshold[ tag]],
			{
				{newTVar_Association, tunif_List} :> {{}, {}, {v -> t}, tunif, newTVar},
				_ -> $Failed
			}
		],
	(*else*)
		$Failed
	]

updateUPSyn[ system_, {new_, newMeta_, {}, tinst_, tvar_}, tag_] :=
	(
		$unifFix[ tag] = $unifFix[ tag] /. tinst;
		$unifMeta[ tag] = Join[ $unifMeta[ tag], newMeta] /. tinst;
		If[ tvar =!= Null, $unifTVar[ tag] = tvar];
		$unifSubst[ tag] = replaceTypes[ $unifSubst[ tag], tinst];
		$unifTSubst[ tag] = Join[ tinst, $unifTSubst[ tag] /. tinst];
		replaceTypes[ Join[ new, system], tinst]
	)
updateUPSyn[ system_, {new_, newMeta_, inst_, tinst_, tvar_}, tag_] :=
	(
		$unifFix[ tag] = $unifFix[ tag] /. tinst;
		$unifMeta[ tag] = Join[ DeleteCases[ $unifMeta[ tag], Alternatives @@ inst[[All, 1]] -> _], newMeta] /. tinst;
		If[ tvar =!= Null, $unifTVar[ tag] = tvar];
		$unifSubst[ tag] = replaceTypes[ Join[ normalize[ $unifSubst[ tag] /. Cases[ inst, _[ META$[ _, _, True], _]], True], DeleteCases[ inst, META$[ _, _, True] -> _]], tinst];
		$unifTSubst[ tag] = Join[ tinst, $unifTSubst[ tag] /. tinst];
		replaceTypes[ Join[ new, Map[ normalize[ # /. inst, True]&, system, {2}]], tinst]
	)
updateUPSyn[ ___] := $Failed


(* ::Subsection:: *)
(*Auxiliary Functions*)

(* Before calling 'scanFR', '$unifTempData[ tag]' shall be set to the 4-tuple '{var, dep, args, argnum}', where
	- 'var' is the meta-variable which is the head of the flexible expression,
	- 'dep' are the dependencies of 'var',
	- 'args' is the argument list of 'var', and
	- 'argnum' is the length of 'args'. *)
scanFR[ expr_, tag_] := (scanFR[ expr, 0, {}, tag];)
scanFR[ v_META$, abs_, pos_, tag_] :=
	Which[
		v === First[ $unifTempData[ tag]],
		Throw[ $Failed],
	(*else*)
		MatchQ[ Replace[ v, $unifMeta[ tag]], {_, False}],
		False,
	(*else*)
		True,
		With[ {z = countTrailingElements[ pos, 0]},
			Sow[ {v, abs, {Drop[ pos, -z], z}}, "meta"]
		];
		True
	]
scanFR[ c_FIX$, abs_, pos_, tag_] :=
	If[ mayDependOn[ $unifTempData[ tag][[2]], c, tag],
		False,
	(*else*)
		With[ {i = Position[ $unifTempData[ tag][[3]], c, {1}, 1, Heads -> False]},
			If[ i === {},
				Throw[ $Failed],
			(*else*)
				Sow[ {pos} -> VAR$[ abs + $unifTempData[ tag][[4]] - i[[1, 1]]], "repl"];
				False	
			]
		]
	]
scanFR[ VAR$[ j_Integer], abs_, pos_, tag_] :=
	If[ j < abs,
		False,
	(*else*)
		With[ {i = Position[ $unifTempData[ tag][[3]], VAR$[ j - abs], {1}, 1, Heads -> False]},
			If[ i === {},
				Throw[ $Failed],
			(*else*)
				Sow[ {pos} -> VAR$[ abs + $unifTempData[ tag][[4]] - i[[1, 1]]], "repl"];
				False
			]
		]
	]
scanFR[ Lambda$TM[ _, e_], abs_, pos_, tag_] := (scanFR[ e, abs + 1, Append[ pos, 2], tag]; False)
scanFR[ TYPED$[ e_, _], abs_, pos_, tag_] := (scanFR[ e, abs, Append[ pos, 1], tag]; False)
scanFR[ f_[ a___], abs_, pos_, tag_] :=
	If[ TrueQ[ scanFR[ f, abs, Append[ pos, 0], tag]],
		True,
	(*else*)
		Module[ {i = 0},
			Scan[
				scanFR[ #, abs, Append[ pos, ++i], tag]&,
				{a}
			]
		];
		False
	]
scanFR[ ___] := False

getFreshVarIndex[ _[ s_Symbol, i_Integer, ___], tag_] :=
	Max[ i, Max[ Cases[ $unifMeta[ tag], (_[ s, j_Integer, ___] -> _) :> j]], Max[ Cases[ $unifSubst[ tag], (_[ s, j_Integer, ___] -> _) :> j]]] + 1
getFreshVarIndex[ s_Symbol, tag_] :=
	getFreshVarIndex[ META$[ s, -1], tag]

getFreshVar[ tag_] :=
	With[ {fvi = $unifFreshVarIndex[ tag]},
	With[ {i0 = First[ fvi]},
		If[ IntegerQ[ i0],
			$unifFreshVarIndex[ tag] = ReplacePart[ fvi, 1 -> i0 + 1];
			META$[ fvi[[2, 1]], i0, True],
		(*else*)
			With[ {i = getFreshVarIndex[ fvi[[2]], tag]},
				$unifFreshVarIndex[ tag] = ReplacePart[ fvi, 1 -> i + 1];
				META$[ fvi[[2, 1]], i, True]
			]
		]
	]]

mayDependOn[ dep_, tag_][ c_] :=
	mayDependOn[ dep, c, tag]	(* useful for pattern matching *)
mayDependOn[ _, b:Except[ _FIX$], _] :=
	!MatchQ[ b, _VAR$]
mayDependOn[ Infinity, _, _] :=
	True
mayDependOn[ dep_Integer, c_, tag_] :=
	MatchQ[ Position[ $unifFix[ tag], c -> _, {1}, 1, Heads -> False], {{i_ /; i <= dep}}]
mayDependOn[ {dep_Integer, ex_List}, c_, tag_] :=
	With[ {p = Position[ $unifFix[ tag], c -> _, {1}, 1, Heads -> True]},
		If[ p === {},
			False,
		(*else*)
			p[[1, 1]] <= dep && !MemberQ[ ex, p[[1, 1]]]
		]
	]

mayNotDependOn[ dep_, tag_][ c_] :=
	!mayDependOn[ dep, c, tag]

intersectDependencies[ m1_, m2_, meta_] :=
	intersectDependencies[ Last[ Replace[ m1, meta]], Last[ Replace[ m2, meta]]]
intersectDependencies[ dep:{(_Integer|Infinity)...}] :=
	Min @@ dep
intersectDependencies[ dep:{(_List|Infinity)..}] :=
	With[ {dep0 = DeleteCases[ dep, Infinity]},
		{Min @@ dep0[[All, 1]], Union @@ dep0[[All, 2]]}
	]
intersectDependencies[ dep_List] :=
	intersectDependencies[ Min @@ Cases[ dep, _Integer], intersectDependencies[ Cases[ dep, _List]]]
intersectDependencies[ dep1:(_Integer|Infinity), dep2:(_Integer|Infinity)] :=
	Min[ dep1, dep2]
intersectDependencies[ 0, _List] :=
	0
intersectDependencies[ Infinity, dep2_List] :=
	dep2
intersectDependencies[ dep1_, {}] :=
	dep1
intersectDependencies[ dep1:(_Integer|Infinity), {dep2:(_Integer|Infinity), ex2_List}] :=
	{Min[ dep1, dep2], ex2}
intersectDependencies[ dep1_List, dep2:(_Integer|Infinity)] :=
	intersectDependencies[ dep2, dep1]
intersectDependencies[ {dep1:(_Integer|Infinity), ex1_List}, {dep2:(_Integer|Infinity), ex2_List}] :=
	{Min[ dep1, dep2], Union[ ex1, ex2]}

(* 'argsFromPos[ pos, expr]' returns the argument list of the subexpression of 'expr'
	specified by 'pos'. 'pos' may be given either as a list of integers, or as a pair
	'{base, zeros}'. *)
argsFromPos[ {_List, 0}, _] :=
	{}
argsFromPos[ {{}, zeros_Integer}, expr_] :=
	argsFromPos[ zeros, expr, {}]
argsFromPos[ {base_List, zeros_Integer}, expr_] :=
	argsFromPos[ zeros, Extract[ expr, base], {}]
argsFromPos[ pos_List, expr_] :=
	With[ {l = Length[ pos]},
	With[ {i = countTrailingElements[ pos, 0, l]},
		If[ i === 0,
			{},
		(*else*)
			argsFromPos[ i, If[ i === l, expr, Extract[ expr, Drop[ pos, -i]]], {}]
		]
	]]
argsFromPos[ 0, _, acc_] :=
	acc
argsFromPos[ n_Integer, expr_, acc_] :=
	argsFromPos[ n - 1, Head[ expr], Join[ List @@ expr, acc]]

countTrailingElements[ l_List, patt_] :=
	countTrailingElements[ l, patt, Length[ l]]
countTrailingElements[ l_List, patt_, len_] :=
	Replace[ Position[ Reverse[ l], Except[ patt], {1}, 1, Heads -> False], {{} -> len, {{i_}} :> i - 1}]

deleteTrailingElements[ l_List, patt_] :=
	l[[;; Replace[ Position[ Reverse[ l], Except[ patt], {1}, 1, Heads -> False], {{} -> 0, {{i_}} :> -i}]]]

End[]

EndPackage[]
