(* ::Package:: *)

(* Theorema
    Copyright (C) 2017 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Core`Tools`HORewriting`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]
Needs[ "Theorema`Language`"]

Begin["`Private`"]

(* ::Section:: *)
(*Simple (for Parsing and Pretty-Printing)*)

(* 'generateSimpleRewriteRule[ mode][ lhs, rhs, tvar, Hold[ ctxt]]' generates a Mathematica transformation rule that
	turns 'lhs' into 'rhs'; outermost lambda abstractions in 'lhs' are stripped off and turned into free variables.
	Argument 'mode' specifies whether higher-order (pattern) matching shall be employed to instantiate free term
	variables ("higher-order"), whether mere syntactic matching shall be applied ("syntactic"), or whether 'lhs' and
	'rhs' are in fact types ("type"). AT THE MOMENT, "higher-order" AND "syntactic" HAVE THE SAME EFFECT!
	In any case, types are only matched rather than unified, and all type variables occurring in 'rhs' must also appear
	in 'lhs'; if this condition is violated, the function returns '$Failed'.
	'ctxt' is a global variable holding the "runtime" context information: before the rules generated by this function
	are applied, 'ctxt' must be set to the list '{tf, tv}', where 'tf' and 'tv' are the sort constraints at runtime. *)
generateSimpleRewriteRule[ "type"][ lhs_, rhs_, _, _] :=
	With[ {tvars = DeleteDuplicates[ Cases[ lhs, _TVAR$, {0, Infinity}, Heads -> True]]},
	With[ {tvarPattRepl = MapIndexed[ (#1 -> ToExpression[ "t$" <> ToString[ First[ #2]] <> "$$$_"])&, tvars]},
	With[ {tvarInstRepl = Replace[ tvarPattRepl, (l_ -> r_Pattern) :> (l -> Extract[ r, {1}, Hold]), {1}]},
	With[ {rhs1 = (HoldComplete @@ {rhs /. tvarInstRepl}) /. Hold[ x_] :> x},
		If[ MemberQ[ rhs1, _TVAR$, {0, Infinity}, Heads -> True],
			$Failed,
		(*else*)
			RuleDelayed @@ Prepend[ rhs1, lhs /. tvarPattRepl]
		]
	]]]]
generateSimpleRewriteRule[ _][ lhs_, rhs_, tvar_, ctxt_Hold] :=
	With[ {ac = abstractionCount[ lhs, 0, Infinity], lhs1 = stripAbstractions[ lhs]},
	With[ {termVars = Table[ ToExpression[ "x$" <> ToString[ i] <> "$$$", InputForm, Hold], {i, 1, ac}]},
		Replace[ makeBoundTypePattern[ If[ ac === 0, lhs1, mapFreeVars[ lhs1, (Pattern @@ Append[ termVars[[#2 - (#3 + 1)]], Blank[]])&]], rhs, False],
			{patt_, tvarInstRepl_, boundRepl0_} :>
				With[ {boundRepl = Replace[ boundRepl0, (l_ -> r_) :> (TYPRNG$[ l, t_] :> TYPRNG$[ r, t]), {1}]},
				With[ {rhs1 = (HoldComplete @@ {applyCurried[ rhs /. boundRepl /. tvarInstRepl, termVars, {}]}) /. Hold[ x_] :> x},
					If[ MemberQ[ rhs1, _TVAR$, {0, Infinity}, Heads -> True],
						$Failed,
					(*else*)
						RuleDelayed @@ Prepend[ rhs1, makeSortPatternCond[ patt, tvarInstRepl, tvar, ctxt]]
					]
				]]
		]
	]]

makeSortPatternCond[ patt_, tvarInstRepl_List, tvar_, ctxt_Hold] :=
	makeSortPatternCond[
		patt,
		Cases[ tvarInstRepl,
			(x_TVAR$ -> _[ v_, ___]) :>
				With[ {sort = Lookup[ tvar, x, getDefaultSort[]]},
					Hold[ v] -> sort /; sort =!= {}
				]
		],
		ctxt
	]
makeSortPatternCond[ patt_, repl_List, Hold[ ctxt_Symbol]] :=
	Replace[ makeSortPatternCond[ patt, repl, Hold[ {tfix$$$, tvar$$$}]],
		h_[ p_, c_] /; h === Condition :>
			ReplacePart[ Condition[ p, {{tfix$$$ = ctxt[[-2]], tvar$$$ = ctxt[[-1]]}, c}], {2, 0} -> With]
	]
makeSortPatternCond[ patt_, repl_List, Hold[ {tf_, tv_}]] :=
	With[ {cond0 = Join @@ Cases[ repl, (_[ v_, ___] -> sort:Except[ {}, _List]) :> Hold[ hasSort[ tf][ v, sort, tv, Infinity] =!= False]]},
		If[ MatchQ[ cond0, _[]],
			patt,
		(*else*)
			ReplacePart[ Condition[ patt, cond0], {2, 0} -> And]
		]
	]

(* 'makeBoundTypePattern[ lhs, rhs, unifTypes]' replaces types (resp. type variables, depending on 'unifTypes'), as
	well as names of bound variables, in 'lhs' by Mathematica pattern variables. Bound variables having a name that
	also appears in 'rhs' are replaced by a named pattern.
	In addition to the pattern-object, the function also returns the output of 'makeAlphaEquivFun[ lhs, unifyTypes]'
	and the renaming substitution of bound variables. *)
makeBoundTypePattern[ lhs_, rhs_, unifyTypes_] :=
	With[ {bnames = DeleteDuplicates[ Cases[ rhs, TYPRNG$[ sym:Except[ Null, _Symbol], _] :> sym, {0, Infinity}, Heads -> True]]},
	With[ {bpos =
			DeleteCases[
				Map[ (# -> Append[ FirstPosition[ lhs, TYPRNG$[ #, _], {}, {0, Infinity}, Heads -> True], 1])&, bnames],
				_ -> {1}
			]},
	With[ {bPatts = Table[ ToExpression[ "v$" <> ToString[ k] <> "$$$_"], {k, 1, Length[ bpos]}]},
		Replace[ makeAlphaEquivFun[ lhs, unifyTypes],
			{patt_, rest___} :>
				{ReplacePart[ patt, MapThread[ Last[ #1] -> #2 &, {bpos, bPatts}]], rest, MapThread[ ReplacePart[ #1, 2 -> Extract[ #2, {1}, Hold]]&, {bpos, bPatts}]}
		]
	]]]

(* 'makeAlphaEquivFun[ expr_]' returns a function 'f' that, when applied to an expression 'b', an association of sort
	constraints on a.b.f. constants, and a pair '{tvar, tsubst}' consisting of an association of sort constraints on
	type variables and a type substitution, checks whether 'expr' and 'b' are alpha-equivalent under 'tsubst' and, in
	case of success, returns the pair of new sort constraints and new type substitation (just as 'alphaEquiv' does).
	First experiments suggest that this variant is faster than 'alphaEquiv' if expressions do not grow too big
	(< 95000 atoms), but of course generating 'f' also takes some time and storing 'f' consumes some memory. Therefore,
	use this variant if alpha-equivalence tests w.r.t. 'expr' have to be performed many times. *)
makeAlphaEquivFun[ expr_] :=
	Replace[ makeAlphaEquivFun[ expr, True],
		{
			{patt_, Hold[], {}} :> Function[ {b, tfix, acc}, If[ MatchQ[ b, patt], acc, False]],
			{patt_, typPattVars_Hold, types_List} :>
				With[ {test = If[ Length[ types] <= 50, unifyTypeListsRec, unifyTypeListsIt]},
					Function[ {b, tfix, acc},
						If[ Last[ acc] === {} && expr === b,
							acc,
						(*else*)
							Replace[ b,
								{
									ReplacePart[ RuleDelayed[ patt, test[ typPattVars, types, tfix, acc]], {2, 1, 0} -> List],
									_ -> False
								}
							]
						]
					]
				]
		}
	]
makeAlphaEquivFun[ expr_, True] :=
	With[ {tpos0 = Position[ expr, _TYPED$], rpos0 = Position[ expr, _TYPRNG$]},
	With[ {tpos1 = DeleteCases[ tpos0, {___, 1}]},	(* there is no need to introduce pattern variables for these types, see theorem below *)
	With[ {tpos =
			MapThread[
				Function[ {p, e},
					Replace[ Last[ e], {a_List :> Sequence @@ MapIndexed[ {Join[ p, {2, First[ #2]}], #1}&, a], t_ :> {Append[ p, 2], t}}]
				],
				{tpos1, Extract[ expr, tpos1]}
			],
			rpos = MapThread[ {Append[ #1, 2], Last[ #2]}&, {rpos0, Extract[ expr, rpos0]}]},
		(* names of pattern variables *MUST* end with "$" (otherwise Mathematica adds "$" itself) *)
	With[ {typPattVars = Table[ ToExpression[ "t$" <> ToString[ k] <> "$$$", InputForm, Hold], {k, 1, Length[ tpos] + Length[ rpos]}]},
		{
			ReplacePart[ expr,
				Join[
					MapThread[ (First[ #1] -> (Pattern @@ Append[ #2, Blank[]])) &, {Join[ tpos, rpos], typPattVars}],
					Cases[ tpos0, {pre___, 1} :> ({pre, 1, 2} -> Blank[])],
					Map[ Append[ #, 1] -> Blank[] &, rpos0]
				]
			],
			Flatten[ Hold @@ typPattVars, 1, Hold],
			Join[ tpos, rpos][[All, 2]]
		}
	]]]]
makeAlphaEquivFun[ expr_, False] :=
	With[ {tpos0 = Position[ expr, _TYPED$], rpos0 = Position[ expr, _TYPRNG$]},
	With[ {tpos = Append[ #, 2]& /@ Join[ DeleteCases[ tpos0, {___, 1}], rpos0]},
	With[ {types = Extract[ expr, tpos]},
	With[ {tvars = DeleteDuplicates[ Cases[ types, _TVAR$, {0, Infinity}, Heads -> True]]},
	With[ {tvarPattRepl = MapIndexed[ (#1 -> ToExpression[ "t$" <> ToString[ First[ #2]] <> "$$$_"])&, tvars]},
		{
			ReplacePart[ expr,
				Join[
					MapThread[ (#1 -> (#2 /. tvarPattRepl))&, {tpos, types}],
					Cases[ tpos0, {pre___, 1} :> ({pre, 1, 2} -> Blank[])],
					Map[ Append[ #, 1] -> Blank[] &, rpos0]
				]
			] /. tvarPattRepl,
			Replace[ tvarPattRepl, (l_ -> r_Pattern) :> (l -> Extract[ r, {1}, Hold]), {1}]
		}
	]]]]]

(*
Theorem:
If 'term' is some well-typed and fully type-annotated term, 'X' is the set of all type variables appearing anywhere in
'term' (including types of a.b.f. constants and meta variables), and 'term1' is obtained from 'term' by replacing all
subexpressions 'f[a::t]' by 'f[a]' (i.e. removing certain type annotations from atoms), then every type variable in 'X'
still appears in 'term1'.
Proof.
When replacing a term 'f[a::t]' by 'f[a]', 't' is the type of 'a::t', and the type of 'f' is something like 't -> s'.
Clearly, all type variables appearing in the type of a term also appear somehwere in the type annotations within the
term (provided it is fully type-annotated), and so all type variables in 't -> s' appear in 'f'. This in particular
implies that all type variables in 't' appear in 'f' and hence in 'f[a]', which finishes the proof. QED
*)


(* ::Section:: *)
(*Equational Rewriting*)

(* ::Subsection:: *)
(*General*)

$recItThreshold = 50;

patternMatchQ[ patt_, term_, ctxt:{_, _, meta_, tfix_, tvar_}] :=
	With[ {vars = getPositions[ patt, _META$]},
	With[ {repl = MapIndexed[ First[ #1] -> VAR$[ First[ #2], False] &, vars]},
	With[ {rl = generateEqRewriteRule[ {ReplacePart[ patt, repl], {}, Replace[ vars[[All, 2]], meta, {1}][[All, 1]]}, {None}, tfix, tvar, False, 0, Length[ vars]]},
		applyEqRewriteRule[ term, Null, rl, {}, ctxt] =!= $Failed
	]]]

(* 'applyEqRewriteRule[ expr, fml, loose, ctxt]' applies the equational rewrite-rule 'fml' to the given expression
	'expr' making use of the rewrite-rule-data generated by 'generateEqRewriteRule' and stored in 'fml'. 'loose' and
	'ctxt' describe the current typing-context ('ctxt' is a 5-tuple '{idx, fix, meta, tfix, tvar}').
	The function returns either '$Failed' or a 6-tuple '{inst1, inst2, {newTVar, tsubst}, bren, idx, test}', where
	- 'inst1' is a list of terms representing the instantiation of the universal quantifiers in 'fml'; elements could
		also be 'Automatic'.
	- 'inst2' is a list of terms representing the terms that must be abstracted over in 'expr' in order to achieve
		alpha-equivalence; elements cannot be 'Automatic' but may involve loose bound variables.
	- 'tsubst' is a type substitution.
	- 'newTVar' is the new association of sort constraints on type variables.
	- 'bren' is a renaming-substitution of bound variables in the rewrite-rule.
	- 'idx' is the new type variable index, i.e. the original one plus the number of type variables in the rewrite-rule
		(if these variables shall be lifted).
	- 'test' is either 'True' or a binary predicate; in the latter case, the predicate is applied to the original
		expression and to the new expression to determine whether the rule shall really be applied or not.
	The terms in 'inst1' and 'inst2' may involve subexpressions of the form 'POS$[ pos, shift]', where 'pos' is a list
	of positions referring to alpha-equivalent subexpressions of 'expr' and 'shift' is an integer specifying a shift of
	the de Bruijn indices of *free* variables in these (alpha-equivalent) subexpressions. *)
applyEqRewriteRule[ expr_, fml:FML$[ _, f_, __], loose_, ctxt_] :=
	applyEqRewriteRule[ expr, f, getOptionalComponent[ fml, "rwData"], loose, ctxt]
applyEqRewriteRule[ expr_, _, data:(_Rule|_RuleDelayed), loose_, ctxt_] :=
	Quiet[ Check[ Replace[ {expr, loose, ctxt}, {data, _ -> $Failed}], $Failed]]
applyEqRewriteRule[ ___] := $Failed

(* 'generateEqRewriteRule[ fml, lift, tfix, tvar]' generates a Mathematica transformation rule that can be used to
	efficiently apply the equational rewrite-rule extracted from the (quantified, conditional) Core equality 'fml'
	(e.g. by 'applyEqRewriteRule'). The transformation rule maps triples '{term, loose, ctxt}' consisting of subject
	term 'term', list of types of loose bound variables 'loose', and typing-context 'ctxt', to a 7-tuple; see the
	description of 'applyEqRewriteRule' for the precise format and meaning of the output. *)
generateEqRewriteRule[ FML$[ _, f_, _, data___], tfix_] :=
	generateEqRewriteRule[ f, Replace[ "TypeVarPositions", {data, _ -> True}], tfix]
generateEqRewriteRule[ STERM$[ f_, _, tvar_Association], lift_, tfix_] :=
	generateEqRewriteRule[ f, lift, tfix, tvar]
generateEqRewriteRule[ f_, lift_, tfix_] :=
	generateEqRewriteRule[ f, lift, tfix, Association[]]
generateEqRewriteRule[ fml_, lift_, tfix_Association, tvar_Association] :=
	Replace[ stripAll[ fml, {}],
		{
			{fml1_, univ_List} :>
				Replace[ toAssmsGoal[ fml1, {}, Infinity],
					{
						{conds___, TYPED$[ CoreEqual$TM, {_}][ lhs_][ rhs_]} :>
							Replace[ stripAbstractions[ lhs, {}],
								{
									{a_, lambda_List} :>
										With[ {uc = Length[ univ], lc = Length[ lambda]},
										With[ {patt =
												{
													If[ univ === {} === lambda,
														a,
													(*else*)
														(* We temporarily change the representation of bound variables: those variables that are universally
															quantified are represented by 'VAR$[ i, True]', where integer 'i' refers to a type in 'univ',
															and those variables that are bound by a lambda-abstraction are represented by 'VAR$[ i, False]',
															where integer 'i' refers to a type in 'lambda'.
															Note that the indices are everywhere the same, regardless of the number of abstractions in whose
															scope they appear! Hence, they may not be considered de Bruijn indices any more. *)
														mapFreeVars[ a, With[ {diff = #3 - #2}, If[ diff < lc, VAR$[ lc - diff, False], VAR$[ uc + lc - diff, True]]]&]
													],
													Reverse[ univ],
													Reverse[ lambda]
												},
											lift1 =
												Switch[ lift,
													_List,
													Replace[ lift, ({p_List, ___} -> sort_) :> (extractSingle[ fml, p] -> sort), {1}],
													True,
													(* here one could probably also use 'Cases', but to be on the safe side we use the same function as everywhere *)
													With[ {defSort = getDefaultSort[]},
														Map[ Last[ #] -> Lookup[ tvar, Last[ #], defSort] &, getPositions[ fml, _TVAR$]]
													],
													_,
													False
												]},
											generateEqRewriteRule[
												patt,
												{rhs, conds},
												tfix,
												If[ ListQ[ lift], Association[ lift1], tvar],
												lift1,
												uc,
												lc
											]
										]],
									_ -> $Failed
								}
							],
						_ -> $Failed
					}
				],
			_ -> $Failed
		}
	]
generateEqRewriteRule[ patt:{lhs_, _, _}, rc:{rhs0_, conds___}, tfix_Association, tvar_Association, lift_, uc_Integer, lc_Integer] :=
	Which[
		rhs0 === None,
		generateEqRewriteRuleAux[ patt, rc, lift, True],
	(*else*)
		MatchQ[ First[ getHeadAtom[ lhs]], VAR$[ _Integer, _]],
		generateEqRewriteRuleAux[ patt, rc, lift, termGreaterQ],	(* allows rules such as '\x::UNIT. x === \x. U' *)
	(*else*)
		True,
		(* Detect "permutative" rules, i.e. rules where the lhs and rhs are equal up to renaming of variables, and where
			variables that would really have to be "renamed" do not appear in the conditions.
			Examples:
			- 'Plus === \x y. y + x'
			- '\x y. y + x === Plus'
			- '\x y. Subtract[ x - y] === \x y z. (x - z) - y'
			- '\A x y. insert[ y, insert[ x, A]] === \A x y. insert[ x, insert[ y, A]]'. *)
		Replace[ stripAbstractions[ rhs0, {}],
			{
				{rhs_, lambdaR_List} :>
					Replace[ getPermutation[ lhs, rhs, lc, Length[ lambdaR], tfix, tvar],
						{
							False :> generateEqRewriteRuleAux[ patt, rc, lift, True],
							repl_List :>
								With[ {max = Max[ lc, Length[ lambdaR]]},
								With[ {vars = Cases[ repl, (v:VAR$[ i_, u_] -> j_) /; (TrueQ[ u] && i + j =!= uc + max) || (u === False && i + j =!= max) :> v]},
									If[ vars === {},
										(* 'lhs' and 'rhs' are alpha-equivalent --> result does not change *)
										$Failed,
									(*else*)
										(* variables in 'lhs' and 'rhs' correspond to each other 1-1
											--> in general, result changes, but rule can perhaps be applied to result again (depending on conditions) *)
										With[ {uvars = Cases[ vars, VAR$[ i_, True] :> (uc + max - i)]},
											generateEqRewriteRuleAux[
												patt,
												rc,
												lift,
												If[ uvars === {} || !containsVar[ {conds}, MemberQ[ uvars, #3 - #2]&], termGreaterQ, UnsameQ]
											]
										]
									]
								]],
							(* every free variable in 'lhs' corresponds to free variable in 'rhs', but not 1-1
								--> in general, rule cannot be applied to result again, but in some cases result does not even change *)
							_ :> generateEqRewriteRuleAux[ patt, rc, lift, UnsameQ]
						}
					],
				_ -> $Failed
			}
		]
	]

generateEqRewriteRuleAux[ {lhs_, univ_, lambda_}, {rhs_, ___}, lift_, test_] :=
	Replace[ getStructuralVarInfo[ lhs, lambda, False],
		{lhs1_, lHO_, lFO_, lMap_} :>
			Replace[ getStructuralVarInfo[ lhs1, univ, True],
				{lhs2_, uHO_, uFO_, uMap_} :>
					Replace[ makeBoundTypePattern[ lhs2, rhs, True],
						{patt_, typPattVars_, types_, bRepl_} :>
							With[ {tvarRepl =
									If[ ListQ[ lift],
										MapIndexed[
											(First[ #1] -> TVAR$[ (First[ #2] - 1) + Hold[ idx$$$]])&,
											lift	(* list of type variables must be in right order *)
										],
									(*else*)
										{}
									]},
								generateEqRewriteRuleAux[
									{patt, uHO /. tvarRepl, uFO, uMap, lHO /. tvarRepl, lFO, lMap},
									With[ {tvar =
											If[ ListQ[ lift],
												Hold[ Append][ Hold[ tvar$$$], DeleteCases[ MapThread[ (Last[ #1] -> Last[ #2])&, {tvarRepl, lift}], _ -> getDefaultSort[]]],
											(*else*)
												Hold[ tvar$$$]
											]},
										If[ types === {},
											{tvar, {}},
										(*else*)
											With[ {fun = If[ Length[ types] <= $recItThreshold, unifyTypeListsRec, unifyTypeListsIt]},
												Hold[ fun][
													types /. tvarRepl,
													List @@ (Hold /@ typPattVars),
													Hold[ tfix$$$],
													{tvar, {}}
												]
											]
										]
									],
									Length[ tvarRepl],
									bRepl,
									test
								]
							]
					]
			]
	]
(* no variables *)
generateEqRewriteRuleAux[ {patt_, {}, {}, _, {}, {}, _}, typAcc_, len_, bren_, test_] :=
	With[ {rhs =
			If[ ListQ[ typAcc],
				HoldComplete[ {{}, {}, typAcc, bren, idx$$$ + len, test}],
			(*else*)
				HoldComplete[
					With[ {res = typAcc},
						{{}, {}, res, bren, idx$$$ + len, test} /; ListQ[ res]
					]
				]
			]},
		RuleDelayed @@ Prepend[ rhs /. Hold[ x_] :> x, {patt, _, {idx$$$_, _, _, tfix$$$_, tvar$$$_}}]
	]
(* no higher-order variables *)
generateEqRewriteRuleAux[ {patt_, {}, uFO_, uMap_, {}, lFO_, lMap_}, typAcc_, len_, bren_, test_] :=
	With[ {foae = generateFOAECode[ uFO, lFO, Hold[ tfix$$$], typAcc],
			uInst = generateInstanceCode[ {}, uFO, uMap, {}, 0],
			lInst = generateInstanceCode[ {}, lFO, lMap, {}, 0]},
	With[ {rhs =
			HoldComplete[
				With[ {res = Catch[ {uInst, lInst, foae, bren, idx$$$ + len, test}]},
					res /; ListQ[ res]
				]
			]},
		RuleDelayed @@ Prepend[ rhs /. Hold[ x_] :> x, {patt, _, {idx$$$_, _, _, tfix$$$_, tvar$$$_}}]
	]]
(* higher-order (and first-order) variables *)
generateEqRewriteRuleAux[ {patt_, uHO_, uFO_, uMap_, lHO_, lFO_, lMap_}, typAcc_, len_, bren_, test_] :=
	With[ {foae = generateFOAECode[ uFO, lFO, Hold[ tfix$$$], typAcc]},
	With[ {ap =
			Hold[ getArgPositions][
				Join[
					Map[ ReplacePart[ First[ #], -1 -> True]&, uHO],
					Map[ ReplacePart[ First[ #], -1 -> False]&, lHO]
				],
				foae
			],
			aPatts = Table[ ToExpression[ "a$" <> ToString[ k] <> "$$$_List"], {k, 1, Length[ uHO] + Length[ lHO]}]},
	With[ {hoae =
			With[ {hoaef = If[ Length[ uHO] + Length[ lHO] <= $recItThreshold, hoAlphaEquivRec, hoAlphaEquivIt]},
				Hold[ hoaef][
					MapThread[
						Function[ {varInfo, aPatt},
							Prepend[ Last[ varInfo], ReplacePart[ Most[ First[ varInfo]], 2 -> Extract[ aPatt, {1}, Hold]]]
						],
						{Join[ uHO, lHO], aPatts}
					],
					Hold[ tfix$$$],
					Hold[ typAcc0$$$]
				]
			]},
	With[ {uInst = generateInstanceCode[ uHO, uFO, uMap, aPatts, 0],
			lInst = generateInstanceCode[ lHO, lFO, lMap, aPatts, Length[ uHO]]},
	With[ {rhs =
			HoldComplete[
				With[ {res = Catch[ Replace[ ap, {aPatts, typAcc0$$$_List} :> With[ {tsubst$$$ = hoae}, {uInst, lInst, tsubst$$$, bren, idx$$$ + len, test}]]]},
					res /; ListQ[ res]
				]
			]},
		RuleDelayed @@ Prepend[ rhs /. Hold[ x_] :> x, {patt, _, {idx$$$_, _, _, tfix$$$_, tvar$$$_}}]
	]]]]]

(*
	g[ \x y z. h[ F[z, x], F[ y]], F, h[ X, X]]:
	
	g[ \rx_ _ rz_. h[ X$1$1$_, X$1$2$_], X$1$3$_, h[ y$1$1$_, y$1$2$_] :>
		With[ {res =
				Catch[
					Replace[ getArgPositions[ {{X$1$1$, {VAR$[0], VAR$[2]}, 2, 3, True}}, foAlphaEquiv[ {{False, {y$1$1$, 0}, {y$1$2$, 0}}}, unifyTypeLists[ ...]]],
						{{a$1$_List}, tsubst0$_List} :>
							With[ {tsubst$ = hoAlphaEquiv[ {{{X$1$1$, a$1$, 2, 3}, {X$1$2$, {VAR$[1]}, 1, 3}, {X$1$3$, {}, 0, 0}}}, tsubst0$]},
								{{toInstance[ X$1$3$, a$1$, 2, 3, Reverse[ {rz, rx}]] /. tsubst$}, {POS$[ ..., -0]}, tsubst$, test}
							]
					]
				]},
			res /; ListQ[ res]
		]
	
	
	g[ X, h[ Y, X]]:
	
	g[ y$1$1$_, h[ x$2$1$_, y$1$2$_]] :>
		With[ {res =
				Catch[
					{{POS$[ ..., 0]}, {POS$[ ..., -0]}, foAlphaEquiv[ {{False, {y$1$1$, 0}, {y$1$2$, 0}}, {True, {y$2$1$, 0}}}, unifyTypeLists[ ...], test}
				]},
			res /; ListQ[ res]
		]
	
	g[ a, b]:
	
	g[ a, b] :>
		With[ {res = unifyTypeLists[ ...]},
			{{}, {}, res, test} /; ListQ[ res]
		]
*)

generateFOAECode[ {}, {}, _, typAcc_] :=
	typAcc
generateFOAECode[ univ0_, lambda0_, tfix_, typAcc_] :=
	With[ {univ = DeleteCases[ univ0, {}], lambda = DeleteCases[ lambda0, {}]},
	With[ {foaef = If[ Length[ univ] + Length[ lambda] <= $recItThreshold, foAlphaEquivRec, foAlphaEquivIt]},
		Hold[ foaef][
			Join[
				Map[ Prepend[ Rest /@ #, True]&, univ],
				Map[ Prepend[ Rest /@ #, False]&, lambda]
			],
			tfix,
			typAcc
		]
	]]

generateInstanceCode[ ho_List, fo_List, map_, aPatts_, offset_] :=
	Replace[ map,
		{
			{1, i_} :>
				Hold[ replaceTypes][
					ReplacePart[ Hold[ toInstance] @@ ho[[i, 1]], 2 -> Extract[ aPatts[[i + offset]], {1}, Hold]],
					Hold[ Last[ tsubst$$$]]
				],
			{2, i_} :> Replace[ fo[[i]], {{} -> Automatic, aux_List :> POS$[ aux[[All, 1]], -aux[[1, -1]]]}]
		},
		{1}
	]

termGreaterQ[ x_, y_] := (x =!= y && OrderedQ[ {y, x}])

getPermutation[ a_, b_, lca_, lcb_, tfix_, tvar_] :=
	Replace[
		Which[
			lca === lcb,
			If[ a === b,
				{{}, tvar, {}},
			(*else*)
				getPermutation[ a, b, 0, tfix, {{}, {tvar, {}}}]
			],
		(*else*)
			lca < lcb,
			getPermutation[ Fold[ #1[ #2] &, a, Table[ VAR$[ k, False], {k, lca + 1, lcb}]], b, 0, tfix, {{}, {tvar, {}}}],
		(*else*)
			True,
			With[ {add = lca - lcb},
				getPermutation[ a, Fold[ #1[ #2] &, mapFreeVars[ b, VAR$[ #3 + add]&], Table[ VAR$[ k], {k, add - 1, 0, -1}]], 0, tfix, {{}, {tvar, {}}}]
			]
		],
		{
			{repl_List, {_Association, _List}} :> repl,
			{_, {_, False}}|{_, False} -> False,
			_ -> True
		}
	]
getPermutation[ a:VAR$[ _, _], b_, abs_, _, {repl_, {tvar_, tunif_List}}] :=
	Replace[ b,
		{
			VAR$[ j_] :>
				With[ {diff = j - abs},
				With[ {newRepl =
						If[ ListQ[ repl],
							Replace[ Replace[ a, repl],
								{
									diff -> repl,
									_Integer -> False,
									_ :> If[ MemberQ[ repl, _ -> diff], False, Prepend[ repl, a -> diff]]
								}
							],
						(*else*)
							repl
						]},
					If[ NonNegative[ diff],
						{newRepl, {tvar, tunif}},
					(*else*)
						{newRepl, {tvar, False}}
					]
				]],
			_ -> {repl, {tvar, False}}
		}
	]
getPermutation[ a:(_VAR$|_FIX$|_META$), b_, _, _, {repl_, {tvar_, tunif_List}}] :=
	If[ a === b,
		{repl, {tvar, tunif}},
	(*else*)
		{repl, {tvar, False}}
	]
getPermutation[ _, _VAR$|_FIX$|_META$, _, _, {repl_, {tvar_, _List}}] :=
	{repl, {tvar, False}}
getPermutation[ x:TYPED$[ a_, s_], y_, abs_, tfix_Association, {repl_, {tvar_Association, tunif_List}}] :=
	Replace[ y,
		{
			TYPED$[ b_, t_] :>
				getPermutation[
					a,
					b,
					abs,
					tfix,
					{repl, compareTypes[ HoldComplete[ x], HoldComplete[ y], hasSort[ tfix], {tvar, tunif}, 0]}
				],
			_ -> {repl, {tvar, False}}
		}
	]
getPermutation[ _, _TYPED$, _, _, {repl_, {tvar_, _List}}] :=
	{repl, {tvar, False}}
getPermutation[ Lambda$TM[ RNG$[ TYPRNG$[ _, s_]], a_], b0_, abs_, tfix_Association, {repl_, {tvar_Association, tunif_List}}] :=
	Replace[ b0,
		{
			Lambda$TM[ RNG$[ TYPRNG$[ _, t_]], b_] :>
				getPermutation[
					a,
					b,
					abs + 1,
					tfix,
					{repl, compareTypes[ s, t, hasSort[ tfix], {tvar, tunif}, 0]}
				],
			_ -> {repl, {tvar, False}}
		}
	]
getPermutation[ _, _Lambda$TM, _, _, {repl_, {tvar_, _List}}] :=
	{repl, {tvar, False}}
getPermutation[ f_[ a_], b0_, abs_, tfix_Association, {repl_, {tvar_, tunif_List}}] :=
	Replace[ b0,
		{
			g_[ b_] :> getPermutation[ a, b, abs, tfix, getPermutation[ f, g, abs, tfix, {repl, {tvar, tunif}}]],
			_ -> {repl, {tvar, False}}
		}
	]
getPermutation[ a_, b_, _, _, {repl_, {tvar_, tunif_List}}] :=
	If[ a === b,
		{repl, {tvar, tunif}},
	(*else*)
		{repl, {tvar, False}}
	]
getPermutation[ _, _, _, _, {repl_, {tvar_, _}}] :=
	{repl, {tvar, False}}
getPermutation[ _, _, _, _, {repl_, _}] :=
	{repl, False}

(* 'getStructuralVarInfo' returns '{patt, ho, fo, map}', where
	- 'patt' is 'lhs' with free variables replaced by pattern variables,
	- 'ho' is structural information for proper higher-order variables; it is a list of elements of the form
		'{{Hold[ pname], args, argnum, abs, argtyps}, {{Hold[ pname0], args0, argnum0, abs0}...}}'; the 'argnum' of the
		first 5-tuple '{Hold[ pname], ...}' is >= all other 'argnum's.
	- 'fo' is structural information for "first-order-like" variables; it is a list of elements of the form
		'{{pos, Hold[ pname], abs}..}'.
	- 'map' is a list of pairs '{cat, ind}' where 'cat' is either '1' (referring to higher-order variables) or '2'
		(referring to first-order variables), and 'ind' is an index in the respective list 'ho' or 'fo'. *)
getStructuralVarInfo[ lhs_, {}, _] :=
	{lhs, {}, {}, {}}
getStructuralVarInfo[ lhs_, types_, u_] :=
	Replace[
		Reap[
			Module[ {isHO, foNum = 0, hoNum = 0, max, k},
			With[ {hoName = If[ TrueQ[ u], "X", "Y"], foName = If[ TrueQ[ u], "x", "y"]},
				MapIndexed[
					Function[ {typ, i},
						isHO = False;
						With[ {pa =
								Map[
									Function[ p0,
									With[ {n = LengthWhile[ Reverse[ p0], # === 0 &]},
									With[ {p = Drop[ p0, -n]},
										If[ n =!= 0, isHO = True];
										{p, getArgs[ extractSingle[ lhs, p], n]}
									]]],
									Position[ lhs, VAR$[ First[ i], u]]
								]},
							If[ isHO && AllTrue[ pa, With[ {args = Last[ #]}, MatchQ[ args, {VAR$[ _Integer]...}] && DuplicateFreeQ[ args]]&],
								max = -1;
								With[ {j = ToString[ ++hoNum]},
								With[ {aux =
										MapIndexed[
											With[ {p = First[ #1], args = Last[ #1], jj = First[ #2]},
											With[ {len = Length[ args]},
												If[ len > max, max = len; k = jj];
												{
													p,
													ToExpression[ hoName <> "$" <> j <> "$" <> ToString[ jj] <> "$$$", InputForm, Hold],
													args,
													len,
													countAbstractionsFromPos[ p, lhs]
												}
											]]&,
											pa
										]},
									Sow[
										{
											Append[ aux[[k]], Reverse[ Most[ flattenTypes[ typ, {}, aux[[k, 4]]]]]],
											Delete[ aux, k]
										},
										True
									];
									{1, hoNum}
								]],
							(*else*)
								With[ {j = ToString[ ++foNum]},
									Sow[
										MapIndexed[
											{
												Join[ First[ #], Table[ 0, {Length[ Last[ #]]}]],
												ToExpression[ foName <> "$" <> j <> "$" <> ToString[ First[ #2]] <> "$$$", InputForm, Hold]
											}&,
											pa
										],
										False
									]
								];
								{2, foNum}
							]
						]
					],
					types
				]
			]],
			{True, False}
		],
		{map_, {ho0_List, fo0_List}} :>
			With[ {ho = Join @@ ho0, fo = Join @@ fo0},
				{
					ReplacePart[ lhs,
						Join[
							Join @@ Map[
								Function[ occs,
									Map[
										{First[ #1]} -> (Pattern @@ Append[ #[[2]], Blank[]]) &,
										Prepend @@ Reverse[ occs]
									]
								],
								ho
							],
							Join @@ Map[
								{First[ #]} -> (Pattern @@ Append[ Last[ #], Blank[]]) &,
								fo,
								{2}
							]
						]
					],
					Map[
						Function[ occs,
							{Rest[ First[ occs]], Rest /@ Last[ occs]}
						],
						ho
					],
					Map[
						With[ {pos = First[ #]}, {pos, Last[ #], countAbstractionsFromPos[ pos, lhs]}]&,
						fo,
						{2}
					],
					map
				}
			]
	]


(* ::Subsection:: *)
(*Runtime Functions*)

(* General assumptions: terms are in beta-eta normal form. *)

(*
	- Check whether no illegal bound variables appear in 'term': legal variables are those in 'args',
		those bound inside 'term', and those bound outside of the innermost 'abs' abstractions if 'univ' is 'False'.
		Throw '$Failed' if an illegal variable appears.
	- Collect the positions of the arguments in 'term' and return them.
	- Collect the positions of all loose bound variables that are not subterms of arguments, and return them, too.
		Note that this list must be empty if 'univ' is 'True'!
*)
getArgPositions[ args_List, tsubst_List] :=
	{Map[ getArgPositions, args], tsubst}
getArgPositions[ _List, tsubst_] :=
	tsubst
getArgPositions[ {term_, args_List, argnum_Integer, abs_Integer, univ_}] :=
	Replace[
		Last[
			Reap[
				With[ {fun = If[ TrueQ[ univ], Throw[ $Failed]&, Sow[ #, False]&]},
					scanVars[ term,
						Function[ {p, a, i},
							If[ i >= a,
								If[ i >= a + abs,
									fun[ p],
								(*else*)
									Sow[ {First[ FirstPosition[ args, VAR$[ i - a], Throw[ $Failed], {1}, Heads -> False]], p -> a}, True]
								]
							]
						]
					]
				],
				{True, False}
			]
		],
		{argpos0_List, loosepos0_List} :>
			With[ {argpos1 =
					Append[
						Replace[
							GatherBy[ Join @@ argpos0, First],
							all:{{j_, _}, ___} :>
								(j -> Replace[ GatherBy[ all[[All, 2]], Last], all2:{_ -> a_, ___} :> (all2[[All, 1]] -> a), {1}]),
							{1}
						],
						_ -> {}
					]},
			With[ {argpos = Table[ Replace[ j, argpos1], {j, 1, argnum}]},
				{
					(* 'argpos' is a list of elements of the form '{(pos -> abs)...}', where 'pos' is a list of positions and
						'abs' is the number of abstractions in whose scope the subterms at the respective positions occur;
						this number is the same for all positions in 'pos'. In short, all positions are grouped together by
						precisely this number. *)
					argpos,
					(* Compute the number of trailing arguments that could be "eta reduced", e.g. if the formal pattern is
						'F[x][y]' and the actual term is 'f[y]', and 'y' does not appear in 'f', then 'y' could be dropped,
						meaning that the formal pattern becomes 'F[x]' and the term becomes 'f'. *)
					Module[ {i = 0}, LengthWhile[ Reverse[ argpos], (MatchQ[ #, {{{0..., 1}} -> 0}] && Length[ #[[1, 1, 1]]] === ++i)&]],
					Join @@ loosepos0
					(* In addition to 'maxdrop' one could also return the number of outermost eta-redexes in 'a', if terms
						are not assumed to be in eta normal form already.
						Example: Match '\x. g[ \y z. f[y, x, z], f[x]]' against the pattern '\x. g[ \y. F[y, x], F[x]]'. *)
				}
			]]
	]

(* Abstract over the given positions 'argpos' in 'term', and update the indices of the loose bound variables. *)
toInstance[ term_, {argpos0_List, maxdrop_Integer, loosepos_List}, argnum0_, abs_, loose_List] :=
	With[ {argpos = Drop[ argpos0, -maxdrop], argnum = argnum0 - maxdrop},
		Fold[
			Lambda$TM[ RNG$[ TYPRNG$[ Null, #2]], #1]&,
			extractSingle[
				ReplacePart[
					With[ {offset = argnum - abs},
						If[ offset === 0,
							term,
						(*else*)
							MapAt[ VAR$[ First[ #] + offset]&, term, loosepos]
						]
					],
					Join @@ MapIndexed[
						Function[ {pos, j0},
							If[ pos === {},
								(* For some strange reason, 'MapAt[ f, {}, {All, 2}]' does not simplify, so we have to consider this case separately. *)
								{},
							(*else*)
								With[ {j = argnum - First[ j0]},
									MapAt[ VAR$[ # + j]&, pos, {All, 2}]
								]
							]
						],
						argpos
					]
				],
				Table[ 0, {maxdrop}]
			],
			Drop[ loose, maxdrop]
		]
	]

alphaEquivListRec[ {f_, r___}, tfix_, acc_List, fun_] :=
	alphaEquivListRec[ {r}, tfix, fun[ f, tfix, acc], fun]
alphaEquivListRec[ _, _, acc_List, _] :=
	acc
alphaEquivListRec[ _, _, acc_, _] :=
	Throw[ acc]

alphaEquivListIt[ l_List, tfix_, acc0_List, fun_] :=
	Module[ {acc = acc0},
		Scan[
			(
				acc = fun[ #, tfix, acc];
				If[ !ListQ[ acc], Throw[ acc]]
			)&,
			l
		];
		acc
	]
alphaEquivListIt[ _, _, acc_List, _] :=
	acc
alphaEquivListIt[ _, _, acc_, _] :=
	Throw[ acc]
	
alphaEquivList[ fun_][ {master_, term_, terms___}, tfix_, acc_List] :=
	alphaEquivList[ fun][ {master, terms}, tfix, fun[ master, term, tfix, acc]]
alphaEquivList[ _][ _, _, acc_List] :=
	acc
alphaEquivList[ _][ _, _, acc_] :=
	Throw[ acc]

hoAlphaEquivRec[ l_, tfix_, acc_] := alphaEquivListRec[ l, tfix, acc, alphaEquivList[ hoAlphaEquiv]]

hoAlphaEquivIt[ l_, tfix_, acc_] := alphaEquivListIt[ l, tfix, acc, alphaEquivList[ hoAlphaEquiv]]

(* 'hoAlphaEquivAux' is sufficiently general as to accommodate non-patterns, too. *)
(* Check whether 'a' and 'b' are alpha-equivalent, and if so, return the type substitution.
	Otherwise, throw '$Failed'. *)
hoAlphaEquiv[ {a_, {argpos_, maxdrop_, loosepos_}, anum_, aabs_}, {b_, args_, bnum_, babs_}, tfix_Association, acc_] :=
	(* 'bnum <= anum' *)
	With[ {babs2 = abstractionCount[ b, 0, anum - bnum]},
		If[ babs2 + bnum + maxdrop >= anum,
			With[ {a1 =
					ReplacePart[
						With[ {offset = babs + babs2 - aabs},
							If[ offset === 0,
								a,
							(*else*)
								MapAt[ VAR$[ First[ #] + offset]&, a, loosepos]
							]
						],
						Join[
							Join @@ MapThread[
								Function[ {pos, arg},
								With[ {op =
										Function @@ {
											reap[
												scanVars[
													arg,
													Function[ {x, y, z}, If[ z >= y, Sow[ {x} -> VAR$[ z + Placeholder]]]]
												]
											] /. Placeholder -> Slot[ 1]
										}},		(* 'op' could be pre-computed at "compile time" ... *)
									(* 'pos' is of the form '{(p -> abs)...}' *)
									ReplacePart[ pos, {i_, 2} :>
										With[ {offset = pos[[i, 2]] + babs2},
											If[ offset === 0, arg, ReplacePart[ arg, op[ offset]]]
										]
									]
								]],
								{Take[ argpos, bnum], args}
							],
							Join @@ MapIndexed[
								Function[ {pos, j0},
								With[ {j = First[ j0]},
									ReplacePart[ pos, {i_, 2} :>
										VAR$[ pos[[i, 2]] + babs2 - j]
									]
								]],
								Take[ argpos, {bnum + 1, bnum + babs2}]
							]
						]
					]},
			With[ {a2 = extractSingle[ a1, Table[ 0, {anum - (bnum + babs2)}]],
					b2 = extractSingle[ b, Table[ 2, {babs2}]]},
				alphaEquiv[
					HoldComplete[ a2],
					HoldComplete[ b2],
					tfix,
					acc
				]
			]],
		(*else*)
			Throw[ $Failed]
		]
	]

foAlphaEquivRec[ l_, tfix_, acc_] := alphaEquivListRec[ l, tfix, acc, foAlphaEquiv]

foAlphaEquivIt[ l_, tfix_, acc_] := alphaEquivListIt[ l, tfix, acc, foAlphaEquiv]

(*
	- Check whether no illegal bound variables appear in 'term', analogous to 'getArgPositions'.
	- Check whether the given 'term' and 'terms' are all alpha-equivalent to each other, and if so, return the type
		substitution. Otherwise, throw '$Failed'.
*)
foAlphaEquiv[ {univ_, {term_, abs_}, terms___}, tfix_Association, acc_List] :=
	With[ {loosepos =
			reap[
				With[ {fun = If[ TrueQ[ univ], Throw[ $Failed]&, Sow[ #]&]},
					scanVars[ term,
						Function[ {p, a, i},
							If[ i >= a,
								If[ i >= a + abs,
									fun[ p],
								(*else*)
									Throw[ $Failed]
								]
							]
						]
					]
				]
			]},
		alphaEquivList[ foAlphaEquivAux][ {{term, abs, loosepos}, terms}, tfix, acc]
	]

foAlphaEquivAux[ {a_, aabs_, loosepos_}, {b_, babs_}, tfix_Association, acc_List] :=
	alphaEquiv[
		With[ {offset = babs - aabs},
			If[ offset === 0,
				HoldComplete[ a],
			(*else*)
				HoldComplete @@ {MapAt[ VAR$[ First[ #] + offset]&, a, loosepos]}
			]
		],
		HoldComplete[ b],
		tfix,
		acc
	]


(* ::Section:: *)
(*Reduction Rewriting*)

(* 'generateReductionRule[ fml, univ]' returns a function 'f' that, when applied to a type variable index 'idx', a
	number of meta variables 'metaNum', and a symbol 'X', returns a list that can be passed as the first argument to 
	function 'applyReductionRule'. *)
generateReductionRule[ fml:FML$[ _, f_, _, info___], univ_] :=
	Replace[ "arData",
		{
			info,
			_ :> generateReductionRule[ f, univ, Replace[ "TypeVarPositions", {info, _ -> True}]]
		}
	]
generateReductionRule[ f_, univ_] :=
	generateReductionRule[ f, univ, False]
generateReductionRule[ STERM$[ fml_, _, tvar_Association], univ_, lift_] :=
	generateReductionRule[ fml, univ, lift, tvar]
generateReductionRule[ fml_, univ_, lift_] :=
	generateReductionRule[ fml, univ, lift, Association[]]
generateReductionRule[ fml_, univ_, lift_, tvar_Association] :=
	Replace[ toAssmsGoalStripAll[ fml],
		{{G_, n_Integer}, A0_List} :>
			If[ n === 0 || TrueQ[ univ],
				With[ {A1 = Join @@ A0,
						lift1 =
							Switch[ lift,
								_List,
								Replace[ lift, ({p_List, ___} -> sort_) :> (extractSingle[ fml, p] -> sort), {1}],
								True,
								(* here one could probably also use 'Cases', but to be on the safe side we use the same function as everywhere *)
								With[ {defSort = getDefaultSort[]},
									Map[ Last[ #] -> Lookup[ tvar, Last[ #], defSort] &, getPositions[ fml, _TVAR$]]
								],
								_,
								False
							]},
				With[ {tvarRepl =
						If[ ListQ[ lift1],
							MapIndexed[
								(First[ #1] -> TVAR$[ (First[ #2] - 1) + Hold[ 1]])&,
								lift1	(* list of type variables must be in right order *)
							],
						(*else*)
							{}
						]},
					Catch[
						Function @@ {
							Join[
								{
									STERM$[
										If[ n === 0,
											G,
										(*else*)
											mapFreeVars[ G,
												If[ MatchQ[ #1, {___, 0}],
													(* Meta variables must not appear at head positions.
														This check alone does not suffice, since meta variables could be heads in the premises.
														So, if they are not instantiated, we obtain unifiation problems with meta variables at
														head positions. *)
													Throw[ $Failed&],
												(*else*)
													META$[ Hold[ 3], n + #2 + Hold[ 2] - (#3 + 1)]
												]&
											]
										] /. tvarRepl,
										Association[],
										If[ ListQ[ lift1],
											Association @@ DeleteCases[ MapThread[ (Last[ #1] -> Last[ #2])&, {tvarRepl, lift1}], getDefaultSort[]],
										(*else*)
											Association[]
										]
									],
									A1[[All, 1]] /. tvarRepl,
									Join @@@ Last[
										Reap[
											MapIndexed[
												Function[
													{x, y},
													Replace[ x,
														{A_, m:Except[ 0, _Integer]} :>
															scanVars[ A, If[ #3 >= #2, Sow[ Join[ y, #1], m + #2 - (#3 + 1)]]&]
													];
												],
												A1
											],
											Range[ 0, n - 1]
										]
									],
									Hold[ 1],
									Hold[ 2],
									Hold[ 3],
									Length[ tvarRepl],
									n
								}
							]
						} /. Hold -> Slot
					]
				]],
			(*else*)
				$Failed&
			]
	]

toAssmsGoalStripAll[ G_] :=
	Reap[ toAssmsGoalStripAll[ G, 0]]
toAssmsGoalStripAll[ TYPED$[ CoreImplies$TM, {}][ A_][ R_], n_] :=
	(Sow[ {A, n}]; toAssmsGoalStripAll[ R, n])
toAssmsGoalStripAll[ TYPED$[ CoreAll$TM, {_}][ Lambda$TM[ _RNG$, A_]], n_] :=
	toAssmsGoalStripAll[ A, n + 1]
toAssmsGoalStripAll[ TYPED$[ CoreAll$TM, {_}][ A_], n_] :=
	toAssmsGoalStripAll[ mapFreeVars[ A, VAR$[ #3 + 1]&][ VAR$[ 0]], n + 1]
toAssmsGoalStripAll[ G_, n_] := {G, n}

applyReductionRule[ thm:THM$[ k_, args___], goalData_, goal_, origIdx_, tfix_, tvar_, spec_] :=
	With[ {data = getThmData[ thm]},
		applyReductionRule[
			Lookup[
				data,
				"arData",
				With[ {t = getTheorems[ True][ k][ args]}, generateReductionRule[ First[ t], True, Last[ t]]]
			] @@ goalData,
			If[ MemberQ[ Lookup[ data, "tags", {}], "autored?"], {False}, {}],
			goal,
			origIdx,
			tfix,
			tvar,
			spec
		]
	]
applyReductionRule[ fml_FML$, goalData_, goal_, origIdx_, tfix_, tvar_, spec_] :=
	applyReductionRule[
		generateReductionRule[ fml, True] @@ goalData,
		If[ MemberQ[ getOptionalComponent[ fml, "tags"], "autored?"], {False}, {}],
		goal,
		origIdx,
		tfix,
		tvar,
		spec
	]
applyReductionRule[ k:{_String, _String}, goalData_, goal_, origIdx_, tfix_, tvar_, spec_] :=
	applyReductionRule[ getFormulas[ True][ k], goalData, goal, origIdx, tfix, tvar, spec]
applyReductionRule[ {STERM$[ rule_, _, ruleTVar_], A_List, varPos_List, idx_, metaNum_, X_, typeVarCount_, termVarCount_}, {safe___}, goal_, origIdx_, tfix_, tvar_, spec_] :=
	Replace[ unifyTermsSyn[ goal, rule, 0, tfix, {{}, Join[ tvar, ruleTVar], {}}, X, origIdx],
		{
			{inst0_, tv_Association, tinst_} :>
				With[ {inst = inst0 /. tinst},
				With[ {inst1 = Table[ Replace[ META$[ X, i], inst], {i, metaNum, metaNum + termVarCount - 1}],
						oldTypeInstPos = Position[ tinst, TVAR$[ _?(# < idx &), ___] -> _, {1}, Heads -> False]},
				With[ {oldTypeInst = Extract[ tinst, oldTypeInstPos],
						oldTermInst = Cases[ inst, HoldPattern[ META$[ X, _?(# < metaNum &)] -> _]]},
					{
						idx + typeVarCount,
						metaNum + termVarCount,
						tv,
						{spec, inst1, idx},		(* all type variables not below 'idx' have to be lifted into the current context *)
						ReplacePart[ replaceTypes[ A, Delete[ tinst, oldTypeInstPos]], MapThread[ Rule, {varPos, inst1}]],
						If[ oldTypeInst === {} === oldTermInst,
							Identity,
						(*else*)
							replaceTypes[ #, oldTypeInst] /. oldTermInst &
						],
						safe
					}
				]]],
			_ -> $Failed
		}
	]
applyReductionRule[ ___] :=
	$Failed

(* Purely syntactic unification of two terms. Type- and term substitution are computed somewhat separately, which means
	that the type substitution is not automatically applied to the term substitution!
	We do not use 'unify' from "HOUnification.m", since we only want to instantiate meta variables whose name is 'X';
	this cannot be achieved easily with 'unify'.
	ATTENTION! Meta variables must not appear at head positions, since otherwise we would have to unify types (c.f.
	"HOunification.m"). Example: unify '(X::BOOL->BOOL)[Y]' and '(isPrime::NAT->BOOL)[0]'. *)
unifyTermsSyn[ v:META$[ X_, _], b_, abs_, tfix_Association, {inst_, tvar_, tinst_}, X_, constThreshold_] :=
	With[ {w = Replace[ v, inst]},
		If[ w === v,
			With[ {b0 = b /. inst},
				If[ v === b0,
					{inst, tvar, tinst},
				(*else*)
					If[ FreeQ[ b0, v] && (abs === 0 || !containsVar[ b0, #3 >= #2 &]),
						{Prepend[ inst /. v -> b0, v -> b0], tvar, tinst},
					(*else*)
						$Failed
					]
				]
			],
		(*else*)
			unifyTermsSyn[ w, b, abs, tfix, {inst, tvar, tinst}, X, constThreshold]
		]
	]
unifyTermsSyn[ a_, v:META$[ X_, _], abs_, tfix_Association, acc_List, X_, constThreshold_] :=
	unifyTermsSyn[ v, a, abs, tfix, acc, X, constThreshold]
unifyTermsSyn[ a:(_FIX$|_VAR$), b_, _, _, acc_List, _, _] :=
	If[ a === b,
		acc,
	(*else*)
		$Failed
	]
unifyTermsSyn[ _, _FIX$|_VAR$, _, _, _, _, _] :=
	$Failed
unifyTermsSyn[ Lambda$TM[ RNG$[ TYPRNG$[ _, typ1_]], a_], b0_, abs_, tfix_Association, {inst_, tvar_Association, tinst_List}, X_, constThreshold_] :=
	Replace[ b0,
		{
			Lambda$TM[ RNG$[ TYPRNG$[ _, typ2_]], b_] :>
				With[ {tunif = unifyTypes[ typ1, typ2, hasSort[ tfix], {tvar, tinst}, constThreshold]},
					If[ ListQ[ tunif],
						unifyTermsSyn[ a, b, abs + 1, tfix, Prepend[ tunif, inst], X, constThreshold],
					(*else*)
						$Failed
					]
				],
			_ -> $Failed
		}
	]
unifyTermsSyn[ a0:TYPED$[ a_, _], b0_, abs_, tfix_Association, {inst_, tvar_Association, tinst_List}, X_, constThreshold_] :=
	Replace[ b0,
		{
			TYPED$[ b_, _] :>
				With[ {tunif = compareTypes[ HoldComplete[ a0], HoldComplete[ b0], hasSort[ tfix], {tvar, tinst}, constThreshold]},
					If[ ListQ[ tunif],
						unifyTermsSyn[ a, b, abs, tfix, Prepend[ tunif, inst], X, constThreshold],
					(*else*)
						$Failed
					]
				],
			_ -> $Failed
		}
	]
unifyTermsSyn[ f_[ a_], b0_, abs_, tfix_Association, acc_List, X_, constThreshold_] /; !MatchQ[ f, META$[ X, _]] :=
	Replace[ b0,
		{
			(g:Except[ META$[ X, _]])[ b_] :> unifyTermsSyn[ a, b, abs, tfix, unifyTermsSyn[ f, g, abs, tfix, acc, X, constThreshold], X, constThreshold],
			_ -> $Failed
		}
	]
unifyTermsSyn[ a_, b_, _, _, acc_List, _, _] :=
	If[ a === b,
		acc,
	(*else*)
		$Failed
	]
unifyTermsSyn[ _, _, _, _, _, _, _] :=
	$Failed

End[]

EndPackage[]
