(* ::Package:: *)

(* Theorema
    Copyright (C) 2017 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Core`Tools`IO`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]
Needs[ "Theorema`Language`"]

Begin["`Private`"]

initIO[ initAbbrevs_] :=
	(
		$exportSymbols = Association[];		(* '$exportSymbols' maps theory-names to sequences of symbols inside 'Hold'. *)
		$loadedFiles = Association[];		(* '$loadedFiles' maps theory-names to lists of file names. *)
		Which[
			TrueQ[ initAbbrevs],
			$dirAbbrevs = loadDirectoryAbbrevs[ FileNameJoin[ {$UserBaseDirectory, "Applications", "Theorema", "Environment.txt"}]],
		(*else*)
			Head[ $dirAbbrevs] =!= Association,
			$dirAbbrevs = Association[]
		];
	)

(* ::Section:: *)
(*Import/Export*)

(* The sole purpose of 'loadFile' is to load Mathematica packages containing implementations of functions that are
	needed by, say, parsing- and printing rules.
	ATTENTION! Do not assign values to Core variables in such packages (neither by setting variables to new values,
	nor by adding new elements to lists)! Otherwise, these assignments may happen twice when loading theory images,
	possibly leading to lists containing duplicate elements. If for some reason values shall be set in an external
	file rather than directly in a notebook, use 'Get' to load the file (but beware that conversely files loaded by
	'Get' shall not contain implementations).
	Summary:
		assignments to Core variables         --> theory notebooks or files loaded by 'Get'
		implementations and other assignments --> files loaded by 'loadFile' *)
SetAttributes[ loadFile, HoldAll];
loadFile[ file_] :=
	Replace[ If[ MatchQ[ Hold[ file], Hold[ _FileNameJoin]], #1&, makeFileName][ file, "m"],
		{
			f_String :>
				If[ FileExistsQ[ f],
					With[ {srcLabel = System`EvalTheoryName[]},
					With[ {files = Lookup[ $loadedFiles, srcLabel, {}]},
						If[ !MemberQ[ files, f],
							AssociateTo[ $loadedFiles, srcLabel -> Append[ files, f]]
						]
					]];
					Get[ f],
				(*else*)
					coreNotification[ {MSG$[ coreMessageToList, 1085, {f}, "Fatal"]}, False];
					$Failed
				],
			_ -> $Failed
		}
	]

System`LoadFile = loadFile;

SetAttributes[ addExportSymbols, HoldAll];
addExportSymbols[ args___] :=
	Replace[ partitionExportList[ DeleteDuplicates[ Hold[ args]]],
		{expSym_Hold, noSym_Hold, protected_Hold} :>
			(
				If[ noSym =!= Hold[], coreNotification[ {MSG$[ coreMessageToList, 245, List @@ (HoldForm /@ noSym), "Warning"]}, False]];
				If[ protected =!= Hold[], coreNotification[ {MSG$[ coreMessageToList, 246, List @@ (HoldForm /@ protected), "Warning"]}, False]];
				If[ expSym === Hold[],
					$exportSymbols,
				(*else*)
					With[ {srcLabel = System`EvalTheoryName[]},
					With[ {syms = Lookup[ $exportSymbols, srcLabel, Hold[]]},
						AssociateTo[ $exportSymbols, srcLabel -> DeleteDuplicates[ Join[ syms, expSym]]]
					]]
				]
			)
	]

System`ExportCore = addExportSymbols;

(* 'exportImage' uses 'DumpSave'. According to the documentation:
	"Files written by 'DumpSave' can only be read on the same type of computer system on which they were written." *)
exportImage[] :=
	exportImage[ Automatic, True, True]
exportImage[ thy_] :=
	exportImage[ thy, True, True]
exportImage[ thy_, cumulative_] :=
	exportImage[ thy, cumulative, True]
exportImage[ thy_, cumulative_, True] :=
	Replace[ Reap[ exportImage[ thy, cumulative, False], "Message"],
		{ret_, msg_List} :> (coreNotification[ msg]; ret)
	]
exportImage[ thy0_, cumulative0_, False] :=
	With[ {thy =
			Switch[ thy0,
				Automatic,
				getTheory[],
				_String,
				If[ StringMatchQ[ thy0, "Source" ~~ __],
					thy0,
				(*else*)
					makeTheoryFileName[ thy0, True]
				],
				_Path,
				makeTheoryFileName[ thy0, True],
				_,
				$Failed
			]},
		If[ StringQ[ thy] && setTheory[ thy] =!= $Failed,
			With[ {baseFile =
					With[ {aux = sourceFile[ thy]},
						FileNameJoin[ {DirectoryName[ aux], FileBaseName[ aux]}]
					]},
				If[ exportTheoryImage[ baseFile <> ".mx", cumulative0] === $Failed,
					$Failed,
				(*else*)
					With[ {cumulative = cumulative0 && getImportedTheories[] =!= {} && MemberQ[ Rest[ getTheoryScope[]], First[ getImportedTheories[]]]},
					Module[ {exportThy},
						Scan[
							(exportThy[ #] = True)&,
							If[ TrueQ[ cumulative],
								DeleteCases[ getTheoryScope[], Alternatives @@ getImportedTheories[]],
							(*else*)
								getTheoryScope[]
							]
						];
						exportThy[ _] = False;
						
						$imgDep =
							If[ TrueQ[ cumulative],
								FileNameJoin[ {DirectoryName[ getImageFile[]], FileBaseName[ getImageFile[]]}] <> "-extra1.mx",
							(*else*)
								None
							];
						$imgCont1 =
							Join[
								(* flat associations ('thy -> val') *)
								KeySelect[ exportThy] /@ {
									$loadedFiles,
									$exportSymbols
								},
								(* other associations that can be combined by 'AssociateTo' on import *)
								{
									KeySelect[ $thmLabelData, Composition[ exportThy, Last]],
									Select[ $labelKeyMapping, Composition[ exportThy, Last]],
									Select[ $labelThmMapping, Composition[ exportThy, Last]],
									Select[ $theoryMapping, exportThy]
								},
								(* constant patterns (no negative scope conditions) *)
								DeleteCases[ #, Except[ HoldPattern[ h_[ p_, inTheoryScopeQ[ t_]]] /; h === Condition && exportThy[ t]]]& /@
								{
									$condDefaultBinderInputPrinting,
									$condDefaultBinderPrinting,
									$uncondDefaultBinderInputPrinting,
									$uncondDefaultBinderPrinting,
									$curriedHeads,
									$typeHeads,
									$proofCommandSeparators,
									$theoryCommandSeparators
								},
								(* associations mapping keys to COREVAR$-like structures *)
								DeleteCases[
									Map[
										thinCoreVar[ exportThy],
										{
											$proofMethodMapping,
											$formulaVariantMapping,
											$fullTypeDisplayMapping,
											$abbrevConstDecls
										},
										{2}
									],
									_[ ___, {}|Association[], _],
									{2}
								],
								(* RULEC$ *)
								thinRULEC$[ exportThy] /@
								{
									$autoredRules,
									$compRules,
									$congRules,
									$elimRules,
									$introRules,
									$simpRules,
									$symRules,
									$transRules
								},
								(* other *)
								{
									DeleteCases[
										Select[ #, Composition[ exportThy, Last, First, Last]]& /@
											KeySelect[ $globalFactLists, Composition[ inTheoryScopeQ, Last, First, Last]],
												(* We must consider all fact lists currently in scope, even if they originate
													from theories in 'getImportedTheories[]'. This is because new facts might
													have been added to such lists later on. *)
										{}
									],
									Cases[ $criticalTheoryCommands, HoldPattern[ h_[ _, c_] -> _] /; h === Condition && c],
									$theoremIndex
								},
								(* COREVAR$ *)
								thinCoreVar[ exportThy] /@ getCoreVar[]
							];
						With[ {syms = Hold @@ Join @@ Values[ KeySelect[ $exportSymbols, exportThy]]},
							Replace[ partitionExportList[ DeleteDuplicates[ syms]],	(* We check all this again, just to be sure. *)
								{expSym0_Hold, noSym_Hold, protected_Hold} :>
									With[ {expSym = Join[ Hold[ $imgDep, $imgCont1], expSym0]},
										If[ noSym =!= Hold[], coreNotification[ {MSG$[ coreMessageToList, 245, List @@ (HoldForm /@ noSym), "Warning"]}, False]];
										If[ protected =!= Hold[], coreNotification[ {MSG$[ coreMessageToList, 246, List @@ (HoldForm /@ protected), "Warning"]}, False]];
										DumpSave @@ Prepend[ List @@@ Hold[ expSym], baseFile <> "-extra1.mx"]
									]
							]
						];
						
						$imgCont2 =
							Join[
								(* flat lists of conditional rules *)
								Map[
									thinScopeCondition[ inTheoryScopeQ, False],
									{
										$proofCommandInterpretationRules1,
										$proofCommandInterpretationRules2,
										$theoryCommandInterpretationRules1,
										$theoryCommandInterpretationRules2,
										$abbrevTermFoldingRules,
										$abbrevTermUnfoldingRules,
										$abbrevTypeFoldingRules,
										$abbrevTypeUnfoldingRules,
										$binaryTermParsingRules,
										$binaryTypeParsingRules,
										$binaryTermInputPrintingRules,
										$binaryTermPrintingRules,
										$binaryTypePrintingRules
									},
									{2}
								],
								(* nested lists of conditional rules *)
								DeleteCases[
									Map[
										thinScopeCondition[ inTheoryScopeQ, True],
										{
											$curriedDBParsingRules,
											$curriedDBInputPrintingRules,
											$curriedDBPrintingRules,
											$curriedNominalInputPrintingRules,
											$curriedNominalPrintingRules,
											$postParsingRules,
											$typeParsingRules,
											$typePrintingRules,
											$uncurriedNominalParsingRules,
											$uncurriedNominalInputPrintingRules,
											$uncurriedNominalPrintingRules
										},
										{2}
									],
									{}|{Replace|All|Repeated},
									{2}
								],
								(* other *)
								{
									DeleteCases[ thinScopeCondition[ inTheoryScopeQ, False] /@ $abbrevTypeDecls, {}]
								}
							];
						DumpSave[ baseFile <> "-extra2.mx", $imgCont2];
						True
					]]
				]
			],
		(*else*)
			$Failed
		]
	]

importImage[ file_] :=
	importImage[ file, True]
importImage[ file_, True] :=
	Replace[ Reap[ importImage[ file, False], "Message"],
		{ret_, msg_List} :> (coreNotification[ msg]; ret)
	]
importImage[ file_String, False] :=
	With[ {baseFile0 =
			With[ {aux =
					Switch[ file,
						_String,
						If[ StringMatchQ[ file, "Source" ~~ __],
							sourceFile[ file],
						(*else*)
							(* We do not call 'makeTheoryFileName' here, because typically 'file' is already a proper theory name. *)
							file
						],
						_Path,
						makeTheoryFileName[ file],
						_,
						$Failed
					]},
				If[ StringQ[ aux],
					FileNameJoin[ {DirectoryName[ aux], FileBaseName[ aux]}],
				(*else*)
					$Failed
				]
			]},
	With[ {baseFile1 = If[ baseFile0 === $Failed, $Failed, importTheoryImage[ baseFile0 <> ".mx"]]},
		If[ StringQ[ baseFile1],
			With[ {baseFile = FileNameJoin[ {DirectoryName[ baseFile1], FileBaseName[ baseFile1]}],
					cvs = getCoreVar[]},
				$imgDep = baseFile <> "-extra1.mx";
				Replace[
					Reap[
						While[ StringQ[ $imgDep],
							Check[
								Sow[ $imgDep, "File"]; Get[ $imgDep]; Sow[ $imgCont1, "Contents"],
								Return[ False]
							]
						],
						{"File", "Contents"}
					],
					{
						{Null, {{files_List}, {conts_List}}} :>
							(
								$imgDep = None;
								Catch[
									With[ {cvNum = Length[ cvs]},
										Scan[
											Function[ fc,
												With[ {cont = Last[ fc]},
													If[ Length[ cont] === cvNum + 29,
														Scan[ Get, Join @@ Values[ cont[[1]]]];
														(* flat associations ('thy -> val') *)
														Scan[
															(AssociateTo @@ Append[ First[ #], Last[ #]])&,
															Transpose[
																{
																	{
																		Hold @ $loadedFiles,
																		Hold @ $exportSymbols,
																		Hold @ $thmLabelData,
																		Hold @ $labelKeyMapping,
																		Hold @ $labelThmMapping,
																		Hold @ $theoryMapping
																	},
																	Take[ cont, 6]
																}
															]
														];
														(* constant patterns (no negative scope conditions) *)
														Scan[
															(Set @@ Append[ First[ #], Join[ #[[1, 1]], Last[ #]]])&,
															Transpose[
																{
																	{
																		Hold @ $condDefaultBinderInputPrinting,
																		Hold @ $condDefaultBinderPrinting,
																		Hold @ $uncondDefaultBinderInputPrinting,
																		Hold @ $uncondDefaultBinderPrinting,
																		Hold @ $curriedHeads,
																		Hold @ $typeHeads,
																		Hold @ $proofCommandSeparators,
																		Hold @ $theoryCommandSeparators
																	},
																	Take[ cont, {7, 14}]
																}
															]
														];
														(* associations mapping keys to COREVAR$-like structures *)
														Scan[
															(Set @@ Append[ First[ #], Merge[ {#[[1, 1]], Last[ #]}, combineCoreVars]])&,
															Transpose[
																{
																	{
																		Hold @ $proofMethodMapping,
																		Hold @ $formulaVariantMapping,
																		Hold @ $fullTypeDisplayMapping,
																		Hold @ $abbrevConstDecls
																	},
																	Take[ cont, {15, 18}]
																}
															]
														];
														(* RULEC$ *)
														Scan[
															combineRULEC$,
															Transpose[
																{
																	{
																		Hold @ $autoredRules,
																		Hold @ $compRules,
																		Hold @ $congRules,
																		Hold @ $elimRules,
																		Hold @ $introRules,
																		Hold @ $simpRules,
																		Hold @ $symRules,
																		Hold @ $transRules
																	},
																	Take[ cont, {19, 26}]
																}
															]
														];
														(* other *)
														$globalFactLists = Merge[ {$globalFactLists, cont[[27]]}, Join @@ # &];
														$criticalTheoryCommands = Join[ $criticalTheoryCommands cont[[28]]];
														$theoremIndex = cont[[29]];
														(* COREVAR$ *)
														Scan[ combineCoreVars, Transpose[ {cvs, Take[ cont, -cvNum]}]];
														$imgDep = First[ fc],
													(*else*)
														sowMessage[ 1086, {First[ fc]}, "Fatal"];
														$imgDep = $Failed;
														Throw[ Null]
													]
												];
											],
											Reverse[ Transpose[ {files, conts}]]
										]
									]
								];
								If[ StringQ[ $imgDep],
									Check[
										$imgDep = StringReplace[ $imgDep, pre__ ~~ "-extra1.mx" ~~ EndOfString :> pre <> "-extra2.mx"];
										Get[ $imgDep];
										If[ Length[ $imgCont2] === 25,
											(* flat lists of conditional rules, appended to defaults *)
											Scan[
												(Set @@ Append[ First[ #], Join[ #[[1, 1]], Last[ #]]])&,
												Transpose[
													{
														{
															Hold @ $proofCommandInterpretationRules1,
															Hold @ $proofCommandInterpretationRules2,
															Hold @ $theoryCommandInterpretationRules1,
															Hold @ $theoryCommandInterpretationRules2
														},
														Take[ $imgCont2, 4]
													}
												]
											];
											(* flat lists of conditional rules, prepended to defaults *)
											Scan[
												(Set @@ Append[ First[ #], Join[ Last[ #], #[[1, 1]]]])&,
												Transpose[
													{
														{
															Hold @ $abbrevTermFoldingRules,
															Hold @ $abbrevTermUnfoldingRules,
															Hold @ $abbrevTypeFoldingRules,
															Hold @ $abbrevTypeUnfoldingRules,
															Hold @ $binaryTermParsingRules,
															Hold @ $binaryTypeParsingRules,
															Hold @ $binaryTermInputPrintingRules,
															Hold @ $binaryTermPrintingRules,
															Hold @ $binaryTypePrintingRules
														},
														Take[ $imgCont2, {5, 13}]
													}
												]
											];
											(* nested lists of conditional rules *)
											{
												$curriedDBParsingRules,
												$curriedDBInputPrintingRules,
												$curriedDBPrintingRules,
												$curriedNominalInputPrintingRules,
												$curriedNominalPrintingRules,
												$postParsingRules,
												$typeParsingRules,
												$typePrintingRules,
												$uncurriedNominalParsingRules,
												$uncurriedNominalInputPrintingRules,
												$uncurriedNominalPrintingRules
											} = Take[ $imgCont2, {14, 24}];
											(* other *)
											$abbrevTypeDecls = Merge[ {$imgCont2[[25]], $abbrevTypeDecls}, Join @@ # &];
											$imgDep,
										(*else*)
											sowMessage[ 1086, {$imgDep}, "Fatal"];
											$Failed
										],
										$Failed
									],
								(*else*)
									$imgDep
								]
							),
						_ -> $Failed
					}
				]
			],
		(*else*)
			baseFile1
		]
	]]


(* ::Section:: *)
(*Further Public Functions*)

addTheoryDirs[ dirs_] :=
	addTheoryDirs[ dirs, System`EvalTheoryName[], System`EvalNotebook[]]
addTheoryDirs[ dirs_, thy_] :=
	addTheoryDirs[ dirs, thy, System`EvalNotebook[]]
addTheoryDirs[ dirs_List, thy_String, nb_NotebookObject] :=
	setCoreVar[ $theoryDirs, DeleteDuplicates[ Join[ makeFileName[ dirs, "", {nb}, $dirAbbrevs, Identity], getCoreVar[ $theoryDirs]]], thy]
addTheoryDirs[ dir_String, thy_, nb_] :=
	addTheoryDirs[ {dir}, thy, nb]

System`AddTheoryDirectories = addTheoryDirs;

makeTheoryFileName[ orig_] :=
	makeTheoryFileName[ orig, False, System`EvalNotebook[]]
makeTheoryFileName[ orig_, addPrefix_] :=
	makeTheoryFileName[ orig, addPrefix, System`EvalNotebook[]]
makeTheoryFileName[ orig_, addPrefix_, nb_NotebookObject] :=
	With[ {res = makeFileName[ orig, "nb", Join[ getCoreVar[ $theoryDirs], {Most[ $defaultTheoryImage], nb}]]},
		If[ ListQ[ res],
			If[ TrueQ[ addPrefix],
				Replace[ resolveSymbolicLinks /@ res, s_String :> sourceLabel[ s], {1}],
			(*else*)
				resolveSymbolicLinks /@ res
			],
		(*else*)
			With[ {res1 = resolveSymbolicLinks[ res]},
				If[ TrueQ[ addPrefix] && StringQ[ res1], sourceLabel[ res1], res1]
			]
		]
	]

(* By default, 'orig' is assumed to be a file name in Unix format, but this can be changed (e.g. by passing an explicit
	'Path[ file, os]' structure). *)
makeFileName[ orig_] :=
	makeFileName[ orig, ""]
makeFileName[ orig_, ext_] :=
	makeFileName[ orig, ext, {System`EvalNotebook[]}]
makeFileName[ orig_, ext_, path:{___, _NotebookObject|_String, ___}, rest___] :=
	makeFileName[
		orig,
		ext,
		Replace[ path,
			{
				nb_NotebookObject :>
					Replace[ Quiet[ NotebookFileName[ nb], NotebookFileName::nosv],
						{
							f_String :> Most[ FileNameSplit[ f]],
							_ -> Sequence[]
						}
					],
				f_String :> FileNameSplit[ f]
			},
			{1}
		],
		rest
	]
makeFileName[ orig_, ext_, path_List] :=
	makeFileName[ orig, ext, path, $dirAbbrevs]
makeFileName[ orig_, ext_, path_List, abbrevs_] :=
	makeFileName[ orig, ext, path, abbrevs, FileNameJoin]
makeFileName[ orig_, ext_String, path_List, abbrevs_, post_] :=
	makeFileName[ orig, ext, path, abbrevs, post, "Unix"]
makeFileName[ Path[ orig_String, os_String], ext_String, path_List, abbrevs_, post_, _String] :=
	makeFileName[ orig, ext, path, abbrevs, post, os]
makeFileName[ orig_String, ext_String, path_List, abbrevs_, post_, os_String] :=
	With[ {fun = Function[ {x, f}, Replace[ x, {a___, b_, "..", c___} :> f[ {a, c}, f]]]},
			(* That's an elegant way to define a recursive pure function. *)
	With[ {addExt =
			If[ ext === "",
				fun[ #, fun]&,
			(*else*)
				Function[ l0,
					With[ {l1 = fun[ l0, fun]},
						If[ l1 === {},
							l1,
						(*else*)
							With[ {last = Last[ l1]},
								Which[
									FileExtension[ last] =!= "",
									l1,
								(*else*)
									StringMatchQ[ last, __ ~~ "."],
									ReplacePart[ l1, -1 -> last <> ext],
								(*else*)
									True,
									ReplacePart[ l1, -1 -> last <> "." <> ext]
								]
							]
						]
					]
				]
			]},
		Replace[ FileNameSplit[ orig, OperatingSystem -> os],
			{
				{""|(_String?(StringMatchQ[ #, __ ~~ ":"]&) /; os === "Windows"), rest___} :> post[ addExt[ {$RootDirectory, rest}]],
				{".", rest___} :> post[ addExt[ Join[ If[ path === {}, {}, Last[ path]], {rest}]]],
				{s:("~"|_String?(StringMatchQ[ #, "$" ~~ __]&)), rest___} :>
					With[ {aux = If[ ListQ[ abbrevs], Replace[ s, abbrevs], Lookup[ abbrevs, s, Null]]},
						post[ addExt[ Join[ aux, {rest}]]] /; ListQ[ aux]
					],
				s_List :>
					If[ path === {},
						post[ addExt[ s]],
					(*else*)
						Catch[
							Scan[
								With[ {aux = addExt[ Join[ #, s]]},
									If[ FileExistsQ[ FileNameJoin[ aux]], Throw[ post[ aux]]]
								]&,
								Most[ path]
							];
							post[ addExt[ Join[ Last[ path], s]]]
						]
					]
			}
		]
	]]
makeFileName[ orig_List, ext_String, path_List, abbrevs_, post_, os_String] :=
	makeFileName[ #, ext, path, abbrevs, post, os]& /@ orig

(* Here, 'orig' is supposed to be a file name in the format of the current OS (unless specified otherwise). *)
shortenFileName[ orig_String] :=
	shortenFileName[ orig, $OperatingSystem]
shortenFileName[ orig_String, os_String] :=
	shortenFileName[ orig, os, $dirAbbrevs]
shortenFileName[ Path[ orig_String, os_String], _String, abbrevs0_] :=
	shortenFileName[ orig, os, abbrevs0]
shortenFileName[ orig_String, os_String, abbrevs0_] :=
	With[ {abbrevs = Reverse[ SortBy[ Reverse /@ DeleteCases[ If[ ListQ[ abbrevs0], Identity, Normal][ abbrevs0], _ -> {}], Composition[ Length, First]]],
			l = FileNameSplit[ orig, OperatingSystem -> os]},
	With[ {len = Length[ l]},
		FirstCase[
			abbrevs,
			(y_List -> x_String) :>
				With[ {ly = Length[ y]},
					FileNameJoin[ Prepend[ Drop[ l, ly], x], OperatingSystem -> os] /; ly <= len && Take[ l, ly] === y
				],
			orig
		]
	]]


(* ::Section:: *)
(*Auxiliary Functions*)

resolveSymbolicLinks[ file_List] :=
	With[ {res = FileInformation[ FileNameJoin[ file], "AbsoluteFileName"]},
		If[ res === {}, $Failed, FileNameSplit[ res]]
	]
resolveSymbolicLinks[ file_String] :=
	With[ {res = FileInformation[ file, "AbsoluteFileName"]},
		If[ res === {}, $Failed, res]
	]
	(* 'FileInformation' is an undocumented built-in function that can be used to retrieve information about files and
	directories. In particular, 'FileInformation[ file, "AbsoluteFileName"]' resolves all symbolic links in 'file'
	and returns its absolute file name, if it exists, or '{}' otherwise. *)

loadDirectoryAbbrevs[ file_String] :=
	With[ {tmaDirList = FileNameSplit[ Theorema`$TheoremaDirectory]},
	With[ {initAbbrevs =
				Join[
					{
						"~" -> FileNameSplit[ $HomeDirectory],
						"$Components" -> Join[ tmaDirList, {"Theorema", "Components"}],
						"$Core" -> Join[ tmaDirList, {"Theorema", "Core"}],
						"$Theorema" -> Append[ tmaDirList, "Theorema"],
						"$Theories" -> Join[ tmaDirList, {"Theorema", "CoreTheories"}],
						"$Provers" -> Join[ tmaDirList, {"Theorema", "CoreProvers"}]
					},
					Map[
						With[ {fb = FileBaseName[ #]}, "$" <> fb -> Join[ tmaDirList, {"Theorema", "CoreTheories", fb}]]&,
						Select[ FileNames[ All, FileNameJoin[ {Theorema`$TheoremaDirectory, "Theorema", "CoreTheories"}]], DirectoryQ]
					]
				]},
		If[ FileExistsQ[ file],
			Module[ {abbrevs = Association[ initAbbrevs]},
				Scan[
					Replace[ #,
						{
							Hold[ lhs_String = rhs_] :>
								If[ StringMatchQ[ lhs, "$" ~~ __],
									Replace[ makeFileName[ rhs, "", {tmaDirList}, abbrevs, Identity],
										l_List :> AssociateTo[ abbrevs, lhs -> l]
									]
								],
							Hold[ lhs_ = rhs_] :> (lhs = rhs)
						}
					]&,
					Check[ ReadList[ file, Hold[ Expression]], {}]
				];
				abbrevs
			],
		(*else*)
			Association[ initAbbrevs]
		]
	]]

partitionExportList[ args_Hold] :=
	With[ {noSym = Position[ args, Except[ _Symbol], {1}, Heads -> False]},
	With[ {args1 = Delete[ args, noSym]},
	With[ {protected = Position[ args1, s_Symbol /; MemberQ[ Attributes[ s], ReadProtected], {1}, Heads -> False]},
		{Delete[ args1, protected], Part[ args, noSym[[All, 1]]], Part[ args1, protected[[All, 1]]]}
	]]]

thinCoreVar[ exportThy_][ cv_] :=
	thinCoreVar[ cv, exportThy]
thinCoreVar[ Hold[ cv_], exportThy_] :=
	thinCoreVar[ cv, exportThy]
thinCoreVar[ COREVAR$[ v___, assoc_Association, def_], exportThy_] :=
	COREVAR$[ v, KeySelect[ assoc, exportThy], def]
thinCoreVar[ {assoc_Association, def_}, exportThy_] :=
	{KeySelect[ assoc, exportThy], def}
thinCoreVar[ {l_List, def_}, exportThy_] :=
	{Select[ l, Composition[ exportThy, First]], def}
thinCoreVar[ x_, _] := x

combineCoreVars[ {cv_}] := cv
combineCoreVars[ {cv1_, cv2_}] :=
	combineCoreVars[ cv1, cv2]
combineCoreVars[ Hold[ cv1_Symbol], cv2_] :=
	(cv1 = combineCoreVars[ cv1, cv2])
combineCoreVars[ COREVAR$[ ___, assoc1_Association, _], COREVAR$[ val___, assoc2_Association, def_]] :=
	COREVAR$[ val, Join[ assoc1, assoc2], def]
combineCoreVars[ {assoc1_Association, _}, {assoc2_Association, def_}] :=
	{Join[ assoc1, assoc2], def}
combineCoreVars[ {l1_List, _}, {l2_List, def_}] :=
	{DeleteDuplicatesBy[ Join[ l2, l1], First], def}

thinRULEC$[ exportThy_][ rc_] :=
	thinRULEC$[ rc, exportThy]
thinRULEC$[ RULEC$[ c_Association, v_Association, depth_], exportThy_] :=
	RULEC$[ KeySelect[ c, exportThy], KeySelect[ v, exportThy], depth]
thinRULEC$[ RULEC$[ c_Association, _, v_Association, _, _, _, depth_Integer], exportThy_] :=
	thinRULEC$[ RULEC$[ c, v, depth], exportThy]

combineRULEC$[ {rc1_, rc2_}] :=
	combineRULEC$[ rc1, rc2]
combineRULEC$[ Hold[ rc1_Symbol], rc2_] :=
	(rc1 = combineRULEC$[ rc1, rc2])
combineRULEC$[ RULEC$[ c1_Association, v1_Association, depth_], RULEC$[ c2_Association, v2_Association, depth_]] :=
	RULEC$[ Join[ c1, c2], Join[ v1, v2], depth]
combineRULEC$[ RULEC$[ c_Association, _, v_Association, _, _, _, depth_Integer], rc2_] :=
	combineRULEC$[ RULEC$[ c, v, depth], rc2]

thinScopeCondition[ exportThy_, keepUncond_][ expr_] :=
	thinScopeCondition[ expr, exportThy, keepUncond, True]
thinScopeCondition[ (h:(Rule|RuleDelayed))[ l_, r_], exportThy_, keepUncond_, True] :=
	With[ {aux = {thinScopeCondition[ l, exportThy, keepUncond, False]}},	(* if 'l' is a rule itself, we must not dig further into the LHS of 'l' *)
		If[ aux === {},
			Sequence @@ {},
		(*else*)
			h[ First[ aux], r]
		]
	]
thinScopeCondition[ expr:(Replace|All|Repeated), _, _, True] := expr
thinScopeCondition[ l_List, exportThy_, keepUncond_, True] :=
	thinScopeCondition[ exportThy, keepUncond] /@ l
thinScopeCondition[ expr_, exportThy_, keepUncond_, _] :=
	Replace[ expr,
		{
			h2_[ patt_, cond_] /; h2 === Condition :>
				With[ {pos = FirstPosition[ Hold[ Null, cond], HoldPattern[ inTheoryScopeQ[ _, ___List]], $Failed]},
					Which[
						ListQ[ pos],
						Replace[ Extract[ Hold[ Null, cond], pos, Hold],
							{
								Hold[ inTheoryScopeQ[ thy_]] :>
									If[ TrueQ[ exportThy[ thy]],
										expr,
									(*else*)
										Sequence @@ {}
									],
								Hold[ inTheoryScopeQ[ thy_, excpt_]] :>
									If[ TrueQ[ exportThy[ thy]],
										Condition @@ ReplacePart[
											Hold[ patt, cond],
											With[ {newExcpt = Select[ excpt, exportThy]},
												If[ newExcpt === {},
													pos :> inTheoryScopeQ[ thy],
												(*else*)
													pos :> inTheoryScopeQ[ thy, newExcpt]
												]
											]
										],
									(*else*)
										Sequence @@ {}
									]
							}
						],
					(*else*)
						TrueQ[ keepUncond],
						expr,
					(*else*)
						True,
						Sequence @@ {}
					]
				],
			_ :> If[ TrueQ[ keepUncond], expr, Sequence @@ {}]
		}
	]


End[]

EndPackage[]
