(* ::Package:: *)

(* Theorema
    Copyright (C) 2017 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Core`Tools`Reasoning`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]
Needs[ "Theorema`Language`"]
Needs[ "Theorema`Core`HOUnification`"]

Begin["`Private`"]


(* ::Section:: *)
(*Initialization*)

If[ !ValueQ[ $proofMode], $proofMode = 2];

If[ !ValueQ[ $preProof],
	$preProof =
		Function[ fml,
			restrictRULEC$[ $autoredRules];
			restrictRULEC$[ $elimRules];
			restrictRULEC$[ $introRules];
			restrictRULEC$[ $symRules];
			restrictRULEC$[ $transRules];
			restrictRULEC$[ $simpRules];
			restrictRULEC$[ $congRules];
			restrictRULEC$[ $compRules];
			Theorema`CoreProvers`CoreProofMethods`$rewritingSimpRules = Null;
			Theorema`CoreProvers`CoreProofMethods`$rewritingCongRules = Null;
			Theorema`CoreProvers`CoreProofMethods`$rewritingCompRules = Null;
			Theorema`CoreProvers`CoreProofMethods`$autoReduceRules = Null;
			$currentScope = {key@fml, getOptionalComponent[ fml, "pos"]};
			$proofGoal = fml	(* to achieve composability, we return 'fml' *)
		]
];


(* ::Section:: *)
(*Lambda Calculus*)

(* 'alphaEquivOffset' is similar to 'alphaEquiv'. The main differences are that free variables do not need to have
	identical indices in the two expressions, but their indices must differ by a constant offset, and that types are
	completely ignored. *)
alphaEquivOffset[ a:(_FIX$|_META$), b_, _, _] :=
	a === b
alphaEquivOffset[ VAR$[ i_Integer], b_, abs_, offset_] :=
	Replace[ b,
		{
			VAR$[ j_] :> If[ i < abs, i === j, i - j === offset],
			_ -> False
		}
	]
alphaEquivOffset[ _, _VAR$|_FIX$|_META$, _, _] :=
	False
alphaEquivOffset[ x:TYPED$[ a_, s_], y_, abs_, offset_] :=
	Replace[ y,
		{
			TYPED$[ b_, s] :> alphaEquivOffset[ a, b, abs, offset],
			_ -> False
		}
	]
alphaEquivOffset[ _, _TYPED$, _, _] :=
	False
alphaEquivOffset[ Lambda$TM[ RNG$[ TYPRNG$[ _, s_]], a_], b0_, abs_, offset_] :=
	Replace[ b0,
		{
			Lambda$TM[ RNG$[ TYPRNG$[ _, s]], b_] :> alphaEquivOffset[ a, b, abs + 1, offset],
			_ -> False
		}
	]
alphaEquivOffset[ _, _Lambda$TM, _, _] :=
	False
alphaEquivOffset[ f_[ a_], b0_, abs_, offset_] :=
	Replace[ b0,
		{
			g_[ b_] :> alphaEquivOffset[ f, g, abs, offset] && alphaEquivOffset[ a, b, abs, offset],
			_ -> False
		}
	]
alphaEquivOffset[ a_, b_, _, _] :=
	a === b

simpleMatchQ[ _, Blank$TM[]] :=
	True
simpleMatchQ[ TYPED$[ TrueQ$TM, {}][ expr_], patt0_] :=
	simpleMatchQ[
		expr,
		Replace[ patt0,
			{
				TYPED$[ TrueQ$TM, {}][ patt_] :> patt,
				TrueQ$TM[ patt_] :> patt
			}
		]
	]
simpleMatchQ[ TYPED$[ expr_, typ1_], TYPED$[ patt_, typ2_]] :=
	(typ2 === TVAR$[] || typ1 === typ2) && simpleMatchQ[ expr, patt]
simpleMatchQ[ TYPED$[ expr_, _], patt_] :=
	simpleMatchQ[ expr, patt]
simpleMatchQ[ expr_, TYPED$[ patt_, typ_]] :=
	(typ === TVAR$[] || getType[ expr, {}, {}, {}] === typ) && simpleMatchQ[ expr, patt]
simpleMatchQ[ expr:(_VAR$|_FIX$|_META$), patt_] :=
	expr === patt
simpleMatchQ[ Lambda$TM[ RNG$[ TYPRNG$[ _, typ1_]], expr_], patt0_] :=
	Replace[ patt0,
		{
			Lambda$TM[ RNG$[ TYPRNG$[ _, typ2_]], patt_] :> (typ2 === TVAR$[] || typ1 === typ2) && simpleMatchQ[ expr, patt],
			_ -> False
		}
	]
simpleMatchQ[ _, _VAR$|_FIX$|_META$|_Lambda$TM] :=
	False
simpleMatchQ[ f_[ a_], patt_] :=
	Replace[ patt,
		{
			g_[ b_] :> simpleMatchQ[ f, g] && simpleMatchQ[ a, b],
			_ -> False
		}
	]
simpleMatchQ[ expr_, patt_] :=
	expr === patt

(* 'termToTermPatt' is similar to 'simpleMatchQ' in that it is also used for quickly checking alpha-equivalence of two
	terms without looking at types. The main difference between the two functions is that the argument 'term' of
	'termToTermPatt' must be the result of type inference, i.e. all atoms--and only these--are type-annotated; the
	pattern in 'simpleMatchQ' does not need to satisfy this requirement. *)
termToTermPatt[ term_] :=
	term /. {TYPED$[ atom_, _] :> TYPED$[ atom, Blank[]], _RNG$ -> Blank[ RNG$]}

containsVar[ expr_, pred_] :=
	Catch[ scanVars[ expr, pred, Throw[ True]&]; False]

mapVars[ l_List, pred_, fun_] :=
	Map[ mapVars[ #, pred, fun]&, l]
mapVars[ expr_, pred_, fun_] :=
	ReplacePart[ expr, reap[ scanVars[ expr, pred, Sow[ {#1} -> fun[ #1, #2, #3]]&]]]

mapFreeVars[ l_List, fun_] :=
	Map[ mapFreeVars[ #, fun]&, l]
mapFreeVars[ expr_, fun_] :=
	ReplacePart[ expr, reap[ scanVars[ expr, If[ TrueQ[ #3 >= #2], Sow[ {#1} -> fun[ #1, #2, #3]]]&]]]

unfoldTriangularSubst[ {}, ___] := {}
unfoldTriangularSubst[ {f_, r___}] :=
	normalize[ Fold[ Append[ #1 /. #2, #2] &, {f}, {r}]]
unfoldTriangularSubst[ subst_List, patt_] :=
	With[ {init = LengthWhile[ subst, !MatchQ[ First[ #], patt] &]},
		unfoldTriangularSubst[ Drop[ subst, init], patt, True]
	]
unfoldTriangularSubst[ {f_, r___}, patt_, True] :=
	normalize[ Fold[ With[ {tmp = #1 /. #2}, If[ MatchQ[ First[ #2], patt], Append[ tmp, #2], tmp]]&, {f}, {r}]]

substFree[ expr_, subst_List] :=
	With[ {newSubst = Replace[ subst, (i_ -> inst_) :> {i, inst, reap[ scanVars[ inst, If[ #3 >= #2, Sow[ {#1} -> VAR$[ #3]]]&]]}, {1}]},
		ReplacePart[
			expr,
			reap[
				scanVars[
					expr,
					(#3 >= #2)&,
					Function[ {pos, abs, i},
						With[ {new = Cases[ newSubst, {i - abs, inst_, rep_} :> {inst, rep}, {1}, 1]},
							If[ new =!= {},
								With[ {inst = new[[1, 1]], rep = new[[1, 2]]},
									Sow[ {pos} -> If[ abs === 0, inst, ReplacePart[ inst, Replace[ rep, (p_ -> VAR$[ j_]) :> (p -> VAR$[ j + abs]), {1}]]]]
								]
							]
						]
					]
				]
			]
		]
	]

(* Attention! 'applyCurried' assumes its third argument to be in reversed order,
	like lists of types of loose bound variables originating from stripping off binders.
	That means that the binders can simply be reconstructed by calling 'applyCurried'. *)
applyCurried[ h_, a_, abs_] :=
	applyCurried[ h, a, abs, False, False]
applyCurried[ h_, a:{___, VAR$[ 0]}, abs:Except[ {}, _List], fullEtaReduce_, varNames_] :=
	With[ {absNum = Length[ abs]},
	Module[ {i = -1},
		Scan[ If[ ++i >= absNum || # =!= VAR$[ i], Return[]]&, Reverse[ a]];
		applyCurried[ h, Drop[ a, -i], i, abs, fullEtaReduce, varNames]
	]]
applyCurried[ h_, a_List, abs_List, fullEtaReduce_, varNames_] :=
	applyCurried[ h, a, 0, abs, fullEtaReduce, varNames]
applyCurried[ h_[ VAR$[ n_]], {}, n_, abs:Except[ {}, _List], True, varNames_] :=
	With[ {absNum = Length[ abs]},
	Module[ {i = n},
		applyCurriedAux[ NestWhile[ Head, h, (++i < absNum && MatchQ[ #, _[ VAR$[ i]]])&], {}, i, abs, varNames]
	]]
applyCurried[ h_, a_List, 0, abs_List, _, varNames_] :=
	Fold[ If[ TrueQ[ varNames], Lambda$TM[ RNG$[ TYPRNG$ @@ #2], #1] &, Lambda$TM[ RNG$[ TYPRNG$[ Null, #2]], #1] &], Fold[ applyFunction, h, a], abs]
applyCurried[ h_, a_List, n_, abs_List, _, varNames_] :=
	applyCurriedAux[ h, a, n, abs, varNames]

applyCurriedAux[ h_, a_List, n_Integer, abs_List, varNames_] :=
	With[ {fun = If[ TrueQ[ varNames], Lambda$TM[ RNG$[ TYPRNG$ @@ #2], #1] &, Lambda$TM[ RNG$[ TYPRNG$[ varNames, #2]], #1] &]},
	Module[ {index = n},
		Catch[
			With[ {func =
					Which[
						#3 === #2,
						Throw[ Fold[ fun, Fold[ applyFunction, h, Join[ a, Table[ VAR$[ k], {k, n - 1, 0, -1}]]], abs]],
						
						#3 > #2,
						Sow[ {#1} -> #3];
						If[ #3 - #2 < index, index = #3 - #2]
					]&},
			With[ {reph0 = reap[ scanVars[ h, func]],
					repa0 =
						Map[
							reap[ scanVars[ #, func]]&,
							a
						]},
			With[ {reph = Replace[ reph0, (p_ -> j_) :> (p -> VAR$[ j - index]), {1}],
					repa = Replace[ repa0, (p_ -> j_) :> (p -> VAR$[ j - index]), {2}]},
				Fold[
					fun,
					Fold[ applyFunction, ReplacePart[ h, reph], Join[ MapThread[ ReplacePart[ #1, #2]&, {a, repa}], Table[ VAR$[ n - k], {k, index + 1, n}]]],
					Drop[ abs, index]
				]
			]]]
		]
	]]

getArgs[ expr_, n_Integer?NonNegative] :=
	Extract[ expr, Table[ Append[ Table[ 0, {k}], 1], {k, n - 1, 0, -1}]]

splitHeadArgs[ expr_] := splitHeadArgs[ expr, {}]
splitHeadArgs[ h_Lambda$TM, {}] := {h, {}}
splitHeadArgs[ Lambda$TM[ _, e_], {a_, aa___}] := splitHeadArgs[ betaReduce[ e, a], {aa}]
splitHeadArgs[ h:(_VAR$|_META$|_FIX$|_TYPED$|_Complex|_DirectedInfinity|_Rational), acc_] := {h, acc}
splitHeadArgs[ h_[ a___], acc_] := splitHeadArgs[ h, Join[ {a}, acc]]
splitHeadArgs[ h_, acc_] := {h, acc}

abstractionCount[ Lambda$TM[ _RNG$, a_], m_Integer, n_?Positive] :=
	abstractionCount[ a, m + 1, n - 1]
abstractionCount[ _, m_, _] :=
	m

stripAbstractions[ expr_] :=
	NestWhile[ Last, expr, MatchQ[ #, Lambda$TM[ _RNG$, _]] &]
stripAbstractions[ Lambda$TM[ RNG$[ TYPRNG$[ _, t_]], a_], acc_] :=
	stripAbstractions[ a, Prepend[ acc, t]]
stripAbstractions[ expr_, acc_] :=
	{expr, acc}

(* 'betaReduceRestricted[ expr, n]' reduces the *at most* 'n' outermost beta-redexes in 'expr'. 'n' could be 'Infinity', in
	which case all outermost beta-redexes are reduced. *)
betaReduceRestricted[ expr_, n_?Positive] :=
	betaReduceRestricted[ expr, n, {}, True]
betaReduceRestricted[ expr_, _] :=
	expr
betaReduceRestricted[ Lambda$TM[ _RNG$, e_], n_?Positive, {a_, aa___}, _] :=
	betaReduceRestricted[ betaReduce[ e, a], n - 1, {aa}, False]
betaReduceRestricted[ (h:Except[ Lambda$TM|VAR$|META$|FIX$|TYPED$|Complex|DirectedInfinity|Rational])[ a___], n_, acc_, True] :=
	betaReduceRestricted[ h, n, Join[ {a}, acc], True]
betaReduceRestricted[ expr_, _, args_List, _] :=
	Fold[ #1[ #2] &, expr, args]	(* we do not use 'applyCurried' nor 'applyFunction', because there must not be beta-reductions *)

(* 'betaExpand[ expr, typs, n]' returns the expression '(\x1 x2 ... xn. expr0)[ x1, x2, ..., xn]' that beta-reduces to 'expr'.
	'betaExpand[ expr, typs, pos]' returns the expression '(\x. expr0)[ expr|_pos]', if no variable bound in 'expr' is free in
	'expr|_pos', and if the subexpressions of 'expr' at the given positions are all alpha-equivalent to each other (note that
	'pos' is supposed to be a *list* of positions). *)
betaExpand[ expr_, 0|{}, _, _, _] :=
	expr
betaExpand[ expr_, n_Integer, loose_List, _List, _List] :=
	Fold[
		#1[ #2] &,
		Fold[ Lambda$TM[ RNG$[ TYPRNG$[ Null, #2]], #1] &, mapVars[ expr, (#3 >= #2 + n)&, VAR$[ #3 + n]&], Take[ loose, n]],
		Table[ VAR$[ n - i], {i, 1, n}]
	]
betaExpand[ expr_, pos:{p1_List, pr___}, loose_List, fix_List, meta_List] :=
	Replace[ validExtract[ expr, p1, {}],
		{
			{sub_, newLoose_List} :>
				With[ {typ = getType[ sub, Join[ newLoose, loose], fix, meta], n = Length[ newLoose]},
					If[ typ === $Failed || (n =!= 0 && containsVar[ sub, (#3 < n)&]),
						$Failed,
					(*else*)
						Catch[
							Lambda$TM[ RNG$[ TYPRNG$[ Null, typ]],
								mapVars[
									ReplacePart[ expr,
										Prepend[
											Map[
												Replace[ validExtract[ expr, #, {}],
													{
														{sub2_, l2_List} :>
															With[ {n2 = Length[ l2]},
																If[ alphaEquivOffset[ sub, sub2, 0, n - n2],
																	{#} -> VAR$[ n2],
																(*else*)
																	Throw[ $Failed]
																]
															],
														_ :> Throw[ $Failed]
													}
												]&,
												{pr}
											],
											{p1} -> VAR$[ n]
										]
									],
									(#3 >= #2 && #1 =!= p1 && !MemberQ[ {pr}, #1]) &,
									VAR$[ #3 + 1]&
								]
							][ mapFreeVars[ sub, VAR$[ #3 - n]&]]
						]
					]
				],
			_ -> $Failed
		}
	]

(* 'etaReduceRestricted[ expr, n]' reduces *at most* the 'n' outermost eta-redexes in 'expr'. 'n' could be 'Infinity', in
	which case all outermost eta-redexes are reduced. *)
etaReduceRestricted[ expr_, n_?Positive] :=
	etaReduceRestricted[ expr, n, {}]
etaReduceRestricted[ expr_, _] :=
	expr
etaReduceRestricted[ (term:Except[ Complex|DirectedInfinity|Rational|Lambda$TM])[ VAR$[ 0]], n_?Positive, rngs:Except[ {}]] :=
	With[ {res = etaReduce[ term, True]},
		etaReduceRestricted[ ReplacePart[ term, res], n - 1, Rest[ rngs]] /; ListQ[ res]
	]
etaReduceRestricted[ Lambda$TM[ RNG$[ r_TYPRNG$], expr_], n_, rngs_List] :=
	etaReduceRestricted[ expr, n, Prepend[ rngs, r]]
etaReduceRestricted[ expr_, _, rngs_List] :=
	Fold[ Lambda$TM[ RNG$[ #2], #1] &, expr, rngs]

(* 'etaExpand[ expr, loose]' fully eta-expands 'expr'. *)
etaExpand[ expr_, loose_List, fix_List, meta_List] :=
	etaExpand[ expr, loose, fix, meta, {}]
etaExpand[ v:VAR$[ i_Integer], loose_, _, _, args_] :=
	etaExpand[ v, args, Most[ Drop[ flattenTypes[ loose[[i + 1]], {}], Length[ args]]]]
etaExpand[ c_FIX$, _, fix_, _, args_] :=
	etaExpand[ c, args, Most[ Drop[ flattenTypes[ Replace[ c, fix], {}], Length[ args]]]]
etaExpand[ m_META$, _, _, meta_, args_] :=
	etaExpand[ m, args, Most[ Drop[ flattenTypes[ First[ Replace[ m, meta]], {}], Length[ args]]]]
etaExpand[ expr:TYPED$[ _, _], _, _, _, args_] :=
	etaExpand[ expr, args, Most[ Drop[ flattenTypes[ getFullTypeNoSorts[ expr], {}], Length[ args]]]]
etaExpand[ Lambda$TM[ rng:RNG$[ TYPRNG$[ _, typ_]], expr_], loose_, fix_, meta_, args_] :=
	etaExpand[ Lambda$TM[ rng, etaExpand[ expr, Prepend[ loose, typ], fix, meta, {}]], args, {}]
etaExpand[ h_[ a___], loose_, fix_, meta_, args_] :=
	etaExpand[ h, loose, fix, meta, Join[ Map[ etaExpand[ #, loose, fix, meta, {}]&, {a}], args]]
etaExpand[ expr_, _, _, _, args_] :=
	etaExpand[ expr, args, {}]
etaExpand[ expr_, args_, typs_] :=
	etaExpand[ Fold[ #1[ #2] &, expr, args], typs]
etaExpand[ expr_, {}] :=
	expr
etaExpand[ expr_, typs_] :=
	With[ {n = Length[ typs]},
		Fold[
			Lambda$TM[ RNG$[ TYPRNG$[ Null, #2]], #1] &,
			Fold[ #1[ #2] &, mapFreeVars[ expr, VAR$[ #3 + n]&], Table[ VAR$[ n - i], {i, 1, n}]],
			typs
		]
	]


(* ::Section:: *)
(*Formulas as Rules*)

(* Collections of formula-rules are stored as 'RULEC$'-expressions of the form 'RULEC$[ c, v, depth]', where both 'c'
	and 'v' are associations with elements of the form 'thy -> rules'. 'thy' is a string specifying the theory the
	rules in 'rules' originate from (including the "Source:"-prefix; "THM$" indicates that the respective rules are
	registered theorems), and 'rules' is an association ultimately mapping characteristic constants to sequences of
	formula-keys (but possibly with deeper nesting, e.g. associations of associations of  sequences). The depth of the
	nesting is specified by 'depth'; it must be at least '1'.
	The difference between the two lists 'c' and 'v' lies in the fact that the *outermost* characteristic constants in
	'c' are not 'VAR$', whereas those in 'v' are precisely 'VAR$'. Since the outermost characteristic constants in 'v'
	are all 'VAR$', the first level of nesting in 'v' is skipped; that's also the reason why 'depth' must be stored
	explicitly: if it is '1', 'v' must map theories to sequences with head 'Hold', otherwise it must map them again to
	associations.
	Note that the order of the elements in 'c' and 'v', as well as those in 'rules', is completely random.
	
	Example:
		'RULEC$[ <|"thy1" -> <|c1 -> Hold[ k1, k2], c2 -> Hold[ k3]|>, "thy2" -> <|c1 -> Hold[ k4]|>|>,
				 <|"thy1" -> Hold[ k5, k6, k7], "thy3" -> Hold[ k8]|>]'
	
	'RULEC$'-structures can be extended to structures of the form 'RULEC$[ c, cc, v, vc, ci, vi, depth]', where 'c' and
	'v' have the same meaning as above, 'cc' and 'vc' are the rules originating from the current theory and coming
	before the current proof goal, and 'ci' and 'vi' are the indices of the theories in 'c' and 'v', respectively, the
	current theory depends upon. *)

(* The associations in 'RULEC$-structures must be such that no value is an individual element, i.e. every value must be
	an association itself or a sequence (with head 'Hold') that shall be iterated over; of course, empty/singleton
	associations/sequences are allowed.
	ATTENTION! Although sequences have head 'Hold', if their elements shall be kept unevaluated they should be put
	inside 'HoldComplete'! Some functions, like 'addRulesToRULEC$' and 'iterateRULEC$', evaluate elements of sequences. *)

makeRULEC$[ depth_Integer] :=
	makeRULEC$[ Association[], Association[], depth]
makeRULEC$[ c_Association, v_Association, depth_Integer] :=
	RULEC$[ Association[], Association[], depth]

(* 'iterateRULEC$[ it, rc]' iterates over the extended 'RULEC$'-structure 'rc', using 'incrementRuleIterator'. *)
SetAttributes[ iterateRULEC$, HoldFirst];
iterateRULEC$[ it_, RULEC$[ c_Association, cc_, v_Association, vc_, ci_List, vi_List, _Integer]] :=
	Catch[
		While[ True,
			Replace[ it,
				{
					{1, j_Integer, it0_, orig_List} :>
						With[ {out = (it = it0; incrementRuleIterator[ it, c[[ci[[j]]]]])},
							Which[
								it =!= $Failed,
								it = {1, j, it, orig};
								Throw[ out],
							(*else*)
								j >= 2,
								it = {1, j - 1, orig, orig},
							(*else*)
								True,
								it = {2, orig, orig}
							]
						],
					{2, it0_, orig_List} :>
						With[ {out = (it = it0; incrementRuleIterator[ it, cc])},
							If[ it =!= $Failed,
								it = {2, it, orig};
								Throw[ out],
							(*else*)
								With[ {origR = If[ orig === {}, {}, Rest[ orig]]},
									it = If[ vi === {}, {4, origR, origR}, {3, Length[ vi], origR, origR}]
								]
							]
						],
					{3, j_Integer, it0_, orig_List} :>
						With[ {out = (it = it0; incrementRuleIterator[ it, v[[vi[[j]]]]])},
							Which[
								it =!= $Failed,
								it = {3, j, it, orig};
								Throw[ out],
							(*else*)
								j >= 2,
								it = {3, j - 1, orig, orig},
							(*else*)
								True,
								it = {4, orig, orig}
							]
						],
					{_, it0_, orig_List} :>
						With[ {out = (it = it0; incrementRuleIterator[ it, vc])},
							If[ it =!= $Failed,
								it = {4, it, orig};
								Throw[ out],
							(*else*)
								Throw[ it = $Failed]
							]
						],
					{Key[ VAR$], l___} :> (it = If[ vi === {}, {4, {l}, {l}}, {3, Length[ vi], {l}, {l}}]),
					l_List :> (it = If[ ci === {}, {2, l, l}, {1, Length[ ci], l, l}]),
					_ :> Throw[ it = $Failed]
				}
			]
		]
	]

(* Sometimes, especially when dynamically adding and removing rules on-the-fly (as in rewriting), it is better to first
	convert 'RULEC$'-structures into "flat" associations where rules are not grouped together according to their
	theories, and where only the rules currently in scope are present. *)
associationFromRULEC$[ RULEC$[ c_Association, cc_, v_Association, vc_, ci_List, vi_List, depth_Integer]] :=
	With[ {a1 = mergeRuleAssociations[ Prepend[ Values[ c][[ci]], cc]],
			a2 = mergeRuleAssociations[ Prepend[ Values[ v][[vi]], vc]]},
		(* When iterating over a 'RULEC$'-structure, first 'c' and then 'cc' is visited. Hence, since flat associations are
			iterated over in reverse order, we do 'Prepend[ Values[ c][[ci]], cc]' rather than 'Append', to preserve the
			original iteration-order. *)
		If[ a2 === Hold[] || a2 === Association[],
			a1,
		(*else*)
			Prepend[ a1, a2]
		]
	]

SetAttributes[ extendRULEC$, HoldFirst];
extendRULEC$[ rc_Symbol, False] :=
	Replace[ rc,
		{
			RULEC$[ c_Association, v_Association, depth_Integer] :>
				(rc = RULEC$[ c, Association[], v, If[ depth === 1, Hold[], Association[]], Range[ Length[ c]], Range[ Length[ v]], depth]),
			_ :> rc
		}
	]
extendRULEC$[ rc_Symbol, ___] :=
	Replace[ rc,
		{
			RULEC$[ c_Association, v_Association, depth_Integer] :>
				(
					rc = Replace[ getTheoryScope[],
						{
							{} :>
								RULEC$[
									c,
									Association[],
									v,
									If[ depth === 1, Hold[], Association[]],
									First /@ Position[ Keys[ c], "THM$", {1}, 1, Heads -> False],
									First /@ Position[ Keys[ v], "THM$", {1}, 1, Heads -> False],
									depth
								],
							{cur_String, deps___} :>
								RULEC$[
									c,
									remRulesFromAssociationByPatt[ Lookup[ c, cur, Association[]], _List?(Composition[ Not, inScopeQ])],
									v,
									remRulesFromAssociationByPatt[ Lookup[ v, cur, If[ depth === 1, Hold[], Association[]]], _List?(Composition[ Not, inScopeQ])],
									First /@ Position[ Keys[ c], "THM$"|_?(MemberQ[ {deps}, #]&), {1}, Infinity, Heads -> False],
									First /@ Position[ Keys[ v], "THM$"|_?(MemberQ[ {deps}, #]&), {1}, Infinity, Heads -> False],
									depth
								]
						}
					]
				),
			_ :> rc
		}
	]

SetAttributes[ restrictRULEC$, HoldFirst];
restrictRULEC$[ rc_Symbol] :=
	Replace[ rc,
		{
			RULEC$[ c_Association, _, v_Association, _, _, _, depth_Integer] :> (rc = RULEC$[ c, v, depth]),
			_ :> rc
		}
	]

SetAttributes[ addRulesToRULEC$, HoldFirst];
addRulesToRULEC$[ rc_Symbol][ spec_Rule] :=
	addRulesToRULEC$[ rc, spec]
addRulesToRULEC$[ rc_Symbol, order:Except[ _Rule], rest___][ spec_Rule] :=
	addRulesToRULEC$[ rc, spec, order, rest]
addRulesToRULEC$[ rc_Symbol, spec_Rule] :=
	addRulesToRULEC$[ rc, spec, Prepend]
addRulesToRULEC$[ rc_Symbol, spec_Rule, order_] :=
	addRulesToRULEC$[ rc, spec, order, None]
addRulesToRULEC$[ rc_Symbol, {} -> _, _, _] :=
	rc
addRulesToRULEC$[ rc_Symbol, _ -> Hold[], _, _] :=
	rc
addRulesToRULEC$[ rc_Symbol, {c_, rest___} -> entries_Hold, order_, dupF_] :=
	Module[ {l},
		rc =
			If[ c === VAR$,
				l = restrictRULEC$[ rc][[2]];
				addRulesToRULEC$[ l, {rest} -> entries, order, dupF, If[ Last[ rc] === 1, Hold[], Association[]]];
				rc = ReplacePart[ rc, 2 -> l],
			(*else*)
				l = First[ restrictRULEC$[ rc]];
				addRulesToRULEC$[ l, {c, rest} -> entries, order, dupF, Association[]];
				rc = ReplacePart[ rc, 1 -> l]
			]
	]
addRulesToRULEC$[ rc_Symbol, c_List -> entry_, order_, dupF_] :=
	addRulesToRULEC$[ rc, c -> Hold[ entry], order, dupF]
addRulesToRULEC$[ assoc_Symbol, c_List -> entries_Hold, order_, dupF_, def_] :=
	With[ {thmPos = Position[ entries, _THM$, {1}, Heads -> False]},
	Module[ {l},
		If[ thmPos =!= {},
			l = Lookup[ assoc, "THM$", def];
			AssociateTo[ assoc, "THM$" -> addRulesToAssociation[ l, c -> Part[ entries, First /@ thmPos], order, dupF]]
		];
		Scan[
			Function[ entries0,
			With[ {thy = entries0[[1, 2]]},
				l = Lookup[ assoc, thy, def];
				AssociateTo[ assoc, thy -> addRulesToAssociation[ l, c -> (Hold @@ entries0), order, dupF]]
			]],
			GatherBy[ List @@ Delete[ entries, thmPos], Last]
		]
	]]

SetAttributes[ remRulesFromRULEC$, HoldFirst];
remRulesFromRULEC$[ rc_Symbol][ spec_Rule] :=
	remRulesFromRULEC$[ rc, spec]
remRulesFromRULEC$[ rc_Symbol, missingF:Except[ _Rule]][ spec_Rule] :=
	remRulesFromRULEC$[ rc, spec, missingF]
remRulesFromRULEC$[ rc_Symbol, spec_Rule] :=
	remRulesFromRULEC$[ rc, spec, None]
remRulesFromRULEC$[ rc_Symbol, {} -> _, _] :=
	rc
remRulesFromRULEC$[ rc_Symbol, _ -> Hold[], _] :=
	rc
remRulesFromRULEC$[ rc_Symbol, {c_, rest___} -> entries_Hold, missingF_] :=
	Module[ {l},
		rc =
			If[ c === VAR$,
				l = Last[ restrictRULEC$[ rc]];
				remRulesFromRULEC$[ l, {rest} -> entries, missingF, If[ Last[ rc] === 1, Hold[], Association[]]];
				rc = ReplacePart[ rc, 2 -> l],
			(*else*)
				l = First[ restrictRULEC$[ rc]];
				remRulesFromRULEC$[ l, {c, rest} -> entries, missingF, Association[]];
				rc = ReplacePart[ rc, 1 -> l]
			]
	]
remRulesFromRULEC$[ rc_Symbol, c_List -> entry_, missingF_] :=
	remRulesFromRULEC$[ rc, c -> Hold[ entry], missingF]
remRulesFromRULEC$[ assoc_Symbol, c_List -> entries_Hold, missingF_, def_] :=
	With[ {thmPos = Position[ entries, _THM$, {1}, Heads -> False]},
	Module[ {l},
		If[ thmPos =!= {},
			l = Lookup[ assoc, "THM$", def];
			remRulesFromAssociation[ l, c -> Extract[ entries, thmPos], missingF];
			If[ l === def,
				KeyDropFrom[ assoc, "THM$"],
			(*else*)
				AssociateTo[ assoc, "THM$" -> l]
			]
		];
		Scan[
			Function[ entries0,
			With[ {thy = entries0[[1, 2]]},
				l = Lookup[ assoc, thy, def];
				remRulesFromAssociation[ l, c -> (Hold @@ entries0), missingF];
				If[ l === def,
					KeyDropFrom[ assoc, thy],
				(*else*)
					AssociateTo[ assoc, thy -> l]
				]
			]],
			GatherBy[ List @@ Delete[ entries, thmPos], Last]
		]
	]]

SetAttributes[ remRulesByPattFromRULEC$, HoldFirst];
remRulesByPattFromRULEC$[ rc_Symbol, patt_] :=
	(
		restrictRULEC$[ rc];
		rc = MapAt[ remRulesFromAssociationByPatt[ patt], rc, {{1}, {2}}]
	)

(* 'incrementRuleIterator' increments the given iterator and returns the next element, if possible; otherwise it sets
	the iterator to '$Failed' and returns '$Failed'. If the given iterator is an initial "pseudo iterator", it is
	automatically initialized by a proper iterator.
	Note that sequences and associations are iterated through in reverse order, starting from the last element.
	When iterating through a sequence whose elements are a mixture of associations and non-associations, like in
	'Hold[ 1, <|x->Hold[ 2]|>, 3]', the first element visited is '3', followed by '2', followed by '1', i.e. the
	association as a whole is not visited. *)
SetAttributes[ incrementRuleIterator, HoldFirst];
incrementRuleIterator[ it_, assoc_] :=
	Replace[ it,
		{
			{{pre___, i_Integer}, orig_List} :>
				If[ i <= 1,
					it = {{pre}, orig};
					incrementRuleIterator[ it, assoc],
				(*else*)
					it = Join[ {pre, i - 1}, With[ {l = Length[ {pre}] + 1}, If[ l < Length[ orig], Drop[ orig, l], {}]]];
					initRuleIterator[
						it,
						assoc,
						{},
						orig,
						assoc
					]
				],
			{{pre___, Key[ c_]}, orig_List} :>
				If[ c === VAR$,
					it = {{pre}, orig};
					incrementRuleIterator[ it, assoc],
				(*else*)
					it = Join[ {pre, Key[ VAR$]}, With[ {l = Length[ {pre}] + 1}, If[ l < Length[ orig], Drop[ orig, l], {}]]];
					initRuleIterator[
						it,
						assoc,
						{},
						orig,
						assoc
					]
				],
			{{}, _List} :> (it = $Failed),
			_List :> initRuleIterator[ it, assoc],
			_ :> (it = $Failed)
		}
	]

(* 'initRuleIterator' takes an initial "pseudo iterator" of an association and sets it to a proper iterator of that
	association that can be used in functions like 'incrementRuleIterator'. Such an interator is a pair consisting of
	a position specification (which might involve 'Key[_]') and the original initial pseudo iterator. *)
SetAttributes[ initRuleIterator, HoldFirst];
initRuleIterator[ it_, assoc_] :=
	initRuleIterator[ it, assoc, {}, it, assoc]
initRuleIterator[ it_, assoc_, itPre_List, initAll_List, assocAll_] :=
	Replace[ it,
		{
			{pre___, All, post___} :>
				With[ {aux = extractSingle[ assoc, {pre}, Hold[]]},
					If[ MatchQ[ aux, (Association|Hold)[]],
						it = {Join[ itPre, {pre, 1}], initAll};
						incrementRuleIterator[ it, assocAll],
					(*else*)
						it = {post};
						initRuleIterator[ it, Last[ aux], Join[ itPre, {pre, Length[ aux]}], initAll, assocAll]
					]
				],
			init_List :>
				With[ {aux = extractSingle[ assoc, init, Hold[]]},
					Switch[ aux,
						(Association|Hold)[],
						it = {Join[ itPre, init, {1}], initAll};
						incrementRuleIterator[ it, assocAll],
					(*else*)
						_Association,
						it = {};
						initRuleIterator[ it, Last[ aux], Join[ itPre, init, {Length[ aux]}], initAll, assocAll],
					(*else*)
						_Hold,
						it = {Join[ itPre, init, {Length[ aux]}], initAll};
						Last[ aux],
					(*else*)
						_,
						it = {Join[ itPre, init], initAll};
						aux
					]
				]
		}
	]

(* 'extractRelevantRules[ assoc, {c}]' extracts from 'assoc' all rules relevant for 'c'. It can only deal with one
	level of nesting, i.e. the values of 'assoc' must be sequences with head 'Hold'. *)
extractRelevantRules[ assoc_Association, {}] :=
	Join @@ Values[ assoc]
extractRelevantRules[ assoc_Association, {Key[ VAR$]}] :=
	Lookup[ assoc, VAR$, Hold[]]
extractRelevantRules[ assoc_Association, {Key[ c_]}] :=
	Join[ Lookup[ assoc, c, Hold[]], Lookup[ assoc, VAR$, Hold[]]]	(* first 'c', then 'VAR$' *)

(* 'addRulesToAssociation' ensures that entries with key 'VAR$' come at the beginning of each association. *)
SetAttributes[ addRulesToAssociation, HoldFirst];
addRulesToAssociation[ assoc_][ spec_Rule] :=
	addRulesToAssociation[ assoc, spec]
addRulesToAssociation[ assoc_, order:Except[ _Rule], rest___][ spec_Rule] :=
	addRulesToAssociation[ assoc, spec, order, rest]
addRulesToAssociation[ assoc_, spec_Rule] :=
	addRulesToAssociation[ assoc, spec, Prepend]
addRulesToAssociation[ assoc_, spec_Rule, order_] :=
	addRulesToAssociation[ assoc, spec, order, None]
addRulesToAssociation[ assoc_, {} -> entries_Hold, order_, dupF_] :=
	If[ Head[ assoc] === Hold,
		With[ {newP = Position[ entries, x_ /; !MemberQ[ assoc, HoldPattern[ x]], {1}, Heads -> False]},
		With[ {new = entries[[newP[[All, 1]]]]},
			If[ dupF =!= None, dupF[ Delete[ entries, newP], Length[ newP]]];
			If[ new === Hold[],
				assoc,
			(*else*)
				assoc = If[ order === Append, Join[ assoc, new], Join[ new, assoc]]
			]
		]],
	(*else*)
		assoc
	]
addRulesToAssociation[ assoc_, {c_} -> entries_Hold, order_, dupF_] :=
	With[ {l = Lookup[ assoc, c, Hold[]]},
	With[ {newP = Position[ entries, x_ /; !MemberQ[ l, HoldPattern[ x]], {1}, Heads -> False]},
	With[ {new = entries[[newP[[All, 1]]]]},
		If[ dupF =!= None, dupF[ Delete[ entries, newP], Length[ newP]]];
		If[ new === Hold[],
			assoc,
		(*else*)
			If[ c === VAR$, PrependTo, AssociateTo][ assoc, c -> If[ order === Append, Join[ l, new], Join[ new, l]]]
		]
	]]]
addRulesToAssociation[ assoc_, {c_, rest__} -> entries_Hold, order_, dupF_] :=
	Module[ {l = Lookup[ assoc, c, Association[]]},
		addRulesToAssociation[ l, {rest} -> entries, order, dupF];
		If[ c === VAR$, PrependTo, AssociateTo][ assoc, c -> l]
	]
addRulesToAssociation[ assoc_, c_List -> entry_, order_, dupF_] :=
	addRulesToAssociation[ assoc, c -> Hold[ entry], order, dupF]

SetAttributes[ remRulesFromAssociation, HoldFirst];
remRulesFromAssociation[ assoc_][ spec_Rule] :=
	remRulesFromAssociation[ assoc, spec]
remRulesFromAssociation[ assoc_, missingF:Except[ _Rule]][ spec_Rule] :=
	remRulesFromAssociation[ assoc, spec, missingF]
remRulesFromAssociation[ assoc_, spec_Rule] :=
	remRulesFromAssociation[ assoc, spec, None]
remRulesFromAssociation[ assoc_, {} -> entries_Hold, missingF_] :=
	If[ Head[ assoc] === Hold,
		With[ {new = DeleteCases[ assoc, Alternatives @@@ HoldPattern[ entries]]},
			If[ missingF =!= None,
				With[ {missing = Select[ entries, Function[ x, !MemberQ[ assoc, HoldPattern[ x]], {HoldFirst}]]},
					missingF[ missing, Length[ entries] - Length[ missing]]
				]
			];
			assoc = new
		],
	(*else*)
		assoc
	]
remRulesFromAssociation[ assoc_, {c_} -> entries_Hold, missingF_] :=
	With[ {l = Lookup[ assoc, c, Hold[]]},
	With[ {new = DeleteCases[ l, Alternatives @@@ HoldPattern[ entries]]},
		If[ missingF =!= None,
			With[ {missing = Select[ entries, Function[ x, !MemberQ[ l, HoldPattern[ x]], {HoldFirst}]]},
				missingF[ missing, Length[ entries] - Length[ missing]]
			]
		];
		If[ new === Hold[],
			KeyDropFrom[ assoc, c],
		(*else*)
			If[ c === VAR$, PrependTo, AssociateTo][ assoc, c -> new]
		]
	]]
remRulesFromAssociation[ assoc_, c:{_} -> entry_, missingF_] :=
	remRulesFromAssociation[ assoc, c -> Hold[ entry], missingF]
remRulesFromAssociation[ assoc_, {c_, rest__} -> entries_, missingF_] :=
	Module[ {l = Lookup[ assoc, c, Association[]]},
		remRulesFromAssociation[ l, {rest} -> entries, missingF];
		If[ l === Association[],
			KeyDropFrom[ assoc, c],
		(*else*)
			If[ c === VAR$, PrependTo, AssociateTo][ assoc, c -> l]
		]
	]

(* Attention! In contrast to 'remRulesFromAssociation', 'remRulesFromAssociationByPatt' is *not* destructive, i.e. it
	does not change the given association in place.
	Also note that 'remRulesFromAssociationByPatt' works for associations, sequences with head 'Hold', and lists alike. *)
remRulesFromAssociationByPatt[ patt_][ assoc_] :=
	remRulesFromAssociationByPatt[ assoc, patt]
remRulesFromAssociationByPatt[ assoc_Association, patt_] :=
	DeleteCases[ remRulesFromAssociationByPatt[ patt] /@ assoc, _[]]
remRulesFromAssociationByPatt[ assoc_, patt_] :=
	DeleteCases[ assoc, patt]

mergeRuleAssociations[ {c_}] :=
	c
mergeRuleAssociations[ l:{__Association}] :=
	Merge[ l, mergeRuleAssociations]
mergeRuleAssociations[ {l__Hold}] :=
	Join[ l]

(* 'getRuleConsts' takes a formula 'f', a category 'cat' and a boolean value 'isRule' as input and returns a list of
	constants 'c' that are characteristic for 'f' viewed as a rule/goal of category 'cat'. The elements of 'c' are
	proper untyped Core constants, 'Lambda$TM', 'VAR$' or a.b.f. constants; 'VAR$' comprises both bound variables and
	meta variables.
	'cat' may be one of the following:
	- "intro": The characteristic constant of an introduction rule is the atomic head of the atomic consequent of the
		rule after stripping off all universal quantifiers.
	- "elim": The characteristic constant of an elimination rule is the atomic head of the first premise of the rule
		after stripping off all universal quantifiers.
	- "simp": The characteristic constant of a simplification rule is the atomic head of the left-hand-side of the
		rule, which must be a (quantified, conditional) Core equality.
	- "trans": At the moment "trans" is like "elim", but in the future this might change.
	- 'None': The characteristic constant of category 'None' is the atomic head of the given term, even if it is
		'CoreAll', 'CoreImplies'. The outermost 'TrueQ' is skipped, though.
	- 'Head': 'Head' is like 'None', except that the constant may also be 'TrueQ'. *)
getRuleConsts[ cat_, isRule_][ f_] :=
	getRuleConsts[ f, cat, isRule]
getRuleConsts[ FML$[ _, f_, __], cat_, isRule_] :=
	getRuleConsts[ f, cat, isRule]
getRuleConsts[ k:{_String, _String}, cat_, isRule_] :=
	Replace[ getFormulas[ False][ k],
		{
			FML$[ _, f_, __] :> getRuleConsts[ f, cat, isRule],
			_ -> {}
		}
	]
getRuleConsts[ THM$[ k_, args___], cat_, isRule_] :=
	Replace[ getTheorems[ False][ k][ args],
		{
			{f_, _List} :> getRuleConsts[ f, cat, isRule],
			_ -> {}
		}
	]
getRuleConsts[ (h:(TYPED$[ CoreAll$TM, _]))[ Lambda$TM[ _, f_]], cat_, isRule_] :=
	If[ cat === None || cat === Head,
		getRuleConsts[ h, cat, isRule],
	(*else*)
		getRuleConsts[ f, cat, isRule]
	]
getRuleConsts[ (h:(TYPED$[ CoreAll$TM, _]))[ f_], cat_, isRule_] :=
	If[ cat === None || cat === Head,
		getRuleConsts[ h, cat, isRule],
	(*else*)
		getRuleConsts[ f[ VAR$[ 0]], cat, isRule]	(* We do not care about de Bruijn indices here, so there is no need to adjust them in 'f'. *)
	]
getRuleConsts[ f:((h:(TYPED$[ CoreImplies$TM, _]))[ a_][ _]), cat_, isRule_] :=
	Which[
		cat === None || cat === Head,
		getRuleConsts[ h, cat, isRule],
	(*else*)
		cat === "elim" || cat === "trans",
		getRuleConsts[ a, cat, isRule],
	(*else*)
		True,
		getRuleConsts[ stripAssms[ f], cat, isRule]
	]
getRuleConsts[ (h:(TYPED$[ CoreImplies$TM, _]))[ a_], cat_, isRule_] :=
	Which[
		cat === None || cat === Head,
		getRuleConsts[ h, cat, isRule],
	(*else*)
		cat === "elim" || cat === "trans",
		getRuleConsts[ a, cat, isRule],
	(*else*)
		isRule,
		{VAR$},
	(*else*)
		True,
		{}
	]
getRuleConsts[ (h:(TYPED$[ CoreEqual$TM, _]))[ f_], cat_, isRule_] :=
	If[ cat === "simp",
		getRuleConsts[ NestWhile[ Last, f, MatchQ[ #, Lambda$TM[ _, _]]&], Head, isRule],
	(*else*)
		getRuleConsts[ h, cat, isRule]
	]
getRuleConsts[ (h:(TYPED$[ TrueQ$TM, _]))[ f_], cat_, isRule_] :=
	Which[
		cat === None,
		getRuleConsts[ f, Head, isRule],
	(*else*)
		cat === Head,
		getRuleConsts[ h, cat, isRule],
	(*else*)
		True,
		getRuleConsts[ f, cat, isRule]
	]
getRuleConsts[ TYPED$[ f_, _], cat_, isRule_] :=
	getRuleConsts[ f, cat, isRule]
getRuleConsts[ Lambda$TM[ _, f_], cat_, isRule_] :=
	If[ TrueQ[ isRule],
		{Lambda$TM},
	(*else*)
		True,
		{Key[ Lambda$TM]}
	]
getRuleConsts[ _VAR$|_META$, _, isRule_] :=
	If[ TrueQ[ isRule], {VAR$}, {}]
getRuleConsts[ c:(_FIX$|_Symbol|_Integer|_Rational|_Real|_Complex|_DirectedInfinity), _, isRule_] :=
	{If[ TrueQ[ isRule], c, Key[ c]]}
getRuleConsts[ h_[ _], cat_, isRule_] :=
	getRuleConsts[ h, cat, isRule]

groupRulesByRuleConsts[ rules_List, cat_, isRule_] :=
	GroupBy[ rules, getRuleConsts[ cat, isRule], Apply[ Hold, #]&]


(* ::Section:: *)
(*Proof Situations*)

allCount[ _, m_, _?NonPositive, _] :=
	m
allCount[ TYPED$[ CoreImplies$TM, _][ _][ R_], m_Integer, n_, True] :=
	allCount[ stripAssms[ R], m, n, True]
allCount[ TYPED$[ CoreAll$TM, _][ Lambda$TM[ _RNG$, R_]], m_Integer, n_, dropPrems_] :=
	allCount[ R, m + 1, n - 1, dropPrems]
allCount[ TYPED$[ CoreAll$TM, _][ _], m_Integer, _, _] :=
	m + 1
allCount[ _, m_Integer, _, _] := m

assmCountSkipAll[ _, m_, _?NonPositive, _] := m
assmCountSkipAll[ TYPED$[ CoreAll$TM, _][ Lambda$TM[ _RNG$, R_]], m_Integer, n_, expandLastAll_] :=
	assmCountSkipAll[ R, m, n, expandLastAll]
assmCountSkipAll[ TYPED$[ CoreAll$TM, _][ TYPED$[ CoreImplies$TM, _][ _]], m_Integer, _, expandLastAll_] :=
	If[ TrueQ[ expandLastAll], m + 1, m]
assmCountSkipAll[ TYPED$[ CoreImplies$TM, _][ _][ R_], m_Integer, n_, expandLastAll_] :=
	assmCountSkipAll[ R, m + 1, n - 1, expandLastAll]
assmCountSkipAll[ _, m_, _, _] := m

aaCount[ TYPED$[ CoreImplies$TM, _][ _][ R_], assms_Integer, all_Integer, maxAssms_?Positive, maxAll_] :=
	aaCount[ R, assms + 1, all, maxAssms - 1, maxAll]
aaCount[ TYPED$[ CoreAll$TM, _][ Lambda$TM[ _RNG$, R_]], assms_Integer, all_Integer, maxAssms_, maxAll_?Positive] :=
	aaCount[ R, assms, all + 1, maxAssms, maxAll - 1]
aaCount[ TYPED$[ CoreAll$TM, _][ R_], assms_Integer, all_Integer, maxAssms_, maxAll_?Positive] :=
	aaCount[ R[ VAR$[ 0]], assms, all + 1, maxAssms, maxAll - 1]	(* we don't care about de Bruijn indices here, so there is no need to adjust those in 'R' *)
aaCount[ _, assms_Integer, all_Integer, _, _] :=
	{assms, all}

(* 'extractAtomicGoal' is like 'aaCount', but also returns the remaining goal. 'aaCount' is faster since it does not
	adjust de Bruijn indices. *)
extractAtomicGoal[ TYPED$[ CoreImplies$TM, _][ _][ R_], assms_Integer, all_Integer, maxAssms_?Positive, maxAll_] :=
	extractAtomicGoal[ R, assms + 1, all, maxAssms - 1, maxAll]
extractAtomicGoal[ TYPED$[ CoreAll$TM, _][ Lambda$TM[ _RNG$, R_]], assms_Integer, all_Integer, maxAssms_, maxAll_?Positive] :=
	extractAtomicGoal[ R, assms, all + 1, maxAssms, maxAll - 1]
extractAtomicGoal[ TYPED$[ CoreAll$TM, _][ R_], assms_Integer, all_Integer, maxAssms_, maxAll_?Positive] :=
	extractAtomicGoal[ mapFreeVars[ R, VAR$[ #3 + 1]&][ VAR$[ 0]], assms, all + 1, maxAssms, maxAll - 1]
extractAtomicGoal[ G_, assms_Integer, all_Integer, _, _] :=
	{G, assms, all}

stripAll[ TYPED$[ CoreAll$TM, _][ Lambda$TM[ _RNG$, expr_]]] :=
	stripAll[ expr]
stripAll[ TYPED$[ CoreAll$TM, _][ expr_]] :=
	mapFreeVars[ expr, VAR$[ #3 + 1]&][ VAR$[ 0]]
stripAll[ expr_] :=
	expr
stripAll[ TYPED$[ CoreAll$TM, _][ Lambda$TM[ RNG$[ TYPRNG$[ _, t_]], expr_]], acc_] :=
	stripAll[ expr, Prepend[ acc, t]]
stripAll[ TYPED$[ CoreAll$TM, {t_}][ expr_], acc_] :=
	{mapFreeVars[ expr, VAR$[ #3 + 1]&][ VAR$[ 0]], Prepend[ acc, t]}
stripAll[ expr_, acc_] :=
	{expr, acc}

takeAssms[ G_, l_List, n_] :=
	Join[ l, reap[ takeAssms[ G, n]]]
takeAssms[ TYPED$[ CoreImplies$TM, {}][ A_][ R_], n_?Positive] :=
	(Sow[ A]; takeAssms[ R, n - 1])
takeAssms[ _, _] := Null

(* 'dropAssms' does not check whether there are sufficiently many assumptions to be dropped. *)
dropAssms[ G_, i_Integer?Positive] :=
	Extract[ G, Table[ 1, {i}]]
dropAssms[ G_, 0] :=
	G

fromAssmsGoal[ {}] :=
	fromAssmsGoal[ {}, TYPED$[ CoreEqual$TM, {FUN$TM[ FORM$TM, FUN$TM[ FORM$TM, FORM$TM]]}][ TYPED$[ CoreEqual$TM, {FORM$TM}]][ TYPED$[ CoreEqual$TM, {FORM$TM}]]]
fromAssmsGoal[ {A___, G_}] :=
	fromAssmsGoal[ {A}, G]
fromAssmsGoal[ A_, G_] :=
	Fold[ TYPED$[ CoreImplies$TM, {}][ #2][ #1]&, G, Reverse[ A]]

(* 'findPositions[ expr, terms, tfix, tvar, allowedIdx]' scans 'expr' and collects the positions of all subexpressions
	that are alpha-equivalent to one of the given terms. If type variables have to be instantiated in order to
	establish alpha-equivalence, the indices of all type variables that are being instantiated must be >= 'allowedIdx'.
	The function returns a pair '{{newTVar, inst}, res}', where 'inst' is the final type substitution and 'res' is the
	list of positions; each element of 'res' is of the form 'i -> pos', where 'i' is the index of the term in list
	'terms' and 'pos' is itself a list, consisting of elements of the form 'p -> abs', where 'p' is the actual position
	and 'abs' is the number of abstractions in whose scope 'p' appears. *)
findPositions[ _, {}, _, tvar_, _] :=
	{{tvar, {}}, {}}
findPositions[ expr_, terms_List, tfix_Association, tvar_, allowedIdx_] :=
	Block[ {$instantiateTypes = True},
		Reap[ Module[ {i}, findPositions[ expr, terms, 0, {}, tfix, {tvar, {}}, allowedIdx, Hold[ i]]], _Integer, Rule]
	]
findPositions[ _VAR$, _, _, _, _, acc_List, __] :=
	acc
findPositions[ expr_, terms_, abs_, pos_, tfix_Association, acc_List, allowedIdx_, Hold[ i_]] :=
	Catch[
		i = 0;
		Scan[
			With[ {a = (++i; alphaEquiv[ HoldComplete[ #], HoldComplete[ expr], tfix, acc])},
				If[ ListQ[ a] && (allowedIdx === 0 || !MemberQ[ Last[ a], (TVAR$[ j_Integer, ___] /; j < allowedIdx) -> _]),
					Sow[ pos -> abs, i];
					Throw[ a]
				]
			]&,
			terms
		];
		findPositions[ expr, terms, abs, pos, tfix, acc, allowedIdx, Hold[ i], False]
	]
findPositions[ _FIX$|_META$|_TYPED$, _, _, _, _, acc_List, _, _, False] :=
	acc
findPositions[ Lambda$TM[ _, expr_], terms_, abs_, pos_, tfix_Association, acc_List, allowedIdx_, i_Hold, False] :=
	findPositions[ expr, terms, abs + 1, Append[ pos, 2], tfix, acc, allowedIdx, i]
findPositions[ h_[ a_], terms_, abs_, pos_, tfix_Association, acc_List, allowedIdx_, i_Hold, False] :=
	findPositions[ a, terms, abs, Append[ pos, 1], tfix, findPositions[ h, terms, abs, Append[ pos, 0], tfix, acc, allowedIdx, i], allowedIdx, i]
findPositions[ _, _, _, _, _, acc_List, _, _, False] :=
	acc

(* 'findAtomPositions' is like 'findPositions', but only searches for atoms (i.e. 'FIX$', 'META$', 'TYPED$').
	No type substituation is computed, meaning that terms must agree literally. *)
findAtomPositions[ _, {}] :=
	{}
findAtomPositions[ expr_, atoms_List] :=
	Last[ Reap[ findAtomPositions[ expr, atoms, 0, {}], _Integer, Rule]]
findAtomPositions[ expr:(_FIX$|_META$|_TYPED$), atoms_, abs_, pos_] :=
	(
		Replace[ Position[ atoms, expr, {1}, 1, Heads -> False],
			{{i_}} :> Sow[ pos -> abs, i]
		];
	)
findAtomPositions[ Lambda$TM[ _, expr_], atoms_, abs_, pos_] :=
	findAtomPositions[ expr, atoms, abs + 1, Append[ pos, 2]]
findAtomPositions[ (h:Except[ VAR$])[ a_], atoms_, abs_, pos_] :=
	(
		findAtomPositions[ h, atoms, abs, Append[ pos, 0]];
		findAtomPositions[ a, atoms, abs, Append[ pos, 1]];
	)
findAtomPositions[ _, _, _, _] := Null

(* 'extractFormulas' returns either '$Failed' or a list of 'STERM$'-structures. *)
extractFormulas[ {}, idx_] :=
	{{}, idx}
extractFormulas[ fmls_List, idx0_Integer] :=
	Module[ {idx = idx0},
		Catch[
			{
				Map[
					Replace[ extractFormula[ #, idx],
						{
							{f_, i_Integer} :> (idx = i; f),
							_ :> Throw[ $Failed]
						}
					]&,
					fmls
				],
				idx
			}
		]
	]

getProofSitData[ PRFSIT$[ _, _, _, _, _, _, _, _, data_Association]] :=
	{data}
getProofSitData[ PRFSIT$[ _, _, _, _, _, _, _, _]] :=
	{Association[]}
getProofSitData[ PRFSIT$[ _, _, _, _, _, _, _, _, data_Association], k_, def_] :=
	Lookup[ data, k, def]
getProofSitData[ _, _, def_] :=
	def

setProofSitData[ PRFSIT$[ _, _, _, _, _, _, _, _, data_Association], new_, del_] :=
	{Append[ KeyDrop[ data, del], new]}
setProofSitData[ _PRFSIT$, new_List, _] :=
	{Association[ new]}
setProofSitData[ _PRFSIT$, new_Association, _] :=
	{new}
setProofSitData[ data_Association, new_, del_] :=
	{Append[ KeyDrop[ data, del], new]}

getTermAbbrevList[ ps_] :=
	Module[ {i},
		{
			Cases[ Normal[ getProofSitData[ ps, "termAbbrevs", {}]],
				(a_ -> {t_, tvp_List}) :>
					If[ tvp === {},
						a -> t,
					(*else*)
						With[ {repl = MapIndexed[ First[ #1] -> TVAR$[ i + First[ #2] - 1] &, tvp],
								tv = DeleteCases[ MapIndexed[ TVAR$[ i + First[ #2] - 1] -> Last[ #1] &, tvp], _ -> getDefaultSort[]],
								n = Length[ tvp]},
							a :> With[ {res = ReplacePart[ t, repl]}, i += n; STERM$[ res, Association[], Association @@ tv]]
						]
					]
			],
			Hold[ i]
		}
	]
getTermAbbrevList[ ps_, False] :=
	Cases[ Normal[ getProofSitData[ ps, "termAbbrevs", {}]],
		(a_ -> {t_, tvp_List}) :> (a -> ReplacePart[ t, MapIndexed[ First[ #1] -> TVAR$[] &, tvp]])
	]

(* ATTENTION! Like 'checkType' and 'inferTypes', 'makeCoreExpr' returns the association of *new* a.b.f. type constants,
	and the *new* association of type variables (including both new and old type variables). *)
makeCoreExpr[ term_, typ_, ps:PRFSIT$[ _, _, idx_, fix_, meta_, tfix_, tvar_, __], args__] :=
	makeCoreExpr[ term, typ, idx, fix, meta, tfix, tvar, getTermAbbrevList[ ps], args]
makeCoreExpr[ term0_EXPR$, typ_, idx0_Integer, fix_List, meta_List, tfix_Association, tvar_Association, abbrevs_, specgen_, post_, _] :=
	Replace[ liftTypeVars[ makeSTERM[ term0], idx0],
		{term1_STERM$, idx1_} :>
			Replace[ unfoldTermAbbrevs[ term1, abbrevs, idx1],
				{term2_, idx2_} :> parseCoreTermPost[ inferTypes[ term2, typ, idx2, idx0, {}, fix, meta, tfix, tvar, specgen], post]
			]
	]
makeCoreExpr[ term0_, typ0_, idx_Integer, fix_List, meta_List, tfix_Association, tvar_Association, _, _, _, param_] :=
	With[ {term = makeSTERM[ term0]},
		If[ param === Check || param === Replace,
			Replace[ checkType[ term, typ0, {}, fix, meta, tfix, tvar],
				{typ_, ts_List, m_, tf_Association, tv_Association} :>
					{{typ, ReplacePart[ If[ param === Replace, replaceTypes[ term, ts], term], {2 -> tf, 3 -> tv}], Max[ idx, m], {}, {}}, ts}
						(* note that type constants and -variables in 'term' are not updated *)
			],
		(*else*)
			With[ {termTyp = getType[ First[ term], {}, fix, meta, getAllConstDecls[ True]]},
				Which[
					termTyp === $Failed,
					$Failed,
				(*else*)
					typ0 === Blank[] || typ0 === termTyp,
					{{termTyp, ReplacePart[ term, 3 -> tvar], idx, {}, {}}, {}},
				(*else*)
					$instantiateTypes,
					Replace[ unifyTypes[ termTyp, typ0, hasSort[ tfix], {tvar, {}}, 0],
						{
							{tv_Association, ts_List} :> {{termTyp /. ts, ReplacePart[ replaceTypes[ term, ts], 3 -> tv], idx, {}, {}}, ts},
							_ -> $Failed
						}
					],
				(*else*)
					True,
					$Failed
				]
			]
		]
	]

makeCoreForms[ {}, args___] :=
	Replace[ makeCoreExpr[ STERM$[ CoreEqual$TM, Association[], Association[]], Blank[], args],
			(* We must call 'makeCoreExpr' to properly generalize/specialize type constants/variables. *)
		{{_, STERM$[ _, tfix_, tvar_], rest__}, ts_List} :> {{FORM$TM, STERM$[ {}, tfix, tvar], rest}, ts}
	]
makeCoreForms[ forms_List, args___] :=
	Replace[ makeCoreExpr[ If[ MemberQ[ forms, _EXPR$], EXPR$, STERM$] @@ makeSingleSTERM[ fromAssmsGoal[ Replace[ forms, EXPR$[ x___] :> STERM$[ x], {1}]]], FORM$TM, args],
		{{typ_, STERM$[ term_, tfix_, tvar_], rest__}, ts_List} :> {{typ, STERM$[ toAssmsGoal[ term, {}, Infinity], tfix, tvar], rest}, ts}
	]
makeCoreForms[ (h:(STERM$|EXPR$))[ {}, tfix0_Association, tvar0_Association], args___] :=
	Replace[ makeCoreExpr[ h[ CoreEqual$TM, tfix0, tvar0], Blank[], args],
			(* We must call 'makeCoreExpr' to properly generalize/specialize type constants/variables. *)
		{{_, STERM$[ _, tfix_, tvar_], rest__}, ts_List} :> {{FORM$TM, STERM$[ {}, tfix, tvar], rest}, ts}
	]
makeCoreForms[ (h:(STERM$|EXPR$))[ forms_List, tfix0_Association, tvar0_Association], args___] :=
	Replace[
		makeCoreExpr[
			If[ h === EXPR$ || MemberQ[ forms, _EXPR$], EXPR$, STERM$] @@
				makeSingleSTERM[ STERM$[ fromAssmsGoal[ Replace[ forms, EXPR$[ x___] :> STERM$[ x], {1}]], tfix0, tvar0]],
			FORM$TM,
			args
		],
		{{typ_, STERM$[ term_, tfix_, tvar_], rest__}, ts_List} :> {{typ, STERM$[ toAssmsGoal[ term, {}, Infinity], tfix, tvar], rest}, ts}
	]
makeCoreForms[ form_, args___] :=
	makeCoreExpr[ form, FORM$TM, args]

(* 'makeCorePattern[ patts, typ, idx, fix, meta, tfix, tvar, abbrevs, specgen, post, master]' returns a 9-tuple
	'{newPatts, newTyp, newIdx, newFix, newMeta, newTFix, newTVar, tsubst, {midx, Hold[ i]}}', where
	- 'newPatts' is the list of new patterns,
	- 'newTyp' is the type of the patterns (they must all have the same type),
	- 'newIdx' is the new type variable index,
	- 'newFix', 'newMeta', 'newTFix' and 'newTVar' are the lists/associations of new a.b.f. term constants, meta
		variables, a.b.f. type constants, and type variables, respectively,
	- 'tsubst' is the type substitution which needs to be applied in order to make 'newPatts' type-check,
	- 'midx' is the common index of the new meta variables originating from named syntactic variables, and
	- 'i' is the current index of the anonymous syntactic variables. *)
makeCorePattern[ expr_, typ_, ps:PRFSIT$[ _, _, idx_, fix_, meta_, tfix_Association, tvar_Association, __], args__] :=
	makeCorePattern[ expr, typ, idx, fix, meta, tfix, tvar, getTermAbbrevList[ ps], args]
makeCorePattern[ patt0_EXPR$, typ0_, idx0_Integer, fix_List, meta_List, tfix_Association, tvar_Association, abbrevs_, specgen_, post_, param0_] :=
	Replace[ liftTypeVars[ makeSTERM[ patt0], idx0],
		{patt1_STERM$, idx1_} :>
			Replace[ unfoldTermAbbrevs[ patt1, abbrevs, idx1],
				{patt2_, idx2_} :>
					Replace[ replaceSyntacticByMeta[ patt2, param0],
						{patt_, param_List} :>
							Replace[ parseCoreTermPost[ inferTypes[ patt, typ0, idx2, idx0, {}, fix, meta, tfix, tvar, specgen], post],
								If[ ListQ[ param0],
									{
										{{typ_, STERM$[ p_, tf_, tv_], i_, fx_, m_}, {}} :> {{p}, typ, i, fx, m, tf, tv, {}, param},
										_ :> (sowMessage[ 201, {}, "Warning"]; {{}, typ0, idx0, {}, {}, {}, {}, tvar, {}, param})
									},
								(*else*)
									{{typ_, STERM$[ p_, tf_, tv_], i_, fx_, m_}, ts_} :> {{p}, typ, i, fx, m, tf, tv, ts, param}
								]
							]
					]
			]
	]
makeCorePattern[ patts:{___EXPR$}, typ0_, idx0_Integer, fix0_, meta0_, tfix0_Association, tvar0_Association, abbrevs_, specgen_, post_, param0_] :=
	Module[ {typ = typ0, idx = idx0, fix = fix0, meta = meta0, tfix = tfix0, tvar = tvar0, subst = {}, param = param0},
		Catch[
			With[ {out =
					Join @@ Map[
						Function[ P,
							Replace[ makeCorePattern[ P, typ, idx, fix, meta, tfix, tvar, abbrevs, specgen, post, param],
								{
									{p_List, t_, i_, fx_, m_, tf_Association, tv_Association, ts_List, pm_} :>
										(
											param = pm;
											typ = t;
											idx = i;
											fix = Join[ fix /. ts, fx];
											meta = Join[ meta /. ts, m];
											tfix = Join[ tfix, tf];		(* keys of 'tfix' and 'tf' are disjoint *)
											tvar = tv;
											subst = Join[ subst /. ts, ts];
											p
										),
									_ :> Throw[ $Failed]
								}
							]
						],
						patts
					]},
				{
					replaceTypes[ out, subst],
					typ,
					idx,
					Drop[ fix, Length[ fix0]],
					Drop[ meta, Length[ meta0]],
					KeyDrop[ tfix, Keys[ tfix0]],
					tvar,
					subst,
					param
				}
			]
		]
	]
makeCorePattern[ patt_, typ_, idx_Integer, fix_List, meta_List, _, tvar_Association, _, specgen_, _, param_] :=
	{
		If[ patt === "_", {}, {patt}],
		typ,
		idx,
		{},
		{},
		Association[],
		tvar,
		{},
		If[ ListQ[ param],
			param,
		(*else*)
			{Max[ If[ IntegerQ[ param], param, 0], Max @@ Cases[ patt, META$[ _, i_Integer] :> i, {0, Infinity}, Heads -> True] + 1], Module[ {i = 0}, Hold[ i]]}
		]
	}

(* Remarks on 'makeCoreExprPatt':
	- In principle it is possible to instantiate existing meta variables using this function, but this has not been
		tested yet. In fact, there might be some problems if type variables in the master term shall be specialized,
		because instantiating meta variables might introduce hidden polymorphism then.
		BETTER NOT INSTANTIATE META VARIABLES THROUGH TERM ABBREVIATIONS!
	- Both the master term and the patterns are beta reduced, since unification requires its input to be in beta
		normal form. This is independen of the value of 'post'; 'post' is applied to the master term at the very end,
		after instantiating meta variables; the instances of the syntactic variables are automatically normalized in
		any case (happens in 'unfoldTriangularSubst'). *)
makeCoreExprPatt[ expr_, typ_, ps:PRFSIT$[ _, _, idx_, fix_, meta_, tfix_, tvar_, __], args__] :=
	makeCoreExprPatt[ expr, typ, idx, fix, meta, tfix, tvar, getTermAbbrevList[ ps], args]
makeCoreExprPatt[ PEXPR$[ term0_, patts0__], typ_, idx_Integer, fix_List, meta_List, tfix_Association, tvar_Association, abbrevs_, specgen_, post_, param_] :=
	Replace[ makeCoreExpr[ term0, typ, idx, fix, meta, tfix, tvar, abbrevs, specgen, post, param],
		{{typ1_, STERM$[ term_, tfix1_Association, tvar1_Association], idx1_, fix1_, meta1_}, ts_} :>
			With[ {fix2 = fix /. ts, tfix2 = Join[ tfix, tfix1]},	(* keys of 'tfix' and 'tfix1' are disjoint *)
				Replace[
					makeCorePattern[
						replaceTypes[ {patts0}, ts],
						typ1, idx1, Join[ fix2, fix1], {}, tfix2, tvar1, abbrevs, "gen", betaReduce,
						{Max[ 0, Max @@ Join[ meta[[All, 1, 2]], meta1[[All, 1, 2]]]], Module[ {i = 0}, Hold[ i]]}
					],
					{
						{patts_List, _, idx2_, {}, meta2_, Association[], tvar2_Association, {}, {midx_, _}} :>
							With[ {termRed = If[ post === betaReduce || (post === normalize && TrueQ[ $autoNormalize]), term, betaReduce[ term]]},
								Replace[
									unifierToAbbrevs[
										unifyFirst[ Map[ {termRed, #}&, patts], Join[ meta /. ts, meta1, meta2], Join[ fix2, fix1], tfix1, tvar2, Automatic, "AllowFlexFlex" -> False],
										Alternatives @@ meta[[All, 1]],
										Alternatives @@ meta1[[All, 1]],
										midx
									],
									{
										{mInst_, m1Inst_, newAbbrevs_, rem_, tinst_, tvar3_Association} :>
											{
												{typ1, STERM$[ replaceTypes[ term, tinst] /. m1Inst, tfix1, tvar3], idx2, fix1 /. tinst, rem},
												Join[ ts /. tinst, tinst],
												mInst,
												newAbbrevs
											},
										_ :> (sowMessage[ 267, {}, "Warning"]; {{typ1, STERM$[ term, tfix1, tvar2], idx1, fix1, meta1}, ts, {}, {}})
									}
								]
							],
						_ :> (sowMessage[ 268, {}, "Warning"]; {{typ1, STERM$[ term, tfix1, tvar1], idx1, fix1, meta1}, ts, {}, {}})
					}
				]
			]
	]
makeCoreExprPatt[ args__] :=
	Replace[ makeCoreExpr[ args], out_List :> Join[ out, {{}, {}}]]

(* 'unifierToAbbrevs[ unif, mPatt, m1Patt, midx]' takes a unifier 'unif', two meta-variable patterns 'mPatt' and
	'm1Patt', and a meta-variable index 'midx', and returns either '$Failed' if the given unifier is no 4-tuple with an
	empty list of flex-flex equations, or a 6-tuple '{mInst, m1Inst, abbrevs, rem, tinst, tvar}', where:
	- 'mInst' and 'm1Inst' are the lists of instantiations of the meta variables matching 'mPatt' and 'm1Patt',
		respectively; these instantiations have been unfolded already!
	- 'abbrevs' is the list of instantiations of the meta variables whose index equals 'midx', where the instances do
		not contain *any* meta variables, and where the meta variables on the left-hand-sides are replaced by the
		corresponding syntactic variables. The RHSs of the transformation rules are 'STERM$'-structures.
	- 'rem' is the list of remaining meta variables that really appear on the right-hand-sides in any of the
		aforementioned instantiations and which do not match 'mPatt'.
	- 'tinst' is the type instantiation of the unifier.
	- 'tvar' are the sort constraints on type variables from 'unif'. *)
unifierToAbbrevs[ {{}, rem0_List, inst0_List, tinst_List, tvar_Association}, mPatt_, m1Patt_, midx_Integer] :=
	(
		With[ {uninst = Cases[ rem0, ((x:META$[ _, midx]) -> _) :> x]},
			If[ uninst =!= {},
				sowMessage[ 269, uninst, "Warning"]
			]
		];
		With[ {inst = unfoldTriangularSubst[ inst0, META$[ _, midx]|mPatt|m1Patt]},
		With[ {smPos = Position[ inst, HoldPattern[ META$[ _, midx] -> _], {1}, Heads -> False]},
		With[ {abbrevs =
				Cases[
					Extract[ inst, smPos],
					((x:META$[ y_, _]) -> t_) /;
						With[ {val = FreeQ[ t, META$]},
							If[ !val, sowMessage[ 270, {x}, "Warning"]];
							val
						] :> (SVAR$[ y] -> STERM$[ t, Association[], tvar])
				],
				remInst = Delete[ inst, smPos]},
		With[ {remPatt = Alternatives @@ DeleteDuplicates[ Cases[ remInst[[All, 2]], Except[ mPatt, _META$], {0, Infinity}, Heads -> True]],
				mPos = Position[ remInst, HoldPattern[ mPatt -> _], {1}, Heads -> False]},
			{
				Extract[ remInst, mPos],
				Delete[ remInst, mPos],
				abbrevs,
				If[ remPatt === Alternatives[], {}, Cases[ rem0, HoldPattern[ remPatt -> _]]],
				tinst,
				tvar
			}
		]]]]
	)
unifierToAbbrevs[ ___] := $Failed

unfoldTermAbbrevs[ term_, {abbrevs_List, Hold[ i_]}, idx_] :=
	{i = idx; unfoldTermAbbrevs[ term, abbrevs], i}
unfoldTermAbbrevs[ term_, abbrevs_List, idx_] :=
	{unfoldTermAbbrevs[ term, abbrevs], idx}
unfoldTermAbbrevs[ term_, {}] :=
	makeSTERM[ term]
unfoldTermAbbrevs[ STERM$[ term0_, tfix_, tvar_], abbrevs_List] :=
	Replace[ unfoldTermAbbrevs[ term0, abbrevs],
		STERM$[ term_, tf_, tv_] :> STERM$[ term, intersectSorts[ tfix, tf], intersectSorts[ tvar, tv]]
	]
unfoldTermAbbrevs[ term0_, abbrevs_List] :=
	makeSingleSTERM[ term0 /. abbrevs]

replaceSyntacticByMeta[ STERM$[ patt_, pattTFix_Association, pattTVar_Association], param_] :=
	Replace[ replaceSyntacticByMeta[ patt, param],
		{newPatt_, newParam_} :> {STERM$[ newPatt, pattTFix, pattTVar], newParam}
	]
replaceSyntacticByMeta[ patt_, param:{midx_Integer, Hold[ i_Symbol]}] :=
	If[ FreeQ[ patt, META$],
		{patt /. {SVAR$[ x_Symbol] :> META$[ x, midx], SVAR$[] :> META$[ Theorema`Knowledge`VAR$X$TM, midx + (++i)]}, param},
	(*else*)
		$Failed
	]
replaceSyntacticByMeta[ patt_, param_] :=
	With[ {max = Max[ If[ IntegerQ[ param], param, 0], Max @@ Cases[ patt, META$[ _, i_Integer] :> i, {0, Infinity}, Heads -> True] + 1]},
	Module[ {i = 0},
		{patt /. {SVAR$[ x_Symbol] :> META$[ x, max], SVAR$[] :> META$[ Theorema`Knowledge`VAR$X$TM, max + (++i)]}, {max, Hold[ i]}}
	]]
replaceSyntacticByMeta[ patt_] :=
	replaceSyntacticByMeta[ patt, 0]

makeSTERM[ EXPR$[ e_]] :=
	makeSTERM[ e]
makeSTERM[ EXPR$[ e_, tf_Association]] :=
	STERM$[ e, tf, Association[]]
makeSTERM[ e:EXPR$[ _, _Association, _Association]] :=
	STERM$ @@ e
makeSTERM[ e:STERM$[ _, _Association, _Association]] :=
	e
makeSTERM[ e_] :=
	STERM$[ e, Association[], Association[]]

makeSingleSTERM[ term_] :=
	With[ {stermPos = Position[ term, STERM$[ _, _Association, _Association]]},
		If[ stermPos === {},
			STERM$[ term, Association[], Association[]],
		(*else*)
			With[ {sterms = Extract[ term, stermPos]},
				STERM$[ MapAt[ First, term, stermPos], intersectSorts @@ sterms[[All, 2]], intersectSorts @@ sterms[[All, 3]]]
			]
		]
	]


(* ::Section:: *)
(*Proving*)

parseAddRemOnlySpec[ spec_List, suffixes_List, src_String] :=
	Catch[
		Module[ {major = 1, minor = 0, majorN = 2, minorN = 0},
		With[ {cases =
				Join[
					Hold[ Placeholder,
						":",
						If[ Positive[ majorN],
							major = majorN;
							minor = minorN;
							majorN = 0,
						(*else*)
							sowCPMMessage[ 1013, "Fatal", src];
							Throw[ $Failed]
						]
					],
					Join @@ MapIndexed[
						With[ {i = First[ #2] + 1, only = "only" <> #1, add = "add" <> #1, rem = "rem" <> #1},
							Hold[ only, majorN = i; minorN = 0, add, majorN = i; minorN = 1, rem, majorN = i; minorN = 2]
						]&,
						suffixes
					],
					Hold[
						_List,
						If[ major === 1,
							Sow[ Placeholder, 1],
						(*else*)
							Sow[ minor -> DeleteCases[ Placeholder, ","], major];
							majorN = 0
						],
						_,
						If[ major === 1,
							Sow[ Placeholder, 1],
						(*else*)
							Sow[ minor -> {Placeholder}, major];
							majorN = 0
						]
					]
				]},
		With[ {fun = (Switch @@@ Function[ cases]) /. Placeholder -> Slot[ 1]},
			Replace[ Last[ Reap[ Scan[ fun, spec], Range[ Length[ suffixes] + 1]]],
				{gen_List, facts___List} :> Prepend[ makeAddRemOnlyList[ src] /@ {facts}, gen]
			]
		]]]
	]

makeAddRemOnlyList[ src_][ spec_List] :=
	With[ {aux = Replace[ SplitBy[ Join @@ spec, First], all:{op_ -> _, ___} :> (op -> Join @@ all[[All, 2]]), {1}]},
	With[ {lastOnly = First[ FirstPosition[ Reverse[ aux], 0 -> _, {Null}, {1}, Heads -> False]]},
		If[ IntegerQ[ lastOnly] && lastOnly < Length[ aux],
			sowCPMMessage[ 1016, "Warning", src];
			Take[ aux, -lastOnly],
		(*else*)
			aux
		]
	]]

(* 'succ' is supposed to be a function that takes 2 arguments 'spec' and 'fml'. 'spec' is a list of pairs of elements
	whose first component can be
	- an integer, referring to an implicit hypothesis,
	- a key, referring to a formula in the global theory context,
	- a theorem specifier, or
	- a variant specifier.
	The lists group all elements together that originate from the same single fact specification by means of 'ruleCat'
	and/or because a single label refers to several formulas. Elements that originate from other single fact
	specifications (e.g. fact collections) are *not* grouped together!
	'err' is the error function. It gets as arguments the specifier that caused the error and a flag indicating the
	kind of the error: '0' means that the fact was not found, '1' means that the fact comes after the current proof
	goal, and '2' means that the literal fact is ill-typed. *)
scanFactSpec[ l_List, ps_PRFSIT$, succ_, err_] :=
	scanFactSpec[ l, ps, None, succ, err]
scanFactSpec[ l_List, ps_PRFSIT$, ruleCat_, succ_, err_] :=
	scanFactSpec[ l, ps, ruleCat, False, succ, err]
scanFactSpec[ l0_List, ps:PRFSIT$[ _, hyps_, __], ruleCat_, includeOrig_, succ_, err_] :=
	Module[ {l = l0},
		Block[ {$scanFactSpecHypNum = Length[ hyps] + 1,
				$scanFactSpecHyps = hyps,
				$scanFactSpecFactLists = getProofSitData[ ps, "factLists", Association[]],
				$scanFactSpecCalcList = Replace[ getProofSitData[ ps, "calc", $Failed], {ky:{_, _}, _} :> ky, {1}],
				$scanFactSpecProofSit = ps,
				$scanFactSpecRuleCat = ruleCat,
				$scanFactSpecIncludeOrig = includeOrig,
				$scanFactSpecSuccFun = succ,
				$scanFactSpecErrFun = err},
			While[ l =!= {},
				l = Join[ singleFactSpec[ First[ l]], Rest[ l]]
			]
		]
	]

(* This is pretty much the same as 'addSingleAssm' ... *)
singleFactSpec[ i:Except[ 0, _Integer]] :=
	With[ {j = If[ Negative[ i], $scanFactSpecHypNum + i, i]},
		If[ 1 <= j < $scanFactSpecHypNum,
			$scanFactSpecSuccFun[ {{j, $scanFactSpecHyps[[j]]}}],
		(*else*)
			$scanFactSpecErrFun[ i, 0]
		];
		{}
	]
singleFactSpec[ k:{_String?(StringMatchQ[ #, "ID" ~~ __] &), _String?(StringMatchQ[ #, "Source" ~~ __] &)}] :=
	With[ {p = First[ FirstPosition[ $scanFactSpecHyps, _[ k, __], {$Failed}, {1}, Heads -> False]]},
		If[ p === $Failed,
			With[ {fml = getFormulas[ True][ k]},
				If[ fml === $Failed,
					$scanFactSpecErrFun[ k, 0],
				(*else*)
					If[ inScopeQ[ fml],
						$scanFactSpecSuccFun[ {{k, fml}}],
					(*else*)
						$scanFactSpecErrFun[ k, 1]
					]
				]
			],
		(*else*)
			$scanFactSpecSuccFun[ {{p, $scanFactSpecHyps[[p]]}}]
		];
		{}
	]
singleFactSpec[ STRING$[ l_String]] :=
	With[ {goalSrc = Quiet[ Check[ $proofGoal[[1, 2]], None]]},
	With[ {ky = getFactListKeys[ l, 1, goalSrc]},
		Replace[ If[ ky === {}, $Failed, Lookup[ $globalFactLists, Key @@ ky, $Failed]],
			{
				facts_List :> (scanFactSpec[ Cases[ facts, {f_, _List?inScopeTSQ} :> f]]; {}),
				_ :>
					With[ {p = Reverse[ formulaPosition[ l, Infinity, None, $scanFactSpecHyps]]},
					With[ {fun =
							Function[ fml,
								With[ {k = key@fml},
								With[ {q = First[ FirstPosition[ $scanFactSpecHyps, _[ k, __], {$Failed}, {1}, Heads -> False]]},
									If[ q === $Failed,
										{k, fml},
									(*else*)
										{q, $scanFactSpecHyps[[q]]}
									]
								]]
							]},
						If[ p === {},
							With[ {fmls0 = Reverse[ selectFormulas[ l, 1, goalSrc]]},
								If[ fmls0 === {},
									singleFactSpec[ THM$[ l]],
								(*else*)
									With[ {fmls =
											Join @@ Map[
												With[ {k = key@#},
													Replace[ getRuleFormulas[ $scanFactSpecRuleCat, #, $scanFactSpecIncludeOrig, True, True, True],
														{
															{FML$[ k, __], rules___} :> Prepend[ fun /@ {rules}, {k, #}],
															rules_List :> fun /@ rules
														}
													]
												]&,
												fmls0
											]},
										If[ fmls === {},
											$scanFactSpecErrFun[ STRING$[ l], 1],
										(*else*)
											$scanFactSpecSuccFun[ fmls]
										]
									]
								]
							],
						(*else*)
							$scanFactSpecSuccFun[
								(* the list below cannot be empty *)
								Join @@ Map[
									With[ {hyp = $scanFactSpecHyps[[First[ #]]]},
										Replace[ getRuleFormulas[ $scanFactSpecRuleCat, hyp, $scanFactSpecIncludeOrig, False, True, True],
											{
												{FML$[ key@hyp, __], rules___} :> Prepend[ fun /@ {rules}, {First[ #], hyp}],
												rules_List :> fun /@ rules
											}
										]
									]&,
									p
								]
							]
						];
						{}
					]]
			}
		]
	]]
singleFactSpec[ e0_EXPR$] :=
	Replace[ makeCoreExpr[ e0, FORM$TM, $scanFactSpecProofSit, None, normalize, None],	(* This is necessary for unfolding (global) abbreviations. *)
		{
			{{_, STERM$[ e_, _, _], _, _, _}, _} :>
				With[ {patt = termToTermPatt[ e]},	(* We completely ignore type- and sort information when retrieving literal facts. *)
				With[ {p = First[ FirstPosition[ $scanFactSpecHyps, _[ _, patt, ___], {$Failed}, {1}, Heads -> False]]},
					If[ p === $Failed,
						$scanFactSpecErrFun[ e0, 0],
					(*else*)
						$scanFactSpecSuccFun[ {{p, $scanFactSpecHyps[[p]]}}]
					];
					{}
				]],
			_ :> ($scanFactSpecErrFun[ e0, 2]; {})
		}
	]
singleFactSpec[ thm0_THM$] :=
	(
		Replace[ fastThmFromLabel[ thm0],
			{
				thm:THM$[ k_, args___] :>
					Replace[ getTheorems[ True][ k][ args],
						{
							l_List :> $scanFactSpecSuccFun[ {{thm, l}}],
							_ :> $scanFactSpecErrFun[ thm0, 0]
						}
					],
				_ :> $scanFactSpecErrFun[ thm0, 0]
			}
		];
		{}
	)
singleFactSpec[ coll_String] :=
	If[ coll === "calculation",
		Replace[ $scanFactSpecCalcList,
			{
				{a_} :> singleFactSpec[ a],
				l_List :> l,
				_ :> ($scanFactSpecErrFun[ coll, 0]; {})
			}
		],
	(*else*)
		Replace[ Lookup[ $scanFactSpecFactLists, coll, $Failed],
			{
				{a_} :> singleFactSpec[ a],
				l_List :> l,
				_ :> ($scanFactSpecErrFun[ coll, 0]; {})
			}
		]
	]
singleFactSpec[ LABELED$[ a_, ___]] :=
	singleFactSpec[ a]
singleFactSpec[ v_VARIANT$] :=
	($scanFactSpecSuccFun[ {{v, Null}}]; {})
singleFactSpec[ {a_}] :=
	singleFactSpec[ a]
singleFactSpec[ a_List] :=
	a
singleFactSpec[ a_] :=
	($scanFactSpecErrFun[ a, 0]; {})

tryTerminalProofSteps[ hyps_, idx_, fix_, meta_, tfix_, tvar_][ G_] :=
	tryTerminalProofSteps[ G, hyps, idx, fix, meta, tfix, tvar]
tryTerminalProofSteps[ G_List, hyps_, idx_, fix_, meta_, tfix_, tvar_] :=
	Catch[
		Map[
			With[ {res = tryTerminalProofSteps[ #, hyps, idx, fix, meta, tfix, tvar]},
				If[ res === $Failed, Throw[ res], res]
			]&,
			G
		]
	]
tryTerminalProofSteps[ G_, hyps_, idx_, fix_, meta_, tfix_, tvar_] :=
	Catch[
		Scan[
			With[ {res = #[ G, hyps, idx, fix, meta, tfix, tvar]},
				If[ res =!= $Failed, Throw[ res]]
			]&,
			$terminalProofSteps
		];
		$Failed
	]

(* 'mapAllInline[ cmd, sub, n]' is an auxiliary function that "copies" a given sequence of proof commands and inserts
	additional commands, such that finally the given commands are applied to the first 'sub' subgoals of the current
	proof situation. Of course, this is only correct if the given commands neither open new subgoals, nor close
	subgoals, nor rearrange subgoals.
	'n' must be the total number of subgoals (but it may be completely arbitrary if 'sub' is a non-negative integer);
	'sub' may be 'All' or "all". *)
mapAllInline[ Hold[], _, _] :=
	Hold[]
mapAllInline[ cmd_Hold, All|"all", n_Integer] :=
	mapAllInline[ cmd, n, n]
mapAllInline[ _Hold, 0, _] :=
	Hold[]
mapAllInline[ cmd_Hold, 1, _] :=
	cmd
mapAllInline[ cmd0_Hold, sub_Integer, _] :=
	With[ {cmd = Append[ cmd0, INF$[ "rearrangeG", 1, sub]]},
		Join @@ Table[ cmd, {sub}]
	]

(* 'introFixCommandList[ G, fix, all]' constructs a sequence of proof commands that introduce fresh a.b.f. constants
	for the Core universal quantifiers in 'G'. The names of these constants stem from the names of the bound variables.
	Parameter 'all' specifies whether all universal quantifiers shall be introduced, or whether only the outermost
	prefix of quantifiers shall be introduced.
	The function returns a pair consisting of the new number of explicit assumptions (after introducing the constants),
	and the sequence of proof commands. *)
introFixCommandList[ G_, fix_, all_] :=
	Replace[ Reap[ introFixCommandListAux[ G, 0, all]],
		{ac_Integer, vars0_List} :>
			{
				ac,
				With[ {vars = Join @@ vars0},
					If[ vars === {},
						Hold[],
					(*else*)
						With[ {max = Max[ -1, Max @@ fix[[All, 1, 2]]]},
							Hold @@ MapIndexed[ INF$[ "allI", FIX$[ #1, max + First[ #2]]]&, vars]
						]
					]
				]
			}
	]

introFixCommandListAux[ TYPED$[ CoreAll$TM, {_}][ Lambda$TM[ RNG$[ TYPRNG$[ v_Symbol, _]], A_]], n_, all_] :=
	(Sow[ v]; introFixCommandListAux[ A, n, all])
introFixCommandListAux[ TYPED$[ CoreImplies$TM, {}][ _][ G_], n_, True] :=
	introFixCommandListAux[ G, n + 1, True]
introFixCommandListAux[ G_, n_, _] := assmCount[ G, n, Infinity]

(* 'makeFactsExplicitInline[ l, sub, cmd, perm]' is an auxiliary function that constructs a sequence of proof commands
	for making the known facts given by list 'l' explicit in the first 'sub' subgoals of a proof situation.
	- 'l' is a list whose elements may be
		- keys, referring to formulas in the global theory context,
		- strings inside 'STRING$', referring to either theorems or labels of formulas in the global environment,
		- plain strings referring to global fact collections; ATTENTION! This differs from 'addKnowledgeImplicit' etc.!
		- instances of theorem schemas inside 'THM$'.
	- 'sub' is an integer specifying the number of subgoals the facts shall be made explicit in; it cannot be 'All'!
	- 'cmd' is an arbitrary sequence of proof commands that are appended to the generated sequence in case of success.
	- 'perm' is a boolean value specifying whether the facts shall be added permanently to the implicit hypotheses, or
		whether they shall immediately be removed again *after they have been made explicit* (but still before applying
		'cmd').
	Note that the function returns '$Failed' if at least one fact is not found. Otherwise, the order of facts in the
	subgoals is precisely the same as their order in 'l'.
	Also note that formulas are (temporarily) added to the hypotheses even if they are there already, since the
	function does not know which hypotheses are already available. Hence, 'perm' should usually be 'False'.
	In contrast to proof method 'addKnowledgeExplicit', this function is well-suited for automatic proof methods that
	depend on some facts of the background theory. *)
makeFactsExplicitInline[ l_List, sub_Integer, cmd_Hold, perm_] :=
	If[ l === {} || NonPositive[ sub],
		cmd,
	(*else*)
		Catch[
			Join[ Hold @@ Join @@ Flatten[ Last[ Reap[ makeFactsExplicitAux[ l, sub, perm]]]], cmd]
		]
	]

makeFactsExplicitAux[ {}, _, _] := Null
makeFactsExplicitAux[ {pre___, k:{_String?(StringMatchQ[ #, "ID" ~~ __] &), _String?(StringMatchQ[ #, "Source" ~~ __] &)}}, sub_Integer, perm_] :=
	Replace[ getFormulas[ True][ k],
		{
			f_FML$ :>
				If[ inScopeQ[ f],
					Sow[
						If[ TrueQ[ perm],
							Join[ Hold[ INF$[ "insert", k]], mapAllInline[ Hold[ INF$[ "makeExplicit", 1]], sub, sub]],
						(*else*)
							mapAllInline[ Hold[ INF$[ "formula", k]], sub, sub]
						]
					];
					makeFactsExplicitAux[ {pre}, sub, perm],
				(*else*)
					Throw[ $Failed]
				],
			_ :> Throw[ $Failed]
		}
	]
makeFactsExplicitAux[ {pre___, STRING$[ src_String, lbl_String]}, sub_Integer, perm_] :=
	With[ {k = fastKeyFromLabel[ {src, lbl}]},
		If[ ListQ[ k] && inScopeQ[ k],
			Sow[
				If[ TrueQ[ perm],
					Join[ Hold[ INF$[ "insert", k]], mapAllInline[ Hold[ INF$[ "makeExplicit", 1]], sub, sub]],
				(*else*)
					mapAllInline[ Hold[ INF$[ "formula", k]], sub, sub]
				]
			];
			makeFactsExplicitAux[ {pre}, sub, perm],
		(*else*)
			Throw[ $Failed]
		]
	]
makeFactsExplicitAux[ {pre___, coll_String}, sub_Integer, perm_] :=
	With[ {ky = getFactListKeys[ coll, 1, Quiet[ Check[ $proofGoal[[1, 2]], None]]]},
		Replace[ If[ ky === {}, $Failed, Lookup[ $globalFactLists, Key @@ ky, $Failed]],
			{
				facts_List :>
					(
						makeFactsExplicitAux[ Reverse[ Cases[ facts, {f_, _List?inScopeTSQ} :> f]], sub, perm];
						makeFactsExplicitAux[ {pre}, sub, perm]
					),
				_ :> Throw[ $Failed]
			}
		]
	]
makeFactsExplicitAux[ {pre___, thm0_THM$}, sub_Integer, perm_] :=
	Replace[ fastThmFromLabel[ thm0],
		{
			thm:THM$[ k_, ___] :>
				If[ getTheorems[ True][ k] === $Failed,
					Throw[ $Failed],
				(*else*)
					Sow[ mapAllInline[ Hold[ INF$[ "theorem", thm]], sub, sub]];
					makeFactsExplicitAux[ {pre}, sub, perm]
				],
			_ :> Throw[ $Failed]
		}
	]
makeFactsExplicitAux[ {__}, _, _] :=
	Throw[ $Failed]

(* 'fastKeyFromLabel' accepts all arguments that 'formulaKeys' accepts. *)
fastKeyFromLabel[ l_] :=
	Lookup[
		$labelKeyMapping,
		Key[ l],
		With[ {k = formulaKeys[ l, 1]},
			If[ k === {},
				$Failed,
			(*else*)
				AssociateTo[ $labelKeyMapping, l -> First[ k]];
				First[ k]
			]
		]
	]

fastThmFromLabel[ THM$[ l_String, args___]] :=
	With[ {k = fastThmFromLabel[ l]},
		If[ k === $Failed,
			$Failed,
		(*else*)
			THM$[ k, args]
		]
	]
fastThmFromLabel[ thm_THM$] := thm
fastThmFromLabel[ l_] :=
	Lookup[
		$labelThmMapping,
		Key[ l],
		With[ {k = formulaKeys[ l, 1, None, getTheorems[ False]]},
			If[ k === {},
				$Failed,
			(*else*)
				AssociateTo[ $labelThmMapping, l -> First[ k]];
				First[ k]
			]
		]
	]

(* 'makeInstantiation[ inst, term, head, cond, stripImpl, ps, mode]' attempts to turn the term-list 'inst' into a list
	of properly type-annotated terms whose types agree with the types of the outermost binders with head 'head' in
	'term'. 'stripImpl' specifies whether Core implication shall be stripped off, and 'ps' and 'mode' are the current
	proof situation and proof mode, respectively. The elements of 'inst' may be
	- 'Automatic' or "$", in which case a fresh meta variable is generated, or
	- "_", in which case the outermost binder is skipped and the instantiation resumes with the next binder, or
	- an arbitrary term, possibly inside 'EXPR$'; if it is inside 'EXPR$', types are inferred, otherwise it is simply
		type-checked.
	The last element of 'inst' may be one of the above possibilities wrapped inside 'RepeatedNull', in which case
	the last element is replicated as many times as there are outermost 'head' binders in the part of 'term' that has
	not been consumed yet.
	Binder 'head' can bind only one variable, i.e. its type must be of the form '[_ -> ](a_ -> b_) -> c_'.
	The result is '$Failed' or a 8-tuple '{newInst, skipPos, newIdx, newMeta, newTFix, newTVar, abbrevs, tsubst}',
	where
	- 'newInst' is the new list of terms (which could also be pairs of the form '{Skip, typ}'),
	- 'skipPos' is the index of the last element in 'newInst' before the first '{Skip, typ}', or '-1' if no such
		element exists,
	- 'newIdx' is the new type variable index,
	- 'newMeta' is the list of new meta variables with types,
	- 'newTFix' and 'newTVar' are the new associations of a.b.f. type constants and type variables, repectively,
	- 'abbrevs' is the list of new term abbreviations, and
	- 'tsubst' is a type substitution that has to be applied to the current proof situation. *)
makeInstantiation[ inst_List, term_, head_, cond_, stripImpl_, ps:PRFSIT$[ _, _, idx_, fix_, meta_, tfix_, tvar_, __], mode_] :=
	Replace[
		Reap[ Catch[ Block[ {$instAbbrevs = getTermAbbrevList[ ps], $instNewAbbrevs = {}},
				{
					makeInstantiationAux[
						inst,
						term,
						{idx, fix, meta, tfix},
						{},
						Association[],
						tvar,
						{},
						If[ TrueQ[ cond],
							{
								TYPED$[ head, {typ_}][ _][ A_] :> {typ, A},
								TYPED$[ head, {typ_}][ _] :> {typ, Null},
								_ -> $Failed
							},
						(*else*)
							{
								TYPED$[ head, {typ_}][ A_] :> {typ, A},
								TYPED$[ head, {typ_}] :> {typ, Null},
								_ -> $Failed
							}
						],
						stripImpl,
						False,
						mode
					],
					DeleteDuplicatesBy[ $instNewAbbrevs, First]
				}
			]],
			"inst"
		],
		{
			{{{{newIdx_, _, _, _}, newMeta_List, tsubst_List, newTFix_Association, newTVar_Association}, abbrevs_List}, inst0_List} :>
				With[ {fc = Length[ fix]},
				Module[ {allMeta = Join[ meta, newMeta], skipPos = -1},
				With[ {inst1 =
						MapIndexed[
							Replace[ #1,
								{
									{Automatic, name_, typ_} :>
										With[ {j = Max[ 0, Max @@ Cases[ allMeta, (META$[ name, k_Integer] -> _) :> k] + 1]},
											AppendTo[ allMeta, META$[ name, j] -> {typ, fc}];
											(* 'tsubst' does not need to be applied to 'typ', since meta variables are introduced before "instT" is applied *)
											META$[ name, j]
										],
									{Skip, typ_} :> (If[ Negative[ skipPos], skipPos = First[ #2] - 1]; {Skip, typ /. tsubst}),
									expr_ :> replaceTypes[ expr, tsubst]
								}
							]&,
							Join @@ inst0
						]},
					{inst1, skipPos, newIdx, Drop[ allMeta, Length[ meta]], newTFix, newTVar, abbrevs, tsubst}
				]]],
			_ -> $Failed
		}
	]

makeInstantiationAux[ {("_"...)|(h_[ "_"] /; h === RepeatedNull)}, _, ctxt_, newMeta_, newTFix_, tvar_, tsubst_, ___] :=
	{ctxt, newMeta, tsubst, newTFix, tvar}
makeInstantiationAux[ inst_List, Lambda$TM[ _RNG$, A_], args__] :=
	makeInstantiationAux[ inst, A, args]
makeInstantiationAux[ inst_List, TYPED$[ CoreImplies$TM, {}][ _][ A_], ctxt_, newMeta_, newTFix_, tvar_, tsubst_, rl_List, stripImpl_, args__] :=
	Which[
		TrueQ[ stripImpl],
		makeInstantiationAux[ inst, stripAssms[ A], ctxt, newMeta, newTFix, tvar, tsubst, rl, stripImpl, args],
	(*else*)
		MatchQ[ inst, {h_[ _] /; h === RepeatedNull}],
		{ctxt, newMeta, tsubst, newTFix, tvar},
	(*else*)
		True,
		sowMessage[ 266, {}, "Fatal"];
		Throw[ $Failed]
	]
makeInstantiationAux[ {h_[ inst_] /; h === RepeatedNull}, A_, ctxt_, newMeta_, newTFix_, tvar_, tsubst_, rl_List, args___] :=
	If[ ListQ[ Replace[ A, rl]],
		makeInstantiationAux[ {inst, h[ inst]}, A, ctxt, newMeta, newTFix, tvar, tsubst, rl, args],
	(*else*)
		{ctxt, newMeta, tsubst, newTFix, tvar}
	]
makeInstantiationAux[ {Automatic|"$", r___}, term_, ctxt_, newMeta_, newTFix_, tvar_, tsubst_, rl_List, args___] :=
	Replace[ Replace[ term, rl],
		{
			{typ_, A_} :>
				With[ {name = Replace[ A, {Lambda$TM[ RNG$[ TYPRNG$[ n:Except[ Null, _Symbol], _]], _] :> n, _ -> Theorema`Knowledge`VAR$X$TM}]},
					Sow[ {Automatic, name, typ}, "inst"];
					makeInstantiationAux[ {r}, A, ctxt, newMeta, newTFix, tvar, tsubst, rl, args]
				],
			_ :> (sowMessage[ 266, {}, "Fatal"]; Throw[ $Failed])
		}
	]
makeInstantiationAux[ inst:{"_", r__}, term_, ctxt_, newMeta_, newTFix_, tvar_, tsubst_, rl_List, stripImpl_, _, mode_] :=
	Replace[ Replace[ term, rl],
		{
			{typ_, A_} :>
				(Sow[ {Skip, typ}, "inst"]; makeInstantiationAux[ {r}, A, ctxt, newMeta, newTFix, tvar, tsubst, rl, stripImpl, True, mode]),
			_ :> (sowMessage[ 266, {}, "Fatal"]; Throw[ $Failed])
		}
	]
makeInstantiationAux[ {f0_, r___}, term_, {idx_, fix_, meta_, tfix_}, newMeta_, newTFix_, tvar_, tsubst_, rl_List, stripImpl_, postSkip_, mode_] :=
	Replace[ Replace[ term, rl],
		{
			{typ_, A_} :>
				Replace[
					makeCoreExprPatt[
						replaceTypes[ f0, tsubst],
						typ /. tsubst,
						idx,
						fix,
						Join[ meta, newMeta],
						Join[ tfix, newTFix],	(* keys are disjoint *)
						tvar,
						$instAbbrevs,
						None,
						normalize,
						If[ postSkip && mode =!= 2, Check, None]
					],
					{
						{{_, STERM$[ f_, tf_, tv_], newIdx_, {}, newMeta0_List}, tinst_, inst_, abb_} :>
							If[ $instantiateTypes || tinst === {},
								Sow[ f, "inst"];
								$instNewAbbrevs = Join[ abb, $instNewAbbrevs];
								makeInstantiationAux[
									{r},
									A,
									{newIdx, fix /. tinst, meta /. tinst, tfix},
									Join[ newMeta /. tinst, newMeta0],
									Join[ newTFix, tf],		(* keys are disjoint *)
									tv,
									Join[ tsubst /. tinst, tinst],
									rl,
									stripImpl,
									postSkip,
									mode
								],
							(*else*)
								sowMessage[ 265, {}, "Fatal"];
								Throw[ $Failed]
							],
						_ :> (sowMessage[ 265, {}, "Fatal"]; Throw[ $Failed])
					}
				],
			_ :> (sowMessage[ 266, {}, "Fatal"]; Throw[ $Failed])
		}
	]
makeInstantiationAux[ _, _, __] :=
	(sowMessage[ 266, {}, "Fatal"]; Throw[ $Failed])

End[]

EndPackage[]
