(* ::Package:: *)

(* Theorema
    Copyright (C) 2010 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Core`Trusted`Consistency`"]

Needs[ "Theorema`Language`"]	(* only to have symbols in right context *)
Needs[ "Theorema`Common`"]		(* only to have symbols in right context *)
Needs[ "Theorema`Core`Trusted`"]

MessageName[ $globalContext, "usage", "English"] = "";	(* TODO: For debugging only -- remove when publish. *)
MessageName[ $traceDependencies, "usage", "English"] = "$traceDependencies specifies whether dependencies of type- and constant definitions
														shall be traced (and checked for acyclicity) in exploration mode as well.
														Normally this value is 'False', because in exploration mode too many things can
														be changed that significantly affect dependencies (e.g. re-declaring a constant).
														Hence, this value should only be set to 'True' for testing purposes.";

Begin[ "`Private`"]

(* The global theory context is a collection of symbols 'thyMode', 'thyCounter', 'ctxtOpen', 'typeDefs', 'lockedConsts',
	'constDecls', 'constDefs', 'fmls', 'thms', 'thmInsts', 'typeDefsEnabled', 'typeOverloadingEnabled',
	'constOverloadingEnabled', 'constOverloaded', 'thyDeps', 'thyScope', 'thyDistF', 'importedThys', 'imageFile'.
	Below, 'cnt' always refers to the value of the theory counter at the time when the respective entry was added.
	
	'typeDefs' is an association that maps any type constructor 'head' that has been declared/defined in the current
		*session* to a non-empty list of all declared/defined types with head 'head'. The list has elements of the form
		'typ -> {cnt, key}' or 'typ -> {cnt, key, pred, tvp, abs, rep, deps}, where 'typ' is the actual type of the
		declaration/definition, 'key' is its key, 'pred', 'abs' and 'rep' are the predicate and type morphisms,
		respectively, of the definition, 'tvp' are the type variable positions with sort contraints in 'pred', and
		'deps' is the list of (relevant) constants the definition depends upon. No two types are unifiable.
		Note that all types must be well-formed, which in particular implies that there must be no type constructors
		with empty argument-lists: 'constr[]'.
		As in Isabelle, type declarations/definitions are not constrained by sorts; only characteristic axioms of type
		definitions may incorporate sort constraints.
	
	'lockedConsts' is an association that maps any constant 'c' that has been locked in the current *session* to a pair
		'{cnt, key}'. It is not possible to only lock certain type instances of a constant.
	
	'constDecls' is an association that maps any constant 'c' that has been declared in the current *session* to a
		triple '{cnt, key, typ}', where 'typ' is the declared type of 'c' (including some further information, e.g. the
		positions of type variables with sort constraints) and 'key' is a key like in formulas.
		Note that keys of declarations do not need to be unique; they are merely used to determine the source of a
		declaration.
	
	'constDefs' is an association that maps any constant 'c' that has been defined in the current *session* to a
		non-empty list of all definitions for type instances of 'c'; 'c' is a plain constant without type annotation.
		The list has elements of the form 'typ -> {key, deps}', where 'typ' is the *short* type of the respective
		definition and 'key' is its key. In strict mode, no two types in the list are unifiable.
	
	'fmls' behaves as follows:
	- 'fmls[]' returns a nested list of key-label pairs of all formulas currently in the context. Key-label pairs
		are grouped according to the source-information in the key.
	- 'fmls[ k]' returns '{cnt, f, stat}', where 'f' is the formula with key 'k', if it is in the context, and 'stat'
		is status information.
		Otherwise, if the formula is not in the context, it returns '{$Failed, $Failed, $Failed}'.
		Note that 'cnt' refers to the time when the formula was added, not when it was proved, although proving
		formulas increments the theory counter!
		The status information is an integer with the following meaning:
		 '0': unproved axiom (no definition),
		 '1': formula with invalid proof,
		 '2': formula with cyclic proof where the formula itself is used directly,
		 '3': formula with incomplete proof,
		 '4': formula with skipped proof,
		 '5': theorem with complete, valid proof (in non-strict mode the proof could be circular, though),
		'11': definition with some illegal type appearing on the RHS (e.g. an unknown type or a type variable that does
				not appear on the LHS),
		'12': recursive definition (where an instance of the constant to be defined appears on the RHS explicitly),
		'13': cyclic definition (UNUSED),
		'14': overlapping definition (where the LHS overlaps with the LHS of another definition);
				note that the status of the other overlapping formula(s) is *not* changed to '14', and also note that
				changing the other overlapping formula(s) later might render the current one non-overlapping!
		'15': well-formed definition (in exploration mode the definition could still be cyclic, though),
		'21': axiom originating from type definition with invalid non-emptiness proof,
		'22': axiom originating from type definition with cyclic non-emptiness proof,
		'23': axiom originating from type definition with incomplete non-emptiness proof,
		'24': axiom originating from type definition with skipped non-emptiness proof,
		'25': axiom originating from type definition with complete and valid non-emptiness proof.
	
	'thms' behaves as follows:
	- 'thms[]' returns a nested list of key-label pairs of all theorem schemas currently in the context. Key-label
		pairs are grouped according to the source-information in the key.
	- 'thms[ k]' returns the theorem schema with key 'k', of the form '{cnt, statements, prfs, indices}', if it is in
		the context. Otherwise, it returns '{$Failed, $Failed, $Failed, $Failed}'.
		Instances of statements are expected to be pairs '{f, tvp}', where 'f' is the actual formula and 'tvp' is the
		list of type variable positions.
	
	'thmInsts' is a flat list of all theorem instances used in the current session. It is only filled in strict theory
		mode.
	
	'thyDeps' is an association, mapping theories (specified by the file names of their respective notebooks, including
		prefix "Source:", although this is not enforced by 'manipulateContext') to lists of their ancestors.
		All currently loaded theories are contained in the association.
	
	'thyScope' is a flat list (possibly empty) of all theories currently in scope. The first element of the list, if
		there is one, depends on all other elements in the list; the remaining elements are ordered by their shortest
		distance from the root element. This does *not* mean that whenever some theory A depends on theory B, A comes
		before B in the list. But if B can be reached from the root only via A, then indeed A comes before B.
	
	'thyDistF' is a function that, for each theory-string 'thy', returns the index of 'thy' in 'thyScope' minus 1, or
		'Infinity' if 'thy' does not occur in 'thyScope'.
		Of course, the behavior of 'thyDistF' could also be achieved by something like 'FirstPosition', but 'thyDistF'
		is *much* faster.
	
	'importedThys' is a (possibly empty) flat list of theories, like 'thyScope'. If empty, no theory images have been
		imported so far. The order of the elements in the list is like in 'thyScope'.
*)

(* Return value of 'manipulateContext':
In general, 'manipulateContext' may return anything. There are, however, two return values with some special meaning:
- '$Failed' indicates that something went wrong, i.e. that the requested context manipulation could not be carried out
	completely (but maybe only partially).
- 'False' means that no substantial changes to the context were made. For instance, defining a type that is defined
	in exactly the same way already, changing the meta-data of a formula, or aborting the re-declaration of a constant
	due to 'overwriteF' all result in 'False'.
	ATTENTION! Changing a key (no matter whether it is a formula, a constant declaration, a type definition, etc.) *is*
	a substantial change and therefore *never* results in 'False'!
*)

(* General remark concerning (overloaded) type- and constant definitions:
Definitions of overloaded constants are tricky, as one must be very careful to detect all mutual dependencies of types
and constants in order to rule out cyclic definitions and, hence, inconsistencies. Therefore, we closely follow the
approach implemented in Isabelle, described in

[1] Ondrej Kuncar and Andrei Popescu. Comprehending Isabelle/HOL's Consistency. ESOP'17, LNCS 10201, pp 724-749.
	Springer (2017).

[2] Ondrej Kuncar. Correctness of Isabelle's Cyclicity Checker: Implementability of Overloading in Proof Assistants.
	CPP'15, ACM (2015).

There is one subtle point, though: both papers [1,2] refer to Isabelle/HOL, i.e. *classical* higher-order order logic
with infinity, whereas in our case definitions are implemented on the Core level, i.e. in a *constructive* setting.
However, the proof given in [1] proceeds by a proof-theoretic argument, which under normal circumstances should carry
over readily to our situation; a formal proof of this claim is still missing, though.
Anyway, if someone wants to set up a logic where unrestricted type definitions in the sense of "typedef" are unwanted,
it is easily possible to disable this feature by "typdef off"; this prevents type definitions entirely.
Moreover, overloading type/constant definitions can be disabled, too, by "type overloading off" and
"const overloading off", respectively.

There is actually another question about how all this is really implemented in Isabelle: [2] presents a uniform
treatment of type- and constant definitions. However, type definitions apparently are only possible in Isabelle/HOL ...
but according to Isar-Ref, constant definitions are implemented in Pure ...? So, who is responsible for enforcing
consistent constant definitions: Pure or HOL? And do type definitions still yield plain unchecked axioms from the Pure
point of view?
*)


(* ::Section:: *)
(*Manipulating the Context*)

If[ !MemberQ[ Attributes[ manipulateContext], Protected],
	With[ {thyMode = Unique[ "$thyMode$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			thyCounter = Unique[ "$thyCounter$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			ctxtOpen = Unique[ "$ctxtOpen$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			typeDefs = Unique[ "$typeDefs$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			lockedConsts = Unique[ "$lockedConsts$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			constDecls = Unique[ "$constDecls$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			constDefs = Unique[ "$constDefs$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			fmls = Unique[ "$fmls$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			thms = Unique[ "$thms$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			thmInsts = Unique[ "$thmInsts$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			typeDefsEnabled = Unique[ "$typeDefsEnabled$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			typeOverloadingEnabled = Unique[ "$typeOverloadingEnabled$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			constOverloadingEnabled = Unique[ "$constOverloadingEnabled$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			constOverloaded = Unique[ "$constOverloaded$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			thyDeps = Unique[ "$thyDeps$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			thyScope = Unique[ "$thyScope$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			thyDistF = Unique[ "$thyDistF$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			importedThys = Unique[ "$importedThys$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			imageFile = Unique[ "$imageFile$" <> StringJoin[ ToString /@ RandomInteger[ 9, {10}]] <> "$"],
			constPatt = (_Symbol|_Integer|_Rational|_Real|_Complex|_DirectedInfinity)},
		manipulateContext[ "typedecl", typ:Except[ _STERM$|$Failed], k:{_String, _String}, overwriteF_] /; ctxtOpen :=
			(
				$msgID = None;
				addTypeDeclaration[ thyMode, Hold[ thyCounter], Hold[ ctxtOpen], Hold[ typeDefs], Hold[ lockedConsts], Hold[ constDecls], fmls, typeOverloadingEnabled, typ, k, overwriteF]
			);
		manipulateContext[ "typedef", typ:Except[ _STERM$|$Failed], pred:STERM$[ _, Association[], _Association], prf_Hold, abs:constPatt, rep:constPatt,
				k:{_String, _String}, lbls:{_String, _String, _String}, data_List, overwriteF_, check_] /; ctxtOpen :=
			(
				$msgID = None;
				addTypeDefinition[ thyMode, Hold[ thyCounter], Hold[ ctxtOpen], Hold[ typeDefs], Hold[ lockedConsts], Hold[ constDecls], Hold[ constDefs], fmls,
					Hold[ thmInsts], typeDefsEnabled, typeOverloadingEnabled, constOverloadingEnabled, typ, pred, prf, abs, rep, k, lbls, data, overwriteF, check]
			);
		manipulateContext[ "lock", const:constPatt, k:{_String, _String}] /; ctxtOpen:=
			($msgID = None; addLock[ thyMode, Hold[ thyCounter], Hold[ lockedConsts], Hold[ constDefs], const, k]);
		manipulateContext[ "decl", const:STERM$[ TYPED$[ constPatt, Except[ $Failed]], _Association, _Association], k:{_String, _String}, overwriteF_] /; ctxtOpen :=
			($msgID = None; addDeclaration[ thyMode, Hold[ thyCounter], Hold[ ctxtOpen], Hold[ typeDefs], lockedConsts, Hold[ constDecls], Hold[ constDefs], fmls, const, k, overwriteF]);
		manipulateContext[ "thm schema", k:{_String, _String}, lbl_String, statement:Except[ $Failed], proof:Except[ $Failed], index:Except[ $Failed]] /; ctxtOpen :=
			($msgID = None; addTheoremSchema[ thyMode, Hold[ thyCounter], thms, thyDistF, k, lbl, statement, proof, index]);
		manipulateContext[ "axiom", fml_FML$, tvar_Association, isDef_] /; ctxtOpen :=
			(
				$msgID = None;
				(* In case of a definition, allow to skip checking termination in exploration mode (analogous to 'check' in "proof"). *)
				addAxiom[ thyMode, Hold[ thyCounter], Hold[ ctxtOpen], Hold[ typeDefs], lockedConsts, Hold[ constDecls], Hold[ constDefs], fmls,
					constOverloadingEnabled, Hold[ constOverloaded], thyDistF, fml, tvar, isDef]
			);
		manipulateContext[ "proof", k_, prf_Hold, check_] /; ctxtOpen :=
			($msgID = None; addProof[ thyMode, Hold[ thyCounter], Hold[ ctxtOpen], fmls, Hold[ thmInsts], k, prf, check]);
		manipulateContext[ "facts", prf_Hold, src_String, idFun_, goalKey_List, goalData_List, overwriteF_] /; ctxtOpen :=
			($msgID = None; addFactsByProof[ thyMode, Hold[ thyCounter], Hold[ ctxtOpen], fmls, Hold[ thmInsts], thyDistF, prf, src, idFun, goalKey, goalData, overwriteF]);
		manipulateContext[ "typedef off"] /; ctxtOpen :=
			($msgID = None; disableTypeDefs[ thyMode, Hold[ thyCounter], Hold[ typeDefs], Hold[ typeDefsEnabled]]);
		manipulateContext[ "type overloading off"] /; ctxtOpen :=
			($msgID = None; disableTypeOverloading[ thyMode, Hold[ thyCounter], Hold[ typeDefs], Hold[ typeOverloadingEnabled]]);
		manipulateContext[ "const overloading off"] /; ctxtOpen :=
			($msgID = None; disableConstOverloading[ thyMode, Hold[ thyCounter], Hold[ constOverloadingEnabled], constOverloaded]);
		manipulateContext[ "rem type", typ_] /; ctxtOpen :=
			($msgID = None; remTypeDefinition[ thyMode, Hold[ thyCounter], Hold[ typeDefs], Hold[ lockedConsts], Hold[ constDecls], fmls, typ]);
		manipulateContext[ "rem lock", const_] /; ctxtOpen :=
			($msgID = None; remLock[ thyMode, Hold[ thyCounter], Hold[ lockedConsts], const]);
		manipulateContext[ "rem decl", const_] /; ctxtOpen :=
			($msgID = None; remDeclaration[ thyMode, Hold[ thyCounter], lockedConsts, Hold[ constDecls], Hold[ constDefs], fmls, const]);
		manipulateContext[ "rem thm", k_] /; ctxtOpen :=
			($msgID = None; remTheoremSchema[ thyMode, Hold[ thyCounter], thms, k]);
		manipulateContext[ "fml stat", k_, stat_Integer] /; ctxtOpen :=
			($msgID = None; changeFormulaStat[ thyMode, Hold[ thyCounter], fmls, k, stat]);
		manipulateContext[ "rem fml", k_] /; ctxtOpen :=
			($msgID = None; remFormula[ thyMode, Hold[ thyCounter], Hold[ constDefs], fmls, k]);
		manipulateContext[ "rem defs", const_] /; ctxtOpen :=
			($msgID = None; remConstDefs[ thyMode, Hold[ thyCounter], Hold[ constDefs], fmls, const]);
		manipulateContext[ "typedef on"] /; ctxtOpen :=
			($msgID = None; enableTypeDefs[ thyMode, Hold[ thyCounter], Hold[ typeDefsEnabled]]);
		manipulateContext[ "type overloading on"] /; ctxtOpen :=
			($msgID = None; enableTypeOverloading[ thyMode, Hold[ thyCounter], Hold[ typeOverloadingEnabled]]);
		manipulateContext[ "const overloading on"] /; ctxtOpen :=
			($msgID = None; enableConstOverloading[ thyMode, Hold[ thyCounter], Hold[ constOverloadingEnabled]]);
		manipulateContext[ "change deps", thy_String, ancestors:{___String}] /; ctxtOpen :=
			($msgID = None; changeTheoryDeps[ thyMode, Hold[ thyDeps], Hold[ thyScope], thyDistF, importedThys, thy, ancestors]);
		manipulateContext[ "set thy", thy_String] /; ctxtOpen :=
			($msgID = None; setCurrentTheory[ Hold[ thyDeps], Hold[ thyScope], thyDistF, thy]);
		manipulateContext[ act_String, args___] :=
			(
				$msgID = None;
				If[ TrueQ[ ctxtOpen], sowMessage[ 1017, {act, args}, "Fatal"], sowMessage[ 1018, {act}, "Fatal"]];
				$Failed
			);
		manipulateContext[ args___] :=
			($msgID = None; sowMessage[ 1001, {"manipulateContext", args}, "Fatal"]; $Failed);
		
		getTheoryMode[] := thyMode;
		getTheoryCounter[] := thyCounter;
		getTypeDefs[] := typeDefs;
		getTypeDefs[ False][ head_] := getTypeDefsAux[ typeDefs][ head];
		getTypeDefs[ True][ head_] := getTypeDefsAux[ typeDefs, thyDistF][ head];
		getLockedConsts[] := lockedConsts;
		getLockedConsts[ const_] := Lookup[ lockedConsts, const, $Failed];
		getConstDecls[] := constDecls;
		getConstDecls[ False][ const_] := getConstDeclsAux[ constDecls, Last][ const];
		getConstDecls[ True][ const_] := getConstDeclsAux[ constDecls, Last, thyDistF][ const];
		getConstDefs[] := constDefs;
		getConstDefs[ False][ const_] := getConstDefsAux[ constDefs][ const];
		getConstDefs[ True][ const_] := getConstDefsAux[ constDefs, thyDistF][ const];
		getFmlEnv[] := fmls[];
		getFmlEnv[ False][ k_] := fmls[ k];
		getFmlEnv[ True][ k:{_, s_String}] := If[ IntegerQ[ thyDistF[ s]], fmls[ k], {$Failed, $Failed, $Failed}];
		getThmEnv[] := thms[];
		getThmEnv[ False][ k_] := thms[ k];
		getThmEnv[ True][ k:{_, s_String}] := If[ IntegerQ[ thyDistF[ s]], thms[ k], {$Failed, $Failed, $Failed, $Failed}];
		getDefaultSort[] := {};
			(* The reason why the default sort cannot be changed in the current implementation is because in some places in the implementation
				sort constraints on type variables are not treated consistently:
				Sometimes a type variable not occurring in an association means that no sort constraints are imposed on that variable at all
				(e.g. because the variable has not appeared so far), sometimes it means that the variable has default sort.
				Anyway, the default sort only affects proofs (and in particular the size of proof situations), not the theory context! *)
		typeDefsEnabledQ[] := typeDefsEnabled;
		typeOverloadingEnabledQ[] := typeOverloadingEnabled;
		constOverloadingEnabledQ[] := constOverloadingEnabled;
		constOverloadedQ[] := constOverloaded;
		getTheoryDeps[] := thyDeps;
		getTheoryDeps[ thy_] := Lookup[ thyDeps, thy, $Failed];
		getTheoryScope[] := thyScope;
		getTheoryRootDist[ thy_] := thyDistF[ thy];
		getImportedTheories[] := importedThys;
		getImageFile[] := imageFile;
		
		checkTheoremInsts[] := checkTheoremInsts[ True, True];
		checkTheoremInsts[ scopeOnly_] := checkTheoremInsts[ scopeOnly, True];
		checkTheoremInsts[ scopeOnly_, discard_] /; ctxtOpen :=
			(
				$msgID = None;
				checkTheoremInstsAux[ Hold[ thmInsts], thyDistF, scopeOnly, discard]
			);
		checkTheoremInsts[ args___] := ($msgID = None; sowMessage[ 1001, {"checkTheoremInsts", args}, "Fatal"]; $Failed);
		
		exportTheoryImage[ file_String] := exportTheoryImage[ file, True];
		exportTheoryImage[ file_String, cumulative_] /; ctxtOpen :=
			(
				$msgID = None;
				exportTheoryImageAux[
					thyMode, thyCounter, typeDefs, lockedConsts, constDecls, constDefs, fmls, thms, thmInsts,
					typeDefsEnabled, typeOverloadingEnabled, constOverloadingEnabled, constOverloaded,
					thyDeps, thyScope, importedThys, imageFile,
					file, cumulative
				]
			);
		exportTheoryImage[ args___] := ($msgID = None; sowMessage[ 1001, {"exportTheoryImage", args}, "Fatal"]; $Failed);
		
		importTheoryImage[ file_String] /; ctxtOpen :=
			(
				$msgID = None;
				importTheoryImageAux[
					Hold[
						thyCounter, typeDefs, lockedConsts, constDecls, constDefs, fmls, thms, thmInsts,
						typeDefsEnabled, typeOverloadingEnabled, constOverloadingEnabled, constOverloaded,
						thyDeps, thyScope, thyDistF, importedThys, imageFile
					],
					file
				]
			);
		importTheoryImage[ args___] := ($msgID = None; sowMessage[ 1001, {"importTheoryImage", args}, "Fatal"]; $Failed);
		
		restartSession[] := restartSession[ False];
		restartSession[ strict_] :=
			(
				$msgID = None;
				restartSessionAux[
					Hold[
						thyMode, thyCounter, ctxtOpen, typeDefs, lockedConsts, constDecls, constDefs,
						fmls, thms, thmInsts, typeDefsEnabled, typeOverloadingEnabled, constOverloadingEnabled, constOverloaded,
						thyDeps, thyScope, thyDistF, importedThys, imageFile
					],
					strict
				]
			);
		restartSession[ args___] := ($msgID = None; sowMessage[ 1001, {"restartSession", args}, "Fatal"]; $Failed);
		
		$globalContext = Hold[ thyMode, ctxtOpen, typeDefs, lockedConsts, constDecls, constDefs, fmls, thms, thmInsts, thyDeps, thyScope, thyDistF, importedThys];
		
		SetAttributes[
			{
				getTheoryMode, getTheoryCounter, getTypeDefs, getLockedConsts, getConstDecls, getConstDefs, getFmlEnv,
				getThmEnv, getDefaultSort, typeDefsEnabledQ, typeOverloadingEnabledQ, constOverloadingEnabledQ,
				constOverloadedQ, getTheoryDeps, getTheoryScope, getTheoryRootDist, getImportedTheories, getImageFile,
				manipulateContext, checkTheoremInsts, exportTheoryImage, importTheoryImage, restartSession
			},
			{Protected, ReadProtected, Locked}
		];
	]
]

restartSessionAux[ Hold[ thyMode_, thyCounter_, ctxtOpen_, typeDefs_, lockedConsts_, constDecls_, constDefs_, fmls_, thms_, thmInsts_,
		typeDefsEnabled_, typeOverloadingEnabled_, constOverloadingEnabled_, constOverloaded_, thyDeps_, thyScope_, thyDistF_, importedThys_,
		imageFile_], strict_] :=
	(
		thyMode = strict;
		thyCounter = 0;
		ctxtOpen = True;
		Clear[ fmls, thms, thyDistF];
		typeDefs = Association[];
		lockedConsts = Association[];
		constDecls = Association[];
		constDefs = Association[];
		fmls[] = {};
		fmls[ _] = {$Failed, $Failed, $Failed};
		thms[] = {};
		thms[ _] = {$Failed, $Failed, $Failed, $Failed};
		thmInsts = {};
		typeDefsEnabled = True;
		typeOverloadingEnabled = True;
		constOverloadingEnabled = True;
		constOverloaded = False;
		thyDeps = Association[];
		thyScope = {};
		thyDistF[ _] = Infinity;
		importedThys = {};
		imageFile = None;
	)

getProtectedConsts[] := $protectedConsts;

If[ !ValueQ[ $protectedConsts],
	$protectedConsts =
		{Lambda$TM, CoreImplies$TM, CoreAll$TM, CoreEqual$TM, TVAR$, TFIX$, TYPED$, VAR$, FIX$, META$, SVAR$, RNG$, TYPRNG$, SEQ$, SEQ0$, SEQ1$, List, Hold, $Failed, Null}
	(* 'TrueQ' is not included in '$protectedConsts', since it is perfectly fine to declare/define it.
		The various tools, however, treat 'TrueQ' in a special way, so one should only declare/define
		it if one really knows what one does. *)
];


(* ::Subsection:: *)
(*Exporting and Importing Theory Images*)

exportTheoryImageAux[ thyMode_, thyCounter_, typeDefs_, lockedConsts_, constDecls_, constDefs_, fmls_, thms_, thmInsts_,
		typeDefsEnabled_, typeOverloadingEnabled_, constOverloadingEnabled_, constOverloaded_, thyDeps_, thyScope_,
		importedThys_, imageFile_, file_, cumulative0_] :=
	Which[
		FileExtension[ file] =!= "mx",
		sowMessage[ 1080, {}, "Fatal"];
		$Failed,
	(*else*)
		thyMode =!= True,
		sowMessage[ 1081, {}, "Fatal"];
		$Failed,
	(*else*)
		thyScope === {},
		sowMessage[ 1082, {}, "Fatal"];
		$Failed,
	(*else*)
		True,
		With[ {cumulative = cumulative0 && importedThys =!= {} && MemberQ[ Rest[ thyScope], First[ importedThys]]},
		Module[ {exportThy},
			Scan[
				(exportThy[ #] = True)&,
				If[ TrueQ[ cumulative],
					DeleteCases[ thyScope, Alternatives @@ importedThys],
				(*else*)
					thyScope
				]
			];
			exportThy[ _] = False;
			If[ MemberQ[ thmInsts, THM$[ {_, _String?exportThy}, ___]],
				sowMessage[ 1083, {}, "Fatal"];
				$Failed,
			(*else*)
				$imgDep = If[ TrueQ[ cumulative], imageFile, None];
				$imgCont =
					{
						thyCounter,
						DeleteCases[ Cases[ #, HoldPattern[ _ -> {_, {_, _String?exportThy}, ___}]]& /@ typeDefs, {}],
						Select[ lockedConsts, Composition[ exportThy, Last, Last]],
						Select[ constDecls, exportThy[ #[[2, 2]]]&],
						DeleteCases[ Cases[ #, HoldPattern[ _ -> {{_, _String?exportThy}, _}]]& /@ constDefs, {}],
						Cases[ fmls[], l:{{{_, _String?exportThy}, _}, ___} :> Map[ Composition[ fmls, First], l]],
						Cases[ thms[], l:{{{_, _String?exportThy}, _}, ___} :> Map[ {thms[ First[ #]], #}&, l]],
							(* The functions implementing the theorem schemas must be exported into the extra-logical images. *)
						typeDefsEnabled,
						typeOverloadingEnabled,
						constOverloadingEnabled,
						constOverloaded,
						KeySelect[ thyDeps, exportThy],
						thyScope	(* 'importedThys' is set to this when importing the image. *)
					};
				DumpSave[ file, {$imgDep, $imgCont}];
				True
			]
		]]
	]

importTheoryImageAux[ Hold[ thyCounter_, typeDefs_, lockedConsts_, constDecls_, constDefs_, fmls_, thms_, thmInsts_,
		typeDefsEnabled_, typeOverloadingEnabled_, constOverloadingEnabled_, constOverloaded_, thyDeps_, thyScope_, thyDistF_,
		importedThys_, imageFile_], file_] :=
	Which[
		FileExtension[ file] =!= "mx",
		sowMessage[ 1080, {}, "Fatal"];
		$Failed,
	(*else*)
		!(thyCounter === 0 && importedThys === {}),
		sowMessage[ 1084, {}, "Fatal"];
		$Failed,
	(*else*)
		!FileExistsQ[ file],
		sowMessage[ 1085, {file}, "Fatal"];
		$Failed,
	(*else*)
		True,
		$imgDep = file;
		Replace[
			Reap[
				While[ StringQ[ $imgDep],
					Check[
						Sow[ $imgDep, "File"]; Get[ $imgDep]; Sow[ $imgCont, "Contents"],
						Return[ False]
					]
				],
				{"File", "Contents"}
			],
			{
				{Null, {{files_List}, {conts_List}}} :>
					(
						imageFile = None;
						importedThys = {};
						Catch[
							Scan[
								Function[ fc,
									Replace[ Last[ fc],
										{
											{thyCounter0_Integer, typeDefs0_Association, lockedConsts0_Association, constDecls0_Association,
												constDefs0_Association, fmls0_List, thms0_List, typeDefsEnabled0_, typeOverloadingEnabled0_,
												constOverloadingEnabled0_, constOverloaded0_, thyDeps0_Association, thyScope0_List} :>
												(
													thyCounter = thyCounter0;
													typeDefs = Merge[ {typeDefs, typeDefs0}, Join @@ # &];
													AssociateTo[ lockedConsts, lockedConsts0];
													AssociateTo[ constDecls, constDecls0];
													constDefs = Merge[ {constDefs, constDefs0}, Join @@ # &];
													Scan[ With[ {k = #[[2, 1]]}, fmls[ k] = #]&, fmls0, {2}];
													fmls[] = Join[ Map[ List @@ (#[[2, {1, 3}]]) &, fmls0, {2}], fmls[]];
													Scan[ With[ {k = #[[2, 1]]}, thms[ k] = First[ #]]&, thms0, {2}];
													thms[] = Join[ Map[ Last, thms0, {2}], thms[]];
													typeDefsEnabled = typeDefsEnabled0;
													typeOverloadingEnabled = typeOverloadingEnabled0;
													constOverloadingEnabled = constOverloadingEnabled0;
													constOverloaded = constOverloaded0;
													AssociateTo[ thyDeps, thyDeps0];
													importedThys = thyScope0;
													imageFile = First[ fc];
												),
											_ :>
												(
													sowMessage[ 1086, {First[ fc]}, "Warning"];
													Throw[ imageFile]
												)
										}
									];
								],
								Reverse[ Transpose[ {files, conts}]]
							];
							imageFile
						]
					),
				_ -> $Failed
			}
		]
	]


(* ::Subsection:: *)
(*Additions*)

addTypeDeclaration[ thyMode_, Hold[ thyCounter_], ctxtOpen_Hold, Hold[ typeDefs_], lockedConsts_Hold, constDecls_Hold, fmls_, typeOverloadingEnabled_, typ0_, k_, overwriteF_] :=
	Replace[ liftTypeVars[ typ0, 0, False],
		{typ_, tvarNum_Integer} :>
			Replace[ prepareTypeForDecl[ thyMode, ctxtOpen, Hold[ typeDefs], lockedConsts, constDecls, fmls, typeOverloadingEnabled, typ, tvarNum, k, {}, overwriteF],
				{
					{head_Symbol, None} :>
						With[ {defs = getTypeDefsAux[ typeDefs][ head]},
							AssociateTo[ typeDefs, head -> Append[ defs, typ -> {++thyCounter, k}]];
							True
						],
					{_Symbol, r_} :> r
				}
			]
	]

(* The type variables appearing in 'pred' must be precisely 'TVAR$[ 0]' through 'TVAR$[ arity - 1]' (without names). *)
addTypeDefinition[ thyMode_, Hold[ thyCounter_], ctxtOpen_Hold, Hold[ typeDefs_], lockedConsts_Hold, constDecls_Hold, constDefs_Hold, fmls_, thmInsts_Hold,
		typeDefsEnabled_, typeOverloadingEnabled_, constOverloadingEnabled_, typ0_, STERM$[ pred0_, Association[], pred0TVar_],
		prf_Hold, abs_, rep_, k:{id_, src_}, lbls_, data_, overwriteF_, check_] :=
	Which[
		typeDefsEnabled =!= True,
		sowMessage[ 1056, {typ0}, "Fatal"];
		$Failed,
	(*else*)
		MemberQ[ $protectedConsts, abs],
		sowMessage[ 1041, {typ0, abs}, "Fatal"];
		$Failed,
	(*else*)
		MemberQ[ $protectedConsts, rep],
		sowMessage[ 1041, {typ0, rep}, "Fatal"];
		$Failed,
	(*else*)
		abs === rep,
		sowMessage[ 1042, {typ0}, "Fatal"];
		$Failed,
	(*else*)
		True,
		Replace[ liftTypeVars[ STERM$[ {typ0, pred0}, Association[], pred0TVar], 0, False],
			{STERM$[ {typ_, pred_}, Association[], predTVar_Association], tvarNum_Integer} :>
				Replace[ prepareTypeForDecl[ thyMode, ctxtOpen, Hold[ typeDefs], lockedConsts, constDecls, fmls, typeOverloadingEnabled, typ, tvarNum, k, {pred, abs, rep, _}, overwriteF],
					{
						{head_Symbol, None} :>
							Replace[
								addTypeDefinitionAux[
									True, thyMode, Hold[ thyCounter], ctxtOpen, getTypeDefsAux[ typeDefs], lockedConsts, constDecls, constDefs, fmls, thmInsts,
									constOverloadingEnabled, typ, tvarNum, STERM$[ pred, Association[], predTVar], prf, abs, rep, k, lbls, data, check
								],
								{
									{tvp_List, deps_List} :>
										(
											(* 'thyCounter' has been incremented already *)
											AssociateTo[ typeDefs, head -> Append[ getTypeDefsAux[ typeDefs][ head], typ -> {thyCounter, k, pred, tvp, abs, rep, deps}]];
											{head, typ}
										),
									_ -> $Failed
								}
							],
						{_Symbol, r_} :>
							(* Even if the type has been defined already we check the non-emptiness proof and update the context,
								but we do not do any other checks. *)
							Replace[
								addTypeDefinitionAux[
									False, thyMode, Hold[ thyCounter], ctxtOpen, getTypeDefsAux[ typeDefs], lockedConsts, constDecls, constDefs, fmls, thmInsts,
									constOverloadingEnabled, typ, tvarNum, pred, prf, abs, rep, k, lbls, data, check
								],
								{
									_List -> r,
									_ -> $Failed
								}
							]
					}
				]
		]
	]

addLock[ thyMode_, Hold[ thyCounter_], Hold[ lockedConsts_], Hold[ constDefs_], const_, k_] :=
	If[ MemberQ[ $protectedConsts, const],
		False,
	(*else*)
		If[ getConstDefsAux[ constDefs][ const] === {},
			With[ {cnt = First[ Lookup[ lockedConsts, const, {$Failed}]]},
				If[ IntegerQ[ cnt],
					AssociateTo[ lockedConsts, const -> {cnt, k}];
					Replace,
				(*else*)
					AssociateTo[ lockedConsts, const -> {++thyCounter, k}];
					Append
				]
			],
		(*else*)
			sowMessage[ 1043, {const}, "Fatal"];
			$Failed
		]
	]

addDeclaration[ thyMode_, Hold[ thyCounter_], Hold[ ctxtOpen_], Hold[ typeDefs_], lockedConsts_, Hold[ constDecls_], constDefs_Hold, fmls_,
		STERM$[ TYPED$[ const_, typ_], tfix0_, tvar0_], k_, overwriteF_] :=
	Which[
		MemberQ[ $protectedConsts, const] || ListQ[ Lookup[ lockedConsts, const, $Failed]],
		sowMessage[ 1009, {const}, "Fatal"];
		$Failed,
	(*else*)
		isLegalType[
			typ,
			getTypeDefsAux[ typeDefs],
			True&,
			False&,
			sowMessage[ 1036, {const, #}, "Fatal"]&,
			sowMessage[ 1037, {const, #}, "Fatal"]&
		],
		(* In constant declarations, names of type variables must disappear (otherwise 'getDeclaredType' is incorrect). *)
		With[ {newType =
				Which[
					const =!= TrueQ$TM,
					Replace[ generalizeTypes @@ Insert[ liftTypeVars[ STERM$[ typ, tfix0, tvar0], 0, False], {}, 2],
						{STERM$[ newType0_, _, tvar_], idx_} :>
							{newType0, idx, Table[ Position[ newType0, TVAR$[ i]] -> Lookup[ tvar, TVAR$[ i], getDefaultSort[]], {i, 0, idx - 1}]}
							(* positions must be in the right order *)
					],
				(*else*)
					!AllTrue[ typeDefs, MatchQ[ #, {(_ -> {_, _})...}]&],	(* 'AllTrue' only considers the values of an association. *)
					sowMessage[ 1055, {}, "Fatal"];
					$Failed,
				(*else*)
					MatchQ[ typ, FUN$TM[ _?(FreeQ[ #, FORM$TM|_TVAR$|_TFIX$]&), FORM$TM]],
					{typ, 0, {}},
				(*else*)
					True,
					sowMessage[ 1012, {}, "Fatal"];
					$Failed
				]},
			If[ ListQ[ newType],
				With[ {decl = getConstDeclsAux[ constDecls, All][ const]},
					If[ decl === $Failed,
						AssociateTo[ constDecls, const -> {++thyCounter, k, newType}];
						Append,
					(*else*)
						Which[
							Last[ decl] === newType,
							If[ k =!= decl[[2]], AssociateTo[ constDecls, const -> {First[ decl], k, newType}]];
							Replace,
						(*else*)
							thyMode,
							sowMessage[ 1007, {const}, "Fatal"];
							$Failed,
						(*else*)
							sameCellQ[ k, decl[[2]]] || Block[ {ctxtOpen = False}, overwriteF[ const, decl[[-1, 1]]]],
							If[ TrueQ[ remConstDefsAux[ constDefs, fmls, const]],
								AssociateTo[ constDecls, const -> {++thyCounter, k, newType}];
								Replace,
							(*else*)
								$Failed
							],
						(*else*)
							True,
							False
						]
					]
				],
			(*else*)
				$Failed
			]
		],
	(*else*)
		True,
		$Failed
	]

addTheoremSchema[ thyMode_, Hold[ thyCounter_], thms_, thyDistF_, k:{_String, src_String}, l_String, statement_, proof_, index_] :=
	Switch[ thms[ k],
		{$Failed, $Failed, $Failed, $Failed},
		thms[ k] = {++thyCounter, statement, proof, index};
		With[ {i = First[ FirstPosition[ thms[], {{{_, src}, _}, ___}, {0}, {1}, Heads -> False]]},
			If[ i === 0,
				PrependTo[ thms[], {{k, l}}],
			(*else*)
				thms[] = Insert[ thms[], {k, l}, {i, 1}]
			]
		];
		Append,
	(*else*)
		{_, statement, proof, index},
		Replace[ getKeyLabelPos[ thms[], k],
			{
				{i:Except[ 0], j:Except[ 0]} :>
					If[ thyDistF[ src] === 0 || l === Extract[ thms[], {i, j, 2}],
						thms[] = ReplacePart[ thms[], {i, j, 2} -> l];
						Replace,
					(*else*)
						sowMessage[ 1074, {k}, "Fatal"];
						$Failed
					],
				_ -> Replace
			}
		],
	(*else*)
		_,
		If[ TrueQ[ thyMode],
			sowMessage[ 1011, {k}, "Fatal"];
			$Failed,
		(*else*)
			thms[ k] = {++thyCounter, statement, proof, index};
			Replace
		]
	]

(* 'addAxiom' can also be used to change the meta-data of existing formulas (even if they are proved theorems). *)
addAxiom[ thyMode_, thyCounter_Hold, ctxtOpen_Hold, Hold[ typeDefs_], lockedConsts_, constDecls_Hold, constDefs_Hold, fmls_,
		constOverloadingEnabled_, constOverloaded_Hold, thyDistF_, fml:FML$[ k_, f_, l_String, data0___], tvar_, isDef_] :=
	Replace[ fmls[ k],
		{
			{cnt_Integer, FML$[ k, f, l0_, pre___, "TypeVarPositions" -> tvp_List, post___], stat_Integer?NonNegative} /; ((10 < stat <= 20) === isDef) :>
				With[ {data = DeleteCases[ FML$[ data0], "TypeVarPositions" -> _]},
					If[ thyDistF[ Last[ k]] === 0 || (l === l0 && Sort[ data] === Sort[ FML$[ pre, post]]),
						(* Note that in exploration mode 'f' might not type-check any more (because a constant was
							re-declared). Nevertheless, we do not check this here, since anyway an ill-typed formula would
							remain in the context. *)
						fmls[ k] =
							{
								cnt,
								Join[
									FML$[ k, f, l,
										"TypeVarPositions" -> Map[ First[ #] -> Lookup[ tvar, extractSingle[ f, #[[1, 1]]], getDefaultSort[]] &, tvp]
									],
									data
								],
								stat
							};
						If[ l0 =!= l,
							Replace[ getKeyLabelPos[ fmls[], k],
								{i:Except[ 0], j:Except[ 0]} :> (fmls[] = ReplacePart[ fmls[], {i, j, 2} -> l])
							]
						];
						False,
					(*else*)
						sowMessage[ 1073, {k}, "Fatal"];
						$Failed
					]
				],
			{$Failed, $Failed, $Failed} :>
				addAxiomAux[ thyMode, thyCounter, ctxtOpen, getTypeDefsAux[ typeDefs], lockedConsts, constDecls, constDefs, fmls, constOverloadingEnabled,
					constOverloaded, thyDistF, fml, tvar, isDef],
			fml0_ :>
				Which[
					thyMode,
					sowMessage[ 1005, {k}, "Fatal"];
					$Failed,
				(*else*)
					remFormulaAux[ constDefs, fmls, fml0],
					addAxiomAux[ thyMode, thyCounter, ctxtOpen, getTypeDefsAux[ typeDefs], lockedConsts, constDecls, constDefs, fmls, constOverloadingEnabled,
						constOverloaded, thyDistF, fml, tvar, isDef],
				(*else*)
					True,
					$Failed
				]
		}
	]

(* In strict theory mode, 'addProof' only succeeds if the formula 'f' to be proved was added just before. This rules
	out circular arguments, since all other formulas that are used in the proof of 'f' had existed (and had been
	proved) already *before* 'f' was added, meaning their proofs cannot depend on it. Still, 'f' itself could possibly
	be used in its own proof, but 'checkProof' would detect such a situation and report it. *)
addProof[ thyMode_, Hold[ thyCounter_], Hold[ ctxtOpen_], fmls_, thmInsts_Hold, k_, prf_Hold, check_] :=
	Replace[ fmls[ k],
		{
			{cnt_Integer, fml_FML$, stat_Integer?NonNegative} /; stat <= 10 :>
				If[ thyMode =!= True || cnt === thyCounter,
					With[ {res = If[ TrueQ[ thyMode || check], Block[ {ctxtOpen = False}, Identity[ (#1&)[ checkProof[ thyMode, thmInsts, fml, prf], "Message"]]], 5]},
						If[ res < 4, sowMessage[ 1050 - res, {k}, "Warning"]];
						++thyCounter;
						fmls[ k] = {cnt, fml, res};
						True
					],
				(*else*)
					sowMessage[ 1051, {k}, "Fatal"];
					$Failed
				],
			{_Integer, _FML$, _} :> (sowMessage[ 1050, {k}, "Fatal"]; $Failed),
			_ :> (sowMessage[ 1004, {k}, "Fatal"]; $Failed)
		}
	]

addFactsByProof[ thyMode_, thyCounter_Hold, Hold[ ctxtOpen_], fmls_, Hold[ thmInsts_], thyDistF_, prf_Hold, src_String, idFun_, goalKey_List, goalData_List, overwriteF_] :=
	Replace[ Block[ {ctxtOpen = False}, obtainFactsFromProof[ prf, src, idFun, goalKey, goalData]],
		{
			{{}, _, _} -> {},
			{new_List, usedThms_List, trust_} :>
				(
					If[ TrueQ[ thyMode], thmInsts = Union[ thmInsts, usedThms]];
					With[ {stat = If[ TrueQ[ trust], 5, 4]},
						Select[ new, addFormula[ thyMode, thyCounter, Hold[ ctxtOpen], fmls, thyDistF, #, stat, overwriteF] &]
					]
				),
			_ :> (sowMessage[ 1052, {}, "Fatal", "facts"]; $Failed)
		}
	]

disableTypeDefs[ _, Hold[ thyCounter_], Hold[ typeDefs_], Hold[ typeDefsEnabled_]] :=
	Which[
		!typeDefsEnabled,
		False,
	(*else*)
		NoneTrue[ typeDefs, MemberQ[ #, _ -> {_, _, __}]&],
		++thyCounter;
		typeDefsEnabled = False;
		True,
	(*else*)
		True,
		sowMessage[ 1054, {}, "Fatal"];
		$Failed
	]

disableTypeOverloading[ _, Hold[ thyCounter_], Hold[ typeDefs_], Hold[ typeOverloadingEnabled_]] :=
	Which[
		!typeOverloadingEnabled,
		False,
	(*else*)
		AllTrue[ typeDefs, MatchQ[ #, {_Symbol|(_Symbol[ x___TVAR$] /; DuplicateFreeQ[ {x}]) -> _}|{}]&],
		++thyCounter;
		typeOverloadingEnabled = False;
		True,
	(*else*)
		True,
		sowMessage[ 1060, {}, "Fatal"];
		$Failed
	]

disableConstOverloading[ _, Hold[ thyCounter_], Hold[ constOverloadingEnabled_], constOverloaded_] :=
	Which[
		!constOverloadingEnabled,
		False,
	(*else*)
		!constOverloaded,
		(* 'constOverloaded' tells whether a constant has been overloaded in the current session. Inferring this from
			'constDefs' and 'constDecls' is cumbersome (note that one must ensure that all defined constants were
			defined immediately after declaring them). *)
		++thyCounter;
		constOverloadingEnabled = False;
		True,
	(*else*)
		True,
		sowMessage[ 1064, {}, "Fatal"];
		$Failed
	]


(* ::Subsection:: *)
(*Deletions*)

(* 'remTypeDefinition' can also be used to remove type declarations. *)
remTypeDefinition[ thyMode_, Hold[ thyCounter_], Hold[ typeDefs_], lockedConsts_, constDecls_Hold, fmls_, typ_] :=
	If[ TrueQ[ thyMode],
		sowMessage[ 1045, {typ}, "Fatal"];
		$Failed,
	(*else*)
		With[ {head = Replace[ typ, {h_Symbol :> h, (h_Symbol)[ __] :> h, _ -> $Failed}]},
			Which[
				head === $Failed,
				sowMessage[ 1053, {typ}, "Fatal"];
				$Failed,
			(*else*)
				MemberQ[ $protectedConsts, head] || head === FORM$TM || head === FUN$TM || head === STRING$TM,
				sowMessage[ 1046, {head}, "Fatal"];
				$Failed,
			(*else*)
				True,
				With[ {defs = getTypeDefsAux[ typeDefs][ head], max = getMaxTVar[ typ] + 1},
				With[ {overlaps = Position[ defs, (t_ -> _) /; ListQ[ unifyTypes[ typ, shiftTypeVars[ t, max], None&, {None, {}}, 0]], {1}, Heads -> False]},
					Which[
						overlaps === {},
						sowMessage[ 1021, {typ}, "Warning"];
						False,
					(*else*)
						remTypeDefinitionsAux[ Hold[ typeDefs], lockedConsts, constDecls, fmls, defs, head, overlaps],
						++thyCounter;
						True,
					(*else*)
						True,
						$Failed
					]
				]]
			]
		]
	]

remLock[ thyMode_, Hold[ thyCounter_], Hold[ lockedConsts_], const_] :=
	Which[
		thyMode,
		sowMessage[ 1023, {const}, "Fatal"];
		$Failed,
	(*else*)
		MemberQ[ $protectedConsts, const],
		sowMessage[ 1024, {const}, "Fatal"];
		$Failed,
	(*else*)
		Lookup[ lockedConsts, const, $Failed] === $Failed,
		sowMessage[ 1025, {const}, "Warning"];
		False,
	(*else*)
		True,
		++thyCounter;
		KeyDropFrom[ lockedConsts, const];
		True
	]

remDeclaration[ thyMode_, Hold[ thyCounter_], lockedConsts_, Hold[ constDecls_], constDefs_Hold, fmls_, const_] :=
	Which[
		thyMode,
		sowMessage[ 1026, {const}, "Fatal"];
		$Failed,
	(*else*)
		MemberQ[ $protectedConsts, const] || ListQ[ Lookup[ lockedConsts, const, $Failed]],
		sowMessage[ 1027, {const}, "Fatal"];
		$Failed,
	(*else*)
		getConstDeclsAux[ constDecls, All][ const] === $Failed,
		sowMessage[ 1028, {const}, "Warning"];
		False,
	(*else*)
		remConstDefsAux[ constDefs, fmls, const],
		++thyCounter;
		KeyDropFrom[ constDecls, const];
		True,
	(*else*)
		True,
		$Failed
	]

remTheoremSchema[ thyMode_, Hold[ thyCounter_], thms_, k_] :=
	Which[
		thyMode,
		sowMessage[ 1013, {k}, "Fatal"];
		$Failed,
	(*else*)
		thms[ k] === {$Failed, $Failed, $Failed, $Failed},
		sowMessage[ 1014, {k}, "Warning"];
		False,
	(*else*)
		True,
		++thyCounter;
		thms[ k] =.;
		Replace[ getKeyLabelPos[ thms[], k],
			{i:Except[ 0], j:Except[ 0]} :>
				(thms[] =
					If[ j === 1 && MatchQ[ thms[][[i]], {_}],
						Delete[ thms[], i],
					(*else*)
						Delete[ thms[], {i, j}]
					])
		];
		True
	]

changeFormulaStat[ thyMode_, Hold[ thyCounter_], fmls_, k_, stat_] :=
	Which[
		thyMode,
		sowMessage[ 1006, {k}, "Fatal"];
		$Failed,
	(*else*)
		0 <= stat <= 5,
		Replace[ fmls[ k],
			{
				{$Failed, $Failed, $Failed} :> (sowMessage[ 1004, {k}, "Warning"]; False),
				{_, _, stat} -> False,
				{cnt_, fml_, _} :>
					(
						fmls[ k] = {cnt, fml, stat};
						++thyCounter;
						True
					)
			}
		],
	(*else*)
		True,
		sowMessage[ 1008, {k, stat}, "Fatal"];
		$Failed
	]

remFormula[ thyMode_, Hold[ thyCounter_], constDefs_Hold, fmls_, k_] :=
	Which[
		thyMode,
		sowMessage[ 1003, {k}, "Fatal"];
		$Failed,
	(*else*)
		remFormulaAux[ constDefs, fmls, fmls[ k]],
		++thyCounter;
		True,
	(*else*)
		True,
		sowMessage[ 1004, {k}, "Warning"];
		False
	]

remConstDefs[ thyMode_, Hold[ thyCounter_], Hold[ constDefs_], fmls_, const_] :=
	With[ {defs = getConstDefsAux[ constDefs][ const]},
		Which[
			defs === {},
			False,
		(*else*)
			thyMode,
			(* From the purely logical point of view nothing speaks against removing constant definitions even in strict
				theory mode (i.e. regarding the corresponding formulas as plain axioms). However, it would be necessary
				to re-compute all dependencies then, which is very time consuming. *)
			sowMessage[ 1044, {const}, "Fatal"];
			$Failed,
		(*else*)
			remConstDefsAux[ Hold[ constDefs], fmls, const, defs],
			++thyCounter;
			True,
		(*else*)
			True,
			$Failed
		]
	]

enableTypeDefs[ thyMode_, Hold[ thyCounter_], Hold[ typeDefsEnabled_]] :=
	Which[
		typeDefsEnabled,
		False,
	(*else*)
		thyMode,
		sowMessage[ 1057, {}, "Fatal"];
		$Failed,
	(*else*)
		True,
		++thyCounter;
		typeDefsEnabled = True;
		True
	]

enableTypeOverloading[ thyMode_, Hold[ thyCounter_], Hold[ typeOverloadingEnabled_]] :=
	Which[
		typeOverloadingEnabled,
		False,
	(*else*)
		thyMode,
		sowMessage[ 1062, {}, "Fatal"];
		$Failed,
	(*else*)
		True,
		++thyCounter;
		typeOverloadingEnabled = True;
		True
	]

enableConstOverloading[ thyMode_, Hold[ thyCounter_], Hold[ constOverloadingEnabled_]] :=
	Which[
		constOverloadingEnabled,
		False,
	(*else*)
		thyMode,
		sowMessage[ 1067, {}, "Fatal"];
		$Failed,
	(*else*)
		True,
		++thyCounter;
		constOverloadingEnabled = True;
		True
	]

changeTheoryDeps[ thyMode_, Hold[ thyDeps_], Hold[ thyScope_], thyDistF_, importedThys_, thy_, ancestors0_List] :=
	If[ MemberQ[ importedThys, thy],
		sowMessage[ 1059, {thy}, "Fatal"];
		$Failed,
	(*else*)
		With[ {ancestors = Union[ ancestors0]},
		With[ {deps = Lookup[ thyDeps, thy, $Failed]},
			Which[
				deps === ancestors,
				False,
			(*else*)
				deps === $Failed,
				If[ MemberQ[ ancestors, thy],
					sowMessage[ 1058, {}, "Fatal"];
					$Failed,
				(*else*)
					AssociateTo[ thyDeps, thy -> ancestors];
					thyDeps = Join[ Association @@ (# -> {}& /@ ancestors), thyDeps];
					True
				],
			(*else*)
				TrueQ[ thyMode] && !SubsetQ[ ancestors, deps],
				sowMessage[ 1061, {}, "Fatal"];
				$Failed,
			(*else*)
				True,
				With[ {newDeps = Append[ Join[ Association @@ (# -> {}& /@ ancestors), thyDeps], thy -> ancestors]},
					Which[
						!ListQ[ accumulateTheories[ ancestors, {thy}, newDeps]],
						sowMessage[ 1058, {}, "Fatal"];
						$Failed,
					(*else*)
						IntegerQ[ thyDistF[ thy]],
						thyDeps = newDeps;
						thyScope = accumulateTheories[ {First[ thyScope]}, {}, thyDeps];
						Clear[ thyDistF];
						Module[ {i = -1}, Scan[ (thyDistF[ #] = (++i))&, thyScope]];
						thyDistF[ _] = Infinity;
						True,
					(*else*)
						True,
						thyDeps = newDeps;
						True
					]
				]
			]
		]]
	]

setCurrentTheory[ Hold[ thyDeps_], Hold[ thyScope_], thyDistF_, thy_] :=
	If[ thyDistF[ thy] === 0,
		False,
	(*else*)
		With[ {deps = Lookup[ thyDeps, thy, $Failed]},
			If[ deps === $Failed,
				AssociateTo[ thyDeps, thy -> {}];
				thyScope = {thy},
			(*else*)
				thyScope = accumulateTheories[ deps, {thy}, thyDeps]
			];
			Clear[ thyDistF];
			Module[ {i = -1}, Scan[ (thyDistF[ #] = (++i))&, thyScope]];
			thyDistF[ _] = Infinity;
			True
		]
	]


(* ::Section:: *)
(*Auxiliary Functions*)

(* ::Subsection:: *)
(*Types*)

SetAttributes[ getTypeDefsAux, HoldAll];
getTypeDefsAux[ typeDefs_][ head_] :=
	Lookup[ typeDefs, head, {}]
getTypeDefsAux[ typeDefs_, thyDistF_][ head_] :=
	Cases[ Lookup[ typeDefs, head, {}], HoldPattern[ _ -> {_, {_, _?(Composition[ IntegerQ, thyDistF])}, ___}]]

prepareTypeForDecl[ thyMode_, Hold[ ctxtOpen_], Hold[ typeDefs_], lockedConsts_, constDecls_Hold, fmls_, typeOverloadingEnabled_, typ_, tvarNum_Integer, k_, rhs_, overwriteF_] :=
	With[ {head = Replace[ typ, {(h_Symbol)[ __] :> h, Except[ _Symbol] -> $Failed}]},
		Which[
			head === $Failed,
			sowMessage[ 1053, {typ}, "Fatal"];
			$Failed,
		(*else*)
			MemberQ[ $protectedConsts, head] || head === FORM$TM || head === FUN$TM || head === STRING$TM,
			sowMessage[ 1019, {head}, "Fatal"];
			$Failed,
		(*else*)
			TrueQ[ typeOverloadingEnabled] || typ === head || (MatchQ[ typ, head[ ___TVAR$]] && DuplicateFreeQ[ typ]),
			With[ {defs = getTypeDefsAux[ typeDefs][ head]},
			With[ {overlaps =
					If[ TrueQ[ typeOverloadingEnabled],
						Position[ defs, (t_ -> _) /; ListQ[ unifyTypes[ typ, shiftTypeVars[ t, tvarNum], None&, {None, {}}, 0]], {1}, Heads -> False],
					(*else*)
						Array[ List, Length[ defs], 1]
					],
					legal =
						(* must be evaluated *after* removing existing type definitions *)
						Hold[
							isLegalType[
								(* 'FORM$TM' and 'FUN$TM' are dummy type constructors, since the original head 'head' must be removed. *)
								Replace[ typ, {_Symbol :> FORM$TM, (_Symbol)[ a_, b___] :> Fold[ FUN$TM, a, {b}]}],
								getTypeDefsAux[ typeDefs],
								True&,
								False&,
								sowMessage[ 1030, {typ, #}, "Fatal"]&,
								sowMessage[ 1031, {typ, #}, "Fatal"]&
							]
						]},
				Which[
					overlaps === {},
					If[ TrueQ[ ReleaseHold[ legal]],
						{head, None},
					(*else*)
						$Failed
					],
				(*else*)
					MatchQ[ overlaps, {_}] && With[ {d = Extract[ defs, First[ overlaps]]}, First[ d] === typ && MatchQ[ Drop[ Last[ d], 2], rhs]],
					With[ {kp = Join[ First[ overlaps], {2, 2}]},
						If[ Extract[ defs, kp] === k,
							{head, False},
						(*else*)
							AssociateTo[ typeDefs, head -> ReplacePart[ defs, kp -> k]];
							{head, True}
						]
					],
				(*else*)
					thyMode,
					sowMessage[ 1020, {typ}, "Fatal"];
					$Failed,
				(*else*)
					AllTrue[ Extract[ defs, overlaps], sameCellQ[ #[[2, 2]], k]&] ||
						Block[ {ctxtOpen = False}, overwriteF[ Extract[ defs, overlaps][[All, 1]]]],
					If[ TrueQ[ remTypeDefinitionsAux[ Hold[ typeDefs], lockedConsts, constDecls, fmls, defs, head, overlaps]] && TrueQ[ ReleaseHold[ legal]],
						{head, None},
					(*else*)
						$Failed
					],
				(*else*)
					True,
					{head, False}
				]
			]],
		(*else*)
			True,
			sowMessage[ 1016, {typ}, "Fatal"];
			$Failed
		]
	]

addTypeDefinitionAux[ check_, thyMode_, Hold[ thyCounter_], Hold[ ctxtOpen_], typeDefF_, Hold[ lockedConsts_], Hold[ constDecls_], constDefs_Hold, fmls_,
		thmInsts_Hold, constOverloadingEnabled_, typ_, tvarNum_, predAll:STERM$[ pred_, Association[], _], prf_, abs_, rep_, k:{id_String, src_String}, lbls_, data0_, checkPrf_] :=
	Which[
		check && ListQ[ getConstDeclsAux[ constDecls, All][ abs]],
		sowMessage[ 1029, {typ, abs}, "Fatal"];
		False,
	(*else*)
		check && ListQ[ getConstDeclsAux[ constDecls, All][ rep]],
		sowMessage[ 1029, {typ, rep}, "Fatal"];
		False,
	(*else*)
		check && ListQ[ Lookup[ lockedConsts, abs, $Failed]],
		sowMessage[ 1041, {typ, abs}, "Fatal"];
		False,
	(*else*)
		check && ListQ[ Lookup[ lockedConsts, rep, $Failed]],
		sowMessage[ 1041, {typ, abs}, "Fatal"];
		False,
	(*else*)
		True,
		With[ {ot = Replace[ getConstDeclsAux[ constDecls, Last][ TrueQ$TM], {{FUN$TM[ ot0_, FORM$TM], 0, {}} :> ot0, _ -> FORM$TM}]},
			Replace[ Block[ {$strictTypeChecking = True}, checkType[ predAll, Blank[], {}, {}, {}, Association[], Association[]]],
				{
					{FUN$TM[ rt_, ot], {}, _, Association[], tvar_Association} :>
						If[ ot === FORM$TM || FreeQ[ rt, FORM$TM],
							With[ {tvars = Array[ TVAR$, tvarNum, 0]},
								Replace[
									Which[
										check && (TrueQ[ thyMode] || TrueQ[ $traceDependencies]),
										checkDependency[
											thyCounter, typeDefF, lockedConsts, constDefs, fmls, constOverloadingEnabled, Null, typ, pred, False, tvars, tvarNum
										],
									(*else*)
										check && quickcheckDependency[ typeDefF, Null, typ, pred, tvars, tvarNum] === 11,
										False,
									(*else*)
										True,
										{}
									],
									deps_List :>
										With[ {fk = Table[ {id <> "." <> ToString[ i], src}, {i, 1, 3}],
												tj = If[ ot === FORM$TM, Identity, TYPED$[ TrueQ$TM, {}]]},
											If[ !check || fmls[ fk[[1]]] === {$Failed, $Failed, $Failed} === fmls[ fk[[2]]] === fmls[ fk[[3]]],
												With[ {typA = FUN$TM[ rt, typ], typR = FUN$TM[ typ, rt],
														data = FML$ @@ DeleteCases[ data0, "TypeVarPositions" -> _],
														stat =
															If[ TrueQ[ thyMode || checkPrf],
																Block[ {ctxtOpen = False},
																	checkProof[
																		thyMode,
																		thmInsts,
																		FML$[ {"", src},
																			(* '/\thesis::ot. (/\x::rt. tj[ pred[ x]] ==> tj[ thesis]) ==> tj[ thesis]' *)
																			TYPED$[ CoreAll$TM, {ot}][
																				Lambda$TM[ RNG$[ TYPRNG$[ Null, ot]],
																					TYPED$[ CoreImplies$TM, {}][
																						TYPED$[ CoreAll$TM, {rt}][
																							Lambda$TM[ RNG$[ TYPRNG$[ Null, rt]],
																								TYPED$[ CoreImplies$TM, {}][ tj[ applyFunction[ pred, VAR$[ 0], False]]][ tj[ VAR$[ 1]]]
																							]
																						]
																					][
																						tj[ VAR$[ 0]]
																					]
																				]
																			],
																			""
																		],
																		prf
																	]
																],
															(*else*)
																5
															] + 20},
													If[ stat < 24, sowMessage[ 1038, {typ}, "Warning"]];
													++thyCounter;
													addTypeMorphism[ thyCounter, Hold[ lockedConsts], Hold[ constDecls], abs, typA, tvars, tvar, k];
													addTypeMorphism[ thyCounter, Hold[ lockedConsts], Hold[ constDecls], rep, typR, tvars, tvar, k];
													MapThread[
														(fmls[ #1] = {thyCounter, Join[ FML$[ #1, #3, #2, "TypeVarPositions" -> makeTypeVarPositions[ #3, tvar]], data], stat};)&,
														{fk, lbls,
															With[ {abs0 = TYPED$[ abs, tvars], rep0 = TYPED$[ rep, tvars]},
																{
																	(* '/\x::typ. tj[ pred[ rep[ x]]]' *)
																	TYPED$[ CoreAll$TM, {typ}][
																		Lambda$TM[ RNG$[ TYPRNG$[ Null, typ]], tj[ applyFunction[ pred, rep0[ VAR$[ 0]], False]]]
																	],
																	(* '/\x::typ. abs[ rep[ x]] === x' *)
																	TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ RNG$[ TYPRNG$[ Null, typ]],
																		TYPED$[ CoreEqual$TM, {typ}][ abs0[ rep0[ VAR$[ 0]]]][ VAR$[ 0]]
																	]],
																	(* '/\x::rt. tj[ pred[ x]] ==> rep[ abs[ x]] === x' *)
																	TYPED$[ CoreAll$TM, {rt}][ Lambda$TM[ RNG$[ TYPRNG$[ Null, rt]],
																		TYPED$[ CoreImplies$TM, {}][
																			tj[ applyFunction[ pred, VAR$[ 0], False]]
																		][
																			TYPED$[ CoreEqual$TM, {rt}][ rep0[ abs0[ VAR$[ 0]]]][ VAR$[ 0]]
																		]
																	]]
																}
															]
														}
													];
													With[ {i = First[ FirstPosition[ fmls[], {{{_, src}, _}, ___}, {0}, {1}, Heads -> False]]},
														If[ i === 0,
															PrependTo[ fmls[], Transpose[ {fk, lbls}]],
														(*else*)
															fmls[] = MapAt[ DeleteDuplicates[ Join[ Transpose[ {fk, lbls}], #]]&, fmls[], i]
														]
													];
													{makeTypeVarPositions[ pred, tvar], deps}
												],
											(*else*)
												sowMessage[ 1039, {typ}, "Fatal"];
												False
											]
										]
								]
							],
						(*else*)
							sowMessage[ 1022, {typ, rt}, "Fatal"];
							False
						],
					_ :> (sowMessage[ 1040, {typ}, "Fatal"]; False)
				}
			]
		]
	]

addTypeMorphism[ thyCounter_Integer, Hold[ lockedConsts_], Hold[ constDecls_], morph_, typ_, tvarList_List, tvar_, k_] :=
	(
		(* type variables in 'typ' are 'TVAR$[ 0]' through 'TVAR$[ tvarNum - 1]' *)
		AssociateTo[ constDecls, morph -> {thyCounter, k, {typ, Length[ tvarList], Map[ Position[ typ, #] -> Lookup[ tvar, #, getDefaultSort[]] &, tvarList]}}];
		If[ Lookup[ lockedConsts, morph, $Failed] === $Failed, AssociateTo[ lockedConsts, morph -> {thyCounter, k}]];
	)

remTypeDefinitionsAux[ Hold[ typeDefs_], lockedConsts_, constDecls_Hold, fmls_, defs_List, head_, indices_List] :=
	Catch[
		Scan[
			If[ remTypeDefinitionsAux[ lockedConsts, constDecls, fmls, Last[ #]] =!= True, Throw[ False]]&,
			Extract[ defs, indices]
		];
		With[ {newDefs = Delete[ defs, indices]},
			If[ newDefs === {},
				KeyDropFrom[ typeDefs, head],
			(*else*)
				AssociateTo[ typeDefs, head -> newDefs]
			]
		];
		True
	]
remTypeDefinitionsAux[ Hold[ lockedConsts_], Hold[ constDecls_], fmls_, {_, k:{id_, src_}, _, _, abs_, rep_, _}] :=
	(
		KeyDropFrom[ constDecls, {abs, rep}];
		KeyDropFrom[ lockedConsts, {abs, rep}];
		With[ {fk = Table[ {id <> "." <> ToString[ j], src}, {j, 1, 3}]},
			Quiet[ Scan[ fmls[ #] =. &, fk], Unset::norep];
			With[ {j = First[ FirstPosition[ fmls[], {{{_, src}, _}, ___}, {0}, {1}, Heads -> False]]},
				If[ j =!= 0,
					fmls[] = DeleteCases[ MapAt[ DeleteCases[ #, {Alternatives @@ fk, _}]&, fmls[], j], {}]
				]
			]
		];
		True
	)
remTypeDefinitionsAux[ _, _, _, _] :=
	True


(* ::Subsection:: *)
(*Formulas*)

addAxiomAux[ thyMode_, Hold[ thyCounter_], ctxtOpen_Hold, typeDefF_, lockedConsts_, Hold[ constDecls_], constDefs_Hold, fmls_, constOverloadingEnabled_,
		constOverloaded_, thyDistF_, FML$[ k_, f_, l_, data___], tvar_, isDef_] :=
	Replace[
		Block[ {$strictTypeChecking = If[ TrueQ[ isDef], True, $strictTypeChecking],
				$reportUndeclaredConsts = True},	(* if 'isDef' is 'True', this has no effect *)
			Reap[ checkType[ STERM$[ f, Association[], tvar], FORM$TM, {}, {}, {}, Association[], Association[]], "undeclared"]
		],
		{
			{{_, {}, _, Association[], newTVar_Association}, undecl0_List} :>
				With[ {fml = FML$[ k, f, l, "TypeVarPositions" -> makeTypeVarPositions[ f, newTVar], Sequence @@ DeleteCases[ {data}, "TypeVarPositions" -> _]],
						undecl = DeleteDuplicates[ Flatten[ undecl0, 1]]},
					If[ undecl =!= {}, sowMessage[ 1028, undecl, "Warning"]];
					If[ TrueQ[ isDef],
						With[ {stat =
								checkDefinition[
									thyMode, thyCounter, typeDefF, lockedConsts, getConstDeclsAux[ constDecls, All],
									constDefs, fmls, constOverloadingEnabled, constOverloaded, f, k, l
								]},
							If[ IntegerQ[ stat],
								Replace[ addFormula[ thyMode, Hold[ thyCounter], ctxtOpen, fmls, thyDistF, fml, stat, Drop[ #1, 2]&],
									Except[ $Failed|False] :> Extract[ f, {{0, 1, 1}, {0, 1, 2}}]
								],
							(*else*)
								$Failed
							]
						],
					(*else*)
						addFormula[ thyMode, Hold[ thyCounter], ctxtOpen, fmls, thyDistF, fml, 0, Drop[ #1, 2]&]
					]
				],
			_ :> (sowMessage[ 1010, {k, l}, "Fatal"]; $Failed)
		}
	]

addFormula[ thyMode_, Hold[ thyCounter_], Hold[ ctxtOpen_], fmls_, thyDistF_, fml:FML$[ k:{_, src_}, f_, ___, tvp:("TypeVarPositions" -> _), ___], stat_, overwriteF_] :=
	With[ {fml0 = fmls[ k]},
		If[ fml0 === {$Failed, $Failed, $Failed},
			Replace[ Block[ {ctxtOpen = False}, overwriteF[ fml, fml0]],	(* 'overwriteF' may still change the label and formula data, but not the formula itself *)
				{
					_[ l_String, data___] :>
						With[ {fml1 = Join[ FML$[ k, f, l, tvp], DeleteCases[ FML$[ data], "TypeVarPositions" -> _]]},
							fmls[ k] = {++thyCounter, fml1, stat};
							With[ {i = First[ FirstPosition[ fmls[], {{{_, src}, _}, ___}, {0}, {1}, Heads -> False]]},
								If[ i === 0,
									PrependTo[ fmls[], {{k, l}}],
								(*else*)
									fmls[] = Insert[ fmls[], {k, l}, {i, 1}]
								]
							];
							True
						],
					_ -> False
				}
			],
		(*else*)
			(* "facts" only *)
			With[ {unsame = (fml0[[2, 2]] =!= f)},
				If[ TrueQ[ thyMode] && unsame,
					sowMessage[ 1005, {k}, "Fatal"];
					$Failed,
				(*else*)
					Replace[ Block[ {ctxtOpen = False}, overwriteF[ fml, fml0]],
						{
							_[ l_String, data0___] :>
								With[ {data = DeleteCases[ FML$[ data0], "TypeVarPositions" -> _]},
									If[ thyDistF[ src] === 0 || (l === fml0[[2, 3]] && data === DeleteCases[ Drop[ fml0[[2]], 3], "TypeVarPositions" -> _]),
										With[ {fml1 = Join[ FML$[ k, f, l, tvp], data]},
											If[ unsame,
												fmls[ k] = {++thyCounter, fml1, stat},
											(*else*)
												fmls[ k] = {First[ fml0], fml1, Max[ stat, Last[ fml0]]}
											];
											If[ fml0[[2, 3]] =!= l,
												Replace[ getKeyLabelPos[ fmls[], k],
													{i:Except[ 0], j:Except[ 0]} :> (fmls[] = ReplacePart[ fmls[], {i, j, 2} -> l])
												]
											];
											True
										],
									(*else*)
										sowMessage[ 1073, {k}, "Fatal"];
										$Failed
									]
								],
							_ -> False
						}
					]
				]
			]
		]
	]

obtainFactsFromProof[ prf_Hold, src_String, idFun_, goalKey_List, goalData_List] :=
	Replace[
		executeProofCommands[
			prf,
			initiateProof[ FML$[ goalKey, TYPED$[ CoreAll$TM, {FORM$TM}][ Lambda$TM[ RNG$[ TYPRNG$[ Null, FORM$TM]], VAR$[ 0]]], "", Sequence @@ goalData], {}, 2],
			2,
			True
		],
		{
			{PRFSTAT$[ _, {PRFSIT$[ _, hyps_List, __]}, _, usedThms_List, ___], trust_} :>
				With[ {defSource = sourceLabel[]},
				With[ {new =
						MapIndexed[
							If[ #1[[1, 2]] === defSource,
								With[ {newID = idFun[ #1, First[ #2]]},
									If[ StringQ[ newID],
										ReplacePart[ #1, 1 -> {newID, src}],
									(*else*)
										ReplacePart[ #1, {1, 2} -> src]
									]
								],
							(*else*)
								#1
							]&,
							Cases[ hyps, FML$[ _, f_  /; FreeQ[ f, _FIX$|_TFIX$], __]]
							(* We also include formulas that are in the global context already, since their meta-data might have changed. *)
						]},
					{new, usedThms, trust}
				]],
			_ -> $Failed
		}
	]

SetAttributes[ getConstDeclsAux, HoldAll];
getConstDeclsAux[ constDecls_, Last][ c_] :=
	Last[ Lookup[ constDecls, c, {$Failed}]]
getConstDeclsAux[ constDecls_, All][ c_] :=
	Lookup[ constDecls, c, $Failed]
getConstDeclsAux[ constDecls_, sel_, thyDistF_][ c_] :=
	Replace[ Lookup[ constDecls, c, $Failed],
		decl:{_, {_, s_String}, typ_} :>
			Which[
				!IntegerQ[ thyDistF[ s]],
				$Failed,
			(*else*)
				sel === Last,
				typ,
			(*else*)
				True,
				decl
			]
	]

SetAttributes[ getConstDefsAux, HoldAll];
getConstDefsAux[ constDefs_][ c_] :=
	Lookup[ constDefs, c, {}]
getConstDefsAux[ constDefs_, thyDistF_][ c_] :=
	Cases[ Lookup[ constDefs, c, {}], HoldPattern[ _ -> {{_, _String?(Composition[ IntegerQ, thyDistF])}, ___}]]

(* If overloading is enabled, we cannot expect all constants appearing on the RHS of a definition to be defined
	already. For instance, one could abstractly define 'Power' in terms of '1' and 'Times', although '1' and 'Times'
	are not defined in full generality themselves (only certain type instances may be).
	This is in contrast to types, where overloading (in that sense) is not possible. Hence, it is reasonable to expect
	and require all types appearing in the predicate of a type definition to be declared/defined already, since this
	facilitates a lot of things considerably.
	Note that sort constraints are completely ignored when checking for acyclicity. *)
checkDefinition[ thyMode_, thyCounter_, typeDefF_, lockedConsts_, constDeclF_, Hold[ constDefs_], fmls_, constOverloadingEnabled_, Hold[ constOverloaded_],
		TYPED$[ CoreEqual$TM, _][ TYPED$[ const:(_Symbol|_Integer|_Rational|_Real|_Complex|_DirectedInfinity), ct_List]][ term_], k_, _] :=
	If[ MemberQ[ $protectedConsts, const] || ListQ[ Lookup[ lockedConsts, const, $Failed]],
		sowMessage[ 1015, {const, ct}, "Fatal"];
		False,
	(*else*)
		(* There is no need to check whether 'typ' is an instance of the declared type of 'const', since the whole
			formula type-checks strictly. *)
		With[ {tvars = Cases[ ct, _TVAR$, {0, Infinity}, Heads -> True]},
		With[ {maxType = Max[ -1, Max @@ tvars[[All, 1]]] + 1},
		With[ {overlaps = Cases[ getConstDefsAux[ constDefs][ const], (t_ -> {k0_, _}) /; ListQ[ unifyTypes[ ct, shiftTypeVars[ t, maxType], None&, {None, {}}, 0]] :> k0]},
			If[ overlaps =!= {} && (TrueQ[ thyMode] || TrueQ[ $traceDependencies]),
				sowMessage[ 1069, {const, ct, overlaps}, "Fatal"];
				False,
			(*else*)
				With[ {justDeclared = First[ constDeclF[ const]] === thyCounter},
				With[ {overloaded = !(justDeclared && MatchQ[ ct, {___TVAR$}] && DuplicateFreeQ[ ct])},
					If[ TrueQ[ constOverloadingEnabled] || !overloaded,
						constOverloaded = constOverloaded || overloaded;
						If[ TrueQ[ thyMode] || TrueQ[ $traceDependencies],
							(* If 'justDeclared' is 'True', there is no need to perform a full termination check, since 'const' cannot
								occur in any other (type- or constant) definitions anyway. *)
							With[ {out =
									checkDependency[
										thyCounter, typeDefF, lockedConsts, Hold[ constDefs], fmls, constOverloadingEnabled, const, ct, term, !justDeclared, tvars, maxType
									]},
								If[ ListQ[ out],
									AssociateTo[ constDefs, const -> Append[ getConstDefsAux[ constDefs][ const], ct -> {k, out}]];
									15,
								(*else*)
									False
								]
							],
						(*else*)
							With[ {out = quickcheckDependency[ typeDefF, const, ct, term, tvars, maxType]},
								If[ IntegerQ[ out],
									AssociateTo[ constDefs, const -> Append[ getConstDefsAux[ constDefs][ const], ct -> {k, {}}]];
									If[ overlaps === {},
										out,
									(*else*)
										sowMessage[ 1070, {const, ct, overlaps}, "Warning"];
										Min[ out, 14]
									],
								(*else*)
									False
								]
							]
						],
					(*else*)
						sowMessage[ 1066, {const, ct}, "Fatal"];
						False
					]
				]]
			]
		]]]
	]
checkDefinition[ __, k_, l_] :=
	(sowMessage[ 1002, {k, l}, "Fatal"]; False)

remFormulaAux[ constDefs_Hold, fmls_, {_Integer, FML$[ k_, f_, __], stat_}] :=
	Catch[
		If[ TrueQ[ 10 < stat <= 20],
			(* definition *)
			Replace[ f[[0, 1]], TYPED$[ const_, _] :> If[ remConstDefinitionAux[ constDefs, const, k] =!= True, Throw[ False]]]
		];
		Quiet[ fmls[ k] =., Unset::norep];
		Replace[ getKeyLabelPos[ fmls[], k],
			{i:Except[ 0], j:Except[ 0]} :>
				(fmls[] =
					If[ j === 1 && MatchQ[ fmls[][[i]], {_}],
						Delete[ fmls[], i],
					(*else*)
						Delete[ fmls[], {i, j}]
					])
		];
		True
	]
remFormulaAux[ _, _, _] :=
	False

remConstDefinitionAux[ Hold[ constDefs_], const_, k_] :=
	With[ {defs = getConstDefsAux[ constDefs][ const]},
	With[ {i = First[ FirstPosition[ defs, _ -> {k, _}, {0}, {1}, Heads -> False]]},
		If[ i =!= 0,
			If[ i === 1 && MatchQ[ defs, {_}],
				KeyDropFrom[ constDefs, const],
			(*else*)
				AssociateTo[ constDefs, const -> Delete[ defs, i]]
			]
		];
		True
	]]

(* The difference between 'remConstDefinitionAux' and 'remConstDefsAux' is that the former removes *one* definition
	with a particular key (leaving formulas completely unchanged), whereas the latter removes *all* definitions and
	sets the status of the corresponding formulas to '0'. *)
remConstDefsAux[ Hold[ constDefs_], fmls_, const_] :=
	remConstDefsAux[ Hold[ constDefs], fmls, const, getConstDefsAux[ constDefs][ const]]
remConstDefsAux[ Hold[ constDefs_], fmls_, const_, defs_List] :=
	(
		Scan[
			With[ {k = #[[2, 1]]},
				Replace[ fmls[ k],
					{cnt_Integer, fml_FML$, stat_Integer} /; 10 < stat <= 20 :> (fmls[ k] = {cnt, fml, 0};)
				]
			]&,
			defs
		];
		KeyDropFrom[ constDefs, const];
		True
	)


(* ::Subsection:: *)
(*Cyclicity Checker*)

(* Returns either '$Failed' or '0' (not composable) or '1' (cyclic) or '2' (illegal object on RHS) or the list of
	constant instances 'term' depends upon (unless constant overloading is disabled, in which case the list is empty). *)
checkDependency[ thyCounter_, typeDefF_, lockedConsts_, constDefs_, fmls_, constOverloadingEnabled_, const_, typ_, term_] :=
	checkDependency[ thyCounter, typeDefF, lockedConsts, constDefs, fmls, constOverloadingEnabled, const, typ, term, True]
checkDependency[ thyCounter_, typeDefF_, lockedConsts_, constDefs_, fmls_, constOverloadingEnabled_, const_, typ_, term_, full_] :=
	checkDependency[ thyCounter, typeDefF, lockedConsts, constDefs, fmls, constOverloadingEnabled, const, typ, term, full,
		DeleteDuplicates[ Cases[ typ, _TVAR$, {0, Infinity}, Heads -> True]]]
checkDependency[ thyCounter_, typeDefF_, lockedConsts_, constDefs_, fmls_, constOverloadingEnabled_, const_, typ_, term_, full_, tvars_List] :=
	checkDependency[ thyCounter, typeDefF, lockedConsts, constDefs, fmls, constOverloadingEnabled, const, typ, term, full, tvars,
		Max[ -1, Max @@ tvars[[All, 1]]] + 1]
checkDependency[ thyCounter_, typeDefF_, lockedConsts_, Hold[ constDefs_], fmls_, constOverloadingEnabled_, const_, typ_, term_, full_, tvars_List, shift_Integer] :=
	Replace[ collectRelevantObjects[ typeDefF, lockedConsts, const, typ, term, tvars],
		{typeDeps_List, constDeps_List} :>
			If[ TrueQ[ constOverloadingEnabled],
				Module[ {done = Association[]},
					Replace[ checkSingleDependency[ thyCounter, typeDefF, getConstDefsAux[ constDefs], fmls, typeDeps, constDeps, const, typ, Hold[ done], shift],
						{consts_List, check_List} :>
							If[ TrueQ[ full] && check =!= {},
								(* In strict theory mode, one could postpone the checks below until the very end,
									meaning that all definitions are checked only *twice*: once when they are added,
									to collect all constants they depend upon, and once more at the end to also get
									hold of dependencies on constants that were defined later.
									This should be considerably more efficient. *)
								
								If[ const =!= Null,	(* security check; 'const' should not be 'Null' *)
									(* temporarily add new definition to 'constDefs' *)
									With[ {defs = Append[ getConstDefsAux[ constDefs][ const], typ -> {Null, consts}]},
										AssociateTo[ constDefs, const -> defs];
										AssociateTo[ done, const -> Append[ Lookup[ done, const, {}], Length[ defs]]]
									]
								];
								
								With[ {out =
										Catch[
											Module[ {todo = check},
												While[ todo =!= {},
													With[ {fst = First[ todo]},
													With[ {c = First[ fst], i = Last[ fst]},
													With[ {def = getConstDefsAux[ constDefs][ c][[i]]},
														(* 'def' is of the form 't -> {k, deps}' *)
														AssociateTo[ done, c -> Append[ Lookup[ done, c, {}], i]];
														Replace[
															checkSingleDependency[
																thyCounter, typeDefF, getConstDefsAux[ constDefs], fmls, {}, def[[2, -1]],
																c, First[ def], Hold[ done]
																(* const, typ, Hold[ done], maxType, getMaxTVar[ First[ def]] + 1 *)
															],
															{
																{_List, new_List} :> (todo = Union[ Rest[ todo], new]),
																res_ :> Throw[ res]
															}
														]
													]]]
												]
											];
											consts
										]},
									
									If[ const =!= Null,
										(* remove new definition again *)
										With[ {defs = getConstDefsAux[ constDefs][ const]},
											If[ MatchQ[ defs, {_}],
												KeyDropFrom[ constDefs, const],
											(*else*)
												AssociateTo[ constDefs, const -> Most[ defs]]
											]
										]
									];
									
									out
								],
							(*else*)
								consts
							]
					]
				],
			(*else*)
				If[ const =!= Null && MemberQ[ constDeps, TYPED$[ const, _]],
					1,
				(*else*)
					{}
				]
			]
	]

(* For exploration mode only. *)
quickcheckDependency[ typeDefF_, const_, typ_, term_, tvars_List, shift_Integer] :=
	Replace[ collectRelevantObjects[ typeDefF, Association[], const, typ, term, tvars, "Warning"],
		{
			{_List, constDeps_List} :>
				Replace[
					Catch[
						isAcyclic[ const, typ, shift, "Warning"][ constDeps];
						True
					],
					{
						True -> 15,
						_ -> 12
					}
				],
			_ -> 11
		}
	]

(* 'collectRelevantObjects[ typeDefF, lockedConsts, constDefs, const, typ, term]' scans 'term' and collects all
	types and constants occurring in it that are relevant for tracing dependencies in the definition
	'const::typ === term'. 'typ' is supposed to be the *short* type of 'const'.
	It returns '2' if some illegal object appears in 'term', and otherwise a pair '{typeDeps, constDeps}', where
	- 'typeDeps' is a list of elements of the form 'TYPE$[ h, i, unif]',
	- 'constDeps' is a list of constant instances.
	In fact, the same function can also be used for type definitions; there, the given 'const' must be 'Null', and
	'typ' must be a full type. *)
collectRelevantObjects[ typeDefF_, lockedConsts_, const_, typ_, term_] :=
	collectRelevantObjects[ typeDefF, lockedConsts, const, typ, term, DeleteDuplicates[ Cases[ typ, _TVAR$, {0, Infinity}, Heads -> True]]]
collectRelevantObjects[ typeDefF_, lockedConsts_, const_, typ_, term_, tvars_List] :=
	collectRelevantObjects[ typeDefF, lockedConsts, const, typ, term, tvars, "Fatal"]
collectRelevantObjects[ typeDefF_, lockedConsts_, const_, typ_, term_, tvars_List, msgType_] :=
	With[ {allConsts =
			DeleteDuplicates[
				Cases[ term,
					(* 'CoreAll' and 'CoreEqual' must be included because of their types. *)
					TYPED$[ _Symbol|_Integer|_Rational|_Real|_Complex|_DirectedInfinity, _List],
					{0, Infinity},
					Heads -> True
				]
			]},
	With[ {allTypes = DeleteDuplicates[ Join[ getFullTypeNoSorts /@ allConsts, Cases[ term, TYPRNG$[ _, t_] :> t, {0, Infinity}, Heads -> True]]]},
	With[ {typeDeps =
			isLegalTypeList[
				allTypes,
				True,
				typeDefF,
				MemberQ[ tvars, #]&,
				False&,
				If[ const === Null, sowMessage[ 1032, {typ, #}, "Fatal"]&, sowMessage[ 1034, {const, typ, #}, msgType]&],
				If[ const === Null, sowMessage[ 1033, {typ, #}, "Fatal"]&, sowMessage[ 1035, {const, typ, #}, msgType]&]
			]},
		If[ ListQ[ typeDeps],
			(* 'typeDeps' contains the instances of defined types occurring in 'term'. Each element is of the form
				'TYPE$[ h, i, unif]', where 'h' is the head of type 't', 'i' is the position of 't' in
				'typdeDefs[ h]', and 'unif' is the corresponding unifier (all LHSs in 'unif' have index >= 'max', and
				all type variables on RHSs have index < 'max').
				Note that not all defined types are listed in 'defTypes'; for instance, if 'K[K[a_]]' appears in
				'term', 'K[a_]' is not included in 'defTypes' since it appears under a non-built-in type.
				Moreover, if 'K[a_]' is defined solely in terms of built-in/declared types and built-in/locked
				constants (i.e. its constant-dependency-list is empty), it is not listed either since it cannot be part
				of a cyclic definition for sure. *)
			{typeDeps, DeleteCases[ allConsts, TYPED$[ CoreImplies$TM|CoreAll$TM|CoreEqual$TM|(_?(Lookup[ lockedConsts, #, $Failed] === $Failed &)), _]]},
		(*else*)
			2
		]
	]]]

(* Returns either '$Failed' or '0' (not composable) or '1' (cyclic) or a pair '{consts, check}', where
	- 'consts' is the list of constant instances 'term' (directly or indirectly) depends upon, which are themselves not
		defined (yet). Each element in the list is either a typed constant or of the form 'CONST$[ _TYPED$, cnt]'.
	- 'check' is a flat list of constant definitions that need to be checked additionally. Each element is of the form
		'DEF$[ c, i]', where 'c' is a plain constant and 'i' is the index of the respective definition in the list
		'constDefs[ c]'. *)
(* When called on a new definition, the value of 'ignore' must be empty! Only when checking old definitions again, its
	value may be non-empty. *)
checkSingleDependency[ thyCounter_, typeDefF_, constDefF_, fmls_, typeDeps_List, constDeps_List] :=
	checkSingleDependency[ thyCounter, typeDefF, constDefF, fmls, typeDeps, constDeps, Null, Null]
checkSingleDependency[ thyCounter_, typeDefF_, constDefF_, fmls_, typeDeps_List, constDeps_List, const_, typ_] :=
	Module[ {ignore = Association[]},
		checkSingleDependency[ thyCounter, typeDefF, constDefF, fmls, typeDeps, constDeps, const, typ, Hold[ ignore]]
	]
checkSingleDependency[ thyCounter_, typeDefF_, constDefF_, fmls_, typeDeps_List, constDeps_List, const_, typ_, ignore_Hold] :=
	checkSingleDependency[
		thyCounter, typeDefF, constDefF, fmls, typeDeps, constDeps, const, typ, ignore,
		getMaxTVar[ typ] + 1
	]
checkSingleDependency[ thyCounter_, typeDefF_, constDefF_, fmls_, typeDeps_List, constDeps_List, const_, typ_, Hold[ ignore_], shift_Integer] :=
	Module[ {visited},
	With[ {res =
			Catch[
				Module[ {consts =
							isAcyclic[ const, typ, shift, "Fatal"][
								Union[
									constDeps,
									DeleteDuplicates[ Flatten[ Replace[ typeDeps, TYPE$[ h_, i_Integer, subst_List] :> (typeDefF[ h][[i, 2, -1]] /. subst), {1}], 1]]
								]
							],	(* 'isAcyclic' does not return 'True' or 'False', but simply throws '1' if it detects a cycle. *)
						j, new},
					(* We cannot rely on 'maxDeps <= maxTyp'! *)
					visited[ _] = {};
					Replace[
						Reap[
							Last @ Reap[
								While[ consts =!= {},
									consts =
										With[ {fst = Replace[ First[ consts], c_TYPED$ :> CONST$[ c, 0]]},
										With[ {n = fst[[1, 1]], t = fst[[1, 2]]},
											If[ MemberQ[ visited[ n], t],
												(* Here we must compare for syntactic identity of the types, since even if one of them
													can be renamed to the other, one might lead to a cycle whereas the other does not.
													For instance, let 'typ = K1[ a_, K2[ b_], b_]' and assume we already processed the
													type 't1 = K1[ b_, b_, a_]' (which is not unifiable with 'typ', nor is it an instance
													theoreof, meaning there is no cycle). However, if we now consider
													't2 = K1[ a_, a_, b_]', then 'typ' and 't2' *are* unifiable (via '{a_ -> K2[ b_]}'),
													although 't2' can be obtained from 't1' via the renaming '{a_ -> b_, b_ -> a_}'. *)
												Rest[ consts],
											(*else*)
												PrependTo[ visited[ n], t];
												Union[
													isAcyclic[ const, typ, shift, "Fatal"][
														expandDepRel[
															thyCounter,
															constDefF,
															fmls,
															const,
															typ,
															Hold[ ignore, j, new],
															shift,
															fst
														]
													],
													Rest[ consts]
												]
											]
										]];
								],
								_DEF$,
								#1 &
							],
							True
						],
						{check_List, final_List} :>
							{
								Flatten[ final, 1],
								DeleteCases[ check, DEF$[ c_, i_] /; MemberQ[ Lookup[ ignore, c, {}], i]]
							}
					]
				]
			]},
		Clear[ visited];
		res
	]]

expandDepRel[ thyCounter_, constDefF_, fmls_, const_, typ_, Hold[ ignore_, j_, new_], shift_, CONST$[ TYPED$[ c_, t_], cnt_Integer]] :=
	(
		j = 0;
		new = Null;
		With[ {ignoreInd = Lookup[ ignore, c, {}]},
		With[ {ignoreNew =
				Position[
					constDefF[ c],
					(t0_ -> {k_, deps_}) /;
						If[ ++j; cnt =!= 0 && TrueQ[ First[ fmls[ k]] < cnt],	(* Must not be replaced by 'fmls[...] <= cnt'! ('K[a_] -> a_') *)
							False,
						(*else*)
							Replace[ unifyTypes[ shiftTypeVars[ t0, shift], t, None&, {None, {}}, 0],
								{
									{_, unif_List} :>
										Which[
											shift === 0 || AllTrue[ unif, #[[1, 1]] >= shift &],
											With[ {RHSs = unif[[All, 2]]},
												If[ MatchQ[ RHSs, {___TVAR$}] && DuplicateFreeQ[ RHSs],
													If[ MemberQ[ ignoreInd, j],
														False,
													(*else*)
														(* Because of orthogonality, 'new' can be set at most once. *)
														new = deps /. If[ unif === {} || shift === 0, unif, MapAt[ # - shift &, unif, {All, 1, 1}]];
														True
													],
												(*else*)
													new = deps /. If[ unif === {} || shift === 0, unif, MapAt[ # - shift &, unif, {All, 1, 1}]];
													False
												]
											],
										(*else*)	(* There does not exist a unifier that only instantiates type variables with index >= 'shift'. *)
											deps === {},
											(* If 'deps' is empty, we relax the requirement of composability; furthermore, there is no need to
												check 'DEF$[ c, j]'. *)
											False,
										(*else*)
											DuplicateFreeQ[
												Cases[ unif,
													(TVAR$[ _?(# >= shift &), ___] -> r_) :>
														If[ MatchQ[ r, _TVAR$], r, sowMessage[ 1071, {const, typ, unif}, "Error"]; Throw[ 0]]
												]
											],
											(* There exists a unifier that only instantiates type variables with index < 'shift'. *)
											If[ !MemberQ[ ignoreInd, j], Sow[ Null, DEF$[ c, j]]];
											False,
										(*else*)
											True,
											sowMessage[ 1071, {const, typ, unif}, "Error"];
											Throw[ 0]
										],
									_ -> False
								}
							]
						],
					{1},
					Heads -> False
				][[All, 1]]},
			If[ ignoreNew =!= {}, AssociateTo[ ignore, c -> Join[ ignoreInd, ignoreNew]]];
			If[ ListQ[ new],
				new,
			(*else*)
				If[ ignoreNew === {} === ignoreInd,
					Sow[ CONST$[ TYPED$[ c, t], thyCounter], True],
				(*else*)
					Sow[ TYPED$[ c, t], True]
				];
				{}
			]
		]]
	)
expandDepRel[ ___] :=
	Throw[ $Failed]

isAcyclic[ Null, _, _, _][ arg_] :=
	arg
isAcyclic[ const_, typ_, shift_, msgType_][ arg_List] :=
	(Scan[ isAcyclic[ const, typ, shift, msgType], arg]; arg)
isAcyclic[ const_, typ_, shift_, msgType_][ CONST$[ ci_, _]] :=
	isAcyclic[ const, typ, shift, msgType][ ci]
isAcyclic[ const_, typ1_, shift_, msgType_][ arg:TYPED$[ const_, typ2_]] :=
	Replace[ unifyTypes[ typ1, shiftTypeVars[ typ2, shift], None&, {None, {}}, 0],
		{
			{_, unif_List} :>
				Which[
					AllTrue[ unif, #[[1, 1]] < shift &],
					sowMessage[ 1072, {const, typ1, unif}, msgType];
					Throw[ 1],
				(*else*)	(* There does not exist a unifier that only instantiates type variables with index < 'shift'. *)
					shift === 0 ||
						DuplicateFreeQ[
							Cases[ unif, (TVAR$[ _?(# < shift &), ___] -> r_) :> If[ MatchQ[ r, _TVAR$], r, sowMessage[ 1071, {const, typ1, unif}, msgType]; Throw[ 0]]]
						],
					(* There exists a unifier that only instantiates type variables with index >= 'shift'. *)
					arg,
				(*else*)
					True,
					sowMessage[ 1071, {const, typ1, unif}, msgType];
					Throw[ 0]
				],
			_ -> arg
		}
	]
isAcyclic[ _, _, _, _][ arg_] :=
	arg


(* ::Subsection:: *)
(*Theories*)

(* 'accumulateTheories[ next, thys, thyDeps]' returns '$Failed' if it detects a cycle involving the first element of
	'thys', but it does not detect other cycles (but it never loops). *)
accumulateTheories[ {}, thys_List, _] :=
	thys
accumulateTheories[ {t_, rest___}, thys_List, thyDeps_] :=
	With[ {res = Lookup[ thyDeps, t, $Failed]},
		Which[
			!ListQ[ res],
			res,
		(*else*)
			res =!= {} && MatchQ[ thys, {Alternatives @@ res, ___}],
			$Failed,
		(*else*)
			True,
			accumulateTheories[ Join[ {rest}, DeleteCases[ res, Alternatives @@ thys | Alternatives[ rest]]], Append[ thys, t], thyDeps]
		]
	]


(* ::Subsection:: *)
(*Other*)

getPositions[ STERM$[ expr_, _Association, _Association], patt_] :=
	With[ {p = Position[ expr, patt]},
		Replace[ GatherBy[ MapThread[ Rule, {Prepend[ #, 1]& /@ p, Extract[ expr, p]}], Last], all:{_ -> e_, ___} :> (all[[All, 1]] -> e), {1}]
	]
getPositions[ expr_, patt_] :=
	With[ {p = Position[ expr, patt]},
		Replace[ GatherBy[ MapThread[ Rule, {p, Extract[ expr, p]}], Last], all:{_ -> e_, ___} :> (all[[All, 1]] -> e), {1}]
	]

(* 'makeTypeVarPositions' can be used to create the "TypeVarPositions" entry of the 'FML$'-structure.
	ATTENTION! If the given term is an 'STERM$'-structure, the resulting positions refer to the *plain* term, not to
	the whole structure! Be careful when using this function in connection with 'specializeTypes' etc. *)
makeTypeVarPositions[ STERM$[ term_, _, tvar_Association]] :=
	makeTypeVarPositions[ term, tvar]
makeTypeVarPositions[ term_] :=
	makeTypeVarPositions[ term, Association[]]
makeTypeVarPositions[ term_, tvar_Association] :=
	With[ {p = Position[ term, _TVAR$]},
		Replace[ GatherBy[ Thread[ p -> Extract[ term, p]], Last], all:{_ -> v_, ___} :> (all[[All, 1]] -> Lookup[ tvar, v, getDefaultSort[]]), {1}]
	]

getKeyLabelPos[ coll_List, k:{_, src_String}] :=
	With[ {i = First[ FirstPosition[ coll, {{{_, src}, _}, ___}, {0}, {1}, Heads -> False]]},
	With[ {j = First[ FirstPosition[ coll[[i]], {k, _}, {0}, {1}, Heads -> False]]},
		{i, j}
	]]

checkProof[ goal_, prf_Hold] :=
	checkProof[ False, Hold[ Null], goal, prf]
checkProof[ thyMode_, thmInsts_Hold, goal_, prf_Hold] :=
	checkProof[ thyMode, thmInsts, goal, prf, 2]
checkProof[ thyMode_, Hold[ thmInsts_], goal:FML$[ k_, __], prf_Hold, prfMode_] :=
	Replace[ executeProofCommands[ prf, initiateProof[ goal, {}, prfMode], prfMode, True],
		{
			{PRFSTAT$[ _, open_List, usedKeys_List, usedThms_List, ___], trust_} :>
				(
					If[ TrueQ[ thyMode], thmInsts = Union[ thmInsts, usedThms]];
					Which[
						MemberQ[ usedKeys, k],
						2,
					(*else*)
						open =!= {},
						3,
					(*else*)
						trust,
						5,
					(*else*)
						True,
						4
					]
				),
			_ -> 1
		}
	]

checkTheoremInstsAux[ Hold[ thmInsts_], thyDistF_, scopeOnly_, discard_] :=
	Replace[ checkTheorems[ If[ TrueQ[ scopeOnly], Cases[ thmInsts, THM$[ {_, _?(Composition[ IntegerQ, thyDistF])}, ___]], thmInsts]],
		{
			{num_Integer, missing_List, type_List, proof_List, order_List, axiom_List} :>
				(
					If[ missing =!= {}, sowMessage[ 1075, missing, "Fatal"]];
					If[ type =!= {}, sowMessage[ 1076, type, "Fatal"]];
					If[ proof =!= {}, sowMessage[ 1077, proof, "Fatal"]];
					If[ order =!= {}, sowMessage[ 1078, order, "Fatal"]];
					If[ axiom =!= {}, sowMessage[ 1079, axiom, "Warning"]];
					If[ TrueQ[ discard],
						thmInsts =
							Join[
								If[ TrueQ[ scopeOnly], DeleteCases[ thmInsts, THM$[ {_, _?(Composition[ IntegerQ, thyDistF])}, ___]], {}],
								missing,
								type,
								proof,
								order,
								axiom
							]
					];
					If[ missing === {} === type === proof === order,
						num,
					(*else*)
						$Failed
					]
				),
			_ -> $Failed
		}
	]


(* ::Section:: *)
(*Locking Symbols*)

(*
Scan[
	SetAttributes[ #, {Protected, Locked}]&,
	{
		exportTheoryImageAux, importTheoryImageAux, restartSessionAux, addTypeDeclaration, addTypeDefinition,
		addDeclaration, addAxiom, addFactsByProof, disableTypeDefs, disableTypeOverloading, disableConstOverloading,
		addProof, remTypeDefinition, remDeclaration, remTheoremSchema, remFormula, enableTypeDefs,
		enableTypeOverloading, enableConstOverloading, getTypeDefsAux, prepareTypeForDecl, addTypeDefinitionAux,
		addTypeMorphism, remTypeDefinitionsAux, addAxiomAux, addFormula, checkDefinition, remFormulaAux,
		remConstDefinitionAux, remConstDefsAux, checkDependency, quickcheckDependency, collectRelevantObjects,
		checkSingleDependency, expandDepRel, isAcyclic, isLegalTypeAux, getKeyLabelPos, checkTheoremInstsAux,
		$protectedConsts
	}	(* TODO: Make sure that all private functions are listed here. *)
]
*)

End[]

EndPackage[]
