(* ::Package:: *)

(* Theorema 
    Copyright (C) 2010 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Core`Trusted`Reasoning`"]

Needs[ "Theorema`Language`"]	(* only to have symbols in right context *)
Needs[ "Theorema`Common`"]		(* only to have symbols in right context *)
Needs[ "Theorema`Core`Trusted`"]
Needs[ "LazyList`", FileNameJoin[ {$TheoremaDirectory, "Theorema", "Components", "LazyList.m"}]]

Begin["`Private`"]

(* Input instruction sequence:
0. Transform the Theorema expression into a fully curried, fully short-type-annotated Core expression with de Bruijn indices.
	The object logic has to take care of this, e.g. in 'parseCoreTermPre'. The basic inferences can only handle such expressions.
1. Type-check the Core expression.
	This happens in 'checkType'.
2. Normalize the expression.
	This happens in 'normalize'; the correctness of 'normalize' relies on the fact that the expression has been type-checked already.
*)

(* Types and type annotations:
Types are formed according to the grammar

t ::= TVAR$[ Integer] | TVAR$[ Integer, Symbol] | TFIX$[ Symbol, Integer] | c | t[ t_1, ..., t_n]

Apparently, type variables are characterized by an index and, optionally, a name, whereas a.b.f. type constants are characterized
either by a symbol and, optionally, an index, or by an index only. All components of a type variable/constant identify that object,
i.e. 'TVAR$[ 1, a]' is different from 'TVAR$[ 1, b]' is different from 'TVAR$[ 2, b]'.
Whenever a new expression enters a proof situation, its types are automatically lifted into the current proof context; for instance,
this means that the indices of the type variables are adjusted and that some a.b.f. constants are turned into variables.

Type annotations are of the form 'TYPED$[ expr, type]', where 'expr' is an arbitrary expression, and where 'type' is either a full
type w.r.t. to the above grammar, or a list of such full types. In the latter case, 'expr' must be a declared constant whose
declared type involves *exactly* as many type variables as the length of 'type'; the full type of 'expr' is then just the instance of
the declared type under the instance-list 'type'.
*)

(* Proof situations (PRFSIT$):
Proof situations are of the form

PRFSIT$[ G, hyps, idx, fix, meta, tfix, tvar, fid, ___]

where
- 'G' is a list of Core expressions of type 'BOOL' constituting the subgoals,
- 'hyps' is a list of Theorema formulas constituting the hypotheses,
- 'idx' is an integer specifying the type variable index, i.e. all type variables in 'G' have index < 'idx',
- 'fix' is a list of rules, assigning to each a.b.f. term constant appearing in 'G' or 'hyps' its type,
- 'meta' is a list of rules, assigning to each meta variable that appears in 'G' or 'hyps' its type and an integer specifying the
	elements from the list 'fix' it may depend on, and
- 'tfix' is a list of rules, assigning to *each* a.b.f. type constant that appeared in the proof so far the its sort constraint,
- 'tvar' is an association, mapping each type variable that appears in 'G' to its sort constraint.
	ATTENTION! Not all type variables might appear in 'tvar'; if some variable does not occur, it simply means that it does not
	have any sort constraints.
- 'fid' is an integer specifying the number of automatically generated formula-keys so far (not critical for the correctness).

The meaning of term meta variables 'm' and type variables 'a_', 'b_', 'c_' in a proof situation 'H_1,...,H_n |- G' can be made
explicit by informally introducing quantifiers like 'EX c_. EX m. ((ALL a_. H_1(a_)),...,(ALL b_. H_n(b_)) ==> G(c_,m))'. Hence,
type variables are not shared among hypotheses and subgoals, and term meta variables may not appear in the hypotheses at all.
*)

(* Lists of proof situations (PRFSTAT$):
Lists of proof situations are stored as 'PRFSTAT$', of the form

PRFSTAT$[ id, sits, keys, thms, (obj)]

where
- 'id' is a "branch-identifier" that is used for keeping track of alternative branches (not critical!),
- 'sits' is a *list* referring to all open proof situations ('PRFSIT$'), possibly empty,
- 'keys' is a list of formula keys (possibly empty),
- 'thms' is a list of theorems,
- 'obj' is an optional proof object; if it is present, 'sits' is the list of positions of open proof situations in 'obj', otherwise it
	is the list of open proof situations themselves.

The original goal follows from the *conjunction* of all proof situations in 'sits', under the additional assumptions corresponding to
'keys' and 'thms'.
*)

(* Remarks about regular types:
- Regular types may introduce hidden polymorphism, e.g. with 'c::(a___ -> BOOL) === \x::(a___). isUnit[ undefined::a_]':
	'c[]' is of type 'BOOL', but unfolding the equality gives an expression that involves the type variable 'a_'.
	-> The ConsistencyChecker should reject such definitions and require every type variable on the RHS to appear on the LHS
	*outside* of "..." and "+".
- Automatic currying of expressions in default internal representation, as well as type inference, could be difficult/impossible.
	Example: plain 'f[ a, b]' could be interpreted as
	 1. 'f[ a][ b]' with 'f :: a_ -> b_ -> c_'
	 2. 'f[ a][ b]' with 'f :: a___ -> b_ -> c_'
	 3. 'f[ a][ b]' with 'f :: a_ -> b___ -> c_'
	 4. 'f[ a][ b]' with 'f :: a___ -> b___ -> c_'
	 5. 'f[ a, b]' with 'f :: a_ * b_ -> c_'
	 6. 'f[ a, b]' with 'f :: a___ -> c_'
	 7. 'f[ a, b]' with 'f :: (a_ * b_)... -> c'
	 8. 'f[ a, b]' with 'f :: (a___ * b_)... -> c'
	 9. 'f[ a, b]' with 'f :: (a_ * b___)... -> c'
	10. 'f[ a, b]' with 'f :: (a___ * b___)... -> c'
	Possible solution: Only introduce one unranked flat symbol, e.g. 'Sequence$TM', that must be used to construct sequences of
	terms. 'Sequence$TM' shall be flat only in the sense that nested 'Sequence$TM's shall be flattened out, but if 'Sequence$TM'
	appears as argument of another symbol, no flattening should happen. 'Sequence$TM' may only appear as the head of an
	expression, not as a symbol on its own; the type of 'Sequence$TM[ ...]' is then 'a___' (provided the types of all arguments
	can be unified).
*)

(* Proof methods:
Proof methods are (parameterized) functions that take as input a proof situation ('PRFSIT$'), an integer indicating the total
number of open proof situations, and a number in the range 0-2 indicating whether the proof is done in fast mode without tracing
(0), in fast mode with tracing enabled (1), or in kernel mode (without tracing; 2).
They return
- a 4-tuple consisting of a new proof situation or a flat list thereof, a list of keys, a list of theorems, and a list of new
	proof commands (mode '0'), or
- a 4-tuple consisting of a new proof situation or a proof node, a list of keys, a list of theorems, and a list of proof commands
	(mode '1'), or
- a list of new proof commands together with parameters (any mode), or
- a *lazy* list thereof (any mode).
In any mode, an empty lazy list indicates failure.

In any case, lists of proof commands have head 'Hold' rather than 'List'!
*)

(* Known leaks:
- Proof methods might find out the name of the variable holding the local environment (e.g. using 'Names'), and modify it.
- Theorems provided by the object logic might depend on global variables, that by default have a value that makes the theorems
	indeed provable, but that might be set to some other value (e.g. by proof methods) that renders the "theorems" invalid.
	The same is true for formulas in the environment.
*)


(* ::Section:: *)
(*Parameters*)

(* The parameters in this section have an influence on the operational behavior of the inference rules and tacticals,
	but they do not affect their correctness.
	Names of parameters MUST start with "$"! *)

If[ !ValueQ[ $autoNormalize], $autoNormalize = True];

If[ !ValueQ[ $branchCombinator], $branchCombinator = 1];

If[ !ValueQ[ $instantiateTypes], $instantiateTypes = True];


(* ::Section:: *)
(*Rule Application & Proof Checking*)

(* 'checkTheorems[ thms]' checks whether the given list of theorems are all either well-formed, proved theorems, or well-formed axioms.
	It returns a 6-tuple '{number, missing, type, proof, order, axiom}' where
	- 'number' is the total number of theorems that have been checked,
	- 'missing' is the list of all theorems that do not exist,
	- 'type' is the list of all theorems that are either ill-typed or whose type-variable-positions are wrong,
	- 'proof' is the list of all well-formed theorems that have an incorrect proof or no proof at all,
	- 'order' is the list of all well-formed, proved theorems that depend on other theorems with an invalid index, and
	- 'axiom' is the list of all unproved theorems, i.e. all theorems whose proofs are empty.
	Note that all theorems appearing in the proof of another theorem are automatically checked themselves, too! *)
checkTheorems[ thms_List] :=
	Join @@@ FlattenAt[ Reap[ checkTheorems[ Replace[ thms, t:{_, _String} :> THM$[ t], {1}], {}], {"missing", "type", "proof", "order", "axiom"}], {2}]
checkTheorems[ {}, done_List] := Length[ done]
checkTheorems[ {thm:THM$[ k_, args___], rest___}, done_List] :=
	checkTheorems[
		Join[
			{rest},
			Replace[ getThmEnv[ True][ k],
				{_, stats_, prfs_, indices_} :>
					Replace[ stats[ args],
						{
							{stat_, tvp_List} :>
								With[ {tvp2 = Map[ (First[ #] -> extractSingle[ stat, #[[1, 1]]])&, tvp]},	(* 'tvp' maps positions to sorts, 'tvp2' maps positions to variables *)
								With[ {stat1 = First[ extractFormula[ stat, tvp, 0]]},
									Replace[ checkType[ stat1, Blank[], {}, {}, {}, Association[], Association[]],
										{
											{FORM$TM, {}, _Integer, Association[], tvar_Association} :>
												With[ {tvp0 = getPositions[ stat, _TVAR$]},
													If[ tvp0[[All, 1]] === tvp[[All, 1]] &&	(* here we might not have syntactic identity, but only equality as sets of sets *)
															And @@ MapThread[ Lookup[ tvar, extractSingle[ First[ stat1], #1[[1, 1]]], defaultSort[]] === Last[ #2] &, {tvp0, tvp}],
														Replace[ prfs[ args],
															{
																Hold[] :> (Sow[ thm, "axiom"]; {}),
																prf_Hold :>
																	Replace[ executeProofCommands[ prf, initiateThmProof[ ReplacePart[ stat1, 3 -> tvar], tvp2], 2, True],
																		{
																			{PRFSTAT$[ _, {}, {}, thms_List, ___], True} :>
																				With[ {index = indices[ args]},
																					If[ MatchQ[ index, _$Failed],
																						Sow[ thm, "order"],
																					(*else*)
																						Catch[
																							Scan[
																								If[ compareThmIndices[ Last[ getThmEnv[ False][ First[ #]]] @@ Rest[ #], index] =!= Less,
																									Throw[ Sow[ thm, "order"]]
																								]&,
																								thms
																							]
																						]
																					];
																					DeleteCases[ thms, Alternatives @@ done]
																				],
																			_ :> (Sow[ thm, "proof"]; {})
																		}
																	],
																_ :> (Sow[ thm, "proof"]; {})
															}
														],
													(*else*)
														Sow[ thm, "type"];
														{}
													]
												],
											_ :> (Sow[ thm, "type"]; {})
										}
									]
								]],
							_ :> (Sow[ thm, "missing"]; {})
						}
					]
			]
		],
		Append[ done, thm]
	]
checkTheorems[ {x_, rest___}, done_List] :=
	(
		Sow[ x, "missing"];
		checkTheorems[
			{rest},
			Append[ done, x]
		]
	)

(* 'executeProofCommands[ cmd, {state, trust}, mode, True]', in case of success, returns a pair '{s, t}', where
	's' is
	- a single proof state with a proof object ('PRFOBJ$') or '$Failed', if 'mode' is '1', or
	- a single proof state with a flat list of proof situations or '$Failed', if 'mode' is different from '1',
	and 't' is a value indicating the "trust" one can put into the proof: 'True' means that no subgoal has been
	skipped, and 'False' means that at least one subgoal has been skipped. Additionally, the function sows a list
	of messages of the form 'i -> msgList', where 'i' is the index of the element of 'cmd' that triggered the
	messages and 'msgList' is a list of messages of the form '{tag, msg}', where 'tag' is a tag (like "Error",
	"Warning" etc.) and 'msg' is the actual message.
	'executeProofCommands[ cmd, state, mode, False]' does basically the same, but only returns a state without
	'trust' and does not reap/sow any messages (i.e. all messages sown in subroutines are passed to the calling
	function just as they are).
	Further definitions are given below. *)
executeProofCommands[ cmd_Hold, ps:{___PRFSIT$}, mode_, toplevel_] :=
	executeProofCommands[ cmd, PRFSTAT$[ $msgID, ps, {}, {}], mode, toplevel]
executeProofCommands[ _Hold, state:Except[ _PRFSTAT$], _, False] :=
	state
executeProofCommands[ cmd_Hold, state_PRFSTAT$, mode_, True] :=
	executeProofCommands[ cmd, {state, True}, mode, True]
(* 'executeProofCommands' is implemented in an iterative way rather than a recursive one. This is because otherwise
	even moderatly-sized proofs would make the internal stack overflow, and thus Mathematica crash. *)
executeProofCommands[ cmd_Hold, {state_PRFSTAT$, trust_}, mode_, True] :=
	Block[ {$IterationLimit = Infinity},
	Module[ {i = 0, s = state, t = trust},
		Catch[
			Scan[
				(
					s = executeProofCommands[ #, s, mode, Hold[ t], ++i]
				)&,
				cmd
			];
			{s, t}
		]
	]]
executeProofCommands[ _Hold, state_, _, True] :=
	state
(* "skip" *)
executeProofCommands[ INF$[ "skip", ___], PRFSTAT$[ id_, {f_List, r___}, keys_, thms_, po_PRFOBJ$], 1, Hold[ t_], i_Integer] :=
	(t = False; PRFSTAT$[ id, {r}, keys, thms, MapAt[ toStandardPRFSIT, po, f]])
executeProofCommands[ INF$[ "skip1", ___], PRFSTAT$[ id_, {f_List, r___}, keys_, thms_, po_PRFOBJ$], 1, Hold[ t_], i_Integer] :=
	With[ {ps = Extract[ po, f]},
		If[ MatchQ[ ps, PRFSIT$[ {__}, __]],
			t = False;
			PRFSTAT$[ id, {f, r}, keys, thms, Delete[ po, Join[ f, {1, 1}]]],
		(*else*)
			Sow[ i -> {MSG$[ coreMessageToList, 9, {}, "Fatal"]}, "Message"];
			Throw[ $Failed]
		]
	]
executeProofCommands[ INF$[ "skip", ___], PRFSTAT$[ id_, {PRFSIT$[ _, hyps_List, __], r___}, keys_List, po__], _Integer, Hold[ t_], _Integer] :=
	With[ {defSource = sourceLabel[]},
		t = False;
		(* although the subgoal is skipped, we nevertheless collect its hypotheses *)
		PRFSTAT$[ id, {r}, Union[ keys, DeleteCases[ key /@ hyps, {_, defSource}]], po]
	]
executeProofCommands[ INF$[ "skip1", ___], PRFSTAT$[ id_, {PRFSIT$[ {_, rem___}, ctxt__], r___}, po__], _Integer, Hold[ t_], _Integer] :=
	(t = False; PRFSTAT$[ id, {PRFSIT$[ {rem}, ctxt], r}, po])
executeProofCommands[ INF$[ "skip"|"skip1", ___], _, _, _Hold, i_Integer] :=
	(
		Sow[ i -> {MSG$[ coreMessageToList, 9, {}, "Fatal"]}, "Message"];
		Throw[ $Failed]
	)

(* We define the six functions 'executeProofCommands0', 'executeProofCommands1', 'executeProofCommands2', 'executeProofCommands0op',
	'executeProofCommands1op' and 'executeProofCommands2op' that, depending on the number of arguments, either apply a list of proof
	commands to a proof state and return a lazy list of pairs of commands and states (2 arguments), or take a lazy list of such pairs as
	input and return a lazy list of states only (1 argument).
	The suffixes '0', '1' etc. indicate the mode the function operates in. The suffix 'op' indicates that the function is merely an
	auxiliary function, namely an operator form of the corresponding function without suffix 'op'.
	The reason for putting the 'mode'-information into the names of the functions instead of additional arguments is because the
	information does not change in recursive calls, and since when working with lazy lists it is desirable to keep the argument lists of
	functions as small as possible (without any non-changing arguments). *)

(* operator forms *)
executeProofCommands0op[ cmd2_Hold, PRFSTAT$[ id_, goals2_List, keys2_List, thms2_List, po___]][ {goals1_List, keys1_List, thms1_List, cmd1_Hold}, i_] :=
	{Join[ cmd1, cmd2], PRFSTAT$[ Append[ id, i], Join[ goals1, Rest[ goals2]], Union[ keys1, keys2], Union[ thms1, thms2], po]}
executeProofCommands0op[ cmd2_Hold, PRFSTAT$[ id_, goals2_List, keys2_List, thms2_List, po___]][ {goals1_List, keys1_List, thms1_List, cmd1_Hold}] :=
	{Join[ cmd1, cmd2], PRFSTAT$[ id, Join[ goals1, Rest[ goals2]], Union[ keys1, keys2], Union[ thms1, thms2], po]}
executeProofCommands0op[ cmd2_Hold, PRFSTAT$[ id_, goals2_List, keys2_List, thms2_List, po___]][ {ps_PRFSIT$, keys1_List, thms1_List, cmd1_Hold}, i_] :=
	{Join[ cmd1, cmd2], PRFSTAT$[ Append[ id, i], ReplacePart[ goals2, 1 -> ps], Union[ keys1, keys2], Union[ thms1, thms2], po]}
executeProofCommands0op[ cmd2_Hold, PRFSTAT$[ id_, goals2_List, keys2_List, thms2_List, po___]][ {ps_PRFSIT$, keys1_List, thms1_List, cmd1_Hold}] :=
	{Join[ cmd1, cmd2], PRFSTAT$[ id, ReplacePart[ goals2, 1 -> ps], Union[ keys1, keys2], Union[ thms1, thms2], po]}

executeProofCommands1op[ cmd2_Hold, PRFSTAT$[ id_, pos:{f_, ___}, keys2_List, thms2_List, po_PRFOBJ$]][ {ps_PRFSIT$, keys1_List, thms1_List, cmd1_Hold}, i_] :=
	{
		Join[ cmd1, cmd2],
		PRFSTAT$[ Append[ id, i], pos, Union[ keys1, keys2], Union[ thms1, thms2], ReplacePart[ po, f -> ps]]
	}
executeProofCommands1op[ cmd2_Hold, PRFSTAT$[ id_, {f_List, r___}, keys2_List, thms2_List, po_PRFOBJ$]][ {node_?isProofNode, keys1_List, thms1_List, cmd1_Hold}, i_] :=
	{
		Join[ cmd1, cmd2],
		PRFSTAT$[ Append[ id, i], Join[ getNewRelevantSits[ node, f], {r}], Union[ keys1, keys2], Union[ thms1, thms2], propagateProofValues[ ReplacePart[ po, f -> node]]]
	}
executeProofCommands1op[ cmd2_Hold, PRFSTAT$[ id_, pos:{f_, ___}, keys2_List, thms2_List, po_PRFOBJ$]][ {ps_PRFSIT$, keys1_List, thms1_List, cmd1_Hold}] :=
	{
		Join[ cmd1, cmd2],
		PRFSTAT$[ id, pos, Union[ keys1, keys2], Union[ thms1, thms2], ReplacePart[ po, f -> ps]]
	}
executeProofCommands1op[ cmd2_Hold, PRFSTAT$[ id_, {f_List, r___}, keys2_List, thms2_List, po_PRFOBJ$]][ {node_?isProofNode, keys1_List, thms1_List, cmd1_Hold}] :=
	{
		Join[ cmd1, cmd2],
		PRFSTAT$[ id, Join[ getNewRelevantSits[ node, f], {r}], Union[ keys1, keys2], Union[ thms1, thms2], propagateProofValues[ ReplacePart[ po, f -> node]]]
	}

With[ {basic = "psData"|"insert"|"hypData"|"makeExplicit"|"makeImplicit"|"drop"|"split"|"rearrangeG"|"cut"|
				"meta"|"inst"|"sorts"|"instT"|"allE"|"allI"|"forgetFix"|"impE"|"refl"|"alphabetaeta"|"replace"|"ext"|"eqI",
		struct = "if"|"close"|"theorem"|"formula"|"forget"|"rearrangePS"},
	Scan[
		Function[ mode,
			With[ {name = "executeProofCommands" <> ToString[ mode]},
			With[ {epc = Symbol[ name], epcOp = Symbol[ name <> "op"]},
				(* main function *)
				executeProofCommands[ cmd_, state_, mode, _Hold, i_Integer] :=
					Replace[ Reap[ executeProofCommands[ Hold[ cmd], state, mode, False], "Message"],
						{
							{ps:PRFSTAT$[ id_, __], msg_List} :>
								(
									With[ {il = Length[ id]},
									With[ {msg1 = Cases[ msg, {_List?(With[ {l = Length[ #]}, If[ l <= il, Take[ id, l] === #, Take[ #, il] === id]]&), m_MSG$} :> m, {2}]},
										If[ msg1 =!= {}, Sow[ i -> trimMessageList[ msg1], "Message"]]
									]];
									ps
								),
							{_, msg_List} :>
								(
									With[ {msg1 = Cases[ msg, {_List, m_MSG$} :> m, {2}]},
										If[ msg1 === {},
											Sow[ i -> {MSG$[ coreMessageToList, 10]}, "Message"],
										(*else*)
											Sow[ i -> trimMessageList[ msg1], "Message"]
										]
									];
									Throw[ $Failed]
								)
						}
					];
				executeProofCommands[ cmd_Hold, state_PRFSTAT$, mode, False] :=
					Replace[ Reap[ lFirst[ epc[ LL[ {cmd, state}, LL[]]]], "GC"],
						{res_, gc_List} :> (Scan[ ReleaseHold, Join @@ gc]; res)
					];
					
				epc[ branches0_LL] :=
					Module[ {branches = branches0},
						Catch[
							While[ True,
								branches =
									Replace[ branches,
										{
											LL[ {Hold[], state_PRFSTAT$}, b_] :>
												Throw[ LL[ state, epc[ b]]],
											LL[ {Hold[ "commit", rest___], state_PRFSTAT$}, _] :>
												LL[ {Hold[ rest], state}, LL[]],
												(* Attention! If "commit" is applied and the current branch fails, the whole toplevel proof command fails.
													This is because the alternative branches are stored in a flat lazy list, not in a structured tree.
													'andTac[ .., "!", ...]' achieves local commits by employing "if". *)
											LL[ {cmd_Hold, state_PRFSTAT$}, b_] :>
												Switch[ $branchCombinator, 1, lJoin, 2, lInterleave, _, lMerge[ $branchCombinator]][ epc[ cmd, state], b],
											LL[ _, b_] :> b,
											_ :> Throw[ LL[]]
										}
									]
							]
						]
					];
	
				(* operator forms *)
				epcOp[ cmd2_Hold, stat_PRFSTAT$][ cmd1_Hold, i_] := {Join[ cmd1, cmd2], Insert[ stat, i, {1, -1}]};
				epcOp[ cmd2_Hold, stat_PRFSTAT$][ cmd1_Hold] := {Join[ cmd1, cmd2], stat};
				epcOp[ _, _][ __] := $Failed;
				epcOp[ cmd_Hold][ stat_PRFSTAT$, i_] := {cmd, Insert[ stat, i, {1, -1}]};	(* needed for "if" *)
				epcOp[ _][ _, _] := $Failed;
				
				epc[ Hold[ method_String, ___], PRFSTAT$[ id_, __]] :=
						(
							$msgID = id;
							sowMessage[ 1, {method}, "Fatal"];
							LL[]
						);
				epc[ Hold[ INF$[ "if", cond_, true_, false_, discard___], rest___], state_] :=
					With[ {new =
							If[ {discard} === {False},
								epc[ LL[ {Hold[ cond], state}, LL[]]],								(* keep messages *)
							(*else*)
								First @ Reap[ epc[ LL[ {Hold[ cond], state}, LL[]]], "Message"]		(* discard messages *)
							]},	(* 'new' is a lazy list of proof states *)
						If[ lEmptyQ[ new],
							LL[ {Hold[ false, rest], state}, LL[]],
						(*else*)
							lMapIndexed[ epcOp[ Hold[ true, rest]], new]
						]
					];
				epc[ Hold[ INF$[ "rearrangePS", i_Integer, j_Integer], rest___], PRFSTAT$[ id_, ps_List, data__]] :=
					With[ {len = Length[ ps]},
					With[ {i0 = If[ Negative[ i], len + 1 + i, i], j0 = If[ Negative[ j], len + 1 + j, j]},
						If[ 1 <= i0 <= len && 1 <= j0 <= len,
							LL[ {Hold[ rest], PRFSTAT$[ id, Insert[ Delete[ ps, i], ps[[i]], j], data]}, LL[]],
						(*else*)
							LL[]
						]
					]];
				
				If[ mode =!= 1,
					(* fast- and kernel mode only; see below for definitions for trace mode *)
					epc[ Hold[ INF$[ "close"], rest___], PRFSTAT$[ id_, {PRFSIT$[ {}, hyps_List, __], r___}, keys_List, thms_]] :=
						With[ {defSource = sourceLabel[]},
							(* The keys of hypotheses are added to 'keys' here instead of "insert", since otherwise the keys would be stored twice (in 'hyps' and in 'keys'). *)
							LL @@ {{Hold[ rest], PRFSTAT$[ id, {r}, Union[ keys, DeleteCases[ key /@ hyps, {_, defSource}]], thms]}, LL[]}
						];
					epc[ Hold[ INF$[ "close"], ___], PRFSTAT$[ id_, __]] :=
						(
							$msgID = id;
							sowMessage[ 2, {}, "Fatal"];
							LL[]
						);
					epc[ Hold[ INF$[ "theorem", thm:THM$[ k_, args___]], rest___], PRFSTAT$[ id_, {ps:PRFSIT$[ {__}, __], r___}, keys_List, thms_List]] :=
						(
							$msgID = id;
							Replace[ getThmEnv[ True][ k][[2]][ args],
								{
									{f_, tvp_List} :>
										LL @@ {
											{
												Hold[ rest],
												PRFSTAT$[ id, {addAssmToProofSit[ f, tvp, ps], r}, keys, If[ MemberQ[ thms, thm], thms, Append[ thms, thm]]]
											},
											LL[]
										},
									_$Failed :> (sowMessage[ 3, {thm}, "Fatal"]; LL[]),
									err_ :> (sowMessage[ 8, {"theorem", {thm}, ToString[ err]}, "Fatal"]; LL[])
								}
							]
						);
					epc[ Hold[ INF$[ "formula", k_], rest___], PRFSTAT$[ id_, {ps:PRFSIT$[ {__}, __], r___}, keys_List, thms_List]] :=
						(
							$msgID = id;
							Replace[ getFmlEnv[ True][ k][[2]],
								{
									FML$[ _, f_, __, "TypeVarPositions" -> tvp_List, ___] :>
										LL @@ {
											{
												Hold[ rest],
												PRFSTAT$[ id, {addAssmToProofSit[ f, tvp, ps], r}, If[ MemberQ[ keys, k], keys, Append[ keys, k]], thms]
											},
											LL[]
										},
									_ :> (sowMessage[ 3, {k}, "Fatal"]; LL[])
								}
							]
						);
					epc[ Hold[ INF$[ "theorem"|"formula", args___], ___], PRFSTAT$[ id_, {PRFSIT$[ {}, __], ___}, __]] :=
						(
							$msgID = id;
							sowMessage[ 4, {args}, "Fatal"];
							LL[]
						);
					epc[ Hold[ INF$[ "forget", i_Integer?Positive], rest___], PRFSTAT$[ id_, {PRFSIT$[ g_List, hyps_List, ctxt__], r___}, keys_List, thms_]] /; i <= Length[ hyps] :=
						With[ {defSource = sourceLabel[]},
							LL @@ {{Hold[ rest], PRFSTAT$[ id, {PRFSIT$[ g, Delete[ hyps, i], ctxt], r}, Union[ keys, DeleteCases[ {key[ hyps[[i]]]}, {_, defSource}]], thms]}, LL[]}
						];
					epc[ Hold[ rl:INF$[ method_String, params___], rest___], state:PRFSTAT$[ id_, goals:{f_, r___}, keys_, thms_]] :=
						(
							$msgID = id;
							Switch[ method,
								basic,
								With[ {rule = rl},	(* evaluation *)
								With[ {new = basicInference[ rule, f]},
									Switch[ new,
										_List,
										LL @@ {{Hold[ rest], PRFSTAT$[ id, Join[ new, {r}], keys, thms]}, LL[]},
										_String,
										sowMessage[ 8, {method, params, new}, "Fatal"];
										LL[],
										_,
										sowMessage[ 7, {method, params}, "Fatal"];
										LL[]
									]
								]],
								
								struct,
								sowMessage[ 7, {method, params}, "Fatal"];
								LL[],
								
								_,
								sowMessage[ 4, {method}, "Fatal"];
								LL[]
							]
						);
					epc[ Hold[ method_, rest___], state:(PRFSTAT$[ id_, ps:{f_, ___}, __])] :=
						With[ {new = ($msgID = id; method[ f, Length[ ps], mode])},
							If[ MatchQ[ new, _LL],
								lMapIndexed[ epcOp[ Hold[ rest], state], new],
							(*else*)
								LL @@ {epcOp[ Hold[ rest], state][ new], LL[]}
							]
						];
					epc[ _, _] := LL[]
				]
			]]
		],
		{0, 1, 2}
	];
	
	executeProofCommands1[ Hold[ INF$[ "close"], rest___], PRFSTAT$[ id_, {f_List, r___}, keys_, thms_, po_PRFOBJ$]] :=
		Switch[ Extract[ po, f],
			PRFSIT$[ {}, _List, __],
			(* There is no need to update the keys, because normally formulas that have been used at some point are stored in the corresponding PRFINFO$ anyway,
				and after all this is only trace mode ... *)
			LL @@ {{Hold[ rest], PRFSTAT$[ id, {r}, keys, thms, ReplacePart[ po, f -> makeTERMINALNODE[ makePRFINFO[ name -> "close"], proved]]]}, LL[]},
			
			Theorema`Provers`Common`Private`TERMINALNODE$[ _, proved],
			LL[ {Hold[ rest], PRFSTAT$[ id, {r}, keys, thms, po]}, LL[]],
			
			_,
			$msgID = id;
			sowMessage[ 2, {}, "Fatal"];
			LL[]
		];
	executeProofCommands1[ Hold[ INF$[ cmd:"theorem"|"formula", spec_], rest___], PRFSTAT$[ id_, pos:{fst_List, ___}, keys_List, thms_List, po_PRFOBJ$]] :=
		(
			$msgID = id;
			Replace[ Extract[ po, fst],
				{
					PRFSIT$[ {}, __] :> (sowMessage[ 4, {spec}, "Fatal"]; LL[]),
					ps_PRFSIT$ :>
						If[ cmd === "theorem",
							Replace[ getThmEnv[ True][ First[ spec]][[2]] @@ Rest[ spec],
								{
									{thm_, tvp_List} :>
										LL @@ {
											{
												Hold[ rest],
												PRFSTAT$[ id,
													pos,
													keys,
													If[ MemberQ[ thms, spec], thms, Append[ thms, spec]],
													ReplacePart[ po, fst -> addAssmToProofSit[ thm, tvp, ps]]
												]
											},
											LL[]
										},
									_$Failed :> (sowMessage[ 3, {spec}, "Fatal"]; LL[]),
									err_ :> (sowMessage[ 8, {"theorem", {spec}, ToString[ err]}, "Fatal"]; LL[])
								}
							],
						(*else*)
							Replace[ getFmlEnv[ True][ spec][[2]],
								{
									FML$[ _, f_, __, "TypeVarPositions" -> tvp_List, ___] :>
										LL @@ {
											{
												Hold[ rest],
												PRFSTAT$[ id,
													pos,
													If[ MemberQ[ keys, spec], keys, Append[ keys, spec]],
													thms,
													ReplacePart[ po, fst -> addAssmToProofSit[ f, tvp, ps]]
												]
											},
											LL[]
										},
									_ :> (sowMessage[ 3, {spec}, "Fatal"]; LL[])
								}
							]
						]
				}
			]
		);
	executeProofCommands1[ Hold[ INF$[ "forget", i_Integer?Positive], rest___], PRFSTAT$[ id_, pos:{f_List, ___}, keys_List, thms_, po_PRFOBJ$]] :=
		With[ {f0 = Append[ f, 2]},
		With[ {hyps = Extract[ po, f0]},
			If[ i <= Length[ hyps],
				With[ {defSource = sourceLabel[]},
					LL @@ {{Hold[ rest], PRFSTAT$[ id, pos, Union[ keys, DeleteCases[ {key[ hyps[[i]]]}, {_, defSource}]], thms, ReplacePart[ po, f0 -> Delete[ hyps, i]]]}, LL[]}
				],
			(*else*)
				$msgID = id;
				sowMessage[ 6, {i}, "Fatal"];
				LL[]
			]
		]];
	executeProofCommands1[ Hold[ rl:INF$[ method_String, params___], rest___], state:PRFSTAT$[ id_, {f_List, r___}, keys_, thms_, po_PRFOBJ$]] :=
		(
			$msgID = id;
			Switch[ method,
				basic,
				With[ {ps = Extract[ po, f], rule = rl},
				With[ {new = basicInference[ rule, ps]},
					Switch[ new,
						_List,
						With[ {info = makePRFINFO[ name -> method, "params" -> Rest[ rule]]},
						With[ {node = If[ new === {}, makeTERMINALNODE[ info, proved], makeANDNODE[ info, new]]},
							LL @@ {
								{
									Hold[ rest],
									PRFSTAT$[ id, Join[ getNewRelevantSits[ node, f], {r}], keys, thms, propagateProofValues[ ReplacePart[ po, f -> node]]]
								},
								LL[]
							}
						]],
						_String,
						sowMessage[ 8, {method, params, new}, "Fatal"];
						LL[],
						_,
						sowMessage[ 7, {method, params}, "Fatal"];
						LL[]
					]
				]],
				
				struct,
				sowMessage[ 7, {method, params}, "Fatal"];
				LL[],
				
				_,
				sowMessage[ 5, {method}, "Fatal"];
				LL[]
			]
		)
]

executeProofCommands1[ Hold[ method_, rest___], state:(PRFSTAT$[ id_, pos:{f_List, ___}, __, po_PRFOBJ$])] :=
	With[ {ps = Extract[ po, f]},
	With[ {new = ($msgID = id; method[ ps, Length[ pos], 1])},
		If[ MatchQ[ new, _LL],
			lMapIndexed[ executeProofCommands1op[ Hold[ rest], state], new],
		(*else*)
			LL @@ {executeProofCommands1op[ Hold[ rest], state][ new], LL[]}
		]
	]]
executeProofCommands1[ _, _] := LL[]

addAssmToProofSit[ a_, tvp_List, PRFSIT$[ {g_, goals___}, hyps_, idx_, fix_, meta_, tfix_, tvar_, ctxt__]] :=
	Replace[ extractFormula[ a, tvp, idx],
		{STERM$[ newA_, _, newTVar_Association], newIdx_} :>
			PRFSIT$[
				{TYPED$[ CoreImplies$TM, {}][ newA][ g], goals},
				hyps,
				newIdx,
				fix,
				meta,
				tfix,
				Join[ tvar, newTVar],	(* keys are disjoint, 'newTVar' has not yet been added to 'tvar' *)
				ctxt
			]
	]


(* ::Section:: *)
(*Setup*)

(* 'initiateProof[ f, assms, mode]' starts a new proof of formula 'f'. 'f' may be arbitrary, as long as its type is 'FORM$TM'
	and it does not involve any free term variables, a.b.f. term constants or meta variables.
	'assms' can be an arbitrary list of keys of formulas which shall be used as assumptions. *)
initiateProof[ fml_FML$, assms_List, 1] :=
	With[ {res = initiateProof[ fml, assms, 0]},
		If[ MatchQ[ res, PRFSTAT$[ _, {PRFSIT$[ {_}, _List, __]}, {}, {}, _]],
			initFormulaLabel[];
			PRFSTAT$[
				{},
				{{2}},
				{},
				{},
				Last[ res],
				With[ {ps = res[[2, 1]]},
					PRFOBJ$[
						makePRFINFO[ name -> initialProofSituation, generated -> Prepend[ ps[[2]], fml], id -> "OriginalPS"],
						ps,
						pending
					]
				]
			],
		(*else*)
			res
		]
	]
initiateProof[ fml:FML$[ {_, src_String}, __], assms_List, _Integer] :=
	If[ getTheoryRootDist[ src] === 0,
		Replace[ extractFormula[ fml, 0],
			{sf:STERM$[ f_, _, _], _} :>
				Replace[ Reap[ Replace[ checkType[ sf, FORM$TM, {}, {}, {}, Association[], Association[]],
								{
									{_, {}, _, Association[], tv_Association} :> tv,
									_ -> "typeChecking"
								}
							], "Message"],
					{
						{tv_Association, msg0_List} :>
							With[ {am = addAssumptions[ assms], msg1 = (Join @@ msg0)[[All, 2]]},
							With[ {fSpec = specializeTypes[ STERM$[ f, Association[], tv], {}]},
							With[ {ps = PRFSIT$[ {fSpec[[1, 1]]}, First[ am], 0, {}, {}, Last[ fSpec], Association[], 0], msg = Join[ msg1, Last[ am]]},
								If[ msg =!= {}, Sow[ 0 -> msg, "Message"]];
								$msgID = {};	(* should be set to IDs of branches later; only set here to ensure it is not 'None' *)
								$preProof[ fml];
								PRFSTAT$[
									{},
									{ps},
									{},
									{}
								]
							]]],
						{"typeChecking", msg_List} :> (Sow[ 0 -> Append[ (Join @@ msg)[[All, 2]], MSG$[ coreMessageToList, 11, {}, "Fatal"]], "Message"]; $Failed),
						{_, msg_List} :> (Sow[ 0 -> Append[ (Join @@ msg)[[All, 2]], MSG$[ coreMessageToList, 12, {}, "Fatal"]], "Message"]; $Failed)
					}
				]
		],
	(*else*)
		Sow[ 0 -> MSG$[ coreMessageToList, 16, {src}, "Fatal"], "Message"];
		$Failed
	]
initiateProof[ _, _List, _Integer] :=
	(
		Sow[ 0 -> {MSG$[ coreMessageToList, 13, {}, "Fatal"]}, "Message"];
		$Failed
	)

addAssumptions[ fmls_List] :=
	With[ {res =
			Map[
				With[ {k = key@#},
					Replace[ getFmlEnv[ True][ k][[2]], Except[ _FML$] :> MSG$[ coreMessageToList, 14, {k}, "Warning"]]
				]&,
				fmls
			]},
		{Cases[ res, _FML$], Cases[ res, _MSG$]}
	]

(* ::Section:: *)
(*Inference Rules*)

(* The set of basic inferences is not the smallest, or most concise, that would theoretically be possible. This means
	that some of the inferences are strictly speaking superfluous, because they could well be replaced by combinations
	of other inferences and theorems. Furthermore, some basic inferences could be weakened; for instance, it would
	suffice if "allE" could only instantiate quantifiers that do not appear in the scope of premises.
	The design of the inference kernel reflects a reasonable tradeoff between efficiency on the one hand, and
	simplicity on the other hand. After all, the inference rules are still comparatively "simple" and comprehensible by
	humans, who should have no difficulties verifying that what the inferences express is really reasonable, and that
	they are correctly implemented by the given programs. *)

(* ::Subsection:: *)
(*Structural Rules*)

(* "psData" changes the optional information stored in proof situations; this information does not have any influence
	on the correctness of the reasoning machinery. *)
basicInference[ INF$[ "psData", data_List], PRFSIT$[ G_, hyps_, idx_, fix_, meta_, tfix_, tvar_, fid_, ___]] :=
	{Join[ PRFSIT$[ G, hyps, idx, fix, meta, tfix, tvar, fid], PRFSIT$ @@ data]}

(* "insert" adds a formula from the background theory to the hypotheses of the current proof situation. *)
basicInference[ INF$[ "insert", k_], PRFSIT$[ G_List, hyps_List, ctxt__]] :=
	With[ {fml = getFmlEnv[ True][ k][[2]]},
		Switch[ fml,
			_FML$,
			{PRFSIT$[ G, Prepend[ hyps, fml], ctxt]},
			
			_String,
			fml,
			
			_,
			$Failed
		]
	]

(* "hypData" changes the meta-data, i.e. label and optional parameters (except "TypeVarPositions"), of a hypothesis. *)
basicInference[ INF$[ "hypData", i_Integer, lbl_String, data___], ps_PRFSIT$] :=
	If[ MemberQ[ {data}, "TypeVarPositions" -> _],
		"invalidData",
	(*else*)
		Replace[
			Replace[ Quiet[ Check[ ps[[2, i]], $Failed]],
				{
					FML$[ k_, f_, l_String, ___, "TypeVarPositions" -> tvp_List, ___] :> FML$[ k, f, lbl, "TypeVarPositions" -> tvp, data],
					FML$[ k_, f_, l_String, ___] :> FML$[ k, f, lbl, data]
				}
			],
			{
				new_FML$ :> {ReplacePart[ ps, {2, i} -> new]},
				_ -> "invalidHypIndex"
			}
		]
	]
	
(* "makeExplicit" turns an arbitrary formula from the hypotheses into an explicit assumption of the first subgoal.
	Note that the formula still remains among the hypotheses, too. *)
basicInference[ INF$[ "makeExplicit", i_Integer], PRFSIT$[ {G_, R___}, hyps_List, idx_, fix_, meta_, tfix_Association, tvar_Association, ctxt__]] :=
	Replace[ extractFormula[ Quiet[ Check[ hyps[[i]], $Failed]], idx],
		{
			{STERM$[ new_, Association[], tv_Association], newIdx_Integer} :>
				{PRFSIT$[
					{TYPED$[ CoreImplies$TM, {}][ new][ G], R},
					hyps,
					newIdx,
					fix,
					meta,
					tfix,
					Join[ tvar, tv],	(* keys are disjoint *)
					ctxt
				]},
			_ -> "invalidHypIndex"
		}
	]
	
(* "makeImplicit" moves one assumption of the first and only subgoal into the list of hypotheses.
	The assumption still remains in the subgoal, too.
	Note that the assumption may neither contain meta variables nor polymorphic a.b.f. term constants, and it may not
	be polymorphic itself. *)
basicInference[ INF$[ "makeImplicit", i_Integer?Positive], PRFSIT$[ {G_}, hyps_List, idx_, fix_, meta_, tfix_, tvar_, fid_, ctxt___]] :=
	Replace[ extractAssm[ G, i],
		{
			$Failed -> "invalidAssmIndex",
			assm_ :>
				If[ FreeQ[ assm, _META$|_TVAR$],
					If[ freeOfPolymorphicFix[ assm, fix],
						{PRFSIT$[
							{G},
							Prepend[ hyps, FML$[ makeFormulaKey[ fid], assm, "?", "TypeVarPositions" -> {}]],
							idx,
							fix,
							meta,
							tfix,
							tvar,
							fid + 1,
							ctxt
						]},
					(*else*)
						"polymorphicFix"
					],
				(*else*)
					"assmContainsVar"
				]
		}
	]

(* "drop" drops an arbitrary assumption of the first subgoal. Together with "makeImplicit" this corresponds to something like "impI".
	Note that the assumption being dropped may contain arbitrary term meta variables and/or type variables. *)
basicInference[ INF$[ "drop", i_Integer?Positive], PRFSIT$[ {G_, R___}, ctxt__]] :=
	If[ assmCount[ G, 0, i] === i,
		{PRFSIT$[
			{If[ i === 1, First[ G], With[ {pos = Table[ 1, {i - 1}]}, ReplacePart[ G, pos -> Extract[ G, Append[ pos, 1]]]]], R},
			ctxt
		]},
	(*else*)
		"invalidAssmIndex"
	]

(* "split" splits the current proof situation into two proof situations, by extracting the first subgoal.
	This inference rule is only applicable if
	- the sets of term meta variables in the subgoal that is extracted and in the remaining subgoals are disjoint,
	- the sets of type variables in the subgoal that is extracted and in the remaining subgoals are disjoint. *)
basicInference[ INF$[ "split"], PRFSIT$[ {G_, R0___}, hyps_List, idx_, fix_List, meta_List, ctxt__]] :=
	With[ {R = {R0}},
	With[ {metaG = DeleteDuplicates[ Cases[ G, _META$, {0, Infinity}, Heads -> True]],
			metaR = DeleteDuplicates[ Cases[ R, _META$, {0, Infinity}, Heads -> True]]},
	With[ {metaGPatt = Alternatives @@ metaG},
		If[ MemberQ[ metaR, metaGPatt],
			"subgoalContainsMeta",
		(*else*)
			With[ {fixG = Replace[ DeleteDuplicates[ Cases[ G, _FIX$, {0, Infinity}, Heads -> True]], fix, {1}],
					fixR = Replace[ DeleteDuplicates[ Cases[ R, _FIX$, {0, Infinity}, Heads -> True]], fix, {1}]},
			With[ {tvarG = Alternatives @@ DeleteDuplicates[ Cases[ {G, fixG, Replace[ metaG, meta, {1}][[All, 1]]}, _TVAR$, {0, Infinity}, Heads -> True]]},
				If[ FreeQ[ {R, fixR, Replace[ metaR, meta, {1}][[All, 1]]}, tvarG],
					{
						PRFSIT$[
							{G},
							hyps,
							idx,
							fix,
							Cases[ meta, _[ metaGPatt, _]],
							ctxt
						],
						PRFSIT$[
							R,
							hyps,
							idx,
							fix,
							DeleteCases[ meta, _[ metaGPatt, _]],
							ctxt
						]
					},
				(*else*)
					"subgoalContainsTypeVar"
				]
			]]
		]
	]]]

basicInference[ INF$[ "rearrangeG", i_Integer?Positive, j_Integer?Positive], PRFSIT$[ G_List, ctxt__]] /; Max[ i, j] <= Length[ G] :=
	{PRFSIT$[ Insert[ Delete[ G, i], G[[i]], j], ctxt]}

(* "cut" sets up a subproof of an arbitrary formula and adds this formula to the list of hypotheses of the original
	goal. "cut" is special in the sense that all type variables in the cut-formula are specialized before it becomes a
	new subgoal. However, it is only applicable if the subgoal is free of meta variables and of polymorphic a.b.f. term
	constants.

		   [C]
		   	.
		   	.
		   	.
	 C		G
	----------
		G
*)
basicInference[ INF$[ "cut", form:STERM$[ f_, _Association, _Association]],
		PRFSIT$[ G_List, hyps_, idx_, fix_, meta_, tfix_Association, tvar_Association, fid_, ctxt___]] :=
	Replace[ checkType[ form, FORM$TM, {}, fix, {}, tfix, Association[]],
		(* since 'checkType' is called with the empty list of meta variables, 'form' does not contain *any* meta variable *)
		{
			{_, _List, max_Integer, newTFix_Association, tv_Association} :>
				If[ freeOfPolymorphicFix[ f, fix],
					With[ {formNew = STERM$[ f, newTFix, tv]},
					With[ {tvp = getPositions[ formNew, _TVAR$]},
					With[ {fSpec = specializeTypes[ formNew, Join[ Keys[ tfix], Keys[ newTFix]], tvp]},
						{
							PRFSIT$[
								{fSpec[[1, 1]]},
								hyps,
								0,		(* there are no type variables in 'fSpec' *)
								fix,
								{},
								Join[ tfix, newTFix, Last[ fSpec]],
								Association[],
								fid,
								ctxt
							],
							PRFSIT$[
								G,
								Prepend[ hyps,
									FML$[
										makeFormulaKey[ fid],
										f,
										"?",
										"TypeVarPositions" -> Map[ (Rest /@ First[ #]) -> Lookup[ tv, Last[ #], getDefaultSort[]] &, tvp]	(* 'Rest' because leading '1' must be dropped *)
									]
								],
								idx,
								fix,
								meta,
								Join[ tfix, newTFix],
								tvar,
								fid + 1,
								ctxt
							]
						}
					]]],
				(*else*)
					"polymorphicFix"
				],
			_ -> "typeChecking"
		}
	]

(* ::Subsection:: *)
(*Quantifier Rules*)

basicInference[ INF$[ "meta", m_META$, typ:Except[ _List], dep_Integer?NonNegative], PRFSIT$[ G_List, hyps_List, idx_, fix_, meta_, tfix_, tvar_, ctxt__]] :=
	If[ isWellFormedAtom[ HoldComplete[ m]] && MemberQ[ meta, m -> _],
		"invalidMeta",
	(*else*)
		If[ dep <= Length[ fix],
			Replace[ isWellFormedType[ typ, tfix, tvar],
				{
					{{}, newTVar_List} :>
						{PRFSIT$[
							G,
							hyps,
							Max[ idx, Max[ newTVar[[All, 1]]] + 1],
							fix,
							Append[ meta, m -> {typ, dep}],
							tfix,
							tvar,
							ctxt
						]},
					_ -> "illFormedType"
				}
			],
		(*else*)
			"invalidDependency"
		]
	]

(* "inst" instantiates a meta variable in the proof goal. The instance must not contain fresh a.b.f. term constants or
	meta variables, it must type-check, its type must agree with the type of the variable, and it must not depend on
	illegal a.b.f. constants.
	
	 G(t)
	-------
	 G(?x)
*)
basicInference[ INF$[ "inst", m_META$ -> inst:STERM$[ inst0:Except[ $Failed], _Association, _Association]],
		PRFSIT$[ G_List, hyps_List, idx_, fix_, meta_, tfix_Association, tvar_Association, ctxt__]] :=
	With[ {info = Replace[ m, meta]},
		If[ ListQ[ info],
			With[ {ci = Last[ info], newMeta = DeleteCases[ meta, m -> _]},
				Replace[ checkType[ inst, First[ info], {}, Take[ fix, ci], newMeta, tfix, tvar],
					{
						{_, tinst_List, max_Integer, newTFix_Association, tv_Association} :>
							{PRFSIT$[
								normalize[ replaceTypes[ G, tinst] /. m -> inst0, $autoNormalize && MatchQ[ inst0, _Lambda$TM]],
								hyps,
								Max[ idx, max],
								fix /. tinst,
								newMeta /. tinst,
								Join[ tfix, newTFix],
								KeyDrop[ tv, tinst[[All, 1]]],
								ctxt
							]},
						_ -> "typeChecking"
					}
				]
			]
		]
	]

(* "sorts" adds further sort constraints to the type variables in the current proof context. *)
basicInference[ INF$[ "sorts", newTVar_Association], PRFSIT$[ G_List, hyps_List, idx_, fix_, meta_, tfix_, tvar_, ctxt__]] :=
	If[ idx === 0 || isSubsort[ KeySelect[ newTVar, First[ #] < idx &], tvar],
		{PRFSIT$[ G, hyps, Max[ idx, 1 + Max[ Keys[ newTVar][[All, 1]]]], fix, meta, tfix, newTVar, ctxt]},
	(*else*)
		"incompatibleSorts"
	]

(* "instT" instantiates type variables in the proof goal.
	The instances are *not* lifted to the current context, and must be compatible with sort constraints on the type
	variables to be instantiated.

	 G(t_)
	--------
	 G(?x_)
*)
basicInference[ INF$[ "instT", inst0:{(_TVAR$ -> _)...}], PRFSIT$[ G_List, hyps_List, idx_, fix_, meta_, tfix_, tvar_, ctxt__]] :=
	With[ {newTFix0 = isWellFormedType[ inst0, tfix]},
		If[ ListQ[ newTFix0],
			With[ {m = Max[ idx, 1 + Max[ Cases[ inst0, TVAR$[ i_Integer, ___] :> i, {0, Infinity}, Heads -> True]]]},
			Replace[ eliminateAnonymousVars[ inst0, m],
				{inst_List, max_Integer} :>
					With[ {lhs = inst[[All, 1]], rhs = inst[[All, 2]]},
						If[ FreeQ[ rhs, Alternatives @@ lhs],
							Replace[
								hasSortList[ hasSortTFix[ tfix]][
									rhs,
									Lookup[ tvar, lhs, getDefaultSort[]],
									{Association[ (# -> {})& /@ newTFix0], KeyDrop[ tvar, inst[[All, 1]]]},
									0
								],
								{
									{newTFix_Association, newTVar_Association} :>
										{PRFSIT$[
											replaceTypes[ G, inst],
											hyps,
											max,
											fix /. inst,
											meta /. inst,
											Join[ tfix, newTFix],		(* keys are disjoint *)
											newTVar,
											ctxt
										]},
									_ -> "incompatibleSorts"
								}
							],
						(*else*)
							"varInInst"
						]
					]
			]],
		(*else*)
			"illFormedType"
		]
	]
	
(* "allE" instantiates the outermost universal quantifier in an arbitrary assumption of the first subgoal by a given
	term. The term must satisfy similar requirements as the instance in "inst"; in particular, it must not contain
	fresh a.b.f. term constants or meta variables.

		(A1 ==> ... ==> An ==> P(t)) ==> G
	-----------------------------------------
	 (A1 ==> ... ==> An ==> /\x. P(x)) ==> G
*)
basicInference[ INF$[ "allE", i_Integer?Positive, inst:STERM$[ inst0_, _Association, _Association]],
		PRFSIT$[ {G_, R___}, hyps_, idx_, fix_, meta_, tfix_Association, tvar_Association, ctxt__]] :=
	Replace[ stripAssms[ extractAssm[ G, i], 0],
		{
			{TYPED$[ CoreAll$TM, {typ_}][ A_], n_Integer} :>
				Replace[ checkType[ inst, typ, {}, fix, meta, tfix, tvar],
					{
						{_, tinst_List, max_Integer, newTFix_Association, tv_Association} :>
							{PRFSIT$[
								replaceTypes[
									{
										ReplacePart[ G,
											Join[ assmPosition[ i], Table[ 1, {n}]] ->
												normalize[ applyFunction[ A, inst0], $autoNormalize && MatchQ[ inst0, _Lambda$TM]]
										],
										R
									},
									tinst
								],
								hyps,
								Max[ idx, max],
								fix /. tinst,
								meta /. tinst,
								Join[ tfix, newTFix],
								KeyDrop[ tv, tinst[[All, 1]]],
								ctxt
							]},
						_ -> "typeChecking"
					}
				],
			_ -> "invalidAssm"
		}
	]
	
(* "allI" replaces the outermost quantifier, possibly ocurring as a consequent of implications, by a fresh a.b.f. constant.

	   A1 ==> ... ==> An ==> P(x0)
	---------------------------------
	 A1 ==> ... ==> An ==> /\x. P(x)
*)
basicInference[ INF$[ "allI", const_FIX$], PRFSIT$[ {G_, R___}, hyps_List, idx_, fix_List, ctxt__]] :=
	If[ isWellFormedAtom[ HoldComplete[ const]] && !MemberQ[ fix, _[ const, _]],
		Replace[ stripAssms[ G, 0],
			{
				{TYPED$[ CoreAll$TM, {typ_}][ G0_], n_Integer} :>
					{PRFSIT$[
						{ReplacePart[ G, {Table[ 1, {n}]} -> applyFunction[ G0, const]], R},
						hyps,
						idx,
						Append[ fix, const -> typ],
						ctxt
					]},
				_ -> "noQuantifier"
			}
		],
	(*else*)
		"invalidConstant"
	]

(* "forgetFix" deletes the given a.b.f. constants from the list 'fix', provided they do not appear in the proof
	situation. Strictly speaking, this proof method is not necessary, of course, but sometimes cleaning up the context
	(e.g. after applying (automatic) proof methods that temporarily introduce new constants) is desirable.
	Note that the basic inference is applicable even if some (or all) of the given constants are not contained in 'fix'
	at all, and that the dependency-information of meta variables is updated as to ensure that no meta variable may
	depend on one of the deleted constants. *)
basicInference[ INF$[ "forgetFix", consts_List], PRFSIT$[ G_List, hyps_List, idx_, fix_List, meta_List, ctxt__]] :=
	With[ {patt = Alternatives @@ consts},	(* we do not check whether 'consts' consists of 'FIX$' only, since this does not matter anyway *)
		If[ FreeQ[ {G, hyps}, patt],
			With[ {pos = Position[ fix, patt -> _, {1}, Heads -> False]},
			With[ {min = Min @@ pos - 1},
				{PRFSIT$[ G, hyps, idx, Delete[ fix, pos], If[ IntegerQ[ min], Replace[ meta, (m_ -> {t_, d_}) :> (m -> {t, Min[ d, min]}), {1}], meta], ctxt]}
			]],
		(*else*)
			"constOccurs"
		]
	]

(* ::Subsection:: *)
(*Backchaining*)

(* "impE" backchains the first subgoal w.r.t. an arbitrary assumption.

	 A1 ==> ... ==> An ==> C1	...		A1 ==> ... ==> An ==> Cm
	-------------------------------------------------------------
		(C1 ==> ... ==> Cm ==> G) ==> A1 ==> ... ==> An ==> G
*)
basicInference[ INF$[ "impE", i_Integer?Positive, n_Integer?NonNegative], PRFSIT$[ {G_, R___}, hyps_List, idx_Integer, fix_List, meta_List, tfix_Association, tvar_Association, ctxt__]] :=
	With[ {acGoal = assmCount[ G, 0, Infinity]},
		If[ acGoal >= n + i,
			With[ {rule = Extract[ G, assmPosition[ i]]},	(* we already know that 'i' is a valid index *)
			With[ {acRule = assmCount[ rule, 0, Infinity]},
				If[ acRule >= n,
					With[ {agRule = toAssmsGoal[ rule, {}, acRule - n], pos = Table[ 1, {acGoal - n - 1}]},
					Replace[ alphaEquiv[ Last[ agRule], Extract[ G, Append[ pos, 1]], tfix, tvar],
						{
							{tv_Association, tinst_List} :>
								With[ {g0 = If[ i === 1, First[ G], With[ {pos0 = Table[ 1, {i - 1}]}, ReplacePart[ G, pos0 -> Extract[ G, Append[ pos0, 1]]]]]},
									{PRFSIT$[
										replaceTypes[ Join[ If[ pos === {}, Most[ agRule], Map[ ReplacePart[ g0, pos -> #]&, Most[ agRule]]], {R}], tinst],
										hyps,
										idx,
										fix /. tinst,
										meta /. tinst,
										tfix,
										KeyDrop[ tv, tinst[[All, 1]]],
										ctxt
									]}
								],
							_ -> "notAlphaEquiv"
						}
					]],
				(*else*)
					"invalidAssm"
				]
			]],
		(*else*)
			"invalidAssmIndex"
		]
	]

(* ::Subsection:: *)
(*Equality Rules*)

(* "refl" closes the first subgoal, provided its consequent asserts the equality of two alpha-equivalent terms.

	-------------------------------
	 A1 ==> ... ==> An ==> a === a
*)
basicInference[ INF$[ "refl"], PRFSIT$[ {G_, R___}, hyps_List, idx_Integer, fix_List, meta_List, tfix_Association, tvar_Association, ctxt__]] :=
	Replace[ stripAssms[ G],
		{
			TYPED$[ CoreEqual$TM, _][ a1_][ a2_] :>
				Replace[ alphaEquiv[ a1, a2, tfix, tvar],
					{
						{tv_Association, tinst_List} :> {PRFSIT$[ replaceTypes[ {R}, tinst], hyps, idx, fix /. tinst, meta /. tinst, tfix, KeyDrop[ tv, tinst[[All, 1]]], ctxt]},
						_ -> "notAlphaEquiv"
					}
				],
			_ -> "invalidGoal"
		}
	]

(* "alphabetaeta" replaces an expression occurring at an arbitrary position in the first subgoal by an alpha-beta-eta equivalent one.
	The replacement may happen under a binder, in the presence of bound variables, and the new expression is normalized *only* for
	checking alpha-beta-eta equivalence with the old expression.

	 G(b)
	------	if a =abe= b
	 G(a)
*)
basicInference[ INF$[ "alphabetaeta", pos_List, term_, normOrig_, normNew_],
		PRFSIT$[ {G_, R___}, hyps_List, idx_Integer, fix_List, meta_List, tfix_Association, tvar_Association, ctxt__]] :=
	With[ {el = validExtract[ G, pos, {}]},
		If[ ListQ[ el],
			With[ {e = normalize[ First[ el], normOrig], loose = Last[ el]},
				Replace[
					If[ normNew === False,
						(* there is no need to type-check 'term', since it must be alpha-equivalent to 'e' anyway *)
						{
							Null,
							{},
							Catch[
								Max[
									Max @@ Cases[ term,
										tv_TVAR$ :> If[ MatchQ[ tv, TVAR$[ _Integer?NonNegative]|TVAR$[ _Integer?NonNegative, _Symbol]], First[ tv], Throw[ $Failed]],
										{0, Infinity},
										Heads -> True
									],
									-1
								] + 1
							],
							Association[],
							Association[]
						},
					(*else*)
						checkType[ STERM$[ term, Association[], Association[]], Blank[], loose, fix, meta, tfix, tvar]
					],
					{
						{_, {}, max_Integer, Association[], tv0_Association} :>
							With[ {term1 = normalize[ term, normNew]},
								Replace[ alphaEquiv[ e, term1, tfix, tv0],
									{
										{tv_Association, tinst_List} :>
											{PRFSIT$[
												replaceTypes[ {If[ pos === {}, term, ReplacePart[ G, pos -> term]], R}, tinst],
												hyps,
												Max[ idx, max],
												fix /. tinst,
												meta /. tinst,
												tfix,
												KeyDrop[ tv, tinst[[All, 1]]],
												ctxt
											]},
										_ -> "notAlphaEquiv"
									}
								]
							],
						_ -> "typeChecking"
					}
				]
			],
		(*else*)
			"invalidPosition"
		]
	]

(* "replace" replaces a valid subexpression of the first subgoal by an equal one, where the equality must be asserted
	by an arbitrary explicit assumption.

	 a === b ==> G(b)
	------------------
	 a === b ==> G(a)
	 
	 Note: In Isabelle, there is no inference rule for arbitrary replacements w.r.t. Pure equalities. Instead, they have rules
	 for reflexivity, symmetry, transitivity, beta-eta-conversion ('beta_conversion') and combinations ('combination'), which
	 boils down to the same.
*)
basicInference[ INF$[ "replace", i_Integer?Positive, pos_List], PRFSIT$[ {G_, R___}, hyps_List, idx_Integer, fix_List, meta_List, tfix_Association, tvar_Association, ctxt__]] :=
	Replace[ extractAssm[ G, i],
		{
			TYPED$[ CoreEqual$TM, _][ a_][ b_] :>
				With[ {e = validExtract[ G, pos]},
					If[ e =!= $Failed,
						Replace[ alphaEquiv[ a, e, tfix, tvar],
							{
								{tv_Association, tinst_List} :>
									{PRFSIT$[
										replaceTypes[
											{
												With[ {new = ReplacePart[ G, pos -> b]},	(* 'pos' cannot be '{}' *)
												With[ {conv = getMaxConversionContext[ pos, MatchQ[ a, _Lambda$TM] || MatchQ[ b, _Lambda$TM]]},
													If[ conv === 0,
														new,
													(*else*)
														MapAt[ normalize, new, {Drop[ pos, -conv]}]
													]
												]],
												R
											},
											tinst
										],
										(* Normalizing only the part that is affected by the replacement leads to complications with backchaining:
											beta-expanding the expression first might leave some eta-redexes after the replacement. *)
										hyps,
										idx,
										fix /. tinst,
										meta /. tinst,
										tfix,
										KeyDrop[ tv, tinst[[All, 1]]],
										ctxt
									]},
								_ -> "notAlphaEquiv"
							}
						],
					(*else*)
						"invalidPosition"
					]
				],
			$Failed -> "invalidAssmIndex",
			_ -> "invalidAssm"
		}
	]

(* "ext" expresses functional extensionality, i.e. it reduces functional equalities to equalities with explicitly
	universally quantified variables.

	 A1 ==> ... ==> An ==> /\y1...yi. a[ y1, ..., yi] === b[ y1, ..., yi]
	----------------------------------------------------------------------
			A1 ==> ... ==> An ==> a === (b::t1_ -> ... -> ti -> s)
*)
basicInference[ INF$[ "ext", i_Integer?Positive], PRFSIT$[ {G_, R___}, ctxt__]] :=
	Replace[ stripAssms[ G, 0],
		{
			{TYPED$[ CoreEqual$TM, {typ_}][ a_][ b_], n_Integer} :>
				Replace[ flattenTypes[ typ, {}, i],
					{
						l_List :>
							{PRFSIT$[
								{
									ReplacePart[ G,
										{Table[ 1, {n}]} ->
											Fold[
												TYPED$[ CoreAll$TM, {#2}][ Lambda$TM[ RNG$[ TYPRNG$[ Null, #2]], #1]] &,
												TYPED$[ CoreEqual$TM, {Last[ l]}][
													Fold[ applyFunction, a, Table[ VAR$[ k], {k, i - 1, 0, -1}]]
												][
													Fold[ applyFunction, b, Table[ VAR$[ k], {k, i - 1, 0, -1}]]
												],
												Reverse[ Most[ l]]
											]
									],
									R
								},
								ctxt
							]},
						_ -> "noFunctionType"
					}
				],
			_ -> "invalidGoal"
		}
	]

(* "eqI" is the introduction rule for equality between terms of FORM-type. It might appear a bit strange that this rule
	is included, but it is needed, for instance, in "HOL" to prove the various atomization- and reflection theorems.
	Also, it is an axiom of "Isabelle/Pure", called "equal_intr" ("thm.ml").

	 A1 ==> ... ==> An ==> a ==> b		A1 ==> ... ==> An ==> b ==> a
	------------------------------------------------------------------
				A1 ==> ... ==> An ==> a === (b::FORM)
*)
basicInference[ INF$[ "eqI"], PRFSIT$[ {G_, R___}, ctxt__]] :=
	Replace[ stripAssms[ G, 0],
		{
			{TYPED$[ CoreEqual$TM, {FORM$TM}][ a_][ b_], n_Integer} :>
				{PRFSIT$[
					{
						ReplacePart[ G, {Table[ 1, {n}]} -> TYPED$[ CoreImplies$TM, {}][ a][ b]],
						ReplacePart[ G, {Table[ 1, {n}]} -> TYPED$[ CoreImplies$TM, {}][ b][ a]],
						R
					},
					ctxt
				]},
			_ -> "invalidGoal"
		}
	]


(* ::Subsection:: *)
(*Auxiliary Functions*)

compareThmIndices[ {}, {}] = Equal
compareThmIndices[ {}, {__}] = Less
compareThmIndices[ {__}, {}] = Greater
compareThmIndices[ {a_, b___}, {c_, d___}] :=
	Replace[ compareThmIndices[ a, c], Equal :> compareThmIndices[ {b}, {d}]]
compareThmIndices[ a_List, b_] :=
	compareThmIndices[ a, {b}]
compareThmIndices[ a_, b_List] :=
	compareThmIndices[ {a}, b]
compareThmIndices[ a:(_Integer|_Rational|_Real), a_] =
	Equal
compareThmIndices[ a:(_Integer|_Rational|_Real), b:(_Integer|_Rational|_Real)] :=
	If[ a < b, Less, Greater]
compareThmIndices[ _, _] = Indeterminate

makeFormulaKey[ i_Integer] :=
	makeFormulaKey[ i, "none"]
makeFormulaKey[ i_Integer, src_String] :=
	makeFmlKey[ "core" <> ToString[ i], src]

initiateThmProof[ f_STERM$, tvp_List] :=
	Replace[ specializeTypes[ f, {}, tvp],
		{STERM$[ f0_, _, _], tfix_Association} :>
			PRFSTAT$[
				{},
				{PRFSIT$[ {f0}, {}, 0, {}, {}, tfix, Association[], 0]},
				{},
				{}
			]
	]

toStandardPRFSIT[ PRFSIT$[ goals_List, hyps_, idx_, fix_, meta_, tfix_, tvar_, fid_, ctxt___], opts___?OptionQ] :=
	PRFSIT$[
		FML$[ makeFormulaKey[ fid], makeConjunction[ goals, And$TM], "G\[NumberSign]" <> Theorema`Language`FormulaManipulation`Private`defLabel[]],
		hyps,
		ToString[ Unique[ "PRFSIT$"]],
		"CoreData" -> {idx, fix, meta, tfix, tvar},
		ctxt,
		opts
	]

assmCount[ _, m_, _?NonPositive] :=
	m
assmCount[ TYPED$[ CoreImplies$TM, _][ _][ R_], m_Integer, n_] :=
	assmCount[ R, m + 1, n - 1]
assmCount[ _, m_, _] := m

toAssmsGoal[ G_, l_List, 0] :=
	Append[ l, G]
toAssmsGoal[ G_, l_List, n_] :=
	Join[ l, reap[ toAssmsGoal[ G, n]]]
toAssmsGoal[ TYPED$[ CoreImplies$TM, {}][ A_][ R_], n_?Positive] :=
	(Sow[ A]; toAssmsGoal[ R, n - 1])
toAssmsGoal[ G_, _] := (Sow[ G];)

stripAssms[ G_] :=
	NestWhile[ First, G, MatchQ[ #, TYPED$[ CoreImplies$TM, _][ _][ _]]&]
stripAssms[ G_, n_Integer] :=
	Module[ {m = n},
		{NestWhile[ First, G, If[ MatchQ[ #, TYPED$[ CoreImplies$TM, _][ _][ _]], ++m; True, False]&], m}
	]

getMaxConversionContext[ pos_List, lam_] :=
	getMaxConversionContext[ pos, lam, $autoNormalize]
getMaxConversionContext[ pos_List, lam_, norm_] :=
	If[ lam === True === norm,
		LengthWhile[ Reverse[ pos], # =!= 1 &],
	(*else*)
		0
	]

(* 'extractAssm' ensures that the given index indeed refers to an existing assumption. This cannot be achieved by
	combining 'assmPosition' with 'Extract'! *)
extractAssm[ TYPED$[ CoreImplies$TM, {}][ A_][ R_], i_Integer] :=
	If[ i > 1,
		extractAssm[ R, i - 1],
	(*else*)
		A
	]
extractAssm[ _, _] :=
	$Failed

SetAttributes[ assmPosition, Listable];
assmPosition[ i_Integer?Positive] :=
	Join[ Table[ 1, {i - 1}], {0, 1}]

validExtract[ expr_, pos_] :=
	validExtract[ expr, pos, False]
validExtract[ expr_, pos_, loose_List] :=
	Replace[ Reap[ validExtract[ expr, pos, True], 0],
		{
			{$Failed, _} -> $Failed,
			{part_, l_} :> {part, Join[ Reverse[ Join @@ l], loose]}
		}
	]
validExtract[ expr_, {}, True|False] :=
	expr
validExtract[ _VAR$|_FIX$|_META$, Except[ {}], _, True|False] :=
	$Failed
validExtract[ TYPED$[ expr_, _], {i_Integer, rest___}, sow:(True|False)] :=
	If[ i === 1,
		validExtract[ expr, {rest}, sow],
	(*else*)
		$Failed
	]
validExtract[ Lambda$TM[ RNG$[ TYPRNG$[ _, t_]], expr_], {i_Integer, rest___}, sow:(True|False)] :=
	If[ i === 2,
		If[ sow, Sow[ t, 0]];
		validExtract[ expr, {rest}, sow],
	(*else*)
		$Failed
	]
validExtract[ h_[ _], {0, rest___}, sow:(True|False)] :=
	validExtract[ h, {rest}, sow]
validExtract[ _[ a_], {1, rest___}, sow:(True|False)] :=
	validExtract[ a, {rest}, sow]
validExtract[ ___] :=
	$Failed

(* The last argument of 'applyFunction' is a boolean value that indicates whether there might be a variable in 'f' that
	is bound outside of 'f'. If not, and the argument the function is applied on is 'VAR$[0]', no beta reduction needs
	to be performed. *)
applyFunction[ f_, a_] :=
	applyFunction[ f, a, True]
applyFunction[ Lambda$TM[ _, e_], VAR$[ 0], False] :=
	e
applyFunction[ Lambda$TM[ _, e_], a_, _] :=
	betaReduce[ e, a]
applyFunction[ f_, a_, _] :=
	f[ a]

(* 'alphaEquiv[ a, b, tfix, tvar]' checks whether 'a' and 'b' are alpha-equivalent on the term level and unifiable (or
	identical, depending on the value of '$instantiateTypes') on the type level. The function either returns a pair
	consisting of a new sort-association of type variables and a type-unifier, or 'False'. *)
alphaEquiv[ a_HoldComplete, b_HoldComplete, tfix_Association, tvar_Association] :=
	If[ a === b,	(* 'SameQ' is extremely fast, so it should be fine to test for syntactic identity first *)
		{tvar, {}},
	(*else*)
		alphaEquiv[ a, b, tfix, {tvar, {}}]
	]
alphaEquiv[ (List|Hold)[ a___], (List|Hold)[ b___], tfix_Association, tvar_Association] :=
	alphaEquiv[ HoldComplete[ a], HoldComplete[ b], tfix, tvar]
alphaEquiv[ a_, b_, tfix_Association, tvar_Association] :=
	alphaEquiv[ HoldComplete[ a], HoldComplete[ b], tfix, tvar]
alphaEquiv[ HoldComplete[], b_, _, acc_List] :=
	If[ b === HoldComplete[],
		acc,
	(*else*)
		False
	]
alphaEquiv[ HoldComplete[ a_, aa__], b0_, tfix_Association, acc_List] :=
	Replace[ b0,
		{
			HoldComplete[ b_, bb__] :>
				alphaEquiv[ HoldComplete[ aa], HoldComplete[ bb], tfix, alphaEquiv[ HoldComplete[ a], HoldComplete[ b], tfix, acc]],
			_ -> False
		}
	]
alphaEquiv[ a:HoldComplete[ _VAR$|_FIX$|_META$], b_, _, acc_List] :=
	If[ a === b,
		acc,
	(*else*)
		False
	]
alphaEquiv[ _, HoldComplete[ _VAR$|_FIX$|_META$], _, _List] :=
	False
alphaEquiv[ x:HoldComplete[ TYPED$[ a_, _]], y_, tfix_Association, acc_List] :=
	Replace[ y,
		{
			HoldComplete[ TYPED$[ b_, _]] :>
				alphaEquiv[
					HoldComplete[ a],
					HoldComplete[ b],
					tfix,
					compareTypes[ x, y, hasSort[ tfix], acc, If[ TrueQ[ $instantiateTypes], 0, Infinity]]
				],
			_ -> False
		}
	]
alphaEquiv[ _, HoldComplete[ _TYPED$], _, _List] :=
	False
alphaEquiv[ HoldComplete[ Lambda$TM[ RNG$[ TYPRNG$[ _, s_]], a_]], b0_, tfix_Association, acc_List] :=
	Replace[ b0,
		{
			HoldComplete[ Lambda$TM[ RNG$[ TYPRNG$[ _, t_]], b_]] :>
				alphaEquiv[
					HoldComplete[ a],
					HoldComplete[ b],
					tfix,
					compareTypes[ s, t, hasSort[ tfix], acc, If[ TrueQ[ $instantiateTypes], 0, Infinity]]
				],
			_ -> False
		}
	]
alphaEquiv[ _, HoldComplete[ _Lambda$TM], _, _List] :=
	False
alphaEquiv[ HoldComplete[ f_[ a_]], b0_, tfix_Association, acc_List] :=
	Replace[ b0,
		{
			HoldComplete[ g_[ b_]] :>
				alphaEquiv[ HoldComplete[ a], HoldComplete[ b], tfix, alphaEquiv[ HoldComplete[ f], HoldComplete[ g], tfix, acc]],
			_ -> False
		}
	]
alphaEquiv[ a_HoldComplete, b_HoldComplete, _, acc_List] :=
	If[ a === b,
		acc,
	(*else*)
		False
	]
alphaEquiv[ _, _, _, _] :=
	False

(* 'extractFormula' always returns either '$Failed' or a pair involving a 'STERM$'-structure. *)
extractFormula[ FML$[ _, f_, _, data___], idx_Integer] :=
	extractFormula[ f, Replace[ "TypeVarPositions", {data, "TypeVarPositions" -> Null}], idx]
extractFormula[ f_, tvp_List, idx_Integer] :=
	{
		STERM$[
			ReplacePart[ f, MapIndexed[ First[ #1] -> TVAR$[ idx + First[ #2] - 1] &, tvp]],
			Association[],
			With[ {list = DeleteCases[ MapIndexed[ TVAR$[ idx + First[ #2] - 1] -> Last[ #1] &, tvp], _ -> getDefaultSort[]]},
				Association[ list]
			]
		],
		idx + Length[ tvp]
	}
extractFormula[ f_, _, idx_Integer] :=
	liftTypeVars[ STERM$[ f, Association[], Association[]], idx]
extractFormula[ ___] := $Failed

(* 'normalize[ expr]' transforms 'expr' into beta-eta normal form, if '$autoNormalize' is 'True'. *)
normalize[ expr_] :=
	normalize[ expr, $autoNormalize]
normalize[ expr_, True] :=
	etaReduce[ betaReduce[ expr]]
normalize[ expr_, _] :=
	expr

(* 'betaReduce' does not use 'substFree', for efficiency reasons.
	Beta- and eta-reduction work for uncurried expressions, too, because unification needs it. *)
betaReduce[ expr_] :=
	expr //. {
				Lambda$TM[ _, e_][ a_] :> betaReduce[ e, a],
				Lambda$TM[ _, e_][ a_, aa__] :> betaReduce[ e, a][ aa]
			}
betaReduce[ e_, a_] :=
	With[ {fa = reap[ scanVars[ a, If[ TrueQ[ #3 >= #2], Sow[ {#1} -> VAR$[ #3]]]&]]},
		ReplacePart[
			e,
			reap[
				scanVars[
					e,
					Function[ {pos, abs, i},
						Which[
							i === abs,
							Sow[ {pos} -> If[ abs === 0, a, ReplacePart[ a, Replace[ fa, (p_ -> VAR$[ j_]) :> (p -> VAR$[ j + abs]), {1}]]]],
							i > abs,
							Sow[ {pos} -> VAR$[ i - 1]]
						]
					]
				]
			]
		]
	]

etaReduce[ expr_] :=
	expr //. {
				Lambda$TM[ _, (term:Except[ Complex|DirectedInfinity|Rational|Lambda$TM])[ VAR$[ 0]]] :>
					With[ {res = etaReduce[ term, True]},
						ReplacePart[ term, res] /; ListQ[ res]
					],
				Lambda$TM[ _, (term:Except[ Complex|DirectedInfinity|Rational|Lambda$TM])[ aa__, VAR$[ 0]]] :>
					With[ {term0 = term[ aa]},
					With[ {res = etaReduce[ term0, True]},
						ReplacePart[ term0, res] /; ListQ[ res]
					]]
			}
etaReduce[ term_, True] :=
	Catch[
		reap[ scanVars[ term, Which[ #2 === #3, Throw[ $Failed], #3 > #2, Sow[ {#1} -> VAR$[ #3 - 1]]]&]]
	]

freeOfPolymorphicFix[ expr_, fix_List] :=
	With[ {poly = Cases[ fix, (c_ -> _?(MemberQ[ #, _TVAR$, {0, Infinity}, Heads -> True]&)) :> c]},
		poly === {} || FreeQ[ expr, Alternatives @@ poly]
	]

(* 'scanVars[ expr, fun]' scans 'expr' and applies 'fun' to all triples 'pos, abs, v', where
	- 'v' is the de Bruijn index of a variable occurring in 'expr',
	- 'pos' is its position, and
	- 'abs' is the number or abstractions in whose scope it appears. *)
scanVars[ l_List, args__] :=
	Scan[ scanVars[ #, args]&, l]
scanVars[ expr_, fun_] :=
	Scan[
		fun[ #, countAbstractionsFromPos[ #, expr], First[ extractSingle[ expr, #]]]&,
		Position[ expr, VAR$[ _]]
	]
scanVars[ expr_, pred_, fun_] :=
	Scan[
		With[ {v = First[ extractSingle[ expr, #]], abs = countAbstractionsFromPos[ #, expr]},
			If[ TrueQ[ pred[ #, abs, v]], fun[ #, abs, v]]
		]&,
		Position[ expr, VAR$[ _]]
	]

countAbstractionsFromPos[ pos_List, _] :=
	Count[ pos, 2]

SetAttributes[ reap, HoldFirst];
reap[ expr_] :=
	With[ {res = Last[ Reap[ expr]]},
		If[ res === {},
			res,
		(*else*)
			Last[ res]
		]
	]

trimMessageList[ msg_List] :=
	With[ {len = Length[ msg]},
		If[ len > 6,
			Append[ Take[ msg, 5], MSG$[ coreMessageToList, 15, {len - 5}]],
		(*else*)
			msg
		]
	]


(* ::Section:: *)
(*Locking Symbols*)

(*
Scan[
	SetAttributes[ #, {Protected, Locked}]&,
	{
		executeProofCommands0op, executeProofCommands1op, executeProofCommands2op,
		executeProofCommands0, executeProofCommands1, executeProofCommands2,
		addAssmToProofSit,
		initiateThmProof, toStandardPRFSIT, addAssumptions,
		extractAssm, trimMessageList, freeOfPolymorphicFix
	}	(* TODO: Make sure that all private functions are listed here. *)
]

Scan[
	With[ {sym = Symbol[ #]},
		SetAttributes[ sym, {Protected, Locked}]
	]&,
	Names[ "LazyList`*"]
]

Scan[
	With[ {sym = Symbol[ "LazyList`Private`" <> #]},
		SetAttributes[ sym, {Protected, Locked}]
	]&,
	{"conjPredicate", "isMember", "generateFair", "generateBest", "generateDF", "generateBF", "pruneNewBF"}
	(* TODO: Make sure that all private functions are listed here. *)
]
*)

End[]

EndPackage[]
