(* ::Package:: *)

(* Theorema
    Copyright (C) 2017 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Core`Trusted`Types`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Language`"]

Begin["`Private`"]

If[ !ValueQ[ $strictTypeChecking], $strictTypeChecking = False];
If[ !ValueQ[ $reportUndeclaredConsts], $reportUndeclaredConsts = False];

(* ::Section:: *)
(*Public Functions*)

(* 'checkType[ term, target, loose, fix, meta, tfix, tvar]' type-checks the given term in the given typing context. No
	type inference is applied; instead, 'term' must be fully type-annotated already.
	The function returns either '$Failed' or a 5-tuple '{typ, tinst, idx, newTFix, newTVar}' consisting of
	- the type 'typ' of 'expr',
	- a type instantiation 'tinst' that unifies the original type of 'term' and 'target', if '$instantiateTypes' is
		'True' (note that 'typ' is the *original* type of 'term', not the instantiated one!), or '$Failed' if the types
		cannot be unified,
	- the smallest number 'idx' such that every type variable appearing in 'term' explicitly (i.e. not in the types of
		meta variables) has index < 'idx',
	- the association of new a.b.f. type constants 'newTFix', and
	- the new association of type variables; this is basically the merge of 'tvar' and the association passed along
		with 'term'. Type variables that are instantiated by 'tinst' are still included in 'newTVar', but sorts are
		adjusted where necessary. *)
checkType[ STERM$[ term_, termTFix_Association, termTVar_Association], target:Except[ _List], loose_List, fix_List, meta_List, tfix_Association, tvar_Association] :=
	If[ AllTrue[ Keys[ termTFix], With[ {s = Lookup[ termTFix, #]}, s === Lookup[ tfix, #, s]]&],
		Replace[ checkTypeAux[ HoldComplete[ term], fix, meta, tfix, loose, -1, KeyComplement[ {termTFix, tfix}], termTVar, Association[], tvar],
			{t_, max_Integer, _, _, newTFix0_Association, newTVar0_Association} :>
				Which[
					target === Blank[] || target === t,
					{t, {}, max + 1, newTFix0, newTVar0},
				(*else*)
					$instantiateTypes,
					Replace[ unifyTypes[ t, target, hasSortTFix[ tfix], {{newTFix0, newTVar0}, {}}, 0],
						{
							{{newTFix_Association, newTVar_Association}, tunif_List} :> {t, tunif, max + 1, newTFix, newTVar},
							_ :> {t, $Failed, max + 1, newTFix0, newTVar0}
						}
					],
				(*else*)
					True,
					{t, $Failed, max + 1, newTFix0, newTVar0}
				]
		],
	(*else*)
		$Failed
	]
checkType[ term_STERM$, fix_List, meta_List, tfix_Association, tvar_Association] :=
	checkType[ term, Blank[], {}, fix, meta, tfix, tvar]
checkType[ term:Except[ _STERM$], args___] :=
	checkType[ STERM$[ term, Association[], Association[]], args]

(* 'isLegalTypeList' (and 'isLegalType') check whether all type constructors appearing in the given type(s) have been
	declared/defined already. *)
isLegalTypeList[ typs_List, typeDefs_, legalVars_, legalFix_, varErr_, undefErr_] :=
	isLegalTypeList[ typs, False, typeDefs, legalVars, legalFix, varErr, undefErr]
isLegalTypeList[ typs_List, sow_, typeDefs_, legalVars_, legalFix_, varErr_, undefErr_] :=
	Catch[
		If[ TrueQ[ sow],
			Last[ Reap[ Scan[ isLegalTypeAux[ typeDefs, legalVars, legalFix, varErr, undefErr, True], typs], _TYPE$, #1&]],
		(*else*)
			Scan[ isLegalTypeAux[ typeDefs, legalVars, legalFix, varErr, undefErr, False], typs];
			True
		]
	]

isLegalType[ typ_, typeDefs_, legalVars_, legalFix_, varErr_, undefErr_] :=
	isLegalType[ typ, False, typeDefs, legalVars, legalFix, varErr, undefErr]
isLegalType[ typ_, sow_, typeDefs_, legalVars_, legalFix_, varErr_, undefErr_] :=
	Catch[
		If[ TrueQ[ sow],
			Last[ Reap[ isLegalTypeAux[ typeDefs, legalVars, legalFix, varErr, undefErr, True][ typ], _TYPE$, #1&]],
		(*else*)
			isLegalTypeAux[ typeDefs, legalVars, legalFix, varErr, undefErr, False][ typ]
		]
	]

isLegalTypeAux[ _, legalVars_, _, varErr_, _, _][ v:(TVAR$[ _Integer]|TVAR$[ _Integer, _Symbol])] :=
	If[ TrueQ[ legalVars[ v]],
		True,
	(*else*)
		varErr[ v];
		Throw[ False]
	]
isLegalTypeAux[ _, _, legalFix_, varErr_, _, _][ v:(TFIX$[ _Symbol, _Integer])] :=
	If[ TrueQ[ legalFix[ v]],
		True,
	(*else*)
		varErr[ v];
		Throw[ False]
	]
isLegalTypeAux[ _, _, _, varErr_, _, _][ v:(_TVAR$|_TFIX$)] :=
	(varErr[ v]; Throw[ False])
isLegalTypeAux[ typeDefs_, legalVars_, legalFix_, varErr_, undefErr_, sow_][ FUN$TM[ s_, t_]] :=
	isLegalTypeAux[ typeDefs, legalVars, legalFix, varErr, undefErr, sow][ s] && isLegalTypeAux[ typeDefs, legalVars, legalFix, varErr, undefErr, sow][ t]
isLegalTypeAux[ _, _, _, _, _, _][ FORM$TM|STRING$TM] :=
	True
isLegalTypeAux[ typeDefs_, legalVars_, legalFix_, varErr_, undefErr_, sow_][ h_[ a__]] :=
	Module[ {match},
	With[ {i = First[ FirstPosition[ typeDefs[ h], (t_ -> _) /; ListQ[ match = matchType[ t, h[ a]]], {0}, {1}, Heads -> False]]},
		If[ i === 0,
			undefErr[ h[ a]];
			Throw[ False],
		(*else*)
			Scan[ isLegalTypeAux[ typeDefs, legalVars, legalFix, varErr, undefErr, False], {a}];
			If[ TrueQ[ sow] && !MatchQ[ typeDefs[ h][[i, 2]], {_, _}|{_, _, _, _, _, _, {}}],	(* '{_, _}' could also represent a single key *)
				Sow[ Null, TYPE$[ h, i, match]]
			];
			True
		]
	]]
isLegalTypeAux[ typeDefs_, _, _, _, undefErr_, sow_][ t_] :=
	With[ {i = First[ FirstPosition[ typeDefs[ t], t -> _, {0}, {1}, Heads -> False]]},
		If[ i === 0,
			undefErr[ t];
			Throw[ False],
		(*else*)
			If[ TrueQ[ sow] && !MatchQ[ typeDefs[ t][[i, 2]], {_, _}|{_, _, _, _, _, _, {}}],
				Sow[ Null, TYPE$[ t, i, {}]]
			];
			True
		]
	]

(* 'isWellFormedType[ type, tfix]' checks whether
	- 'type' is free of TYPED$, VAR$, FIX$, META$, SVAR$, RNG$, TYPRNG$, SEQ$, SEQ0$, SEQ1$, $Failed,
	- everything except possibly the head of 'type' is free of List,
	- every type variable must have either no arguments, one integer argument, or an integer- and a symbol argument,
	- every a.b.f. type constant must have one symbol- and one integer argument.
	In case of success, the function returns the list of all a.b.f. type constants appearing in 'type' but not in
	'tfix'. Note that 'isWellFormedType' does not check whether every occurrence of 'TVAR$' and 'TFIX$' has arguments
	in the first place. Argument-less occurrences are treated just like any other constant, without having any special
	meaning.
	'isWellFormedType[ type, tfix, tvar]' basically does the same, but additionally requires that no anonymous type
	variables appear in 'type'. It returns either a pair '{newTFix, newTVar}' consisting of the list of new a.b.f. type
	constants and the list of new type variables, or 'False'. *)
isWellFormedType[ type_List, args___] :=
	isWellFormedType[ Hold @@ type, args]
isWellFormedType[ type_] :=
	FreeQ[ type, TYPED$|VAR$|FIX$|META$|SVAR$|RNG$|TYPRNG$|SEQ$|SEQ0$|SEQ1$|List|$Failed] &&
	FreeQ[ type, Except[ _[]|_[ _Integer]|_[ _Integer, _Symbol], _TVAR$]] &&
	FreeQ[ type, Except[ _[ _Symbol, _Integer], _TFIX$]]
isWellFormedType[ type_, tfix_Association] :=
	If[ FreeQ[ type, TYPED$|VAR$|FIX$|META$|SVAR$|RNG$|TYPRNG$|SEQ$|SEQ0$|SEQ1$|List|$Failed] &&
			FreeQ[ type, Except[ _[]|_[ _Integer]|_[ _Integer, _Symbol], _TVAR$]],
		With[ {f = DeleteDuplicates[ Cases[ type, _TFIX$?(Lookup[ tfix, #, True]&), {0, Infinity}, Heads -> True]]},
			If[ MatchQ[ f, {(_[ _Symbol, _Integer])...}],
				f,
			(*else*)
				False
			]
		]
	]
isWellFormedType[ type_, tfix_Association, tvar_Association] :=
	If[ FreeQ[ type, TYPED$|VAR$|FIX$|META$|SVAR$|RNG$|TYPRNG$|SEQ$|SEQ0$|SEQ1$|List|$Failed],
		Catch[
			{
				DeleteDuplicates[
					Cases[
						type,
						c:TFIX$[ args___] /; Lookup[ tfix, c, True] :> Replace[ {args}, {{_Symbol, _Integer} -> c, _ :> Throw[ False]}],
						{0, Infinity},
						Heads -> True
					]
				],
				DeleteDuplicates[
					Cases[
						type,
						v:TVAR$[ args___] /; Lookup[ tvar, v, True] :> (Replace[ {args}, {{_Integer, _Symbol} -> v, {_Integer} -> v, _ :> Throw[ False]}]),
						{0, Infinity},
						Heads -> True
					]
				]
			}
		],
	(*else*)
		False
	]

isWellFormedAtom[ HoldComplete[ VAR$[ _Integer?NonNegative]]] :=
	True
isWellFormedAtom[ HoldComplete[ (FIX$|META$)[ s_Symbol, _Integer]]] :=
	!MemberQ[ {TYPED$, VAR$, FIX$, META$, SVAR$, TVAR$, TFIX$, SEQ$, SEQ0$, SEQ1$, TYPRNG$, RNG$, Lambda$TM}, s]
isWellFormedAtom[ _] :=
	False

(* 'shiftTypeVars[ term, offset]' simply adds 'offset' to all (non-anonymous) type variables appearing in 'term'. *)
shiftTypeVars[ expr_, 0] :=
	expr
shiftTypeVars[ STERM$[ expr_, tfix_, tvar_Association], offset_Integer] :=
	STERM$[ shiftTypeVars[ expr, offset], tfix, KeyMap[ Function[ x, MapAt[ (# + offset)&, x, 1]], tvar]]
shiftTypeVars[ expr_, offset_Integer] :=
	expr /. TVAR$[ i_Integer, r___] :> TVAR$[ i + offset, r]

(* 'liftTypeVariables[ term, idx]' transfers all type variables appearing in 'term' into the context characterized by
	'idx'. This means that
	- every anonymous type variable gets a fresh index,
	- the index of every indexed type variable is adjusted to avoid clashes.
	The function returns the pair '{newTerm, newIdx}'. *)
liftTypeVars[ expr_, idx_Integer] :=
	liftTypeVars[ expr, idx, True]
liftTypeVars[ STERM$[ expr_, tfix_, tvar_Association], idx_Integer, keepNames_] :=
	With[ {repl = typeLiftingReplacement[ expr, idx, keepNames]},
		eliminateAnonymousVars[
			STERM$[ replaceTypes[ expr, repl], tfix, KeyMap[ Replace[ #, repl]&, tvar]],
			idx + Length[ repl],
			TVAR$
		]
	]
liftTypeVars[ expr_, idx_Integer, keepNames_] :=
	With[ {repl = typeLiftingReplacement[ expr, idx, keepNames]},
		eliminateAnonymousVars[
			replaceTypes[ expr, repl],
			idx + Length[ repl],
			TVAR$
		]
	]

typeLiftingReplacement[ expr_, idx_Integer, keepNames_] :=
	With[ {tvars = DeleteDuplicates[ Cases[ expr, TVAR$[ _Integer, _Symbol]|TVAR$[ _Integer], {0, Infinity}, Heads -> True]]},
	With[ {symPos = Position[ tvars, TVAR$[ _Integer, _Symbol], {1}, Heads -> False]},
	With[ {intSymRepl =
			MapIndexed[
				If[ TrueQ[ keepNames], #1 -> TVAR$[ First[ #2] + idx - 1, Last[ #1]] &, #1 -> TVAR$[ First[ #2] + idx - 1] &],
				Extract[ tvars, symPos]
			]},
	With[ {n0 = idx + Length[ intSymRepl] - 1},
	With[ {intRepl =
			MapIndexed[
				#1 -> TVAR$[ First[ #2] + n0] &,
				Delete[ tvars, symPos]
			]},
		Join[ intSymRepl, intRepl]
	]]]]]

eliminateAnonymousVars[ expr_] :=
	eliminateAnonymousVars[ expr, getMaxTVar[ expr] + 1]
eliminateAnonymousVars[ expr_, n_Integer] :=
	eliminateAnonymousVars[ expr, n, TVAR$]
eliminateAnonymousVars[ STERM$[ TVAR$[], tfix_, tvar_], n_Integer, h_] :=
	{STERM$[ h[ n], tfix, tvar], n + 1}
eliminateAnonymousVars[ STERM$[ expr_, tfix_, tvar_], n_Integer, h_] :=
	With[ {repl = MapIndexed[ {#1} -> h[ First[ #2] + n - 1] &, Position[ expr, TVAR$[]]]},
		{STERM$[ ReplacePart[ expr, repl], tfix, tvar], n + Length[ repl]}
	]
eliminateAnonymousVars[ TVAR$[], n_Integer, h_] :=
	{h[ n], n + 1}
eliminateAnonymousVars[ expr_, n_Integer, h_] :=
	With[ {repl = MapIndexed[ {#1} -> h[ First[ #2] + n - 1] &, Position[ expr, TVAR$[]]]},
		{ReplacePart[ expr, repl], n + Length[ repl]}
	]

(* 'generalizeTypes[ term, tfix, idx]' turns every a.b.f. type constant appearing anywhere in 'term' but not in 'tfix'
	into a fresh type variable. 'STERM$'-structures are updated accordingly, by moving entries from 'tfix' to 'tvar'.
	The function returns the pair '{newTerm, newIdx}'.
	Note that 'tfix' is a plain list of a.b.f. type constants, not an association! *)
generalizeTypes[ expr_] :=
	generalizeTypes[ expr, {}]
generalizeTypes[ expr_, tfix_List] :=
	generalizeTypes[ expr, tfix, getMaxTVar[ expr] + 1]
generalizeTypes[ expr_, tfix_List, idx_Integer] :=
	With[ {repl = typeGeneralizationReplacement[ expr, tfix, idx]},
		{
			swapSortAssociations[ repl, {}][ replaceTypes[ expr, repl]],
			idx + Length[ repl]
		}
	]
generalizeTypes[ expr_, _List, idx_Integer, pos_List] :=
	{
		(* 'pos' is assumed to be a list with elements of the form '{p1, p2, ...} -> TFIX$[ ...]'. *)
		With[ {repl = MapIndexed[ Last[ #1] -> (TVAR$ @@ Prepend[ Most[ Last[ #1]], First[ #2] + idx - 1]) &, pos]},
			swapSortAssociations[ repl, {}][ ReplacePart[ expr, MapThread[ First[ #1] -> Last[ #2] &, {pos, repl}]]]
		],
		idx + Length[ pos]
	}

typeGeneralizationReplacement[ STERM$[ expr_, _Association, _Association], tfix_List, idx_Integer] :=
	typeGeneralizationReplacement[ expr, tfix, idx]
typeGeneralizationReplacement[ expr_, tfix_List, idx_Integer] :=
	typeGeneralizationReplacement[ DeleteDuplicates[ Cases[ expr, _TFIX$?(!MemberQ[ tfix, #]&), {0, Infinity}, Heads -> True]], idx]
typeGeneralizationReplacement[ tfixes_List, idx_Integer] :=
	Module[ {i = idx - 1},
		Replace[ tfixes, c:TFIX$[ sym_Symbol, _Integer] :> (c -> TVAR$[ ++i, sym]), {1}]
	]

(* 'specializeTypes[ term, tfix]' turns every *non-anonymous* type variable appearing anywhere in 'term' into an
	a.b.f. type constant that does not appear in 'tfix'. It returns the pair '{newTerm, newTFix}'. If 'term' is a
	'STERM$'-structure, the sort constraints are updated accordingly. *)
specializeTypes[ expr_, tfix_List] :=
	specializeTypes[ expr, tfix, getPositions[ expr, _TVAR$]]
specializeTypes[ expr:STERM$[ _, tf_Association, tv_Association], tfix_List, tvp_List] :=
	With[ {repl =
			(* 'tvp' is assumed to be a list with elements of the form '{p1, p2, ...} -> TVAR$[ ...]'. *)
			typeSpecializationReplacementAux[
				Cases[ tvp, _[ _, v:TVAR$[ _Integer]] :> v],
				tfix,
				Join @@ Map[
					Function[
						p,
						With[ {sym = p[[1, 2, 2]]},
							With[ {n = Max[ -1, Max @@ Cases[ tfix, TFIX$[ sym, i_Integer] :> i]]},
								MapIndexed[ Last[ #1] -> TFIX$[ sym, First[ #2] + n] &, p]
							]
						]
					],
					GatherBy[ Cases[ tvp, _[ _, TVAR$[ _Integer, _Symbol]]], Composition[ Last, Last]]
				]
			]},
		{
			swapSortAssociations[ {}, repl][ ReplacePart[ expr, Map[ (First[ #] -> Replace[ Last[ #], repl])&, tvp]]],
			With[ {list = Thread[ repl[[All, 2]] -> Lookup[ tv, repl[[All, 1]], getDefaultSort[]]]},	(* entries with default sort must be kept! *)
				Association[ list]
			]
		}
	]
specializeTypes[ expr_, tfix_List, tvp_List] :=
	With[ {repl =
			(* 'tvp' is assumed to be a list with elements of the form '{p1, p2, ...} -> TVAR$[ ...]'. *)
			typeSpecializationReplacementAux[
				Cases[ tvp, _[ p_, TVAR$[ _Integer]] :> p],
				tfix,
				Join @@ Map[
					Function[
						p,
						With[ {sym = p[[1, 2, 2]]},
							With[ {n = Max[ -1, Max @@ Cases[ tfix, TFIX$[ sym, i_Integer] :> i]]},
								MapIndexed[ First[ #1] -> TFIX$[ sym, First[ #2] + n] &, p]
							]
						]
					],
					GatherBy[ Cases[ tvp, _[ _, TVAR$[ _Integer, _Symbol]]], Composition[ Last, Last]]
				]
			]},
		{
			ReplacePart[ expr, repl],
			If[ getDefaultSort[] === {},
				Association[],
			(*else*)
				With[ {list = (# -> {})& /@ repl[[All, 2]]},
					Association[ list]
				]
			]
		}
	]

typeSpecializationReplacement[ STERM$[ expr_, _, _], tfix_List] :=
	typeSpecializationReplacement[ expr, tfix]
typeSpecializationReplacement[ expr_, tfix_List] :=
	With[ {tvars = DeleteDuplicates[ Cases[ expr, TVAR$[ _Integer, _Symbol]|TVAR$[ _Integer], {0, Infinity}, Heads -> True]]},
	With[ {symPos = Position[ tvars, TVAR$[ _Integer, _Symbol], {1}, Heads -> False]},
		typeSpecializationReplacementAux[
			Delete[ tvars, symPos],
			tfix,
			Join @@ Map[
				Function[
					varList,
					With[ {sym = varList[[1, 2]]},
						With[ {n = Max[ -1, Max @@ Cases[ tfix, TFIX$[ sym, i_Integer] :> i]]},
							MapIndexed[ #1 -> TFIX$[ sym, First[ #2] + n] &, varList]
						]
					]
				],
				GatherBy[ Extract[ tvars, symPos], Last]
			]
		]
	]]

(* ATTENTION! 'swapSortAssociations' deletes default sort constraints from 'tfix', i.e. the resulting association of
	sort constraints on a.b.f. type constants might lack some of these constants. *)
swapSortAssociations[ fixToVar_List, varToFix_List][ STERM$[ expr_, tfix_Association, tvar_Association]] :=
	With[ {tfixes = fixToVar[[All, 1]], tvars = varToFix[[All, 1]]},
	With[ {varSorts = Lookup[ tvar, tvars, Default],
			fixSorts = Lookup[ tfix, tfixes, Default]},
	With[ {varDefPos = Position[ varSorts, Default, {1}, Heads -> False],
			fixDefPos = Position[ fixSorts, Default, {1}, Heads -> False]},
	With[ {newTFix = Append[ KeyDrop[ tfix, tfixes], MapThread[ (Last[ #1] -> #2)&, {Delete[ varToFix, varDefPos], Delete[ varSorts, varDefPos]}]],
			newTVar = Append[ KeyDrop[ tvar, tvars], MapThread[ (Last[ #1] -> #2)&, {Delete[ fixToVar, fixDefPos], Delete[ fixSorts, fixDefPos]}]]},
		STERM$[ expr, newTFix, newTVar]
	]]]]
swapSortAssociations[ _, _][ expr_] :=
	expr

(* 'replaceTypes' is only correct if only type variables and a.b.f. constants are replaced.
	Otherwise it would be necessary to transform all short types into full types first. *)
replaceTypes[ repl_][ term_] :=
	replaceTypes[ term, repl]
replaceTypes[ term_, {}] :=
	term
replaceTypes[ STERM$[ term_, tfix_, tvar_], repl_] :=
	STERM$[ term /. repl, tfix, tvar]
replaceTypes[ term_, repl_] :=
	term /. repl	(* this seems to be more efficient than retrieving the positions of types first, and then replacing only there *)

(* 'getFullTypeNoSorts' assumes that all sort constraints imposed by the declaration of the given constants are met by
	the given short type. 'getFullType' does not assume that. *)
getFullTypeNoSorts[ type_List, HoldComplete[ CoreImplies$TM], _] :=
	If[ type === {},
		FUN$TM[ FORM$TM, FUN$TM[ FORM$TM, FORM$TM]],
	(*else*)
		$Failed
	]
getFullTypeNoSorts[ type_List, HoldComplete[ CoreAll$TM], _] :=
	Replace[ type,
		{
			{t_} :> FUN$TM[ FUN$TM[ t, FORM$TM], FORM$TM],
			_ -> $Failed
		}
	]
getFullTypeNoSorts[ type_List, HoldComplete[ CoreEqual$TM], _] :=
	Replace[ type,
		{
			{t_} :> FUN$TM[ t, FUN$TM[ t, FORM$TM]],
			_ -> $Failed
		}
	]
getFullTypeNoSorts[ type_List, HoldComplete[ _String], _] :=
	If[ type === {},
		STRING$TM,
	(*else*)
		$Failed
	]
getFullTypeNoSorts[ type_List, HoldComplete[ expr_], decls_] :=
	Replace[ decls[ expr],
		{
			{typ_, num_Integer, pos_List} :>
				If[ Length[ type] === num,
					ReplacePart[ typ, MapThread[ Rule[ First[ #1], #2]&, {pos, type}]],
				(*else*)
					$Failed
				],
			_ -> $Failed
		}
	]
getFullTypeNoSorts[ typ_List, expr_Hold, decls_] :=
	getFullTypeNoSorts[ typ, HoldComplete @@ expr, decls]
getFullTypeNoSorts[ typ_List, expr_, decls_] :=
	getFullTypeNoSorts[ typ, HoldComplete[ expr], decls]
getFullTypeNoSorts[ typ_, _, _] := typ
getFullTypeNoSorts[ typ_, expr_] :=
	getFullTypeNoSorts[ typ, expr, getConstDecls[ True]]
getFullTypeNoSorts[ HoldComplete[ TYPED$[ expr_, typ_]]] :=
	getFullTypeNoSorts[ typ, HoldComplete[ expr]]
getFullTypeNoSorts[ t_TYPED$] :=
	getFullTypeNoSorts[ HoldComplete[ t]]

getFullType[ type_List, HoldComplete[ CoreImplies$TM], _, tvar_, _] :=
	If[ type === {},
		{tvar, FUN$TM[ FORM$TM, FUN$TM[ FORM$TM, FORM$TM]]},
	(*else*)
		$Failed
	]
getFullType[ type_List, HoldComplete[ CoreAll$TM], _, tvar_, _] :=
	Replace[ type,
		{
			{t_} :> {tvar, FUN$TM[ FUN$TM[ t, FORM$TM], FORM$TM]},
			_ -> $Failed
		}
	]
getFullType[ type_List, HoldComplete[ CoreEqual$TM], _, tvar_, _] :=
	Replace[ type,
		{
			{t_} :> {tvar, FUN$TM[ t, FUN$TM[ t, FORM$TM]]},
			_ -> $Failed
		}
	]
getFullType[ type_List, HoldComplete[ _String], _, tvar_, _] :=
	If[ type === {},
		{tvar, STRING$TM},
	(*else*)
		$Failed
	]
getFullType[ type_List, HoldComplete[ expr_], tfix_, tvar_, decls_] :=
	Replace[ decls[ expr],
		{
			{typ_, num_Integer, tvp_List} :>
				If[ Length[ type] === num,
					With[ {univSortPos = Position[ tvp, _ -> {}, {1}, Heads -> False]},
					With[ {newTVar = hasSortList[ tfix][ Delete[ type, univSortPos], Delete[ tvp, univSortPos][[All, 2]], tvar, 0]},
						If[ newTVar === $Failed,
							$Failed,
						(*else*)
							{newTVar, ReplacePart[ typ, MapThread[ Rule[ First[ #1], #2]&, {tvp, type}]]}
						]
					]],
				(*else*)
					$Failed
				],
			_ -> $Failed
		}
	]

(* 'getDeclaredType[ sym, idx]' returns a 5-tuple '{full, short, tvar, flex, i}', where 'full' is the full declared
	type of 'sym', 'short' is the short version of 'full' (i.e. either 'full' itself or a list of types), 'tvar' is the
	association of sort constraints on the type variables, 'flex' is a boolean value specifying whether 'typ' is
	"flexible" (i.e. a type inferred from constant declarations as opposed to an explicitly annotated type) or not, and
	'i' is the new index. *)
getDeclaredType[ expr_, idx_Integer] :=
	getDeclaredType[ expr, idx, getConstDecls[ True]]
getDeclaredType[ HoldComplete[ CoreImplies$TM], idx_Integer, _] :=
	{FUN$TM[ FORM$TM, FUN$TM[ FORM$TM, FORM$TM]], {}, Association[], False, idx}
getDeclaredType[ HoldComplete[ CoreAll$TM], idx_Integer, _] :=
	{FUN$TM[ FUN$TM[ TVAR$[ idx], FORM$TM], FORM$TM], {TVAR$[ idx]}, If[ getDefaultSort[] === {}, Association[], Association[ TVAR$[ idx] -> {}]], False, idx + 1}
getDeclaredType[ HoldComplete[ CoreEqual$TM], idx_Integer, _] :=
	{FUN$TM[ TVAR$[ idx], FUN$TM[ TVAR$[ idx], FORM$TM]], {TVAR$[ idx]}, If[ getDefaultSort[] === {}, Association[], Association[ TVAR$[ idx] -> {}]], False, idx + 1}
getDeclaredType[ HoldComplete[ _String], idx_Integer, _] :=
	{STRING$TM, {}, False, Association[], idx}
getDeclaredType[ HoldComplete[ expr_], 0, decls_] :=
	Replace[ decls[ expr],
		{
			{typ_, num_Integer, tvp_List} :>
				With[ {short = Array[ TVAR$, num, 0]},
					{typ, short, Association @@ DeleteCases[ Thread[ short -> tvp[[All, 2]]], _ -> getDefaultSort[]], $strictTypeChecking =!= True, num}
				],
			_ :>
				If[ TrueQ[ $strictTypeChecking],
					$Failed,
				(*else*)
					If[ TrueQ[ $reportUndeclaredConsts], Sow[ expr, "undeclared"]];
					{TVAR$[ 0], TVAR$[ 0], Association[], True, 1}
				]
		}
	]
getDeclaredType[ HoldComplete[ expr_], idx_Integer, decls_] :=
	Replace[ decls[ expr],
		{
			{typ_, num_Integer, tvp_List} :>
				With[ {short = Array[ TVAR$, num, idx]},
					{
						ReplacePart[ typ, MapIndexed[ Function[ {p, i}, Map[ Append[ #, 1]&, First[ p]] -> idx + First[ i] - 1], tvp]],	(* 'tvp' refers to entire 'TVAR$[...]'-expression! *)
						short,
						Association @@ DeleteCases[ Thread[ short -> tvp[[All, 2]]], _ -> getDefaultSort[]],
						$strictTypeChecking =!= True,
						idx + num
					}
				],
			_ :>
				If[ TrueQ[ $strictTypeChecking],
					$Failed,
				(*else*)
					If[ TrueQ[ $reportUndeclaredConsts], Sow[ expr, "undeclared"]];
					{TVAR$[ idx], TVAR$[ idx], Association[], True, idx + 1}
				]
		}
	]

getTypeInstance[ const_, typ_, tfix_, tvar_] :=
	getTypeInstance[ const, typ, tfix, tvar, getConstDecls[ True]]
getTypeInstance[ CoreImplies$TM, typ_, _, tvar_, _] :=
	If[ typ === FUN$TM[ FORM$TM, FUN$TM[ FORM$TM, FORM$TM]],
		{tvar, {}},
	(*else*)
		$Failed
	]
getTypeInstance[ CoreAll$TM, typ_, _, tvar_, _] :=
	Replace[ typ,
		{
			FUN$TM[ FUN$TM[ t_, FORM$TM], FORM$TM] :> {tvar, {t}},
			_ -> $Failed
		}
	]
getTypeInstance[ CoreEqual$TM, typ_, _, tvar_, _] :=
	Replace[ typ,
		{
			FUN$TM[ t_, FUN$TM[ t_, FORM$TM]] :> {tvar, {t}},
			_ -> $Failed
		}
	]
getTypeInstance[ _String, typ_, _, tvar_, _] :=
	If[ typ === STRING$TM,
		{tvar, {}},
	(*else*)
		$Failed
	]
getTypeInstance[ {typ0_, _Integer, tvp_List}, typ_, tfix_, tvar_, _] :=
	With[ {patts = Table[ ToExpression[ "x$$$" <> ToString[ i] <> "_"], {i, 1, Length[ tvp]}]},
	With[ {univSortPos = Position[ tvp, _ -> {}, {1}, Heads -> False]},
	With[ {sorts = Delete[ tvp, univSortPos][[All, 2]], allVars = (Hold @@ patts)[[All, 1]]},
	With[ {vars = Delete[ allVars, univSortPos]},
		Replace[ typ,
			{
				RuleDelayed @@ Prepend[
					ReplacePart[
						Hold[
							With[ {newTVar$$$ = hasSortList[ tfix][ vars, sorts, tvar, 0]},
								{newTVar$$$, allVars} /; newTVar$$$ =!= $Failed
							]
						],
						{{1, 1, 1, 2, 1, 0}, {1, 2, 1, 2, 0}} -> List
					],
					ReplacePart[ typ0, MapThread[ Rule[ First[ #1], #2]&, {tvp, patts}]]
				],
				_ -> $Failed
			}
		]
	]]]]
getTypeInstance[ const:Except[ _List], typ_, tfix_, tvar_, decls_] :=
	Replace[ decls[ const],
		{
			l_List :> getTypeInstance[ l, typ, tfix, tvar, decls],
			_ -> {tvar, typ}
		}
	]
getTypeInstance[ _, _, _, _, _] :=
	$Failed

(* 'flattenTypes[ t, {}]' flattens out nested function types and returns the list of all argument-types followed by the
	atomic result type. The argument-types may of course be function-types.
	'flattenTypes[ t, {}, n]' flattens out precisely the first 'n' nested function types; returns '$Failed' if there
	are less than 'n' functions. *)
flattenTypes[ t_] :=
	flattenTypes[ t, {}]
flattenTypes[ FUN$TM[ s_, t_], acc_List] :=
	flattenTypes[ t, Append[ acc, s]]
flattenTypes[ t_, {}] :=
	{t}
flattenTypes[ t_, acc_] :=
	Append[ acc, t]
flattenTypes[ t_, {}, 0] :=
	{t}
flattenTypes[ t_, acc_, 0] :=
	Append[ acc, t]
flattenTypes[ FUN$TM[ s_, t_], acc_List, n_Integer?Positive] :=
	flattenTypes[ t, Append[ acc, s], n - 1]
flattenTypes[ _, _List, _Integer?Positive] :=
	$Failed

(* 'compareTypes[ a, b, sortFun, acc, constThreshold]' checks whether the types (of) 'a' and 'b' are equal or unifiable
	(under 'acc'), respectively, depending on the value of 'constThreshold'. It returns
	- the new accumulator in case of success, or
	- 'First' if the type of 'a' had to be obtained using 'getFullTypeNoSorts', but the result was '$Failed', or
	- 'Last' is the type of 'b' had to be obtained using 'getFullTypeNoSorts', but the result was '$Failed', or
	- '$Failed' if the unification / identity-testing fails. *)
compareTypes[ HoldComplete[ TYPED$[ _, s_List]], HoldComplete[ TYPED$[ _, t_List]], sortFun_, acc:{_, tinst_List}, constThreshold_] :=
	Which[
		tinst === {} && s === t,
		acc,
	(*else*)
		IntegerQ[ constThreshold],
		If[ Length[ s] === Length[ t],
			unifyTypes[ s, t, sortFun, acc, constThreshold],
		(*else*)
			$Failed
		],
	(*else*)
		True,
		$Failed
	]
compareTypes[ a:HoldComplete[ TYPED$[ _, _List]], HoldComplete[ TYPED$[ _, t_]], sortFun_, acc:{_, tinst_List}, constThreshold_] :=
	With[ {s = getFullTypeNoSorts[ a]},
		Which[
			s === $Failed,
			First,
		(*else*)
			tinst === {} && s === t,
			acc,
		(*else*)
			IntegerQ[ constThreshold],
			unifyTypes[ s, t, sortFun, acc, constThreshold],
		(*else*)
			True,
			$Failed
		]
	]
compareTypes[ HoldComplete[ TYPED$[ _, s_]], b:HoldComplete[ TYPED$[ _, _List]], sortFun_, acc:{_, tinst_List}, constThreshold_] :=
	With[ {t = getFullTypeNoSorts[ b]},
		Which[
			t === $Failed,
			Last,
		(*else*)
			tinst === {} && s === t,
			acc,
		(*else*)
			IntegerQ[ constThreshold],
			unifyTypes[ s, t, sortFun, acc, constThreshold],
		(*else*)
			True,
			$Failed
		]
	]
compareTypes[ HoldComplete[ TYPED$[ _, s_]], HoldComplete[ TYPED$[ _, t_]], sortFun_, acc_List, constThreshold_] :=
	compareTypes[ s, t, sortFun, acc, constThreshold]
compareTypes[ s_, t_, sortFun_, acc:{_, tinst_List}, constThreshold_] :=
	Which[
		tinst === {} && s === t,
		acc,
	(*else*)
		IntegerQ[ constThreshold],
		unifyTypes[ s, t, sortFun, acc, constThreshold],
	(*else*)
		True,
		$Failed
	]

(* 'unifyTypes' only orients its arguments if necessary. This means that, if possible, the resulting type-substitution
	only substitutes type-variables appearing in the first type -- even if this necessitates sorts to be adjusted that
	otherwise could remain as they are.
	All type variables with index < 'constThreshold' are treated as constants.
	The given sort constraints on type variables may be an association or 'None'; in the latter case, sorts are
	completely ignored. *)
unifyTypes[ v:TVAR$[ i_, ___], b_, sortFun_, {sortCtxt_, inst_List}, constThreshold_] /; i >= constThreshold :=
	With[ {w = Replace[ v, inst]},
		If[ w === v,
			With[ {b0 = b /. inst},
				If[ v === b0,
					{sortCtxt, inst},
				(*else*)
					If[ FreeQ[ b0, v],
						With[ {newSortCtxt = sortFun[ b0, v, sortCtxt, constThreshold]},
							If[ newSortCtxt === $Failed,
								$Failed,
							(*else*)
								{newSortCtxt, Prepend[ inst /. v -> b0, v -> b0]}
							]
						],
					(*else*)
						$Failed
					]
				]
			],
		(*else*)
			unifyTypes[ w, b, sortFun, {sortCtxt, inst}, constThreshold]
		]
	]
unifyTypes[ a_, v:TVAR$[ i_, ___], sortFun_, acc:{_, _List}, constThreshold_] /; i >= constThreshold :=
	unifyTypes[ v, a, sortFun, acc, constThreshold]
unifyTypes[ a_TFIX$, b_, _, acc_List, _] :=
	If[ a === b,
		acc,
	(*else*)
		$Failed
	]
unifyTypes[ _, _TFIX$, _, _List, _] :=
	$Failed
unifyTypes[ {}, {}, _, acc_List, _] :=
	acc
unifyTypes[ {a_, aa___}, {b_, bb___}, sortFun_, acc_List, constThreshold_] :=
	unifyTypes[ {aa}, {bb}, sortFun, unifyTypes[ a, b, sortFun, acc, constThreshold], constThreshold]
unifyTypes[ f_[ args1___], g_[ args2___], sortFun_, acc_List, constThreshold_] :=
	With[ {a = {args1}, b = {args2}},
		If[ f === g && Length[ a] === Length[ b],
			unifyTypes[ a, b, sortFun, acc, constThreshold],
		(*else*)
			$Failed
		]
	]
unifyTypes[ a_, b_, _, acc_List, _] :=
	If[ a === b,
		acc,
	(*else*)
		$Failed
	]
unifyTypes[ _, _, _, _, _] :=
	$Failed

(* 'matchType[ patt, typ]' tries to match 'patt' against 'typ', i.e. it tries to find a type substitution 's' such that
	'patt /. s === typ'. Note that the sets of type variables on the LHSs and RHSs of 's' might not be disjoint! This
	means that, say, 'TVAR$[0]' matches 'LIST[TVAR$[0]]', although the two types are not unifiable.
	Note that the types are matched on a purely syntactical level, i.e. no sort constraints are taken into account. *)
matchType[ patt_, typ_] :=
	First[ matchType[ patt, typ, {{}, {}}]]
matchType[ v_TVAR$, typ_, {inst_List, self_List}] :=
	If[ MemberQ[ self, v],
		If[ typ === v,
			{inst, self},
		(*else*)
			{$Failed, $Failed}
		],
	(*else*)
		With[ {w = Replace[ v, inst]},
			Which[
				w === v,
				If[ w === typ,
					{inst, Prepend[ self, v]},
				(*else*)
					{Prepend[ inst, v -> typ], self}
				],
			(*else*)
				w === typ,
				{inst, self},
			(*else*)
				True,
				{$Failed, $Failed}
			]
		]
	]
matchType[ c_TFIX$, typ_, is:{_List, _List}] :=
	If[ c === typ,
		is,
	(*else*)
		{$Failed, $Failed}
	]
matchType[ {}, {}, is_] :=
	is
matchType[ {a_, aa___}, {b_, bb___}, is:{_List, _List}] :=
	matchType[ {aa}, {bb}, matchType[ a, b, is]]
matchType[ f_[ args1___], g_[ args2___], is:{_List, _List}] :=
	With[ {a = {args1}, b = {args2}},
		If[ f === g && Length[ a] === Length[ b],
			matchType[ a, b, is],
		(*else*)
			{$Failed, $Failed}
		]
	]
matchType[ _, _, is_] :=
	is

getMaxTVar[ STERM$[ expr_, _Association, _Association]] :=
	getMaxTVar[ expr]
getMaxTVar[ expr_] :=
	Max[ -1, Max @@ Cases[ expr, TVAR$[ i_Integer, ___] :> i, {0, Infinity}, Heads -> True]]


(* ::Subsection:: *)
(*Sorts*)

(* The currently implemented sort logic is based on "complement sorts". A complement sort is a list of type
	constructors such that a type 't' is of sort 's' iff no type constructor in 's' appears in 't', with whatever arity
	at whatever position.
	Some functions in this subsection are implemented in a sufficiently general way as to handle also other kinds of
	sorts, others are not. The latter functions are explicitly marked as such. *)

intersectSorts[ assocs0___Association] :=
	With[ {defSort = getDefaultSort[]},
		Replace[ If[ defSort === {}, DeleteCases[ {assocs0}, Association[]], {assocs0}],
			{
				{} -> Association[],
				{assoc_} :> assoc,
				assocs_List :>
					If[ defSort === {},
						Merge[ assocs, intersectSorts],
					(*else*)
						DeleteCases[ Merge[ assocs, intersectSorts[ Length[ assocs], defSort]], defSort]
					]
			}
		]
	]
intersectSorts[ sorts_List] :=
	Union @@ sorts
intersectSorts[ num_Integer, defSort_][ sorts_List] :=
	With[ {un = Union @@ sorts},
		If[ Length[ sorts] === num,
			un,
		(*else*)
			Union[ defSort, un]
		]
	]

(* 'sub' is the subsort, i.e. the sort with *more* constraints. *)
isSubsort[ sub_Association, super_Association] :=
	With[ {defSort = getDefaultSort[], superList = Normal[ super]},
		AllTrue[ superList, isSubsort[ Lookup[ sub, First[ #], defSort], Last[ #]]&] &&
		(defSort === {} || AllTrue[ KeySelect[ sub, (!MemberQ[ superList, # -> _])&], isSubsort[ #, defSort]&])
	]
isSubsort[ sub_List, super_List] :=
	SubsetQ[ sub, super]	(* as sets, 'super' must be a subset of 'sub' *)

(* This function can only handle complement sorts at the moment. *)
(* hasSort[ tfix][ typ, sort, tvar, constThreshold]' checks whether the given type 'typ' has the given sort 'sort',
	possibly after constraining the sorts of some type variables with index >= 'constThreshold'.
	In case of success, the function returns the new association of sort constraints on type variables; otherwise, it
	returns '$Failed'. *)
hasSort[ tfix_][ typ_, v_TVAR$, tvar_, args___] :=
	hasSort[ tfix][ typ, Lookup[ tvar, v, getDefaultSort[]], tvar, args]
hasSort[ tfix_][ typ_, c_TFIX$, tvar_, args___] :=
	hasSort[ tfix][ typ, Lookup[ tfix, c, getDefaultSort[]], tvar, args]
hasSort[ _][ _, {}, tvar_, ___] :=
	tvar
hasSort[ tfix_][ typ_, sort_List, tvar_] :=
	hasSort[ tfix][ typ, sort, tvar, 0]
hasSort[ tfix_][ typ_, sort_List, tvar_, constThreshold_] :=
	If[ FreeQ[ typ, Alternatives @@ sort],
		With[ {tvars = DeleteDuplicates[ Cases[ typ, _TVAR$, {0, Infinity}, Heads -> True]]},
		With[ {cpos = Position[ tvars, _[ i_, ___] /; i < constThreshold, {1}, Heads -> False]},
			If[
				AllTrue[
					DeleteDuplicates[
						Join[
							Lookup[ tfix, DeleteDuplicates[ Cases[ typ, _TFIX$, {0, Infinity}, Heads -> True]], getDefaultSort[]],
							Lookup[ tvar, Extract[ tvars, cpos], getDefaultSort[]]
						]
					],
					isSubsort[ #, sort]&
				],
			(*then*)
				If[ constThreshold === Infinity,
					tvar,
				(*else*)
					With[ {tvars1 = Delete[ tvars, cpos]},
					With[ {list = DeleteCases[ MapThread[ #1 -> Union[ #2, sort] &, {tvars1, Lookup[ tvar, tvars1, getDefaultSort[]]}], _ -> getDefaultSort[]]},
						Join[ KeyDrop[ tvar, tvars1], Association[ list]]		(* keys are disjoint *)
					]]
				],
			(*else*)
				$Failed
			]
		]],
	(*else*)
		$Failed
	]

(* This function can only handle complement sorts at the moment. *)
(* 'hasSortTFix' is similar to 'hasSort', but in addition to sort constraints on type variables it also expects sort
	constraints on a.b.f. type constants in the sort context, which it also adjusts. *)
hasSortTFix[ ctxtTFix_][ typ_, v_TVAR$, {tfix_, tvar_}, args___] :=
	hasSortTFix[ ctxtTFix][ typ, Lookup[ tvar, v, getDefaultSort[]], {tfix, tvar}, args]
hasSortTFix[ ctxtTFix_][ typ_, c_TFIX$, {tfix_, tvar_}, args___] :=
	hasSortTFix[ ctxtTFix][ typ, Lookup[ ctxtTFix, c, Lookup[ tfix, c, getDefaultSort[]]], {tfix, tvar}, args]
hasSortTFix[ _][ _, {}, sortCtxt_, ___] :=
	sortCtxt
hasSortTFix[ ctxtTFix_][ typ_, sort_List, sortCtxt_] :=
	hasSortTFix[ ctxtTFix][ typ, sort, sortCtxt, 0]
hasSortTFix[ ctxtTFix_][ typ_, sort_List, {tfix_, tvar_}, constThreshold_] :=
	If[ FreeQ[ typ, Alternatives @@ sort],
		With[ {tfixes = DeleteDuplicates[ Cases[ typ, _TFIX$, {0, Infinity}, Heads -> True]],
				tvars = DeleteDuplicates[ Cases[ typ, _TVAR$, {0, Infinity}, Heads -> True]]},
		With[ {cpos = Position[ tvars, _[ i_, ___] /; i < constThreshold, {1}, Heads -> False]},
			If[
				AllTrue[
					DeleteDuplicates[
						Join[
							DeleteCases[ Lookup[ ctxtTFix, tfixes, $Failed], $Failed],
							Lookup[ tvar, Extract[ tvars, cpos], getDefaultSort[]]
						]
					],
					isSubsort[ #, sort]&
				],
			(*then*)
				With[ {tvars1 = Delete[ tvars, cpos]},
				With[ {list = DeleteCases[ MapThread[ #1 -> Union[ #2, sort] &, {tvars1, Lookup[ tvar, tvars1, getDefaultSort[]]}], _ -> getDefaultSort[]]},
					{
						Map[ Union[ #, sort]&, tfix],
						Join[ KeyDrop[ tvar, tvars1], Association[ list]]		(* keys are disjoint *)
					}
				]],
			(*else*)
				$Failed
			]
		]],
	(*else*)
		$Failed
	]

hasSortList[ tfix_Association] :=
	hasSortList[ hasSort[ tfix]]
hasSortList[ _][ {}, {}, sortAcc_, _] :=
	sortAcc
hasSortList[ fun_][ {typ_}, {sort_}, sortAcc_, constThreshold_] :=
	fun[ typ, sort, sortAcc, constThreshold]
hasSortList[ fun_][ types_List, sorts_List, sortAcc_, constThreshold_] :=
	Catch[
		Fold[
			If[ #1 === $Failed, Throw[ $Failed], fun[ First[ #2], Last[ #2], #1, constThreshold]]&,
			sortAcc,
			Transpose[ {types, sorts}]
		]
	]


(* ::Section:: *)
(*Private Functions*)

(* 'checkTypeAux[ term, fix, meta, tfix, loose, max, termTFix, termTVar, newTFix, tvar]' type-checks 'term' in the
	typing context specified by 'fix', 'meta', 'tfix' and 'loose'. The function either returns '$Failed', or a 6-tuple
	consisting of the full type of 'term', the maximum type-variable-index found so far, the association of all a.b.f.
	type constants that have not been found so far, the association of type variables that have not been found so far,
	the association of all new a.b.f. type constants, and the new association of type variables. *)
checkTypeAux[ HoldComplete[ TYPED$[ term_, type_]], _, _, tfix_, loose_, max_, termTFix_, termTVar_, newTFix_, newTVar_] :=
	Replace[ isWellFormedType[ type, Join[ tfix, newTFix], newTVar],
		{
			{newTFix0_List, newTVar0_List} :>
				With[ {newMax = Max[ max, Max[ newTVar0[[All, 1]]] + 1],
						newTermTFix = KeyDrop[ termTFix, newTFix0],
						newTermTVar = KeyDrop[ termTVar, newTVar0],
						newTFix1 = Join[ newTFix, Association @@ Map[ (# -> Lookup[ termTFix, #, getDefaultSort[]]) &, newTFix0]],	(* keys are disjoint *)
									(* We cannot 'KeyTake[ ...]' here, since 'newTFix1' must contain *all* a.b.f. type constants,
										not only those with non-default sorts. *)
						newTVar1 = Join[ newTVar, KeyTake[ termTVar, newTVar0]]},	(* keys are disjoint *)
					If[ ListQ[ type],
						Replace[ getFullType[ type, HoldComplete[ term], Join[ tfix, newTFix1], newTVar1, getConstDecls[ True]],
							{
								{newTVar2_, full_} :> {full, newMax, newTermTFix, newTermTVar, newTFix1, newTVar2},
								_ :> (sowMessage[ 110, {term, type}]; $Failed)
							}
						],
					(*else*)
						Replace[ getTypeInstance[ term, type, Join[ tfix, newTFix1], newTVar1],
							{
								{newTVar2_, _List} :> {type, newMax, newTermTFix, newTermTVar, newTFix1, newTVar2},
								{newTVar2_, _} :>
									If[ TrueQ[ $strictTypeChecking],
										sowMessage[ 109, {term}];
										$Failed,
									(*else*)
										{type, newMax, newTermTFix, newTermTVar, newTFix1, newTVar2}
									],
								_ :>
									If[ TrueQ[ $strictTypeChecking],
										sowMessage[ 111, {term, type}];
										$Failed,
									(*else*)
										{type, newMax, newTermTFix, newTermTVar, newTFix1, newTVar1}
									]
							}
						]
					]
				],
			_ :> (sowMessage[ 102, {type}]; $Failed)
		}
	]
checkTypeAux[ HoldComplete[ c:(_Complex|_DirectedInfinity|_Rational)], __] :=
	(sowMessage[ 112, {c}]; $Failed)
checkTypeAux[ v:HoldComplete[ _VAR$], _, _, _, loose_, args__] :=
	If[ isWellFormedAtom[ v],
		With[ {i = v[[1, 1]]},
			If[ i < Length[ loose],
				{loose[[i + 1]], args},
			(*else*)
				sowMessage[ 104, {First[ v]}];
				$Failed
			]
		],
	(*else*)
		sowMessage[ 103, {v}];
		$Failed
	]
checkTypeAux[ term:HoldComplete[ c_FIX$], fix_, _, _, _, args__] :=
	If[ isWellFormedAtom[ term],
		With[ {t0 = Replace[ c, fix]},
			If[ t0 === c,
				$Failed,
			(*else*)
				{t0, args}
			]
		],
	(*else*)
		sowMessage[ 103, {c}];
		$Failed
	]
checkTypeAux[ term:HoldComplete[ v_META$], _, meta_, _, _, args__] :=
	If[ isWellFormedAtom[ term],
		With[ {t0 = Replace[ v, meta]},
			If[ ListQ[ t0],	(* 't0' is of the form '{type, constIndex}' *)
				{First[ t0], args},
			(*else*)
				$Failed
			]
		],
	(*else*)
		sowMessage[ 103, {v}];
		$Failed
	]
checkTypeAux[ HoldComplete[ Lambda$TM[ RNG$[ TYPRNG$[ var_, typ_]], expr_]], fix_, meta_, tfix_, loose_, max_, termTFix_, termTVar_, newTFix_, newTVar_] :=
	If[ MatchQ[ var, _Symbol],
		Replace[ isWellFormedType[ typ, Join[ tfix, newTFix], newTVar],
			{
				{newTFix0_List, newTVar0_List} :>
					Replace[
						checkTypeAux[
							HoldComplete[ expr],
							fix,
							meta,
							tfix,
							Prepend[ loose, typ],
							Max[ max, Max[ newTVar0[[All, 1]]] + 1],
							KeyDrop[ termTFix, newTFix0],
							KeyDrop[ termTVar, newTVar0],
							Join[ newTFix, Association @@ Map[ (# -> Lookup[ termTFix, #, getDefaultSort[]]) &, newTFix0]],
							Join[ newTVar, KeyTake[ termTVar, newTVar0]]
						],
						{rt_, res__} :> {FUN$TM[ typ, rt], res}
					],
				_ :> (sowMessage[ 102, {typ}]; $Failed)
			}
		],
	(*else*)
		sowMessage[ 108, {var}];
		$Failed
	]
checkTypeAux[ HoldComplete[ f_[ a_]], fix_, meta_, tfix_, loose_, args__] :=
	Replace[ checkTypeAux[ HoldComplete[ f], fix, meta, tfix, loose, args],
		{
			{FUN$TM[ at_, rt_], res0__} :>
				Replace[ checkTypeAux[ HoldComplete[ a], fix, meta, tfix, loose, res0],
					{
						{at, res__} :> {rt, res},
						{ta_, __} :> (sowMessage[ 105, {FUN$TM[ at, rt], ta}]; $Failed)
					}
				],
			_ -> $Failed
		}
	]
checkTypeAux[ HoldComplete[ term:(TYPED$|VAR$|FIX$|META$|SVAR$|TVAR$|TFIX$|SEQ$|SEQ0$|SEQ1$|TYPRNG$|RNG$|Lambda$TM)], __] :=
	(sowMessage[ 101, SymbolName[ term]]; $Failed)
checkTypeAux[ HoldComplete[ c_], __] :=
	(sowMessage[ 112, {c}]; $Failed)

typeSpecializationReplacementAux[ {}, tfix_, repl_] :=
	repl
typeSpecializationReplacementAux[ vars_, tfix_, repl_] :=
	Join[
		repl,
		With[ {defSym = Theorema`Knowledge`VAR$\[Alpha]$TM},
		With[ {fixIdx = Max[ -1, Max @@ Join[ Cases[ tfix, TFIX$[ defSym, i_Integer] :> i], Cases[ repl, (_ -> TFIX$[ defSym, i_Integer]) :> i]]],
				fixSyms =
					DeleteCases[
						Map[
							("VAR$" <> # <> "$TM")&,
							{"\[Alpha]", "\[Beta]", "\[Gamma]", "\[Delta]", "\[Epsilon]", "\[Zeta]", "\[CurlyTheta]",
								"\[Kappa]", "\[Mu]", "\[Nu]", "\[Xi]", "\[Sigma]", "\[CurlyPhi]", "\[Chi]", "\[Psi]"}
						],
						Alternatives @@ Join[ Cases[ tfix, TFIX$[ s_Symbol, _] :> SymbolName[ s]], Cases[ repl, (_ -> TFIX$[ s_Symbol, _]) :> SymbolName[ s]]]
					]},
		With[ {fixLen = Length[ fixSyms]},
			MapIndexed[
				#1 -> With[ {i = First[ #2]},
					If[ i <= fixLen,
						TFIX$[ Symbol[ "Theorema`Knowledge`" <> fixSyms[[i]]], 0],
					(*else*)
						TFIX$[ defSym, fixIdx + i - fixLen]
					]
				]&,
				vars
			]
		]]]
	]


(* ::Section:: *)
(*Locking Symbols*)

(*
Scan[
	SetAttributes[ #, {Protected, Locked}]&,
	{
		typeSpecializationReplacementAux, checkTypeAux
	}	(* TODO: Make sure that all private functions are listed here. *)
]
*)

End[]

EndPackage[]
