(* Core proof methods for automatically reducing goals of certain forms. *)

(* The following symbols appear in "Theorema`CoreProvers`CoreProofMethods`"-context. *)
Scan[
	With[ {$$name = #[[1]], $$desc = #[[2]]},
	With[ {$$args = Join[ ToExpression[ "Theorema`CoreProvers`CoreProofMethods`" <> $$name, InputForm, Hold], Hold[ "usage", "English"]]},
		Set @@ ReplacePart[ Hold[ $$args, $$desc], {1, 0} -> MessageName];
	]]&,
	{
		{"$autoReduceRules", ""}
	}
]

(* ::Subsection:: *)
(*Reduction*)

(* 'goal' is a plain term, 'assms' is a list of plain terms, 'hyps' is a list of 'FML$'-structures. *)
reduceGoal[ goal_, assms_List, hyps_, idx_, tfix_, tvar_, allowSkip_, rest___] :=
	Module[ {X, fun},
		(* For memoization, 'idx0', 'metaNum0' and 'tvar0' should be irrelevant ... *)
		fun[ idx0_, metaNum0_, tvar0_, goal0_] := (fun[ _, _, _, goal0] = reduceGoalSingle[ {idx0, metaNum0, X}, goal0, idx, tfix, tvar0, assms, hyps, rest]);
		searchBackwardsProof[
			{goal},
			idx,
			0,	(* This number refers to the meta variables with name 'X', not to existing ones! *)
			tvar,
			fun,
			acceptNoMeta[ X, allowSkip],
			allowSkip,
			Infinity	(* As soon as the total number of alternative branches reaches the given value, 'ParallelMap' is applied. There seems to be quite some overhead, though. *)
		]
	]

reduceGoalSingle[ goalData:{idx_, metaNum_, X_}, goal_, origIdx_, tfix_, tvar_, assms_, hyps_] :=
	Catch[
		MapIndexed[
			With[ {res = applyReductionRule[ {makeSTERM[ #1], {}, {}, idx, metaNum, X, 0, 0}, {}, goal, origIdx, tfix, tvar, #2]},
				If[ ListQ[ res], Throw[ {res}]]
			]&,
			assms
		];
		MapIndexed[
			With[ {hyp = extractFormula[ #1, idx]},
			With[ {res = applyReductionRule[ {First[ hyp], {}, {}, idx, metaNum, X, Last[ hyp] - idx, 0}, {}, goal, origIdx, tfix, tvar, First[ #2]]},
				If[ ListQ[ res], Throw[ {res}]]
			]]&,
			hyps
		];
		SortBy[
			Cases[
				Map[
					applyReductionRule[ #, goalData, goal, origIdx, tfix, tvar, #]&,
					List @@ extractRelevantRules[ $autoReduceRules, getRuleConsts[ goal, None, False]]	(* we need the atomic head *)
				],
				{_, _, _, _, _, __}
			],
			Length[ #[[4]]]&
		]
	]
reduceGoalSingle[ goalData:{idx_, metaNum_, X_}, goal_, origIdx_, tfix_, tvar_, assms_, hyps_, hypFactIndices_, hypRuleIndices_, globRules_Association] :=
	Catch[
		MapIndexed[
			With[ {res = applyReductionRule[ {makeSTERM[ #1], {}, {}, idx, metaNum, X, 0, 0}, {}, goal, origIdx, tfix, tvar, #2]},
				If[ ListQ[ res], Throw[ {res}]]
			]&,
			assms
		];
		Map[
			With[ {hyp = extractFormula[ hyps[[#]], idx]},
			With[ {res = applyReductionRule[ {First[ hyp], {}, {}, idx, metaNum, X, Last[ hyp] - idx, 0}, {}, goal, origIdx, tfix, tvar, #]},
				If[ ListQ[ res], Throw[ {res}]]
			]]&,
			hypFactIndices
		];
		SortBy[
			Cases[
				Join[
					Map[
						applyReductionRule[ hyps[[#]], goalData, goal, origIdx, tfix, tvar, #]&,
						hypRuleIndices
					],
					Map[
						applyReductionRule[ #, goalData, goal, origIdx, tfix, tvar, #]&,
						List @@ extractRelevantRules[ globRules, getRuleConsts[ goal, None, False]]		(* we need the atomic head *)
					]
				],
				{_, _, _, _, _, __}
			],
			Length[ #[[4]]]&
		]
	]

(* There must not be new meta variables, even if all subgoals can be proved. This strong condition could be weakened,
	but then we would have to mess around with the types of meta variables ... *)
acceptNoMeta[ X_, allowSkip_][ hist_] :=
	If[ !MemberQ[ hist, {_, _, {_, _?(!FreeQ[ #, META$[ X, _]]&), _}, ___}],
		With[ {skipPos = Cases[ hist, {_, p_, Skip, ___} :> p]},
			Which[
				skipPos === {},
				Throw[ hist],
			(*else*)
				allowSkip && !MemberQ[ hist, {_, p_, _, False} /; With[ {lp = Length[ p]}, MemberQ[ skipPos, _?(lp < Length[ #] && Take[ #, lp] === p &)]]],
				Sow[ hist]
			]
		]
	]


(* ::Subsection:: *)
(*Proof Methods*)

If[ !MemberQ[ $terminalProofSteps, autoReduceTerminalStep], AppendTo[ $terminalProofSteps, autoReduceTerminalStep]];

autoReduceTerminalStep[ G_, hyps_, idx_, _, _, tfix_, tvar_] :=
	With[ {res = (initAutoReduceRules[]; autoReduceAux[ G, Placeholder, hyps, idx, tfix, tvar, False])},
		If[ res === $Failed,
			res,
		(*else*)
			Function[ res] /. Placeholder -> Slot[ 1]
		]
	]

(* 'arVariant[ args, factID, keyHandler, lbl]' is an auxiliary proof method used for equipping the given formulas with
	reduction rules. The formulas themselves are left unchanged. *)
arVariant[ args_List, factID_, keyHandler_, lbl_, add_, rem_][ ps_, __] :=
	Hold @@ {
		updateAbbrevLists[ {}, {}, {}, {factID}],
		variantIterator[
			getProofSitData[ ps, factID, {}],
			arVariantFun,
			keyHandler,
			lbl,
			add,
			rem
		]
	}

arVariantFun[ i_, PRFSIT$[ _, hyps_, __]] :=
	With[ {hyp = hyps[[i]]},
		If[ !MemberQ[ hyp, "arData" -> _],
			With[ {arData = generateReductionRule[ hyp, True]},
				If[ arData === $Failed,
					Null,
				(*else*)
					Append[ DeleteCases[ List @@ Drop[ hyp, 3], "TypeVarPositions" -> _], "arData" -> arData]
				]
			],
		(*else*)
			Null
		]
	]

(* 'autoReduce[ spec][ ps, __]' auto-reduces the first subgoal in 'ps' w.r.t. facts tagged by "autored" or "autored?"
	and facts explicitly specified in 'spec' (which is as in rewriting, i.e. of the form "only: add: rem:").
	'spec' may also determine whether subgoals may be skipped (default) or not. *)
autoReduce[ spec_][ ps_PRFSIT$, args__] :=
	Replace[ parseARSpec[ spec, ps, "autoReduce"],
		{hyps_List, glob_Association, variants_List, allowSkip_} :>
			With[ {id = getFreshFactID[]},
				If[ variants === {},
					autoReduce[ id, {hyps, glob, allowSkip}][ ps, args],
				(*else*)
					Hold[
						addKnowledgeAux[ variants, addKeysToList[ id, False], Hold[], 3, None],
						autoReduce[ id, {hyps, glob, allowSkip}]
					]
				]
			]
	]
autoReduce[ argsID_, {hypIndices0_List, glob_Association, allowSkip_}][ ps:PRFSIT$[ {G_, ___}, hyps_, idx_, _, _, tfix_, tvar_, __], __] :=
	With[ {argKeys0 = getProofSitData[ ps, argsID, {}]},
	With[ {tempKeys = Cases[ argKeys0, {k:{_, _}, False} :> k], argKeys = Replace[ argKeys0, {k:{_, _}, _} :> k, {1}]},
	With[ {newHyps =
			Cases[
				Map[
					With[ {i = First[ FirstPosition[ hyps, FML$[ #, __], {$Failed}, {1}, Heads -> False]]},
						If[ !IntegerQ[ i], sowCPMMessage[ 158, {#}, "Warning"]];
						i
					]&,
					argKeys
				],
				_Integer
			]},
	With[ {hypIndices = Join[ hypIndices0, newHyps]},
	With[ {res = autoReduceAux[ G, 0, hyps, idx, tfix, tvar, allowSkip, Delete[ Range[ Length[ hyps]], List /@ hypIndices], hypIndices, glob]},
		If[ MatchQ[ res, _Hold],
			Join[ res, If[ tempKeys === {}, Hold[], Hold[ forgetHyps[ tempKeys]]]],
		(*else*)
			$Failed
		]
	]]]]]

autoReduceAux[ G0_, addAssms_, hyps_, idx_, tfix_, tvar_, allowSkip_, rest___] :=
	With[ {ag = toAssmsGoal[ G0, {}, Infinity]},
	With[ {G = Last[ ag]},
		Replace[ reduceGoal[ G, Most[ ag], hyps, idx, tfix, tvar, allowSkip, rest],
			{
				hist_List :>
					With[ {n = Length[ ag] - 1},
						constructAutoReductionProof[ hist, n, addAssms]
					],
				_ -> $Failed
			}
		]
	]]

parseARSpec[ spec_List, ps:PRFSIT$[ _, hyps_, __], src_String] :=
	Replace[ parseAddRemOnlySpec[ spec, {""}, src],
		{gen_List, facts_List} :>
			Catch[
				Append[
					Block[ {$getReductionRulesPS = ps,
							$getReductionRulesHypNum = Length[ hyps] + 1,
							$getReductionRulesErrFun = (sowCPMMessage[ Switch[ #2, 0, 173, 1, 174, _, 19], {#1}, "Fatal"]; Throw[ $Failed])&},
						getReductionRules[ facts]
					],
					!MemberQ[ gen, "noSkip"|{"noSkip"}]
				]
			]
	]

getReductionRules[ {0 -> only_List, rest___}] :=
	Module[ {glob = Association[]},
		Replace[ transposeEmpty[ getReductionRulesAux[ only, True], 4],
			{_, addHyps_, addGlob_List, addVariants_} :>
				getReductionRules[
					{rest},
					Join @@ addHyps,
					Scan[ addRulesToAssociation[ glob, Prepend, None], Normal[ groupRulesByRuleConsts[ Join @@ addGlob, "intro", True]]];
					Hold[ glob],
					Join @@ addVariants
				]
		]
	]
getReductionRules[ spec_] :=
	Module[ {glob = initAutoReduceRules[]},
		getReductionRules[ spec, {}, Hold[ glob], {}]
	]
getReductionRules[ {}, hyps_, Hold[ glob_], variants_] :=
	{hyps, glob, variants}
getReductionRules[ {1 -> add_List, rest___}, hyps_, Hold[ glob_], variants_] :=
	Module[ {dup},
	With[ {aux =
			Fold[
				Function[ {old, new},
					Replace[ new,
						{k_, addHyps_, addGlob_, addVariants_} :>
							With[ {out =
									{
										dup = (addVariants === {});
										addToList[ First[ old], addHyps, (dup = False; #)&, Identity, False],
										Join[ Last[ old], addVariants]
									}},
								Scan[ addRulesToAssociation[ glob, Prepend, If[ #1 === Hold[], dup = False]&], Normal[ groupRulesByRuleConsts[ addGlob, "intro", True]]];
								If[ TrueQ[ dup], sowCPMMessage[ 171, If[ ListQ[ k], {k}, {}], "Warning"]];
								out
							]
					]
				],
				{hyps, variants},
				getReductionRulesAux[ add, True]
			]},
		getReductionRules[ {rest}, First[aux], Hold[ glob], Last[ aux]]
	]]
getReductionRules[ {2 -> rem_List, rest___}, hyps_, Hold[ glob_], variants_] :=
	Module[ {missing},
	With[ {aux =
			Fold[
				Function[ {old, new},
					Replace[ new,
						{k_, remHyps_, remGlob_} :>
							With[ {out = (missing = True; remFromList[ old, remHyps, (missing = False; #)&, Identity, False])},
								Scan[ remRulesFromAssociation[ glob, If[ #2 =!= 0, missing = False]&], Normal[ groupRulesByRuleConsts[ remGlob, "intro", True]]];
								If[ TrueQ[ missing], sowCPMMessage[ 172, If[ ListQ[ k], {k}, {}], "Warning"]];
								out
							]
					]
				],
				hyps,
				getReductionRulesAux[ rem, False]
			]},
		getReductionRules[ {rest}, aux, Hold[ glob], variants]
	]]

getReductionRulesAux[ l_List, allowVariants_] :=
	Join @@ Last[
		Reap[
			scanFactSpec[ l, $getReductionRulesPS, "autored", !allowVariants, getReductionRulesAux[ Null, allowVariants], $getReductionRulesErrFun],
			True
		]
	]
getReductionRulesAux[ Null, allowVariants_][ l_List] :=
	Sow[
		Prepend[
			Join @@@ Last[
				Reap[
					Scan[ getReductionRulesAux[ allowVariants], l],
					Range[ 1, If[ TrueQ[ allowVariants], 3, 2]]
				]
			],
			FirstCase[ l, {_, FML$[ k_, __]} :> k, None]
		],
		True
	]
getReductionRulesAux[ allowVariants_][ {k_List, fml_FML$}] :=
	If[ TrueQ[ allowVariants] && getOptionalComponent[ fml, "arData"] === {},
		Sow[ VARIANT$[ k, arVariant, {}], 3],
	(*else*)
		With[ {p = First[ FirstPosition[ $getReductionRulesPS[[2]], _[ k, __], {$Failed}, {1}, Heads -> False]]},
			If[ p === $Failed,
				Sow[ k, 2],
			(*else*)
				Sow[ p - $getReductionRulesHypNum, 1]
			]
		]
	]
getReductionRulesAux[ allowVariants_][ {i_Integer, fml_FML$}] :=
	If[ TrueQ[ allowVariants] && getOptionalComponent[ fml, "arData"] === {},
		Sow[ VARIANT$[ key@fml, arVariant, {}], 3],
	(*else*)
		Sow[ i - $getReductionRulesHypNum, 1]
	]
getReductionRulesAux[ allowVariants_][ {thm_THM$, _}] :=
	If[ TrueQ[ allowVariants] && Lookup[ getThmData[ thm], "arData", $Failed] === $Failed,
		Sow[ VARIANT$[ thm, arVariant, {}], 3],
	(*else*)
		Sow[ thm, 2]
	]
getReductionRulesAux[ allowVariants_][ {v_VARIANT$, _}] :=
	If[ TrueQ[ allowVariants],
		Sow[ VARIANT$[ v, arVariant, {}], 3],
	(*else*)
		sowCPMMessage[ 175, "Warning"]
	]

initAutoReduceRules[] :=
	If[ MatchQ[ $autoReduceRules, _Association],
		$autoReduceRules,
	(*else*)
		(* Note the difference between '$autoReduceRules' and '$autoredRules'! *)
		extendRULEC$[ $autoredRules];
		$autoReduceRules = associationFromRULEC$[ $autoredRules]
	]


(* ::Subsection:: *)
(*Proof Generation*)

constructAutoReductionProof[ hist0_, ac_Integer, addAssms_] :=	(* 'addAssms' might be instantiated by a symbolic value! *)
	Catch[
		With[ {subgoalGroups =
				Module[ {multi = DeleteDuplicates[ Cases[ hist0, {p_List, _, Null, ___} :> p]]},
					If[ multi === {},
						{},
					(*else*)
						First[
							Last[
								Reap[
									While[ multi =!= {},
										With[ {i =
												Position[ multi,
													p_List /; With[ {lp = Length[ p]}, !MemberQ[ multi, _List?(lp < Length[ #] && Take[ #, lp] === p &)]],
													{1},
													Heads -> False
												]},
											Sow[ Extract[ multi, i], True];
											multi = Delete[ multi, i];
										]
									],
									True
								]
							]
						]
					]
				],
				envKeys = DeleteDuplicates[ Cases[ hist0, {_, _, {k:{_String, _String}, _, _}, ___} :> k]]},
		With[ {newHyps = Length[ envKeys]},
		Module[ {newAssms = 0,
				hist =
					If[ newHyps === 0 === addAssms,
						hist0,
					(*else*)
						Replace[ hist0,
							{
								{g_, p_, {k:{_String, _String}, inst_, lift_}, ___} :>
									{g, p, {newHyps + 1 - First[ FirstPosition[ envKeys, k, Throw[ $Failed], {1}, Heads -> False]], inst, lift}},
								{g_, p_, {i_Integer, inst_, lift_}, ___} :>
									{g, p, {i + newHyps, inst, lift}},
								{g_, p_, {{i_Integer}, inst_, lift_}, ___} :>
									{g, p, {{i + addAssms}, inst, lift}}
							},
							{1}
						]
					]},
			Hold @@ Join[
				INF$[ "insert", #]& /@ envKeys,
				Join @@ Map[
					Function[ subgoals,
						With[ {i = Map[ First[ FirstPosition[ hist, {_, #, __}, Throw[ $Failed], {1}, Heads -> False]]&, subgoals], na = newAssms + ac + addAssms},
							newAssms += Length[ i];
							{
								setupSubgoalsMP[ hist[[i, 1]], -1],	(* 'i' is a list of indices! *)
								autoredProof[
									Join @@ MapIndexed[
										Function[ {s, j},
											With[ {pos = hist[[s, 2]], newRed = {j + na, {}, Infinity}},
											With[ {lpos = Length[ pos]},
											With[ {sub = Position[ hist, {_, _List?(lpos < Length[ #] && Take[ #, lpos] === pos &), __}, {1}, Heads -> False]},
											With[ {ret = Prepend[ Extract[ hist, sub][[All, 3]], hist[[s, 3]]]},
												hist = Replace[ Delete[ hist, sub], {pos, p_, Null, ___} :> {pos, p, newRed}, {1}];
												hist[[ s, 3]] = newRed;
												ret
											]]]]
										],
										i
									],
									0
								]
							}
						]
					],
					subgoalGroups
				],
				{autoredProof[ hist[[All, 3]], newHyps]}
			]
		]]]
	]

(* 'autoredProof[ seq, forget]' is an auxiliary proof method that iterates through 'seq' and applies the corresponding
	proof methods. 'seq' is supposed to originate from 'constructAutoReduceProof', and 'forget' determines how many
	implicit hypotheses shall be forgotten in the end. *)
autoredProof[ {}, forget_Integer][ PRFSIT$[ G_, __], __] :=
	If[ G === {},
		Hold[],
	(*else*)
		Hold @@ Table[ INF$[ "forget", 1], {forget}]
	]
autoredProof[ {first_, rest___}, forget_][ PRFSIT$[ G_, _, idx_, __], __] :=
	Hold @@ Append[
		Replace[ first,
			{
				Skip :> With[ {len = Length[ G]}, If[ len <= 1, {}, {INF$[ "rearrangeG", 1, len]}]],
				{thm_THM$, inst_List, lift_} :> Prepend[ autoredRed[ 1, inst, lift, idx], INF$[ "theorem", thm]],
				{i_Integer, inst_List, lift_} :> Prepend[ autoredRed[ 1, inst, lift, idx], INF$[ "makeExplicit", i]],
				{{i_Integer}, inst_List, lift_} :> autoredRed[ i, inst, lift, idx]
			}
		],
		autoredProof[ {rest}, forget]
	]

autoredRed[ i_Integer, inst0_List, lift_, idx_Integer] :=
	With[ {diff = idx - lift},
	With[ {inst = If[ Positive[ diff], inst0 /. TVAR$[ j_Integer /; j >= lift, rest___] :> TVAR$[ j + diff, rest], inst0]},
		Append[ INF$[ "allE", i, makeSTERM[ #]]& /@ inst, INF$[ "impE", i, 0]]
	]]


(* ::Subsection:: *)
(*Search*)

(* A state is structure of the form 'STATE$[ idx, metaNum, tvar, hist][ goals...]' where 'idx' and 'metaNum' are the
	current type variable index and number of meta variables, respectively, 'tvar' are the current sort constraints on
	type variables, and 'hist' is a list of 3/4-tuples '{goal, pos, red, (safe)}', where 'goal' is either the actual
	formula or the position ('pos') of a preceding entry whose goal equals the actual formula, 'pos' is the position of
	'goal' in the derivation tree, 'red' is information on how to reduce the goal (e.g. by which fact), and 'safe'
	specifies whether the reduction 'red' is "safe", i.e. reduces 'goal' to a collection of new goals whose conjunction
	is equivalent to 'goal'. 'safe' can be omitted, in which case it defaults to 'True'. *)

(* 'searchBackwardsProof' returns either '$Failed' or the complete history of a backwards proof of the given list of
	goals. If there are several candidates, the first with the least number of skipped goals is returned. The given
	function 'fun' is expected to return a list of 6/7-tuples
	'{newIdx, newMetaNum, newTVar, red, newGoals, op, (safe)}', where
	- 'newIdx' and 'newMetaNum' are the type variable index and number of meta variables, respectively, of each new
		state,
	- 'newTVar' is the new association of sort constraints,
	- 'red' is information on how to reduce the original goal in order to arrive at 'newGoals',
	- 'newGoals' is the list of new goals of each new state (could be empty), and
	- 'op' is a function that shall be applied to all existing goals (both in the state and in the history; could be
		'Identity'),
	- 'safe' specifies whether the reduction is "safe" or not; if omitted, it is assumed to be 'True'. *)
searchBackwardsProof[ goals_List, idx_, metaNum_, tvar_, fun_, acc_, allowSkip_, Infinity] :=
	searchBackwardsProof[ fun, acc, allowSkip][ STATE$[ idx, metaNum, tvar, {}] @@ MapIndexed[ List, goals]]
searchBackwardsProof[ goals_List, idx_, metaNum_, tvar_, fun_, acc_, allowSkip_, parallelizeNum_] :=
	Module[ {sit = {STATE$[ idx, metaNum, tvar, {}] @@ MapIndexed[ List, goals]}},
		Catch[
			selectBackwardsProof[
				Join[
					Join @@ Last[
						Reap[	(* Everything sown below is consumed by the 'Reap', even if something is thrown and caught by the enclosing 'Catch'. *)
							While[ sit =!= {} && Length[ sit] < parallelizeNum,
								sit = Join[ Rest[ sit], searchBackwardsProof[ First[ sit], fun, acc, allowSkip]]
							];
						]
					],
					If[ sit === {}, {}, ParallelMap[ searchBackwardsProof[ fun, acc, allowSkip], sit]]
				],
				$Failed
			]
		]
	]
searchBackwardsProof[ fun_, acc_, allowSkip_][ state_] :=
	Module[ {sit = {state}},
		Catch[
			selectBackwardsProof[
				Join @@ Last[
					Reap[
						While[ sit =!= {},
							sit = Join[ Rest[ sit], searchBackwardsProof[ First[ sit], fun, acc, allowSkip]]
						];
					]
				],
				$Failed
			]
		]
	]
searchBackwardsProof[ STATE$[ _, _, _, hist_][], _, acc_, allowSkip_] :=
	(acc[ hist]; {})
searchBackwardsProof[ STATE$[ idx_, metaNum_, tvar_Association, hist_][ {goal_, pos_List}, goals___], fun_, acc_, allowSkip_] :=
	With[ {goals1 = {goals},
			(* Maybe the rigid pattern '{goal, _}' could be replaced by something weaker that takes renamings of meta
				variables and type variables into account. But still one must be careful, since such variables might
				also occur elsewhere and then must be instantiated by different terms/types. *)
			v = First[ FirstPosition[ hist, {goal, __}, {$Failed}, {1}, Heads -> False]]},
		If[ IntegerQ[ v],
			With[ {p0 = hist[[v, 2]]},
			With[ {lp = Length[ p0]},
				If[ Quiet[ Check[ Take[ pos, lp] === p0, False]],
					(* cycle detected *)
					If[ TrueQ[ allowSkip],
						(* If entries of 'hist' are deleted, 'metaNum' might eventually be larger than necessary. *)
						skipGoal[
							idx,
							metaNum,
							tvar,
							ReplacePart[ Take[ hist, v], {-1, 3} -> Skip],
							DeleteCases[ goals1, {_, p1_} /; Quiet[ Check[ Take[ p1, lp] === p0, False]]],
							acc
						],
					(*else*)
						{}
					],
				(*else*)
					With[ {hist1 = Append[ hist, {p0, pos, Null}]},
						If[ goals1 === {},
							acc[ hist1]; {},
						(*else*)
							{STATE$[ idx, metaNum, tvar, hist1][ goals]}
						]
					]
				]
			]],
		(*else*)
			Replace[ fun[ idx, metaNum, tvar, goal],
				{
					l:Except[ {}, _List] :>
						MapIndexed[
							Replace[ #1,
								{
									{newIdx_, newMetaNum_, newTVar_Association, red_, newGoals_List, op_, safe___} :>
										With[ {hist1 =
													Append[
														If[ op === Identity || hist === {}, hist, MapAt[ op, hist, {All, {1, 3}}]],
														{op[ goal], pos, op[ red], safe}
													]},
											If[ newGoals === {} === goals1,
												acc[ hist1]; Sequence @@ {},
											(*else*)
												STATE$[ newIdx, newMetaNum, newTVar, hist1] @@ Join[
													MapIndexed[ {#1, Join[ pos, #2]}&, newGoals],
													If[ op === Identity || goals1 === {}, goals1, MapAt[ op, goals1, {All, 1}]]
												]
											]
										],
									_ -> Sequence[]
								}
							]&,
							l
						],
					_ :>
						If[ TrueQ[ allowSkip],
							skipGoal[ idx, metaNum, tvar, Append[ hist, {goal, pos, Skip}], goals1, acc],
						(*else*)
							{}
						]
				}
			]
		]
	]

skipGoal[ idx_, metaNum_, tvar_, hist_, goals_List, acc_] :=
	If[ goals === {},
		acc[ hist]; {},
	(*else*)
		{STATE$[ idx, metaNum, tvar, hist] @@ goals}
	]

selectBackwardsProof[ l_, def_] :=
	selectBackwardsProof[ l, True, Infinity, 0, def]
selectBackwardsProof[ {}, _, _, _, out_] := out
selectBackwardsProof[ {first_, rest___}, skipOrig_, skipNum_, skipDepth_, out_] :=
	With[ {so = MemberQ[ first, {_, {}|{_}, Skip, ___}]},
		If[ so && !skipOrig,
			selectBackwardsProof[ {rest}, skipOrig, skipNum, skipDepth, out],
		(*else*)
			With[ {sn = Count[ first, {_, _, Skip, ___}]},
				Which[
					sn === 0,
					first,
				(*else*)
					so || skipNum < sn,
					selectBackwardsProof[ {rest}, skipOrig, skipNum, skipDepth, out],
				(*else*)
					True,
					With[ {sd = Total[ Cases[ first, {_, _, p_, Skip, ___} :> Length[ p]]]},	(* Instead of / in addition to 'Total' one could also try 'Max' and/or 'Min'. *)
						If[ !so || sn < skipNum || skipDepth < sd,
							selectBackwardsProof[ {rest}, so, sn, sd, first],
						(*else*)
							selectBackwardsProof[ {rest}, skipOrig, skipNum, skipDepth, out]
						]
					]
				]
			]
		]
	]
