(* Core proof methods for calculational (i.e. symmetry and transitivity) reasoning. *)

(* ::Subsection:: *)
(*Symmetry*)

(* 'symmetricVariant[ args, factID, keyHandler, lbl, add, rem]' is an auxiliary proof method for obtaining the
	symmetric variants of known facts. This is accomplished by "returning", for each given fact 'fact', the first
	successful 'sym[OF fact]', where 'sym' iterates through all known facts (implicit hypotheses, theorems, background
	knowledge) declared as symmetry rules by tag "sym".
	Note: In contrast to Isabelle, simply the *first* successful variant is returned, even if there are more than one.
	It is also possible to specify the desired symmetry rules explicitly; in that case, the proof method only iterates
	through the given rules (which do not need to be declared as symmetry rules, of course). *)
symmetricVariant[ {}, factID_, keyHandler_, lbl_, add_, rem_][ ps_, __] :=
	With[ {factKeys = getProofSitData[ ps, factID, {}]},
		Hold[
			updateAbbrevLists[ {}, {}, {}, {factID}],
			variantIterator[ factKeys, symmetricVariantFun[], keyHandler, lbl, add, rem]
		]
	]
symmetricVariant[ rules_List, factID_, keyHandler_, lbl_, add_, rem_][ __] :=
	(* shall be unfolded only when applied *)
	With[ {id = getFreshFactID[]},
		$addKnowledgeSource = 3;
		Hold[
			addKnowledgeAux[ rules, addKeysToList[ id, False], Hold[], Automatic, "sym"],
			symmetricVariant[ id, factID, keyHandler, lbl, add, rem]
		]
	]
symmetricVariant[ ruleID_, factID_, keyHandler_, lbl_, add_, rem_][ ps_, __] :=
	With[ {factKeys = getProofSitData[ ps, factID, {}], ruleKeys0 = getProofSitData[ ps, ruleID, {}]},
	With[ {tempRules = Cases[ ruleKeys0, {k:{_, _}, False} :> k], ruleKeys = Replace[ ruleKeys0, {k:{_, _}, _} :> k, {1}]},
		Join[
			Hold[
				updateAbbrevLists[ {}, {}, {}, {factID, ruleID}],
				variantIterator[ factKeys, symmetricVariantFun[ ruleKeys], keyHandler, lbl, add, rem]
			],
			If[ tempRules === {}, Hold[], Hold[ forgetHyps[ tempRules]]]
		]
	]]

symmetricVariantFun[][ i_, PRFSIT$[ _, hyps_, _, fix_, _, tfix_Association, __]] :=
	Replace[ extractFormula[ hyps[[i]], 0],
		{STERM$[ fml_, Association[], tvar1_Association], idx_} :>
			Catch[
				Module[ {j = 0},
					Scan[
						Function[ sym0,
							++j;
							If[ MemberQ[ getOptionalComponent[ sym0, "tags"], "sym"],
								Replace[ extractFormula[ sym0, idx],
									{STERM$[ sym_, Association[], tvar2_Association], _} :>
										Replace[ resolveImplicationOFAuto[ {fml}, sym, hyps, idx, fix, tfix, Join[ tvar1, tvar2]],	(* keys are disjoint *)
											{new_STERM$, pd_List} :> Throw[ {{new, OFVariantProof[ j, {i}, pd]}}]
										]
								]
							]
						],
						hyps
					];
					j = getRuleConsts[ fml, "elim", False];
					extendRULEC$[ $symRules];
					While[
						Replace[ iterateRULEC$[ j, $symRules],
							{
								k_List :>
									(
										If[ !MemberQ[ hyps, FML$[ k, __]],
											With[ {sym0 = getFormulas[ True][ k]},
												Replace[ extractFormula[ sym0, idx],
													{STERM$[ sym_, Association[], tvar2_Association], _} :>
														Replace[ resolveImplicationOFAuto[ {fml}, sym, hyps, idx, fix, tfix, Join[ tvar1, tvar2]],
															{new_STERM$, pd_List} :> Throw[ {{new, symmetricVariantProof[ i, k, pd]}}]
														]
												]
											]
										];
										True
									),
								thm_THM$ :>
									(
										Replace[ getThmInst[ True, idx][ thm],
											{STERM$[ sym_, Association[], tvar2_Association], _} :>
												Replace[ resolveImplicationOFAuto[ {fml}, sym, hyps, idx, fix, tfix, Join[ tvar1, tvar2]],
													{new_STERM$, pd_List} :> Throw[ {{new, symmetricVariantProof[ i, thm, pd]}}]
												]
										];
										True
									),
								_ -> False
							}
						]
					]
				];
				sowCPMMessage[ 131, "Fatal", "symmetric"];
				$Failed
			]
	]
symmetricVariantFun[ keys_List][ i_, PRFSIT$[ _, hyps_, _, fix_, _, tfix_Association, __]] :=
	Replace[ extractFormula[ hyps[[i]], 0],
		{STERM$[ fml_, Association[], tvar1_Association], idx_} :>
			Catch[
				Scan[
					Function[ k,
						Replace[ Position[ hyps, FML$[ Replace[ k, {ky:{_, _}, _} :> ky], __], {1}, 1, Heads -> False],
							{{j_Integer}} :>
								Replace[ extractFormula[ hyps[[j]], idx],
									{STERM$[ sym_, Association[], tvar2_Association], _} :>
										Replace[ resolveImplicationOFAuto[ {fml}, sym, hyps, idx, fix, tfix, Join[ tvar1, tvar2]],	(* keys are disjoint *)
											{new_STERM$, pd_List} :> Throw[ {{new, OFVariantProof[ j, {i}, pd]}}]
										]
								]
						]
					],
					keys
				];
				sowCPMMessage[ 132, "Fatal", "symmetric"];
				$Failed
			]
	]

symmetricVariantProof[ i_Integer, k_List, pd_][ n_Integer] :=
	Prepend[
		OFVariantProof[ 0, {i}, pd][ n + 1],
		INF$[ "insert", k]
	]
symmetricVariantProof[ i_Integer, thm_THM$, pd_][ n_Integer] :=
	OFVariantProof[ {INF$[ "theorem", thm]}, {i + n}, pd]


(* ::Subsection:: *)
(*Transitivity*)

(* 'transitiveVariant[ args, factID, keyHandler, lbl, add, rem]' is an auxiliary proof method for obtaining the
	transitive variant of the given facts altogether. This means that unlike 'symmetricVariant', 'OFVariant',
	'ofVariant' etc., 'transitiveVariant' does not consider the given facts individually and for each fact produces one
	new fact, but instead searches for suitable transitivity theorems it can apply to the whole list of facts at once
	(using 'OFVariant', of course). Apart from that, the proof method proceeds similarly as 'symmetricVariant'; in
	particular, it is also possible to specify the transitivity rules to be tried explicitly. *)
transitiveVariant[ {}, factKeys_List, keyHandler_, lbl_, add_, rem_][ args__] :=
	If[ factKeys === {},
		Hold[],
	(*else*)
		transitiveVariantAux[ factKeys, Null, keyHandler, lbl, add][ args]
	]
transitiveVariant[ {}, factID:Except[ _List], keyHandler_, lbl_, add_, rem_][ ps:PRFSIT$[ _, hyps_, __], __] :=
	With[ {factKeys = getProofSitData[ ps, factID, {}]},
		Prepend[
			If[ factKeys === {}, Hold[], Hold[ transitiveVariantAux[ factKeys, Null, keyHandler, lbl, add]]],
			updateAbbrevLists[ {}, {}, {}, {factID}]
		]
	]
transitiveVariant[ args_List, facts_, keyHandler_, lbl_, add_, rem_][ __] :=
	If[ facts === {},
		Hold[],
	(*else*)
		(* shall be unfolded only when applied *)
		With[ {id = getFreshFactID[]},
			$addKnowledgeSource = 3;
			Hold[
				addKnowledgeAux[ args, addKeysToList[ id, False], Hold[], Automatic, "trans"],
				transitiveVariant[ id, facts, keyHandler, lbl, add, rem]
			]
		]
	]
transitiveVariant[ ruleID:Except[ _List], factKeys_List, keyHandler_, lbl_, add_, rem_][ ps_, __] :=
	With[ {ruleKeys = getProofSitData[ ps, ruleID, {}]},
		Hold[ updateAbbrevLists[ {}, {}, {}, {ruleID}], transitiveVariantAux[ factKeys, ruleKeys, keyHandler, lbl, add]]
	]
transitiveVariant[ ruleID:Except[ _List], factID:Except[ _List], keyHandler_, lbl_, add_, rem_][ ps_, __] :=
	With[ {factKeys = getProofSitData[ ps, factID, {}], ruleKeys = getProofSitData[ ps, ruleID, {}]},
		Prepend[
			If[ factKeys === {}, Hold[], Hold[ transitiveVariantAux[ factKeys, ruleKeys, keyHandler, lbl, add]]],
			updateAbbrevLists[ {}, {}, {}, {factID, ruleID}]
		]
	]

transitiveVariantAux[ factKeys_List, ruleKeys_, keyHandler_, lbl0_, tags_][ PRFSIT$[ G_, hyps_, idx_, fix_, meta_, tfix_Association, tvar_Association, fid_, ctxt___], ___, mode_] :=
	Catch[
		With[ {factIndices =
				Reverse[
					Map[
						Replace[ Position[ hyps, FML$[ Replace[ #, {ky:{_, _}, _} :> ky], __], {1}, 1, Heads -> False],
							{
								{{i_}} :> i,
								_ :> Throw[ $Failed]
							}]&,
						factKeys
					]
				],
				ruleIndices =
					If[ ListQ[ ruleKeys],
						Map[
							Replace[ Position[ hyps, FML$[ Replace[ #, {ky:{_, _}, _} :> ky], __], {1}, 1, Heads -> False],
								{
									{{i_}} :> i,
									_ :> Throw[ $Failed]
								}]&,
							ruleKeys
						],
					(*else*)
						Null
					]},
			Replace[ transitiveVariantFun[ Reverse[ Part[ hyps, factIndices]], If[ ListQ[ ruleIndices], ruleIndices, Null], hyps, fix, tfix],
				{
					{new_STERM$, j_, pd_} :>
						With[ {tempIndices =
								Union[
									Extract[ factIndices, Position[ factKeys, {{_, _}, False}, {1}, Heads -> False]],
									If[ ListQ[ ruleIndices], Extract[ ruleIndices, Position[ ruleKeys, {{_, _}, False}, {1}, Heads -> False]], {}]
								],
								lbl = If[ StringQ[ lbl0], lbl0, "?"]},
							If[ mode === 2,
								Join[
									Hold[ INF$[ "cut", new]],
									transitiveVariantProof[ factIndices, j, pd],
									Hold @@ MapIndexed[ INF$[ "forget", #1 + 2 - First[ #2]]&, tempIndices],
									If[ lbl === "?" && tags === {}, Hold[], Hold[ INF$[ "hypData", 1, lbl, "tags" -> tags]], Hold[]],
									Hold[ transitiveVariantAux[ keyHandler]]
								],
							(*else*)
								(* no separate definition for trace mode *)
								With[ {ky = makeFormulaKey[ fid]},
								With[ {newPS =
										PRFSIT$[
											G,
											Prepend[
												Delete[ hyps, tempIndices],
												FML$[ ky, First[ new], lbl, "tags" -> tags, "TypeVarPositions" -> makeTypeVarPositions[ new]]
											],
											idx, fix, meta, Join[ tfix, new[[2]]], tvar, fid + 1, ctxt
										]},
									{
										newPS,
										{},
										{},
										Hold[ transitiveVariantAux[ keyHandler]]
									}
								]]
							]
						],
					_ -> $Failed
				}
			]
		]
	]
transitiveVariantAux[ keyHandler_][ ps:PRFSIT$[ _, {FML$[ ky_, __], ___}, __], args__] :=
	keyHandler[ {{ky, False}}][ ps, args]

(* 'transitiveVariantFun' is not suitable for 'variantIterator'! *)
transitiveVariantFun[ facts_List, Null, hyps_List, fix_List, tfix_Association] :=
	Replace[ extractFormulas[ facts, 0],
		{fmls0_List, idx_} :>
			Catch[
				With[ {fmls1 = makeSingleSTERM[ fmls0]},
				With[ {fmls = First[ fmls1], tvar1 = Last[ fmls1]},
				Module[ {j = 0},
					Scan[
						Function[ trans0,
							++j;
							If[ MemberQ[ getOptionalComponent[ trans0, "tags"], "trans"],
								Replace[ extractFormula[ trans0, idx],
									{STERM$[ trans_, Association[], tvar2_Association], _} :>
										Replace[ resolveImplicationOFAuto[ fmls, trans, hyps, idx, fix, tfix, Join[ tvar1, tvar2]],		(* keys are disjoint *)
											{new_STERM$, pd_List} :> Throw[ {new, j, pd}]
										]
								]
							]
						],
						hyps
					];
					j = getRuleConsts[ First[ fmls], "trans", False];
					extendRULEC$[ $transRules];
					While[
						Replace[ iterateRULEC$[ j, $transRules],
							{
								k_List :>
									(
										If[ !MemberQ[ hyps, FML$[ k, __]],
											With[ {trans0 = getFormulas[ True][ k]},
												Replace[ extractFormula[ trans0, idx],
													{STERM$[ trans_, Association[], tvar2_Association], _} :>
														Replace[ resolveImplicationOFAuto[ fmls, trans, hyps, idx, fix, tfix, Join[ tvar1, tvar2]],
															{new_STERM$, pd_List} :> Throw[ {new, k, pd}]
														]
												]
											]
										];
										True
									),
								thm_THM$ :>
									(
										Replace[ getThmInst[ True, idx][ thm],
											{STERM$[ trans_, Association[], tvar2_Association], _} :>
												Replace[ resolveImplicationOFAuto[ fmls, trans, hyps, idx, fix, tfix, Join[ tvar1, tvar2]],
													{new_STERM$, pd_List} :> Throw[ {new, thm, pd}]
												]
										];
										True
									),
								_ -> False
							}
						]
					]
				]]];
				sowCPMMessage[ 133, "Fatal", "transitive"];
				$Failed
			]
	]
transitiveVariantFun[ facts_List, ruleIndices_List, hyps_, fix_List, tfix_Association] :=
	Replace[ extractFormulas[ facts, 0],
		{fmls0_List, idx_} :>
			Catch[
				With[ {fmls1 = makeSingleSTERM[ fmls0]},
				With[ {fmls = First[ fmls1], tvar1 = Last[ fmls1]},
					Scan[
						Function[ j,
							Replace[ extractFormula[ hyps[[j]], idx],
								{STERM$[ trans_, Association[], tvar2_Association], _} :>
									Replace[ resolveImplicationOFAuto[ fmls, trans, hyps, idx, fix, tfix, Join[ tvar1, tvar2]],		(* keys are disjoint *)
										{new_STERM$, pd_} :> Throw[ {new, j, pd}]
									]
							]
						],
						ruleIndices
					]
				]];
				sowCPMMessage[ 134, "Fatal", "transitive"];
				$Failed
			]
	]

(* 'transitiveVariantProof' is not suitable for 'variantIterator'! *)
transitiveVariantProof[ indices_List, j_Integer, pd_] :=
	OFVariantProof[ j, indices, pd][ 0]
transitiveVariantProof[ indices_List, thm_THM$, pd_] :=
	OFVariantProof[ {INF$[ "theorem", thm]}, indices, pd]
transitiveVariantProof[ indices_List, k_List, pd_] :=
	Prepend[
		OFVariantProof[ 0, indices, pd][ 1],
		INF$[ "insert", k]
	]

(* 'resolveImplicationOFAuto' simply calls 'resolveImplicationOF' with syntactic unification, with as many "$"s
	prepended to 'args' as possible.
	The reason why we use syntactic unification rather than general higher-order unification here can best be
	illustrated in the following example:
	Assume rule "== trans forward" (i.e. '/\P s t. s == t ==> P[t] ==> P[s]') and facts
	'3 == s[2]' and 's[2] <= s[s[2]]'. Then there are in total 4 possible higher-order unifiers that instantiate 'P' by
	- '\x. x <= s[x]'
	- '\x. s[2] <= s[x]',
	- '\x. x <= s[s[2]]', and
	- '\x. s[2] <= s[s[2]]', respectively.
	Convincing 'resolveImplicationOF' that the third one is the "right" one seems to be at least difficult ... *)
resolveImplicationOFAuto[ args_List, subj_, hyps_, idx_, fix_, tfix_Association, tvar_Association] :=
	With[ {ac = assmCountSkipAll[ subj, -Length[ args], Infinity, False]},
		If[ Negative[ ac],
			$Failed,
		(*else*)
			resolveImplicationOF[ Join[ Table[ "$", {ac}], args], subj, "syntactic", hyps, idx, fix, tfix, tvar]
		]
	]
	
	
(* ::Subsection:: *)
(*Accumulating Facts and Combining them by Transitivity*)

(* 'combineCalculation[ rules, final][ ps, __]' updates list "calc" by combining its elements and the current facts
	contained in "this" by a suitable transitivity rule, employing 'transitiveVariant'. If 'rules' is empty, the
	transitivity rule is searched among all currently available knowledge declared as "trans", whereas otherwise it is
	searched among the elements of 'rules' only.
	'final' specifies whether this is the last step in a calculation or not; in the former case, "calc" is cleared.
	List "calc" cannot be treated as a normal entry in "factLists", because it must also store permanent/temporary
	information. Otherwise, we would permanently store all intermediate results in a chain of
	transitivity-calculations. *)
combineCalculation[ rules_List, final_][ ps_PRFSIT$, args__] :=
	With[ {factLists = getProofSitData[ ps, "factLists", Association[]]},
	With[ {this = Lookup[ factLists, "this", $Failed]},
		If[ ListQ[ this],
			With[ {calc = getProofSitData[ ps, "calc", $Failed]},
				Which[
					ListQ[ calc],
					With[ {forget = Cases[ calc, {ky:{_, _}, False} :> ky]},
						If[ TrueQ[ final],
							Hold[
								(* transitivity shall only generate *one* new fact, that's why we can use 'initializeKeyList' here *)
								transitiveVariant[ rules, Join[ calc, this], initializeKeyList[ "chainedFacts", False], Automatic, {}, {}],
								forgetHyps[ forget],
								updateAbbrevLists[ {}, {}, {}, {"calc"}]
							],
						(*else*)
							Hold[
								(* transitivity shall only generate *one* new fact, that's why we can use 'initializeKeyList' here *)
								transitiveVariant[ rules, Join[ calc, this], initializeKeyList[ "calc", False], Automatic, {}, {}],
								forgetHyps[ forget]
							]
						]
					],
				(*else*)
					final,
					sowCPMMessage[ 135, "Fatal", "finally"];
					$Failed,
				(*else*)
					True,
					updateAbbrevLists[ {}, {}, {"calc" -> this}, {}][ ps, args]
				]
			],
		(*else*)
			sowCPMMessage[ 1009, "Warning", "also"];
			Hold[]
		]
	]]

(* 'accumulateCalculation[ final][ ps, __]' updates list "calc" by simply adding the keys contained in list "this".
	Although "calc" could be treated just as a normal entry in "factLists" here, for the sake of uniformity it is
	treated exactly as in 'combineCalculation'. *)
accumulateCalculation[ final_][ ps_PRFSIT$, args__] :=
	With[ {factLists = getProofSitData[ ps, "factLists", Association[]]},
	With[ {this = Lookup[ factLists, "this", $Failed]},
		If[ ListQ[ this],
			With[ {calc = getProofSitData[ ps, "calc", $Failed]},
				Which[
					ListQ[ calc],
					If[ TrueQ[ final],
						updateAbbrevLists[ {}, {}, {"chainedFacts" -> Join[ calc, this]}, {"calc"}],
					(*else*)
						updateAbbrevLists[ {}, {}, {"calc" -> Join[ calc, this]}, {}]
					][ ps, args],
				(*else*)
					final,
					sowCPMMessage[ 135, "Fatal", "ultimately"];
					$Failed,
				(*else*)
					True,
					updateAbbrevLists[ {}, {}, {"calc" -> this}, {}][ ps, args]
				]
			],
		(*else*)
			sowCPMMessage[ 1009, "Warning", "moreover"];
			Hold[]
		]
	]]
