(* Core proof methods for transforming formulas into certain formats. *)

(* ::Subsection:: *)
(*Parameters*)

(* The following symbols appear in "Theorema`CoreProvers`CoreProofMethods`"-context. *)
Scan[
	With[ {$$name = #[[1]], $$desc = #[[2]]},
	With[ {$$args = Join[ ToExpression[ "Theorema`CoreProvers`CoreProofMethods`" <> $$name, InputForm, Hold], Hold[ "usage", "English"]]},
		Set @@ ReplacePart[ Hold[ $$args, $$desc], {1, 0} -> MessageName];
	]]&,
	{
		{"analyzeStructure", ""},
		{"makeCoreCanonicalFormat", ""},
		{"reconstructFormula", ""},
		{"univToLambdaAdjustDB", ""},
		{"$canonicalFormatConstructor",
			"$canonicalFormatConstructor is a function that takes as input a formula 'f', two integers 'i' and 'n', and
			a list 'args', and is supposed to return a list of pairs '{new, prf}', where the 'new' are the canonical
			formats of 'f' and the 'prf' are the proofs of the respective 'new' from 'f', assuming that 'f' is the
			'i'-th explicit assumption and there are 'n' further explicit assumptions after it. The 'prf' may appear
			inside 'HoldComplete' to prevent constructing proofs in fast mode. 'args' is precisely the list of
			arguments of the respective 'VARIANT$' wrapper."},
		{"$equationalFormatConstructor",
			"$equationalFormatConstructor is like '$canonicalFormatConstructor' but shall try to obtain (quantified,
			conditional) equalities. It is called with one additional argument 'univ' specifying whether the results
			are allowed to be universally quantified or not."}
	}
]


(* ::Subsection:: *)
(*Direct Conversion*)

(* 'replaceByEquiv' is an auxiliary function that replaces an explicit assumption of the first subgoal of the current
	proof situation by an equivalent formula, where the equivalence is justified by a proof. *)
replaceByEquiv[ i_Integer, new_, _, PRFSIT$[ {G_, R___}, ctxt__], _, 0] :=
	{
		PRFSIT$[ {ReplacePart[ G, assmPosition[ i] -> new], R}, ctxt],
		{},
		{},
		Hold[]
	}
replaceByEquiv[ i_Integer, new_, prf_, ps:PRFSIT$[ {G0_, ___}, __], tag_, 1] :=
	Replace[ replaceByEquiv[ i, new, prf, ps, tag, 0],
		{
			{sit:PRFSIT$[ {G_, ___}, __], keys_List, thms_List, cmd_Hold} :>
				{
					makeANDNODE[ makePRFINFO[ name -> tag, used -> {{G0}}, generated -> {{G}}, "param" -> i], sit],
					keys,
					thms,
					cmd
				},
			_ -> $Failed
		}
	]
replaceByEquiv[ i_Integer, new_, prf_Hold, _, _, _] :=
	Prepend[ Append[ prf, INF$[ "drop", i]], setupSubgoalsMP[ new, i]]
replaceByEquiv[ i_Integer, new_, HoldComplete[ prf_], ps_, tag_, mode_] :=
	replaceByEquiv[ i, new, prf, ps, tag, mode]
replaceByEquiv[ _, _, _, _, _, _] :=
	$Failed


(* ::Subsection:: *)
(*Canonical Format*)

If[ !ValueQ[ $canonicalFormatConstructor], $canonicalFormatConstructor = makeCoreCanonicalFormat];

(* 'toCanonicalFormat[ i][ ps, ___]' transforms an arbitrary assumption of the first subgoal of 'ps', specified by 'i',
	into its canonical format, making use of '$canonicalFormatConstructor'. If '$canonicalFormatConstructor' returns
	more than one result, only the first one is obtained. *)
toCanonicalFormat[ i0_Integer][ ps:PRFSIT$[ {G_, ___}, __], ___, mode_] :=
	With[ {ac = assmCount[ G, 0, Infinity]},
	With[ {i = If[ Negative[ i0], ac + 1 + i0, i0]},
		If[ 1 <= i <= ac,
			Replace[ $canonicalFormatConstructor[ Extract[ G, assmPosition[ i]], i, 0, {}],
				{
					{} -> Hold[],
					{{canon_, prf_}, ___} :> replaceByEquiv[ i, canon, prf, ps, toCanonicalFormat, mode],
					_ -> $Failed
				}
			],
		(*else*)
			sowCPMMessage[ 111, {i0}, "Fatal"];
			$Failed
		]
	]]

(* 'canonVariant[ args, factID, keyHandler, lbl, add, rem]' is an auxiliary proof method used for obtaining the
	canonical format of hypotheses of the current proof situation, making use of '$canonicalFormatConstructor'.
	If '$canonicalFormatConstructor' returns more than one result, all of them are obtained. *)
canonVariant[ args_List, factID_, keyHandler_, lbl_, add_, rem_][ ps_, __] :=
	Hold @@ {
		updateAbbrevLists[ {}, {}, {}, {factID}],
		variantIterator[ getProofSitData[ ps, factID, {}], canonVariantFun[ args], keyHandler, lbl, add, rem]
	}

canonVariantFun[ args_][ i_, PRFSIT$[ _, hyps_, __]] :=
	Replace[ extractFormula[ hyps[[i]], 0],
		{STERM$[ fml_, _, tvar_], _} :>
			Replace[ $canonicalFormatConstructor[ fml, 1, 0, args],
				{
					{} -> Null,
					out:{{_, _}..} :>
						MapAt[ STERM$[ #, Association[], tvar]&, MapAt[ canonVariantProofWrapper[ i, #]&, out, {All, 2}], {All, 1}],
					_ -> $Failed
				}
			]
	]
	
canonVariantProofWrapper[ i_Integer, prf_Hold][ n_Integer] :=
	Prepend[ Append[ prf, INF$[ "close"]], INF$[ "makeExplicit", i + n]]
canonVariantProofWrapper[ i_Integer, HoldComplete[ prf_]][ n_Integer] :=
	canonVariantProofWrapper[ i, prf][ n]
canonVariantProofWrapper[ _, _][ _] :=
	$Failed


(* ::Subsubsection:: *)
(*Functions for Obtaining the Core Canonical Format*)

(* 'makeCoreCanonicalFormat[ form, x, y, eq]' constructs the canonical format 'form0' of the given formula 'form', and
	returns it together with its proof (assuming that 'form' is the 'x'-th explicit assumption and there are 'y'
	further assumptions after it). Alternatively, if 'form' is in canonical format already, the function simply returns
	'{}'.
	'eq' is a boolean value specifying whether the canonical format must be a Core equality or not; if 'eq' is 'True'
	but the formula is no equality, the function returns '$Failed'.
	The canonical format of a Core formula is obtained as follows:
	- First, the 'form' is brought to prenex normal form, i.e. all universal quantifiers are moved to the front.
	- If the atomic consequent is an equality and 'eq' is 'True', all universal quantifiers binding variables that
		appear in the equality but in no single premise are removed, and the variable they bind becomes bound by an
		explicit abstraction. This means that the equality becomes a functional equality. *)
makeCoreCanonicalFormat[ form_, x_, y_, _List] :=
	makeCoreCanonicalFormat[ form, x, y, False]
makeCoreCanonicalFormat[ form:Except[ _List], x_, y_, eq:(True|False)] :=
	makeCoreCanonicalFormat[ Reap[ analyzeStructure[ form, 0, 0, Infinity, True], {0, 1, 2, 3}], x, y, eq]
makeCoreCanonicalFormat[ {{A_, prenex_}, {loose_, indices_, prems_, premsScope_}}, x_, y_, eq:(True|False)] :=
	makeCoreCanonicalFormat[ A, Join @@ loose, Join @@ indices, Join @@ prems, Join @@ premsScope, prenex, x, y, eq]
makeCoreCanonicalFormat[ A_, loose_List, indices_List, prems_List, premsScope_List, prenex_, x_, y_, eq_] :=
	If[ TrueQ[ eq] && !MatchQ[ A, TYPED$[ CoreEqual$TM, {_}][ _][ _]],
		$Failed,
	(*else*)
		Replace[ univToLambdaAdjustDB[ eq, A, loose, indices, prems, premsScope],
			{A1_, loose1_, prems1_, good_} :>
				If[ TrueQ[ prenex] && good === {},
					{},
				(*else*)
					{{
						reconstructFormula[ fromAssmsGoal[ prems1, A1], loose1, {}, {}],
						With[ {ll = Length[ loose1], lp = Length[ prems1]},
							HoldComplete[ getCoreCanonicalFormatProof[ x, y, ll, lp, good]]
						]
					}}
				]
		]
	]

reconstructFormula[ A_, loose_List, {}, {}] :=
	Fold[ TYPED$[ CoreAll$TM, {Last[ #2]}][ Lambda$TM[ RNG$[ #2], #1]]&, A, Reverse[ loose]]
reconstructFormula[ A_, loose_List, {ppre___, prem_}, {spre___, scope_}] :=
	reconstructFormula[
		TYPED$[ CoreImplies$TM, {}][ prem][ Fold[ TYPED$[ CoreAll$TM, {Last[ #2]}][ Lambda$TM[ RNG$[ #2], #1]]&, A, Reverse[ Drop[ loose, scope]]]],
		Take[ loose, scope],
		{ppre},
		{spre}
	]

(* 'univToLambdaAdjustDB[ univToLambda, term, loose, indices, prems, premsScope]' replaces as many universally
	quantified variables from 'loose' by lambda abstractions on either side of 'term', if 'term' is an equality and
	'univToLambda' is 'True'. It also adjusts the de Bruijn indices in 'term' and 'prems' in such a way that all
	(remaining) universal quantifiers move to the front. *)
univToLambdaAdjustDB[ True, A:(TYPED$[ CoreEqual$TM, {typ_}][ a_][ b_]), loose_List, indices_List, prems_List, premsScope_List] :=
	With[ {max = Max @@ reap[ scanVars[ {a, b}, If[ TrueQ[ #3 >= #2], Sow[ #3 - #2]]&]] + 1},
		If[ Positive[ max],
			With[ {premsNum = Length[ prems]},
			Module[ {i = indices[[-max]] + 1, good = Range[ max - 1, 0, -1], j = 1},
				Catch[
					Scan[
						Function[ l,
							If[ i <= premsNum && good =!= {},
								With[ {new = DeleteDuplicates[ reap[ scanVars[ Take[ prems, {i, l}], If[ TrueQ[ 0 <= #3 - #2 < j], Sow[ #3 + max - (#2 + j)]]&]]]},
									good = DeleteCases[ good, Alternatives @@ new];
									i = l + 1;
									++j
								],
							(*else*)
								Throw[ Null]
							]
						],
						Take[ indices, -max + 1]
					]
				];
				If[ i <= premsNum && good =!= {},
					With[ {new = DeleteDuplicates[ reap[ scanVars[ Drop[ prems, i - 1], If[ TrueQ[ 0 <= #3 - #2 < max], Sow[ #3 - #2]]&]]]},
						good = DeleteCases[ good, Alternatives @@ new]
					]
				];
				(* 'good' contains the de Bruijn indices of those variables that do not appear in premises, in descending order *)
				With[ {allNum = Length[ indices]},
				With[ {goodIndices = allNum - good,		(* 'goodIndices' is in ascending order *)
						num = Length[ good]},
				With[ {goodPos = List /@ goodIndices},
				With[ {goodLoose = Reverse[ Extract[ loose, goodPos]]},
					{
						With[ {repl =
							If[ num === 0 || (j = First[ good]) + 1 === num,
								{},
							(*else*)
								DeleteCases[
									Join[
										MapIndexed[
											With[ {l = First[ #2]},
											With[ {res = Prepend[ Table[ k -> k + l, {k, #1 + 1, j - 1}], j -> num - l]},
												j = #1;
												Sequence @@ res
											]]&,
											Rest[ good]
										],
										{j -> 0},
										Table[ k -> k + num, {k, 0, j - 1}]
									],
									z_ -> z_
								]
							]},
							TYPED$[ CoreEqual$TM, {Fold[ FUN$TM[ Last[ #2], #1]&, typ, goodLoose]}][
								applyCurried[ If[ repl === {}, a, mapFreeVars[ a, VAR$[ #2 + Replace[ #3 - #2, repl]]&]], {}, goodLoose, True, True]
							][
								applyCurried[ If[ repl === {}, b, mapFreeVars[ b, VAR$[ #2 + Replace[ #3 - #2, repl]]&]], {}, goodLoose, True, True]
							]
						],
						Delete[ loose, goodPos],
						MapThread[
							Function[ {p, origNum},
								With[ {repl =
										With[ {gi = Reverse[ Select[ goodIndices, (# <= origNum &)]]},
										With[ {giNum = Length[ gi]},
										With[ {newNum = allNum - origNum - (num - giNum)},
											If[ gi === {} && newNum === 0,
												{},
											(*else*)
												i = 0;
												Join[
													DeleteCases[
														MapIndexed[
															With[ {res = Table[ k -> k - First[ #2] + newNum + 1, {k, i, origNum - # - 1}]},
																i = origNum - # + 1;
																Sequence @@ res
															]&,
															Reverse[ gi]
														],
														z_ -> z_
													],
													Table[ k -> k - giNum + newNum, {k, i, origNum - 1}]
												]
											]
										]]]},
									If[ repl === {}, p, mapFreeVars[ p, VAR$[ #2 + Replace[ #3 - #2, repl]]&]]
								]
							],
							{prems, premsScope}
						],
						goodIndices
					}
				]]]]
			]],
		(*else*)
			{A, loose, prems, {}}
		]
	]
univToLambdaAdjustDB[ _, A_, loose_List, _List, prems_List, premsScope_List] :=
	{
		A,
		loose,
		With[ {allNum = Length[ loose]},
			MapThread[
				Function[ {p, s},
					If[ s >= allNum,
						p,
					(*else*)
						mapFreeVars[ p, VAR$[ #3 + allNum - s]&]
					]
				],
				{prems, premsScope}
			]
		],
		{}
	}

analyzeStructure[ TYPED$[ CoreImplies$TM, {}][ A_][ B_], m_, n_, maxAssms_?Positive, prenex_] :=
	(Sow[ A, 2]; Sow[ m, 3]; analyzeStructure[ B, m, n + 1, maxAssms - 1, prenex])
analyzeStructure[ TYPED$[ CoreAll$TM, {_}][ Lambda$TM[ RNG$[ tr_TYPRNG$], A_]], m_, n_, maxAssms_, prenex_] :=
	(Sow[ tr, 0]; Sow[ n, 1]; analyzeStructure[ A, m + 1, n, maxAssms, prenex && n === 0])
analyzeStructure[ TYPED$[ CoreAll$TM, {_}][ TYPED$[ CoreEqual$TM, {typ_}][ a_]], _, n_, _, prenex_] :=
	{
		TYPED$[ CoreEqual$TM, {FUN$TM[ typ, typ]}][
			Lambda$TM[ RNG$[ TYPRNG$[ Null, typ]], mapFreeVars[ a, VAR$[ #3 + 1]&]]
		][
			Lambda$TM[ RNG$[ TYPRNG$[ Null, typ]], VAR$[ 0]]
		],
		prenex && n === 0
	}
analyzeStructure[ TYPED$[ CoreAll$TM, {typ_}][ A_], _, n_, _, prenex_] :=
	(Sow[ TYPRNG$[ Null, typ], 0]; Sow[ n, 1]; {A, prenex && n === 0})
analyzeStructure[ A_, _, _, _, prenex_] := {A, prenex}

getCoreCanonicalFormatProof[ i_Integer, _, allNum_, premNum_, good_] :=
	With[ {numGood = Length[ good]},
	Module[ {X, Y, j = 1},
		Hold @@ Join[
			Table[ INF$[ "allI", FIX$[ X, k]], {k, 0, allNum - 1}],
			If[ numGood === 0, {}, {INF$[ "ext", numGood]}],
			Table[ INF$[ "allI", FIX$[ Y, k]], {k, 0, numGood - 1}],
			MapIndexed[
				With[ {k = First[ #2]},
				With[ {res = Append[ Table[ INF$[ "allE", i, makeSTERM[ FIX$[ X, l]]], {l, j - k, #1 - k - 1}], INF$[ "allE", i, makeSTERM[ FIX$[ Y, k - 1]]]]},
					j = #1 + 1;
					Sequence @@ res
				]]&,
				good
			],
			Table[ INF$[ "allE", i, makeSTERM[ FIX$[ X, k]]], {k, j - numGood - 1, allNum - 1}],
			{INF$[ "impE", i, premNum]}
		]
	]]


(* ::Subsection:: *)
(*Equational Format*)

If[ !ValueQ[ $equationalFormatConstructor], $equationalFormatConstructor = makeCoreEquationalFormat];

(* 'toEquationalFormat[ i, univ][ ps, ___]' transforms an arbitrary assumption of the first subgoal of 'ps', specified
	by 'i', into its equational format, making use of '$equationalFormatConstructor'. If '$equationalFormatConstructor'
	returns more than one result, only the first one is obtained.
	'univ' specifies whether the result is allowed to be universally quantified or not. *)
toEquationalFormat[ i_Integer] :=
	toEquationalFormat[ i, True]
toEquationalFormat[ i0_Integer, univ_][ ps:PRFSIT$[ {G_, ___}, __], ___, mode_] :=
	With[ {ac = assmCount[ G, 0, Infinity]},
	With[ {i = If[ Negative[ i0], ac + 1 + i0, i0]},
		If[ 1 <= i <= ac,
			Replace[ $equationalFormatConstructor[ Extract[ G, assmPosition[ i]], i, 0, univ, {}],
				{
					{} -> Hold[],
					{{eq_, prf_}, ___} :> replaceByEquiv[ i, eq, prf, ps, toEquationalFormat, mode],
					_ -> $Failed
				}
			],
		(*else*)
			sowCPMMessage[ 112, {i0}, "Fatal"];
			$Failed
		]
	]]

(* 'eqVariant[ args, factID, keyHandler, lbl, add, rem]' is an auxiliary proof method used for obtaining the equational
	format of hypotheses of the current proof situation, making use of '$equationalFormatConstructor'.
	If '$equationalFormatConstructor' returns more than one result, all of them are obtained. *)
eqVariant[ args_List, factID_, keyHandler_, lbl_, add_, rem_][ ps_, __] :=
	Hold @@ {
		updateAbbrevLists[ {}, {}, {}, {factID}],
		variantIterator[
			getProofSitData[ ps, factID, {}],
			eqVariantFun[ !MemberQ[ args, "ground"|{"ground"}], MemberQ[ args, "rw"|{"rw"}], args],
			keyHandler,
			lbl,
			add,
			rem
		]
	}

eqVariantFun[ univ_, rw_, args_][ i_, PRFSIT$[ _, hyps_, _, _, _, tfix_, __]] :=
	With[ {hyp = hyps[[i]]},
		Replace[ extractFormula[ hyp, 0],
			{STERM$[ fml_, _, tvar_Association], _} :>
				Replace[ $equationalFormatConstructor[ fml, 1, 0, univ, args],
					{
						{} :>
							If[ TrueQ[ rw] && !MemberQ[ hyp, "rwData" -> _],
								With[ {rwData = generateEqRewriteRule[ hyp, tfix]},
									If[ rwData === $Failed,
										Null,
									(*else*)
										Append[ DeleteCases[ List @@ Drop[ hyp, 3], "TypeVarPositions" -> _], "rwData" -> rwData]
									]
								],
							(*else*)
								Null
							],
						out:{{_, _}..} :>
							With[ {tags0 = Intersection[ getOptionalComponent[ hyp, "tags"], {"comp", "cong"}]},
							With[ {tags = If[ tags0 === {}, {}, {"tags" -> tags0}]},
								Which[
									TrueQ[ rw],
									Replace[ out,
										{eq_, prf_} :>
											With[ {rwData = generateEqRewriteRule[ eq, True, tfix, tvar]},
											With[ {data = If[ rwData === $Failed, tags, Append[ tags, "rwData" -> rwData]]},
												If[ data === {},
													(* 'canonVariantProofWrapper' can be used for equational format, too *)
													{STERM$[ eq, Association[], tvar], canonVariantProofWrapper[ i, prf]},
												(*else*)
													{STERM$[ eq, Association[], tvar], canonVariantProofWrapper[ i, prf], data}
												]
											]],
										{1}
									],
								(*else*)
									tags === {},
									MapAt[ STERM$[ #, Association[], tvar]&, MapAt[ canonVariantProofWrapper[ i, #]&, out, {All, 2}], {All, 1}],
								(*else*)
									True,
									Replace[ out, {eq_, prf_} :> {STERM$[ eq, Association[], tvar], canonVariantProofWrapper[ i, prf], tags}, {1}]
								]
							]],
						_ -> $Failed
					}
				]
		]
	]


(* ::Subsubsection:: *)
(*Functions for Obtaining the Core Equational Format*)

makeCoreEquationalFormat[ form_, x_, y_, False, _] :=
	Replace[ makeCoreCanonicalFormat[ form, x, y, True],
		{
			{} :> If[ MatchQ[ form, TYPED$[ CoreAll$TM, _][ _]], $Failed, {}],
			{{canon_, prf_}} :> If[ MatchQ[ canon, TYPED$[ CoreAll$TM, _][ _]], $Failed, {{canon, prf}}]
		}
	]
makeCoreEquationalFormat[ form_, x_, y_, _, _] :=
	makeCoreCanonicalFormat[ form, x, y, True]


(* ::Subsection:: *)
(*Elimination Format*)

(* 'toEliminationFormat[ i, n, dup, force][ ps, ___]' transforms an arbitrary assumption of the first subgoal of 'ps',
	specified by 'i', into its elimination format, making use of 'makeCoreEliminationFormat'. If that function returns
	more than one result, only the first one is obtained.
	'n' specifies how many premises of the rule shall be considered part of the atomic consequent *at most* (i.e. if
	there are not enough premises, then all are taken); negative values are possible, too, and indicate how many
	premises shall *not* be considered part of the atomic consequent.
	'dup' specifies how many of the premises that are not part of the atomic consequent shall be duplicated; it must be
	non-negative or 'All'.
	'force' specifies whether a rule that already is in elimination format (i.e. whose atomic consequent, taking 'j'
	into account, is a bound variable) shall nevertheless be turned into a new elimination rule. *)
toEliminationFormat[ i_Integer] :=
	toEliminationFormat[ i, 0, True]
toEliminationFormat[ i_Integer, n_Integer] :=
	toEliminationFormat[ i, n, True]
toEliminationFormat[ i0_Integer, n_Integer, dup_Integer?NonNegative, force_][ ps:PRFSIT$[ {G_, ___}, __], ___, mode_] :=
	With[ {ac = assmCount[ G, 0, Infinity]},
	With[ {i = If[ Negative[ i0], ac + 1 + i0, i0]},
		If[ 1 <= i <= ac,
			Replace[ makeCoreEliminationFormat[ Extract[ G, assmPosition[ i]], i, n, dup, force],
				{
					{} -> Hold[],
					{{elim_, prf_}, ___} :> replaceByEquiv[ i, elim, prf, ps, toEliminationFormat, mode],
					_ -> $Failed
				}
			],
		(*else*)
			sowCPMMessage[ 112, {i0}, "Fatal"];
			$Failed
		]
	]]

(* 'eliminationVariant[ args, factID, keyHandler, lbl, add, rem]' is an auxiliary proof method used for obtaining the
	elimination format of hypotheses of the current proof situation, making use of 'makeCoreEliminationFormat'.
	If that function returns more than one result, all of them are obtained. *)
eliminationVariant[ args_List, factID_, keyHandler_, lbl_, add_, rem_][ ps_, __] :=
	Hold @@ {
		updateAbbrevLists[ {}, {}, {}, {factID}],
		variantIterator[
			getProofSitData[ ps, factID, {}],
			eliminationVariantFun @@ args,
			keyHandler,
			lbl,
			add,
			rem
		]
	}

eliminationVariantFun[] :=
	eliminationVariantFun[ True, 0, 0]
eliminationVariantFun[ force_] :=
	If[ IntegerQ[ force] || force === All || force === "all" || force === "_",
		eliminationVariantFun[ True, force, 0],
	(*else*)
		eliminationVariantFun[ force, 0, 0]
	]
eliminationVariantFun[ force_, n_] :=
	If[ IntegerQ[ force] || force === All || force === "all" || force === "_",
		eliminationVariantFun[ True, force, n],
	(*else*)
		eliminationVariantFun[ force, n, 0]
	]
eliminationVariantFun[ force_, n_, dup_][ i_, PRFSIT$[ _, hyps_, __]] :=
	Replace[ extractFormula[ hyps[[i]], 0],
		{STERM$[ fml_, _, tvar_Association], _} :>
			Replace[
				makeCoreEliminationFormat[
					fml,
					1,
					Switch[ n, All|"all", -1, "_", 0, _Integer, n, _, 0],
					Switch[ dup, All|"all", All, "_", 0, _Integer?NonNegative, dup, _, 0],
					If[ StringQ[ force], force =!= "?", force]
				],
				{
					{} -> Null,
					out:{{_, _}..} :> MapAt[ STERM$[ #, Association[], tvar]&, MapAt[ canonVariantProofWrapper[ i, #]&, out, {All, 2}], {All, 1}],
					_ -> $Failed
				}
			]
	]


(* ::Subsubsection:: *)
(*Functions for Obtaining the Elimination Format*)

(* 'makeCoreEliminationFormat[ form, i, n, dup, force]' constructs the elimination format of 'form'. 'n', 'dup' and
	'force' have the same meaning as in 'toEliminationFormat'. *)
makeCoreEliminationFormat[ form:Except[ _List], i_, n0_, dup_, force_] :=
	With[ {n = Which[ Negative[ n0], -n0 - 1, n0 === 0, Infinity, True, Max[ assmCountSkipAll[ form, -n0, Infinity, True], 0]]},
		makeCoreEliminationFormat[ Reap[ analyzeStructure[ form, 0, 0, n, True], {0, 1, 2, 3}], i, dup, force]
	]
makeCoreEliminationFormat[ {{A_, _}, {loose_, indices_, prems_, premsScope_}}, i_, dup_, force_] :=
	makeCoreEliminationFormat[ A, Join @@ loose, Join @@ indices, Join @@ prems, Join @@ premsScope, i, dup, force]
makeCoreEliminationFormat[ A_, loose_List, indices_List, prems_List, premsScope_List, i_, dup0_, force_] :=
	If[ force =!= True && MatchQ[ A, _VAR$],
		{},
	(*else*)
		With[ {thesisIndex = If[ indices === {}, 0, Last[ indices]], allNum = Length[ loose] + 1,
				dup = If[ dup0 === All, Length[ prems], dup0]},
		With[ {premsScopeN = Join[ Take[ premsScope, thesisIndex], Drop[ premsScope, thesisIndex] + 1],
				premsN = Join[ Take[ prems, thesisIndex], mapFreeVars[ Drop[ prems, thesisIndex], VAR$[ #3 + 1]&]],
				dupIndex = Min[ dup, thesisIndex]},
		With[ {AN =
				fromAssmsGoal[
					Join[
						{mapFreeVars[ A, VAR$[ #3 + 1]&]},
						MapThread[
							Function[ {p, ps}, With[ {offset = allNum - ps}, mapFreeVars[ p, VAR$[ #3 + offset]&]]],
							{Take[ premsN, dupIndex], Take[ premsScope, dupIndex]}
						],
						Take[ premsN, {dupIndex + 1, dup}]
					],
					VAR$[ 0]
				]},
			{{
				reconstructFormula[ VAR$[ 0], Append[ loose, TYPRNG$[ Theorema`Knowledge`VAR$thesis$TM, FORM$TM]], Append[ premsN, AN], Append[ premsScopeN, allNum]],
				With[ {ll = Length[ loose], lp = Length[ prems],
						ac = assmCount[ A, 0, Infinity], dupAC = assmCount[ #, 0, Infinity]& /@ Take[ prems, dup]},
					HoldComplete[ getCoreEliminationFormatProof[ i, ll, lp, ac, dupAC]]
				]
			}}
		]]]
	]

getCoreEliminationFormatProof[ i_, allNum_, premNum_, ac_, dupAC_List] :=
	Module[ {X},
		Hold @@ Join[
			Table[ INF$[ "allI", FIX$[ X, k]], {k, 0, allNum}],		(* 'allNum' does not include 'thesis' *)
			{INF$[ "impE", i + premNum + 1, 0]},
			Table[ INF$[ "allE", i, makeSTERM[ FIX$[ X, k]]], {k, 0, allNum - 1}],
			{INF$[ "impE", i, premNum + ac]},
			MapIndexed[ INF$[ "impE", i + First[ #2], #1]&, dupAC]
		]
	]


(* ::Subsection:: *)
(*Beta/Eta Conversion*)

(* 'convertBetaEta[ pos, spec, normOrig][ ps, _, mode]' performs beta-/eta- reductions or expansions in the first
	subgoal of 'ps', at the specified positions 'pos' ('pos' is expected to be a *list* of positions). 'spec' can be
	one of
	- "normalize", in which case full normalization, i.e. beta- and eta reductions are performed on the specified
		subexpressions,
	- '{"beta", Reduce}', in which case full beta reduction is performed, but no eta reduction,
	- '{"beta", Reduce, n}', where 'n' is non-negative, in which case the *at most* 'n' outermost beta-redexes are
		reduced,
	- '{"beta", Expand, n}', where 'n' is non-negative, in which case the first 'n' free variables appearing in each of
		the specified subexpressions are abstracted away (e.g. 'f[x,y]' becomes '(\x,y. f[x,y])[x, y]'),
	- '{"beta", Expand, sub}', where 'sub' is a list of positions, in which case the subexpressions at positions 'sub'
		are abstracted away from every of the specified (by 'pos') subexpressions 'expr'; this is only possible if any
		variable bound in 'expr' is also bound in the subexpressions, *and if all the subexpressions are
		alpha-equivalent to each other*,
	- '{"eta", Reduce}', in which case full eta reduction is performed, but no beta reduction,
	- '{"eta", Reduce, n}', where 'n' is a non-negative integer, in which case the at most 'n' outermost eta-redexes
		are reduced,
	- '{"eta", Expand}', in which case full eta-expansion is performed on the specified subexpressions, i.e. every
		function is applied to exactly as many arguments as its type permits.
	'normOrig' is a boolean value indicating whether the original subexpressions need to be normalized before they can
		be tested for alpha-equivalence with the new expressions (just as in "alphabetaeta").
	In fast mode, all these conversions are performed directly, without relying on basic inference "alphabetaeta".
	If at least one of the given positions is invalid, or the specified conversion cannot be applied to at least one
	subexpression, the proof method fails instantly. *)
convertBetaEta[ pos_List, spec_] := convertBetaEta[ pos, spec, True]
convertBetaEta[ pos_List, spec_, normOrig_][ PRFSIT$[ {G_, R___}, hyps_, idx_, fix_, meta_, ctxt__], ___, mode_] :=
	Catch[
		With[ {fun =
				Replace[ getNormalizationFunction[ spec, fix, meta],
					$Failed :>
						(
							sowCPMMessage[ 121, {spec}, "Fatal"];
							Throw[ $Failed]
						)
				]},
		With[ {repl =
				Map[
					Replace[ validExtract[ G, #, {}],
						{
							{expr_, loose_List} :>
								With[ {new = fun[ expr, loose]},
									If[ new === $Failed,
										sowCPMMessage[ 122, #, "Fatal"];
										Throw[ $Failed],
									(*else*)
										{#} -> new
									]
								],
							_ :>
								(
									sowCPMMessage[ 123, #, "Fatal"];
									Throw[ $Failed]
								)
						}
					]&,
					pos
				],
				normNew = (spec =!= "normalize")},
			If[ mode === 2,
				Hold @@ Replace[ repl, ({p_List} -> new_) :> INF$[ "alphabetaeta", p, new, normOrig, normNew], {1}],
			(*else*)
				{PRFSIT$[ {ReplacePart[ G, repl], R}, hyps, idx, fix, meta, ctxt], {}, {}, Hold[]}
			]
		]]
	]

(* 'normVariant[ major][ args, factID, keyHandler, lbl, add, rem]' is an auxiliary proof method used for normalizing or
	beta/eta reducing/expaning hypotheses of the current proof situation. *)
normVariant[ major_][ args_List, factID_, keyHandler_, lbl_, add_, rem_][ ps_PRFSIT$, __] :=
	Catch[
		Hold @@ {
			updateAbbrevLists[ {}, {}, {}, {factID}],
			variantIterator[
				getProofSitData[ ps, factID, {}],
				normVariantFunction[
					If[ major === "normalize",
						"normalize",
					(*else*)
						Replace[ Reverse[ Flatten[ args]],
							{
								{s:("reduce"|"expand")} :> {major, s},
								{"reduce", n_Integer?NonNegative} :> {major, "reduce", n},
								{} -> {major, "reduce"},
								_ :> (sowCPMMessage[ 124, {major, args}, "Fatal", "norm"]; Throw[ $Failed])
							}
						]
					]
				],
				keyHandler,
				lbl,
				add,
				rem
			]
		}
	]

normVariantFunction[ spec_][ i_, PRFSIT$[ _, hyps_, _, fix_, meta_, __]] :=
	Replace[ extractFormula[ hyps[[i]], 0],
		{STERM$[ fml_, _, tvar_], _} :>
			With[ {new = getNormalizationFunction[ spec, fix, meta][ fml, {}]},
				If[ new === fml,
					Null,
				(*else*)
					{{
						STERM$[ new, Association[], tvar],
						With[ {ac = assmCount[ new, 0, Infinity]},
							Hold[ INF$[ "makeExplicit", i + #], convertBetaEta[ {{0, 1}}, spec], INF$[ "impE", 1, ac], INF$[ "close"]]&
						]
					}}
				]
			]
	]

getNormalizationFunction[ "normalize"|{"normalize"}, _, _] :=
	(normalize[ #1, True] &)
getNormalizationFunction[ {"beta", Reduce|"reduce"}, _, _] :=
	(betaReduce[ #1] &)
getNormalizationFunction[ {"beta", Reduce|"reduce", n_}, _, _] :=
	(betaReduceRestricted[ #1, n] &)
getNormalizationFunction[ {"beta", Expand|"expand", arg_}, fix_, meta_] :=
	(betaExpand[ #1, arg, #2, fix, meta] &)
getNormalizationFunction[ {"eta", Reduce|"reduce"}, _, _] :=
	(etaReduce[ #1] &)
getNormalizationFunction[ {"eta", Reduce|"reduce", n_}, _, _] :=
	(etaReduceRestricted[ #1, n] &)
getNormalizationFunction[ {"eta", "expand"}, fix_, meta_] :=
	(etaExpand[ #1, #2, fix, meta] &)
getNormalizationFunction[ _, _, _] := $Failed
