(* Core proof methods handling implications. *)

(* ::Subsection:: *)
(*Elimination*)

(* 'eliminateCoreImplies[ args, i][ ps, ___]' eliminates the implication in the 'i'-th explicit assumption of the first
	subgoal in 'ps' by unifying its premises with the consequents of other explicit assumptions, given by the 'args'.
	'args' is supposed to be a list of indices (possibly negative) referring to explicit assumptions, or "_" indicating
	to skip a premise of the main implication.
	It must be possible to unify the premise-assumption-pairs without being left with flex-flex equations; in that
	case, the first such unifier is chosen, and no backtracking is employed. However, note that all unification
	problems are solved simultaneously.
	Also note that (some of) the premises of the main implication may appear in the scope of universal quantifiers;
	these, and only these, quantifiers are automatically instantiated by meta variables before the unification
	procedure is invoked. *)
eliminateCoreImplies[ args_List] :=
	eliminateCoreImplies[ args, 1]
eliminateCoreImplies[ {pre___, "_"..}, i_] :=
	eliminateCoreImplies[ {pre}, i]
eliminateCoreImplies[ {}, _] :=
	voidTac[]
eliminateCoreImplies[ args_List, i0_Integer][ ps:PRFSIT$[ {G_, ___}, __], rest___] :=
	With[ {num = assmCount[ G, 0, Infinity]},
	With[ {i = If[ Negative[ i0], num + 1 + i0, i0]},
		If[ 1 <= i <= num,
			eliminateCoreImplies[ args, i, num][ ps, rest],
		(*else*)
			sowCPMMessage[ 91, {i0}, "Fatal"];
			$Failed
		]
	]]
eliminateCoreImplies[ args0_List, orig_Integer, num_Integer] :=
	Catch[
		Module[ {max = orig},
		With[ {args =
				Replace[ args0,
					{
						i0_Integer :>
							With[ {i = If[ Negative[ i0], num + 1 + i0, i0]},
								Which[
									i === orig,
									sowCPMMessage[ 92, "Fatal"];
									Throw[ failTac[]],
								(*else*)
									i <= num,
									max = Max[ max, i];
									i,
								(*else*)
									True,
									sowCPMMessage[ 93, {i0}, "Fatal"];
									Throw[ failTac[]]
								]
							]
					},
					{1}
				]},
			eliminateCoreImplies[ args, orig, max, num]
		]]
	]
eliminateCoreImplies[ args_List, orig_Integer, max_Integer, num_Integer][ ps:PRFSIT$[ {G_, ___}, __], ___, 1] :=
	Replace[ executeProofCommands[ Hold[ eliminateCoreImplies[ args, orig, max, num]], {ps}, 0, False],
		{
			PRFSTAT$[ _, {sit:PRFSIT$[ {G0_, ___}, __]}, keys_List, thms_List, ___] :>
				With[ {origpos = assmPosition[ orig],
						assms = Extract[ G, assmPosition[ Cases[ args, _Integer]]]},	(* all indices are positive *)
					{
						makeANDNODE[ makePRFINFO[ name -> eliminateCoreImplies, used -> {{Extract[ G, origpos]}, assms}, generated -> {{Extract[ G0, origpos]}}], sit],
						keys,
						thms,
						Hold[]
					}
				],
			_ -> $Failed
		}
	]
eliminateCoreImplies[ argIndices_List, orig_Integer, max_Integer, num_Integer][ ps:PRFSIT$[ {G_, ___}, _, _, fix_, meta_, tfix_Association, tvar_Association, __], ___, mode_] :=
	With[ {args = Replace[ argIndices, i_Integer :> {Extract[ G, assmPosition[ i]], i}, {1}]},
		Replace[ resolveImplication[ args, Extract[ G, assmPosition[ orig]], fix, meta, tfix, tvar, mode === 0],
			{
				{cmd_List, new_} :>
					Hold @@ Append[ cmd, eliminateCoreImpliesPost[ orig, new]],
				{cmd_List, allE_List, indices_List} :>
					Hold @@ Join[ cmd, Map[ INF$[ "allE", orig, makeSTERM[ normalize[ #]]]&, allE], {eliminateCoreImpliesPost[ orig, indices, max, num]}],
				_ -> $Failed
			}
		]
	]

(* 'eliminateCoreImpliesPost' is an auxiliary proof method that is applied after instantiating the universal
	quantifiers in the subject by suitable terms. *)
eliminateCoreImpliesPost[ orig_, new_][ ps_PRFSIT$, ___, 0] :=
	{
		ReplacePart[ ps, Join[ {1, 1}, assmPosition[ orig]] -> new],
		{},
		{},
		Hold[]
	}
eliminateCoreImpliesPost[ orig_, argIndices_List, max_, num_][ PRFSIT$[ {G_, ___}, __], ___] :=
	With[ {ag = toAssmsGoal[ G, {}, max]},
		Replace[
			Last @ Reap[
				Module[ {i = 0},
					Scan[
						If[ ++i =!= orig,
							Replace[ Cases[ argIndices, {i, b_} :> b, {1}, 1],
								{
									{0} -> Null,
									{j_} :> Sow[ takeAssms[ #1, {}, j], 1],
									_ :> Sow[ #1, 2]
								}
							]
						]&,
						Most[ ag]
					]
				],
				{1, 2}
			],
			{inst1_List, inst2_List} :>
				Hold @@ Join[
					{INF$[ "theorem", THM$[ fastThmFromLabel[ "GMP"], orig, argIndices]]},
					Map[ INF$[ "allE", 1, makeSTERM[ #]]&, Join[ toAssmsGoal[ ag[[orig]], {}, Length[ argIndices]], Flatten[ inst1, 2], Join @@ inst2, {Last[ ag]}]],
					{INF$[ "impE", 1, num]}
				]
		]
	]

resolveImplication[ args_List, subj_, fix_List, meta_List, tfix_Association, tvar_Association, constructFml_] :=
	Replace[ Reap[ resolveImplicationAux[ args, subj, meta, Length[ fix], constructFml], {"unif", "ind"}],
		{
			{{new_, newMeta0_List}, {{unif_List}, ind_List}} :>	(* there must be at least 1 unification problem *)
				Replace[ unifyFirst[ unif, newMeta0, fix, tfix, tvar, Automatic, "AllowFlexFlex" -> False],	(* there may be several unifiers, but we simply take the first one *)
					{
						{{}, newMeta1_, inst_, tinst_, tv_Association} :>
							With[ {oldPos = Position[ inst, m_ /; MemberQ[ meta, m -> _] -> _, {1}, Heads -> False]},
									(* positions of those meta variables that are in the proof situation already *)
							With[ {oldInst = Extract[ inst, oldPos], newInst = Delete[ inst, oldPos]},
							With[ {cmd =
									Join[
										{INF$[ "sorts", tv]},
										If[ tinst === {}, {}, {INF$[ "instT", tinst]}],
										Cases[ newMeta1, (m_ -> {t_, d_}) /; FreeQ[ meta, m -> _, {1}, Heads -> False] :> INF$[ "meta", m, t, d]],
										Cases[ oldInst, (m_ -> i_) :> INF$[ "inst", m -> STERM$[ normalize[ i], Association[], Association[]]]]
									]},
								If[ constructFml,
									{cmd, normalize[ Fold[ ReplaceAll, replaceTypes[ new, tinst], inst]]},
								(*else*)
									{cmd, Fold[ ReplaceAll, Drop[ newMeta0, Length[ meta]][[All, 1]], newInst], Join @@ ind}
								]
							]]],
						_ :> (sowCPMMessage[ 1007, "Fatal", "resolveImplication"]; $Failed)
					}
				],
			_ :> (sowCPMMessage[ 1008, "Fatal", "resolveImplication"]; $Failed)
		}
	]

resolveImplicationAux[ {}, subj_, meta_, _, _] := {subj, meta}
resolveImplicationAux[ args_List, TYPED$[ CoreAll$TM, {typ_}][ A_], meta_, fixNum_, constructFml_] :=
	With[ {name = Replace[ A, {Lambda$TM[ RNG$[ TYPRNG$[ n:Except[ Null, _Symbol], _]], _] :> n, _ -> Theorema`Knowledge`VAR$X$TM}]},
	With[ {j = Max[ 0, Max @@ Cases[ meta, (META$[ name, i_Integer] -> _) :> i] + 1]},
	With[ {m = META$[ name, j]},
		resolveImplicationAux[ args, applyFunction[ A, m], Append[ meta, m -> {typ, fixNum}], fixNum, constructFml]
	]]]
resolveImplicationAux[ {"_", r___}, TYPED$[ CoreImplies$TM, {}][ A_][ R_], meta_, fixNum_, constructFml_] :=
	If[ constructFml,
		Replace[ resolveImplicationAux[ {r}, R, meta, fixNum, True],
			{new_, newMeta_} :> {TYPED$[ CoreImplies$TM, {}][ A][ new], newMeta}
		],
	(*else*)
		Sow[ Skip, "ind"];
		resolveImplicationAux[ {r}, R, meta, fixNum, False]
	]
resolveImplicationAux[ {{arg_, i_Integer}, r___}, TYPED$[ CoreImplies$TM, {}][ A_][ R_], meta_, fixNum_, constructFml_] :=
	Replace[ stripAssms[ arg, 0],
		{A0_, n_} :>
			(
				Sow[ {A, A0}, "unif"];
				If[ constructFml,
					Replace[ resolveImplicationAux[ {r}, R, meta, fixNum, True],
						{new_, newMeta_} :> {ReplacePart[ arg, {Table[ 1, {n}]} -> new], newMeta}
					],
				(*else*)
					Sow[ {i, n}, "ind"];
					resolveImplicationAux[ {r}, R, meta, fixNum, False]
				]
			)
	]
resolveImplicationAux[ _List, _, _, _, _] :=
	$Failed

(* 'eliminateCoreImpliesMult[ inst, n]' eliminates implications in multiple explicit assumptions, falling back to
	'eliminateCoreImplies'. 'n' can be an integer, in which case the first 'n' assumptions are eliminated (or the last
	ones, if 'n' is negative), or a list of integers. *)
eliminateCoreImpliesMult[ _, 0|{}] :=
	voidTac[]
eliminateCoreImpliesMult[ args_, n0_Integer][ PRFSIT$[ {G_, ___}, __], ___] :=
	With[ {num = assmCount[ G, 0, Infinity]},
	With[ {n = If[ Negative[ n0], num + 1 + n0, n0]},
		If[ 0 <= n <= num,
			Hold @@ Table[ eliminateCoreImplies[ args, i, num], {i, 1, n}],
		(*else*)
			sowCPMMessage[ 94, {n0}, "Fatal"];
			$Failed
		]
	]]
eliminateCoreImpliesMult[ args_, p_List][ PRFSIT$[ {G_, ___}, __], ___] :=
	With[ {num = assmCount[ G, 0, Infinity]},
		Hold @@ Map[ eliminateCoreImplies[ args, #, num]&, p]
	]


(* ::Subsection:: *)
(*OF-variant*)

(* 'OFVariant[ args, factID, keyHandler, lbl, add, rem]' is an auxiliary proof method used for eliminating premises of
	hypotheses by resolving them with other hypotheses.
	'OFVariant' is the rough high-level analogue of 'resolveImplication' above. There are important differences,
	though:
	- Nested implications among the premises of the subject/argument facts are treated in a more sophisticated
		way here.
	- Universal quantifiers in the arguments may be instantiated here.
	- Meta-variables left after unification get bound again, since hypotheses cannot contain meta variables. *)
OFVariant[ {args___, "_"...}, factID_, keyHandler_, lbl_, add_, rem_][ __] :=
	(* shall be unfolded only when applied *)
	With[ {id = getFreshFactID[],
			blankPos = Position[ {args}, "_", {1}, Heads -> False]},
	With[ {args1 = Delete[ {args}, blankPos]},
	With[ {autoPos = Position[ args1, "$", {1}, Heads -> False]},
		Hold @@ {
			addKnowledgeAux[ Delete[ args1, autoPos], addKeysToList[ id, False], Hold[], Automatic, None],
			OFVariant[ id, MapIndexed[ (#1 + 1 - First[ #2])&, blankPos], MapIndexed[ (#1 + 1 - First[ #2])&, autoPos], factID, keyHandler, lbl, add, rem]
		}
	]]]
OFVariant[ argsID_, blankInsert_List, autoInsert_List, factID_, keyHandler_, lbl_, add_, rem_][ ps_, __] :=
	With[ {factKeys = getProofSitData[ ps, factID, {}], argKeys0 = getProofSitData[ ps, argsID, {}]},
	With[ {tempArgs = Cases[ argKeys0, {k:{_, _}, False} :> k], argKeys = Replace[ argKeys0, {k:{_, _}, _} :> k, {1}]},
		Join[
			Hold[
				updateAbbrevLists[ {}, {}, {}, {factID, argsID}],
				variantIterator[ factKeys, OFVariantFun[ argKeys, blankInsert, autoInsert], keyHandler, lbl, add, rem]
			],
			If[ tempArgs === {}, Hold[], Hold[ forgetHyps[ tempArgs]]]
		]
	]]

OFVariantFun[ keys_List, blankInsert_List, autoInsert_List][ i_, PRFSIT$[ _, hyps_, _, fix_, _, tfix_Association, __]] :=
	If[ keys === {} === autoInsert,
		Null,
	(*else*)
		With[ {indices = Reverse[ Map[ First[ Position[ hyps, FML$[ #, __], {1}, 1, Heads -> False]]&, keys]]},
			Replace[ extractFormulas[ Extract[ hyps, indices], 0],
				(* We need to revert the list, because the new type-variable indices must be computed in the same order,
					as the hypotheses are made explicit later.
					In principle it would be possible to pre-compute this only once, but then we'd have to parameterize
					'OFVariantFun' over the whole expressions ... *)
				{args0_List, idx1_} :>
					With[ {args1 = makeSingleSTERM[ args0],
							hyp = hyps[[i]]},
					With[ {args = First[ args1], tvar1 = Last[ args1]},
						Replace[ extractFormula[ hyp, idx1],
							{STERM$[ fml_, Association[], tvar2_Association], idx2_} :>
								Replace[
									resolveImplicationOF[
										Insert[ Insert[ Reverse[ args], "$", autoInsert], "_", blankInsert],
										fml, Automatic, hyps, idx2, fix, tfix, Join[ tvar1, tvar2]		(* keys are disjoint *)
									],
									{
										{new_STERM$, pd_List} :> {{new, OFVariantProof[ i, indices[[All, 1]], pd], preserveFormulaTags[ new, hyp, tfix]}},
										j:(0|1|2) :>
											(
												sowCPMMessage[ If[ j === 2, 100, 1008 - j], "Fatal", "OF"];
												$Failed
											),
										_ -> $Failed
									}
								]
						]
					]]
			]
		]
	]

OFVariantProof[ i_Integer, indices_List, pd_][ n_Integer] :=
	OFVariantProof[ {INF$[ "makeExplicit", i + n]}, indices + n, pd]
OFVariantProof[ cmds_List, indices_List, {ac_Integer, meta_List, metaRem_List, metaUn_List, inst_List, argInfo_List, allIndices_List}] :=
	With[ {numRemVars = Length[ metaRem]},
	Module[ {X},
		Join[
			Hold @@ Join[
				Map[ INF$[ "makeExplicit", #]&, indices],
				cmds,	(* 'cmds' must add *exactly 1* explicit assumption *)
				
				(* instantiations *)
				Table[ INF$[ "allI", FIX$[ X, j]], {j, 0, numRemVars - 1}],
				With[ {repl =
						Join[
							Replace[ metaUn, (m_ -> {t_, _}) :> (m -> TYPED$[ Theorema`Knowledge`TYPE$TM, {t}]), {1}],
							MapIndexed[ (#1 -> FIX$[ X, numRemVars - First[ #2]])&, metaRem]
						]},
				With[ {instList0 = Replace[ meta, inst, {1}]},
				With[ {instList1 =
						ReplacePart[ instList0,
							Replace[ getPositions[ instList0, _META$],
								(pos_List -> m_META$) :> (pos -> Replace[ m, repl]),
								{1}
							]
						]},
					Join[
						(* we cannot use "allE" here, because the instances might contain new a.b.f. constants whose types have been fixed in "cut" *)
						If[ allIndices === {}, {}, {instantiateCoreAll[ EXPR$ /@ instList1[[allIndices]], 1]}],
						Cases[ argInfo, {p_Integer, _, _, s_, n_?Positive} :> instantiateCoreAll[ EXPR$ /@ Take[ instList1, {s, s + n - 1}], p + 1]]
					]
				]]]
			],
			OFVariantProof2[ indices, ac, argInfo]
		]
	]]
OFVariantProof2[ indices_List, ac_Integer, argInfo_List] :=
	With[ {numRemPrem = Length[ indices] + Total[ Replace[ argInfo, {{_Integer, d_Integer, _, _, _} :> d, {Skip, _} :> 1, _ -> 0}, {1}]]},
	Module[ {premIndex = Length[ indices]},
		Join[
			(* ((A1 --> ... --> An --> C) --> (P1 --> ... Pi --> A1) --> ... --> (Pj --> ... --> Pm --> Al) --> (P1 --> ... --> Pm --> C)) *)
			Hold[ INF$[ "impE", 1, ac]],
			(* ((P1 --> ... Pi --> A1) --> ... --> (Pj --> ... --> Pm --> Al) --> (P1 --> ... --> Pm --> A1))
																			...
			   ((P1 --> ... Pi --> A1) --> ... --> (Pj --> ... --> Pm --> Al) --> (P1 --> ... --> Pm --> An)) *)
			Join @@ Replace[ argInfo,
				{
					{Skip, elim_} :> Hold @@ {INF$[ "impE", ++premIndex, elim]},
					{p_, drop_, elim_, _, _} :>
						Append[
							If[ drop === 0,
								Hold[],
							(*else*)
								premIndex += drop;
								Hold @@ Table[ INF$[ "drop", premIndex + 1], {numRemPrem - premIndex}]
							],
							INF$[ "impE", p, drop + elim]
						],
					prf_ :> prf[ numRemPrem]
				},
				{1}
			],
			Hold[ INF$[ "close"]]
		]
	]]

(* Since the 'resolveImplication' we need here differs substantially from the one we need in 'eliminateCoreImplies', we
	give an alternative definition (with more arguments) below. The same is true for 'resolveImplicationAux'. *)
resolveImplicationOF[ args_List, subj_, unifMode_, hyps_, idx_, fix_, tfix_Association, tvar_Association] :=
	Replace[ Reap[ Catch[ resolveImplicationOFAux[ args, subj, {}, Length[ fix], 1, 1]], {0, 1, 2, 3, 4, 5}],	(* 0='unif', 1='argInfo', 2='meta', 3='indices', 4='auto', 5='ac' *)
		{
			{new0:Except[ $Failed], {unif_List, {argInfo_List}, {{meta0_List}}, indices_List, auto0_List, {{ac_Integer}}}} :>
				(* 'argInfo' is a list of 5-tuples '{pos, newPrems, elimPrems, allStart, allLength}' and pairs '{Skip, prems}' and positive integers. *)
				Replace[ unifyFirst[ Join @@ unif, meta0, fix, tfix, tvar, unifMode, "AllowFlexFlex" -> False],	(* there may be several unifiers, but we simply take the first one *)
					{
						{{}, meta1_, inst0_, tinst_, tv_Association} :>
							(* there are no polymorphic a.b.f. term constants in 'unif',
								so we do not need to instantiate types in the proof situation *)
							With[ {inst1 = unfoldTriangularSubst[ inst0]},
							With[ {prfs = tryTerminalProofSteps[ normalize[ replaceTypes[ Join @@ auto0, tinst] /. inst1], hyps, idx, fix, meta1, tfix, tv]},
								If[ ListQ[ prfs],
									With[ {new1 = normalize[ replaceTypes[ new0, tinst] /. inst1]},
									With[ {mp = Sort[ findAtomPositions[ new1, meta1[[All, 1]]], (First[ #1] > First[ #2])&]},
									With[ {mpi = mp[[All, 1]]},
									With[ {metaRem = meta1[[mpi]]},
									With[ {new2 =
											Fold[
												TYPED$[ CoreAll$TM, {#2[[2, 1]]}][ Lambda$TM[ RNG$[ TYPRNG$[ #2[[1, 1]], #2[[2, 1]]]], #1]]&,
												ReplacePart[ new1,
													Join @@ MapIndexed[
														With[ {j = First[ #2]},
															Replace[ Last[ #1],
																(p_ -> abs_) :> ({p} -> VAR$[ abs + j - 1]),
																{1}
															]
														]&,
														mp
													]
												],
												metaRem
											]},
										{
											STERM$[ new2, Association[], tv],
											{
												ac,
												meta0[[All, 1]],
												metaRem[[All, 1]],
												Delete[ meta1, List /@ mpi],
												inst1,
												If[ prfs === {}, argInfo, Replace[ argInfo, j_Integer?Positive :> prfs[[j]], {1}]],
												Join @@ indices
											}
										}
									]]]]],
								(*else*)
									2
								]
							]],
						_ -> 1
					}
				],
			_ -> 0
		}
	]

resolveImplicationOFAux[ {}, subj_, meta_, _, _, _] :=
	(
		Sow[ meta, 2];
		Sow[ assmCount[ subj, 0, Infinity], 5];
		subj
	)
resolveImplicationOFAux[ args_List, TYPED$[ CoreAll$TM, {typ_}][ A_], meta_, fixNum_, p_, a_] :=
	With[ {name = Replace[ A, {Lambda$TM[ RNG$[ TYPRNG$[ n:Except[ Null, _Symbol], _]], _] :> n, _ -> Theorema`Knowledge`VAR$X$TM}]},
	With[ {j = Max[ 0, Max @@ Cases[ meta, (META$[ name, i_Integer] -> _) :> i] + 1]},
	With[ {m = META$[ name, j]},
		Sow[ Length[ meta] + 1, 3];
		resolveImplicationOFAux[ args, applyFunction[ A, m], Append[ meta, m -> {typ, fixNum}], fixNum, p, a]
	]]]
resolveImplicationOFAux[ {"_", r___}, TYPED$[ CoreImplies$TM, {}][ A_][ R_], meta_, fixNum_, p_, a_] :=
	(
		Sow[ {Skip, assmCount[ A, 0, Infinity]}, 1];
		TYPED$[ CoreImplies$TM, {}][ A][ resolveImplicationOFAux[ {r}, R, meta, fixNum, p, a]]
	)
resolveImplicationOFAux[ {"$", r___}, TYPED$[ CoreImplies$TM, {}][ A_][ R_], meta_, fixNum_, p_, a_] :=
	(
		Sow[ a, 1];
		Sow[ A, 4];
		resolveImplicationOFAux[ {r}, R, meta, fixNum, p, a + 1]
	)
resolveImplicationOFAux[ {B_, r___}, TYPED$[ CoreImplies$TM, {}][ A_][ R_], meta_, fixNum_, p_, a_] :=
	With[ {allA = allCount[ A, 0, Infinity, True], allB = allCount[ B, 0, Infinity, True]},
		Replace[ Reap[ resolveImplicationOFAux[ Max[ 0, allB - allA], B, meta, fixNum], -1],
			{B1_, {{meta1_List}}} :>
				With[ {acA = assmCount[ A, 0, Infinity], acB = assmCount[ B1, 0, Infinity]},
				With[ {drop = If[ acB >= acA, acB - acA, acB]},
				With[ {B2 = dropAssms[ B1, drop]},
					Sow[ {A, B2}, 0];
					Sow[ {p, drop, acB - drop, Length[ meta] + 1, Max[ 0, allB - allA]}, 1];
					ReplacePart[ B1, {Table[ 1, {drop}]} -> resolveImplicationOFAux[ {r}, R, meta1, fixNum, p + 1, a]]
				]]]
		]
	]
resolveImplicationOFAux[ _List, _, _, _, _, _] :=
	Throw[ $Failed]
	
resolveImplicationOFAux[ 0, rem_, meta_List, _] :=
	(
		Sow[ meta, -1];
		rem
	)
resolveImplicationOFAux[ n_Integer, TYPED$[ CoreImplies$TM, {}][ A_][ R_], meta_List, fixNum_] :=
	TYPED$[ CoreImplies$TM, {}][ A][ resolveImplicationOFAux[ n, R, meta, fixNum]]
resolveImplicationOFAux[ n_Integer, TYPED$[ CoreAll$TM, {typ_}][ A_], meta_List, fixNum_] :=
	With[ {name = Replace[ A, {Lambda$TM[ RNG$[ TYPRNG$[ v:Except[ Null, _Symbol], _]], _] :> v, _ -> Theorema`Knowledge`VAR$X$TM}]},
	With[ {j = Max[ 0, Max @@ Cases[ meta, (META$[ name, i_Integer] -> _) :> i] + 1]},
	With[ {m = META$[ name, j]},
		resolveImplicationOFAux[ n - 1, applyFunction[ A, m], Append[ meta, m -> {typ, fixNum}], fixNum]
	]]]


(* ::Subsection:: *)
(*for-Variant*)

(* 'forVariant[ args, factID, keyHandler, lbl, add, rem]' is an auxiliary proof method used for instantiating
	universally quantified hypotheses and immediately proving premises of the resulting formula. *)
forVariant[ {args0___, "_"...}, factID_, keyHandler_, lbl_, add_, rem_][ ps_PRFSIT$, __] :=
	With[ {factKeys = getProofSitData[ ps, factID, {}], funabbrevs = termListToFunction[ {args0}, ps, "for"]},
		If[ ListQ[ funabbrevs],
			Hold @@ {
				updateAbbrevLists[ {}, Last[ funabbrevs], {}, {factID}],
				variantIterator[ factKeys, forVariantFun[ First[ funabbrevs], funabbrevs[[2]]], keyHandler, lbl, add, rem]
			},
		(*else*)
			$Failed
		]
	]

forVariantFun[ termsF_, tvarNum_][ i_, PRFSIT$[ _, hyps_, _, fix_, _, tfix_Association, __]] :=
	If[ termsF === Null,
		Null,
	(*else*)
		With[ {hyp = hyps[[i]]},
			Replace[ extractFormula[ hyp, 0],
				{STERM$[ fml_, Association[], tvar_Association], idx_} :>
					With[ {terms = termsF[ idx]},	(* 'termsF' is a function that gives an 'STERM$' when applied to a type variable index *)
						Replace[ makeInstanceof[ terms, fml, tfix, tvar],
							{snew1:STERM$[ new1_, tf_Association, tv_Association], pd1_} :>
								With[ {plainTerms = First[ terms]},
								With[ {firstNotSkipped = First[ FirstPosition[ plainTerms, Except[ "_"], {Infinity}, {1}, Heads -> False]],
										termsNum = Length[ plainTerms],
										assms = getAssumptionsWithUnivNum[ fml]},
								With[ {OF =
										Module[ {aux},
											Map[
												Function[ an,
												With[ {a = First[ an], n = Last[ an]},
													If[ n < firstNotSkipped,
														"_",
													(*else*)
														aux = "_";
														Catch[
															scanVars[ a,
																If[ #3 >= #2,
																	With[ {diff = n + #2 - #3},
																		If[ diff <= termsNum && plainTerms[[diff]] =!= "_", aux = "$", Throw[ "_"]]
																	]
																]&
															];
															aux
														]
													]
												]],
												assms
											]
										]},
								With[ {lastNotSkipped = First[ FirstPosition[ Reverse[ OF], "$", {$Failed}, {1}, Heads -> False]],
										newTFix = Join[ tfix, tf]},		(* keys are disjoint *)
									If[ IntegerQ[ lastNotSkipped],
										Replace[ resolveImplicationOF[ Drop[ OF, 1 - lastNotSkipped], new1, Automatic, hyps, idx + tvarNum, fix, newTFix, tv],
											{
												{new2_STERM$, pd2_List} :>
													{{
														new2,
														forVariantProof[ i, Max[ 0, assms[[-lastNotSkipped, 2]] - termsNum], pd1, pd2],
														preserveFormulaTags[ new2, hyp, newTFix]
													}},
												j:(0|1|2) :>
													(
														sowCPMMessage[ If[ j === 2, 100, 1008 - j], "Fatal", "for"];
														$Failed
													),
												_ -> $Failed
											}
										],
									(*else*)
										{{snew1, ofVariantProof[ i, pd1], preserveFormulaTags[ snew1, hyp, newTFix]}}
									]
								]]]]
						]
					]
			]
		]
	]

forVariantProof[ i_, addFix_Integer, {_Integer, inst_List, tsubst_List}, {ac_Integer, _List, _List, _List, _List, argInfo_List, _List}][ n_] :=
	Module[ {X, counter = -1},
		Join[
			(* "of"-part *)
			Hold @@ Join[
				Table[ INF$[ "allI", FIX$[ X, j]], {j, 0, addFix + Count[ inst, "_"] - 1}],
				{INF$[ "makeExplicit", i + n]},
				Replace[ inst, {"_" :> INF$[ "allE", 1, makeSTERM[ FIX$[ X, ++counter]]], TYPED$[ f_, _] :> INF$[ "allE", 1, makeSTERM[ replaceTypes[ f, tsubst]]]}, {1}],
				Table[ INF$[ "allE", 1, makeSTERM[ FIX$[ X, j]]], {j, counter + 1, counter + addFix}]
			],
			(* "OF"-part *)
			OFVariantProof2[ {}, ac, argInfo]
		]
	]

getAssumptionsWithUnivNum[ G_] :=
	Join @@ Last[ Reap[ getAssumptionsWithUnivNum[ G, 0]]]
getAssumptionsWithUnivNum[ TYPED$[ CoreImplies$TM, {}][ A_][ G_], n_] :=
	(Sow[ {A, n}]; getAssumptionsWithUnivNum[ G, n])
getAssumptionsWithUnivNum[ TYPED$[ CoreAll$TM, {_}][ Lambda$TM[ _RNG$, G_]], n_] :=
	getAssumptionsWithUnivNum[ G, n + 1]
getAssumptionsWithUnivNum[ TYPED$[ CoreAll$TM, {_}][ TYPED$[ CoreImplies$TM, {}][ A_]], n_] :=
	(Sow[ {mapFreeVars[ A, VAR$[ #3 + 1]&], n + 1}];)
getAssumptionsWithUnivNum[ _, _] := Null


(* ::Subsection:: *)
(*in-Variant*)

(* 'inVariant[ args, factID, keyHandler, lbl, add, rem]' is an auxiliary proof method used for instantiating type
	variables in the premises of hypotheses and immediately proving these premises afterward.
	The type variables are instantiated "from left to right", i.e. in the order as they are visited in depth-first
	search. Note, however, that type variables appearing in universal quantifiers are *not* taken into account!
	A premise is attempted to be proved iff
	- at least one type variable appearing in it is instantiated (even if after instantiation some type variables still
		remain), and
	- it is ground (i.e. it does not contain any universally quantified variables). *)
inVariant[ {args0___, "_"...}, factID_, keyHandler_, lbl_, add_, rem_][ ps_PRFSIT$, __] :=
	With[ {factKeys = getProofSitData[ ps, factID, {}], funabbrevs = termListToFunction[ {args0}, ps, "in"]},
		If[ ListQ[ funabbrevs],
			Hold @@ {
				updateAbbrevLists[ {}, Last[ funabbrevs], {}, {factID}],
				variantIterator[ factKeys, inVariantFun[ First[ funabbrevs], funabbrevs[[2]]], keyHandler, lbl, add, rem]
			},
		(*else*)
			$Failed
		]
	]

inVariantFun[ termsF_, tvarNum_][ i_, PRFSIT$[ _, hyps_, _, fix_, _, tfix_Association, __]] :=
	If[ termsF === Null,
		Null,
	(*else*)
		With[ {hyp = hyps[[i]]},
			Replace[ extractFormula[ hyp, 0],
				{fml_STERM$, idx_} :>
					With[ {assms = getAssumptionsWithUnivNum[ First[ fml]]},
					Module[ {tsubst = {}},
					With[ {OF =
							Module[ {visited = {}, types = Replace[ First[ termsF[ idx]], TYPED$[ _, typ_] :> typ, {1}]},
								Map[
									Function[ an,
										If[ types === {},
											"_",
										(*else*)
											With[ {a = First[ an], n = Last[ an], typesNum = Length[ types]},
											With[ {tvarsAll = DeleteDuplicates[ Cases[ a, _TVAR$, {0, Infinity}, Heads -> True]]},
											With[ {tvarsNew = Select[ tvarsAll, !MemberQ[ visited, #] &]},
											With[ {newNum = Length[ tvarsNew]},
												visited = Join[ visited, tvarsNew];
												If[ newNum > typesNum,
													tsubst = Join[ tsubst, DeleteCases[ MapThread[ Rule, {Take[ tvarsNew, typesNum], types}], _ -> "_"]];
													types = {},
												(*else*)
													tsubst = Join[ tsubst, DeleteCases[ MapThread[ Rule, {tvarsNew, Take[ types, newNum]}], _ -> "_"]];
													types = Drop[ types, newNum]
												];
												If[ MemberQ[ tvarsAll, _?(MemberQ[ tsubst, # -> _] &)] && (n === 0 || !containsVar[ a, #3 >= #2 &]), "$", "_"]
											]]]]
										]
									],
									assms
								]
							]},
					With[ {new1 = replaceTypes[ fml, tsubst], ac = LengthWhile[ assms, Last[ #] === 0 &]},
					With[ {lastNotSkipped = First[ FirstPosition[ Reverse[ OF], "$", {$Failed}, {1}, Heads -> False]]},
						If[ IntegerQ[ lastNotSkipped],
							Replace[ resolveImplicationOF[ Drop[ OF, 1 - lastNotSkipped], First[ new1], Automatic, hyps, idx + tvarNum, fix, tfix, fml[[3]]],
								{
									{new2_STERM$, pd2_List} :>
										{{new2, OFVariantProof[ i, {}, pd2], preserveFormulaTags[ new2, hyp, tfix]}},
									j:(0|1|2) :>
										(
											sowCPMMessage[ If[ j === 2, 100, 1008 - j], "Fatal", "in"];
											$Failed
										),
									_ -> $Failed
								}
							],
						(*else*)
							(* we let "impE" compute the type substitution, to make sure that type variable indices are correct *)
							{{new1, Hold[ INF$[ "makeExplicit", i + #], INF$[ "impE", 1, ac]]&, preserveFormulaTags[ new1, hyp, tfix]}}
						]
					]]]]]
			]
		]
	]


(* ::Subsection:: *)
(*Introduction*)

(* 'introduceAssms[ assms, n, implicit, ac, skip, split][ ps, ___]' makes the explicit assumptions 'assms' of the first
	subgoal of 'ps' implicit, drops them, and furthermore drops the first 'n' of the remaining explicit assumptions.
	'assms' is supposed to be a (possibly empty) list of elements of the form
	- '{spec, mod}', where 'spec' is either an index (possibly negative) or a literal Theorema formula (possibly inside
		'EXPR$' and/or 'LABELED$'), 'mod' is '0', '1' or '2'; in that case, the specified assumption is made
		implicit ('mod' equals '1' or '2') and dropped ('mod' equals '0' or '2'),
	- 'STRING$[ lbl]', which is a short-hand for '{LABELED$[ i, lbl], 2}', where 'i' is the index of the first assumption
		that has not been considered yet, or
	- 'spec', which is a short-hand for '{spec, 2}'.
	'n' is either a non-negative integer or 'All' or "all", specifying how many of the *remaining* explicit assumptions
	shall be made implicit and/or dropped.
	'implicit' is a boolean value specifying whether the remaining 'n' assumptions shall only be dropped ('False';
	default) or also made implicit ('True').
	'ac' is either a non-negative integer or 'All' or "all", specifying how many of the explicit assumptions of the
	current subgoal can be introduced *at most* (i.e. all others are treated as being part of the atomic goal); 'ac'
	may be larger than the total number of explicit assumptions.
	'skip' specifies whether non-existing assumptions shall simply be skipped with a warning, or make the whole proof
	method fail. No messages are generated if 'skip' is 'False'!
	'split' specifies whether, after possible type instantiations, the first subgoal of 'ps' shall be split off before
	introducing/dropping explicit assumptions. This parameter has its canonical application in connection with proof
	method 'selectSubgoal'.
	The order of the new implicit hypotheses is as follows: first come the assumptions that are introduced explicitly,
	in the order they are introduced (no matter in which order they appear in the goal), and then come all remaining
	assumptions that are also introduced (if any), in the order they appear in the goal.
	The proof method fails if
	- there are more than one subgoals and at least once assumption shall be made implicit,
	- one of the assumptions that shall be made implicit contains a meta variable or a type variable that also appears
		elsewhere. *)
introduceAssms[ assms_List] := introduceAssms[ assms, 0, False, All, True, False]
introduceAssms[ assms_List, n_] := introduceAssms[ assms, n, False, All, True, False]
introduceAssms[ assms_List, n_, implicit_] := introduceAssms[ assms, n, implicit, All, True, False]
introduceAssms[ assms_List, n_, implicit_, ac_] := introduceAssms[ assms, n, implicit, ac, True, False]
introduceAssms[ assms_List, n_, implicit_, ac_, skip_] := introduceAssms[ assms, n, implicit, ac, skip, False]
introduceAssms[ {}, 0, _, _, _, split_][ ps:PRFSIT$[ G_List, __], ___] :=
	Hold @@ makeSplittingCommandList[ split && Length[ G] > 1, ps]
introduceAssms[ assms_List, n_, implicit_, ac_, skip_, split_][ ps:PRFSIT$[ {G_, ___}, hyps_List, __], args___, 1] :=
	Replace[ executeProofCommands[ Hold[ introduceAssms[ assms, n, implicit, ac, skip, split]], {ps}, 0, False],
		{
			PRFSTAT$[ _, sits:{PRFSIT$[ {G0_, ___}, hyps0_List, __], ___}, global_List, thms_List] :>
				{
					makeANDNODE[ makePRFINFO[ name -> introduceAssms, used -> {{G}}, generated -> {Drop[ hyps0, -Length[ hyps]], {G0}}], sits],
					global,
					thms,
					Hold[]
				},
			_ -> $Failed
		}
	]
introduceAssms[ {}, All|"all", implicit_, ac_, _, split_][ ps:PRFSIT$[ {G_, ___}, __], ___] :=
	introduceAssms[ ps, assmCount[ G, 0, Replace[ ac, All|"all" -> Infinity]], implicit, split]
introduceAssms[ {}, n_Integer?Positive, implicit_, ac_, skip_, split_][ ps:PRFSIT$[ {G_, ___}, __], ___] :=
	With[ {m = assmCount[ G, 0, Min[ n, Replace[ ac, All|"all" -> Infinity]]]},
		If[ m < n,
			If[ TrueQ[ skip],
				sowCPMMessage[ 95, "Warning"];
				introduceAssms[ ps, m, implicit, split],
			(*else*)
				$Failed
			],
		(*else*)
			introduceAssms[ ps, m, implicit, split]
		]
	]
introduceAssms[ ps:PRFSIT$[ {G_, R___}, _, _, fix_, _, tfix_Association, __], n_Integer, implicit_, split_] :=
	If[ n === 0,
		Hold[],
	(*else*)
		Hold @@ If[ TrueQ[ implicit],
			With[ {tsubst =
					typeSpecializationReplacement[
						Replace[ DeleteDuplicates[ Cases[ ReplacePart[ G, Table[ 1, {n}] -> Null], _FIX$, {0, Infinity}, Heads -> True]], fix, {1}],
						Keys[ tfix]
					]},
				Join[
					If[ tsubst === {}, {}, sowCPMMessage[ 1002, "Warning", "introduceAssms"]; {INF$[ "instT", tsubst]}],
					makeSplittingCommandList[ split && {R} =!= {}, ps],
					Join @@ Table[ {INF$[ "makeImplicit", 1], INF$[ "drop", 1]}, {n}],
					{updateAbbrevLists[ {"this" -> Range[ n]}, {}, {}, {}]}
				]
			],
		(*else*)
			Join[
				makeSplittingCommandList[ split && {R} =!= {}, ps],
				Table[ INF$[ "drop", 1], {n}],
				{INF$[ "psData", setProofSitData[ ps, {}, {}]]}
			]
		]
	]
introduceAssms[ assms0_List, n0_, implicit_, ac0_, skip_, split_][ ps:PRFSIT$[ {G_, R___}, _, idx_, fix_, meta_, tfix_, tvar_, __], ___] :=
	Catch[
		Block[ {$introAssmsList = Most[ toAssmsGoal[ G, {}, Replace[ ac0, All|"all" -> Infinity]]],
				$introAssmsIdx = idx,
				$introAssmsFix = fix,
				$introAssmsMeta = meta,
				$introAssmsTFix = tfix,
				$introAssmsTVar = tvar,
				$introAssmsIndices = {},
				$introAssmsWarning = False,
				$introAssmsTS = {},
				$introAssmsSubst = {},
				$introAssmsAbbrevs = getTermAbbrevList[ ps],
				$introAssmsNewAbbrevs = {}},
		With[ {ac = Length[ $introAssmsList], multi = ({R} =!= {})},
		With[ {assms = DeleteCases[ introduceAssmsAux[ ac, split =!= True && multi, skip] /@ assms0, Null]},
		With[ {impl1 = Cases[ assms, {_Integer, Except[ 0, _Integer], _String, ___}]},
			If[ $introAssmsWarning, sowCPMMessage[ 96, "Warning"]];
			With[ {len = Length[ $introAssmsIndices]},
			With[ {n =
					Switch[ n0,
						All|"all",
						ac - len,
						_Integer?NonNegative,
						If[ n0 + len > ac,
							sowCPMMessage[ 95, "Warning"];
							ac - len,
						(*else*)
							n0
						],
						_,
						sowCPMMessage[ 97, {n0}, "Fatal"];
						Throw[ $Failed]
					]},
				Replace[
					Reap[
						Module[ {j = 0, m = n, current = 1},
							Join[
								Join @@ Map[
									(
										++j;
										Replace[ #,
											{i_Integer, mod_Integer, _String, ___} :>
												With[ {diff = i - j},
												With[ {min = Min[ m, diff]},
												With[ {d =
														If[ Positive[ min],
															If[ TrueQ[ implicit], Sow[ Range[ j, j + min - 1]]];
															Table[ INF$[ "drop", current], {min}],
														(*else*)
															{}
														]},
													current += Max[ 0, diff - m];
													m = Max[ 0, m - diff];
													j = i;
													Join[ d,
														If[ mod === 0,
															{INF$[ "drop", current]},	(* drop only *)
														(*else*)
															If[ mod === 1,
																++current; {},		(* make implicit only *)
															(*else*)
																{INF$[ "drop", current]}	(* make implicit and drop *)
															]
														]
													]
												]]]
										]
									)&,
									SortBy[ assms, First]
								],
								If[ Positive[ m],
									If[ TrueQ[ implicit], Sow[ Range[ j + 1, j + m]]];
									Table[ INF$[ "drop", current], {m}],
								(*else*)
									{}
								]
							]
						]
					],
					{drop_List, impl_List} :>
						With[ {impl2 = Flatten[ impl, 2], num = Length[ impl1]},
							$introAssmsList = Part[ $introAssmsList, Join[ impl1[[All, 1]], impl2]];
							With[ {allFix = DeleteDuplicates[ Cases[ $introAssmsList, _FIX$, {0, Infinity}, Heads -> True]]},
							With[ {tsubst = typeSpecializationReplacement[ Join[ $introAssmsList, Replace[ allFix, $introAssmsFix, {1}]], Keys[ $introAssmsTFix]]},
								If[ tsubst =!= {},
									sowCPMMessage[ 98, "Warning"];
									$introAssmsTS = Join[ tsubst, $introAssmsTS /. tsubst]
								];
								Hold @@ Join[
									{INF$[ "sorts", $introAssmsTVar]},
									If[ $introAssmsTS === {}, {}, {INF$[ "instT", $introAssmsTS]}],
									Cases[ $introAssmsSubst, (m_ -> inst_) :> INF$[ "inst", m -> STERM$[ normalize[ inst], Association[], Association[]]]],
									makeSplittingCommandList[ split && multi, ps],
									INF$[ "makeImplicit", #]& /@ Reverse[ impl2],
									Join @@ Replace[
										Reverse[ impl1],
										{i_, _, lbl_String, data___} :> Join[ {INF$[ "makeImplicit", i]}, If[ {lbl, data} === {"?"}, {}, {INF$[ "hypData", 1, lbl, data]}]],
										{1}
									],
									{updateAbbrevLists[ {"this" -> Range[ num]}, Prepend[ $introAssmsNewAbbrevs, Hold[ ellipsisAbbrev[ num]]], {}, {}]},
									drop
								]
							]]
						]
				]
			]]
		]]]]
	]

introduceAssmsAux[ ac_, multi_, skip_][ {LABELED$[ spec_, lbl_String, data___], mod_Integer}] :=
	introduceAssmsAux[ ac, multi, skip][ {spec, mod, lbl, data}]
introduceAssmsAux[ ac_, multi_, skip_][ {LABELED$[ spec_, _, data___], mod_Integer}] :=
	introduceAssmsAux[ ac, multi, skip][ {spec, mod, "?", data}]
introduceAssmsAux[ ac_, multi_, skip_][ {spec_, mod_Integer}] :=
	introduceAssmsAux[ ac, multi, skip][ {spec, mod, "?"}]
introduceAssmsAux[ ac_, multi_, skip_][ {spec_Integer?Negative, mod_Integer, data__}] :=
	introduceAssmsAux[ ac, multi, skip][ {ac + spec + 1, mod, data}]
introduceAssmsAux[ ac_, multi_, skip_][ all:{spec_Integer, mod_Integer, __}] :=
	Which[
		1 <= spec <= ac,
		Which[
			multi && mod =!= 0,
			sowCPMMessage[ 99, "Fatal"];
			Throw[ $Failed],
		(*else*)
			MemberQ[ $introAssmsIndices, spec],
			If[ TrueQ[ skip],
				Null,
			(*else*)
				Throw[ $Failed]
			],
		(*else*)
			True,
			$introAssmsIndices = Union[ $introAssmsIndices, {spec}];
			all
		],
	(*else*)
		TrueQ[ skip],
		$introAssmsWarning = True;
		Null,
	(*else*)
		True,
		Throw[ $Failed]
	]
introduceAssmsAux[ ac_, multi_, skip_][ {f:(_EXPR$|_PEXPR$), mod_Integer, data__}] :=
	Replace[ makeCoreExprPatt[ f, FORM$TM, $introAssmsIdx, $introAssmsFix, If[ mod === 0, $introAssmsMeta, {}], $introAssmsTFix, $introAssmsTVar, $introAssmsAbbrevs, None, normalize, None],
		{
			{{_, STERM$[ expr_, newTFix_, newTVar_], idx_Integer, {}, {}}, ts_List, {}, abb_} :>
				(
					$introAssmsList = replaceTypes[ $introAssmsList, ts];
					$introAssmsIdx = idx;
					$introAssmsFix = $introAssmsFix /. ts;
					$introAssmsMeta = $introAssmsMeta /. ts;
					$introAssmsTFix = Join[ $introAssmsTFix, newTFix];	(* keys are disjoint *)
					$introAssmsTVar = newTVar;
					$introAssmsTS = Join[ ts, $introAssmsTS /. ts];
					$introAssmsNewAbbrevs = Join[ abb, $introAssmsNewAbbrevs];
					introduceAssmsAux[ ac, multi, skip][ {expr, mod, data}]
				),
			_ :> If[ TrueQ[ skip], $introAssmsWarning = True; Null, Throw[ $Failed]]
		}
	]
introduceAssmsAux[ ac_, multi_, skip_][ {f_, mod_Integer, data__}] :=
	Replace[
		Catch[
			Module[ {i = 0, tag},
				Scan[
					If[ !MemberQ[ $introAssmsIndices, ++i],
						Replace[ lFirst[ unify[ {{#, f}}, $introAssmsMeta, $introAssmsFix, $introAssmsTFix, $introAssmsTVar, tag, "AllowFlexFlex" -> False]],
							{
								{{}, rem_, subst_, ts_, tv_Association} :>
									(
										clearIterator[ tag];
										If[ !MemberQ[ rem, (v_ -> _) /; !MemberQ[ $introAssmsMeta, v -> _]],
											(* there must be no new meta variables *)
											Throw[ {i, subst, ts, tv}]
										]
									),
								_ :> (clearIterator[ tag];)
							}
						]
					]&,
					$introAssmsList
				]
			]
		],
		{
			{i_Integer, subst_List, ts0_List, tv_Association} :>
				With[ {ts = If[ $introAssmsIdx === 0, {}, Cases[ ts0, HoldPattern[ TVAR$[ _Integer?(# < $introAssmsIdx &), ___] -> _]]]},
					$introAssmsList = normalize[ Fold[ ReplaceAll, replaceTypes[ $introAssmsList, ts], subst], TrueQ[ $autoNormalize] && subst =!= {}];
					$introAssmsFix = $introAssmsFix /. ts;
					$introAssmsMeta = DeleteCases[ $introAssmsMeta, (v_ -> _) /; MemberQ[ subst, v -> _]] /. ts;
					$introAssmsTVar = tv;
					$introAssmsTS = Join[ ts, $introAssmsTS /. ts];
					$introAssmsSubst = Join[ $introAssmsSubst /. ts, subst];
					introduceAssmsAux[ ac, multi, skip][ {i, mod, data}]
				],
			_ :> If[ TrueQ[ skip], $introAssmsWarning = True; Null, Throw[ $Failed]]
		}
	]
introduceAssmsAux[ ac_, multi_, skip_][ STRING$[ lbl_String]] :=
	introduceAssmsAux[ ac, multi, skip][
		{Module[ {j = 0}, Catch[ Scan[ If[ # > ++j, Throw[ j]]&, $introAssmsIndices]; j + 1]], 2, lbl}
	]
introduceAssmsAux[ ac_, multi_, skip_][ spec_] :=
	introduceAssmsAux[ ac, multi, skip][ {spec, 2}]

makeSplittingCommandList[ ps:PRFSIT$[ {_, __}, __]] :=
	makeSplittingCommandList[ True, ps]
makeSplittingCommandList[ ps_PRFSIT$] :=
	makeSplittingCommandList[ False, ps]
makeSplittingCommandList[ split_, ps_PRFSIT$] :=
	If[ TrueQ[ split],
		With[ {chained = getProofSitData[ ps, "chainedFacts", {}]},
			If[ chained === {},
				{INF$[ "split"]},
			(*else*)
				{
					INF$[ "psData", setProofSitData[ ps, {}, {"chainedFacts"}]],
					INF$[ "split"],
					INF$[ "psData", setProofSitData[ ps, {"chainedFacts" -> chained}, {}]]
				}
			]
		],
	(*else*)
		{}
	]

(* 'introduceAssmsFix[ case, lbl, consts][ ps, __]' splits off the subgoal of 'ps' corresponding to 'case', updates the
	"cases"-entry in 'ps', introduces the outermost universal quantifiers using the a.b.f. constants in 'consts', and
	finally makes all outermost explicit assumptions implicit, collecting them in collection 'lbl'.
	'case' may be a non-string, in which case the first subgoal is split off without updating "cases", and 'lbl' may be
	'Automatic', in which case it is set to 'case'.
	Note that the proof method also succeeds if 'ps' has only one subgoal, 'consts' is empty and there are no explicit
	assumptions. *)
introduceAssmsFix[] :=
	introduceAssmsFix[ Null]
introduceAssmsFix[ case_] :=
	introduceAssmsFix[ case, Automatic]
introduceAssmsFix[ case_, lbl_] :=
	introduceAssmsFix[ case, lbl, {}]
introduceAssmsFix[ case_, lbl_, consts_List][ ps:PRFSIT$[ {G_, ___}, hyps_List, __], args___, 1] :=
	Replace[ executeProofCommands[ Hold[ introduceAssmsFix[ case, lbl, consts]], {ps}, 0, False],
		{
			PRFSTAT$[ _, sits:{PRFSIT$[ {G0_, ___}, hyps0_List, __], ___}, global_List, thms_List] :>
				{
					makeANDNODE[ makePRFINFO[ name -> introduceAssmsFix, used -> {{G}}, generated -> {Drop[ hyps0, -Length[ hyps]], {G0}}, "data" -> {case, lbl, consts}], sits],
					global,
					thms,
					Hold[]
				},
			_ -> $Failed
		}
	]
introduceAssmsFix[ case_, lbl0_, {}][ ps:PRFSIT$[ G_List, __], ___, Except[ 1]] :=
	Replace[ splitGoalByCase[ case, ps],
		{cmd_Hold, i_Integer} :>
			With[ {assms = assmCount[ G[[i]], 0, Infinity],
					lbl = If[ lbl0 === Automatic, case, lbl0]},
			With[ {hyps = If[ Positive[ assms], Range[ assms, 1, -1], Null]},
				Join[
					cmd,
					Hold @@ Append[
						Join @@ Table[ {INF$[ "makeImplicit", 1], INF$[ "drop", 1]}, {assms}],
						updateAbbrevLists[ Join[ {"this" -> hyps}, If[ StringQ[ lbl], {lbl -> hyps}, {}]], {}, {}, {}]
					]
				]
			]]
	]
introduceAssmsFix[ case_, lbl_, consts_List][ ps:PRFSIT$[ G_List, __], ___, Except[ 1]] :=
	Replace[ splitGoalByCase[ case, ps],
		{cmd_Hold, _Integer} :>
			Join[ cmd, Hold @@ {introduceFix[ consts], introduceAssmsFix[ Null, If[ lbl === Automatic, case, lbl], {}]}]
	]

(* 'splitGoalByCase' is an auxiliary function used for determining which subgoal to split off from the given proof
	situation, and for generating corresponding proof commands. *)
splitGoalByCase[ case_String, ps:PRFSIT$[ G_List, __]] :=
	With[ {cases = getProofSitData[ ps, "cases", {}]},
	With[ {patt = Replace[ case, cases]},
		If[ patt === case,
			sowCPMMessage[ 101, {case}, "Fatal"];
			$Failed,
		(*else*)
			With[ {i = First[ FirstPosition[ G, patt, {0}, {1}, Heads -> False]]},
				If[ i === 0,
					sowCPMMessage[ 102, {case}, "Fatal"];
					$Failed,
				(*else*)
					With[ {newCases = DeleteCases[ cases, case -> _, {1}, 1]},
						{
							Hold @@ Join[
								If[ newCases === {},
									{updateAbbrevLists[ {}, {}, {}, {"cases"}]},
								(*else*)
									{updateAbbrevLists[ {}, {}, {"cases" -> newCases}, {}]}
								],
								If[ i === 1, {}, {INF$[ "rearrangeG", i, 1]}],
								{INF$[ "split"]}
							],
							i
						}
					]
				]
			]
		]
	]]
splitGoalByCase[ Except[ _String], PRFSIT$[ {_}, __]] :=
	{Hold[], 1}
splitGoalByCase[ Except[ _String], PRFSIT$[ {_, __}, __]] :=
	{Hold[ INF$[ "split"]], 1}
splitGoalByCase[ _, _] :=
	$Failed
