(* Core proof methods handling known facts. *)

(* ::Subsection:: *)
(*Adding Knowledge*)

(* 'addKnowledgeImplicit[ l, coll, acc, src, internal, ruleCat][ ps, __]' adds the knowledge specified by the list 'l'
	to the implicit hypotheses of the proof situation 'ps'. 'l' is a list whose elements may be
	- indices, which refer to the implicit hypotheses,
	- keys,
	- labels inside 'STRING$', possibly including also source information (in which case *all* formulas with the
		respective labels in the implicit hypotheses are added, but only *the first* formula in the global
		environment); note that first global fact collections are searched, and only if not fact collection with the
		given name is in scope, then formulas are searched,
	- instances of theorem schemas inside 'THM$' (even if theorems are not included in 'src', see below),
	- literal formulas inside 'EXPR$', referring to known facts among the implicit hypotheses (only the *first*
		matching formula is added); the global environment is ignored regardless of 'src',
	- plain strings referring to *local* fact collections stored in the optional "factLists" parameter of 'ps',
	- the string "calculation", which refers to the optional "calc" parameter of 'ps' (if existing),
	- variants of facts inside 'VARIANT$'.
	'coll' determines which entry in the meta-data of 'ps' shall refer to the keys of the newly added facts afterward.
	If 'coll' is no string, no entry in the meta-data of 'ps' will refer to the keys.
	'acc' is a boolean value specifying whether the keys of the new facts shall be added to 'coll' ('True') or whether
	'coll' shall be replaced by the list of new keys (any other value). It only takes effect if 'coll' is a string.
	'src' specifies where to search for knowledge:
	- '0' means that only implicit hypotheses are searched,
	- '1' means that implicit hypotheses and the global environment are searched,
	- '2' means that implicit hypotheses and registered theorems are searched, and
	- '3' means that all these collections are searched.
	- 'Automatic' means that the value of the last invocation of 'addKnowledgeImplicit' or 'addKnowledgeExplicit' is
	taken.
	'internal' is a boolean value indicating whether 'coll' is supposed to be an "internal" fact list (like
	"chainedFacts") that not only stores the keys of formulas but also permanent/temporary information ('True'), or
	whether it is an ordinary list of facts in "factLists" (any other value). It only takes effect if 'coll' is a
	string.
	'ruleCat' can be a string or 'None'. If it is a string, all facts specified by labels are replaced by their
	corresponding rule version(s) w.r.t. 'ruleCat', if possible. This means that, for instance, instead of adding fact
	"rule 1" its rule version "rule 1 simp 1" is added.
	Note that the proof method fails completely if facts cannot be found or obtained. *)
addKnowledgeImplicit[ l_List, coll_, acc_, src_, internal_, ruleCat_][ ps_PRFSIT$, args___] :=
	(
		(* must be unfolded only when applied *)
		Which[
			IntegerQ[ src],
			$addKnowledgeSource = src,
		(*else*)
			!IntegerQ[ $addKnowledgeSource],
			$addKnowledgeSource = 3
		];
		$addKnowledgeFactID = -1;
		If[ TrueQ[ internal] && StringQ[ coll],
			With[ {meth = addKnowledgeAux[ l, addKeysToList[ coll, False], Hold[], ruleCat]},
				If[ TrueQ[ acc],
					meth[ ps, args],
				(*else*)
					With[ {data = setProofSitData[ ps, {coll -> {}}, {}]},
						Hold[ INF$[ "psData", data], meth]
					]
				]
			],
		(*else*)
			With[ {id = getFreshFactID[]},
				addKnowledgeAux[ l, addKeysToList[ id, False], Hold[ addKnowledgeImplicitPost[ coll, acc, id]], ruleCat][ ps, args]
			]
		]
	)

(* 'addKnowledgeExplicit[ l, sub, src, cmd, perm, ruleCat][ ps, __]' adds the knowledge specified by the list 'l' to
	the explicit assumptions of the first 'sub' subgoals of the proof situation 'ps'.
	'l' has the same meaning as in 'addKnowledgeImplicit'.
	'sub' is an integer or 'All'; if it is an integer, the new knowledge is added to the first 'sub' subgoals.
	'src' has the same meaning as in 'addKnowledgeImplicit'.
	'cmd' is a command sequence that is applied after the knowledge has been added.
	'perm', is a boolean value specifying whether the *new* implicit hypotheses shall be kept after applying
	'cmd', or removed again; new implicit hypotheses might arise because even if new knowledge shall be made explicit,
	it might be necessary to add it to the implicit hypotheses first. The commands in 'cmd' may add and/or remove
	arbitrary hypotheses themselves, without affecting the correct behavior of the proof method.
	'ruleCat' has the same meaning as in 'addKnowledgeImplicit'.
	Note that the proof method fails completely if facts cannot be found or obtained. *)
addKnowledgeExplicit[ l_List, sub:(_Integer|All), src_, cmd_, perm_, ruleCat_][ args__] :=
	(
		(* must be unfolded only when applied *)
		Which[
			IntegerQ[ src],
			$addKnowledgeSource = src,
		(*else*)
			!IntegerQ[ $addKnowledgeSource],
			$addKnowledgeSource = 3
		];
		$addKnowledgeFactID = -1;
		If[ TrueQ[ perm],
			addKnowledgeAux[ l, makeExplicitAssms[ #, sub]&, cmd, ruleCat],
		(*else*)
			With[ {id = getFreshFactID[]},
				addKnowledgeAux[ l, addKeysToList[ id, False], Hold[ addKnowledgeExplicitPost[ cmd, sub, perm, id]], ruleCat]
			]
		][ args]
	)
	
addKnowledgeImplicitPost[ coll_, acc_, id_][ ps:PRFSIT$[ _, hyps_, __], __] :=
	With[ {keys = getProofSitData[ ps, id, {}]},
	With[ {plainKeys = Replace[ keys, {k:{_, _}, _} :> k, {1}]},
		Hold @@ {
			updateAbbrevLists[
				Prepend[
					If[ !StringQ[ coll] || (TrueQ[ acc] && plainKeys === {}),
						{},
					(*else*)
						{coll -> Join[ If[ TrueQ[ acc], Lookup[ getProofSitData[ ps, "factLists", Association[]], coll, {}], {}], plainKeys]}
					],
					"this" -> plainKeys
				],
				If[ plainKeys === {}, {}, Join @@ Cases[ hyps, FML$[ Last[ plainKeys], f_, __] :> ellipsisAbbrev[ f], {1}, 1]],
				{},
				If[ StringQ[ id], {}, {id}]	(* if 'id' is a string, don't remove it (could be "chainedFacts" etc.) *)
			]
		}
	]]

addKnowledgeExplicitPost[ cmd_, sub_, perm_, id_][ ps:PRFSIT$[ _, hyps_, __], __] :=
	With[ {keys = getProofSitData[ ps, id, {}]},
		If[ TrueQ[ perm],
			If[ StringQ[ id],
				cmd,	(* if 'id' is a string, don't remove it (could be "chainedFacts" etc.) *)
			(*else*)
				Prepend[ cmd, INF$[ "psData", setProofSitData[ ps, {}, {id}]]]
			],
		(*else*)
			Join[
				Hold[ INF$[ "psData", setProofSitData[ ps, {}, {id}]], makeExplicitAssms[ keys, sub]],
				cmd,
				With[ {forget = Cases[ keys, {k_List, False} :> k]},
					If[ forget === {}, Hold[], Hold[ forgetHyps[ forget]]]
				]
			]
		]
	]

addKnowledgeAux[ spec0_List, keyHandler_, cmdPost_Hold, src_, ruleCat_][ args__] :=
	(
		(* must be unfolded only when applied *)
		Which[
			IntegerQ[ src],
			$addKnowledgeSource = src,
		(*else*)
			!IntegerQ[ $addKnowledgeSource],
			$addKnowledgeSource = 3
		];
		If[ !IntegerQ[ $addKnowledgeFactID], $addKnowledgeFactID = -1];
		addKnowledgeAux[ spec0, keyHandler, cmdPost, ruleCat][ args]
	)
addKnowledgeAux[ spec0_List, keyHandler_, cmdPost_Hold, ruleCat_][ ps:PRFSIT$[ _, hyps_, __], __] :=
	Block[ {$addKnowledgeHypNum = -Length[ hyps],
			$addKnowledgeKeyHandler = keyHandler,
			$addKnowledgeHyps = hyps,
			$addKnowledgeFactLists = getProofSitData[ ps, "factLists", Association[]],
			$addKnowledgeCalcList = Replace[ getProofSitData[ ps, "calc", $Failed], {ky:{_, _}, _} :> ky, {1}],
			$addKnowledgeProofSit = ps,
			$addKnowledgeRuleCat = ruleCat,
			$addKnowledgeErrFun = (sowCPMMessage[ 11, {#}, "Fatal"]; Throw[ $Failed])&,
			$addKnowledgeErrFun2 = (sowCPMMessage[ 12, {#}, "Fatal"]; Throw[ $Failed])&,
			$addKnowledgeAdded = 0,
			$addKnowledgeCmd2 = Null},
	Module[ {spec = spec0},
		Catch[
			Replace[
				Reap[
					While[ $addKnowledgeCmd2 === Null && spec =!= {},
						spec = Join[ Most[ spec], addSingleAssm[ Last[ spec], Automatic, {}, {}]]
					],
					{0, 1, 2}
				],
				{_, {cmd1_List, ki_List, cmdR_List}} :>
					(* 'ki' stands for "keys and/or indices", 'cmdR' stands for "renaming commands" *)
					Join[
						Hold @@ Flatten[ cmd1, 2],
						If[ ki === {} || keyHandler === Null, Hold[], Hold @@ {keyHandler[ Reverse[ Flatten[ ki, 2]]]}],
						If[ cmdR === {}, Hold[], Hold @@ {renameHyps[ Flatten[ cmdR, 2]]}],
						If[ MatchQ[ $addKnowledgeCmd2, _Hold], $addKnowledgeCmd2, Hold[]],
						If[ spec === {}, cmdPost, Hold[ addKnowledgeAux[ spec, keyHandler, cmdPost, ruleCat]]]
					]
			]
		]
	]]

addSingleAssm[ i:Except[ 0, _Integer], lbl_, add_, rem_] :=
	With[ {j = If[ Negative[ i], -$addKnowledgeHypNum + i + 1, i]},
		If[ 1 <= j <= -$addKnowledgeHypNum,
			With[ {n = $addKnowledgeHypNum + j - 1},
				Sow[ {n}, 1];
				If[ StringQ[ lbl] || !(add === {} === rem), Sow[ {{n, lbl, add, rem}}, 2]]
			];
			{},
		(*else*)
			$addKnowledgeErrFun[ i]
		]
	]
addSingleAssm[ k:{_String?(StringMatchQ[ #, "ID" ~~ __] &), _String?(StringMatchQ[ #, "Source" ~~ __] &)}, lbl_, add_, rem_] :=
	With[ {p = Position[ $addKnowledgeHyps, _[ k, __], {1}, 1, Heads -> False]},
		If[ p === {},
			If[ OddQ[ $addKnowledgeSource],
				With[ {fml = getFormulas[ True][ k]},
					Which[
						fml === $Failed,
						$addKnowledgeErrFun[ k],
					(*else*)
						inScopeQ[ fml],
						++$addKnowledgeAdded;
						Sow[ {INF$[ "insert", k]}, 0];
						Sow[ {{$addKnowledgeHypNum - $addKnowledgeAdded, False}}, 1];
						If[ StringQ[ lbl] || !(add === {} === rem), Sow[ {{$addKnowledgeHypNum - $addKnowledgeAdded, lbl, add, rem}}, 2]];
						{},
					(*else*)
						True,
						$addKnowledgeErrFun2[ k]
					]
				],
			(*else*)
				$addKnowledgeErrFun[ k]
			],
		(*else*)
			With[ {n = $addKnowledgeHypNum + p[[1, 1]] - 1},
				Sow[ {n}, 1];
				If[ StringQ[ lbl] || !(add === {} === rem), Sow[ {{n, lbl, add, rem}}, 2]]
			];
			{}
		]
	]
addSingleAssm[ FML$[ k_List, __], lbl_, add_, rem_] :=
	With[ {p = Position[ $addKnowledgeHyps, _[ k, __], {1}, 1, Heads -> False]},
		(* Passing a formula means that the given formula is certainly in the global theory context and certainly comes
			before the current goal, but it might also be in the implicit hypotheses already. *)
		If[ p === {},
			++$addKnowledgeAdded;
			Sow[ {INF$[ "insert", k]}, 0];
			Sow[ {{$addKnowledgeHypNum - $addKnowledgeAdded, False}}, 1];
			If[ StringQ[ lbl] || !(add === {} === rem), Sow[ {{$addKnowledgeHypNum - $addKnowledgeAdded, lbl, add, rem}}, 2]];
			{},
		(*else*)
			With[ {n = $addKnowledgeHypNum + p[[1, 1]] - 1},
				Sow[ {n}, 1];
				If[ StringQ[ lbl] || !(add === {} === rem), Sow[ {{n, lbl, add, rem}}, 2]]
			];
			{}
		]
	]
addSingleAssm[ STRING$[ l_String], lbl_, add_, rem_] :=
	With[ {goalSrc = Quiet[ Check[ $proofGoal[[1, 2]], None]]},
	With[ {ky = If[ OddQ[ $addKnowledgeSource], getFactListKeys[ l, 1, goalSrc], {}]},
		Replace[ If[ ky === {}, $Failed, Lookup[ $globalFactLists, Key @@ ky, $Failed]],
			{
				facts0_List :>
					With[ {facts1 = Cases[ facts0, {f_, _List?inScopeTSQ} :> f]},
						If[ StringQ[ lbl],
							LABELED$[ #, lbl, "tags" -> add, "remtags" -> rem]& /@ facts1,
						(*else*)
							facts1
						]
					],
				_ :>
					With[ {p = Reverse[ formulaPosition[ l, Infinity, None, $addKnowledgeHyps]]},
						Which[
							p === {},
							With[ {q = If[ OddQ[ $addKnowledgeSource], Reverse[ formulaKeys[ l, 1, Quiet[ Check[ $proofGoal[[1, 2]], None]]]], {}]},
									(* Don't use 'key' or 'source' to access the source of '$proofGoal', because there must not be 'unexpected'. *)
								If[ q === {},
									If[ $addKnowledgeSource >= 2,
										addSingleAssm[ THM$[ l], lbl, add, rem],
									(*else*)
										$addKnowledgeErrFun[ STRING$[ l]]
									],
								(*else*)
									With[ {q1 = getRuleFormulas[ $addKnowledgeRuleCat, q, False, True, True, True]},
										If[ q1 =!= {},
											Sow[ Map[ INF$[ "insert", #]&, q1], 0];
											With[ {len = Length[ q1]},
												$addKnowledgeAdded += len;
												Sow[ Table[ {$addKnowledgeHypNum + i - $addKnowledgeAdded, False}, {i, 0, len - 1}], 1];
												If[ (StringQ[ lbl] && lbl =!= l) || !(add === {} === rem),
													Sow[ Table[ {$addKnowledgeHypNum + i - $addKnowledgeAdded, lbl, add, rem}, {i, 0, len - 1}], 2]
												]
											];
											{},
										(*else*)
											$addKnowledgeErrFun2[ STRING$[ l]]
										]
									]
								]
							],
						(*else*)
							StringQ[ $addKnowledgeRuleCat] && OddQ[ $addKnowledgeSource],
							Flatten[
								Map[
									With[ {hyp = $addKnowledgeHyps[[First[ #]]]},
									With[ {rules = getRuleFormulas[ $addKnowledgeRuleCat, hyp, False, False, True, True]},
										If[ rules === {hyp},
											#,	(* '#' is a list *)
										(*else*)
											rules
										]
									]]&,
									p
								],
								1
							],
						(*else*)
							True,
							Sow[ Map[ ($addKnowledgeHypNum + First[ #] - 1)&, p], 1];
							If[ StringQ[ lbl] || !(add === {} === rem), Sow[ Map[ {$addKnowledgeHypNum + First[ #] - 1, lbl, {}, {}}&, p], 2]];
							{}
						]
					]
			}
		]
	]]
addSingleAssm[ e0_EXPR$, lbl_, add_, rem_] :=
	Replace[ makeCoreExpr[ e0, FORM$TM, $addKnowledgeProofSit, None, normalize, None],	(* This is necessary for unfolding (global) abbreviations. *)
		{
			{{_, STERM$[ e_, _, _], _, _, _}, _} :>
				With[ {patt = termToTermPatt[ e]},	(* We completely ignore type- and sort information when retrieving literal facts. *)
				With[ {p = Reverse[ Position[ $addKnowledgeHyps, _[ _, patt, ___], {1}, 1, Heads -> False]]},
					If[ p === {},
						$addKnowledgeErrFun[ e0],
					(*else*)
						Sow[ Map[ ($addKnowledgeHypNum + First[ #] - 1)&, p], 1];
						If[ StringQ[ lbl] || !(add === {} === rem), Sow[ Map[ {$addKnowledgeHypNum + First[ #] - 1, lbl, add, rem}&, p], 2]];
						{}
					]
				]],
			_ :> (sowCPMMessage[ 19, {e0}, "Fatal"]; Throw[ $Failed])
		}
	]
addSingleAssm[ thm0_THM$, lbl_, add_, rem_] :=
	Replace[ fastThmFromLabel[ thm0],
		{
			thm:THM$[ k_, ___] :>
				If[ getTheorems[ True][ k] === $Failed,
					$addKnowledgeErrFun[ thm0],
				(*else*)
					++$addKnowledgeAdded;
					Sow[ {addTheoremImplicit[ thm, lbl, add, rem]}, 0];
					Sow[ {{$addKnowledgeHypNum - $addKnowledgeAdded, False}}, 1];
					{}
				],
			_ :> $addKnowledgeErrFun[ thm0]
		}
	]
addSingleAssm[ coll_String, lbl_, add_, rem_] :=
	If[ coll === "calculation",
		(* "calculation" always refers to "calc", which cannot be stored as an ordinary list of facts because of the
			permanent/temporary information *)
		Replace[ $addKnowledgeCalcList,
			{
				{a_} :> addSingleAssm[ a, lbl, add, rem],
				l_List :>
					If[ StringQ[ lbl],
						LABELED$[ #, lbl, "tags" -> add, "remtags" -> rem]& /@ l,
					(*else*)
						l
					],
				_ :> (sowCPMMessage[ 13, "Error"]; Throw[ $Failed])
			}
		],
	(*else*)
		Replace[ Lookup[ $addKnowledgeFactLists, coll, $Failed],
			{
				{a_} :> addSingleAssm[ a, lbl, add, rem],
				l_List :>
					If[ StringQ[ lbl],
						LABELED$[ #, lbl, "tags" -> add, "remtags" -> rem]& /@ l,
					(*else*)
						l
					],
				_ :> $addKnowledgeErrFun[ coll]
			}
		]
	]
addSingleAssm[ LABELED$[ a_, lbl1_, data___], lbl2_, add_, rem_] :=
	With[ {add0 = Replace[ "tags", {data, _ -> {}}], rem0 = Replace[ "remtags", {data, _ -> {}}]},
		addSingleAssm[
			a,
			If[ StringQ[ lbl1], lbl1, lbl2],
			DeleteDuplicates[ Join[ DeleteCases[ add, Alternatives @@ rem0], add0]],
			DeleteDuplicates[ Join[ DeleteCases[ rem, Alternatives @@ add0], rem0]]
		]
	]
addSingleAssm[ VARIANT$[ a_, method_, args_], lbl_, add_, rem_] :=
	addSingleAssm[ VARIANT$[ a, method, args, None], lbl, add, rem]
addSingleAssm[ VARIANT$[ a_, method_, args_, ruleCat_], lbl_, add_, rem_] :=
	With[ {factID = getFreshFactID[]},
		$addKnowledgeCmd2 =
			Hold @@ {
				addKnowledgeAux[ {a}, addKeysToList[ factID, False], Hold[], ruleCat],
				method[ args, factID, $addKnowledgeKeyHandler, lbl, add, rem]
			};
		{}
	]
addSingleAssm[ {a_}, lbl_, add_, rem_] :=
	addSingleAssm[ a, lbl, add, rem]
addSingleAssm[ a_List, lbl_, add_, rem_] :=
	If[ lbl === Automatic && add === {} === rem,
		a,
	(*else*)
		LABELED$[ #, lbl, "tags" -> add, "remtags" -> rem]& /@ a
	]
addSingleAssm[ a_, _, _, _] :=
	$addKnowledgeErrFun[ a]
	
(* 'addKeysToList[ id, perm][ list][ ps, ___]' prepends the keys of the implicit hypotheses of 'ps', specified by
	'list', to the optional parameter 'id' of 'ps'; if 'id' does not yet exist, it is initialized by the empty list
	first. The elements of 'id' can be indices, keys, or pairs of the form '{index/key, permanent}', where 'permanent'
	is a boolean value. Indices may be negative, of course, and are replaced by the keys of the hypotheses they refer
	to. If 'perm' is 'True', all pairs in 'list' are replaced by their first element only, forgetting the 'permanent'
	information. *)
addKeysToList[ id_, perm_][ list_][ ps:PRFSIT$[ _, hyps_, __], __] :=
	If[ list === {},
		Hold[],
	(*else*)
		With[ {n = Length[ hyps] + 1},
		With[ {keys =
				If[ TrueQ[ perm],
					Replace[ list,
						{
							i_Integer :> If[ Negative[ i], hyps[[n + i, 1]], hyps[[i, 1]]],
							{i_Integer, _} :> If[ Negative[ i], hyps[[n + i, 1]], hyps[[i, 1]]],
							{k:{_String, _String}, _} :> k
						},
						{1}
					],
				(*else*)
					Replace[ list,
						{
							i_Integer :> If[ Negative[ i], hyps[[n + i, 1]], hyps[[i, 1]]],
							{i_Integer, permanent_} :> {If[ Negative[ i], hyps[[n + i, 1]], hyps[[i, 1]]], permanent}
						},
						{1}
					]
				]},
			Hold @@ {INF$[ "psData", setProofSitData[ ps, {id -> Join[ keys, getProofSitData[ ps, id, {}]]}, {}]]}
		]]
	]

(* 'initializeKeyList[ id, perm][ list][ ps, ___]' proceeds analogously to 'addKeysToList', but it does not *add* keys
	to an existing list, but rather replaces any existing list. This means that it should only be used as a key handler
	if at most *one* new fact is added. *)
initializeKeyList[ id_, perm_][ list_][ ps:PRFSIT$[ _, hyps_, __], __] :=
	If[ list === {},
		Hold @@ {INF$[ "psData", setProofSitData[ ps, {id -> {}}, {}]]},
	(*else*)
		With[ {n = Length[ hyps] + 1},
		With[ {keys =
				If[ TrueQ[ perm],
					Replace[ list,
						{
							i_Integer :> If[ Negative[ i], hyps[[n + i, 1]], hyps[[i, 1]]],
							{i_Integer, _} :> If[ Negative[ i], hyps[[n + i, 1]], hyps[[i, 1]]],
							{k:{_String, _String}, _} :> k
						},
						{1}
					],
				(*else*)
					Replace[ list,
						{
							i_Integer :> If[ Negative[ i], hyps[[n + i, 1]], hyps[[i, 1]]],
							{i_Integer, permanent_} :> {If[ Negative[ i], hyps[[n + i, 1]], hyps[[i, 1]]], permanent}
						},
						{1}
					]
				]},
			Hold @@ {INF$[ "psData", setProofSitData[ ps, {id -> keys}, {}]]}
		]]
	]

getFreshFactID[] := (++$addKnowledgeFactID)


(* ::Subsection:: *)
(*Extracting Sublists of Fact-Lists*)

(* For obtaining variants of formulas, 'method[ args, factID, keyHandler, lbl, add, rem][ ps, ___]' must do the
	following:
	- Obtain the variants of the hypotheses specified by parameter 'factID' of 'ps', using 'args'.
	- If 'lbl' is a string, set the labels of the new hypotheses to 'lbl'. Also set the tags of the new hypotheses to
		'add' and, if a "new" hypothesis is actually an existing one, remove the tags contained in 'rem'. In that case
		'method' should not "return" existing hypotheses, but really create new ones (possibly duplicating them),
		unless the old and the new labels as well as the old and the new tags agree.
	- Collect the keys/indices and markers of the new formulas in a list 'keys' and apply proof method
		'keyHandler[ keys]'; 'keys' must be in the right order (not reversed).
		Newly generated facts must get marked as "temporary", whereas existing facts must get the same marker as in
		'factID'.
	- Forget all hypotheses marked as temporary in 'factID' and not needed any more.
	- Remove 'factID' from the parameters of 'ps'. *)

(* 'partVariant[ args, factID, keyHandler, lbl, add, rem]' is an auxiliary proof method used for extracting lists of
	hypotheses of the current proof situation. *)
partVariant[ ran:(_Integer|_Span), factID_, keyHandler_, lbl_, add_, rem_][ ps:PRFSIT$[ G_, hyps_, idx_, fix_, meta_, tfix_, tvar_, fid_, ctxt___], ___, mode_] :=
	Catch[
		With[ {keys = getProofSitData[ ps, factID, {}]},
		With[ {num = Length[ keys]},
		With[ {indices = If[ IntegerQ[ ran], {ran}, Quiet[ Check[ Part[ Range[ num], ran], sowCPMMessage[ 14, {ran}, "Fatal"]; Throw[ $Failed]]]]},
		With[ {keepKeys = Quiet[ Check[ Part[ keys, indices], sowCPMMessage[ 14, {ran}, "Fatal"]; Throw[ $Failed]]],
				dropKeys = Cases[ Delete[ keys, List /@ indices], {k:{_, _}, False} :> k]},
		With[ {keepNum = Length[ keepKeys], aux = (lbl === Automatic && add === {} === rem)},
		With[ {dropInd = Union[ Map[ Position[ hyps, FML$[ #, __], {1}, 1, Heads -> False][[1]]&, dropKeys]] + If[ aux, 0, keepNum],
				dupInd = If[ aux, {}, Map[ Position[ hyps, FML$[ #, __], {1}, 1, Heads -> False][[1, 1]]&, keepKeys]],
				post = partVariant[ keyHandler, If[ aux, keepKeys, keepNum]]},
			If[ mode === 2,
				Hold @@ Join[
					{updateAbbrevLists[ {}, {}, {}, {factID}]},
					Join @@ MapIndexed[
						With[ {hyp = hyps[[#1]]},
						With[ {fml = First[ extractFormula[ hyp, 0]], newData = changeLabelTags[ lbl, add, rem, hyp]},
							{
								INF$[ "cut", fml],
								INF$[ "makeExplicit", #1 + First[ #2] - 1],
								INF$[ "impE", 1, assmCount[ First[ fml], 0, Infinity]],
								INF$[ "close"],
								Join[ INF$[ "hypData", 1], INF$ @@ DeleteCases[ newData, "TypeVarPositions" -> _]]
							}
						]]&,
						Reverse[ dupInd]
					],
					MapIndexed[ INF$[ "forget", First[ #1] + 1 - First[ #2]]&, dropInd],
					{post}
				],
			(*else*)
				{
					PRFSIT$[
						G,
						Delete[
							Join[
								MapIndexed[
									With[ {newData = changeLabelTags[ lbl, add, rem, #1]},
										Join[ FML$[ makeFormulaKey[ fid + First[ #2] - 1], #1[[2]]], FML$ @@ newData]
									]&,
									hyps[[dupInd]]
								],
								hyps
							],
							dropInd
						],
						idx, fix, meta, tfix, tvar, fid + keepNum, ctxt
					],
					{},
					{},
					Hold[ updateAbbrevLists[ {}, {}, {}, {factID}], post]
				}
			]
		]]]]]]
	]
partVariant[ keyHandler_, keys0_][ ps:PRFSIT$[ _, hyps_, __], args__] :=
	With[ {keys = If[ IntegerQ[ keys0], {key@#, False}& /@ Take[ hyps, keys0], keys0]},
		keyHandler[ keys][ ps, args]
	]


(* ::Subsection:: *)
(*Obtaining Variants of Hypotheses*)

(* 'variantIterator[ facts, fun, keyHandler, lbl, add, rem]' is an auxiliary proof method used for obtaining variants
	of hypotheses in the current proof situation 'ps'.
	'facts' is the list of keys of the known facts whose variants shall be obtained.
	'fun' is a function that obtains the variants: it is called with the two arguments 'i' and 'ps', where 'i' is the
	index of the hypothesis and 'ps' is the whole proof situation. It is supposed to return either '$Failed' or a non-
	empty list of pairs/triples '{new, prf, (data)}' or a list of rules 'data' or 'Null'. A list of pairs/triples
	'{new, prf, (data)}' represents all the obtained variants 'new' (which must be 'STERM$'-structures) together with
	their proofs 'prf', where the 'prf' are functions that, when applied to an integer 'n', return a proof of the
	respective 'new'; 'n' specifies how many additional hypotheses have been added since 'fun' was invoked. A list of
	rules 'data' means that the variant of the formula is identical to the formula itself, but that the meta-data of
	the formula shall be changed. 'Null' means that nothing needs to be done.
	'keyHandler' has the same meaning as always in the context of knowledge insertion.
	'lbl' is the label of the new hypotheses. It can be 'Automatic', a literal string or a function; in the latter
	case, it is applied to the original label and expected to return again either 'Automatic' or a literal string.
	'add' and 'rem' specify the tags that shall be added resp. removed to the list of tags returned by 'fun' (if any).
	Attention! Tags of hypotheses are *not* passed along to obtained variants automatically! If this is desired, 'fun'
	must take care of it by returning corresponding 'data' lists.
	Note that if the new label differs from the original one, or the new tags differ from the original ones, existing
	permanent hypotheses are duplicated even if the obtained variants are identical. *)
variantIterator[ facts_List, fun_, keyHandler_, lbl_, add_, rem_] :=
	variantIterator[ facts, fun, keyHandler, lbl, add, rem, 0, {}]
variantIterator[ {}, _, keyHandler_, _, _, _, n_, keys_][ ps:PRFSIT$[ _, hyps_, __], args__] :=
	keyHandler[ Join[ {key@#, False}& /@ Take[ hyps, n], keys]][ ps, args]
variantIterator[ {pre___, ky_}, fun_, keyHandler_, lbl0_, add_, rem_, n_, keys0_][ ps:PRFSIT$[ G_, hyps_, idx_, fix_, meta_, tfix_, tvar_, fid_, ctxt___], ___, mode_] :=
	With[ {keys = Join[ {key@#, False}& /@ Take[ hyps, n], keys0],
			i = Position[ hyps, FML$[ Replace[ ky, {k:{_, _}, _} :> k], _, __], {1}, 1, Heads -> False][[1, 1]]},
	With[ {hyp = hyps[[i]]},
	With[ {lbl = Switch[ lbl0, Automatic|_String, lbl0, _, lbl0[ label@hyp]]},
		Replace[ fun[ i, ps],
			{
				Null :>
					If[ (StringQ[ lbl] && lbl =!= label@hyp) || !(add === {} === rem),
						changeLabelData[ Null, i, ky, keys, {pre}, fun, keyHandler, lbl, add, rem, ps, mode],
					(*else*)
						Hold @@ {variantIterator[ {pre}, fun, keyHandler, lbl, add, rem, 0, Prepend[ keys, ky]]}
					],
				data:{___Rule} :>
					changeLabelData[ data, i, ky, keys, {pre}, fun, keyHandler, lbl, add, rem, ps, mode],
				l:{{_, __}..} :>
					With[ {len = Length[ l]},
						If[ mode === 2,
							Catch[
								Join[
									Join @@ MapIndexed[
										With[ {prf = #1[[2]][ First[ #2] - 1], newData = adjustLabelTags[ lbl, add, rem, "?", Quiet[ Check[ #1[[3]], {}]]]},
											If[ MatchQ[ prf, _Hold],
												Join[
													Hold[ INF$[ "cut", First[ #1]]],
													prf,
													If[ newData =!= {"?"},
														Hold @@ {INF$ @@ Join[ {"hypData", 1}, newData]},
													(*else*)
														Hold[]
													]
												],
											(*else*)
												Throw[ $Failed]
											]
										]&,
										Reverse[ l]
									],
									If[ MatchQ[ ky, {{_, _}, False}], Hold @@ {INF$[ "forget", i + len]}, Hold[]],
									Hold[ variantIterator[ {pre}, fun, keyHandler, lbl, add, rem, len, keys]]
								]
							],
						(*else*)
							(* no separate definition for trace mode *)
							{
								PRFSIT$[
									G,
									Join[
										MapIndexed[
											With[ {new = First[ #1], newData = adjustLabelTags[ lbl, add, rem, "?", Quiet[ Check[ #1[[3]], {}]]]},
												Join[
													FML$[
														makeFormulaKey[ fid + First[ #2] - 1],
														First[ new]
													],
													FML$ @@ newData,
													FML$[ "TypeVarPositions" -> makeTypeVarPositions[ new]]
												]
											]&,
											l
										],
										If[ MatchQ[ ky, {{_, _}, False}], Delete[ hyps, i], hyps]
									],
									idx, fix, meta, tfix, tvar, fid + len, ctxt
								],
								{},
								{},
								Hold[ variantIterator[ {pre}, fun, keyHandler, lbl, add, rem, len, keys]]
							}
						]
					],
				_ -> $Failed
			}
		]
	]]]

changeLabelData[ data0_, i_, ky_, keys_List, pre_List, fun_, keyHandler_, lbl0_, add_, rem_, PRFSIT$[ G_, hyps_, idx_, fix_, meta_, tfix_, tvar_, fid_, ctxt___], mode_] :=
	With[ {hyp = hyps[[i]]},
	With[ {lbldata = If[ ListQ[ data0], adjustLabelTags[ lbl0, add, rem, label@hyp, data0], adjustLabelTags[ lbl0, add, rem, label@hyp, List @@ Drop[ hyp, 3]]]},
		If[ MatchQ[ ky, {{_, _}, False}],
			With[ {post = variantIterator[ pre, fun, keyHandler, lbl0, add, rem, 0, Prepend[ keys, ky]]},
				If[ mode === 2,
					Hold @@ {
						INF$ @@ Join[ {"hypData", i}, DeleteCases[ lbldata, "TypeVarPositions" -> _]],
						post
					},
				(*else*)
					{
						PRFSIT$[ G, ReplacePart[ hyps, i -> Join[ Take[ hyp, 2], FML$ @@ lbldata]], idx, fix, meta, tfix, tvar, fid, ctxt],
						{},
						{},
						Hold[ post]
					}
				]
			],
		(*else*)
			With[ {fml = First[ extractFormula[ hyp, 0]]},
				If[ mode === 2,
					Hold @@ {
						INF$[ "cut", fml],
						INF$[ "makeExplicit", i],
						INF$[ "impE", 1, assmCount[ First[ fml], 0, Infinity]],
						INF$[ "close"],
						INF$ @@ Join[ {"hypData", 1}, DeleteCases[ lbldata, "TypeVarPositions" -> _]],
						variantIterator[ pre, fun, keyHandler, lbl0, add, rem, 1, keys]
					},
				(*else*)
					(* no separate definition for trace mode *)
					{
						PRFSIT$[
							G,
							Prepend[
								hyps,
								Join[
									FML$[
										makeFormulaKey[ fid],
										First[ fml]
									],
									FML$ @@ lbldata,
									FML$[ "TypeVarPositions" -> Replace[ "TypeVarPositions", Append[ List @@ Drop[ hyp, 3], _ :> makeTypeVarPositions[ fml]]]]
								]
							],
							idx, fix, meta, tfix, tvar, fid + 1, ctxt
						],
						{},
						{},
						Hold[ variantIterator[ pre, fun, keyHandler, lbl0, add, rem, 1, keys]]
					}
				]
			]
		]
	]]

(* In contrast to 'changeLabelTags', 'adjustLabelTags' always returns a list. *)
adjustLabelTags[ Automatic, {}, {}, orig_String, data_List] :=
	Prepend[ data, orig]
adjustLabelTags[ lbl_, add_, rem_, orig_String, data_List] :=
	With[ {tags = Replace[ Replace[ "tags", data], "tags" -> {}],
			newLbl = Switch[ lbl, STRING$[ _String], First[ lbl], _String, lbl, Automatic, orig, _, lbl[ orig]]},
	With[ {newTags = DeleteDuplicates[ Join[ DeleteCases[ tags, Alternatives @@ rem], add]]},
		If[ newTags === {},
			Prepend[ DeleteCases[ data, "tags" -> _], newLbl],
		(*else*)
			Join[ {newLbl, "tags" -> newTags}, DeleteCases[ data, "tags" -> _]]
		]
	]]


(* ::Subsection:: *)
(*Adding Theorems*)

(* 'addTheoremImplicit[ thm, lbl, add, tags][ ps, __]' adds the given theorem, if existing, to the implicit hypotheses
	of 'ps'. 'lbl' is the label of the new hypothesis; it can be a string, 'Automatic' or a function that is applied to
	the name of the theorem. 'add' and 'rem' are the tags that shall be added to resp. removed from the original list
	of tags of the theorem. *)
addTheoremImplicit[ thm:THM$[ k_, args___], lbl0_, add_List, rem_List][ PRFSIT$[ G_, hyps_, idx_, fix_, meta_, tfix_, tvar_, fid_, ctxt___], ___, mode_] :=
	Replace[ getTheorems[ True][ k][ args],
		{
			{f_, tvp_List} :>
				With[ {tags = DeleteDuplicates[ Join[ DeleteCases[ Lookup[ getThmData[ thm], "tags", {}], Alternatives @@ rem], add]],
						lbl =
							Switch[ lbl0,
								STRING$[ _String],
								First[ lbl0],
								_String,
								lbl0,
								_,
								With[ {l = getThmLabel[ thm]},
									If[ lbl0 === Automatic, l, lbl0[ l]]
								]
							]},
					If[ mode === 2,
						With[ {ac = assmCount[ f, 0, Infinity]},
							Hold[ INF$[ "cut", First[ extractFormula[ f, tvp, 0]]], INF$[ "theorem", thm], INF$[ "impE", 1, ac], INF$[ "close"], INF$[ "hypData", 1, lbl, "tags" -> tags]]
						],
					(*else*)
						With[ {fml = FML$[ makeFormulaKey[ fid], f, lbl, "tags" -> tags, "TypeVarPositions" -> tvp]},
						With[ {newPS = PRFSIT$[ G, Prepend[ hyps, fml], idx, fix, meta, tfix, tvar, fid + 1, ctxt]},
							{
								If[ mode === 1,
									makeANDNODE[ makePRFINFO[ name -> addTheoremImplicit, used -> {}, generated -> {fml}, "data" -> thm], newPS],
								(*else*)
									newPS
								],
								{},
								{thm},
								Hold[]
							}
						]]
					]
				],
			_ :> (sowCPMMessage[ 15, {thm}, "Fatal"]; $Failed)
		}
	]


(* ::Subsection:: *)
(*Renaming Hypotheses*)

(* 'renameHyps[ re][ ps, __]' renames the given implicit hypotheses of 'ps'. 're' is supposed to be a list of 4-tuples
	of the form '{spec, lbl, add, rem}', where
	- 'spec' specifies which hypotheses shall be renamed,
	- 'lbl' is the new label (could be 'Automatic', in which the label remains the same, or an arbitrary function that
		is applied to the original label and that must return a string),
	- 'add' and 'rem' are lists of tags that shall be removed resp. added to the tags of the hypotheses.
	The 'spec' can be indices, keys, labels or literal formulas; in the case of labels and literal formulas, *all*
	matching hypotheses are renamed.
	Non-existing hypotheses are simply skipped. *)
renameHyps[ re0_List][ ps:PRFSIT$[ G_, hyps_List, ctxt__], ___, mode_] :=
	With[ {num = Length[ hyps] + 1, warning = (sowCPMMessage[ 16, {#}, "Warning"]; {})&},
	With[ {re =
			Join @@ Replace[ re0,
				{
					{i_Integer, l0_, add_List, rem_List} :>
						With[ {j = If[ Negative[ i], num + i, i]},
							If[ 1 <= j < num,
								{{j, changeLabelTags[ l0, add, rem, hyps[[j]]]}},
							(*else*)
								warning[ i]
							]
						],
					{STRING$[ lbl_String], l0_, add_List, rem_List} :>
						With[ {js = formulaPosition[ lbl, Infinity, None, hyps][[All, 1]]},
							MapThread[ {#1, changeLabelTags[ l0, add, rem, #2]}&, {js, Extract[ hyps, js]}]
						],
					{k_List, l0_, add_List, rem_List} :>
						With[ {js = Position[ hyps, _[ k, __], {1}, 1, Heads -> False]},
							MapThread[ {#1, changeLabelTags[ l0, add, rem, #2]}&, {js, Extract[ hyps, js]}]
						],
					{e0_EXPR$, l0_, add_List, rem_List} :>
						Replace[ makeCoreExpr[ e0, FORM$TM, ps, None, normalize, None],	(* This is necessary for unfolding (global) abbreviations. *)
							{
								{{_, STERM$[ e_, _, _], _, _, _}, _} :>
									With[ {patt = termToTermPatt[ e]},	(* We completely ignore type- and sort information when retrieving literal facts. *)
									With[ {js = Position[ hyps, _[ _, patt, ___], {1}, Heads -> False]},
										MapThread[ {#1, changeLabelTags[ l0, add, rem, #2]}&, {js, Extract[ hyps, js]}]
									]],
								_ :> (sowCPMMessage[ 19, {e0}, "Warning"]; {})
							}
						],
					s_ :> warning[ s]
				},
				{1}
			]},
		If[ mode === 2,
			Hold @@ Cases[ re, {i_, newData_List} :> Join[ INF$[ "hypData", i], INF$ @@ DeleteCases[ newData, "TypeVarPositions" -> _]]],
		(*else*)
			With[ {newPS = PRFSIT$[ G, ReplacePart[ hyps, Cases[ re, {i_, newData_List} :> (i -> Join[ Take[ hyps[[i]], 2], FML$ @@ newData])]], ctxt]},
				{
					If[ mode === 1,
						makeANDNODE[ makePRFINFO[ name -> renameHyps, used -> {}, generated -> {}, "data" -> re], newPS],
					(*else*)
						newPS
					],
					{},
					{},
					Hold[]
				}
			]
		]
	]]

changeLabelTags[ lbl_, add_, rem_, FML$[ _, _, orig_String, data___]] :=
	changeLabelTags[ lbl, add, rem, orig, {data}]
changeLabelTags[ Automatic, {}, {}, _, _] :=
	Null
changeLabelTags[ lbl_, add_, rem_, orig_String, data_List] :=
	With[ {tags = Replace[ Replace[ "tags", data], "tags" -> {}],
			newLbl = Switch[ lbl, STRING$[ _String], First[ lbl], _String, lbl, Automatic, orig, _, lbl[ orig]]},
	With[ {newTags = DeleteDuplicates[ Join[ DeleteCases[ tags, Alternatives @@ rem], add]]},
		Which[
			newLbl === orig && newTags === tags,
			Null,
		(*else*)
			newTags === {},
			Prepend[ DeleteCases[ data, "tags" -> _], newLbl],
		(*else*)
			True,
			Join[ {newLbl, "tags" -> newTags}, DeleteCases[ data, "tags" -> _]]
		]
	]]


(* ::Subsection:: *)
(*Changing Meta-Data of Hypotheses*)

(* 'addToHypData[ data][ ps, ___]' adds the given data to the meta-data of all hypotheses of 'ps'. Data that already
	exists is overwritten. *)
addToHypData[ data_List][ PRFSIT$[ G_, hyps_List, ctxt__], ___, mode_] :=
	If[ mode === 2,
		Hold @@ MapIndexed[
			INF$[ "hypData", First[ #2], #1[[3]], Sequence @@ DeleteDuplicatesBy[ DeleteCases[ Join[ data, List @@ Drop[ #1, 3]], "TypeVarPositions" -> _], First]]&,
			hyps
		],
	(*else*)
		{
			PRFSIT$[ G, Map[ Join[ Take[ #, 3], FML$ @@ DeleteDuplicatesBy[ Join[ DeleteCases[ data, "TypeVarPositions" -> _], List @@ Drop[ #, 3]], First]]&, hyps], ctxt],
			{},
			{},
			Hold[]
		}
	]


(* ::Subsection:: *)
(*Forgetting Knowledge*)

(* 'forgetHyps[ assms][ ps, ___]' removes the implicit assumptions 'assms' from the list of hypotheses of 'ps'.
	'assms' is supposed to be a list of indices, labels, keys or literal formulas inside 'EXPR$'. Non-existing
	assumptions are simply skipped.
	Note that all indices in 'assms' refer to the *current* list of hypotheses; for instance, if the first two
	assumptions shall be removed, 'assms' must be '{1, 2}'. *)
forgetHyps[ {}] :=
	voidTac[]
forgetHyps[ assms:Except[ {}, _List]][ ps:PRFSIT$[ G_, hyps_List, ctxt__], ___, mode_] :=
	With[ {num = Length[ hyps], warning = (sowCPMMessage[ 17, {#}, "Warning"]; {})&},
	With[ {indices =
			Union @@ Replace[ Replace[ assms, LABELED$[ a_, ___] :> a, {1}],
				{
					i_Integer :>
						Which[
							i === 0 || i < -num || i > num,
							warning[ i],
						(*else*)
							Negative[ i],
							{{num + i + 1}},
						(*else*)
							True,
							{{i}}
						],
					STRING$[ lbl_String] :> formulaPosition[ lbl, Infinity, None, hyps],
					k_List :> Position[ hyps, _[ k, __], {1}, 1, Heads -> False],
					e0_EXPR$ :>
						Replace[ makeCoreExpr[ e0, FORM$TM, ps, None, normalize, None],	(* This is necessary for unfolding (global) abbreviations. *)
							{
								{{_, STERM$[ e_, _, _], _, _, _}, _} :>
									With[ {patt = termToTermPatt[ e]},	(* We completely ignore type- and sort information when retrieving literal facts. *)
										Position[ hyps, _[ _, patt, ___], {1}, Heads -> False]
									],
								_ :> (sowCPMMessage[ 19, {e0}, "Warning"]; {})
							}
						],
					s_ :> warning[ s]
				},
				{1}
			]},
		If[ mode === 2,
			Hold @@ MapIndexed[
				INF$[ "forget", First[ #1] + 1 - First[ #2]]&,
				indices
			],
		(*else*)
			{
				PRFSIT$[ G, Delete[ hyps, indices], ctxt],
				{},
				{},
				Hold[]
			}
		]
	]]


(* ::Subsection:: *)
(*Internal Facts*)

(* 'internalFactHandler[ coll, method, expl, spec][ ps, __]' makes the facts in internal collection 'coll' of 'ps'
	explicit in the first/all/no subgoals of 'ps', depending on 'expl', then forgets all temporary facts and removes
	the 'coll' entry from 'ps' depending on the value of 'spec' (see below), and finally applies the given 'method'.
	'spec' may be '0', '1' or '2':
	- '0' means that the facts are kept just as they are, i.e. even the temporary hypotheses are kept and the 'coll'
		entry of 'ps' is kept as well.
	- '1' means that again all hypotheses are kept, but the 'coll' entry is removed.
	- '2' means that the temporary hypotheses are removed, and also the 'coll' entry is removed.
	By default, 'coll' is "chainedFacts".
	Note: If the facts shall be made explicit in the first subgoal, but there is no subgoal, the proof method still
	succeeds without a warning! *)
internalFactHandler[ method_, expl_, spec_] :=
	internalFactHandler[ "chainedFacts", method, expl, spec]
internalFactHandler[ _, method_, None, 0] :=
	method
internalFactHandler[ coll_, method_, expl:(First|All), spec_][ ps:PRFSIT$[ G_List, hyps_List, __], args__] :=
	Catch[
		With[ {keys0 = getProofSitData[ ps, coll, {}]},
			If[ keys0 === {},
				If[ MatchQ[ method, _INF$],
					Hold[ method],
				(*else*)
					method[ ps, args]
				],
			(*else*)
				With[ {indices =
						Map[
							Replace[ Position[ hyps, FML$[ Replace[ #, {ky:{_, _}, _} :> ky], __], {1}, 1, Heads -> False],
								{
									{{i_}} :> i,
									_ :> (sowCPMMessage[ 18, "Fatal"]; Throw[ $Failed])
								}
							]&,
							keys0
						]},
					Join[
						If[ spec === 0, Hold[], Hold @@ {INF$[ "psData", setProofSitData[ ps, {}, {coll}]]}],
						With[ {len = Length[ G]},
							mapAllInline[ Hold @@ Map[ INF$[ "makeExplicit", #]&, Reverse[ indices]], Replace[ expl, First -> Min[ 1, len]], len]
						],
						If[ spec === 2,
							Hold @@ MapIndexed[ INF$[ "forget", #1 + 1 - First[ #2]]&, Union[ Extract[ indices, Position[ keys0, {{_, _}, False}, {1}, Heads -> False]]]],
						(*else*)
							Hold[]
						],
						If[ method === voidTac[], Hold[], Hold[ method]]
					]
				]
			]
		]
	]
internalFactHandler[ coll_, method_, None, 2][ ps:PRFSIT$[ _, hyps_, __], args__] :=
	With[ {keys0 = getProofSitData[ ps, coll, {}]},
		If[ keys0 === {},
			If[ MatchQ[ method, _INF$],
				Hold[ method],
			(*else*)
				method[ ps, args]
			],
		(*else*)
			With[ {indices =
					Map[
						Replace[ Position[ hyps, FML$[ #, __], {1}, 1, Heads -> False], {{{i_}} :> i, _ -> Sequence[]}]&,
						Cases[ keys0, {ky:{_, _}, False} :> ky]
					]},
				Hold @@ Join[
					{INF$[ "psData", setProofSitData[ ps, {}, {coll}]]},
					MapIndexed[ INF$[ "forget", #1 + 1 - First[ #2]]&, Union[ indices]],
					If[ method === voidTac[], {}, {method}]
				]
			]
		]
	]
internalFactHandler[ coll_, method_, None, spec:Except[ 0]][ ps_PRFSIT$, args__] :=
	With[ {keys0 = getProofSitData[ ps, coll, {}]},
		If[ keys0 === {},
			If[ MatchQ[ method, _INF$],
				Hold[ method],
			(*else*)
				method[ ps, args]
			],
		(*else*)
			Hold @@ Join[
				{INF$[ "psData", setProofSitData[ ps, {}, {coll}]]},
				If[ method === voidTac[], {}, {method}]
			]
		]
	]
