(* Core proof methods for equational rewriting and simplification. *)

(* Open Issue:
The fact that expressions are automatically normalized whenever replacements are carried out could lead to problems
with the systematic traversal of expressions in rewriting. Namely, after applying a rewrite-rule and normalizing the
result, the original position may not be valid any more, due to beta- and, in particular, eta reductions that might
affect not only the subexpression that was rewritten but also superexpressions thereof. As a quick-and-dirty solution,
function 'adjustSubposition' ensures that after a rewrite the current position always refers to a valid subexpression,
by removing as many of the trailing elements as necessary and updating 'loose' and 'except' accordingly. This approach,
however, could perhaps miss certain subexpressions of the new expression or rewrite others twice.

Alternatively, one could try to normalize expressions "manually", i.e. setting '$autoNormalize' to 'False' and during
rewriting, whenever the current subexpression is a redex, normalize it (and only it!) explicitly (e.g. using
"alphabetaeta") before applying any rewrite-rules. This would also render the whole normal form business more
efficient, since then one can maintain a list of positions of subexpressions that do not need to be looked at any more.
*)

(* ::Subsection:: *)
(*Public Functions and Parameters*)

(* The following symbols appear in "Theorema`CoreProvers`CoreProofMethods`"-context. *)
Scan[
	With[ {$$name = #[[1]], $$desc = #[[2]]},
	With[ {$$args = Join[ ToExpression[ "Theorema`CoreProvers`CoreProofMethods`" <> $$name, InputForm, Hold], Hold[ "usage", "English"]]},
		Set @@ ReplacePart[ Hold[ $$args, $$desc], {1, 0} -> MessageName];
	]]&,
	{
		{"addToNormalForms", ""},
		{"defaultRewritingMaster", ""},
		{"getMicroStrategy", ""},
		{"renameBoundVars", ""},	(* 'renameBoundVars' is only an auxiliary proof method *)
		{"rewritableQ", ""},
		
		{"$rewritingAddNF", ""},
		{"$rewritingContextDepth", ""},
		{"$rewritingDefaultMicroStrategy", ""},
		{"$rewritingDefaultRewriteAssms", ""},
		{"$rewritingDefaultUseAssms", ""},
		(* '$rewritingSimpRules' etc. are like '$simpRules' etc., except that they only contain the specifications of those formulas/theorems that come before the proof goal. *)
		{"$rewritingCompRules", ""},
		{"$rewritingCongRules", ""},
		{"$rewritingSimpRules", ""},
		{"$rewritingRecursionDepth", ""},
		{"$rewritingMaster", ""},
		{"$rewritingPost", ""},
		{"$rewritingSolver", ""}
	}
]

(* A "rewriter" is a function taking six arguments 'nForms', 'expr', 'pos', subpos', 'loose', 'ctxt', where 'expr' is
	the expression currently being rewritten, 'pos' is an integer specifying which part of the goal 'expr' originates
	from (i.e. explicit assumption or atomic goal), 'subpos' specifies the position within that part, and 'loose' and
	'ctxt' are the current type-context information ('ctxt' is itself a 5-tuple '{idx, fix, meta, tfix, tvar}'.
	It returns either a triple '{cmd, succ, fail}' or a 5-tuple '{pre, post, inst, succ, fail}' or something else to
	indicate failure. If the result is a triple,
	- 'cmd' is an arbitrary sequence of proof commands that are supposed to rewrite 'expr'; 'cmd' might in principle do
		anything, but it must not change the structure of the first subgoal. In particular, it must not add/remove
		explicit assumptions.
	- 'succ' and 'fail' can be either 'Continue', 'Skip' or 'Next'; they specify what to do after applying 'cmd' in
		case of success and failure, respectively. 'Continue' means that the rewriting engine continues by trying the
		next rule, 'Skip' means that the expression and its subexpressions are not rewritten any further, and 'Next'
		means that the whole expression shall not be rewritten further by down-rules, but subexpressions may be
		rewritten.
	If the result is a 5-tuple,
	- 'pre' is an arbitrary sequence of proof commands that shall add precisely one new explicit assumption at the
		beginning of the first subgoal (and possibly instantiate type/meta variables, and possibly add further
		hypotheses), *but apart from that leave the proof situation unchanged*. The assumption being added must be a
		(conditional) Core equality, which is then used by '...RewriteSingle' to rewrite the given expression 'expr'
		(possibly after proving conditions).
	- 'inst' is a list of terms used by '...RewriteSingle' in the same way as 'inst2' returned by 'applyEqRewriteRule'
		is used.
	- 'post' is a sequence of proof commands that are applied after '...RewriteSingle'; they shall somehow "restore"
		the original status of the proof situation, e.g. by removing temporary hypotheses again; the assumption added
		in 'pre', however, is removed automatically and does not need to be taken care of by 'post'!
	- 'succ' and 'fail' have the same meaning as above.
	Neither 'cmd' nor 'pre' nor 'post' must create alternative branches.
	If 'cmd', 'pre' or 'post' do not behave as they should, the whole rewriting engine might unexpectedly fail or crash
	at some point.
	Rewriters roughly correspond to "simplification procedures" in Isabelle. *)

If[ !ValueQ[ $rewritingAddNF], $rewritingAddNF = False];
If[ !ValueQ[ $rewritingContextDepth], $rewritingContextDepth = Infinity];
If[ !ValueQ[ $rewritingDefaultMicroStrategy], $rewritingDefaultMicroStrategy = getMicroStrategy[ "lazy"]];		(* "eager" seems to be slightly faster, but "lazy" is more thorough *)
If[ !ValueQ[ $rewritingDefaultRecursionDepth], $rewritingDefaultRecursionDepth = 1];
If[ !ValueQ[ $rewritingDefaultRewriteAssms], $rewritingDefaultRewriteAssms = True];
If[ !ValueQ[ $rewritingDefaultUseAssms], $rewritingDefaultUseAssms = True];
If[ !ValueQ[ $rewritingRewriters], $rewritingRewriters = {NOMATCH}];
If[ !ValueQ[ $rewritingPost], $rewritingPost = Hold[]];

(* A "master" is a function that is called at the very beginning of rewriting and whenever rewriting a part of the
	subgoal is completed. It gets 'nForms', the whole subgoal, the position of the rewritten part, the list of new meta
	variables, the whole list of implicit hypotheses, and the sort constraints on a.b.f. type constants and type
	variables as arguments and returns either 'Continue' or a sequence of proof commands. In the former case, rewriting
	proceeds with the next part of the subgoal, whereas in the latter case it stops and applies the proof commands
	returned by the master instead. These proof commands shall close the first subgoal without affecting the structure
	of any other subgoals.
	The position passed to the master is either 'Null' (meaning that no rewriting has been done so far), a non-negative
	integer 'i' (meaning that the 'i+1'-th assumption has been completed), or a singleton '{i}' (meaning that the
	atomic goal, after 'i' assumptions, has been completed).
	The purpose of a master is to detect trivial subgoals as quickly as possible, for instance if the atomic goal is an
	equality of two identical terms, or an assumption is 'False' in classical logic. *)
If[ !ValueQ[ $rewritingMaster], $rewritingMaster = defaultRewritingMaster];

(* A "solver" is an arbitrary proof method used for recursively proving conditions of rules; note that this proof
	method gets two additional arguments 'depth' and 'sub': 'depth' is an integer or 'Infinity' specifying how many
	recursions are still allowed to prove conditions of rules recursively, and 'sub' is an integer specifying the
	number of subgoals that must be solved.
	'Automatic' means that conditions shall be proved by rewriting. *)
If[ !ValueQ[ $rewritingSolver], $rewritingSolver = Automatic];


(* ::Subsection:: *)
(*Parsing Specifications*)

parseRWSpec[ spec_List, ps:PRFSIT$[ _, hyps_, __], src_String] :=
	Replace[ parseAddRemOnlySpec[ spec, {"", "rw"}, src],
		{gen_List, facts_List, rewriters_List} :>
			Module[ {all = False,
					introFix = True,
					depth = $rewritingDefaultRecursionDepth,
					microstrat = $rewritingDefaultMicroStrategy,
					useAssms = $rewritingDefaultUseAssms,
					rewriteAssms = $rewritingDefaultRewriteAssms},
			With[ {modFun =
					Switch[ #,
						"all",
						all = True,
						"first",
						all = False,
						"fix",
						introFix = True,
						"noFix",
						introFix = False,
						"useAssms",
						useAssms = True,
						"noAssms",
						useAssms = False,
						"rewriteAssms",
						rewriteAssms = True,
						"noRewriteAssms",
						rewriteAssms = False,
						"noUseAssms",
						sowCPMMessage[ 1015, {#, "noAssms"}, "Warning", src],
						"\[Infinity]",
						depth = Infinity,
						_Integer?NonNegative,
						depth = #,
						_,
						If[ StringQ[ #] && StringMatchQ[ #, DigitCharacter..],
							With[ {tmp = ToExpression[ #]},
								If[ IntegerQ[ tmp] && NonNegative[ tmp],
									depth = tmp,
								(*else*)
									sowCPMMessage[ 1014, {#}, "Warning", src]
								]
							],
						(*else*)
							Replace[ getMicroStrategy[ #],
								{
									val:{_, _, _, _} :> (microstrat = val;),
									_ :> sowCPMMessage[ 1014, {#}, "Warning", src]
								}
							]
						]
					]&},
				Scan[ If[ ListQ[ #], Scan[ modFun, DeleteCases[ #, ","]], modFun[ #]]&, gen];
				Catch[
					Join[
						{getRewriters[ rewriters, $rewritingRewriters]},
						Block[ {$getRewriteRulesPS = ps,
								$getRewriteRulesHypNum = Length[ hyps] + 1,
								$getRewriteRulesErrFun = (sowCPMMessage[ Switch[ #2, 0, 155, 1, 156, _, 19], {#1}, "Fatal"]; Throw[ $Failed])&},
							getRewriteRules[ facts, First[ microstrat]]
						],
						{
							useAssms,
							rewriteAssms,
							microstrat,
							depth,
							all,
							introFix
						}
					]
				]
			]]
	]

getRewriters[ {}, acc_] :=
	acc
getRewriters[ {0 -> only_List, rest___}, _] :=
	getRewriters[ {rest}, Cases[ only, s_String :> Symbol[ "Theorema`CoreProvers`" <> s]]]
getRewriters[ {1 -> add_List, rest___}, acc_] :=
	getRewriters[ {rest}, addToList[ acc, Cases[ add, s_String :> Symbol[ "Theorema`CoreProvers`" <> s]], Identity, (sowCPMMessage[ 153, "Warning"]; #)&, False]]
getRewriters[ {2 -> rem_List, rest___}, acc_] :=
	getRewriters[ {rest}, remFromList[ acc, Cases[ rem, s_String :> Symbol[ "Theorema`CoreProvers`" <> s]], Identity, (sowCPMMessage[ 154, "Warning"]; #)&, False]]

getRewriteRules[ {0 -> only_List, rest___}, normaldown_] :=
	Module[ {globDown = Association[], globUp = Association[]},
		Replace[ transposeEmpty[ getRewriteRulesAux[ only, normaldown, True], 6],
			{_, addHypsDown_, addHypsUp_, addGlobDown_List, addGlobUp_List, addVariants_} :>
				getRewriteRules[
					{rest},
					normaldown,
					Join @@ addHypsDown,
					Join @@ addHypsUp,
					Scan[ addRulesToAssociation[ globDown, Prepend, None], Normal[ groupRulesByRuleConsts[ Join @@ addGlobDown, "simp", True]]];
					Hold[ globDown],
					Scan[ addRulesToAssociation[ globUp, Prepend, None], Normal[ groupRulesByRuleConsts[ Join @@ addGlobUp, "simp", True]]];
					Hold[ globUp],
					Join @@ addVariants
				]
		]
	]
getRewriteRules[ spec_, normaldown_] :=
	Module[ {globDown, globUp},
		{globDown, globUp} = initRewriteRules[ normaldown];
		getRewriteRules[
			spec,
			normaldown,
			{},
			{},
			Hold[ globDown],
			Hold[ globUp],
			{}
		]
	]
getRewriteRules[ {}, _, hypsDown_, hypsUp_, Hold[ globDown_], Hold[ globUp_], variants_] :=
	{hypsDown, hypsUp, globDown, globUp, variants}
getRewriteRules[ {1 -> add_List, rest___}, normaldown_, hypsDown_, hypsUp_, Hold[ globDown_], Hold[ globUp_], variants_] :=
	Module[ {dup},
	With[ {aux =
			Fold[
				Function[ {old, new},
					Replace[ new,
						{k_, addHypsDown_, addHypsUp_, addGlobDown_, addGlobUp_, addVariants_} :>
							With[ {out =
									{
										dup = (addVariants === {});
										addToList[ First[ old], addHypsDown, (dup = False; #)&, Identity, False],
										addToList[ old[[2]], addHypsUp, (dup = False; #)&, Identity, False],
										Join[ Last[ old], addVariants]
									}},
								Scan[ addRulesToAssociation[ globDown, Prepend, If[ #1 === Hold[], dup = False]&], Normal[ groupRulesByRuleConsts[ addGlobDown, "simp", True]]];
								Scan[ addRulesToAssociation[ globUp, Prepend, If[ #1 === Hold[], dup = False]&], Normal[ groupRulesByRuleConsts[ addGlobUp, "simp", True]]];
								If[ TrueQ[ dup], sowCPMMessage[ 151, If[ ListQ[ k], {k}, {}], "Warning"]];
								out
							]
					]
				],
				{hypsDown, hypsUp, variants},
				getRewriteRulesAux[ add, normaldown, True]
			]},
		getRewriteRules[ {rest}, normaldown, First[aux], aux[[2]], Hold[ globDown], Hold[ globUp], Last[ aux]]
	]]
getRewriteRules[ {2 -> rem_List, rest___}, normaldown_, hypsDown_, hypsUp_, Hold[ globDown_], Hold[ globUp_], variants_] :=
	Module[ {missing},
	With[ {aux =
			Fold[
				Function[ {old, new},
					Replace[ new,
						{k_, remHypsDown_, remHypsUp_, remGlobDown_, remGlobUp_} :>
							With[ {out =
									{
										missing = True;
										remFromList[ First[ old], remHypsDown, (missing = False; #)&, Identity, False],
										remFromList[ Last[ old], remHypsUp, (missing = False; #)&, Identity, False]
									}},
								Scan[ remRulesFromAssociation[ globDown, If[ #2 =!= 0, missing = False]&], Normal[ groupRulesByRuleConsts[ remGlobDown, "simp", True]]];
								Scan[ remRulesFromAssociation[ globUp, If[ #2 =!= 0, missing = False]&], Normal[ groupRulesByRuleConsts[ remGlobUp, "simp", True]]];
								If[ TrueQ[ missing], sowCPMMessage[ 152, If[ ListQ[ k], {k}, {}], "Warning"]];
								out
							]
					]
				],
				{hypsDown, hypsUp},
				getRewriteRulesAux[ rem, normaldown, False]
			]},
		getRewriteRules[ {rest}, normaldown, First[aux], Last[ aux], Hold[ globDown], Hold[ globUp], variants]
	]]

getRewriteRulesAux[ l_List, normaldown_, allowVariants_] :=
	Join @@ Last[
		Reap[
			scanFactSpec[ l, $getRewriteRulesPS, "simp", !allowVariants, getRewriteRulesAux[ Null, normaldown, allowVariants], $getRewriteRulesErrFun],
			True
		]
	]
getRewriteRulesAux[ Null, normalDown_, allowVariants_][ l_List] :=
	Sow[
		Prepend[
			Join @@@ Last[
				Reap[
					Scan[ getRewriteRulesAux[ normalDown, allowVariants], l],
					Range[ If[ TrueQ[ allowVariants], 5, 4]]
				]
			],
			FirstCase[ l, {_, FML$[ k_, __]} :> k, None]
		],
		True
	]
getRewriteRulesAux[ normaldown_, allowVariants_][ {k_List, fml_FML$}] :=
	If[ TrueQ[ allowVariants] && getOptionalComponent[ fml, "rwData"] === {},
		Sow[ VARIANT$[ k, eqVariant, {"rw"}], 5],
	(*else*)
		With[ {p = First[ FirstPosition[ $getRewriteRulesPS[[2]], _[ k, __], {$Failed}, {1}, Heads -> False]]},
			If[ p === $Failed,
				Sow[ k, If[ isDownRule[ fml, normaldown], 3, 4]],
			(*else*)
				Sow[ p - $getRewriteRulesHypNum, If[ isDownRule[ fml, normaldown], 1, 2]]
			]
		]
	]
getRewriteRulesAux[ normaldown_, allowVariants_][ {i_Integer, fml_FML$}] :=
	If[ TrueQ[ allowVariants] && getOptionalComponent[ fml, "rwData"] === {},
		Sow[ VARIANT$[ key@fml, eqVariant, {"rw"}], 5],
	(*else*)
		Sow[ i - $getRewriteRulesHypNum, If[ isDownRule[ fml, normaldown], 1, 2]]
	]
getRewriteRulesAux[ normaldown_, allowVariants_][ {thm_THM$, _}] :=
	If[ TrueQ[ allowVariants] && Lookup[ getThmData[ thm], "rwData", $Failed] === $Failed,
		Sow[ VARIANT$[ thm, eqVariant, {"rw"}], 5],
	(*else*)
		Sow[ thm, If[ isDownRule[ thm, normaldown], 3, 4]]
	]
getRewriteRulesAux[ normaldown_, allowVariants_][ {v_VARIANT$, _}] :=
	If[ TrueQ[ allowVariants],
		Sow[ VARIANT$[ v, eqVariant, {"rw"}], 5],
	(*else*)
		sowCPMMessage[ 157, "Warning"]
	]

initRewriteRules[ normaldown_] :=
	(
		If[ !MatchQ[ $rewritingSimpRules, _Association],
			extendRULEC$[ $simpRules];
			$rewritingSimpRules = associationFromRULEC$[ $simpRules]
		];
		If[ !MatchQ[ $rewritingCongRules, _Association],
			extendRULEC$[ $congRules];
			$rewritingCongRules = associationFromRULEC$[ $congRules]
		];
		If[ !MatchQ[ $rewritingCompRules, _Association],
			extendRULEC$[ $compRules];
			$rewritingCompRules = associationFromRULEC$[ $compRules]
		];
		If[ TrueQ[ normaldown],
			{Merge[ {$rewritingCongRules, $rewritingSimpRules}, Apply[ Join, #]&], $rewritingCompRules},
		(*else*)
			{$rewritingCongRules, Merge[ {$rewritingCompRules, $rewritingSimpRules}, Apply[ Join, #]&]}
		]
	)

isDownRule[ fml_FML$, True] :=
	!MemberQ[ getOptionalComponent[ fml, "tags"], "comp"]
isDownRule[ thm_THM$, True] :=
	!MemberQ[ Lookup[ getThmData[ thm], "tags", {}], "comp"]
isDownRule[ fml_FML$, False] :=
	MemberQ[ getOptionalComponent[ fml, "tags"], "cong"]
isDownRule[ thm_THM$, False] :=
	MemberQ[ Lookup[ getThmData[ thm], "tags", {}], "cong"]


(* ::Subsection:: *)
(*Auxiliary Functions*)

adjustSubposition[ expr_, {sublist_, s_}, except:{exlist_, x_}, excepthist_] :=
	Replace[ Reap[ adjustSubposition[ expr, sublist, 0], True],
		{n_, loose0_} :>
			With[ {loose = Reverse[ Join @@ loose0], diff = n - Length[ sublist]},
				If[ Negative[ diff],
					{
						{Take[ sublist, n], Min[ s, n]},
						If[ TrueQ[ excepthist], {Drop[ exlist, diff], Max[ 1, x + diff]}, except],
						loose
					},
				(*else*)
					{{sublist, s}, except, loose}
				]
			]
	]
adjustSubposition[ _, {}, n_] :=
	n
adjustSubposition[ _VAR$|_FIX$|_META$, _, n_] :=
	n
adjustSubposition[ TYPED$[ expr_, _], {i_Integer, rest___}, n_] :=
	If[ i === 1,
		adjustSubposition[ expr, {rest}, n + 1],
	(*else*)
		n
	]
adjustSubposition[ Lambda$TM[ RNG$[ TYPRNG$[ _, t_]], expr_], {i_Integer, rest___}, n_] :=
	If[ i === 2,
		Sow[ t, True];
		adjustSubposition[ expr, {rest}, n + 1],
	(*else*)
		n
	]
adjustSubposition[ h_[ _], {0, rest___}, n_] :=
	adjustSubposition[ h, {rest}, n + 1]
adjustSubposition[ _[ a_], {1, rest___}, n_] :=
	adjustSubposition[ a, {rest}, n + 1]
adjustSubposition[ _, _, n_] :=
	n

(* 'adjustInstantiations[ inst1, inst2, expr, nfpos]' adjusts the given instantiations 'inst1' and 'inst2' by replacing
	'POS$'-terms by the corresponding subexpressions of 'expr' (possibly with shifted de Bruijn indices). *)
adjustInstantiations[ inst1_, inst2_, expr_, loose_] :=
	With[ {pospos = Position[ {inst1, inst2}, POS$[ _List, _Integer]]},
		(* Collect all 'POS$'-expressions in 'inst1' and 'inst2', and replace them by the respective subexpressions of
			'expr' (with de Bruijn indices adjusted accordingly). *)
	With[ {terms =
			Replace[ Extract[ {inst1, inst2}, pospos],
				POS$[ pos_List, shift_Integer] :>
					Replace[ validExtract[ expr, First[ pos], loose],
						{term_, loose0_} :> If[ shift === 0, term, mapFreeVars[ term, VAR$[ #3 + shift]&]]
					],
				{1}
			]},
		ReplacePart[ {inst1, inst2}, Thread[ pospos -> terms]]
	]]

isApplicable[ True, _, _, _, _] := True
isApplicable[ test_, G_, pos_, rhs_, inst_] :=
	With[ {b = Fold[ applyFunction, rhs, inst]},
	With[ {n = abstractionCount[ b, 0, LengthWhile[ Reverse[ pos], (# === 0)&]]},
	With[ {G0 = extractSingle[ G, Drop[ pos, -n]]},
		TrueQ[ test[ G0, normalize[ Fold[ applyFunction, b, getArgs[ G0, n]]]]]
	]]]

SetAttributes[ getInitialRWPos, HoldFirst];
getInitialRWPos[ nForms_, term_] :=
	If[ rewritableQ[ nForms, term],
		{{}, 0, {{}, 0}, initExcept[], True},
	(*else*)
		$Failed
	]

initExcept[] := {{{{}, {}}}, 1}

resizeExcept[ {exlist_, x_}, size_] :=
	{
		If[ size <= x,
			Take[ exlist, size],
		(*else*)
			With[ {len = Length[ exlist]},
				Which[
					len === size,
					exlist,
				(*else*)
					len < size,
					Join[ exlist, Table[ {{}, {}}, {size - len}]],
				(*else*)
					True,
					Take[ exlist, size]
				]
			]
		],
		size
	}

enlargeExcept[ {exlist_, x_}] :=
	{Append[ exlist, {{}, {}}], x + 1}

enlargeSubpos[ {sublist_, s_}, i_] :=
	{Append[ sublist, i], s + 1}


(* ::Subsection:: *)
(*Strategies*)

(* A microstrategy is a 4-tuple '{normaldown, excepthist, nextSucc, nextRWPos}', where
	- 'normaldown' is a boolean value specifying whether "normal" rewrite-rules (i.e. those neither tagged as "comp"
		nor as "cong") shall be used as down-rules or as up-rules.
	- 'excepthist' is a boolean value specifying whether the 'except'-argument of the various functions shall maintain
		the full history of exceptions starting from the root of the current part to the current subexpression, or
		whether it shall only contain 1 single element referring to the current subexpression.
	- 'nextSucc' is a function that determines how to proceed when a rewrite-rule has been applied successfully.
	- 'nextRWPos' is a function that determines how to proceed when a subexpression cannot be rewritten as a whole (any
		more).
	Both 'nextSucc' and 'nextRWPos' get 'nForms' as their first argument, so it is reasonable (though not mandatory) to
	equip them with 'HoldFirst'. *)
getMicroStrategy[ "lazy"] = {True, True, lazyNextSuccess, lazyNextRWPos};
getMicroStrategy[ "eager"] = {False, False, eagerNextSuccess, eagerNextRWPos};


(* ::Subsubsection:: *)
(*Lazy Microstrategy*)

(* Lazy microstrategy:
	When rewriting an explicit assumption or the atomic goal, call it 'e', descend into 'e' and do the following for
	each subexpression 'e0' visited:
	1. Apply all rewrite-rules except those tagged as "comp", yielding 'e1'.
	2. If 'e1' differs from 'e0', "step back" to the context of 'e1' (as far as determined by '$rewritingContextDepth')
		and carry out step 1 for each subexpression on the path leading to 'e1' (but no parallel paths).
	3. If the node of 'e0' is reached without rewriting an intermediate subexpression, try again all down-rules on
		'e1'. If, however, the rule applied to 'e0' yielding 'e1' was a congruence rule, do not try it any more until
		the next non-congruence rule is applied to the expression at that position.
	4. If 'e1' does not differ from 'e0', resume with the descent.
	5. If no descent is possible any more, try all rules tagged as "comp". If a rule is applicable, go back to step 2.
		Otherwise, move further out from the leaves of the expression to its root, trying to apply "comp"-rules at each
		level. *)

SetAttributes[ lazyNextSuccess, HoldFirst];
lazyNextSuccess[ _, down_, {sublist_, s_}, {exlist_, x_}, exnew_] :=
	With[ {sublist0 = Take[ sublist, s], exlist0 = Take[ exlist, x]},
	With[ {exlist1 = If[ exnew === {{}, {}}, exlist0, ReplacePart[ exlist0, -1 -> Join[ Last[ exlist0], exnew, 2]]]},
		Switch[ $rewritingContextDepth,
			0,
			{{sublist0, s}, {exlist1, x}, True},
			Infinity,
			{{sublist0, 0}, {exlist1, 1}, True},
			_,
			With[ {n =
					Module[ {c = $rewritingContextDepth + 1},
						LengthWhile[ Reverse[ sublist0], (If[ # =!= 0, --c]; c =!= 0) &]
					]},
				{{sublist0, s - n}, {exlist1, Max[ 1, x - n]}, True}
			]
		]
	]]

(* 'term' is always the whole part of the subgoal currently under consideration i.e. 'extractSingle[ G, mainpos]'. *)
SetAttributes[ lazyNextRWPos, HoldFirst];
lazyNextRWPos[ nForms_, True, term_, loose_, subpos:{sublist_, _}, except_] :=
	Replace[ extractSingle[ term, sublist],
		{
			_TYPED$|_VAR$|_FIX$|_META$ -> {loose, subpos, except, False},
			Lambda$TM[ RNG$[ TYPRNG$[ _, typ_]], _?(rewritableQ[ nForms])] :>
				{Prepend[ loose, typ], enlargeSubpos[ subpos, 2], enlargeExcept[ except], True},
			(_?(rewritableQ[ nForms]))[ _] :>
				{loose, enlargeSubpos[ subpos, 0], enlargeExcept[ except], True},
			_[ _?(rewritableQ[ nForms])] :>
				{loose, enlargeSubpos[ subpos, 1], enlargeExcept[ except], True},
			_ -> {loose, subpos, except, False}
		}
	]
lazyNextRWPos[ nForms_, False, term_, loose_, {{pre___, 0}, s_}, {exlist_, x_}] :=
	With[ {pp = {pre, 1}},
		Replace[ Extract[ term, pp],
			{
				_?(rewritableQ[ nForms]) :> {loose, {pp, s}, {ReplacePart[ exlist, -1 -> {{}, {}}], x}, True},
				_ :> {loose, {{pre}, s - 1}, {Most[ exlist], x - 1}, False}
			}
		]
	]
lazyNextRWPos[ _, False, _, loose_, {{pre___, 1}, s_}, {exlist_, x_}] :=
	{loose, {{pre}, s - 1}, {Most[ exlist], x - 1}, False}
lazyNextRWPos[ _, False, _, loose_, {{pre___, 2}, s_}, {exlist_, x_}] :=
	{Rest[ loose], {{pre}, s - 1}, {Most[ exlist], x - 1}, False}


(* ::Subsubsection:: *)
(*Eager Microstrategy*)

(* Eager microstrategy:
	When rewriting an explicit assumption or the atomic goal, call it 'e', descend into 'e' and do the following for
	each subexpression 'e0' visited:
	1. Apply all congruence rules to 'e0' repeatedly, until no more rule is applicable (but do not apply the same
		congruence rule twice, of course).
	2. If a congruence rule was applicable, do not descend further into 'e0', but go upward trying all other rules.
	3. If a rule is applicable, descent into the new expression again. *)

SetAttributes[ eagerNextSuccess, HoldFirst];
eagerNextSuccess[ _, down_, sub_, {{ex_}, _}, exnew_] :=
	If[ TrueQ[ down],
		{sub, {{Join[ ex, exnew, 2]}, 1}, True},
	(*else*)
		{sub, initExcept[], True}
	]

SetAttributes[ eagerNextRWPos, HoldFirst];
eagerNextRWPos[ nForms_, True, term_, loose_, subpos:{sublist_, _}, except:{{{{}, {}}}, _}] :=
	(* no congruence rule could be applied -> descend further into the term *)
	Replace[ extractSingle[ term, sublist],
		{
			_TYPED$|_VAR$|_FIX$|_META$ -> {loose, subpos, except, False},
			Lambda$TM[ RNG$[ TYPRNG$[ _, typ_]], _?(rewritableQ[ nForms])] :>
				{Prepend[ loose, typ], enlargeSubpos[ subpos, 2], except, True},
			(_?(rewritableQ[ nForms]))[ _] :>
				{loose, enlargeSubpos[ subpos, 0], except, True},
			_[ _?(rewritableQ[ nForms])] :>
				{loose, enlargeSubpos[ subpos, 1], except, True},
			_ -> {loose, subpos, except, False}
		}
	]
eagerNextRWPos[ _, True, term_, loose_, subpos_, except_] :=
	(* at least 1 congruence rule could be applied -> move upward *)
	{loose, subpos, except, False}
eagerNextRWPos[ nForms_, False, term_, loose_, {{pre___, 0}, s_}, except_] :=
	With[ {pp = {pre, 1}},
		Replace[ Extract[ term, pp],
			{
				_?(rewritableQ[ nForms]) :> {loose, {pp, s}, except, True},
				_ :> {loose, {{pre}, s - 1}, except, False}
			}
		]
	]
eagerNextRWPos[ _, False, _, loose_, {{pre___, 1}, s_}, except_] :=
	{loose, {{pre}, s - 1}, except, False}
eagerNextRWPos[ _, False, _, loose_, {{pre___, 2}, s_}, except_] :=
	{Rest[ loose], {{pre}, s - 1}, except, False}


(* ::Subsection:: *)
(*Generating Rewrite-Rules*)

generateRewriteRules[ f_, univ_, i_, n_, lift_, tfix_, tvar_] :=
	DeleteCases[
		Replace[ $equationalFormatConstructor[ f, i, n, univ, {}],
			{
				{} :> {{Null, Hold[], generateEqRewriteRule[ f, lift, tfix, tvar]}},
				out_List :> Cases[ out, {eq_, prf:(_Hold|_HoldComplete)} :> {eq, prf, generateEqRewriteRule[ eq, lift, tfix, tvar]}],
				_ -> {}
			}
		],
		{_, _, $Failed|_Integer}
	]

(* 'rwVariant[ args, factID, keyHandler, lbl]' is an auxiliary proof method used for equipping the given formulas with
	equational rewrite rules. The formulas themselves are left unchanged. *)
rwVariant[ args_List, factID_, keyHandler_, lbl_, add_, rem_][ ps_, __] :=
	Hold @@ {
		updateAbbrevLists[ {}, {}, {}, {factID}],
		variantIterator[
			getProofSitData[ ps, factID, {}],
			rwVariantFun,
			keyHandler,
			lbl,
			add,
			rem
		]
	}

rwVariantFun[ i_, PRFSIT$[ _, hyps_, _, _, _, tfix_, __]] :=
	With[ {hyp = hyps[[i]]},
		If[ !MemberQ[ hyp, "rwData" -> _],
			With[ {rwData = generateEqRewriteRule[ hyp, tfix]},
				If[ rwData === $Failed,
					Null,
				(*else*)
					Append[ DeleteCases[ List @@ Drop[ hyp, 3], "TypeVarPositions" -> _], "rwData" -> rwData]
				]
			],
		(*else*)
			Null
		]
	]


(* ::Subsection:: *)
(*Normal Forms*)

SetAttributes[ rewritableQ, HoldFirst];
rewritableQ[ nForms_][ term_] :=
	rewritableQ[ nForms, term]
rewritableQ[ nForms_, term_] :=
	!MatchQ[ term, _META$|_VAR$] &&
		FreeQ[ nForms, term /. RNG$[ TYPRNG$[ _, typ_]] :> typ]

SetAttributes[ addToNormalForms, HoldFirst];
addToNormalForms[ nForms_, _META$|_VAR$] :=
	False
addToNormalForms[ nForms_, term0_] :=
	With[ {term = term0 /. RNG$[ TYPRNG$[ _, typ_]] :> typ},	(* names of bound variables do not matter *)
		If[ FreeQ[ nForms, term],
			nForms = Append[ Select[ nForms, FreeQ[ term, #]&], term];
			True,
		(*else*)
			False
		]
	]


(* ::Subsection:: *)
(*NOMATCH*)

SetAttributes[ NOMATCH, HoldFirst];
NOMATCH[ _, TYPED$[ Theorema`Knowledge`NOMATCH$TM, {_, _}][ patt_][ term_], pos_, subpos_, {}, ctxt_] :=
	If[ patternMatchQ[ patt, term, ctxt],
		Skip,
	(*else*)
		{
			Hold @@ {
				INF$[ "formula", fastKeyFromLabel[ {Lookup[ $theoryMapping, "$Minimal", $Failed], "NOMATCH def"}]],
				INF$[ "replace", 1, Join[ Table[ 1, {pos + 1}], subpos, {0, 0}]],
				INF$[ "drop", 1]
			},
			Skip,
			Skip
		}
	]
NOMATCH[ _, _, _, _, _, _] := Continue

If[ !MemberQ[ $terminalProofSteps, NOMATCHTerminalStep], AppendTo[ $terminalProofSteps, NOMATCHTerminalStep]];

NOMATCHTerminalStep[ G_, hyps_, idx_, fix_, meta_, tfix_, tvar_] :=
	Replace[ stripAssms[ G, 0],
		{
			{TYPED$[ Theorema`Knowledge`NOMATCH$TM, {_, _}][ patt_][ term_], n_Integer} :>
				If[ patternMatchQ[ patt, term, {idx, fix, meta, tfix, tvar}],
					$Failed,
				(*else*)
					Function[ a,
						Hold[
							INF$[ "formula", fastKeyFromLabel[ {Lookup[ $theoryMapping, "$Minimal", $Failed], "NOMATCH def"}]],
							INF$[ "replace", 1, Join[ Table[ 1, {a + n + 1}], {0, 0}]],
							INF$[ "refl"]
						]
					]
				],
			_ -> $Failed
		}
	]


(* ::Subsection:: *)
(*Renaming Bound Variables*)

(* 'renameBoundVars[ pos, ren]' is an auxiliary proof method that renames all bound variables under the given position
	'pos' of the first subgoal of the current proof situation according to the given renaming substitution. 'pos' must
	be valid position (this is not checked!), and 'ren' must be a list of elements of the form 'x -> y', where 'x' and
	'y' are symbols. *)
renameBoundVars[ _List, {}][ __] :=
	Hold[]
renameBoundVars[ pos_List, ren0_List][ PRFSIT$[ {G_, R___}, ctxt__], ___, mode_] :=
	Module[ {change = False},
	With[ {ren = Cases[ ren0, (x_Symbol -> y_Symbol) /; x =!= y :> (RNG$[ TYPRNG$[ x, typ_]] :> (change = True; RNG$[ TYPRNG$[ y, typ]]))]},
		If[ ren === {},
			Hold[],
		(*else*)
			With[ {new = extractSingle[ G, pos] /. ren},
				If[ change,
					If[ mode === 2,
						Hold[ INF$[ "alphabetaeta", pos, new, False, False]],
					(*else*)
						{
							PRFSIT$[ {ReplacePart[ G, {pos} -> new], R}, ctxt],
							{},
							{},
							Hold[]
						}
					],
				(*else*)
					Hold[]
				]
			]
		]
	]]


(* ::Subsection:: *)
(*Direct Rewriting*)

Get[ FileNameJoin[ {$TheoremaDirectory, "Theorema", "CoreProvers", "CoreProofMethods", "Rewrite", "Direct.m"}]];


(* ::Subsection:: *)
(*Indirect Rewriting*)

Get[ FileNameJoin[ {$TheoremaDirectory, "Theorema", "CoreProvers", "CoreProofMethods", "Rewrite", "Indirect.m"}]];


(* ::Subsection:: *)
(*Combined Rewriting*)

(* 'rewriteAll[ spec]' is a proof method that combines direct and indirect rewriting: it directly rewrites all subgoals
	in the current proof situation with settings 'depth = 0', 'rewriters = {}', 'useAssms = False',
	'rewriteAssms = True' and 'introFix = False' (irrespective of 'spec'), and indirectly rewrites all chained facts. *)
rewriteAll[ spec_List][ ps_PRFSIT$, __] :=
	Replace[ parseRWSpec[ spec, ps, "rewriteAll"],
		{_List, hypsDown_List, hypsUp_List, globDown_Association, globUp_Association, variants_List,
				_, _, microstrat_, _, _, _} :>
			With[ {id = getFreshFactID[]},
				Hold @@ {
					addKnowledgeAux[ variants, addKeysToList[ id, False], Hold[], 3, None],
					rewriteAll[ id, {{hypsDown, hypsUp}, {globDown, globUp}, microstrat}]
				}
			]
	]
rewriteAll[ argsID_, {{hypsDown_, hypsUp_}, glob_, microstrat_}][ ps:PRFSIT$[ _, hyps_, idx_, fix_, _, tfix_, __], __] :=
	With[ {argKeys0 = getProofSitData[ ps, argsID, {}]},
	With[ {tempArgs = Cases[ argKeys0, {k:{_, _}, False} :> k], argKeys = Replace[ argKeys0, {k:{_, _}, _} :> k, {1}],
			fun = If[ TrueQ[ microstrat[[1]]], If[ MemberQ[ #, "comp"], 2, 1]&, If[ MemberQ[ #, "cong"], 1, 2]&]},
		Replace[
			Last[
				Reap[
					Scan[
						With[ {i = First[ FirstPosition[ hyps, FML$[ #, __], {$Failed}, {1}, Heads -> False]]},
							If[ IntegerQ[ i],
								With[ {h = hyps[[i]]},
									If[ MemberQ[ h, "rwData" -> _],
										Sow[ i, fun[ getOptionalComponent[ h, "tags"]]],
									(*else*)
										sowCPMMessage[ 159, {#}, "Warning"]
									]
								],
							(*else*)
								sowCPMMessage[ 158, {#}, "Warning"]
							]
						]&,
						argKeys
					],
					{1, 2}
				]
			],
			{newHypsDown_, newHypsUp_} :>
				With[ {hypIndices = {Join[ hypsDown, Join @@ newHypsDown], Join[ hypsUp, Join @@ newHypsUp]},
						chainedKeys = getProofSitData[ ps, "chainedFacts", {}]},
				Module[ {usedEnvKeys = {}},
				With[ {list =
						Map[
							Function[ ky0,
								With[ {ky = Replace[ ky0, {k:{_, _}, _} :> k]},
									Replace[ FirstCase[ hyps, f:FML$[ ky, __] :> extractFormula[ f, 0], $Failed],
										{
											{A0_STERM$, _} :>
												Replace[ indirectRewriteAll[ A0, hyps, idx, fix, {}, tfix, usedEnvKeys, {hypIndices, glob, microstrat}],
													{
														{newUsed_List, A_, newTVar_Association, data:Except[ {}]} :>
															(
																usedEnvKeys = newUsed;
																{ky0, STERM$[ A, Association[], newTVar], Replace[ data, {{j_}, rest___} :> {j, rest}, {1}]}
															),
														_ -> Sequence[]
													}
												],
											_ -> Sequence[]
										}
									]
								]
							],
							chainedKeys
						]},
				With[ {keys = list[[All, 1]], envKeysNum = Length[ usedEnvKeys]},
				With[ {keyNum = Length[ keys]},
				With[ {offset = envKeysNum + keyNum},
					Hold @@ Join[
						{updateAbbrevLists[ {}, {}, {}, {argsID}]},
						INF$[ "insert", #]& /@ usedEnvKeys,
						rewriteAllHyp /@ list,
						{
							rewriteAllGoal[
								keyNum,
								envKeysNum,
								usedEnvKeys,
								{
									If[ offset === 0, hypIndices, Replace[ hypIndices, j_Integer?Positive :> j + offset, {2}]],
									glob,
									microstrat
								}
							],
							rewriteAllPost[ keys, keyNum, envKeysNum, tempArgs]
						}
					]
				]]]]]]
		]
	]]

(* 'rewriteAllPost[ oldKeys, envKeysNum, tempKeys]' is an auxiliary proof method applied after rewriting. Its sole
	purpose is to "clean up the context" again, i.e. removing all hypotheses that are not needed any more and updating
	the list of chained facts. *)
rewriteAllPost[ oldKeys_, newNum_, envKeysNum_, tempKeys_][ ps:PRFSIT$[ _, hyps_, __], __] :=
	With[ {n = newNum + 1, tempChained = Cases[ oldKeys, {ky:{_, _}, False} :> ky]},
	With[ {repl = MapIndexed[ #1 -> {key[ hyps[[n - First[ #2]]]], False} &, oldKeys]},
		Hold @@ Join[
			If[ repl === {}, {}, {updateAbbrevLists[ {}, {}, {"chainedFacts" -> Replace[ getProofSitData[ ps, "chainedFacts", {}], repl, {1}]}, {}]}],
			Table[ INF$[ "forget", n], {envKeysNum}],
			If[ tempKeys === {} === tempChained, {}, {forgetHyps[ Join[ tempChained, tempKeys]]}]
		]
	]]

(* 'rewriteAllGoal[ params]' is an auxiliary proof method used for rewriting the goals. *)
SetAttributes[ rewriteAllGoal, HoldFirst];
rewriteAllGoal[ newNum_, envKeysNum_, usedEnvKeys_, {{hypIndicesDown_, hypIndicesUp_}, {globDown_, globUp_}, microstrat_}][ args__] :=
	With[ {n = newNum + envKeysNum + 1},
	Module[ {assmData = {}, iter, nForms = {},
				params =
					{
						{},					(* rewriters *)
						hypIndicesDown,		(* hypIndicesDown *)
						hypIndicesUp,		(* hypIndicesUp *)
						globDown,			(* globDown *)
						globUp,				(* globUp *)
						False,				(* useAssms *)
						True,				(* rewriteAssms *)
						microstrat,			(* microstrat *)
						False,				(* introFix *)
						0,					(* metaNum *)
						Association @@ MapIndexed[ #1 -> n - First[ #2] &, usedEnvKeys]	(* availableEnvKeys *)
					}},
			mapAllTac[ rewriteAllGoal[ {params, assmData, iter, nForms}], All][ args]
	]]
rewriteAllGoal[ {params_, assmData_, iter_, nForms_}][ ps:PRFSIT$[ _, _, _, _, meta_, __], args__] :=
	(
		params[[9]] = Length[ meta];
		directRewriteAll[ {params, assmData, iter, nForms}, 0][ ps, args]
	)

(* 'rewriteAllHyp[ data]' is an auxiliary proof method used for rewriting implicit hypotheses. *)
rewriteAllHyp[ {ky0_, snew:STERM$[ new_, _, _], data_}][ PRFSIT$[ G_, hyps_, idx_, fix_, meta_, tfix_, tvar_, fid_, ctxt___], ___, mode_] :=
	If[ mode === 2,
		With[ {ky = Replace[ ky0, {k:{_, _}, _} :> k]},
		With[ {i = First[ FirstPosition[ hyps, FML$[ ky, __], {$Failed}, {1}, Heads -> False]]},
			Hold[
				INF$[ "cut", snew],
				INF$[ "makeExplicit", i],
				indirectRewriteProof[ data]
			]
		]],
	(*else*)
		{
			PRFSIT$[
				G,
				Prepend[ hyps, FML$[ makeFormulaKey[ fid], new, "?", "TypeVarPositions" -> makeTypeVarPositions[ snew]]],
				idx, fix, meta, tfix, tvar, fid + 1, ctxt
			],
			{},
			{},
			Hold[]
		}
	]
