(* Direct Rewriting. *)

(* Macrostrategy of 'directRewriteAll':
	1. Rewrite the whole subgoal, including all explicit assumptions, using immutable down-rules. Prove conditions of
		rules without any explicit assumptions.
	2. Strip off the first explicit assumption. If assumptions shall be rewritten, rewrite it using immutable rules
		only and prove conditions without any explicit assumptions.
	3. If assumptions shall be used for rewriting, generate canonical forms and rewrite-rules from the first assumption
		and add them right after it.
	4. Proceed with the remaining part of the subgoal, but now also rewrite w.r.t. the rules originating from the first
		explicit assumption and prove conditions with the first assumptions and all of its canonical forms.
	5. Proceed in this fashion until the atomic goal is reached, and rewrite the atomic goal.
	Hence, the order of explicit assumptions is crucial!
*)

(* The value of 'params' is a 11-tuple
	'{rewriters, hypIndicesDown, hypIndicesUp, globDown, globUp, useAssms, rewriteAssms,
		microstrat, introFix, metaNum, availableEnvKeys}';
	this value usually does not change. If 'solver' is omitted, the main rewriting proof method is employed for proving
	conditions. *)
(* The value of 'assmData' is a list of elements of the form '{i, num}' or '{i, data}' or 'False' whose length is
	precisely the number of explicit assumptions stripped off so far. A pair of integers '{i, num}' at position 'j'
	(with 'i >= j') indicates that the 'j'-th assumption is not in normal form, and that its normal forms are the
	'i'-th up to 'i + num - 1'-th assumptions ('num' may be '0', indicating that there are no normal forms); a pair
	'{i, data}' at position 'j' (with 'j >= i') indicates that the 'j'-th assumption is a normal form of the 'i'-th
	one, and 'data' is the data generated by the rule-generator ('i' may be equal to 'j', indicating that the 'j'-th
	assumption is in normal form already); 'False' at position 'j' indicates that the 'j'-th assumption shall neither
	be used for rewriting nor be rewritten itself. *)
SetAttributes[ directRewriteAll, HoldFirst];
directRewriteAll[] := directRewriteAll[ {}]
directRewriteAll[ spec_List][ ps_PRFSIT$, __] :=
	Replace[ parseRWSpec[ spec, ps, "directRewriteAll"],
		{rewriters_List, hypIndicesDown_List, hypIndicesUp_List, globDown_Association, globUp_Association, variants_List,
				useAssms_, rewriteAssms_, microstrat_, depth_, all_, introFix_} :>
			Module[ {assmData, iter, nForms = {},
						params =
							{
								rewriters,			(* rewriters *)
								hypIndicesDown,		(* hypIndicesDown *)
								hypIndicesUp,		(* hypIndicesUp *)
								globDown,			(* globDown *)
								globUp,				(* globUp *)
								useAssms,			(* useAssms *)
								rewriteAssms,		(* rewriteAssms *)
								microstrat,			(* microstrat *)
								introFix,			(* introFix *)
								0,					(* metaNum *)
								Association[]		(* availableEnvKeys *)
							}},
				Join[
					With[ {expl = If[ TrueQ[ all], All, First]},
						Hold[ internalFactHandler[ voidTac[], expl, 0]]
					],
					If[ variants === {},
						If[ TrueQ[ all],
							Hold[ mapAllTac[ directRewriteAll[ {params, assmData, iter, nForms}, Null, depth], All]],
						(*else*)
							Hold[ directRewriteAll[ {params, assmData, iter, nForms}, Null, depth]]
						],
					(*else*)
						Hold[
							addKnowledgeImplicit[ variants, "methodFacts", False, 3, True, None],
							directRewriteAll[ {params, assmData, iter, nForms}, all, depth]
						]
					],
					Hold[ internalFactHandler[ "methodFacts", voidTac[], None, 2]]
				]
			]
	]
directRewriteAll[ {params_, assmData_, iter_, nForms_}, all:(True|False), depth_][ ps:PRFSIT$[ _, hyps_List, __], args__] :=
	With[ {keys = Replace[ getProofSitData[ ps, "methodFacts", {}], {ky:{_, _}, _} :> ky, {1}],
			fun = If[ TrueQ[ params[[8, 1]]], If[ MemberQ[ #, "comp"], 3, 2]&, If[ MemberQ[ #, "cong"], 2, 3]&],
			hypNum = Length[ hyps] + 1},
		params =
			ReplacePart[ params,
				Last[
					Reap[
						Scan[
							With[ {i = First[ FirstPosition[ hyps, FML$[ #, __], {$Failed}, {1}, Heads -> False]]},
								If[ IntegerQ[ i],
									With[ {h = hyps[[i]]},
										If[ MemberQ[ h, "rwData" -> _],
											Sow[ i - hypNum, fun[ getOptionalComponent[ h, "tags"]]],
										(*else*)
											sowCPMMessage[ 159, {#}, "Warning"]
										]
									],
								(*else*)
									sowCPMMessage[ 158, {#}, "Warning"]
								]
							]&,
							keys
						],
						2|3,
						#1 -> Join[ params[[#1]], #2] &
					]
				]
			];
		(* Note that we store indices of theorems and still keep "methodFacts";
			so, there is some redundancy, but this should not be a problem.
			Also note that "methodFacts" does not contain all facts passed to the proof method! *)
		If[ TrueQ[ all],
			mapAllTac[ directRewriteAll[ {params, assmData, iter, nForms}, Null, depth], All],
		(*else*)
			directRewriteAll[ {params, assmData, iter, nForms}, Null, depth]
		][ ps, args]
	]
(* The following is the entry point for each subgoal to attack. *)
directRewriteAll[ {params_, assmData_, iter_, nForms_}, Null, depth_][ PRFSIT$[ {g_, ___}, _, _, fix_, __], __] :=
	If[ TrueQ[ params[[9]]],
		With[ {cmd = Last[ introFixCommandList[ g, fix, True]]},	(* These constants might be generalized again at the end, so we choose "nice" names for them. *)
			Append[ cmd, directRewriteAll[ {params, assmData, iter, nForms}, Length[ fix], depth]]
		],
	(*else*)
		Hold[ directRewriteAll[ {params, assmData, iter, nForms}, 0, depth]]
	]
directRewriteAll[ {params_, assmData_, iter_, nForms_}, keep_Integer, depth_][ PRFSIT$[ G:{g_, ___}, hyps_, _, _, meta_, tfix_, tvar_, __], __] :=
	Replace[ $rewritingMaster[ nForms, g, Null, {}, hyps, tfix, tvar],
		Except[ _Hold] :>
			With[ {len = Length[ G] - 1},
				assmData = {};
				params[[10]] = Length[ meta];
				Hold[
					directRewriteAll[ {params, assmData, iter, nForms}, depth],
					directRewritePost[ assmData, len, keep]
				]
			]
	]
directRewriteAll[ {params_, assmData_, iter_, nForms_}, depth_][ PRFSIT$[ {G_, ___}, hyps_, idx_, fix_, meta_, tfix_, tvar_, __], __] :=
	Replace[ getInitialRWPos[ nForms, G],
		{
			{loose_, pos_, sub_, except_, down_} :>
				directRewriteAll0[
					{params, assmData, iter, nForms},
					down,
					Which[
						down,
						iter = -Length[ params[[1]]];
						0,	(* rewriters are only tried when descending into the expression *)
					(*else*)
						params[[8, 1]],
						iter = -Length[ params[[3]]];
						2,
					(*else*)
						True,
						iter = 1;
						1
					],
					{pos, sub, except, depth, G, hyps, loose, {idx, fix, meta, tfix, tvar}}
				],
			_ -> Hold[]
		}
	]

(* 'directRewriteAllAux[ ...]' is an auxiliary proof method. The version with 'loose' is applied after trying all rules
	on a certain subexpression without success. The other version, without 'loose', is applied after successfully
	applying a rule, in which case 'loose' and 'subpos' must be re-computed, because they might refer to non-existing
	parts of the new expression. *)
SetAttributes[ directRewriteAllAux, HoldFirst];
directRewriteAllAux[ {params_, assmData_, iter_, nForms_}, down_, cat_, loose_, pos_Integer, subpos_, except_, depth_][
		PRFSIT$[ {G_, ___}, hyps_, idx_, fix_, meta_, tfix_, tvar_, __], __] :=
	(* 'except' is a pair '{exlist, i}', where 'exlist' is a list of triples '{hypKeys, thms, envKeys}' specifying
		those rules that shall not be applied to the expression corresponding to the respective prefix of 'subpos' any
		more (because they are congruence rules that have been applied already). *)
	directRewriteAll0[
		{params, assmData, iter, nForms},
		down,
		cat,
		{If[ Negative[ pos], assmCount[ G, 0, Infinity] + pos + 1, pos], subpos, except, depth, G, hyps, loose, {idx, fix, meta, tfix, tvar}}
	]
directRewriteAllAux[ {params_, assmData_, iter_, nForms_}, down_, cat_, pos_Integer, subpos_, except_, depth_][
		PRFSIT$[ {G_, ___}, hyps_, idx_, fix_, meta_, tfix_, tvar_, __], __] :=
	directRewriteAll1[
		{params, assmData, iter, nForms},
		down,
		cat,
		{If[ Negative[ pos], assmCount[ G, 0, Infinity] + pos + 1, pos], subpos, except, depth, G, hyps, {idx, fix, meta, tfix, tvar}}
	]

(* 'directRewriteAll0', 'directRewriteAll1' and 'directRewriteAll2' are no proof methods, but merely auxiliary
	functions. *)
SetAttributes[ directRewriteAll0, HoldFirst];
directRewriteAll0[ {params_, assmData_, iter_, nForms_}, down_, cat_, {pos_, subpos:{sublist_, s_}, except_, depth_, G_, hyps_, loose_, ctxt_}] :=
	With[ {mainPos = Table[ 1, {pos}]},
	With[ {expr = extractSingle[ G, Join[ mainPos, Take[ sublist, s]]]},
		directRewriteAll2[
			{params, assmData, iter, nForms},
			down,
			cat,
			{
				pos,
				subpos,
				except,
				depth,
				mainPos,
				G,
				expr,
				hyps,
				loose,
				ctxt
			}
		]
	]]

SetAttributes[ directRewriteAll1, HoldFirst];
directRewriteAll1[ {params_, assmData_, iter_, nForms_}, down_, cat_, {pos_, subpos_, except_, depth_, G_, hyps_, ctxt_}] :=
	With[ {mainPos = Table[ 1, {pos}]},
	With[ {expr0 = extractSingle[ G, mainPos]},
	With[ {aux = adjustSubposition[ expr0, subpos, except, params[[8, 2]]]},
	With[ {expr = extractSingle[ expr0, Take @@ First[ aux]]},
		directRewriteAll2[
			{params, assmData, iter, nForms},
			down,
			cat,
			{
				pos,
				First[ aux],
				aux[[2]],
				depth,
				mainPos,
				G,
				expr,
				hyps,
				Last[ aux],
				ctxt
			}
		]
	]]]]

SetAttributes[ directRewriteAll2, HoldFirst];
directRewriteAll2[ {params_, assmData_, iter_, nForms_}, down_, cat_,
		data:{pos_, subpos:{sublist_, s_}, except:{exlist_, x_}, depth_, mainpos_, G_, expr_, hyps_, loose_, ctxt_}] :=
	With[ {jj = If[ TrueQ[ down], 0, 1], excur = exlist[[x]], subcur = Take[ sublist, s],
			rec = directRewriteAllNextSuccess[ {params, assmData, iter, nForms}, pos, depth, down, subpos, except, #]&,
			next = evalTac[ directRewriteAllAux[ {params, assmData, iter, nForms}, down, #1, loose, pos, subpos, except, depth], iter = #2]&},
	Switch[ cat,
		0,	(* rewriters *)
		Catch[
			Scan[
				Replace[ ++iter; #[ nForms, expr, pos, sublist, loose, ctxt],
					{
						{pre_Hold, post_Hold, inst_List, succ_, fail_} :>
							Throw[
								Hold @@ {
									INF$[ "if",
										andTac @@ Join[
											pre,
											Hold[ directRewriteSingle[ {params, assmData, nForms}, 0, pos, subcur, inst, depth]],
											post
										],
										directRewriteAllNextRewriter[ {params, assmData, iter, nForms}, succ, True, loose, pos, subpos, except, depth],
										directRewriteAllNextRewriter[ {params, assmData, iter, nForms}, fail, iter, loose, pos, subpos, except, depth]
									]
								}
							],
						{cmd:Hold[ __], succ_, fail_} :>
							Throw[
								Hold @@ {
									INF$[ "if",
										andTac @@ cmd,
										directRewriteAllNextRewriter[ {params, assmData, iter, nForms}, succ, True, loose, pos, subpos, except, depth],
										directRewriteAllNextRewriter[ {params, assmData, iter, nForms}, fail, iter, loose, pos, subpos, except, depth]
									]
								}
							],
						Skip :> Throw[ directRewriteAll2[ {params, assmData, iter, nForms}, False, Next, data]],
						Next :> Throw[ directRewriteAll2[ {params, assmData, iter, nForms}, True, Next, data]]
					}
				]&,
				Take[ First[ params], iter]
			];
			If[ TrueQ[ params[[6]] && params[[8, 1]]],
				iter = 1;
				directRewriteAll2[ {params, assmData, iter, nForms}, down, 1, data],
			(*else*)
				iter = -Length[ params[[2 + jj]]];
				directRewriteAll2[ {params, assmData, iter, nForms}, down, 2, data]
			]
		],
	(*else*)
		1,	(* explicit assumptions *)
		Catch[
			Scan[
				Replace[ ++iter; #,
					{_Integer, ruleData:Except[ _Integer|$Failed]} :>
						Replace[ applyEqRewriteRule[ expr, Extract[ G, assmPosition[ iter]], ruleData, loose, ctxt],
							{{}, inst20_List, {_Association, tinst_List}, _List, _, test_} :>	(* there must not be universally quantified variables *)
								Replace[ adjustInstantiations[ {}, inst20, extractSingle[ G, Join[ mainpos, subcur]], loose],
									{{}, inst2_} :>
										Throw[
											Hold @@ {
												INF$[ "if",
													(* here we do not rename bound variables; if this changes, 'directRewriteSingle' must be adjusted *)
													directRewriteSingle[ {params, assmData, nForms}, tinst, {}, {}, iter, pos, subcur, inst2, test, depth],
													rec[ {{}, {}}],
													next[ 1, iter + 1]	(* 'iter' refers to the current rule *)
												]
											}
										]
								]
						]
				]&,
				Drop[ assmData, --iter]
			];
			iter = -Length[ params[[2 + jj]]];
			directRewriteAll2[ {params, assmData, iter, nForms}, down, 2, data]
		],
	(*else*)
		2,	(* hypotheses *)
		Catch[
			Scan[
				If[ ++iter; MemberQ[ First[ excur], #],
					Null,
				(*else*)
					With[ {fml = hyps[[#]]},
						Replace[ applyEqRewriteRule[ expr, fml, loose, ctxt],
							{inst10_List, inst20_List, {_Association, tinst_List}, bren_List, _, test_} :>
								Replace[ adjustInstantiations[ inst10, inst20, extractSingle[ G, Join[ mainpos, subcur]], loose],
									{inst1_, inst2_} :>
										Throw[
											Hold @@ {
												INF$[ "if",
													andTac[
														INF$[ "makeExplicit", #],
														directRewriteSingle[ {params, assmData, nForms}, tinst, inst1, bren, 0, pos, subcur, inst2, test, depth]
													],
													rec[ {If[ MemberQ[ getOptionalComponent[ fml, "tags"], "cong"], {#}, {}], {}}],
													next[ 2, iter]	(* 'iter' already refers to the next rule *)
												]
											}
										]
								]
						]
					]
				]&,
				Take[ params[[2 + jj]], iter]
			];
			iter = getRuleConsts[ expr, Head, False];	(* we need the atomic head, even if it is 'TrueQ' *)
			directRewriteAll2[ {params, assmData, iter, nForms}, down, 3, data]
		],
	(*else*)
		3,	(* formula environment and theorem instances *)
		Catch[
			With[ {idx = First[ ctxt]},
				While[
					(* 'params[[4 + jj]]' is an association, not a 'RULEC$'-structure! *)
					With[ {k = incrementRuleIterator[ iter, params[[4 + jj]]]},
						Which[
							k === $Failed,
							False,
						(*else*)
							MemberQ[ Last[ excur], k],
							True,
						(*else*)
							ListQ[ k],
							With[ {fml = getFormulas[ True][ k]},
								Replace[ applyEqRewriteRule[ expr, fml, loose, ctxt],
									{inst10_List, inst20_List, {_Association, tinst_List}, bren_List, _, test_} :>
										Replace[ adjustInstantiations[ inst10, inst20, extractSingle[ G, Join[ mainpos, subcur]], loose],
											{inst1_, inst2_} :>
												Throw[
													Hold @@ {
														INF$[ "if",
															Join[
																With[ {j = Lookup[ params[[11]], k, $Failed]},
																	If[ IntegerQ[ j],
																		andTac[ INF$[ "makeExplicit", j]],
																	(*else*)
																		andTac[ INF$[ "formula", k]]
																	]
																],
																andTac[ directRewriteSingle[ {params, assmData, nForms}, tinst, inst1, bren, 0, pos, subcur, inst2, test, depth]]
															],
															rec[ {{}, If[ MemberQ[ getOptionalComponent[ fml, "tags"], "cong"], {k}, {}]}],
															next[ 3, iter]
														]
													}
												]
										]
								]
							];
							True,
						(*else*)
							True,
							Replace[ getThmInst[ True, idx][ k],
								{thm_, _} :>
									Replace[ applyEqRewriteRule[ expr, thm, Lookup[ getThmData[ k], "rwData", $Failed], loose, ctxt],
										{inst10_List, inst20_List, {_Association, tinst_List}, bren_List, _, test_} :>
											Replace[ adjustInstantiations[ inst10, inst20, extractSingle[ G, Join[ mainpos, subcur]], loose],
												{inst1_, inst2_} :>
													Throw[
														Hold @@ {
															INF$[ "if",
																andTac[
																	INF$[ "theorem", k],
																	directRewriteSingle[ {params, assmData, nForms}, tinst, inst1, bren, 0, pos, subcur, inst2, test, depth]
																],
																rec[ {{}, If[ MemberQ[ Lookup[ getThmData[ k], "tags", {}], "cong"], {k}, {}]}],
																next[ 3, iter]
															]
														}
													]
											]
									]
							];
							True
						]
					]
				]
			];
			directRewriteAllNextFailure[
				{params, assmData, iter, nForms},
				down, expr, pos, mainpos, subpos, except, depth, G, hyps, loose, ctxt,
				$rewritingAddNF
			]
		],
	(*else*)
		_,
		directRewriteAllNextFailure[ {params, assmData, iter, nForms}, down, expr, pos, mainpos, subpos, except, depth, G, hyps, loose, ctxt, False]
	]]

(* 'directRewriteAllRec[ params, i, pos, depth]' is an auxiliary proof method that is applied when proving conditions
	of rewrite-rules recursively. 'i' is the position of the assumption that was used for rewriting (could be '0'). *)
SetAttributes[ directRewriteAllRec, HoldFirst];
directRewriteAllRec[ {params_, assmData_, nForms_}, i_, pos_, depth_][ PRFSIT$[ {G_, R___}, hyps_, _, _, meta_, tfix_, tvar_, __], __] :=
	Replace[ $rewritingMaster[ nForms, G, Null, Drop[ meta, params[[10]]], hyps, tfix, tvar],
		Except[ _Hold] :>
			Module[ {newData = assmData, iter = -Length[ First[ params]]},
			With[ {numR = Length[ {R}]},
				If[ i =!= 0,
					newData =
						Replace[
							Delete[ newData, i],
							{{j_Integer, num:Except[ 0, _Integer]} /; j <= i < j + num :> {j, num - 1}, {j_, x_} /; j > i :> {j - 1, x}},
							{1}
						]
				];
				Hold @@ {
					directRewriteAllRec[ {params, newData, iter, nForms}, pos - If[ i === 0, 0, 1], depth - 1],
					directRewritePost[ assmData, numR, -1],
					closeSubgoals[ numR, params[[10]], True]
				}
			]]
	]
directRewriteAllRec[ params_, pos_, depth_][ PRFSIT$[ {G_, ___}, hyps_, idx_, fix_, meta_, tfix_, tvar_, __], __] :=
	directRewriteAll0[
		params,
		True,
		0,
		{pos, {{}, 0}, initExcept[], depth, G, hyps, {}, {idx, fix, meta, tfix, tvar}}
	]

(* 'directRewriteAllNextRewriter' is an auxiliary function called after a rewriter has been applied (successfully or
	not) to a term.
	'succ' is either 'True' or an integer; in the latter case, it is precisely the index of the rewriter to be tried
	next. *)
SetAttributes[ directRewriteAllNextRewriter, HoldFirst];
directRewriteAllNextRewriter[ {params_, assmData_, iter_, nForms_}, spec_, succ_, loose_, pos_, subpos_, except_, depth_] :=
	Switch[ spec,
		Continue,
		If[ TrueQ[ succ],
			directRewriteAllNextSuccess[ {params, assmData, iter, nForms}, pos, depth, True, subpos, except, {{}, {}}],
		(*else*)
			iter = succ;
			directRewriteAllAux[ {params, assmData, iter, nForms}, True, 0, loose, pos, subpos, except, depth]
		],
		Skip,
		directRewriteAllAux[ {params, assmData, iter, nForms}, False, Next, pos, subpos, except, depth],
		_,
		directRewriteAllAux[ {params, assmData, iter, nForms}, True, Next, pos, subpos, except, depth]
	]

(* 'directRewriteAllNextSuccess[ ...]' is an auxiliary proof method that is applied after an expression has been
	rewritten by a rewrite-rule. It basically only resets 'iter', 'cat', 'down', 'subpos' and 'except', and then
	applies the main method 'directRewriteAll' again. *)
SetAttributes[ directRewriteAllNextSuccess, HoldFirst];
directRewriteAllNextSuccess[ {params_, assmData_, iter_, nForms_}, pos_, depth_, down_, subpos_, rest___][ args__] :=
	Replace[ params[[8, 3]][ nForms, down, subpos, rest],
		{
			{newSub_, newExcept_, newDown_} :>
				With[ {newCat =
						Which[
							newDown,
							iter = -Length[ params[[1]]];
							0,	(* rewriters are only tried when descending into the expression *)
						(*else*)
							params[[8, 1]],
							iter = -Length[ params[[3]]];
							2,
						(*else*)
							True,
							iter = 1;
							1
						]},
					directRewriteAllAux[ {params, assmData, iter, nForms}, newDown, newCat, pos, newSub, newExcept, depth]
				][ args],
			_ -> $Failed
		}
	]

(* 'directRewriteAllNextFailure' is an auxiliary proof method that is applied after all rules have been tried on a
	subterm without success. *)
SetAttributes[ directRewriteAllNextFailure, HoldFirst];
directRewriteAllNextFailure[ {params_, assmData_, iter_, nForms_}, down_, expr_, pos_, mainpos_, subpos:{sublist_, s_}, except_, depth_, G_, hyps_, loose_,
		ctxt:{_, fix_, meta_, tfix_, tvar_}, addnf_] :=
	With[ {microstrat = params[[8]]},
		Replace[ getNextDirectRWPos[ nForms, down, G, loose, mainpos, subpos, except, params[[6]], params[[7]], microstrat[[2]], microstrat[[4]]],
			{
				{newLoose_, newSub_, newExcept_, newDown_} :>
					With[ {newCat =
							Which[
								newDown,
								iter = -Length[ params[[1]]];
								0,	(* rewriters are only tried when descending into the expression *)
							(*else*)
								First[ microstrat],
								iter = -Length[ params[[3]]];
								2,
							(*else*)
								True,
								iter = 1;
								1
							]},
						If[ TrueQ[ down] || addnf =!= True || !addToNormalForms[ nForms, expr],
							directRewriteAll0[ {params, assmData, iter, nForms}, newDown, newCat, {pos, newSub, newExcept, depth, G, hyps, newLoose, ctxt}],
						(*else*)
							Hold[ directRewriteAllAux[ {params, assmData, iter, nForms}, newDown, newCat, newLoose, pos, newSub, newExcept, depth]]
						]
					],
				Next :>
					With[ {cont = $rewritingMaster[ nForms, G, pos, Drop[ meta, params[[10]]], hyps, tfix, tvar], pp = pos + 1},
						If[ MatchQ[ cont, _Hold],
							cont,
						(*else*)
							iter = -Length[ First[ params]];
							If[ TrueQ[ params[[6]]],
								With[ {rem = -(assmCount[ dropAssms[ G, pp], 0, Infinity] + 1)},
									Hold[
										assmToRWRules[ assmData, pp],
										directRewriteAllAux[
											{params, assmData, iter, nForms},
											True,
											0,
											{},
											rem,
											{{}, 0},
											initExcept[],
											depth
										]
									]
								],
							(*else*)
								directRewriteAll0[
									{params, assmData, iter, nForms},
									True,
									0,
									{pp, {{}, 0}, initExcept[], depth, G, hyps, {}, ctxt}
								]
							]
						]
					],
				_ :>
					With[ {cont = $rewritingMaster[ nForms, G, {pos}, Drop[ meta, params[[10]]], hyps, tfix, tvar]},
						If[ MatchQ[ cont, _Hold],
							cont,
						(*else*)
							Hold[]
						]
					]
			}
		]
	]
	
SetAttributes[ directRewriteSingle, HoldFirst];
directRewriteSingle[ params_, tinst_, inst1_, bren0_, i0_, pos_, subcur_, inst2_, test_, depth_][ args__] :=
	With[ {meth = directRewriteSingle[ params, i0, pos, subcur, inst2, test, depth],
			bren = DeleteCases[ bren0, x_ -> x_]},
		If[ tinst === {} === inst1 === bren,
			meth[ args],
		(*else*)
			Hold @@ Join[
				If[ tinst === {}, {}, {INF$[ "instT", tinst]}],
				If[ bren === {}, {}, {renameBoundVars[ {0, 1}, bren]}],		(* must happen *before* instantiating variables *)
				If[ inst1 === {}, {}, {instantiateCoreAll[ inst1, 1]}],
				{meth}
			]
		]
	]
directRewriteSingle[ params_, i0_, pos_, subcur_, inst0_, test_, depth_][ PRFSIT$[ {G_, R___}, hyps_, idx_, fix_, meta_, tfix_, tvar_, __], __] :=
	With[ {p0 = pos + If[ i0 === 0, 1, 0]},
	With[ {realpos = Join[ Table[ 1, {p0}], subcur],
			i = If[ i0 === 0, 1, i0]},
		(* 'i0' being '0' means that the first assumption shall be used for rewriting, and that all
			other indices temporarily have to be increased by '1'. *)
	With[ {assm = Extract[ G, assmPosition[ i]]},
		Replace[ stripAssms[ assm, 0],
			{
				{eq:(TYPED$[ CoreEqual$TM, {_}][ a_][ b_]), ac_Integer} :>
					Replace[ replaceAutomaticByMeta[ inst0, a, meta, Length[ fix]],
						{inst_, newMeta_} :>
							Which[
								!isApplicable[ test, G, realpos, b, inst],
								$Failed,
							(*else*)
								ac === 0,
								Hold @@ Join[
									newMeta,
									If[ inst === {}, {}, {INF$[ "alphabetaeta", realpos, Fold[ #1[ #2]&, a, inst], False, True]}],
									{INF$[ "replace", i, Join[ realpos, Table[ 0, {Length[ inst]}]]]},
									If[ i0 === 0, {INF$[ "drop", 1]}, {}]
								],
							(*else*)
								Positive[ depth],
								With[ {solver =
										If[ $rewritingSolver === Automatic,
											{fastFix[ ac], reverseSubgoals[ ac], mapAllTac[ directRewriteAllRec[ params, i0, pos, depth], ac]},
										(*else*)
											{$rewritingSolver[ depth - 1, ac]}
										]},
									Hold @@ Join[
										newMeta,
										{setupSubgoalsMP[ eq, p0], INF$[ "impE", i, 0]},
										solver,
										{directRewriteSingle[ params, Length[ {R}], i0, pos, subcur, inst]}
									]
								],
							(*else*)
								True,
								With[ {subprf = tryTerminalProofSteps[ Extract[ assm, Table[ assmPosition[ j], {j, 1, ac}]], hyps, idx, fix, meta, tfix, tvar],
										num = Length[ {R}]},
									If[ ListQ[ subprf],
										Join[
											Hold @@ newMeta,
											Hold[ setupSubgoalsMP[ eq, p0], INF$[ "impE", i, 0]],
											Join @@ (#[ p0]& /@ subprf),
											Hold[ directRewriteSingle[ params, num, i0, pos, subcur, inst]]
										],
									(*else*)
										$Failed
									]
								]
							]
					],
				_ -> $Failed
			}
		]
	]]]
directRewriteSingle[ _, num_, i0_, pos_, subcur_, inst_][ PRFSIT$[ {G_, R___}, __], __] :=
	(* This version is applied after proving conditions. *)
	If[ Length[ {R}] === num,
		With[ {i = pos + If[ i0 === 0, 1, 0] + 1},	(* The unconditional equality is the 'i'-th assumption. *)
		With[ {realpos = Join[ Table[ 1, {i}], subcur],
				rule = Extract[ G, assmPosition[ i]]},
			Replace[ stripAssms[ rule, 0],
				{
					{TYPED$[ CoreEqual$TM, {_}][ a_][ _], 0} :>
						Hold @@ Join[
							If[ inst === {}, {}, {INF$[ "alphabetaeta", realpos, Fold[ #1[ #2]&, a, inst], False, True]}],
							{INF$[ "replace", i, Join[ realpos, Table[ 0, {Length[ inst]}]]]},
							{INF$[ "drop", i]},
							If[ i0 === 0, {INF$[ "drop", 1]}, {}]
						],
					_ -> $Failed
				}
			]
		]],
	(*else*)
		$Failed
	]

(* 'assmToRWRules[ assmData, pos]' is an auxiliary proof method that generates rewrite-rules from the 'pos'-th explicit
	assumption in the first subgoal of the current proof situation, and updates 'assmData' accordingly. *)
SetAttributes[ assmToRWRules, HoldFirst];
assmToRWRules[ assmData_, pos_Integer][ ps:PRFSIT$[ {G_, ___}, _, _, _, _, tfix_, tvar_, __], args__] :=
	Replace[ generateRewriteRules[ Extract[ G, assmPosition[ pos]], False, pos, 0, False, tfix, tvar],
		{
			{{Null, _, data:Except[ _Integer]}} :> (AppendTo[ assmData, {pos, data}]; Hold[]),
			rules:{{Except[ Null], _, _}..} :>
				With[ {len = Length[ rules]},
					Hold @@ {
						INF$[ "if",	(* either all equational forms can be obtained, or none *)
							solveTac[
								SameQ,
								andTac @@ Join @@ Replace[
									Reverse[ rules],
									{
										{new_, prf_Hold, _} :> Prepend[ prf, setupSubgoalsMP[ new, pos]],
										{new_, HoldComplete[ prf_], _} :> Prepend[ prf, setupSubgoalsMP[ new, pos]]
									},
									{1}
								]
							],
							With[ {newData = Prepend[ Map[ {pos, #[[3]]}&, rules], {pos + 1, len}]},
								voidTac[ assmData = Join[ assmData, newData]]
							],
							voidTac[ AppendTo[ assmData, {pos, 0}]]
						]
					}
				],
			_ :> (AppendTo[ assmData, {pos, 0}]; Hold[])
		}
	]

(* 'directRewritePost[ assmData, num]' is an auxiliary proof method. Its sole purpose is removing all auxiliary
	assumptions that have been generated during rewriting, and then applying '$rewritingPost'. *)
SetAttributes[ directRewritePost, HoldFirst];
directRewritePost[ assmData_, num_Integer, keep0_Integer][ PRFSIT$[ G_List, _, _, fix_, __], __] :=
	With[ {len = Length[ G] - 1},
		Which[
			num === len,
			Join[
				If[ ListQ[ assmData],
					Module[ {c = -1},
						Hold @@ MapIndexed[
							With[ {i = First[ #2]},
								Replace[ #1,
									{
										{Except[ i], Except[ _Integer]} :> INF$[ "drop", i - (++c)],
										_ -> Sequence[]
									}
								]
							]&,
							assmData
						]
					],
				(*else*)
					Hold[]
				],
				With[ {fixNum = Length[ fix]},
				With[ {keep = If[ Negative[ keep0], fixNum - keep0 - 1, keep0]},
					If[ keep < fixNum && TrueQ[ $generalizeFix],
						With[ {goalNum = Length[ G], gfix = Drop[ fix, keep][[All, 1]]},
							Append[ mapAllInline[ Hold @@ {generalizeFix[ gfix]}, goalNum - num, goalNum], INF$[ "forgetFix", gfix]]
						],
					(*else*)
						Hold[]
					]
				]],
				$rewritingPost
			],
		(*else*)
			len > num,
			$Failed,
		(*else*)
			True,
			Hold[]
		]
	]


(* ::Subsubsection:: *)
(*Default Master*)

(* 'defaultRewritingMaster' solves goals of the form 't === X', where 'X' is a new meta variable, by reflexivity.
	It waits until 't' has been fully rewritten. *)
SetAttributes[ defaultRewritingMaster, HoldFirst];
defaultRewritingMaster[ nForms_, G_, {n_Integer}, newmeta_, _, tfix_, tvar_] :=
	Replace[ dropAssms[ G, n],
		{
			TYPED$[ CoreEqual$TM, {_}][ a_][ b_] :>
				Which[
					MatchQ[ a, _META$] && MemberQ[ newmeta, a -> _],
					defaultRewritingMaster[ nForms, b, a],
				(*else*)
					MatchQ[ b, _META$] && MemberQ[ newmeta, b -> _],
					defaultRewritingMaster[ nForms, a, b],
				(*else*)
					True,
					Replace[ alphaEquiv[ a, b, tfix, tvar],		(* there is no reason for postponing this to 'closeSubgoals' *)
						{
							{_Association, _List} :>
								Hold[ INF$[ "refl"]],
							_ -> Continue
						}
					]
				],
			_ -> Continue
		}
	]
defaultRewritingMaster[ _, _, _, _, _, _, _] :=
	Continue
defaultRewritingMaster[ nForms_, t_, X_] :=
	If[ t === X,
		Hold[ INF$[ "refl"]],
	(*else*)
		addToNormalForms[ nForms, t];
		Hold[ INF$[ "inst", X -> STERM$[ t, Association[], Association[]]], INF$[ "refl"]]
	]

	
(* ::Subsubsection:: *)
(*Auxiliary Functions*)

SetAttributes[ getNextDirectRWPos, HoldFirst];
getNextDirectRWPos[ nForms_, True, expr_, {}, mainpos_, {sublist_, 0}, except:{exlist_, _}, useAssms_, rewriteAssms_, excepthist_, stratNextRWPos_] :=
	If[ sublist === {} || MatchQ[ sublist, {0, 1, ___}],
		Replace[ extractSingle[ expr, mainpos],
			{
				TYPED$[ CoreImplies$TM, {}][ A_][ B_] :>
					Which[
						TrueQ[ rewriteAssms] && rewritableQ[ nForms, A],
						If[ TrueQ[ useAssms], nForms = {}];
						{{}, {{0, 1}, 2}, If[ TrueQ[ excepthist], resizeExcept[ except, 2], except], True},
					(*else*)
						rewritableQ[ nForms, B],
						If[ TrueQ[ useAssms], nForms = {}];
						Next,
					(*else*)
						True,
						$Failed
					],
				term_ :>
					If[ sublist === {},
						stratNextRWPos[ nForms, True, term, {}, {sublist, 0}, except],
					(*else*)
						{{}, {sublist, 1}, {exlist, 2}, True}
					]
			}
		],
	(*else*)
		{{}, {sublist, 1}, {exlist, 2}, True}
	]
getNextDirectRWPos[ nForms_, True, expr_, loose_, mainpos_, subpos:{sublist_, s_}, except:{exlist_, x_}, _, _, _, stratNextRWPos_] :=
	Which[
		s === Length[ sublist] || x === Length[ exlist],
		stratNextRWPos[ nForms, True, extractSingle[ expr, mainpos], loose, subpos, except],
	(*else*)
		sublist[[s + 1]] === 2,
		{Prepend[ loose, extractSingle[ expr, Join[ mainpos, Take[ sublist, s], {1, 1, 2}]]], {sublist, s + 1}, {exlist, x + 1}, True},
	(*else*)
		True,
		{loose, {sublist, s + 1}, {exlist, x + 1}, True}
	]
getNextDirectRWPos[ nForms_, False, expr_, {}, mainpos_, subpos:{{}, _}, _, useAssms_, _, _, _] :=
	Replace[ extractSingle[ expr, mainpos],
		{
			TYPED$[ CoreImplies$TM, {}][ _][ _?(rewritableQ[ nForms])] :> (If[ TrueQ[ useAssms], nForms = {}]; Next),
			_ -> $Failed
		}
	]
getNextDirectRWPos[ nForms_, False, expr_, loose_, mainpos_, subpos:{{0, 1}, _}, except:{exlist_, _}, _, _, _, stratNextRWPos_] :=
	Replace[ extractSingle[ expr, mainpos],
		{
			TYPED$[ CoreImplies$TM, {}][ _][ _] :> {loose, {{}, 0}, {Take[ exlist, 1], 1}, False},
			term_ :> stratNextRWPos[ nForms, False, term, loose, subpos, except]
		}
	]
getNextDirectRWPos[ nForms_, False, expr_, loose_, mainpos_, subpos_, except_, _, _, _, stratNextRWPos_] :=
	stratNextRWPos[ nForms, False, extractSingle[ expr, mainpos], loose, subpos, except]

replaceAutomaticByMeta[ inst_List, term_, meta_List, fixNum_Integer] :=
	replaceAutomaticByMeta[ Position[ inst, Automatic, {1}, Heads -> False], inst, term, meta, fixNum, {}]
replaceAutomaticByMeta[ {}, inst_, _, _, _, newMeta_] :=
	{inst, newMeta}
replaceAutomaticByMeta[ {{i_Integer}, rest___}, inst_, term_, meta_, fixNum_, newMeta_] :=
	Replace[ Quiet[ Check[ Extract[ term, Join[ Table[ 2, {i - 1}], {1, 1}]], $Failed]],
		{
			TYPRNG$[ var_Symbol, typ_] :>
				With[ {name = If[ var === Null, Theorema`Knowledge`VAR$X$TM, var]},
				With[ {m =
						META$[ var,
							1 + Max[ -1, Max @@ Cases[ meta, (META$[ name, k_Integer] -> _) :> k], Max @@ Cases[ newMeta, INF$[ _, META$[ name, k_Integer], __] :> k]]
						]},
					replaceAutomaticByMeta[ {rest}, ReplacePart[ inst, i -> m], term, meta, fixNum, Append[ newMeta, INF$[ "meta", m, typ, fixNum]]]
				]],
			_ -> $Failed
		}
	]
