(* Core proof methods handling universal quantifiers. *)

(* ::Subsection:: *)
(*Instantiation*)

(* 'instantiateCoreAll[ inst, pos][ ps, ___]' instantiates the Core-universally-quantified explicit assumption at
	the given position 'pos' by 'inst'. 'inst' is a supposed to be a list whose elements are
	 - 'Automatic' or "$", in which case a the quantifier is instantiated by a fresh meta variable, or
	 - "_", in which case the outermost quantifier is skipped and the instantiation resumes with the next
	 	quantifier, or
	 - an arbitrary term, possibly inside 'EXPR$'; if it is inside 'EXPR$', types are inferred, otherwise it
	 	is simply type-checked.
	 The last element of 'inst' may be one of the above possibilities wrapped inside 'RepeatedNull', in which case
	 all remaining quantifiers are instantiated in the respective way.
	 If at least one quantifier is skipped, the proof method sets up a subgoal using "Modus Ponens". *)
instantiateCoreAll[ inst_List] :=
	instantiateCoreAll[ inst, 1]
instantiateCoreAll[ {("_"...)|(h_[ "_"] /; h === RepeatedNull)}, _] :=
	voidTac[]
instantiateCoreAll[ inst_List, i0_Integer][ ps:PRFSIT$[ {G_, ___}, __], rest___] :=
	With[ {ac = assmCount[ G, 0, Infinity]},
	With[ {i = If[ Negative[ i0], ac + 1 + i0, i0]},
		If[ 1 <= i <= ac,
			instantiateCoreAll[ inst, {i}][ ps, rest],
		(*else*)
			sowCPMMessage[ 71, {i0}, "Fatal"];
			$Failed
		]
	]]
instantiateCoreAll[ inst_List, {i_Integer}][ ps:PRFSIT$[ {G_, ___}, __], ___, 1] :=
	Replace[ executeProofCommands[ Hold[ instantiateCoreAll[ inst, {i}]], {ps}, 0, False],
		{
			PRFSTAT$[ _, {sit:PRFSIT$[ {G0_, ___}, __]}, assms_List, thms_List, ___] :>
				{
					With[ {pos = assmPosition[ i]},
						makeANDNODE[
							makePRFINFO[ name -> instantiateCoreAll, used -> {{Extract[ G, pos]}}, generated -> {{Extract[ G0, pos]}}],
							sit
						]
					],
					assms,
					thms,
					Hold[]
				},
			_ -> $Failed
		}
	]
instantiateCoreAll[ inst_List, {i_Integer}][ ps:PRFSIT$[ {G_, ___}, __], ___, mode_] :=
	Replace[ makeInstantiation[ inst, Extract[ G, assmPosition[ i]], CoreAll$TM, False, True, ps, mode],
		{inst1_List, skipPos_Integer, newIdx_Integer, newMeta_List, newTFix_Association, newTVar_Association, abbrevs_List, tsubst_List} :>
			Hold @@ Join[
				If[ tsubst === {}, {}, {INF$[ "instT", tsubst]}],
				Replace[ newMeta, (m_ -> {t_, d_}) :> INF$[ "meta", m, t, d], {1}],
				If[ abbrevs === {}, {}, {updateAbbrevLists[ {}, abbrevs, {}, {}]}],
				If[ mode === 2,
					Join[
						{INF$[ "sorts", newTVar]},
						INF$[ "allE", i, STERM$[ #, newTFix, Association[]]]& /@ If[ Negative[ skipPos], inst1, Take[ inst1, skipPos]],
						If[ Negative[ skipPos], {}, {instantiateCoreAllPost[ Drop[ inst1, skipPos], i]}]
					],
				(*else*)
					{instantiateCoreAllPost[ inst1, i, newIdx, newTFix]}
				]
			]
	]

(* 'instantiateCoreAllPost' is an auxiliary proof method that is applied if a subgoal has to be set up.
	It does not type-check 'inst', as this is assumed to have happened already in 'instantiateCoreAll'. 'inst', in
	fact, must be a list of well-typed terms or elements of the form '{Skip, typ_}'.
	In kernel mode, this proof method directly sets up a new subgoal; hence, the first element of 'inst' should
	reasonably be '{Skip, _}' (but it does not need to). *)
instantiateCoreAllPost[ inst_List, i_Integer, newIdx_, newTFix_][ PRFSIT$[ {G_, R___}, hyps_List, _, fix_, meta_, _, ctxt__], ___] :=
	(* this definition is used in fast mode *)
	With[ {p = assmPosition[ i]},
	With[ {new = ($prems = 0; makeInstance[ inst, Extract[ G, p]])},
		{
			PRFSIT$[ {ReplacePart[ G, p -> new], R}, hyps, newIdx, fix, meta, newTFix, ctxt],
			{},
			{},
			Hold[]
		}
	]]
instantiateCoreAllPost[ inst_List, i_Integer][ PRFSIT$[ {G_, ___}, __], ___] :=
	(* this definition is used in kernel mode *)
	With[ {new = ($prems = 0; makeInstance[ inst, Extract[ G, assmPosition[ i]]])},
	Module[ {name},
		Hold @@ Join[
			{setupSubgoalsMP[ new, i]},	(* 'new' becomes 'i + 1'-th explicit assumption *)
			Table[ INF$[ "allI", FIX$[ name, j]], {j, 1, Count[ inst, {Skip, _}]}],
			Module[ {j = 0}, Map[ INF$[ "allE", i, makeSTERM[ If[ MatchQ[ #, {Skip, _}], FIX$[ name, ++j], #]]]&, inst]],
			{
				INF$[ "impE", i, $prems],	(* here one could 'testIdentityTac[ ...]' *)
				INF$[ "drop", i]			(* 'new' is now 'i'-th explicit assumption *)
			}
		]
	]]

(* 'makeInstance[ inst, fml]' constructs the instance of 'fml' w.r.t. the given instantiation 'inst'. It returns the
	instantiated formula and sets the auxiliary global variable '$prems' to the number of premises in the instance,
	including those that appear in the scope of skipped quantifiers and those that appear after the last instantiated
	quantifier. *)
makeInstance[ inst_, Lambda$TM[ rng_RNG$, A_]] :=
	Lambda$TM[ rng, makeInstance[ inst, A]]
makeInstance[ {}, fml_] :=
	($prems += assmCount[ fml, 0, Infinity]; fml)
makeInstance[ {{Skip, _}, r__}, TYPED$[ CoreAll$TM, {typ_}][ A_]] :=
	TYPED$[ CoreAll$TM, {typ}][ makeInstance[ {r}, A]]
makeInstance[ {f_, r___}, TYPED$[ CoreAll$TM, {typ_}][ A_]] :=
	makeInstance[ {r}, normalize[ applyFunction[ A, f], $autoNormalize && MatchQ[ f, _Lambda$TM]]]
makeInstance[ inst_, fml:(TYPED$[ CoreImplies$TM, {}][ _][ _])] :=
	With[ {ac = assmCount[ fml, 0, Infinity]},
	With[ {pos = Table[ 1, {ac}]},
		$prems += ac;
		ReplacePart[ fml, pos -> makeInstance[ inst, Extract[ fml, pos]]]
	]]

(* 'instantiateCoreAllMult[ inst, n]' instantiates multiple explicit assumptions, falling back to 'instantiateCoreAll'.
	'n' can be an integer, in which case the first 'n' assumptions are instantiated (or the last ones, if 'n' is
	negative), or a list of integers. *)
instantiateCoreAllMult[ _, 0|{}] :=
	voidTac[]
instantiateCoreAllMult[ inst_, 1] :=
	instantiateCoreAll[ inst]
instantiateCoreAllMult[ inst_, n0_Integer][ ps:PRFSIT$[ {G_, ___}, __], ___] :=
	With[ {num = assmCount[ G, 0, Infinity]},
	With[ {n = If[ Negative[ n0], num + 1 + n0, n0]},
		If[ n <= num,
			Hold @@ Table[ instantiateCoreAll[ inst, i], {i, 1, n}],
			(* Of course, it would be much more efficient to infer the types of 'inst' only once. However, type
				inference takes the types of the respective quantifiers into account, and these types clearly could
				differ between the assumptions to be instantiated. *)
		(*else*)
			sowCPMMessage[ 74, {n0}, "Fatal"];
			$Failed
		]
	]]
instantiateCoreAllMult[ inst_, p_List][ ___] :=
	Hold @@ Map[ instantiateCoreAll[ inst, #]&, p]

(* 'instantiateCoreAllGoal' is an auxiliary proof method that simply instantiates the outermost universal quantifier
	in an explicit assumption by the current goal (i.e. atomic goal and some explicit assumptions). This method has its
	canonical application when instantiating formulas of the form '/\thesis. (...) ==> thesis', where 'thesis' can have
	type 'FORM' or object truth type. In the latter case, the 4-ary version of the proof method must be used, which
	automatically atomizes and reflects back the atomic goal, if necessary. *)
instantiateCoreAllGoal[] := instantiateCoreAllGoal[ 1, 0]
instantiateCoreAllGoal[ i_Integer?Positive] := instantiateCoreAllGoal[ i, 0]
instantiateCoreAllGoal[ i_Integer?Positive, n_Integer?NonNegative][ PRFSIT$[ {G_, ___}, __], ___] :=
	Hold @@ {INF$[ "allE", i, makeSTERM[ dropAssms[ G, n + i]]]}
instantiateCoreAllGoal[ i_Integer?Positive, n_Integer?NonNegative, cmd_Hold, reflectBack_][ ps:PRFSIT$[ {G_, ___}, __], args___] :=
	If[ getObjectTruthType[] === FORM$TM,
		Join[ instantiateCoreAllGoal[ i, n][ ps, args], cmd],
	(*else*)
		Replace[ stripAssms[ G, 0],
			{
				{TYPED$[ TrueQ$TM, {}][ _], m_Integer} :>
					Join[ instantiateCoreAllGoal[ i, m + 1 - i][ ps, args], cmd],
				{term_, m_Integer} :>
					Replace[ $atomizationFunction[ term, {m}, Table[ 1, {m}], {}, G],
						{
							cmd1_Hold :>
								Join[
									cmd1,
									Hold[ Evaluate[ instantiateCoreAllGoal[ i, m + 1 - i]]],
									cmd,
									If[ TrueQ[ reflectBack], Hold[ tryTac[ reflectTerm[ 0, {}]]], Hold[]]
								],
							_ :> (sowCPMMessage[ 79, {}, "Fatal"]; $Failed)
						}
					]
			}
		]
	]


(* ::Subsection:: *)
(*of-variant*)

(* 'ofVariant[ args, factID, keyHandler, lbl, add, rem]' is an auxiliary proof method used for instantiating
	universally quantified hypotheses.
	'ofVariant' is the rough high-level analogue of 'instantiateCoreAll' above. There are important differences,
	though:
	- The terms of the instantiation must not contain any meta variables nor polymorphic a.b.f. term constants.
	- The types of the terms are inferred only once, without taking into account the types of the variables that are
		instantiated. Moreover, the types are inferred in fresh type-variable contexts, meaning that the type variables
		in the terms are treated as being distinct (keep in mind that there must not be polymorphic a.b.f. constants!) *)
ofVariant[ {args0___, "_"...}, factID_, keyHandler_, lbl_, add_, rem_][ ps_PRFSIT$, __] :=
	With[ {factKeys = getProofSitData[ ps, factID, {}], funabbrevs = termListToFunction[ {args0}, ps, "of"]},
		If[ ListQ[ funabbrevs],
			Hold @@ {
				updateAbbrevLists[ {}, Last[ funabbrevs], {}, {factID}],
				variantIterator[ factKeys, ofVariantFun[ First[ funabbrevs]], keyHandler, lbl, add, rem]
			},
		(*else*)
			$Failed
		]
	]

ofVariantFun[ terms_][ i_, PRFSIT$[ _, hyps_, _, fix_, _, tfix_Association, __]] :=
	If[ terms === Null,
		Null,
	(*else*)
		With[ {hyp = hyps[[i]]},
			Replace[ extractFormula[ hyp, 0],
				{STERM$[ fml_, Association[], tvar_Association], idx_} :>
					Replace[ makeInstanceof[ terms[ idx], fml, tfix, tvar],	(* 'terms' is a function that gives an 'STERM$' when applied to a type variable index *)
						{new_STERM$, pd_} :> {{new, ofVariantProof[ i, pd], preserveFormulaTags[ new, hyp, tfix]}}
					]
			]
		]
	]

ofVariantProof[ i_Integer, {ac_Integer, inst_List, tsubst_List}][ n_Integer] :=
	Module[ {X, counter = -1},
		Hold @@ Join[
			Table[ INF$[ "allI", FIX$[ X, j]], {j, 0, Count[ inst, "_"] - 1}],
			{INF$[ "makeExplicit", i + n]},
			(* we can safely use "allE" here, since the instances 'f' neither contain new a.b.f. constants nor new meta variables *)
			Replace[ inst, {"_" :> INF$[ "allE", 1, makeSTERM[ FIX$[ X, ++counter]]], TYPED$[ f_, _] :> INF$[ "allE", 1, makeSTERM[ replaceTypes[ f, tsubst]]]}, {1}],
			{INF$[ "impE", 1, ac], INF$[ "close"]}
		]
	]

makeInstanceof[ STERM$[ inst_, _, tvar2_], A_, tfix_, tvar1_] :=
	With[ {tvar = intersectSorts[ tvar1, tvar2]},	(* Actually, the keys should be disjoint, but just to be sure ... *)
		Replace[ Reap[ Catch[ makeInstanceof[ inst, A, tfix, tvar, {}, 0]], 0],
			{
				{new_, {{{tsubst_List, ac_Integer}}}} :>
					{STERM$[ replaceTypes[ new, tsubst], Association[], tvar], {ac, inst, tsubst}},
				_ -> $Failed
			}
		]
	]
makeInstanceof[ inst_, Lambda$TM[ rng_RNG$, A_], tfix_, tvar_, tsubst_, n_] :=
	Lambda$TM[ rng, makeInstanceof[ inst, A, tfix, tvar, tsubst, n]]
makeInstanceof[ {}, fml_, _, tvar_, tsubst_, n_] :=
	(Sow[ {tsubst, n + assmCount[ fml, 0, Infinity]}, 0]; fml)
makeInstanceof[ {"_", r___}, TYPED$[ CoreAll$TM, {typ_}][ A_], tfix_, tvar_, tsubst_, n_] :=
	TYPED$[ CoreAll$TM, {typ}][ makeInstanceof[ {r}, A, tfix, tvar, tsubst, n]]
makeInstanceof[ {TYPED$[ f_, t_], r___}, TYPED$[ CoreAll$TM, {typ0_}][ A_], tfix_, tvar_, tsubst_, n_] :=
	With[ {typ = typ0 /. tsubst},	(* no need to apply 'tsubst' to 't', since types of instantiations are completely independent of each other *)
		Replace[ unifyTypes[ t, typ, hasSort[ tfix], {tvar, {}}, 0],
			{
				{newTVar_Association, tinst_List} :>
					makeInstanceof[ {r}, normalize[ applyFunction[ A, f], $autoNormalize && MatchQ[ f, _Lambda$TM]], tfix, newTVar, Join[ tsubst /. tinst, tinst], n],
				_ :> (sowCPMMessage[ 75, {t, typ}, "Fatal", "of"]; Throw[ $Failed])
			}
		]
	]
makeInstanceof[ inst_, fml:(TYPED$[ CoreImplies$TM, {}][ _][ _]), tfix_, tvar_, tsubst_, n_] :=
	With[ {ac = assmCount[ fml, 0, Infinity]},
	With[ {pos = Table[ 1, {ac}]},
		ReplacePart[ fml, pos -> makeInstanceof[ inst, Extract[ fml, pos], tfix, tvar, tsubst, n + ac]]
	]]
makeInstanceof[ _, _, _, _, _, _] :=
	(sowCPMMessage[ 73, "Fatal", "of"]; Throw[ $Failed])

(* 'termListToFunction' takes a list of pre-terms and a proof situation, and returns either '$Failed' or a triple
	'{fun, idx, abbrevs}', where 'fun' is a function that, when applied to a type variable index, returns the
	'STERM$'-structure 'STERM$[ terms, tfix, tvar]', where 'terms' is the list of type-annotated terms corresponding to
	the input list. *)
termListToFunction[ terms0_List, ps:PRFSIT$[ _, _, _, fix0_, _, tfix_, __], src_String] :=
	Catch[
		With[ {abbrevs = getTermAbbrevList[ ps], fix = Select[ fix0, FreeQ[ Last[ #], _TVAR$]&]},
		Module[ {newAbbrevs = {}, idx = 0, tvar = Association[]},
		With[ {terms =
				Replace[ terms0,
					f0:Except[ "_"] :>
						Replace[ makeCoreExprPatt[ f0, Blank[], idx, fix, {}, tfix, tvar, abbrevs, "gen", normalize, Check],
							{
								{{t_, STERM$[ f_, Association[], newTVar_], newIdx_, {}, {}}, tinst_, {}, abb_} :>
									(
										idx = newIdx;
										newAbbrevs = Join[ abb, newAbbrevs];
										tvar = newTVar;
										TYPED$[ f, t]
									),
								{{_, _, _, Except[ {}], _, _}, _, _, _} :>
									(sowCPMMessage[ 1003, "Fatal", src]; Throw[ $Failed]),
								{{_, _, _, _, Except[ {}], _}, _, _, _} :>
									(sowCPMMessage[ 1004, "Fatal", src]; Throw[ $Failed]),
								_ :> (sowCPMMessage[ 1005, "Fatal", src]; Throw[ $Failed])
							}
						],
					{1}
				]},
		With[ {fun =
				If[ terms === {},
					Null,
				(*else*)
					With[ {tvarList = Normal[ tvar]},
					With[ {pos = getPositions[ {terms, Null, tvarList}, _TVAR$][[All, 1]]},
					With[ {fun0 =
							ReplacePart[ STERM$[ terms, Association[], tvarList],
								MapIndexed[ Function[ {p, i}, p -> TVAR$[ First[ i] + Slot[ 1] - 1]], pos]
							]},
						MapAt[ Association, Function[ fun0], {1, 3}]
					]]]
				]},
			{fun, idx, DeleteDuplicatesBy[ newAbbrevs, First]}
		]]]]
	]


(* ::Subsection:: *)
(*Generalization*)

(* 'generalizeTerms[ terms, n][ ps, ___]' generalizes all of the terms in 'terms' appearing after the first 'n'
	explicit assumptions of the first subgoal of 'ps' by replacing them by new universally quantified variables.
	Note that new variables are only introduced for terms that actually appear. *)
generalizeTerms[ terms_List] :=
	generalizeTerms[ terms, 0]
generalizeTerms[ {}, _] :=
	voidTac[]
generalizeTerms[ terms0:Except[ {}, _List], n_Integer?NonNegative][ ps:PRFSIT$[ {G_, ___}, _, idx_, fix_, meta_, tfix_, tvar_, ___], args___, mode_] :=
	If[ assmCount[ G, 0, n] === n,
		Catch[
			With[ {abbrevs = getTermAbbrevList[ ps]},
			Module[ {newAbbrevs = {}, newTVar = tvar},
			With[ {terms =
					Module[ {idx0 = idx},
						Map[
							Replace[ makeCoreExprPatt[ #, Blank[], idx0, fix, meta, tfix, newTVar, abbrevs, "gen", normalize, None],	(* MODIFICATION: "gen" was 'None' before. *)
								{
									{{typ_, STERM$[ term_, Association[], tv_Association], i_Integer, {}, {}}, {}, {}, abb_} :>	(* type- and meta instantiation must be empty; this could be weakened *)
										(
											idx0 = i;
											newAbbrevs = Join[ abb, newAbbrevs];
											newTVar = tv;
											{term, typ}
										),
									_ :> (sowCPMMessage[ 76, "Fatal"]; Throw[ $Failed])
								}
							]&,
							terms0
						]
					]},
				newAbbrevs = DeleteDuplicatesBy[ newAbbrevs, First];
				Replace[ findPositions[ dropAssms[ G, n], terms[[All, 1]], tfix, newTVar, idx],
					{
						{_, {}} :>
							(
								If[ newAbbrevs === {},
									Hold[],
								(*else*)
									Hold[ updateAbbrevLists[ {}, newAbbrevs, {}, {}]]
								]
							),
						{{_Association, tsubst_List}, res_List} :>
							With[ {spec = Replace[ SortBy[ res, First], (i_Integer -> pos_List) :> {pos, terms[[i, 2]] /. tsubst}, {1}]},
								If[ mode === 1,
									Replace[ generalizeAt[ spec, n, newAbbrevs][ ps, args, 0],
										{p:PRFSIT$[ {fml_, ___}, __], assms_, thms_, cmd_} :>
											{
												makeANDNODE[ makePRFINFO[ name -> generalizeTerms, used -> {}, generated -> {{fml}}, "terms" -> terms[[All, 1]]], p],
												assms,
												thms,
												cmd
											}
									],
								(*else*)
									generalizeAt[ spec, n, newAbbrevs][ ps, args, mode]
								]
							]
					}
				]
			]]]
		],
	(*else*)
		sowCPMMessage[ 1006, {n}, "Fatal", "generalizeTerms"];
		$Failed
	]

(* 'generalizeFix[ consts, n, newMeta][ ps, ___]' generalizes all of the given a.b.f. term constants in 'consts'
	appearing after the first 'n' explicit assumptions of the first subgoal of 'ps', by replacing them by new
	universally quantified variables.
	'n' may be a non-negative integer or 'Automatic'; in the latter case, the variables are bound after as many
	explicit assumptions as possible.
	'newMeta' is a boolean value specifying whether meta variables depending on constants to be generalized shall be
	replaced by new meta variables that make this dependence explicit (i.e. are explicitly applied to the respective
	constants) before the generalization happens. The meta variables are then replaced everywhere in 'ps', not only in
	the first subgoal, of course. For that reason, the new meta variables do not depend on the constants implicitly as
	well any more, since otherwise applying 'generalizeFix' a second time to a different subgoal would again add the
	same constants as arguments ...
	Applying this proof method with 'newMeta' set to 'False' has the same effect as applying 'generalizeTerms' when all
	terms are a.b.f. constants and no new term abbreviations are given, but this proof method is more efficient. *)
generalizeFix[ consts_List] :=
	generalizeFix[ consts, 0, True]
generalizeFix[ consts_List, n_] :=
	generalizeFix[ consts, n, True]
generalizeFix[ {}, _, _] :=
	voidTac[]
generalizeFix[ consts:Except[ {}, _List], n_, True][ ps:PRFSIT$[ {G_, ___}, _, _, fix_, meta_, __], args__] :=
	Catch[
		With[ {indices = Sort[ Map[ Quiet[ Check[ Position[ fix, # -> _, {1}, 1, Heads -> False][[1, 1]], Throw[ $Failed]]]&, consts]]},
		With[ {list =
				GatherBy[
					Cases[ meta,
						(m_?(!FreeQ[ G, #] &) -> {t_, d:Except[ 0]}) :> With[ {c = LengthWhile[ indices, (# <= d &)]}, (m -> {t, c}) /; c =!= 0]
					],
					Composition[ First, First]
				],
				min = First[ indices] - 1},
			If[ list === {},
				generalizeFix[ consts, n, False][ ps, args],
			(*else*)
				Hold @@ Append[
					Flatten[
						Module[ {k},
							Map[
								Function[ class,
								With[ {name = class[[1, 1, 1]]},
									k = Max[ -1, Max @@ Cases[ meta, (META$[ name, i_] -> _) :> i]];
									Replace[ class,
										(m_ -> {t_, c_}) :>
											With[ {f = fix[[Take[ indices, c]]]},
												{
													INF$[ "meta", META$[ name, ++k], makeFunctionType[ Reverse[ f[[All, 2]]], t], min],
													INF$[ "inst", m -> STERM$[ applyCurried[ META$[ name, k], f[[All, 1]], {}], Association[], Association[]]]
												}
											],
										{1}
									]
								]],
								list
							]
						]
					],
					generalizeFix[ consts, n, False]
				]
			]
		]]
	]
generalizeFix[ consts:{__FIX$}, Automatic, False][ ps:PRFSIT$[ {G_, ___}, _, _, fix_, __], args___, mode_] :=
	Replace[ findAtomPositions[ G, consts],
		{
			{} :> Hold[],
			res_List :>
				Module[ {min = Infinity},
				With[ {spec0 =
						Replace[ SortBy[ res, First],
							(i_Integer -> pos_List) :>
								(
									min = Min[ min, Min @@ Map[ Function[ p, LengthWhile[ First[ p], (# === 1)&]], pos]];
									{pos, Replace[ consts[[i]], fix]}
								),
							{1}
						]},
					min = assmCount[ G, 0, min];
					With[ {spec = If[ min === 0, spec0, Replace[ spec0, {pos_, t_} :> {Map[ (Drop[ First[ #], min] -> Last[ #])&, pos], t}, {1}]]},
						If[ mode === 1,
							Replace[ generalizeAt[ spec, min, {}][ ps, args, 0],
								{p:PRFSIT$[ {fml_, ___}, __], assms_, thms_, cmd_} :>
									{
										makeANDNODE[ makePRFINFO[ name -> generalizeFix, used -> {}, generated -> {{fml}}, "consts" -> consts], p],
										assms,
										thms,
										cmd
									}
							],
						(*else*)
							generalizeAt[ spec, min, {}][ ps, args, mode]
						]
					]
				]]
		}
	]
generalizeFix[ consts:{__FIX$}, n_Integer?NonNegative, False][ ps:PRFSIT$[ {G_, ___}, _, _, fix_, __], args___, mode_] :=
	If[ assmCount[ G, 0, n] === n,
		Replace[ findAtomPositions[ dropAssms[ G, n], consts],
			{
				{} :> Hold[],
				res_List :>
					With[ {spec = Replace[ SortBy[ res, First], (i_Integer -> pos_List) :> {pos, Replace[ consts[[i]], fix]}, {1}]},
						If[ mode === 1,
							Replace[ generalizeAt[ spec, n, {}][ ps, args, 0],
								{p:PRFSIT$[ {fml_, ___}, __], assms_, thms_, cmd_} :>
									{
										makeANDNODE[ makePRFINFO[ name -> generalizeFix, used -> {}, generated -> {{fml}}, "consts" -> consts], p],
										assms,
										thms,
										cmd
									}
							],
						(*else*)
							generalizeAt[ spec, n, {}][ ps, args, mode]
						]
					]
			}
		],
	(*else*)
		sowCPMMessage[ 1006, {n}, "Fatal", "generalizeFix"];
		$Failed
	]

(* 'generalizeAt[ spec, n, abbrevs][ ps, ___]' is an auxiliary proof method for generalizing the terms at the given
	positions, by first abstracting over them and then introducing universal quantifiers. It also adds new term
	abbreviations 'abbrevs'.
	'spec' is supposed to be a list of pairs of the form '{pos, typ}', where 'p' is itself a list of rules of the form
	'p -> n', where 'p' is a position and 'n' is the number of binders in whose scope 'p' appears; 'typ' is the type of
	the subexpressions at the positions 'p' (which must all have the same type, since they must be alpha-equivalent).
	'n' is a non-negative integer specifying where to introduce the new universal quantifiers, i.e. after how many
	explicit assumptions. ATTENTION! The positions in 'pos' refer to the goal with its first 'n' assumptions dropped,
	i.e. they do not have prefix 'Table[ 1, {n}]'!
	In order for the proof method to function correctly, the given positions must all be valid and disjoint, and the
	subexpressions at the given positions must be alpha-equivalent for each group of positions '{pos, typ}'. These
	requirements are not checked but, for the sake of efficiency, simply assumed; if they are violated, the proof method
	may fail, result in an unexpected proof state, or even result in an invalid proof state in fast- and trace mode. *)
generalizeAt[ spec_] :=
	generalizeAt[ spec, 0, {}]
generalizeAt[ spec_, n_] :=
	generalizeAt[ spec, n, {}]
generalizeAt[ {}, _, {}] :=
	voidTac[]
generalizeAt[ {}, _, abbrevs_List] :=
	updateAbbrevLists[ {}, abbrevs, {}, {}]
generalizeAt[ spec_List, n_Integer, abbrevs_List][ ps:PRFSIT$[ {G_, R___}, hyps_, idx_, fix_, meta_, tfix_, tvar_, fid_, ___], ___, mode_] :=
	With[ {rev = Reverse[ spec], pos0 = Table[ 1, {n}]},
	With[ {G1 = If[ n === 0, G, Extract[ G, pos0]]},
	With[ {fml0 =
			Reap[
				ReplacePart[
					G1,
					Join @@ MapIndexed[
						With[ {j = First[ #2] - 1, pos = First[ #1]},
							If[ pos === {},
								Sow[ TYPED$[ Theorema`Knowledge`TYPE$TM, {Last[ #1]}], "inst"],
							(*else*)
								Sow[ First[ Extract[ G1, {pos[[1, 1]]}]], "inst"]
							];
							Replace[ pos, {({} -> i_Integer) :> ({{}} -> VAR$[ i + j]), (p_List -> i_Integer) :> (p -> VAR$[ i + j])}, {1}]
						]&,
						rev
					]
				],
				"inst"
			]},
	With[ {inst = Join @@ Last[ fml0]},
	With[ {fml =
			Fold[
				TYPED$[ CoreAll$TM, {First[ #2]}][ Lambda$TM[ RNG$[ TYPRNG$[ Last[ #2], First[ #2]]], #1]]&,
				First[ fml0],
				MapThread[ {Last[ #1], Replace[ #2, {FIX$[ name_, ___] :> name, _ -> Null}]}&, {rev, inst}]
			]},
		If[ mode === 2,
			Hold @@ Join[
				{
					setupSubgoalsMP[ fml, n],
					INF$[ "rearrangeG", 2, 1]
				},
				Reverse[ Map[ INF$[ "allE", n + 1, makeSTERM[ #]]&, inst]],			(* all elements of 'inst' type-check *)
				{INF$[ "impE", n + 1, assmCount[ G1, 0, Infinity]]},		(* here one could 'testIdentityTac[ ...]' *)
				If[ abbrevs === {}, {}, {updateAbbrevLists[ {}, abbrevs, {}, {}]}]
			],
		(*else*)
			{
				Join[ PRFSIT$[ {ReplacePart[ G, {pos0} -> fml], R}, hyps, idx, fix, meta, tfix, tvar, fid], PRFSIT$ @@ updateAbbrevLists[ {}, abbrevs, {}, {}, ps]],
				{},
				{},
				Hold[]
			}
		]
	]]]]]

(* 'forgetConsts[ n][ ps, __]' removes all a.b.f. constants specified by 'n' from 'ps', employing basic inference
	"forgetFix". In contrast to "forgetFix" the constants to be removed may not only be specified by a list, but also
	by an integer 'n': if 'n' is non-negative, all but the first 'n' constants are removed, otherwise the last '-n'
	constants are removed. *)
forgetConsts[ c_List] :=
	INF$[ "forgetFix", c]
forgetConsts[ n_Integer][ PRFSIT$[ _, _, _, fix_, __], __] :=
	With[ {l =
			If[ Negative[ n],
				Take[ fix, n],
			(*else*)
				Drop[ fix, n]
			][[All, 1]]},
		If[ l === {},
			Hold[],
		(*else*)
			Hold[ INF$[ "forgetFix", l]]
		]
	]


(* ::Subsection:: *)
(*Introduction*)

(* 'introduceFix[ consts][ ps, ___]' introduces the given a.b.f. constants for the Core-universally-quantified goal in
	'ps'. 'consts' is a supposed to be a list whose elements are
	- 'Automatic' or "$", in which case a fresh a.b.f. constant is introduced, or
	- 'RepeatedNull' or "...", in which case fresh a.b.f. constants are introduced for all remaining quantifiers, or
	- "_", in which case a fresh a.b.f. constant is introduced, which however is immediately generalized again using
		'generalizeTerms' after all constants have been introduced,
	- a.b.f. constants, possibly inside 'EXPR$'.
	Note that if too many constants are provided, the proof method simply ignores the remaining ones and issues a
	warning, but it does not fail! *)
introduceFix[ {pre___, RepeatedNull|"...", ___}] :=
	introduceFix[ {pre, RepeatedNull}, Length[ {pre}]]
introduceFix[ {pre___, "_"..}] :=
	introduceFix[ {pre}, Length[ {pre}]]
introduceFix[ consts_List] :=
	introduceFix[ consts, Length[ consts]]
introduceFix[ {}, _Integer] := voidTac[]
introduceFix[ consts_List, len_Integer][ ps:PRFSIT$[ {G_, ___}, __], args___, mode_] :=
	With[ {num = If[ len === 0, 0, allCount[ G, 0, len, True]]},
	With[ {newConsts = Replace[ If[ num < len, sowCPMMessage[ 77, "Warning"]; Take[ consts, num], consts], EXPR$[ c_, ___] :> c, {1}]},
		Which[
			newConsts === {},
			Hold[],
		(*else*)
			mode === 1,
			Replace[ executeProofCommands[ Hold[ introduceFix[ consts, {}, Null]], {ps}, 0, False],
				{
					PRFSTAT$[ _, {sit:PRFSIT$[ {G0_, ___}, __]}, assms_List, thms_List, ___] :>
						{
							makeANDNODE[
								makePRFINFO[ name -> introduceFix, used -> {{}}, generated -> {{Nest[ First, G0, assmCount[ G, 0, Infinity]]}}, "abf" -> consts],
								sit
							],
							assms,
							thms,
							Hold[]
						},
					_ -> $Failed
				}
			],
		(*else*)
			True,
			introduceFix[ newConsts, {}, Null][ ps, args, mode]
		]
	]]
introduceFix[ {}, gen_List, n_] :=
	generalizeFix[ gen, If[ IntegerQ[ n], n, 0], False]
introduceFix[ {f:(Automatic|"$"|RepeatedNull|"_"), r___}, gen_List, n_][ PRFSIT$[ {G_, ___}, _List, _Integer, fix_List, __], ___] :=
	Catch[
		With[ {name =
				Replace[ stripAssms[ G],
					{
						TYPED$[ CoreAll$TM, {_}][ Lambda$TM[ RNG$[ TYPRNG$[ v:Except[ Null, _Symbol], _]], _]] :> v,
						TYPED$[ CoreAll$TM, {_}][ _] :> Theorema`Knowledge`VAR$x$TM,
						_ :> Throw[ Hold[]]		(* can only happen if 'f' is 'RepeatedNull' *)
					}
				]},
		With[ {j = Max[ 0, Max @@ Cases[ fix, (FIX$[ name, i_Integer] -> _) :> i] + 1]},
			Hold @@ {
				INF$[ "allI", FIX$[ name, j]],
				introduceFix[
					If[ f === RepeatedNull || f === "...", {f}, {r}],
					If[ f === "_", Append[ gen, FIX$[ name, j]], gen],
					If[ f === "_" && !IntegerQ[ n], assmCount[ G, 0, Infinity], n]
				]
			}
		]]
	]
introduceFix[ {f_FIX$, r___}, gen_List, n_][ ___] :=
	Hold[ INF$[ "allI", f], introduceFix[ {r}, gen, n]]
introduceFix[ {f_, ___}, _List, _][ ___] :=
	(
		sowCPMMessage[ 78, {f}, "Fatal"];
		$Failed
	)

(* 'fastFix[ sub, qc]' is an auxiliary proof method that introduces a.b.f. constants for the first 'qc' remaining
	universal quantifiers in the first 'sub' subgoals of the given proof situation ('sub' and 'qc' could be 'All' as
	well, with the obvious meaning). The names of the constants are all 'Null', they only differ by their indices; this
	is the reason why the proof method is called "fast".
	If 'qc' is an integer, there must be at least as many universal quantifiers left; this is not checked, though! *)
fastFix[ sub_] :=
	fastFix[ sub, All]
fastFix[ sub_, qc_][ PRFSIT$[ G_, _, _, fix_, __], __] :=
	fastFix[ sub, qc, G, fix]
fastFix[ _, _, {}, _List] :=
	Hold[]
fastFix[ _, 0, _List, _List] :=
	Hold[]
fastFix[ sub0_, All, G_List, fix_List] :=
	With[ {sub = If[ sub0 === All, Length[ G], sub0]},
	Module[ {allc = 0},
	With[ {qc =
			Map[
				With[ {aux = allCount[ #, 0, Infinity, True]},
					allc += aux;
					aux
				]&,
				If[ sub0 === All, G, Take[ G, sub]]
			]},
		If[ allc === 0,
			Hold[],
		(*else*)
			fastFix[ qc, sub, fix]
		]
	]]]
fastFix[ sub0_, qc_Integer?Positive, G_List, fix_List] :=
	fastFix[ qc, If[ sub0 === All, Length[ G], sub0], fix]
fastFix[ qc_, sub_Integer, fix_List] :=
	fastFix[ qc, sub, Max[ -1, Max @@ fix[[All, 1, 2]]]]
fastFix[ {qc_Integer}, 1, max_Integer] :=
	fastFix[ qc, 1, max]
fastFix[ qc_List, sub_Integer, max_Integer] :=
	Module[ {k = max},
		Hold @@ (Join @@ Map[ Append[ Table[ INF$[ "allI", FIX$[ Null, ++k]], {#}], INF$[ "rearrangeG", 1, sub]]&, qc])
	]
fastFix[ qc_Integer, 1, max_Integer] :=
	Hold @@ Table[ INF$[ "allI", FIX$[ Null, k]], {k, max + 1, max + qc}]
fastFix[ qc_Integer, sub_Integer, max_Integer] :=
	Hold @@ (Join @@ Table[ Append[ Table[ INF$[ "allI", FIX$[ Null, k]], {k, max + 1 + i * qc, max + (i + 1) * qc}], INF$[ "rearrangeG", 1, sub]], {i, 0, sub - 1}])
