(* ::Package:: *)

(* Theorema
    Copyright (C) 2017 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`HOL`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Language`"]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]

With[ {lang = "English"},
	MessageName[ addParsingRules, "usage", lang] = "";
	MessageName[ addPrintingRules, "usage", lang] = "";
	MessageName[ declareHOLConstants, "usage", lang] = "";
	MessageName[ holMessageToList, "usage", lang] = "";
	MessageName[ HOLProofLanguageTranslator, "usage", lang] = "";
	MessageName[ HOLTheoryLanguageTranslator, "usage", lang] = "";
	MessageName[ sowHOLMessage, "usage", lang] = "";
	MessageName[ translateToCoreEquality, "usage", lang] = "";
];

Begin["`Private`"]

(* DONE:
- Maybe all the rules below have to be stated both in "Theorema`" and in "Theorema`Computation`" context?
	-> Only state them in "Theorema`" context, otherwise it would be a mess. For computations, use 'renameToStandardContext' etc.
*)

(* ::Section:: *)
(*Messages*)

(* In contrast to Core, in HOL we put prover-related messages and theory-related messages together: 'holMessageToList'. *)

Clear[ holMessageToList];
Map[ Get, FileNames[ "*.m", FileNameJoin[ {Theorema`$TheoremaDirectory, "Theorema", "CoreTheories", "HOL", "Packages", "LanguageData"}]]];

sowHOLMessage[ text_String] :=
	sowMessage[ MSG$[ text]]
sowHOLMessage[ text_String, header_String] :=
	sowMessage[ MSG$[ text, header]]
sowHOLMessage[ id_Integer, header_String] :=
	sowHOLMessage[ id, {}, header]
sowHOLMessage[ id_Integer, header_String, src_String] :=
	sowHOLMessage[ id, {}, header, src]
sowHOLMessage[ args__] :=
	sowMessage[ MSG$[ holMessageToList, args]]


(* ::Section:: *)
(*Input*)

HOLFmlPre[ form_, lbl_, tags_List, _, _] :=
	Replace[ translateToCoreEquality[ form, {}, False],
		{
			0 ->
				{
					form,
					lbl,
					tags,
					Automatic,
					If[ #2 =!= True, $Failed, True]&
				},
			i_Integer :> (sowHOLMessage[ i, {}, "Fatal"]; $Failed),
			def_ :>
				{
					def,
					lbl,
					tags,
					"def",
					If[ #2 =!= True, $Failed, True]&
				}
		}
	]

addParsingRules[ kind_, rules0_List, args___] :=
	With[ {rules =
			Replace[ EvalTheoryName[],
				{
					src_String :> addScopeCondition[ src] /@ rules0,
					_ -> rules0
				}
			]},
		Switch[ kind,
			"type",
			addToRuleBlocks[ $typeParsingRules, rules, args],
			"uncurried nominal",
			addToRuleBlocks[ $uncurriedNominalParsingRules, rules, args],
			"curried db",
			addToRuleBlocks[ $curriedDBParsingRules, rules, args],
			_,
			$Failed
		]
	]
addParsingRules[ kind_, rule:(_Rule|_RuleDelayed), args___] :=
	addParsingRules[ kind, {rule}, args]

(* ::Subsection:: *)
(*Term Parsing*)

parseHOLBinder[ Abbrev$TM[ r_RNG$, t_]] :=
	If[ MatchQ[ r, RNG$[ ABBRVRNG$[ _, __]...]],
		With[ {terms = Last /@ (List @@ r),
				newRng = Map[ RNG$[ TYPRNG$[ varToSymbol[ First[ #]], TVAR$[]]]&, List @@ r]},
			Fold[
				Abbrev$TM[ Lambda$TM[ First[ #2], #1]][ Last[ #2]]&,
				Abbrev$TM[ Lambda$TM[ Last[ newRng], t]][ Last[ terms]],
				Reverse[ Transpose[ {Most[ newRng], Most[ terms]}]]
			]
		],
	(*else*)
		sowHOLMessage[ 52, {r}, "Error", "parsing"];
		Throw[ $Failed]
	]
parseHOLBinder[ q_[ r0_RNG$, c_, f0_]] :=
	Replace[ normalizeRange[ r0, DeleteCases[ {c}, True|TYPED$[ True, {}]], f0],
		{r_List, conds_List, f_} :>
			With[ {args = Reverse[ Transpose[ {r, listToConjDisj /@ conds}]]},
				Fold[
					q[ Lambda$TM[ First[ #2], Last[ #2]]][ Lambda$TM[ First[ #2], #1]]&,
					q[ Lambda$TM[ args[[1, 1]], args[[1, 2]]]][ Lambda$TM[ args[[1, 1]], f]],
					Rest[ args]
				]
			]
	]
parseHOLBinder[ q_] :=
	defaultBinderParser[ q]

makeRuleFunction[ rules_List] :=
	makeRuleFunction[ rules, Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$x$TM, TVAR$[]]], Undefined$TM]]
makeRuleFunction[ {}, def_] :=
	def
makeRuleFunction[ {rule_, rest___}, def_] :=
	makeRuleFunction[ rule, makeRuleFunction[ {rest}, def]]
makeRuleFunction[ Theorema`Knowledge`Case$TM[ rng0_RNG$, cond0_, rule0:RuleDelayed$TM[ _, _]], def_] :=
	Replace[ normalizeRange[ rng0, DeleteCases[ {cond0}, True|TYPED$[ True, {}]], rule0],
		{rng1_List, cond1_List, rule1_} :>
			With[ {nr1 = multiRangeToTuple[ rng1, {listToConjDisj[ Flatten[ cond1, 1]], First[ rule1], Last[ rule1]}]},
			With[ {rng = First[ nr1], expr = Last[ nr1]},
				Theorema`Knowledge`RuleFunction$TM[
					Lambda$TM[ rng, expr[[1]]]
				][
					Lambda$TM[ rng, expr[[2]]]
				][
					Lambda$TM[ rng, expr[[3]]]
				][
					def
				]
			]]
	]
makeRuleFunction[ rule_RuleDelayed$TM, def_] :=
	With[ {var = VAR$[ Symbol[ "Theorema`Knowledge`" <> freshName[ "VAR$x$TM", DeleteDuplicates[ Cases[ rule, VAR$[ x_Symbol] :> x, {0, Infinity}, Heads -> True]]]]]},
		makeRuleFunction[ Theorema`Knowledge`Case$TM[ RNG$[ SIMPRNG$[ var]], True, rule], def]
	]
makeRuleFunction[ e_, _] :=
	(sowHOLMessage[ 53, {e}, "Error", "parsing"]; Throw[ $Failed])
	
multiRangeToTuple[ {rng:RNG$[ _TYPRNG$]}, form_] :=
	{rng, form}
multiRangeToTuple[ rng:{RNG$[ TYPRNG$[ var_, _]], __RNG$}, form_] :=
	With[ {len = Length[ rng]},
	With[ {repl = MapIndexed[ (VAR$[ #1[[1, 1]]] -> accessTuple[ First[ #2] - 1, len, VAR$[ var]])&, rng]},
		{RNG$[ TYPRNG$[ varToSymbol[ var], buildProductType[ rng[[All, 1, 2]], len]]], substituteFree[ form, repl, "checkTypes" -> False, "postprocessing" -> Identity]}
	]]

accessTuple[ m_Integer, n_Integer, acc_List] :=
	If[ n <= 1,
		acc,
	(*else*)
		With[ {q = Quotient[ n, 2]},
			If[ m >= q,
				accessTuple[ m - q, n - q, Prepend[ acc, Theorema`Knowledge`Snd$TM]],
			(*else*)
				accessTuple[ m, q, Prepend[ acc, Theorema`Knowledge`Fst$TM]]
			]
		]
	]
accessTuple[ m_Integer, n_Integer, t:Except[ _List]] :=
	(Composition @@ accessTuple[ m, n, {}])[ t]
	
buildProductType[ types_List, n_Integer] :=
	If[ n === 1,
		First[ types],
	(*else*)
		With[ {q = Quotient[ n, 2]},
			Theorema`Knowledge`PROD$TM[ buildProductType[ Take[ types, q], q], buildProductType[ Drop[ types, q], n - q]]
		]
	]

listToConjDisj[ l_List] :=
	listToConjDisj[ l, And$TM]
listToConjDisj[ {}, Or$TM] :=
	False
listToConjDisj[ {}, And$TM] :=
	True
listToConjDisj[ {c_}, _] :=
	c
listToConjDisj[ l_List, h_] :=
	binaryApply[ h, l]

(* 'translateToCoreEquality[ term, {}, acceptCoreEq]' returns either a Core equality or an integer with the following
	meaning:
	- '0': 'term' is no definition,
	- '1': 'term' is an ill-formed definition,
	- '2': arguments on LHS of 'term' are not distinct,
	- '3': argument on LHS is no variable,
	- '4': RHS contains variables not appearing on LHS.
	'acceptCoreEq' specifies whether Core equalities shall be accepted as definitions. *)
translateToCoreEquality[ TYPED$[ CoreEqual$TM, typ_][ lhs_][ rhs_], loose_, acceptCoreEq_] :=
	If[ TrueQ[ acceptCoreEq],
		translateToCoreEquality[ TYPED$[ EqualDef$TM, typ][ lhs][ rhs], loose, acceptCoreEq],
	(*else*)
		0
	]
translateToCoreEquality[ TYPED$[ EqualDef$TM, {typ_}][ lhs0_][ rhs_], loose_, _] :=
	Replace[ splitHeadArgs[ lhs0],
		{
			{_, {___, x_VAR$, ___, x_, ___}} -> 2,
			{const_, args:{___VAR$}} :>
				With[ {argIndices = args[[All, 1]], num = Length[ args]},
				With[ {repl =
						Append[
							MapIndexed[ (#1 -> num - First[ #2])&, argIndices],
							_ :> Throw[ $Failed]
						],
						revloose = Reverse[ loose[[argIndices + 1]]]},
				With[ {term =
						Catch[
							Fold[
								Lambda$TM[ RNG$[ #2], #1]&,
								If[ loose === {},
									rhs,
								(*else*)
									mapFreeVars[ rhs, VAR$[ #2 + Replace[ #3 - #2, repl]]&]
								],
								revloose
							]
						]},
					If[ term === $Failed,
						4,
					(*else*)
						TYPED$[ CoreEqual$TM, {makeFunctionType[ revloose[[All, 2]], typ]}][ const][ term]
					]
				]]],
			{_, _List} -> 3,
			_ -> 1
		}
	]
translateToCoreEquality[ TYPED$[ TrueQ$TM, {}][ term_], loose_, acceptCoreEq_] :=
	translateToCoreEquality[ term, loose, acceptCoreEq]
translateToCoreEquality[ TYPED$[ Forall$TM, _][ Lambda$TM[ _RNG$, True|TYPED$[ True, _]]][ Lambda$TM[ RNG$[ rng_TYPRNG$], term_]], loose_, acceptCoreEq_] :=
	translateToCoreEquality[ term, Prepend[ loose, rng], acceptCoreEq]
translateToCoreEquality[ TYPED$[ Forall$TM, {typ_}][ Lambda$TM[ _RNG$, True|TYPED$[ True, _]]][ term_], loose_, acceptCoreEq_] :=
	translateToCoreEquality[ mapFreeVars[ term, VAR$[ #3 + 1]&][ VAR$[ 0]], Prepend[ loose, TYPRNG$[ Null, typ]], acceptCoreEq]
translateToCoreEquality[ TYPED$[ CoreAll$TM, _][ Lambda$TM[ RNG$[ rng_TYPRNG$], term_]], loose_, acceptCoreEq_] :=
	translateToCoreEquality[ term, Prepend[ loose, rng], acceptCoreEq]
translateToCoreEquality[ TYPED$[ CoreAll$TM, {typ_}][ term_], loose_, acceptCoreEq_] :=
	translateToCoreEquality[ mapFreeVars[ term, VAR$[ #3 + 1]&][ VAR$[ 0]], Prepend[ loose, TYPRNG$[ Null, typ]], acceptCoreEq]
translateToCoreEquality[ _, _, _] :=
	0

	
(* ::Section:: *)
(*Output*)

addPrintingRules[ kind_, rules0_List, args___] :=
	With[ {rules =
			Replace[ EvalTheoryName[],
				{
					src_String :> addScopeCondition[ src] /@ rules0,
					_ -> rules0
				}
			]},
		Switch[ kind,
			"type",
			addToRuleBlocks[ $typePrintingRules, rules, args],
			"curried db",
			With[ {res = addToRuleBlocks[ $curriedDBPrintingRules, rules, args]},
				If[ ListQ[ res],
					(* Probably there's a more efficient way than re-computing the whole list ... *)
					$curriedDBInputPrintingRules =
						translatePrettyPrintingRulesToInputPrintingRules[ res //. (h:(printHOLBinder|printCoreBinder))[ q_, True] :> h[ q, False]]
				];
				res
			],
			"curried nominal",
			With[ {res = addToRuleBlocks[ $curriedNominalPrintingRules, rules, args]},
				If[ ListQ[ res],
					$curriedNominalInputPrintingRules = translatePrettyPrintingRulesToInputPrintingRules[ res]
				];
				res
			],
			"uncurried nominal",
			With[ {res = addToRuleBlocks[ $uncurriedNominalPrintingRules, rules, args]},
				If[ ListQ[ res],
					$uncurriedNominalInputPrintingRules = translatePrettyPrintingRulesToInputPrintingRules[ res]
				];
				res
			],
			_,
			$Failed
		]
	]
addPrintingRules[ kind_, rule:(_Rule|_RuleDelayed), args___] :=
	addPrintingRules[ kind, {rule}, args]

(* Abbrev, pretty *)
printHOLBinder[ (h:TYPED$[ Abbrev$TM, _])[ Lambda$TM[ RNG$[ TYPRNG$[ x_Symbol, typ_]], t_]][ a_], True] :=
	h[ RNG$[ ABBRVRNG$[ TYPED$[ VAR$[ x], typ], a]], mapVarsSlow[ t, (#2 === #3)&, TYPED$[ VAR$[ #3], typ]&]]
printHOLBinder[ (h:TYPED$[ Abbrev$TM, {ta_, _}])[ f_][ a_], True] :=
	h[ RNG$[ ABBRVRNG$[ TYPED$[ Null, ta], a]], mapFreeVarsSlow[ f, VAR$[ #3 + 1]&][ TYPED$[ VAR$[ 0], ta]]]

(* Abbrev, input *)
printHOLBinder[ (TYPED$[ Abbrev$TM, _]|Abbrev$TM)[ Lambda$TM[ RNG$[ TYPRNG$[ x_Symbol, typ_]], t_]][ a_], False] :=
	Abbrev$TM[ RNG$[ ABBRVRNG$[ VAR$[ x], If[ typ === TVAR$[], a, TYPED$[ a, typ]]]], t]
printHOLBinder[ TYPED$[ Abbrev$TM, {ta_, _}][ f_][ a_], False] :=
	Abbrev$TM[ RNG$[ ABBRVRNG$[ Null, TYPED$[ a, ta]]], mapFreeVarsSlow[ f, VAR$[ #3 + 1]&][ VAR$[ 0]]]
printHOLBinder[ (h:(TYPED$[ Abbrev$TM, _]|Abbrev$TM))[ f_][ a_], False] :=
	h[ RNG$[ ABBRVRNG$[ Null, a]], mapFreeVarsSlow[ f, VAR$[ #3 + 1]&][ VAR$[ 0]]]

(* Case, pretty *)
printHOLBinder[ TYPED$[ Theorema`Knowledge`Case$TM, typ:{_, ttyp_, otyp_}][ c_][ Lambda$TM[ RNG$[ TYPRNG$[ y_, _]], p_]][ Lambda$TM[ RNG$[ TYPRNG$[ x_, t_]], e_]], True] :=
	With[ {rd = TYPED$[ RuleDelayed$TM, {ttyp, otyp}]},
		If[ MatchQ[ c, Lambda$TM[ _, True|TYPED$[ True, {}]]] && !containsVar[ {p, e}, (#2 === #3)&],
			rd[ p, e],
		(*else*)
			printCoreBinder[ TYPED$[ Theorema`Knowledge`Case$TM, typ][ c][ Lambda$TM[ RNG$[ TYPRNG$[ chooseRangeVar[ {x, y}], t]], rd[ p, e]]], True]
		]
	]
printHOLBinder[ TYPED$[ Theorema`Knowledge`Case$TM, typ:{_, ttyp_, otyp_}][ c_][ p_][ Lambda$TM[ r_RNG$, e_]], True] :=
	With[ {rd = TYPED$[ RuleDelayed$TM, {ttyp, otyp}]},
		printCoreBinder[ TYPED$[ Theorema`Knowledge`Case$TM, typ][ c][ Lambda$TM[ r, rd[ p, e]]], True]
	]
printHOLBinder[ TYPED$[ Theorema`Knowledge`Case$TM, typ:{_, ttyp_, otyp_}][ c_][ Lambda$TM[ r_RNG$, p_]][ e_], True] :=
	With[ {rd = TYPED$[ RuleDelayed$TM, {ttyp, otyp}]},
		printCoreBinder[ TYPED$[ Theorema`Knowledge`Case$TM, typ][ c][ Lambda$TM[ r, rd[ p, e]]], True]
	]
printHOLBinder[ TYPED$[ Theorema`Knowledge`Case$TM, typ:{vtyp_, ttyp_, otyp_}][ c_][ p_][ e_], True] :=
	With[ {rd = TYPED$[ RuleDelayed$TM, {ttyp, otyp}]},
		printCoreBinder[ TYPED$[ Theorema`Knowledge`Case$TM, typ][ c][ Lambda$TM[ RNG$[ TYPRNG$[ Null, vtyp]], rd[ p[ VAR$[ 0]], e[ VAR$[ 0]]]]], True]
	]

(* Case, input *)
printHOLBinder[ (h:(TYPED$[ Theorema`Knowledge`Case$TM, _]|Theorema`Knowledge`Case$TM))[ c_][ Lambda$TM[ RNG$[ TYPRNG$[ y_, _]], p_]][ Lambda$TM[ RNG$[ TYPRNG$[ x_, t_]], e_]], False] :=
	If[ MatchQ[ c, Lambda$TM[ _, True|TYPED$[ True, {}]]] && !containsVar[ {p, e}, (#2 === #3)&],
		RuleDelayed$TM[ p, e],
	(*else*)
		printCoreBinder[ h[ c][ Lambda$TM[ RNG$[ TYPRNG$[ chooseRangeVar[ {x, y}], t]], RuleDelayed$TM[ p, e]]], False]
	]
printHOLBinder[ (h:(TYPED$[ Theorema`Knowledge`Case$TM, _]|Theorema`Knowledge`Case$TM))[ c_][ p_][ Lambda$TM[ r_RNG$, e_]], False] :=
	printCoreBinder[ h[ c][ Lambda$TM[ r, RuleDelayed$TM[ p, e]]], False]
printHOLBinder[ (h:(TYPED$[ Theorema`Knowledge`Case$TM, _]|Theorema`Knowledge`Case$TM))[ c_][ Lambda$TM[ r_RNG$, p_]][ e_], False] :=
	printCoreBinder[ h[ c][ Lambda$TM[ r, RuleDelayed$TM[ p, e]]], False]
printHOLBinder[ (h:(TYPED$[ Theorema`Knowledge`Case$TM, _]|Theorema`Knowledge`Case$TM))[ c_][ p_][ e_], False] :=
	printCoreBinder[ h[ c][ Lambda$TM[ RNG$[ TYPRNG$[ Null, TVAR$[]]], RuleDelayed$TM[ p[ VAR$[ 0]], e[ VAR$[ 0]]]]], False]

reconstructHOLRange[
			(TYPED$[ Element$TM, _]|Element$TM)[ VAR$[ 0]][ (TYPED$[ IntegerInterval$TM, _]|IntegerInterval$TM)[ a0_][ b0_][ TYPED$[ True, _]|True][ TYPED$[ True, _]|True]],
			tr_
		] /; !containsVar[ {a0, b0}, (#2 === #3)&] :=
	With[ {ab = mapFreeVarsSlow[ {a0, b0}, VAR$[ #3 - 1]&]},
		{
			If[ TrueQ[ tr],
				Theorema`Knowledge`TSTEPRNG$[ Null, Null, Sequence @@ ab, 1],
			(*else*)
				STEPRNG$[ Null, Sequence @@ ab, 1]
			],
			True
		}
	]
reconstructHOLRange[ (TYPED$[ Element$TM, _]|Element$TM)[ VAR$[ 0]][ A0_], tr_] /; !containsVar[ A0, (#2 === #3)&] :=
	With[ {A = mapFreeVarsSlow[ A0, VAR$[ #3 - 1]&]},
		{
			If[ TrueQ[ tr],
				Theorema`Knowledge`TSETRNG$[ Null, Null, A],
			(*else*)
				SETRNG$[ Null, A]
			],
			True
		}
	]
reconstructHOLRange[
			(TYPED$[ Exists$TM, _]|Exists$TM)[
				Lambda$TM[ _RNG$,
					(TYPED$[ Element$TM, _]|Element$TM)[ VAR$[ 0]][ (TYPED$[ IntegerInterval$TM, _]|IntegerInterval$TM)[ TYPED$[ 0, _]|0][ TYPED$[ Infinity, _]|Infinity][ TYPED$[ True, _]|True][ TYPED$[ False, _]|False]]
				]
			][
				Lambda$TM[ _RNG$,
					((TYPED$[ And$TM, _]|And$TM)[ (TYPED$[ Equal$TM, _]|Equal$TM)[ VAR$[ 1]][ (TYPED$[ Plus$TM, _]|Plus$TM)[ a_][ step_]]][ rel_[ VAR$[ 1]][ b_]])|
					((TYPED$[ And$TM, _]|And$TM)[ rel_[ VAR$[ 1]][ b_], (TYPED$[ Equal$TM, _]|Equal$TM)[ VAR$[ 1]][ (TYPED$[ Plus$TM, _]|Plus$TM)[ a_][ step_]]])
				]
			],
			tr_
		] :=
	With[ {s =
			Switch[ step,
				VAR$[ 0],
				1,
				(TYPED$[ Times$TM, _]|Times$TM)[ VAR$[ 0]][ _],
				First[ step],
				(TYPED$[ Times$TM, _]|Times$TM)[ _][ VAR$[ 0]],
				step[[0, 1]],
				_,
				$Failed
			]},
		With[ {abs = mapFreeVarsSlow[ {a, b, s}, VAR$[ #3 - 2]&]},
			{
				If[ TrueQ[ tr],
					Theorema`Knowledge`TSTEPRNG$[ Null, Null, Sequence @@ abs],
				(*else*)
					STEPRNG$[ Null, Sequence @@ abs]
				],
				True
			}
		] /;
			s =!= $Failed &&
			!containsVar[ {a, b, step}, (#2 + 1 === #3)&] &&
			With[ {relPatt = If[ TrueQ[ Negative[ s]], GreaterEqual$TM, LessEqual$TM]},
				MatchQ[ rel, TYPED$[ relPatt, _]|relPatt]
			] && !containsVar[ {a, b, s}, (#2 === #3)&]
	]
reconstructHOLRange[ (h:(TYPED$[ And$TM, t_]|And$TM))[ a_][ b_], tr_] :=
	With[ {rng = reconstructHOLRange[ a, tr]},
	With[ {newC = Last[ rng]},
		{First[ rng], If[ newC === True, b, h[ newC][ b]]}
	]]
reconstructHOLRange[ TYPED$[ True, _], tr_] :=
	defaultReconstructRange[ True, tr]
reconstructHOLRange[ c_, tr_] :=
	defaultReconstructRange[ c, tr]

tupleToMultiRange[ expr_] := expr	(* TODO *)


(* ::Section:: *)
(*Syntax*)

(* postfix function application *)
MakeBoxes[ TAG$[ Postfix$TM, tag_][ f_[ a_]], System`TheoremaForm] :=
    RowBox[ {MakeBoxes[ a, System`TheoremaForm], Theorema`Language`Syntax`Private`tmaTagBox[ Infix, "//", tag, TAG$], MakeBoxes[ f, System`TheoremaForm]}]
MakeBoxes[ Theorema`Computation`Language`TAG$[ Theorema`Computation`Language`Postfix$TM, tag_][ f_[ a_]], System`TheoremaForm] :=
    RowBox[ {MakeBoxes[ a, System`TheoremaForm], Theorema`Language`Syntax`Private`tmaTagBox[ Infix, "//", tag, Theorema`Computation`Language`TAG$], MakeBoxes[ f, System`TheoremaForm]}]

(* Of course, (typed) set ranges only make sense after introducing sets in theory "Sets.nb", but nevertheless we
	register the corresponding ranges already here. After all, this is only to make standard-Theorema parse typed
	set ranges correctly. *)
registerRange[
	"TSETRNG$",
	{
		RowBox[ {RowBox[ {RowBox[ {"(", "$var", ")"}], "::", t_}], "\[Element]", A_}] :> {t, A},
		RowBox[ {RowBox[ {"$var", "::", t_}], "\[Element]", A_}] :> {t, A},
		RowBox[ {RowBox[ {"(", RowBox[ {RowBox[ {"(", "$var", ")"}], "::", t_}], ")"}], "\[Element]", A_}] :> {t, A},
		RowBox[ {RowBox[ {"(", RowBox[ {"$var", "::", t_}], ")"}], "\[Element]", A_}] :> {t, A}
	},
	{x_, t_, A_} :> RowBox[ {RowBox[ {MakeBoxes[ x], "::", makeTypeBoxes[ t]}], "\[Element]", MakeBoxes[ A]}],
	{x_, _, A_} :> {Element$TM[ x, A]}
];


(* ::Section:: *)
(*Theory Language*)

HOLTheoryLanguageTranslator[ "decl", Hold[], Hold[ consts___, notation_List]] :=
	With[ {spec = If[ notation === {}, {}, getNotationSpec[ notation]]},
		If[ ListQ[ spec],
			Hold[ declareHOLConstants[ {consts}, spec]] /. EXPR$[ e_] :> parseCoreTermPre[ e],
		(*else*)
			$Failed
		]
	]
HOLTheoryLanguageTranslator[ "decl", Hold[], Hold[ consts___]] :=
	HOLTheoryLanguageTranslator[ "decl", Hold[], Hold[ consts, {}]]
HOLTheoryLanguageTranslator[ args__] := defaultTLT[ args]


(* ::Subsection:: *)
(*Constant Declarations*)

(* 'declareHOLConstants' declares a constant or a list thereof to be of a certain type, using 'declareConstantsDirect',
	and additionally adds notation. *)
declareHOLConstants[ consts_List, notation_List][ False, i_, id_, src_, args___] :=
	termNotation[
		3,
		Replace[
			First[
				declareConstantsDirect[
					Replace[ consts, e_EXPR$ :> First[ generalizeTypes[ makeSTERM[ e], {}]], {1}],
					cellIDLabel[ {id, i, #}]&,
					sourceLabel[ src],
					Interactive,
					{src, id}
				]
			],
			{STERM$[ TYPED$[ e_, _], ___] :> e, (STERM$|TYPED$)[ e_, ___] :> e},
			{1}
		],
		notation
	][ False, i, id, src, args]
declareHOLConstants[ consts_List, notation_List][ True, i_, id_, src_, args___] :=
	termNotation[
		3,
		Replace[
			First[ undeclareConstantsDirect[ Replace[ consts, EXPR$[ e_, ___] :> First[ generalizeTypes[ e, {}]], {1}], {src, id}]],
			{EXPR$[ TYPED$[ e_, _], ___] :> e, (EXPR$|TYPED$)[ e_, ___] :> e},
			{1}
		],
		notation
	][ True, i, id, src, args]


(* ::Subsection:: *)
(*Type Definitions*)

(* HOL type definitions depend upon the following theorems:
	"\[Exists]E", "\[DoubleLeftRightArrow]I", "\[Equal] reflect". *)

HOLTypedefFormulaConstructor[ rt_, STERM$[ pred_, tfix_, tvar_], types_] :=
	STERM$[ HOLTypedefFormulaConstructor[ rt, pred, types], tfix, tvar]
HOLTypedefFormulaConstructor[ rt_, pred_, True] :=
	TYPED$[ TrueQ$TM, {}][
		TYPED$[ Exists$TM, {rt}][
			Lambda$TM[ RNG$[ TYPRNG$[ Replace[ pred, {Lambda$TM[ RNG$[ _[ x_, _]], _] :> x, _ -> Null}], rt]], TYPED$[ True, {}]]
		][
			pred
		]
	]
HOLTypedefFormulaConstructor[ rt_, pred_, False] :=
	Exists$TM[
		Lambda$TM[ RNG$[ TYPRNG$[ Null, rt]], True]
	][
		pred
	]

HOLTypedefProofTransformer[ prf_] :=
	With[ {cmd =
			Block[ {$ignoreScope = True},
				makeFactsExplicitInline[
					{STRING$[ Lookup[ $theoryMapping, "$HOL", $Failed], "\[Exists]E"]},
					1,
					Join[
						Hold[
							Theorema`CoreProvers`instantiateCoreAll[ {Automatic, FIX$[ Theorema`Knowledge`VAR$thesis$TM, 0]}],
							Theorema`CoreProvers`backChainSingle[ 1, 0, 1, 1]
						],
						prf
					],
					False
				]
			]},
		If[ MatchQ[ cmd, _Hold],
			Prepend[ cmd, INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$thesis$TM, 0]]],
		(*else*)
			sowHOLMessage[ 11, {STRING$[ "\[Exists]E"]}, "Fatal"];
			Hold[ Theorema`CoreProvers`failTac[]]
		]
	]

addHOLTypedefThms[ rep0_, abs0_, rep_, abs_, pred_, typ_, rt_, tvar_, predPrf_, repName_String, absName_String, id_String, src_String, pos_List,
		repKey_List, repInvKey_List, absInvKey_List] :=
	With[ {trivPred = If[ MatchQ[ predPrf, _Hold], First, Identity],
			holThy = Lookup[ $theoryMapping, "$HOL", $Failed]},
	With[ {predVar0 = If[ trivPred === First, Null, TYPED$[ TrueQ$TM, {}][ applyFunction[ pred, VAR$[ 0], False]]],
			posData = If[ ListQ[ pos], "pos" -> pos, Sequence @@ {}],
			iffIKey = fastKeyFromLabel[ {holThy, "\[DoubleLeftRightArrow]I"}]},
		Which[
			iffIKey === $Failed,
			sowHOLMessage[ 11, {STRING$[ "\[DoubleLeftRightArrow]I"]}, "Error"];
			Table[ False, {6}],	(* Only to inform the calling function that something went wrong. *)
		(*else*)
			fastKeyFromLabel[ {holThy, "\[Equal] reflect"}] === $Failed,
			sowHOLMessage[ 11, {STRING$[ "\[Equal] reflect"]}, "Error"];
			Table[ False, {6}],
		(*else*)
			True,
			Join[
				addCoreTypedefInductThms[ rep0, abs0, rep, abs, pred, typ, rt, tvar, predPrf, repName, absName, id, src, pos, repKey, repInvKey, absInvKey],
				{
					(* "rep cases": '/\P::FORM y::rt. (TrueQ[ pred[ y]] ==>) (/\x::typ. TrueQ[ y == rep[ x]] ==> P) ==> P' *)
					addThmToTheory[
						FML$[ {id <> ".6", src},
							TYPED$[ CoreAll$TM, {FORM$TM}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$P$TM, FORM$TM]],
								TYPED$[ CoreAll$TM, {rt}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$y$TM, rt]],
									trivPred[
										TYPED$[ CoreImplies$TM, {}][
											predVar0
										][
											TYPED$[ CoreImplies$TM, {}][
												TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$x$TM, typ]],
													TYPED$[ CoreImplies$TM, {}][
														TYPED$[ TrueQ$TM, {}][ TYPED$[ Equal$TM, {rt}][ VAR$[1]][ rep[ VAR$[ 0]]]]
													][
														VAR$[ 2]
													]]
												]
											][
												VAR$[ 1]
											]
										]
									]
								]]
							]],
							repName <> " cases",
							posData
						],
						tvar,
						Join[
							Hold @@ {
								INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$P$TM, 0]],
								INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$y$TM, 0]],
								Theorema`CoreProvers`instantiateCoreAll[
									{EXPR$[ abs0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]},
									If[ trivPred === First, 1, 2]
								],
								INF$[ "impE", If[ trivPred === First, 1, 2], 0]
							},
							If[ trivPred === First,
								Hold[],
							(*else*)
								Hold[
									INF$[ "makeImplicit", 1],
									INF$[ "drop", 1]
								]
							],
							Hold[
								Theorema`CoreProvers`setupSubgoal[
									EXPR$[ CoreEqual$TM[ rep0[ abs0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]][ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]
								]
							],
							If[ trivPred === First,
								Join[
									Hold[
										INF$[ "formula", absInvKey],
										INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]],
										INF$[ "impE", 1, 0]
									],
									predPrf
								],
							(*else*)
								Hold[
									INF$[ "makeExplicit", 1],
									INF$[ "formula", absInvKey],
									INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]],
									INF$[ "impE", 1, 1]
								]
							],
							Hold[
								INF$[ "close"],
								INF$[ "makeExplicit", 1],
								(* 'rep[ abs[ y]] === y ==> TrueQ[ y == rep[ abs[ y]]]' *)
								INF$[ "replace", 1, {1, 1, 1}],
								(* 'rep[ abs[ y]] === y ==> TrueQ[ y == y]' *)
								Theorema`CoreProvers`reflectHOLConstant[ Equal$TM, {1}],
								INF$[ "refl"],
								INF$[ "close"]
							]
						],
						False,
						False,
						False,
						True,
						False
					],
					
					(* "abs cases": '/\P::FORM x::typ. (/\y::rt. TrueQ[ x == abs[ y]] ==> (TrueQ[ pred[ y]] ==>) P) ==> P' *)
					addThmToTheory[
						FML$[ {id <> ".7", src},
							TYPED$[ CoreAll$TM, {FORM$TM}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$P$TM, FORM$TM]],
								TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$x$TM, typ]],
									TYPED$[ CoreImplies$TM, {}][
										TYPED$[ CoreAll$TM, {rt}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$y$TM, rt]],
											TYPED$[ CoreImplies$TM, {}][
												TYPED$[ TrueQ$TM, {}][ TYPED$[ Equal$TM, {typ}][ VAR$[ 1]][ abs[ VAR$[ 0]]]]
											][
												trivPred[
													TYPED$[ CoreImplies$TM, {}][
														predVar0
													][
														VAR$[ 2]
													]
												]
											]
										]]
									][
										VAR$[ 1]
									]
								]]
							]],
							absName <> " cases",
							posData
						],
						tvar,
						Join[
							Hold[
								INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$P$TM, 0]],
								INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$x$TM, 0]],
								Theorema`CoreProvers`instantiateCoreAll[
									{EXPR$[ rep0[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]}
								],
								INF$[ "impE", 1, 0],
								INF$[ "formula", repInvKey],
								INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]],
								(* 'abs[ rep[ x]] === x ==> TrueQ[ x == abs[ rep[ x]]]' *)
								INF$[ "replace", 1, {1, 1, 1}],
								(* 'abs[ rep[ x]] === x ==> TrueQ[ x == x]' *)
								Theorema`CoreProvers`reflectHOLConstant[ Equal$TM, {1}],
								INF$[ "refl"]
							],
							If[ trivPred === First,
								Hold[],
							(*else*)
								Hold[
									INF$[ "formula", repKey],
									INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]],
									INF$[ "impE", 1, 0]
								]
							],
							Hold[
								INF$[ "close"]
							]
						],
						False,
						False,
						False,
						True,
						False
					],
					
					(* "rep inject": '/\x y::typ. TrueQ[ rep[ x] == rep[ y] <=> x == y]' *)
					addThmToTheory[
						FML$[ {id <> ".8", src},
							TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$x$TM, typ]],
								TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$y$TM, typ]],
									TYPED$[ TrueQ$TM, {}][
										TYPED$[ Equal$TM, {Theorema`Knowledge`BOOL$TM}][
											TYPED$[ Equal$TM, {rt}][ rep[ VAR$[ 1]]][ rep[ VAR$[ 0]]]
										][
											TYPED$[ Equal$TM, {typ}][ VAR$[ 1]][ VAR$[ 0]]
										]
									]
								]]
							]],
							repName <> " inject",
							posData
						],
						tvar,
						Hold[
							INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$x$TM, 0]],
							INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$y$TM, 0]],
							INF$[ "formula", iffIKey],
							Theorema`CoreProvers`instantiateCoreAll[
								{
									EXPR$[ Equal$TM[ rep0[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]][ rep0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]],
									EXPR$[ Equal$TM[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]][ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]
								}
							],
							INF$[ "impE", 1, 0],
							INF$[ "split"],
							
							(* "==>": 'TrueQ[ rep[ x] == rep[ y]] ==> TrueQ[ x == y]' *)
							Theorema`CoreProvers`reflectHOLConstant[ Equal$TM, {0, 1}],
							Theorema`CoreProvers`reflectHOLConstant[ Equal$TM, {1}],
							(* from here the proof is the same is in Core *)
							INF$[ "makeImplicit", 1],
							INF$[ "drop", 1],
							Theorema`CoreProvers`setupSubgoal[
								EXPR$[ CoreEqual$TM[ abs0[ rep0[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]][ abs0[ rep0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]]]
							],
							INF$[ "makeExplicit", 1],
							(* 'rep[ x] === rep[ y] ==> abs[ rep[ x]] === abs[ rep[ y]]' *)
							INF$[ "replace", 1, {1, 0, 1, 1}],
							INF$[ "refl"],
							INF$[ "close"],
							INF$[ "makeExplicit", 1],
							INF$[ "insert", repInvKey], (* formula is used twice, so we do not apply "formula" *)
							INF$[ "makeExplicit", 1],
							INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]],
							(* 'abs[ rep[ y]] === y ==> abs[ rep[ x]] === abs[ rep[ y]] ==> x === y' *)
							INF$[ "replace", 1, {1, 0, 1, 1}],
							INF$[ "drop", 1],
							INF$[ "makeExplicit", 1],
							INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]],
							(* 'abs[ rep[ x]] === x ==> abs[ rep[ x]] === y ==> x === y' *)
							INF$[ "replace", 1, {1, 0, 1, 0, 1}],
							INF$[ "impE", 2, 0],
							INF$[ "close"],
							
							(* "<==": 'TrueQ[ x == y] ==> TrueQ[ rep[ x] == rep[ y]]' *)
							Theorema`CoreProvers`reflectHOLConstant[ Equal$TM, {0, 1}],
							Theorema`CoreProvers`reflectHOLConstant[ Equal$TM, {1}],
							(* 'x === y ==> rep[ x] === rep[ y]' *)
							INF$[ "replace", 1, {1, 0, 1, 1}],
							INF$[ "refl"],
							INF$[ "close"]
						],
						False,
						False,
						False,
						True,
						False
					],
					
					(* "abs inject": '/\x y::rt. (TrueQ[ pred[ x]] ==> TrueQ[ pred[ y]] ==>) abs[ x] == abs[ y] <=> x == y' *)
					addThmToTheory[
						FML$[ {id <> ".9", src},
							TYPED$[ CoreAll$TM, {rt}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$x$TM, rt]],
								TYPED$[ CoreAll$TM, {rt}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$y$TM, rt]],
									trivPred[
										trivPred[
											TYPED$[ CoreImplies$TM, {}][
												If[ trivPred === First, Null, TYPED$[ TrueQ$TM, {}][ applyFunction[ pred, VAR$[ 1], False]]]
											][
												TYPED$[ CoreImplies$TM, {}][
													predVar0
												][
													TYPED$[ TrueQ$TM, {}][
														TYPED$[ Equal$TM, {Theorema`Knowledge`BOOL$TM}][
															TYPED$[ Equal$TM, {typ}][ abs[ VAR$[ 1]]][ abs[ VAR$[ 0]]]
														][
															TYPED$[ Equal$TM, {rt}][ VAR$[ 1]][ VAR$[ 0]]
														]
													]
												]
											]
										]
									]
								]]
							]],
							absName <> " inject",
							posData
						],
						tvar,
						Join[
							Hold[
								INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$x$TM, 0]],
								INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]
							],
							If[ trivPred === First,
								Hold[],
							(*else*)
								Hold[
									INF$[ "makeImplicit", 1],
									INF$[ "drop", 1],
									INF$[ "makeImplicit", 1],
									INF$[ "drop", 1]
								]
							],
							Hold[
								INF$[ "formula", iffIKey],
								Theorema`CoreProvers`instantiateCoreAll[
									{
										EXPR$[ Equal$TM[ abs0[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]][ abs0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]],
										EXPR$[ Equal$TM[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]][ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]
									}
								],
								INF$[ "impE", 1, 0],
								INF$[ "split"],
								
								(* "==>": 'TrueQ[ abs[ x] == abs[ y]] ==> TrueQ[ x == y]' *)
								Theorema`CoreProvers`reflectHOLConstant[ Equal$TM, {0, 1}],
								Theorema`CoreProvers`reflectHOLConstant[ Equal$TM, {1}],
								INF$[ "makeImplicit", 1],
								INF$[ "drop", 1],
								Theorema`CoreProvers`setupSubgoal[
									EXPR$[ CoreEqual$TM[ rep0[ abs0[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]][ rep0[ abs0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]]]
								],
								INF$[ "makeExplicit", 1],
								(* 'abs[ x] === abs[ y] ==> rep[ abs[ x]] === rep[ abs[ y]]' *)
								INF$[ "replace", 1, {1, 0, 1, 1}],
								INF$[ "refl"],
								INF$[ "close"],
								INF$[ "makeExplicit", 1],
								INF$[ "insert", absInvKey], (* formula is used twice, so we do not apply "formula" *)
								Theorema`CoreProvers`setupSubgoal[
									EXPR$[ CoreEqual$TM[ rep0[ abs0[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]][ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]
								]
							],
							If[ trivPred === First,
								Hold[],
							(*else*)
								Hold[
									INF$[ "makeExplicit", 5]
								]
							],
							Hold[
								INF$[ "makeExplicit", 1],
								INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]
							],
							If[ trivPred === First,
								Join[
									Hold[
										INF$[ "impE", 1, 0]
									],
									predPrf
								],
							(*else*)
								Hold[
									INF$[ "impE", 1, 1]
								]
							],
							Hold[
								INF$[ "close"],
								INF$[ "makeExplicit", 1],
								(* 'rep[ abs[ x]] === x ==> rep[ abs[ x]] === rep[ abs[ y]] ==> x === y' *)
								INF$[ "replace", 1, {1, 0, 1, 0, 1}],
								INF$[ "drop", 1],
								Theorema`CoreProvers`setupSubgoal[
									EXPR$[ CoreEqual$TM[ rep0[ abs0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]][ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]
								]
							],
							If[ trivPred === First,
								Hold[],
							(*else*)
								Hold[
									INF$[ "makeExplicit", 5]
								]
							],
							Hold[
								INF$[ "makeExplicit", 2],
								INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]
							],
							If[ trivPred === First,
								Join[
									Hold[
										INF$[ "impE", 1, 0]
									],
									predPrf
								],
							(*else*)
								Hold[
									INF$[ "impE", 1, 1]
								]
							],
							Hold[
								INF$[ "close"],
								INF$[ "makeExplicit", 1],
								(* 'rep[ abs[ y]] === y ==> x === rep[ abs[ y]] ==> x === y' *)
								INF$[ "replace", 1, {1, 0, 1, 1}],
								INF$[ "impE", 2, 0],
								INF$[ "close"],
								
								(* "<==": 'TrueQ[ x == y] ==> TrueQ[ abs[ x] == abs[ y]]' *)
								Theorema`CoreProvers`reflectHOLConstant[ Equal$TM, {0, 1}],
								Theorema`CoreProvers`reflectHOLConstant[ Equal$TM, {1}],
								(* 'x === y ==> abs[ x] === abs[ y]' *)
								INF$[ "replace", 1, {1, 0, 1, 1}],
								INF$[ "refl"],
								INF$[ "close"]
							]
						],
						False,
						False,
						False,
						True,
						False
					]
				}
			]
		]
	]]
(* In Isabelle/HOL, the following theorems are additionally generated for type definitions:
"Rep_range": 'range rep = {x. pred x}'
"Abs_image": '{abs x | x. pred x} = UNIV' *)


(* ::Section:: *)
(*Proof Language*)

Theorema`CoreProvers`loadProvers[ {"CoreProofMethods", "HOLProofMethods"}];

HOLProofLanguageTranslator[ "define", Hold[], def_Hold] :=
	Replace[ makeFmlList[ True] @@ def,
		{
			{d_} :> {Theorema`CoreProvers`defineHOLConst[ d]},
			_ -> $Failed
		}
	]
HOLProofLanguageTranslator[ args__] := defaultPLT[ args]

End[]

EndPackage[]
